#include "FileManagement.h"

#include "RendererGLUT.h"
#include "RendererOSMesa.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "VKString.h"

#include "ShapeLocalPCA.h"
#include "ShapeTriangleSoup.h"
#include "ShapeDistance.h"
#include "DistEuclidean.h"
#include "TimeProfiler.h"
#include "ShapeGrid.h"


#include "LocFeatAlg.h"
/** Feature Parameters */
FeatParams::FeatParams(FeatureType type, Model3D * model, double radius, const VKString & externalFile)
{
	m_FeatureType = type;
	m_fRadius = radius;
	m_sExternalFile = externalFile;
	m_pModel = model;
	m_fAmplify = -1;
	m_iHistBins = 8;
}

FeatParams::FeatureType FeatParams::Str2Type(const VKString & str)
{
	if (str=="FullLocalPCA")
		return TYPE_FullLocalPCA;
	else if (str=="SimplePCA")
		return TYPE_SimplePCA;
	else if (str=="GeodesicPCA")
		return TYPE_GeodesicPCA;
	else if (str=="OrientedGeodesicPCA")
		return TYPE_OrientedGeodesicPCA;
	else if (str=="GeodesicZVariance")
		return TYPE_GeodesicZVariance;
	else if (str=="Height")
		return TYPE_Height;
	else if (str=="CoAlignedXYZ")
		return TYPE_CoAlignedXYZ;
	else if (str=="AbsCurv")
		return TYPE_AbsCurv;
	else if (str=="AbsCurvGeodesicAvg")
		return TYPE_AbsCurvGeodesicAvg;
	else if (str=="SDF")
		return TYPE_SDF;
	else if (str=="SDFGeodesicAvg")
		return TYPE_SDFGeodesicAvg;
	else if (str=="DistHist")
		return TYPE_DistHist;
	else if (str=="ExternalArff")
		return TYPE_ExternalArff;
	else
		assert(false);
}

/** Feature Algorithms */
ShapeValue * FeatAlgs::ComputeFeatures(const std::vector<FeatParams*> & params)
{
	assert(params.size()>0);
	std::vector<double*> vals;
	std::vector<int> nvals;
	int nvalsCurr;
	int ntotal = 0;
	for (int i=0; i<(int)params.size(); i++)
	{
		double * valsCurr = ComputeFeature(params[i], &nvalsCurr);
		vals.push_back(valsCurr);
		nvals.push_back(nvalsCurr);
		ntotal += nvalsCurr;
	}
	assert(ntotal>0);
	
	assert(params[0]->m_pModel->GetGeomShape("pnts")!=NULL);
	ShapePointCloud * pnts = params[0]->m_pModel->GetGeomShape("pnts")->SafeCastToPointCloud(); assert(pnts!=NULL);
	
	ShapeValue * shapevals = new ShapeValue(pnts, "values", ntotal);
	double * allvalues = new double[pnts->NumPoints() * ntotal];
	for (int p=0; p<pnts->NumPoints(); p++)
	{
		int offset = 0;
		for (int v=0; v<(int)vals.size(); v++)
		{
			for (int d=0; d<nvals[v]; d++)
				allvalues[p*ntotal+offset+d] = vals[v][p*nvals[v] + d];
			offset += nvals[v];
		}
	}
	shapevals->InitCache(ShapeValue::SVTYPE_PerPointSet, -1);
	shapevals->SetPntsValues(allvalues, pnts->NumPoints() * ntotal);
	return shapevals;
}

double * FeatAlgs::ComputeFeature(FeatParams* params, int * nvals)
{
	double * values = NULL;
	*nvals = 0;
	if (params->m_FeatureType == FeatParams::TYPE_Invalid)
		assert(false);
	else if (params->m_FeatureType == FeatParams::TYPE_SimplePCA
			 || params->m_FeatureType == FeatParams::TYPE_GeodesicPCA
			 || params->m_FeatureType == FeatParams::TYPE_FullLocalPCA
			 || params->m_FeatureType == FeatParams::TYPE_OrientedGeodesicPCA)
		values = ComputePCAFeatures(params, nvals);
	else if (params->m_FeatureType == FeatParams::TYPE_GeodesicZVariance)
		values = ComputeZVarianceFeatures(params, nvals);
	else if (params->m_FeatureType == FeatParams::TYPE_Height)
		values = ComputeHeightFeatures(params, nvals);
	else if (params->m_FeatureType == FeatParams::TYPE_CoAlignedXYZ)
		values = ComputeCoAlignedFeatures(params, nvals);
	else if (params->m_FeatureType == FeatParams::TYPE_AbsCurv)
		values = ComputeCurvFeatures(params, nvals);
	else if (params->m_FeatureType == FeatParams::TYPE_AbsCurvGeodesicAvg)
	{
		int tempnvals = 0;
		double * tempvals = ComputeCurvFeatures(params, &tempnvals);
		values = ComputeAverageFeature(tempvals, tempnvals, params, nvals);
	}
	else if (params->m_FeatureType == FeatParams::TYPE_SDF)
		values = ComputeSDFFeatures(params, nvals);
	else if (params->m_FeatureType == FeatParams::TYPE_SDFGeodesicAvg)
	{
		int tempnvals = 0;
		double * tempvals = ComputeSDFFeatures(params, &tempnvals);
		values = ComputeAverageFeature(tempvals, tempnvals, params, nvals);
	}
	else if (params->m_FeatureType == FeatParams::TYPE_DistHist)
		values = ComputeDistHistograms(params, nvals);
	else if (params->m_FeatureType == FeatParams::TYPE_ExternalArff)
		values = ReadExternalFeatures(params, nvals);
	else
		assert(false);
	assert(values!=NULL && *nvals>0);
	return values;
}

double * FeatAlgs::ComputePCAFeatures(FeatParams* params, int * nvals)
{
	assert(params->m_pModel->GetGeomShape("pnts")!=NULL);
	ShapePointCloud * pnts = params->m_pModel->GetGeomShape("pnts")->SafeCastToPointCloud();
	assert(pnts!=NULL);

	assert(params->m_pModel->GetGeomShape("dense")!=NULL);
	ShapePointCloud * dense = params->m_pModel->GetGeomShape("dense")->SafeCastToPointCloud();
	assert(dense!=NULL);
	
	assert(params->m_pModel->GetGeomShape(Model3D::m_sMainGeomName)!=NULL);
	ShapeMesh * mesh = params->m_pModel->GetGeomShape(Model3D::m_sMainGeomName)->SafeCastToMesh();
	assert(mesh!=NULL);
	
	ShapeDistance * dist = NULL;
	
	if (params->m_FeatureType==FeatParams::TYPE_GeodesicPCA ||
		params->m_FeatureType==FeatParams::TYPE_OrientedGeodesicPCA ||
		params->m_FeatureType==FeatParams::TYPE_GeodesicZVariance)
	{
		dist = dense->GetDistanceMetric("metric");
		assert(dist!=NULL);
	}

	ShapeLocalPCA * localPCA = NULL;
	
	if (dist!=NULL)
		localPCA = new ShapeLocalPCA(pnts, "pcavals", params->m_fRadius, dense, dist);
	else
		localPCA = new ShapeLocalPCA(pnts, "pcavals", params->m_fRadius, -1, false);
	localPCA->ComputePntsPCA(true);
	*nvals = 4;
	if (params->m_FeatureType == FeatParams::TYPE_FullLocalPCA)
		*nvals = 7;
	
	ShapeGrid * insideout = NULL;
	if (params->m_FeatureType == FeatParams::TYPE_OrientedGeodesicPCA)
		insideout = new ShapeGrid(mesh, 0, 0.02);
	
	double * values = new double[pnts->NumPoints() * *nvals];
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		double stddev[3];
		R3Triad axis = localPCA->GetPCA(pnts->GetPoint(i), stddev);
		assert(stddev[0]>=stddev[1] && stddev[1]>=stddev[2]);
		
		if (params->m_FeatureType == FeatParams::TYPE_FullLocalPCA)
		{
			values[i * *nvals + 0] = axis[0][0];
			values[i * *nvals + 1] = axis[0][1];
			values[i * *nvals + 2] = axis[1][0];
			values[i * *nvals + 3] = axis[1][1];
			values[i * *nvals + 4] = stddev[0];
			values[i * *nvals + 5] = stddev[1];
			values[i * *nvals + 6] = stddev[2];
		}
		else
		{
			values[i * *nvals + 0] = 1.-stddev[1] / stddev[0];	// high => linear
			values[i * *nvals + 1] = 1.-stddev[2] / stddev[0];   // high => planar (strange)
			R3Vector av = axis[2];
			if (params->m_FeatureType == FeatParams::TYPE_OrientedGeodesicPCA)	// fix orientation for the normal
				av = AlignWithTheNormal(pnts->GetPoint(i), av, insideout);
			else
				av = av.Dot(R3Vector(0,0,1)) < 0 ? -av : av;
			av.Normalize();
			values[i * *nvals + 2] = 1.-acos(av.Dot(R3Vector(0,0,1)))/M_PI;			 // high => normal=up
			values[i * *nvals + 3] = acos(vkAbs(axis[0].Dot(R3Vector(0,0,1))))/M_PI;  // high => direction perpendicular to up
		}
	}
	return values;
}

R3Vector FeatAlgs::AlignWithTheNormal(const ShapePoint & p, const R3Vector & v, ShapeGrid * insideout)
{
	if (insideout==NULL)
		return (v.Dot(p.GetNormal())<0) ? -v : v;	// TODO: perhaps inside-out...
	else
	{
		R3Point pos = p.GetPosition();
		R3Vector norm = p.GetNormal();
		norm.Normalize();
		norm *= 0.5*insideout->m_fGridCellSize;
		double depthAtPoint = insideout->PointDepth(pos-norm);
		double depthAtOffset = insideout->PointDepth(pos+norm);
		if (depthAtPoint < depthAtOffset)
			norm = -norm;
		norm.Normalize();
		
		if (norm.Dot(v)<0)
			return -v;
		else
			return v;
	}
}

double * FeatAlgs::ComputeAverageFeature(const double * inFeat, int inNVals, FeatParams* params, int * nvals)
{
	assert(params->m_fRadius>0);	// need some nhd...
	assert(params->m_pModel->GetGeomShape("pnts")!=NULL);
	ShapePointCloud * pnts = params->m_pModel->GetGeomShape("pnts")->SafeCastToPointCloud();
	assert(pnts!=NULL);
	
	assert(params->m_pModel->GetGeomShape("dense")!=NULL);
	ShapePointCloud * dense = params->m_pModel->GetGeomShape("dense")->SafeCastToPointCloud();
	assert(dense!=NULL);
	ShapeDistance * dist = dense->GetDistanceMetric("metric");
	assert(dist!=NULL);
	
	*nvals = inNVals;
	double * values = new double[pnts->NumPoints() * *nvals];
	TimeProfiler timer;
//	std::cout<<"InFeat("<<inNVals<<"): [";
//	for (int i=0; i<pnts->NumPoints(); i++)
//		std::cout<<inFeat[i]<<" ";
//	std::cout<<"]"<<std::endl;
	
	std::vector<int> dense2sparse;
	for (int i=0; i<dense->NumPoints(); i++)
		dense2sparse.push_back(-1);
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		const PointInSet * closest = dense->GetKdTree()->FindClosest(pnts->PointPosition(i));
		assert(closest!=NULL && closest->id>=0 && closest->id < dense->NumPoints());
		dense2sparse[closest->id] = i;
	}
	
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		timer.WriteProgress("avg", i, pnts->NumPoints());
		std::vector<ShapePoint> neighborhood;
		dist->NearestNeighborhood(pnts->GetPoint(i), params->m_fRadius, &neighborhood);
				
		for (int d=0; d<inNVals; d++)
		{
			double meanval = 0;
			int counter = 0;
			
			for (int nn=0; nn<(int)neighborhood.size(); nn++)
			{
				int denseID = neighborhood[nn].PointID();
				assert(denseID < dense->NumPoints());
				int sparseID = dense2sparse[denseID];
				if (sparseID>=0)
				{
					assert(sparseID<pnts->NumPoints());
					meanval += inFeat[sparseID*inNVals + d];
					counter++;
				}
			}
			//std::cout<<"Point: "<<i<<" - "<<meanval<<" / "<<neighborhood.size()<<std::endl;
			values[i * *nvals + d] = meanval/counter;
		}
	}
	std::cout<<std::endl;
	return values;
}

double * FeatAlgs::ComputeSDFFeatures(FeatParams* params, int * nvals)
{
	double scale = 1.;	// assumed shape is normalized
	
	assert(params->m_pModel->GetGeomShape("pnts")!=NULL);
	ShapePointCloud * pnts = params->m_pModel->GetGeomShape("pnts")->SafeCastToPointCloud();
	assert(pnts!=NULL);
	assert(params->m_pModel->GetGeomShape(Model3D::m_sMainGeomName)!=NULL);
	ShapeMesh * mesh = params->m_pModel->GetGeomShape(Model3D::m_sMainGeomName)->SafeCastToMesh();
	assert(mesh!=NULL);

	*nvals = 1;
	double * values = new double[pnts->NumPoints() * *nvals];
	
	TimeProfiler timer;
	R3Mesh * savePtr=NULL;
	std::vector<int> triangleMap, vertexMap;
	R3MeshSearchTree * searchTree = mesh->CreateSearchTree(&savePtr, triangleMap, vertexMap);
	assert(savePtr!=NULL && searchTree!=NULL);

	for (int i=0; i<pnts->NumPoints(); i++)
	{
		if (i%10==0)
			timer.WriteProgress("sdftimer", i, pnts->NumPoints());
		ShapePoint p = pnts->GetPoint(i);
		R3Vector n = p.GetNormal();
		assert(p.GetMeshGeometry()==mesh);
		
		double v0 = ComputeSDFAtPoint(p,  n, true, savePtr, searchTree);
		if (v0 < 0)
			v0 = ComputeSDFAtPoint(p,  n, false, savePtr, searchTree);

		double v1 = ComputeSDFAtPoint(p, -n, true, savePtr, searchTree);
		if (v1 < 0)
			v1 = ComputeSDFAtPoint(p, -n, false, savePtr, searchTree);

		//std::cout<<"P["<<i<<"]: v0="<<v0<<" v1="<<v1<<std::endl;
		
		double vmin = (v1 < 0 || (v0 >= 0 && v0 < v1)) ? v0 : v1;
		values[i] = (vmin < 0 ? 0 : vmin) * scale;
		//std::cout<<"P["<<i<<"] value="<<values[i]<<std::endl;
	}
	std::cout<<std::endl;
	return values;
}

double FeatAlgs::ComputeSDFAtPoint(const ShapePoint & pnt, const R3Vector & normal, bool only_hit_interior_surfaces,
								   R3Mesh * savePtr, R3MeshSearchTree * searchTree)
{
	double scale = 1.;	// assumed shape is normalized
	static int const NUM_RAYS = 30;
	static R3Vector const CONE_DIRS[NUM_RAYS] = {
        R3Vector( 0.407497f,  0.134086f,  0.903309f),
        R3Vector( 0.128722f, -0.822715f,  0.553688f),
        R3Vector( 0.359298f,  0.686129f,  0.632560f),
        R3Vector( 0.813318f, -0.153956f,  0.561081f),
        R3Vector(-0.707169f,  0.296987f,  0.641647f),
        R3Vector(-0.478783f, -0.543441f,  0.689520f),
        R3Vector( 0.561083f,  0.523634f,  0.641087f),
        R3Vector(-0.380898f, -0.655789f,  0.651811f),
        R3Vector( 0.722588f,  0.472155f,  0.504912f),
        R3Vector(-0.535205f,  0.421235f,  0.732200f),
        R3Vector( 0.577593f, -0.477896f,  0.661817f),
        R3Vector( 0.393652f, -0.144484f,  0.907834f),
        R3Vector( 0.141508f, -0.660777f,  0.737122f),
        R3Vector(-0.274967f, -0.141703f,  0.950954f),
        R3Vector(-0.304935f,  0.474085f,  0.825989f),
        R3Vector( 0.711150f,  0.301383f,  0.635164f),
        R3Vector(-0.696119f, -0.112341f,  0.709082f),
        R3Vector(-0.248327f,  0.381035f,  0.890588f),
        R3Vector(-0.0772394f, 0.0462501f, 0.995939f),
        R3Vector( 0.438024f,  0.656463f,  0.614159f),
        R3Vector(-0.483384f, -0.542477f,  0.687065f),
        R3Vector( 0.498569f, -0.697466f,  0.514752f),
        R3Vector(-0.128902f, -0.0173429f, 0.991506f),
        R3Vector( 0.338759f, -0.318772f,  0.885227f),
        R3Vector(-0.574364f, -0.463887f,  0.674474f),
        R3Vector(-0.459207f,  0.260125f,  0.849390f),
        R3Vector(-0.628057f, -0.528928f,  0.570771f),
        R3Vector( 0.232916f, -0.328273f,  0.915416f),
        R3Vector( 0.231667f, -0.796882f,  0.557951f),
        R3Vector(-0.270612f, -0.809654f,  0.520797f),
	};
		
	R3Vector in = normal;
	in.Normalize();
	R3Vector u, v;
	if (vkAbs(in.X())>vkAbs(in.Y()) && vkAbs(in.X())>vkAbs(in.Z()))
		u = R3Vector(in.Y(), -in.X(), 0);
	else
		u = R3Vector(0, in.Z(), -in.Y());
	u.Normalize();
	v = in % u;
	v.Normalize();
	
	R4Matrix rot(u[0], v[0], in[0], 0,
				 u[1], v[1], in[1], 0,
				 u[2], v[2], in[2], 0,
				 0,    0,    0,     1);
	
	R3Vector offset = 0.001 * scale * in;
	
	double values[NUM_RAYS];
	double weights[NUM_RAYS];
	int num_values = 0;
	R3Point position = pnt.GetPosition();
	for (int i = 0; i < NUM_RAYS; ++i)
	{
        R3Vector dir = rot * CONE_DIRS[i];
        R3Ray ray(position + offset, dir);
		
		R3MeshIntersection closest;
		searchTree->FindIntersection(ray, closest);
		if (closest.type!=R3_MESH_NULL_TYPE)
		{
			assert(closest.face!=NULL);
			if (!only_hit_interior_surfaces || savePtr->FaceNormal(closest.face).Dot(in) >= 0)
			{
				values[num_values] = closest.t;
				weights[num_values] = CONE_DIRS[i][2];  // cos(angle) is just the z-component
				num_values++;
//				std::cout<<"   Isect: "<<closest.t<<std::endl;
			}
		}
	}
	
	if (num_values <= 0)
        return -1.0;  // either the normal is in the wrong direction or this is a 2D surface
	
	// Outlier rejection: reject all values more than one standard deviation from the median
	int mid = num_values / 2;  // integer division takes floor
	std::nth_element(values, values + mid, values + num_values);
	
	// Compute variance
	double sum_values = 0, sum_squares = 0;
	for (int i = 0; i < num_values; ++i)
	{
        sum_values += values[i];
        sum_squares += (values[i] * values[i]);
	}
	
	double avg = sum_values / num_values;
	double var = sum_squares / num_values - avg * avg;
	
	sum_values = 0;
	double sum_weights = 0;
	for (int i = 0; i < num_values; ++i)
	{
		if (pow(values[i] - values[mid], 2.) <= var)
		{
			sum_values += (weights[i] * values[i]);
			sum_weights += weights[i];
		}
	}
	
	double val = 0;
	if (sum_weights > 0)
		val = ((double)sum_values / (sum_weights * scale));
	else  // should never happen
        val =(double)(values[mid] / scale);
	if (val<0)
		return 0;
	else if (val>1)
		return 1;
	return val;
	
}

double * FeatAlgs::ComputeCurvFeatures(FeatParams* params, int * nvals)
{
	double scale = 1.;
	double nhd_radius = 0.05 * scale;
	assert(params->m_pModel->GetGeomShape("pnts")!=NULL);
	ShapePointCloud * pnts = params->m_pModel->GetGeomShape("pnts")->SafeCastToPointCloud();
	assert(pnts!=NULL);

	assert(params->m_pModel->GetGeomShape("dense")!=NULL);
	ShapePointCloud * dense = params->m_pModel->GetGeomShape("dense")->SafeCastToPointCloud();
	assert(dense!=NULL);

	*nvals = 1;
	double * values = new double[pnts->NumPoints()* *nvals];
	
	const R3Kdtree<const PointInSet *> * kdTree = dense->GetKdTree();
	
	std::vector<R3Vector> pntsNormals;
	std::vector<R3Vector> denseNormals;
	PrecomputeNormals(pnts, pntsNormals);
	PrecomputeNormals(pnts, denseNormals);
	
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		R3Vector sum_offsets(0,0,0);
		long num_offsets=0;
		RNArray<const PointInSet*> nhd_points;
		kdTree->FindAll(pnts->PointPosition(i), 0, nhd_radius, nhd_points);
		R3Vector normal = pntsNormals[i];
		R3Point position = pnts->PointPosition(i); // point position
		normal.Normalize();
		
		for (int n=0; n<nhd_points.NEntries(); n++)
		{
			R3Vector pntNorm = denseNormals[nhd_points[n]->id];
			R3Point pntPos = nhd_points[n]->GetPosition(); // neighbor's position
			if (pntNorm.Dot(normal) < 0)  // ignore points on hidden side
				pntNorm *= -1;
			R3Vector offset = pntPos-position;  offset.Normalize();
			sum_offsets += offset;
			num_offsets++;
		}
		
		values[i] = num_offsets > 0 ? (vkAbs(sum_offsets.Dot(normal)) / num_offsets) : 0;	//  abs curvature
	}
	return values;
}

void FeatAlgs::PrecomputeNormals(ShapePointCloud * pnts, std::vector<R3Vector> & pntsNormals)
{
	int fid;
	double b1, b2;
	const ShapeMesh * mesh = NULL;
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		R3Vector normal;
		pnts->GetPoint(i).GetBarycentricCoords(&fid, &b1, &b2, &mesh);
		assert(mesh!=NULL);
		// NOTE: these functions are not doing SLERP -- reimplement!!!!
		R3Vector n0 = mesh->VertexNormal(mesh->VertexOnFace(fid, 0));
		R3Vector n1 = mesh->VertexNormal(mesh->VertexOnFace(fid, 0));
		R3Vector n2 = mesh->VertexNormal(mesh->VertexOnFace(fid, 0));
		
		normal = b1 * n0 + b2 * n1 + (1-b1-b2) * n2;
		
		pntsNormals.push_back(normal);
	}
}


double * FeatAlgs::ComputeCoAlignedFeatures(FeatParams* params, int * nvals)
{
	assert(false);
//	int nvals = 3;
//	vals = new ShapeValue(pnts, "values", nvals);
//	double * values = new double[pnts->NumPoints() * nvals];
//	
//	FillValuesWithPeterXYZ(mesh, alignedMesh, densePoints, pnts, values);
//	
//	vals->InitCache(ShapeValue::SVTYPE_PerPointSet, -1);
//	vals->SetPntsValues(values, pnts->NumPoints() * nvals);
}

double * FeatAlgs::ReadExternalFeatures(FeatParams* params, int * nvals)
{
	assert(params->m_pModel->GetGeomShape("pnts")!=NULL);
	ShapePointCloud * pnts = params->m_pModel->GetGeomShape("pnts")->SafeCastToPointCloud();
	assert(pnts!=NULL);
	
	ShapeValue * vals = new ShapeValue(pnts);
	vals->ReadShapeValues(params->m_sExternalFile.c_str());
	
	std::vector<int> dims;
	dims.push_back((int)params->m_fRadius);
	if (dims[0]<0)
		dims[0] = 0;	// take first value
	assert(dims[0]>=0 && dims[0]<vals->NDims());
	int ndims = (int)dims.size();
	//std::cout<<"NDims: "<<dims.size()<<std::endl;
	
	*nvals = ndims;
	double * features = new double[pnts->NumPoints()* ndims];
	
	//std::cout<<"Radius: "<<params->m_fRadius<<std::endl;
	
	double * minvalues = new double[ndims];
	double * maxvalues = new double[ndims];
	for (int d=0; d<ndims; d++)
	{
		minvalues[d] = RN_INFINITY;
		maxvalues[d] = -RN_INFINITY;
	}
	
	for (int p=0; p<pnts->NumPoints(); p++)
	for (int d=0; d<ndims; d++)
	{
		features[p * ndims + d] = vals->Value(pnts->GetPoint(p), dims[d]);
		minvalues[d] = vkMin(minvalues[d], features[p * ndims + d]);
		maxvalues[d] = vkMax(maxvalues[d], features[p * ndims + d]);
	}
	
	if (params->m_fAmplify>=0)
	{
		for (int p=0; p<pnts->NumPoints(); p++)
		for (int d=0; d<ndims; d++)
		{
			if (params->m_fAmplify>0)
				features[p * ndims + d] = pow(features[p * ndims + d], params->m_fAmplify);
			else if (params->m_fAmplify==0  && (maxvalues[d]-minvalues[d])>0)
				features[p * ndims + d] = (features[p * ndims + d]-minvalues[d]) / (maxvalues[d]-minvalues[d]);
		}
	}
	return features;
}

double * FeatAlgs::ComputeZVarianceFeatures(FeatParams* params, int * nvals)
{
	assert(params->m_pModel->GetGeomShape("pnts")!=NULL);
	ShapePointCloud * pnts = params->m_pModel->GetGeomShape("pnts")->SafeCastToPointCloud();
	assert(pnts!=NULL);

	assert(params->m_pModel->GetGeomShape("dense")!=NULL);
	ShapePointCloud * dense = params->m_pModel->GetGeomShape("dense")->SafeCastToPointCloud();
	assert(dense!=NULL);
	ShapeDistance * dist = dense->GetDistanceMetric("metric");
	assert(dist!=NULL);

	*nvals = 1;
	double * values = new double[pnts->NumPoints() * *nvals];
	TimeProfiler timer;
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		timer.WriteProgress("zvarprog", i, pnts->NumPoints());
		double zvar = 0;
		std::vector<ShapePoint> neighborhood;
		dist->NearestNeighborhood(pnts->GetPoint(i), params->m_fRadius, &neighborhood);
		double zmean = 0;
		for (int p=0; p<(int)neighborhood.size(); p++)
			zmean += neighborhood[p].GetPosition().Z();
		zmean/=(int)neighborhood.size();
		for (int p=0; p<(int)neighborhood.size(); p++)
			zvar += pow(neighborhood[p].GetPosition().Z()-zmean, 2.);
		values[i * *nvals + 0] = sqrt(zvar);
	}
	std::cout<<std::endl;
	return values;
}

double * FeatAlgs::ComputeHeightFeatures(FeatParams* params, int * nvals)
{
	assert(params->m_pModel->GetGeomShape("pnts")!=NULL);
	ShapePointCloud * pnts = params->m_pModel->GetGeomShape("pnts")->SafeCastToPointCloud();
	assert(pnts!=NULL);
	
	assert(params->m_pModel->GetGeomShape(Model3D::m_sMainGeomName)!=NULL);
	ShapeMesh * mesh = params->m_pModel->GetGeomShape(Model3D::m_sMainGeomName)->SafeCastToMesh();
	assert(mesh!=NULL);

	double scale = 1.;
	double ground = mesh->VertexPosition(0).Z();
	for (int i=0; i<mesh->NVertices(); i++)
		ground = vkMin(ground, mesh->VertexPosition(i).Z());
	
	*nvals = 1;
	double * values = new double[pnts->NumPoints() * *nvals];
	for (int i=0; i<pnts->NumPoints(); i++)
		values[i] = scale * (pnts->PointPosition(i).Z() - ground);
	return values;
}

double * FeatAlgs::ComputeDistHistograms(FeatParams* params, int * nvals)
{
	assert(params->m_fRadius>0 && params->m_iHistBins>1);
	
	assert(params->m_pModel->GetGeomShape("pnts")!=NULL);
	ShapePointCloud * pnts = params->m_pModel->GetGeomShape("pnts")->SafeCastToPointCloud();
	assert(pnts!=NULL);
	
	assert(params->m_pModel->GetGeomShape("dense")!=NULL);
	ShapePointCloud * dense = params->m_pModel->GetGeomShape("dense")->SafeCastToPointCloud();
	assert(dense!=NULL);
	const R3Kdtree<const PointInSet*> * kdTree = dense->GetKdTree();
//	ShapeDistance * dist = dense->GetDistanceMetric("metric");
//	assert(dist!=NULL);

	*nvals = params->m_iHistBins;
	double * dh = new double[*nvals * pnts->NumPoints()];
	for (int i=0; i<*nvals * pnts->NumPoints(); i++)
		dh[i] = 0;
	
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		RNArray<const PointInSet*> nhd_points;
		kdTree->FindAll(pnts->PointPosition(i), 0, params->m_fRadius, nhd_points);
		int counter =0;
		for (int j=0; j<nhd_points.NEntries(); j++)
		{
			double dist = R3Distance(pnts->PointPosition(i), nhd_points[j]->GetPosition());
			int bin = (int)(params->m_iHistBins * dist / params->m_fRadius + 0.5);
			if (bin>=0 && bin<params->m_iHistBins)
			{
				dh[i * params->m_iHistBins + bin] += 1.;
				counter++;
			}
		}
		
		// original implementation had no normalization...
		//for (int d=0; d<params->m_iHistBins; d++)
		//	dh[i*params->m_iHistBins + d] /= counter;
	}
	return dh;
}



