#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ShapeDistance.h"
#include "VKString.h"
#include "TimeProfiler.h"
#include "FileManagement.h"
#include "ShapeGrid.h"

int main(int argc, char** argv )
{
	if (argc<2)
	{
		std::cout<<"Usage: ./"<<argv[0]<<" inputDir [params] [optional]"<<std::endl;
		std::cout<<"Params:"<<std::endl;
		std::cout<<" -dirType type : MyFormat (default), ColladaDirs, MeshFiles\n";
		std::cout<<" -areaAdjusted expectedArea \n";
		std::cout<<" -type (see options below)\n";
		std::cout<<"     Read : read file)\n";
		std::cout<<"     IterFar [eucl geod] : iterative farthest (euclidean or geodesic)\n";
		std::cout<<"     Random : randomly sample points proportional to triangle area\n";
		std::cout<<"     AGDExtrema :  extrema of average geodesic distances\n";
		std::cout<<"  X  CurvatureExtrema :  \n";
		std::cout<<" -npoints NumPoints (approximate number of points, ignored by some)\n";
		std::cout<<" -pntsName pntsName\n";
		std::cout<<"Optional:"<<std::endl;
		std::cout<<" -model model or name"<<std::endl;
		std::cout<<" -hollow - prune points inside the mesh"<<std::endl;
		std::cout<<" -mesh meshName (uses \"main\" if not defined)\n";
		std::cout<<" -process : process only if file does not exist yet\n";
		std::cout<<" -writeGeometry : write geometry (does not include it in the model file)\n";
		std::cout<<" -outGeomIsFile : geometry name is filename (relative path)  \n";
		std::cout<<" -outGeomIsReallyFile : geometry name is filename (absolute path)  \n";
		std::cout<<" -manifold\n";
		std::cout<<" -writeTiming filename : timing\n";
		std::cout<<" -randseed num : random seed\n";
		std::cout<<" -vis\n";
		exit(0);
	}
	
	double expectedArea = -1;
	bool hollow = false;
	bool processIfNoFile = false;
	bool visualize = false;
	bool writeGeometry = false;
	bool outGeomIsReallyFile = false;
	ModelDatabase::InputDirFormat dirFormat = ModelDatabase::DBINPUTDIR_MyFormat;
	VKString dbDir(argv[1]);
	VKString outDir(argv[1]);
	VKString dbName = "none";
	VKString pntsName = "none";
	VKString writeTiming = "none";
	int createFlags = ModelDatabase::CREATE_DIST_EUCLIDEAN;
	int readFlags = ModelDatabase::READ_MESH_AS_SOUP | ModelDatabase::READ_MESH_AS_MANIFOLD;
	int writeFlags = ModelDatabase::WRITE_MODEL3D_FILES | ModelDatabase::WRITE_ARG_SINGLE_GEOM;	
		
	// these have to be specified
	int nPnts = 0;
	VKString type = "none";
	VKString distName = "none";
	VKString meshName = Model3D::m_sMainGeomName;
	VKString modelToProcess = "all";
	bool outGeomInMyFormat = true;
	int randseed = -1;
	
	// parse params	
	for (int i=2; i<argc; i++)
	{
		if (strcmp(argv[i], "-dirType")==0)
		{
			i++;
			if (strcmp(argv[i], "MeshFiles")==0)
				dirFormat = ModelDatabase::DBINPUTDIR_MeshFiles;
			else if (strcmp(argv[i], "MyFormat")==0)
				dirFormat = ModelDatabase::DBINPUTDIR_MyFormat;
			else if (strcmp(argv[i], "ColladaDirs")==0)
				dirFormat = ModelDatabase::DBINPUTDIR_ColladaDirs;
			else
				dirFormat = ModelDatabase::DBINPUTDIR_Unknown;
		}
		else if (strcmp(argv[i], "-npoints")==0)
			nPnts = atoi(argv[++i]);
		else if (strcmp(argv[i], "-randseed")==0)
			srand(atoi(argv[++i]));
		else if (strcmp(argv[i], "-hollow")==0)
			hollow = true;
		else if (strcmp(argv[i], "-type")==0)
		{
			i++;
			type = VKString(argv[i]);
			if (type == "IterFar")
			{
				i++;
				if (strcmp(argv[i], "eucl")==0)
				{
					createFlags |= ModelDatabase::CREATE_DIST_EUCLIDEAN;
					distName = ShapeDistance::m_sDefaultEuclidean;
				}
				else if (strcmp(argv[i], "geod")==0)
				{
					createFlags |= ModelDatabase::CREATE_DIST_GEODESIC;
					distName = ShapeDistance::m_sDefaultGeodesic;
				}
				else
					assert(false);
			}
			else if (type == "AGDExtrema")
			{
				createFlags |= ModelDatabase::CREATE_DIST_GEODESIC;
			}
		}
		else if (strcmp(argv[i], "-model")==0)
			modelToProcess = VKString(argv[++i]);
		else if (strcmp(argv[i], "-vis")==0)
			visualize = true;
		else if (strcmp(argv[i], "-pntsName")==0)
			pntsName = VKString(argv[++i]);
		else if (strcmp(argv[i], "-writeGeometry")==0)
			writeGeometry = true;
		else if (strcmp(argv[i], "-mesh")==0)
			meshName = VKString(argv[++i]);
		else if (strcmp(argv[i], "-areaAdjusted")==0)
			expectedArea = atoi(argv[++i]);
		else if (strcmp(argv[i], "-manifold")==0)
		{
			readFlags |= ModelDatabase::READ_MESH_AS_MANIFOLD;
			readFlags &= ~(ModelDatabase::READ_MESH_AS_SOUP);	
		}
		else if (strcmp(argv[i], "-writeTiming")==0)
			writeTiming = VKString(argv[++i]);
		else if (strcmp(argv[i], "-outGeomIsFile")==0)
			outGeomInMyFormat = false;
		else if (strcmp(argv[i], "-outGeomIsReallyFile")==0)
			outGeomIsReallyFile = true;
		else if (strcmp(argv[i], "-process")==0)
			processIfNoFile = false;
		else
		{
			std::cout<<"[ERROR] Unknown arg: "<<argv[i]<<std::endl;
			assert(false);
		}
	}

	if (pntsName=="none")
	{
		std::cout<<"[ERROR] Pnts name not defined. Use -pntsName"<<std::endl;
		return 0;
	}

	if (outGeomInMyFormat)
		pntsName = meshName + "_" + pntsName;
		
	VKStringList allModels;
	if (modelToProcess=="all")
		allModels = ModelDatabase::GetAllModelsInDir(dbDir, dirFormat);
	else
		allModels.push_back(modelToProcess);
	
	VKStringList mName, mFile;
	ModelDatabase::GetModelFiles(dbDir, dirFormat, allModels, mName, mFile);
	assert(mName.count()==allModels.count() && mFile.count()==allModels.count());

	TimeProfiler timer;
	ModelDatabase * db = NULL;
	if (visualize)
		db = new ModelDatabase("visOnly", "visOnly");
	
	if (allModels.count()>1)
		std::cout<<"Creating points for "<<allModels.count()<<" models"<<std::endl;
	TimeProfiler profiler;
	for (int i=0; i<allModels.count(); i++)
	{
		profiler.startedProcess(mName[i]);
		if (allModels.count()>1)
			timer.WriteProgress("FindPnts", i, allModels.count());
		
		VKString mDir = ModelDatabase::GetModelDirname(outDir, mName[i]);
		
		VKString fileWrite;
		if (!outGeomInMyFormat && outGeomIsReallyFile)
			fileWrite = pntsName;
		else if (!outGeomInMyFormat)
			fileWrite = outDir+"/"+pntsName;
		else
			fileWrite = mDir + "/"+ModelDatabase::GetModelGeometryDir()+"/"+pntsName+".pnts";
		
		bool createGeom = !processIfNoFile || !FileManagement::FileExists(fileWrite.c_str());
		if (!createGeom && !visualize)
			continue;
		
		// load model
		Model3D * model = Model3D::CreateFromFile(mFile[i], readFlags, createFlags, mName[i]);
		if (visualize && db!=NULL)
			db->AddModel(model);
		
		GeomShape * shape = model->GetGeomShape(Model3D::m_sMainGeomName);
		assert(shape!=NULL);
		ShapeMesh * mesh = shape->SafeCastToMesh();
		assert(mesh!=NULL);
		//double area = mesh->Area();
		//std::cout<<"Relative areas: "<<expectedArea<<" -> "<<area<<std::endl;
		//std::cout<<"Points: "<<nPnts<<" -> "<<(nPnts/expectedArea*area)<<std::endl;
		//exit(0);
		if (expectedArea>0 && mesh->Area()>0)
			nPnts = nPnts/expectedArea*mesh->Area();

		ShapePointCloud * pnts = NULL;
		// create pnts
		if (createGeom)
		{
			if (model->GeometryExists(pntsName))
			{
				std::cout<<"[WARNING] Model "<<mName[i].c_str();
				std::cout<<" already has "<<pntsName.c_str()<<std::endl;
				std::cout<<"\tIgnoring (geometries not re-computed)"<<std::endl;
				continue;
			}
			if (type=="Read")
				pnts =  (ShapePointCloud*)model->ReadGeometry(GeomShape::SHAPE_TYPE_POINT_CLOUD, 
															  pntsName, fileWrite, meshName);
			else if (type == "IterFar")
				pnts = ShapePointCloud::CreateIterFar(pntsName, meshName, model, nPnts, distName);
			else if (type=="Random")
				pnts = ShapePointCloud::CreateRandom(pntsName, meshName, model, nPnts);
			else if (type=="AGDExtrema")
				pnts = ShapePointCloud::CreateAGDExtrema(pntsName, meshName, model, nPnts);
			else	// todo: CurvatureExtrema
			{
				std::cout<<"\n[ERROR] Unknown points type: "<<type.c_str()<<std::endl;
				return 0;
			}
			if (type!="Read")
				model->AddGeomShape(pntsName, pnts);
		}
		
		if (!model->GeometryExists(pntsName))
			pnts =  (ShapePointCloud*)model->ReadGeometry(GeomShape::SHAPE_TYPE_POINT_CLOUD, 
														  pntsName, fileWrite, meshName);
		assert(pnts!=NULL);
		
		if (hollow)
		{
			double cell = 0.02;
			ShapeGrid * insideout = new ShapeGrid(mesh, 0, cell);
			int nremoved = 0;
			int ncurrent = pnts->NumPoints();

			std::cout<<"Hollowing out the surface:"<<std::endl;
			ShapePointCloud * surfacepnts = new ShapePointCloud(mesh, "tempsurf");
			for (int i=0; i<pnts->NumPoints(); i++)
			{
				if (insideout->PointDepth(pnts->PointPosition(i)) < 4*cell)
					surfacepnts->AddPoint(pnts->GetPoint(i));
			}
			pnts->Clear();
			for (int i=0; i<surfacepnts->NumPoints(); i++)
				pnts->AddPoint(surfacepnts->GetPoint(i));
			
			std::cout<<"Surface Points: "<<pnts->NumPoints()<<" / "<<ncurrent<<std::endl;
		}
		

		profiler.finishedProcess(mName[i], false);
		if (writeGeometry)
			pnts->WriteGeometryFile(fileWrite);
		// cleanup
		if (!visualize || db==NULL)
			delete model;
	}
	if (allModels.count()>1)
		std::cout<<"\n"<<std::flush;
	
	if (writeTiming!="none")
	{
		FileManagement::CreateDirsOnPathToFile(writeTiming.c_str());
		std::ofstream timingStream(writeTiming.c_str());
		assert(timingStream.is_open());
		timingStream<<profiler.profile().c_str()<<"\n";
		timingStream.close();
	}
	
	if (visualize)
	{	
#ifdef VISUALIZATION
		assert(db!=NULL);
		int listFlags = DatabaseVisWindow::LISTGP_Models | DatabaseVisWindow::LISTGP_Geoms;
		listFlags |= DatabaseVisWindow::LISTGP_MeshFlags;
		
		DatabaseVisWindow * dbVis = new DatabaseVisWindow(db, NULL, listFlags, 1, 1, 0, 0, 1.);
		VisWindowRenderer * wndGLUT = new RendererGLUT(argc, argv, dbVis);
#else
		std::cout<<"[WARNING] Visualization disabled (compiling flag)"<<std::endl;
#endif
	}
}


