#include "VisInteractor.h"

VisInteractor::VisInteractor(const Interactor * interactor)
{
	m_pInteractor = interactor;
}

void VisInteractor::FillRenderableArray(const InteractorPose * pose, const R3Affine & xform, const R3Affine & cam,
										std::vector<RenderableFace> & outFaces) const
{
	if (m_pInteractor->GetActorType()==Interactor::TYPE_Meshes)
	{
		const ShapeMesh * mesh = m_pInteractor->GetMesh(pose);
		const R3Affine & canonical = m_pInteractor->GetCanonical(pose);
		if (mesh->NFaces()>0)
		{
			for (int f=0; f<mesh->NFaces(); f++)
			{
				int lid=(int)outFaces.size();
				outFaces.push_back(RenderableFace());
				
				for (int v=0; v<3; v++)
				{
					R3Point p = mesh->VertexPosition(mesh->VertexOnFace(f, v));
					canonical.Apply(p);
					xform.Apply(p);
					outFaces[lid].vtx[v] = p;
				}
			}
		}
		else
		{
			glDisable(GL_LIGHTING);
			glColor3d(.2, .2, .2);
			glBegin(GL_POINTS);
			for (int v=0; v<mesh->NVertices(); v++)
			{
				R3Point p = mesh->VertexPosition(v);
				canonical.Apply(p);
				xform.Apply(p);
				glVertex3d(p.X(), p.Y(), p.Z());
			}
			glEnd();
		}
	}
#ifdef INCLUDE_BULLET
	else if (m_pInteractor->GetActorType()==Interactor::TYPE_RagDoll)
	{
		FillPhysicsRenderableArray(pose, cam, outFaces);
	}
#endif
}

#ifdef INCLUDE_BULLET
void VisInteractor::FillPhysicsRenderableArray(const InteractorPose * pose, const R3Affine & xform,
											   std::vector<RenderableFace> & faces) const
{
	for (int i=0; i<(int)pose->m_pPhysicsRigidBody.size(); i++)
	{
		btRigidBody * body = pose->m_pPhysicsRigidBody[i];
		
		btTransform physicsXForm;
		body->getMotionState()->getWorldTransform (physicsXForm);
		const btVector3 & t = physicsXForm.getOrigin ();
		btQuaternion r = physicsXForm.getRotation ();
		const btVector3 & a = r.getAxis();
		R3Affine newXForm = xform;
		
		newXForm.Translate(R3Vector(t.x(), t.y(), t.z()));
		newXForm.Rotate(R3Vector(a.x(), a.y(), a.z()), r.getAngle());
		
		btCollisionShape * 	shape = body->getCollisionShape ();
		FillShapeRenderableArray(shape, newXForm, faces);
	}
}

void VisInteractor::FillShapeRenderableArray(const btCollisionShape * shape, const R3Affine & xform,
											 std::vector<RenderableFace> & faces) const
{
	switch (shape->getShapeType())
	{
		case COMPOUND_SHAPE_PROXYTYPE:
		{
			const btCompoundShape * cShape = static_cast<const btCompoundShape*>(shape);
			for (int i=0; i<cShape->getNumChildShapes(); i++)
			{
				const btCollisionShape * child = cShape->getChildShape(i);
				const btTransform & childXForm = cShape->getChildTransform (i);
				const btVector3 & t = childXForm.getOrigin ();
				btQuaternion r = childXForm.getRotation ();
				const btVector3 & a = r.getAxis();
				
				R3Affine newXForm = xform;
				newXForm.Translate(R3Vector(t.x(), t.y(), t.z()));
				newXForm.Rotate(R3Vector(a.x(), a.y(), a.z()), r.getAngle());
				FillShapeRenderableArray(child, newXForm, faces);	// todo: convert transform
			}
			break;
		}
		case SPHERE_SHAPE_PROXYTYPE:
		{
			int ff = (int)faces.size();
			FillUnitSphere(faces, 4);
			const btSphereShape * sShape = static_cast<const btSphereShape*>(shape);
			double radius = sShape->getRadius();
			for (int f=ff; f<(int)faces.size(); f++)
			{
				//translate and scale the unit sphere
				for (int d=0; d<3; d++)
				{
					faces[f].vtx[d] *= radius;
					xform.Apply(faces[f].vtx[d]);
				}
				xform.ApplyInverse(faces[f].norm);
			}
			break;
		}
		case CAPSULE_SHAPE_PROXYTYPE:
		case CYLINDER_SHAPE_PROXYTYPE:
		{
			int ff = (int)faces.size();
			double radius, height;
			int axis;
			m_pInteractor->FillHeightAndRadius(shape, radius, height, axis);
			FillCylinderForRendering(radius, height/2, axis, 10, faces);
			for (int f=ff; f<(int)faces.size(); f++)
			{
				for (int d=0; d<3; d++)
					xform.Apply(faces[f].vtx[d]);
				xform.ApplyInverse(faces[f].norm);
			}
			break;
		}
		default:
		{ 
			static bool printedWarning = false;
			if (!printedWarning)
			{
				printedWarning = true;
				std::cout<<"[ERROR] Trying to render unknown shape: "<<shape->getShapeType()<<std::endl;
			}
		}
	}
}
#endif

void VisInteractor::FillUnitSphere(std::vector<RenderableFace> & faces, int iFactor) const
{
    int iPos = 0;
	
    double *arrV = new double[iFactor*iFactor*3];
	
    for (int j= 0; j < iFactor; j ++)
    {
        double theta = (vkPI*j)/(iFactor);
		
        for(int i=0; i<iFactor; i++ )
        {
            iPos = j*iFactor+i;
            float phi = (2*vkPI*i)/(iFactor);
            arrV[iPos*3+0] = (sin(theta)*cos(phi));
            arrV[iPos*3+1] = (sin(theta)*sin(phi));
            arrV[iPos*3+2] = (cos(theta));
		}
    }
	
    int iNext = 0;
	
    for (int j= 0; j < iFactor; j ++)
    {
        for( int i=0; i<iFactor; i++ )
        {
            if (i == iFactor - 1)
                iNext = 0;
            else iNext = i +1;
			
            iPos = (j*iFactor*6)+(i*6);
			
			int fid = (int)faces.size();
			faces.push_back(RenderableFace());
			faces[fid].vtx[0] = R3Point(arrV[(j*iFactor+i)*3+0], arrV[(j*iFactor+i)*3+1], arrV[(j*iFactor+i)*3+2]);
			faces[fid].norm = R3Vector(faces[fid].vtx[0].X(), faces[fid].vtx[0].Y(), faces[fid].vtx[0].Z());
			faces[fid].vtx[1] = R3Point(arrV[(j*iFactor+iNext)*3+0], arrV[(j*iFactor+iNext)*3+1], arrV[(j*iFactor+iNext)*3+2]);
			if (j != iFactor -1)
				faces[fid].vtx[2] = R3Point(arrV[(((j+1)*iFactor)+i)*3+0],arrV[(((j+1)*iFactor)+i)*3+1], arrV[(((j+1)*iFactor)+i)*3+2]);
			else
				faces[fid].vtx[2] = R3Point(0, 0, -1);
			
			int fid2 = (int)faces.size();
			faces.push_back(RenderableFace());
			
			faces[fid2].vtx[0] = faces[fid].vtx[2];
			faces[fid2].norm = R3Vector(faces[fid2].vtx[0].X(), faces[fid2].vtx[0].Y(), faces[fid2].vtx[0].Z());
			
			faces[fid2].vtx[1] = faces[fid].vtx[1];
			
			
            if (j != iFactor - 1)
				faces[fid2].vtx[2] = R3Point(arrV[(((j+1)*iFactor)+iNext)*3+0], arrV[(((j+1)*iFactor)+iNext)*3+1],
											 arrV[(((j+1)*iFactor)+iNext)*3+2]);
            else
				faces[fid2].vtx[2] = R3Point(0,0,-1);
			//faces[fid2].norm[2] = R3Vector(faces[fid2].vtx[2].X(), faces[fid2].vtx[2].Y(), faces[fid2].vtx[2].Z());
			
		}
    }
	delete [] arrV;
}

void VisInteractor::FillCylinderForRendering(double r, double h, int axis, int size, std::vector<RenderableFace> & face) const
{ 
	// non-uniform vertices
	std::vector<R3Point> vertices;
	int a2 = 1, a3 = 2;
	if (axis==1)
		a2 = 0;
	else if (axis==2)
		a3 = 0;
	
	// create vertices
	double pnt[3];
	pnt[a2] = 0;
	pnt[a3] = 0;
	pnt[axis] = -h;
	vertices.push_back(R3Point(pnt[0], pnt[1], pnt[2]));
	pnt[axis] = h;
	vertices.push_back(R3Point(pnt[0], pnt[1], pnt[2]));
	
	for (int i=0; i<size; i++)
	{
		double a = (double)i / (double)size * 2 * vkPI;
		pnt[a2] = r * cos(a);
		pnt[a3] = r * sin(a);
		pnt[axis] = -h;
		vertices.push_back(R3Point(pnt[0], pnt[1], pnt[2]));
		pnt[axis] = h;
		vertices.push_back(R3Point(pnt[0], pnt[1], pnt[2]));
	}
	
	// faces
	for (int i=0; i<size; i++)
	{
		int iCurrBottom = 2+2*i;
		int iCurrTop = 2+(2*i+1);
		
		int iNextBottom = 2+2*((i+1)%size);
		int iNextTop = 2+2*((i+1)%size)+1;
		
		// bottom disk
		int fid = (int)face.size();
		face.push_back(RenderableFace());
		face[fid].vtx[0] = vertices[0];
		face[fid].vtx[1] = vertices[iCurrBottom];
		face[fid].vtx[2] = vertices[iNextBottom];
		
		// upper disk
		fid = (int)face.size();
		face.push_back(RenderableFace());
		face[fid].vtx[0] = vertices[1];
		face[fid].vtx[1] = vertices[iCurrTop];
		face[fid].vtx[2] = vertices[iNextTop];
		
		// sides
		fid = (int)face.size();
		face.push_back(RenderableFace());
		face[fid].vtx[0] = vertices[iCurrBottom];
		face[fid].vtx[1] = vertices[iNextBottom];
		face[fid].vtx[2] = vertices[iNextTop];
		
		fid = (int)face.size();
		face.push_back(RenderableFace());
		face[fid].vtx[0] = vertices[iNextTop];
		face[fid].vtx[1] = vertices[iCurrTop];
		face[fid].vtx[2] = vertices[iCurrBottom];
	}
}