#include "FileManagement.h"
#include "VKString.h"
#include "TimeProfiler.h"

#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ShapeDistance.h"

#include "ShapeTriangleSoup.h"
#include "IKSolver.h"
#include "DBApp.h"
#include <vector>
#include "ShapePoint.h"

#include "CreateGTApp.h"
#include "BindingEnergy.h"
#include "PosePriorMixGauss.h"
#include "PntCorrsPose.h"

CreateGTApp::CreateGTApp(ModelDatabase * db, const VKString & writeFile, const VKString & writeXForm,
						 const VKString & outmesh, bool readPose)
: DBApp(db, NULL)
{
	m_pActor = NULL;
	m_pSolver = NULL;
	m_bRunIK = false;
	m_iSelectedSkeletonLink = -1;
	m_CameraXForm = R3identity_affine;
	m_fGroundTranslate = 0;
	if (writeXForm!="none")
	{
		m_fGroundTranslate = RN_INFINITY;
		//std::cout<<"Normalizing mesh"<<std::endl;
		// rescale models
		for (int i=0; i<db->NumModels(); i++)
		{
			Model3D * m = db->GetModel(i);
			GeomShape * shape = m->GetGeomShape(Model3D::m_sMainGeomName);
			assert(shape!=NULL);
			ShapeMesh * mesh = shape->SafeCastToMesh();
			assert(mesh!=NULL);
			
			R3Point c = mesh->Centroid();
			double r = mesh->AverageRadius(c);
			
			for (int v=0; v<mesh->NVertices(); v++)
			{
				R3Point p = (mesh->VertexPosition(v)+(-c)) / r;
//				p = R3Point(p.Z(), p.X(), p.Y());
				mesh->SetVertexPosition(v, p);
				m_fGroundTranslate = vkMin(m_fGroundTranslate, p.Z());
			}
			if (i==0)
			{
				m_OrigCentroid = c;
				m_fOrigRadius = r;
			}
		}
	}		
	
	m_sWriteXFormFile = writeXForm;
	m_sWriteFile = writeFile;
	m_sWriteMesh = outmesh;
	m_iCurrRegion = 0;
	m_iCurrPoint = 0;
	m_iCurrPose = 0;
	m_iCurrPredictedPose = -1;
//	m_fScale = .6;
//	m_fAngle = -0.5 * M_PI;
	m_fScale = 1.;
	m_fAngle = 0;
	if (!readPose)
	{
		std::cout<<"Creating initial empty pose..."<<std::endl;
		m_Poses.push_back(new PntCorrsPose((Interactor*)NULL));
		m_Poses[m_iCurrPose]->AddMatch(0, TargetPrimitive(TargetPrimitive::PRIM_None));
	}
	UpdatePoseScales();
	
	//SetRenderingFlags(0xFFFFFFFF & (~VIS_Axes) & (~VIS_Transparent));
	SetRenderingFlags(0xFFFFFFFF & (~VIS_Axes) & (~VIS_ScreenshotMode));
}

void CreateGTApp::UpdatePoseScales()
{
	for (int i=0; i<(int)m_Poses.size(); i++)
	{
		m_Poses[i]->m_NormalizingXForm = R3identity_affine;
		m_Poses[i]->m_NormalizingXForm.Scale(m_fScale);
		m_Poses[i]->m_NormalizingXForm.Rotate(R3Vector(0,0,m_fAngle));
		m_Poses[i]->m_NormalizingXForm.Translate(R3Vector(0, 0, -m_fGroundTranslate));
		//std::cout<<"UPDATING: "<<m_fScale<<" "<<m_fAngle<<" "<<m_fGroundTranslate<<std::endl;
	}
}

void CreateGTApp::PrintPoints() const
{
	std::cout<<"Pose: "<<m_iCurrPose<<" / "<<m_Poses.size()<<std::endl;
	if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size())
	{
		PntCorrsPose * pose = m_Poses[m_iCurrPose];
		for (int r=0; r<(int)pose->NMatches(); r++)
		{
			int rid = pose->Region(r);
			std::cout<<"   "<<m_sRegionNames[rid].c_str()<<" - ";
			if (pose->Prim(r).IsType(TargetPrimitive::PRIM_None))
				std::cout<<"Unassigned"<<std::endl;
			else if (pose->Prim(r).IsType(TargetPrimitive::PRIM_Point))
			{
				std::cout<<"Point"<<std::endl;
				int fid;
				double b1, b2;
				const ShapeMesh * mesh;
				R3Point pos = pose->Prim(r).GetPosition();
				pose->Prim(r).GetPoint().GetBarycentricCoords(&fid, &b1, &b2, &mesh);
				std::cout<<"    Pnt"<<": "<<fid<<" "<<b1<<" "<<b2<<" "<<(1-b1-b2)<<" ";
				std::cout<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<std::endl;
			}
			else if (pose->Prim(r).IsType(TargetPrimitive::PRIM_Ground))
			{
				std::cout<<"Ground"<<std::endl;
				R3Point pos = pose->Prim(r).GetPosition();
				std::cout<<"    Pos="<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<std::endl;
			}
			else
				std::cout<<"ERROR: unknown primitive "<<std::endl;
		}
	}
}

bool CreateGTApp::WritePosesFile() const
{
	if (m_sWriteXFormFile!="none" && m_sWriteXFormFile!="identity")
	{
		FileManagement::CreateDirsOnPathToFile(m_sWriteXFormFile.c_str());
		std::ofstream xformStream(m_sWriteXFormFile.c_str());
		assert(xformStream.is_open());
		xformStream<<"Scale "<<m_fScale<<"\n";
		xformStream<<"RotZ "<<m_fAngle<<"\n";
		xformStream<<"OrigCentroid "<<m_OrigCentroid.X()<<" "<<m_OrigCentroid.Y()<<" "<<m_OrigCentroid.Z()<<" "<<"\n";
		xformStream<<"OrigRadius "<<m_fOrigRadius<<"\n";
		xformStream<<"OrigGround "<<m_fGroundTranslate<<"\n";
		xformStream.close();
	}
	
	if (m_Poses.size()==0)
		return true;
	bool allPoints = true;
	if (m_sWriteFile!="none")
		allPoints = PntCorrsPose::WritePosesGTFile(m_sWriteFile, m_Poses, m_sRegionNames, m_sRegionTypes, m_RegionNPoints);

	// write mesh
	if (m_sWriteMesh!="none")
	{
		const ShapeMesh * mesh = m_pDB->GetModel(0)->GetGeomShape(Model3D::m_sMainGeomName)->SafeCastToMesh();
		R3Point centroid = mesh->Centroid();
		double radius = mesh->AverageRadius(centroid);

		ShapeTriangleSoup * outmesh = new ShapeTriangleSoup();
		
		R3Affine canonical = R3identity_affine;
		canonical.Scale(m_fScale);
		canonical.Rotate(R3Vector(0,0,m_fAngle));
		canonical.Translate(R3Vector(0, 0, -m_fGroundTranslate));

		for (int i=0; i<mesh->NVertices(); i++)
		{
			R3Point p = mesh->VertexPosition(i);
			canonical.Apply(p);
			outmesh->AddVertex(p.X(), p.Y(), p.Z());
		}
		
		for (int i=0; i<mesh->NFaces(); i++)
			outmesh->AddFace(mesh->VertexOnFace(i, 0), mesh->VertexOnFace(i, 1), mesh->VertexOnFace(i, 2));
		
		FileManagement::CreateDirsOnPathToFile(m_sWriteMesh.c_str());
		outmesh->WriteGeometryFile(m_sWriteMesh);
		delete outmesh;
	}
	
	return allPoints;
}

void CreateGTApp::SetPredictions(const std::vector<PntCorrsPose * > & predictions)
{
	m_Predictions = predictions;
	for (int i=0; i<(int)m_Predictions.size(); i++)
	{
		assert(i==(int)m_ClassifiedPredicted.size());
		m_ClassifiedPredicted.push_back(std::vector<PointType>());
		m_Pose2ClassifiedPnts[m_Predictions[i]] = i;
		if (m_ClassifiedPredicted[i].size()==0)
		{
			for (int r=0; r<(int)m_Predictions[i]->NMatches(); r++)
				m_ClassifiedPredicted[i].push_back(PT_Prediction);
		}
	}
}

int CreateGTApp::NPredictions() const
{
	return (int)m_Predictions.size();
}

void CreateGTApp::ReadPosesFile()
{
	const ShapeMesh * mesh = NULL;
	if (m_pDB->NumModels()>0)
		mesh = m_pDB->GetModel(0)->GetGeomShape(Model3D::m_sMainGeomName)->SafeCastToMesh();
	double x,y,z;
	VKString tempStr;
	std::ifstream xformStream(m_sWriteXFormFile.c_str());
	if (xformStream.is_open())
	{
		assert(m_sWriteXFormFile!="none" && m_sWriteXFormFile!="identity");
		m_Poses.clear();
		tempStr.readToken(xformStream);		assert(tempStr=="Scale");
		xformStream>>m_fScale;
		tempStr.readToken(xformStream);		assert(tempStr=="RotZ");
		xformStream>>m_fAngle;
		tempStr.readToken(xformStream);		assert(tempStr=="OrigCentroid");
		
		xformStream>>x>>y>>z;
		m_OrigCentroid = R3Point(x, y,z);
		tempStr.readToken(xformStream);		assert(tempStr=="OrigRadius");
		double orig = 0;
		//xformStream>>m_fOrigRadius;
		xformStream>>orig;
		//double origGround=0;
		tempStr.readToken(xformStream);		assert(tempStr=="OrigGround");
		xformStream>>m_fGroundTranslate;	// keep it - want to 
		xformStream.close();
	}
	
	//ReadPosesFile(m_sWriteFile, mesh, m_Poses);
	if (FileManagement::FileExists(m_sWriteFile.c_str()) && m_Poses.size()==0)
	{
		PntCorrsPose::CreatePosesFromGTFile(m_sWriteFile, mesh, "none", m_Poses);
		if (m_pSolver!=NULL)
		{
			for (int i=0; i<(int)m_Poses.size(); i++)
				m_pSolver->FixPoseParameterization(m_Poses[i]);
		}
	}
	if (m_Poses[m_iCurrPose]->NMatches()==0)
		m_Poses[m_iCurrPose]->AddMatch(0, TargetPrimitive(TargetPrimitive::PRIM_None));
	UpdatePoseScales();
}

void CreateGTApp::ReadPriorAsPose(const VKString & priorFile)
{
	PosePrior * prior = PosePrior::ReadPrior(priorFile);
	assert(prior!=NULL && prior->PriorType()==PosePrior::PRIOR_GaussAngles);
	prior->ReadPointStats(priorFile+".pntstats.txt");
	PosePriorGauss * gauss = (PosePriorGauss*)prior;
	PntCorrsPose * pose = new PntCorrsPose((Interactor*)NULL);
	m_iCurrPose = 0;
	m_Poses.push_back(pose);
	m_Poses[m_iCurrPose]->AddMatch(0, TargetPrimitive(TargetPrimitive::PRIM_None));
	m_pSolver->InitPoseToZero(pose);
	for (int i=0; i<(int)gauss->m_AngleDistributions.size(); i++)
	{
		PosePointGauss * gd = gauss->m_AngleDistributions[i];
		int offset = m_pSolver->m_ParamOffsetByLinkID[i];
		if (gd==NULL)
			continue;
		for (int d=0; d<gd->m_iNDims; d++)
			pose->Param(offset+d) = gd->m_aMeans[d];
	}
}


void CreateGTApp::SetRegionsFile(const VKString & regionsFile)
{
	bool ok;
	std::ifstream textStream(regionsFile.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] File does not exist: "<<regionsFile.c_str()<<std::endl;
	assert(textStream.is_open());
	while (!textStream.eof())
	{
		VKString strLine = "";
		strLine.readLine(textStream);
		if (strLine!="")
		{
			if (strLine.startsWith("#"))
				continue;
			
			VKStringList splitLine = strLine.split(" ");
			assert(splitLine.count()>=3);
			m_sRegionNames.push_back(splitLine[0]);
			m_sRegionTypes.push_back(splitLine[1]);
			m_RegionNPoints.push_back(splitLine[2].toInt(&ok));
			assert(ok);
		}
	}
	textStream.close();
}


bool CreateGTApp::KeyPressed(DatabaseVisWindow * window, const char key, int mouseX, int mouseY)
{
	switch(key)
	{
		case '`':
			std::cout<<"Keys:"<<std::endl;
			std::cout<<"  SPACE - select point"<<std::endl;
			std::cout<<"  s/S - choose label for current region"<<std::endl;
			std::cout<<"  ,/. - choose point"<<std::endl;
			std::cout<<"  </> - choose pose"<<std::endl;
			std::cout<<"  ;/' - choose region (from existing)"<<std::endl;
			std::cout<<"  [/] - rotate (around up)"<<std::endl;
			std::cout<<"  -/= - scale"<<std::endl;
			std::cout<<std::endl;
			std::cout<<"  h/H - change selected skeleton joint"<<std::endl;
			std::cout<<"  j/J - change 0th link's parameter"<<std::endl;
			std::cout<<"  k/K - change 1st link's parameter"<<std::endl;
			std::cout<<"  l/L - change 2nd link's parameter"<<std::endl;
			std::cout<<"  n   - next (mean) gaussian"<<std::endl;
			std::cout<<"  g/G - move ground"<<std::endl;
			std::cout<<"  O - freeze / run IK"<<std::endl;
			std::cout<<"  o - single IK"<<std::endl;
			std::cout<<"  x - switch to standing pose"<<std::endl;
			std::cout<<"  z - switch to sitting pose"<<std::endl;
			std::cout<<"  A - print angles"<<std::endl;
			std::cout<<"  + - add pose"<<std::endl;
			std::cout<<"  7-9, &-* - transalte"<<std::endl;
			std::cout<<"  q - Symmetrize using left side"<<std::endl;
			std::cout<<"  Q - symmetrize using right side"<<std::endl;
			std::cout<<std::endl;
			std::cout<<"  1-3, !-# - write / read current pose"<<std::endl;
			std::cout<<"  P - print possible regions"<<std::endl;
			std::cout<<"  p - print points"<<std::endl;
			std::cout<<"  ESC - exit (should proceed to next model)"<<std::endl;
			break;
		case 'P':
			for (int i=0; i<m_sRegionNames.count(); i++)
				std::cout<<(i+1)<<": "<<m_sRegionNames[i].c_str()<<" "<<m_RegionNPoints[i]<<std::endl;
			break;
		case 'A':	// write current angles
			if (m_iCurrPose<0 || m_iCurrPose >= (int)m_Poses.size()
				|| !m_Poses[m_iCurrPose]->IsParamsInit(m_pSolver->NJointParams()))
			{
				std::cout<<"No pose or no params..."<<std::endl;
				return true;
			}
			for (int i=0; i<(int)m_pSolver->m_Links.size(); i++)
			{
				SkeletonLink * l = m_pSolver->m_Links[i];
				VKString j1Str = l->m_pParent->m_sJointName;
				VKString j2Str = l->m_pChild->m_sJointName;
				j1Str = j1Str.CreateStrWithPadding(14, " ");
				j2Str = j2Str.CreateStrWithPadding(14, " ");
				std::cout<<j1Str.c_str()<<" -> "<<j2Str.c_str()<<": ";
				
				int offset = m_pSolver->m_ParamOffsetByLinkID[i];
				VKString allnumStr = "";
				for (int d=0; d<SkeletonLink::NDims(l->m_JointType); d++)
				{
					double num = ((int)(100*m_Poses[m_iCurrPose]->Param(offset+d)/M_PI))/100.;
					VKString numStr = VKString::number(num);
					allnumStr += numStr.CreateStrWithPadding(7, " ");
				}
				if (allnumStr!="")
				{
					allnumStr = allnumStr.CreateStrWithPadding(3*7, " ");
					std::cout<<allnumStr.c_str()<<" x pi";
				}
				std::cout<<std::endl;
			}
			break;
		case '+':	// symmetric flip
		{
			m_Poses.push_back(new PntCorrsPose((Interactor*)NULL));
			m_iCurrPose = (int)m_Poses.size()-1;
			m_Poses[m_iCurrPose]->InitParams(m_pSolver->NJointParams(), m_pSolver->m_SolverFlags);
			assert(m_iCurrPose==1);
			for (int i=0; i<m_pSolver->NJointParams(); i++)
				m_Poses[m_iCurrPose]->Param(i) = m_Poses[0]->Param(i);
			m_Poses[m_iCurrPose]->Param(3) += M_PI;
			
			for (int i=0; i<m_Poses[0]->NMatches(); i++)
			{
				int rid = m_Poses[0]->Region(i);
				VKString newName = m_pSolver->m_RegionNames[rid];
				if (newName.startsWith("left_"))
					newName = newName.replace("left_", "right_");
				else if (newName.startsWith("right_"))
					newName = newName.replace("right_", "left_");
				for (int c=0; c<m_pSolver->m_RegionNames.count(); c++)
					if (m_pSolver->m_RegionNames[c]==newName)
					{
						rid = c;
						break;
					}
				m_Poses[m_iCurrPose]->AddMatch(rid, m_Poses[0]->Prim(i));
			}
			UpdatePoseScales();
			break;
		}
		case 'q':	// symmetrize using left
		case 'Q':	// symmetrize using right
			if (m_iCurrPose<0 || m_iCurrPose >= (int)m_Poses.size()
				|| !m_Poses[m_iCurrPose]->IsParamsInit(m_pSolver->NJointParams()))
			{
				std::cout<<"No pose or no params..."<<std::endl;
				return true;
			}

			if (m_pSolver->m_pPrior!=NULL && m_pSolver->m_pPrior->PriorType() == PosePrior::PRIOR_MultiGauss)
			{
				int left, right;
				PosePriorMixGauss * prior = (PosePriorMixGauss*)m_pSolver->m_pPrior;
				//for (int i=0; i<prior->NSymmetryCouplings(); i++)
				for (int i=0; i<prior->NSymmetryCouplings(); i++)
				{
					prior->SymmCoupling(i, &left, &right);

					if (m_iSelectedSkeletonLink==-1 || m_iSelectedSkeletonLink==left || m_iSelectedSkeletonLink==right)
					{
//						std::cout<<"Symm Coupling: "<<std::endl;
//						SkeletonLink * link = m_pSolver->m_Links[left];
//						VKString descrStr = link->m_pParent->m_sJointName+" -> "+link->m_pChild->m_sJointName;
//						descrStr = descrStr.CreateStrWithPadding(35, " ");
//						std::cout<<"    "<<descrStr.c_str()<<std::endl;
//						if (right>=0)
//						{
//							link = m_pSolver->m_Links[right];
//							VKString descrStr = link->m_pParent->m_sJointName+" -> "+link->m_pChild->m_sJointName;
//							descrStr = descrStr.CreateStrWithPadding(35, " ");
//							std::cout<<"    "<<descrStr.c_str()<<std::endl;
//						}

						int oLeft = m_pSolver->m_ParamOffsetByLinkID[left];
						int oRight = -1;
						if (right>=0)
							oRight = m_pSolver->m_ParamOffsetByLinkID[right];

						int ndims = prior->NParamsForSymmJoint(i);
						for (int d=0; d < ndims; d++)
						{
							double diff = prior->SignedDistToSymmetry(i, d, m_Poses[m_iCurrPose]);
							int pid = prior->SymmParamID(i, d);
							if (key=='q' || oRight<0)
								m_Poses[m_iCurrPose]->Param(oLeft+pid) += diff;
						}
					}
				}
			}
			else
				std::cout<<"No prior"<<std::endl;
//			for (int i=0; i<(int)m_pSolver->m_Links.size(); i++)
//			{
//				assert(m_pSolver->m_Links[i]->m_pChild!=NULL);
//				VKString jTrgName = m_pSolver->m_Links[i]->m_pChild->m_sJointName;
//				int ndims = SkeletonLink::NDims(m_pSolver->m_Links[i]->m_JointType);
//				int oTrg = m_pSolver->m_ParamOffsetByLinkID[i];
//				if (((key=='q' && jTrgName.startsWith("right")) || (key=='Q' && jTrgName.startsWith("left"))) && ndims>0)
//				{
//					VKString jTrgStrp = jTrgName.replace("right", "").replace("left", "");
//					for (int j=0; j<(int)m_pSolver->m_Links.size(); j++)
//					{
//						if (i==j)
//							continue;
//						assert(m_pSolver->m_Links[j]->m_pChild!=NULL);
//						VKString jSrcName = m_pSolver->m_Links[j]->m_pChild->m_sJointName;
//						VKString jSrcStrp = jSrcName.replace("right", "").replace("left", "");
//						if (jTrgStrp==jSrcStrp)
//						{
//							int oSrc = m_pSolver->m_ParamOffsetByLinkID[j];
//							for (int d=0; d<ndims; d++)
//							{
//								double sign = 1.;
//								if (d!=1 && jTrgStrp!="_leg2" && jTrgStrp != "_toe")
//									sign = -1.;
//								m_Poses[m_iCurrPose]->Param(oTrg+d) = sign * m_Poses[m_iCurrPose]->Param(oSrc+d);
//							}
//						}
//					}
//				}
//			}
			break;
		case 't':
			Toggle(VIS_PosesPntNames);
			break;
		case ' ':	// select point
		{
			if (m_iCurrPose<0 || m_iCurrPose >= (int)m_Poses.size())
			{
				std::cout<<"Need to add another pose... currpose="<<m_iCurrPose<<" / "<<m_Poses.size()<<std::endl;
				return true;
			}
			
			std::cout<<"Adding point to pose: "<<m_iCurrPose<<std::endl;
			
			PntCorrsPose * pose = m_Poses[m_iCurrPose];
			
			if (m_iCurrRegion<0 || m_iCurrRegion >= (int)pose->NMatches())
			{
				std::cout<<"Need to pick a region..."<<std::endl;
				return true;
			}
			int rid = pose->Region(m_iCurrRegion);
			assert(rid >=0 && rid < m_sRegionNames.count());

			TargetPrimitive & prim = pose->Prim(m_iCurrRegion);
			
			assert(window->GetDrawableModels().size()>0);
			VisModel3D * mvis = window->GetDrawableModels()[0];
			int cameraID = mvis->SubwindowCameraID(0, window);
			ShapePoint pnt = mvis->QuickIntersection(cameraID, m_CameraXForm, mouseX, mouseY, window);
			if (pnt.Invalid() && m_pGroundPlaneBuffer!=NULL)
			{
				glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_pGroundPlaneBuffer->m_iFrameBufferObject);
				glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
				unsigned char rgba[4];
				glReadPixels(mouseX, mouseY, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, rgba);
				if (rgba[0]==0)
				{
					GLint viewport[4];
					glGetIntegerv(GL_VIEWPORT, viewport);
					
					window->LoadStandardSceneProjectionCamera();
					glMatrixMode(GL_MODELVIEW);
					glLoadIdentity();
					assert(window->GetDrawableModels().size()>0);
					VisModel3D * mvis = window->GetDrawableModels()[0];
					R3Affine camera = mvis->GetCameraTransform(mvis->SubwindowCameraID(0, window));
					camera.Load();
					
					GLdouble modelview[16];
					GLdouble projection[16];
					glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
					glGetDoublev(GL_PROJECTION_MATRIX, projection);
					GLfloat fwinZ;
					glReadPixels(mouseX, mouseY, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &fwinZ);
					GLdouble winX = mouseX, winY = mouseY, winZ = fwinZ;
					GLdouble posX, posY, posZ;
					gluUnProject( winX, winY, winZ, modelview, projection, viewport, &posX, &posY, &posZ);
					window->LoadCurrentBuffer();
					pnt = ShapePoint(R3Point(posX, posY, posZ));
				}
			}

			if(pnt.Invalid())
			{
				std::cout<<"[WARNING] Could not select point (probably clicked on empty space)"<<std::endl;
				break;
			}
			
			assert(m_iCurrPoint==0);
			if (pnt.GetPntType()!=ShapePoint::SHAPE_POINT_MESHLESS)	// point on surface
			{
				int fid; double b1, b2;
				const ShapeMesh * mesh;
				pnt.GetBarycentricCoords(&fid, &b1, &b2, &mesh);
				prim = TargetPrimitive(TargetPrimitive::PRIM_Point, pnt);
				std::cout<<"Selected point: "<<fid<<" "<<b1<<" "<<b2<<std::endl;
			}
			else
			{
				R3Point pos = pnt.GetPosition();
				prim = TargetPrimitive(TargetPrimitive::PRIM_Ground, pnt);
				std::cout<<"Ground point: ["<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<"] "<<std::endl;
			}
			assert(m_RegionNPoints[rid]==1);
			m_iCurrPoint = 0;
			int last = (int)m_Poses[m_iCurrPose]->NMatches()-1;
			assert(last>=0);
			if (!m_Poses[m_iCurrPose]->Prim(last).IsType(TargetPrimitive::PRIM_None))
			{
				m_iCurrRegion = m_Poses[m_iCurrPose]->NMatches();
				m_Poses[m_iCurrPose]->AddMatch(rid+1, TargetPrimitive(TargetPrimitive::PRIM_None));
			}
			
			break;
		}
		case 's':
		case 'S':
		{
			m_Poses[m_iCurrPose]->Region(m_iCurrRegion) += (key=='s' ? 1 : -1);
			if (m_Poses[m_iCurrPose]->Region(m_iCurrRegion) >= NRegions())
				m_Poses[m_iCurrPose]->Region(m_iCurrRegion) = (NRegions()-1);
			if (m_Poses[m_iCurrPose]->Region(m_iCurrRegion) < 0)
				m_Poses[m_iCurrPose]->Region(m_iCurrRegion) = 0;
			VKString name = m_sRegionNames[m_Poses[m_iCurrPose]->Region(m_iCurrRegion)];
			std::cout<<"Renamed Region "<<m_iCurrRegion<<": "<<name.c_str()<<std::endl;
			break;
		}
		case '1':
		case '2':
		case '3':
		case '!':
		case '@':
		case '#':
		{
			if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size() && m_pSolver!=NULL)
			{
				if (key=='1'||key=='2'||key=='3')
				{
					int id = (int)(key - '1');
					while (id>=m_SavedPoses.size())
						m_SavedPoses.push_back(NULL);
					std::cout<<"Saving pose to slot: "<<id<<std::endl;
					m_SavedPoses[id] = m_pSolver->CreateCopy(m_Poses[m_iCurrPose]);
				}
				else if (key=='!'||key=='@'||key=='#')
				{
					int id = 0;
					if (key=='@')
						id=1;
					else if (key=='#')
						id=2;
					if (id < m_SavedPoses.size() && m_SavedPoses[id]!=NULL)
					{
						std::cout<<"Reading pose from slot: "<<id<<std::endl;
						m_Poses[m_iCurrPose] = m_pSolver->CreateCopy(m_SavedPoses[m_iCurrPose]);
					}
				}
			}
			break;
		}
		case '/':
		case '?':	// previous / next pose in predictions
		{
			m_iCurrPredictedPose += (key == '?' ? -1 : 1);
			if (m_iCurrPredictedPose<0)
				m_iCurrPredictedPose = 0;
			while (m_iCurrPredictedPose >= (int)m_Predictions.size())
				m_Predictions.push_back(new PntCorrsPose((Interactor*)NULL));
			std::cout<<"Pose: "<<m_iCurrPredictedPose<<std::endl;
			UpdatePoseScales();
			break;
		}
		case '<':
		case '>':	// previous / next pose
		{
			m_iCurrPose += (key == '<' ? -1 : 1);
			if (m_iCurrPose<0)
				m_iCurrPose = 0;
			while (m_iCurrPose >= (int)m_Poses.size())
				m_Poses.push_back(new PntCorrsPose((Interactor*)NULL));
			std::cout<<"Pose: "<<m_iCurrPose<<std::endl;
			UpdatePoseScales();
			break;
		}
		case ';':
		case '\'':	// previous / next region
			if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size())
			{
				m_iCurrRegion += (key == ';' ? -1 : 1);
				if (m_iCurrRegion<0 || m_iCurrRegion>=(int)m_Poses[m_iCurrPose]->NMatches())
					m_iCurrRegion = 0;
				if (m_iCurrRegion<(int)m_Poses[m_iCurrPose]->NMatches())
				{
					VKString name = m_sRegionNames[m_Poses[m_iCurrPose]->Region(m_iCurrRegion)];
					std::cout<<"Region: "<<m_iCurrRegion<<". Name: "<<name.c_str()<<std::endl;
				}
			}
			else
				std::cout<<"[WARNING] need to add pose"<<std::endl;
			break;
//		case 'c':
//			if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size() && m_iCurrRegion<(int)m_Poses[m_iCurrPose]->m_RegionID.size())
//			{
//				PntCorrsPose * pose = m_Poses[m_iCurrPose];
//				pose->m_RegionID.erase(pose->m_RegionID.begin()+m_iCurrRegion);
//				pose->m_aFeaturePoints.erase(pose->m_aFeaturePoints.begin()+m_iCurrRegion);
//				pose->m_SkeletonParams.clear();
//				pose->m_JointPositions.clear();
//			}
//			else
//				std::cout<<"[WARNING] need to add pose or pick a different region..."<<std::endl;
//			break;
		case 'c':
			if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size()
				&& m_iCurrRegion>=0 && m_iCurrRegion < m_Poses[m_iCurrPose]->NMatches())
				m_Poses[m_iCurrPose]->Clear(m_iCurrRegion--);
			else
				std::cout<<"[WARNING] Cannot clear: "<<m_iCurrPose<<" "<<m_iCurrRegion<<std::endl;
			break;
		case ',':	//
		case '.':	// previous / next point
			if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size() && m_iCurrRegion<(int)m_Poses[m_iCurrPose]->NMatches())
			{
				m_iCurrPoint += (key == ',' ? -1 : 1);
				
				int rid = m_Poses[m_iCurrPose]->Region(m_iCurrRegion);
				int npnts = m_RegionNPoints[rid];
				if (m_iCurrPoint<=0)
					m_iCurrPoint = 0;
				else if (m_iCurrPoint>=npnts)
					m_iCurrPoint = npnts-1;
				std::cout<<"Curr Point: "<<m_iCurrPoint<<" / "<<npnts<<std::endl;
			}
			else
				std::cout<<"[WARNING] need to add pose or a region..."<<std::endl;
			break;
		case 'I':
			if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size())
				m_pSolver->FindBestInitPose(m_Poses[m_iCurrPose]);
			break;
		case 'f':
		case 'O':
		case 'o':
			if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size())
			{
				m_pSolver->SolveIKProblem(m_Poses[m_iCurrPose], key=='O');
				std::cout<<"IKStats: "<<std::endl;
				std::cout<<m_pSolver->m_pIKStats->GetStatsStr().c_str()<<std::endl;
			}
			break;
		case 'x':	// standing
		case 'Z':	// just do the rigid fitting
		case 'z':	// sitting
		case 'a':	// average pose
			if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size() && m_pSolver!=NULL)
			{
				PntCorrsPose * pose = m_Poses[m_iCurrPose];
				
				if (key=='x')
					m_pSolver->InitPoseToZero(pose);
				else if (key=='a')
				{
					VKString sit = "0 0 0 -1.57 -0.685388 -0.557437 -0.118655 -0.371568 ";
					sit += "0.628796 -0.331933 0.0236393 0.135538 -0.00220023 0.391438 -0.661077 1.74485 -0.399031 ";
					sit += "0 0 0 -0.391438 -0.661077 -1.74485 0.399031 -0 0 -0 -0.38369 -1.25328 -0.333965 0.615172 ";
					sit += "-0.249179 -0.0475238 1.52513 0.853733 -0.615172 -0.249179 0.0475238 1.52513 0.853733";
					VKStringList sitList = sit.split(" ");
					assert(sitList.count()==m_pSolver->NJointParams());
					m_pSolver->InitPoseToZero(pose);
					for (int i=0; i<sitList.count(); i++)
						pose->Param(i) = sitList[i].toDouble();
				}
				else if (key=='z')
					m_pSolver->InitPoseToSitting(pose);
//				else if (key=='Z')
//				{
//					// torso
//					pose->Param(4) = 0;
//					pose->Param(5) = -0.5 * M_PI;
//					pose->Param(6) = 0;
//
//					// pelvis
//					pose->Param(24) = 0;
//					pose->Param(25) = -0.5 * M_PI;
//					pose->Param(26) = 0;
//				}
				m_pSolver->FindBestRigidXForm(pose);
			}
			break;
		case 'h':
		case 'H': // change selected skeleton joint
		case 'j':
		case 'J': // update skeleton params (0)
		case 'k':
		case 'K': // update skeleton params (1)
		case 'l':
		case 'L': // update skeleton params (2)
		case 'y':
		case 'Y': // update skeleton rotation 1
		case 'r':
		case 'R': // update skeleton rotation 2
		case '7':
		case '&': // update skeleton x-position
		case '8':
		case '*': // update skeleton y-position
		case '9':
		case '(': // update skeleton z-position
			if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size() && m_pSolver!=NULL)
			{
				m_iSelectedGaussianMean = -1;
				PntCorrsPose * pose = m_Poses[m_iCurrPose];
				if (key=='h' || key=='H')
				{
					m_iSelectedSkeletonLink += key=='h' ? 1 : -1;
					
					if (m_iSelectedSkeletonLink<0)
						m_iSelectedSkeletonLink = 0;
					else if (m_iSelectedSkeletonLink>=(int)m_pSolver->m_Links.size())
						m_iSelectedSkeletonLink = ((int)m_pSolver->m_Links.size())-1;
					SkeletonLink * link = m_pSolver->m_Links[m_iSelectedSkeletonLink];
					assert(link->m_pParent!=NULL);
					std::cout<<"Selected: "<<link->m_pParent->m_sJointName.c_str()<<" -> "<<link->m_pChild->m_sJointName.c_str();
					std::cout<<" Type: "<<SkeletonLink::Joint2Str(link->m_JointType)<<std::endl;
				}
				else if (key=='j' || key=='J' || key=='k' || key=='K' || key=='l' || key=='L')
				{
					if (m_iSelectedSkeletonLink>=0 && m_iSelectedSkeletonLink < (int)m_pSolver->m_ParamOffsetByLinkID.size())
					{
						assert(m_iSelectedSkeletonLink < (int)m_pSolver->m_Links.size());
						SkeletonLink * link = m_pSolver->m_Links[m_iSelectedSkeletonLink];
						assert(link->m_pParent!=NULL);
						int maxoffset = SkeletonLink::NDims(link->m_JointType);
						std::cout<<""<<link->m_pParent->m_sJointName.c_str();
						std::cout<<", "<<SkeletonLink::Joint2Str(link->m_JointType)<<": ";
						int localoffset = 0;
						if (key=='j' || key=='J')
							localoffset = 0;
						else if (key=='k' || key=='K')
							localoffset = 1;
						else if (key=='l' || key=='L')
							localoffset = 2;
						if (localoffset<maxoffset)
						{
							int offset = m_pSolver->m_ParamOffsetByLinkID[m_iSelectedSkeletonLink]+localoffset;
							pose->Param(offset) += 3.14 * 0.05 * ((key=='J'||key=='K'||key=='L')?-1:1);
							std::cout<<"P[ "<<offset<<"] = "<<pose->Param(offset);
						}
						std::cout<<std::endl;
					}
				}
				else if (key=='r' || key=='R')	// rotation
					pose->Param(3) += 3.14 * 0.1 * (key=='R'?-1:1);
				else if (key=='y' || key=='Y')	// rotation
					pose->Param(4) += 3.14 * 0.1 * (key=='Y'?-1:1);
				else if (key=='&' || key=='7')	// translate X
					pose->Param(0) += 0.025 * (key=='&'?-1:1);
				else if (key=='*' || key=='8')	// translate Y
					pose->Param(1) += 0.025 * (key=='*'?-1:1);
				else if (key=='(' || key=='9')	// translate Z
					pose->Param(2) += 0.025 * (key=='('?-1:1);
				else
					std::cout<<"Unknown character: "<<key<<std::endl;
			}
			break;
		case 'n':
			if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size()
				&& m_pSolver!=NULL && m_iSelectedSkeletonLink>=0)
			{
				PntCorrsPose * pose = m_Poses[m_iCurrPose];
				PosePrior * prior = m_pSolver->m_pPrior;
				assert(prior!=NULL);
				PosePointGauss * gauss = NULL;
				if (prior->PriorType()==PosePrior::PRIOR_GaussAngles)	// single mean
				{
					PosePriorGauss * gaussPrior =(PosePriorGauss *)prior;
					gauss = gaussPrior->m_AngleDistributions[m_iSelectedSkeletonLink];
				}
				else if (prior->PriorType()==PosePrior::PRIOR_MultiGauss)	// multiple means
				{
					m_iSelectedGaussianMean++;
					PosePriorMixGauss * mixGaussPrior = (PosePriorMixGauss*)prior;
					MultiGauss * mg = mixGaussPrior->GetDistr(m_iSelectedSkeletonLink);
					if (m_iSelectedGaussianMean < 0 || m_iSelectedGaussianMean >= mg->NGaussians())
						m_iSelectedGaussianMean = 0;
					std::cout<<"Selected mean: "<<m_iSelectedGaussianMean<<" / "<<mg->NGaussians()<<": "<<std::flush;
					if (m_iSelectedGaussianMean < mg->NGaussians())
						gauss = mg->GetGaussian(m_iSelectedGaussianMean);
				}
				if (gauss!=NULL)
				{
					int offset = m_pSolver->m_ParamOffsetByLinkID[m_iSelectedSkeletonLink];
					std::cout<<"Var["<<gauss->m_iNDims<<"]: "<<std::flush;
					for (int d=0; d<gauss->m_iNDims; d++)
					{
						pose->Param(offset + d) = gauss->m_aMeans[d];
						std::cout<<gauss->m_aVariances[d]<<" "<<std::flush;
					}
					std::cout<<std::endl;
					m_pSolver->ForwardKinematics(pose);
				}
				else
				{
					std::cout<<"[WARNING] There is no prior for the selected link"<<std::endl;
					break;
				}
			}
			else
			{
				std::cout<<"[WARNING] Pose or link are not selected -- cannot move to next prior"<<std::endl;
			}
			break;
		case 'p':	// print points
			PrintPoints();
			break;
		case 'v':
			if (m_pSolver->m_pPrior!=NULL && m_pSolver->m_pPrior->PriorType() == PosePrior::PRIOR_MultiGauss)
			{
				PosePriorMixGauss * prior = (PosePriorMixGauss*)m_pSolver->m_pPrior;
				std::cout<<prior->GetStr().c_str()<<std::endl;
			}
			else
				std::cout<<"[WARNING] Unknown prior"<<std::endl;
			break;
		case 'e':
		case 'E':
			if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size() && m_pSolver!=NULL)
			{
				PntCorrsPose * pose = m_Poses[m_iCurrPose];
				PosePrior * prior = m_pSolver->m_pPrior;
				VKString str = BindingEnergy::GetIKOnlyEnergyStr(pose, m_pSolver, prior, key=='E');
				std::cout<<"Energy "<<m_iCurrPose<<": "<<std::endl;
				std::cout<<str.c_str()<<std::endl;
			}
			break;
		case 'g':
		case 'G':
			m_fGroundTranslate += (key == 'g' ? -1 : 1) * (0.05);
			std::cout<<"Ground Plane: "<<m_fGroundTranslate<<std::endl;
			UpdatePoseScales();
			break;
		case 'm':
		case 'M':	// rotate around x
		{
			ShapeMesh * mesh = (ShapeMesh*)m_pDB->GetModel(0)->GetGeomShape(Model3D::m_sMainGeomName)->SafeCastToMesh();
			R3Point centroid = mesh->Centroid();
			double radius = mesh->AverageRadius(centroid);
						
			R3Affine canonical = R3identity_affine;
			canonical.Translate(R3Vector(centroid.X(), centroid.Y(), centroid.Z()));
			double rot = (key=='m' ? 1 : -1)*0.05;
			canonical.Rotate(R3Vector(0,rot,0));
			canonical.Translate(-R3Vector(centroid.X(), centroid.Y(), centroid.Z()));
			
			for (int i=0; i<mesh->NVertices(); i++)
			{
				R3Point p = mesh->VertexPosition(i);
				canonical.Apply(p);
				mesh->SetVertexPosition(i, p );
			}
			
			break;
		}
		case '[':
		case ']':	// rotate around up
			m_fAngle += (key == '[' ? -1 : 1) * (vkPI / 10.);
			std::cout<<"Angle: "<<m_fAngle<<std::endl;
			UpdatePoseScales();
			break;
		case '=':
		case '-':	// scale object
			m_fScale += (key == '-' ? -1 : 1) * (.05);
			UpdatePoseScales();
			std::cout<<"Scale: "<<m_fScale<<std::endl;
			break;
		case 27:	// exit
			exit(0);
			break;
		case 'w':	// write file and exit
			if (WritePosesFile())
				exit(0);
			else
				std::cout<<"Missing some points..."<<std::endl;
			break;
	}
	return true;
}

bool CreateGTApp::Draw(DatabaseVisWindow * window)
{
	double br = 0, bg = 0, bb = 0;
	window->GetBackgroundColor(&br, &bg, &bb);
	glClearColor(br, bg, bb, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	window->LoadStandardSceneProjectionCamera();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	VisModel3D * mvis = NULL;
	if (window->GetDrawableModels().size()>0)
		mvis = window->GetDrawableModels()[0];
	else
	{
		mvis = window->GetEmptyWindowXForm(0);
		static bool init = false;
		int cid = mvis->SubwindowCameraID(0, window);
		mvis->GetCameraTransform(cid);
		if (!init)
			mvis->Scale(cid, -0.5, window);
		init = true;
	}
	R3Affine camera = mvis->GetCameraTransform(mvis->SubwindowCameraID(0, window));
	
//	static bool firstRescale = false;
//	if (!firstRescale)
//	{
//		window->InitializeGL();
//		window->GetDrawableModels()[0]->Scale(0, -.5, window);
//		firstRescale = true;
//		camera = mvis->GetCameraTransform(mvis->SubwindowCameraID(0, window));
//	}
	
	// render ground plane
	glColor3d(.95, .95, 1.);
	glDisable(GL_LIGHTING);
	glBegin(GL_QUADS);
	double size = 10, height = 0;
	if (IsFlagSet(VIS_ScreenshotMode))
		size = 2;
	int grid = 20;
	double w = 0.5/(double)grid *size;
	for (int x=0; x<grid; x++)
	for (int y=0; y<grid; y++)
	{
		double cx = ((double)x / (double)grid - 0.5) * size;
		double cy = ((double)y / (double)grid - 0.5) * size;
		R3Point p1(cx-w, cy-w, -height);		camera.Apply(p1);		glVertex3d(p1.X(), p1.Y(), p1.Z());
		R3Point p2(cx-w, cy+w, -height);		camera.Apply(p2);		glVertex3d(p2.X(), p2.Y(), p2.Z());
		R3Point p3(cx+w, cy+w, -height);		camera.Apply(p3);		glVertex3d(p3.X(), p3.Y(), p3.Z());
		R3Point p4(cx+w, cy-w, -height);		camera.Apply(p4);		glVertex3d(p4.X(), p4.Y(), p4.Z());
	}
	glEnd();
	
	// render ground plane buffer
	window->StartDrawToBuffer();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	glColor3d(0., 0., 0.);
	glBegin(GL_QUADS);
	for (int x=0; x<grid; x++)
	for (int y=0; y<grid; y++)
	{
		double cx = ((double)x / (double)grid - 0.5) * size;
		double cy = ((double)y / (double)grid - 0.5) * size;
		R3Point p1(cx-w, cy-w, -height);		camera.Apply(p1);		glVertex3d(p1.X(), p1.Y(), p1.Z());
		R3Point p2(cx-w, cy+w, -height);		camera.Apply(p2);		glVertex3d(p2.X(), p2.Y(), p2.Z());
		R3Point p3(cx+w, cy+w, -height);		camera.Apply(p3);		glVertex3d(p3.X(), p3.Y(), p3.Z());
		R3Point p4(cx+w, cy-w, -height);		camera.Apply(p4);		glVertex3d(p4.X(), p4.Y(), p4.Z());
	}
	glEnd();
	m_pGroundPlaneBuffer = &(window->EndDrawToBuffer());
	window->LoadCurrentBuffer();

	// render shape and skeleton
	R3Affine canonical = R3identity_affine;
	canonical.Scale(m_fScale);
	canonical.Rotate(R3Vector(0,0,m_fAngle));
	canonical.Translate(R3Vector(0, 0, -m_fGroundTranslate));
	
	R3Affine xform = R3identity_affine;
	xform.Transform(camera);
	xform.Transform(canonical);
	m_CameraXForm = xform;
	
	
	const ShapeMesh * mesh = NULL;
	if (m_pDB->NumModels()>0)
		mesh = m_pDB->GetModel(0)->GetGeomShape(Model3D::m_sMainGeomName)->SafeCastToMesh();
	//R3Point centroid = mesh->Centroid();
	//double radius = mesh->AverageRadius(centroid);
	window->LoadStandardSceneProjectionCamera();
	
	GLfloat surfaceMaterial[4];
	
	// draw correspondence points + skeleton
	if (IsFlagSet(VIS_PosesPnts) && m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size())
		DrawPosePnts(window, m_Poses[m_iCurrPose], xform, camera);
	if (IsFlagSet(VIS_PosesPredictedPnts) && m_iCurrPredictedPose>=0 && m_iCurrPredictedPose<(int)m_Predictions.size())
		DrawPosePnts(window, m_Predictions[m_iCurrPredictedPose], xform, camera);
	bool done = true;
	if (IsFlagSet(VIS_GTSkeleton) && m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size())
		done = done && DrawSkeleton(window, m_Poses[m_iCurrPose], camera);
	if (IsFlagSet(VIS_PosesPredictedPnts) && m_iCurrPredictedPose>=0 && m_iCurrPredictedPose<(int)m_Predictions.size())
		done = done && DrawSkeleton(window, m_Predictions[m_iCurrPredictedPose], camera);

	std::vector<RenderableFace> faces;
	
	if (IsFlagSet(VIS_Model) && mesh!=NULL)
	{
		std::vector<GeomShape*> meshlist;	meshlist.push_back((GeomShape*)mesh);	std::vector<int> ids;	ids.push_back(0);
		window->LoadStandardSceneProjectionCamera();
		mvis->DrawIsectBuffers(window, 0, meshlist, ids, &xform);
		
		window->LoadStandardSceneProjectionCamera();
		// triangles
		window->StandardSurfaceColor(surfaceMaterial);
		if (mesh->NFaces()>0)
		{
			glDisable(GL_CULL_FACE);
			glEnable(GL_NORMALIZE);
			glEnable(GL_RESCALE_NORMAL);
			glEnable(GL_LIGHTING);
			glBegin(GL_TRIANGLES);
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial);
			for (int f=0; f<mesh->NFaces(); f++)
			{
				int fid = (int)faces.size();
				faces.push_back(RenderableFace());
				
				R3Vector n = mesh->FaceNormal(f);
				xform.ApplyInverse(n);
				faces[fid].norm = n;
				faces[fid].r = surfaceMaterial[0];
				faces[fid].g = surfaceMaterial[1];
				faces[fid].b = surfaceMaterial[2];
				faces[fid].a = .25;
				if (!IsFlagSet(VIS_Transparent))
					glNormal3d(n.X(), n.Y(), n.Z());
				for (int v=0; v<3; v++)
				{
					R3Point p = mesh->VertexPosition(mesh->VertexOnFace(f, v));
					//p = (p + (-centroid)) / radius;
					xform.Apply(p);
					faces[fid].vtx[v] = p;
					if (!IsFlagSet(VIS_Transparent))
						glVertex3d(p.X(), p.Y(), p.Z());
				}
			}
			glEnd();
		}
		else
		{
			glDisable(GL_LIGHTING);
			glColor3d(0., 0., 0.);
			glBegin(GL_POINTS);
			for (int v=0; v<mesh->NVertices(); v++)
			{
				R3Point p = mesh->VertexPosition(v);
				xform.Apply(p);
				glVertex3d(p.X(), p.Y(), p.Z());
			}
			glEnd();
		}
	}
	
	if (IsFlagSet(VIS_Reference) && m_pDB->NumModels()>=2)
	{
		glEnable(GL_LIGHTING);
		// reference
		if (m_pDB->NumModels()>1)
		{
			R3Affine refXForm = R3identity_affine;
			refXForm.Translate(R3Vector(3, 0, 0));
			refXForm.Rotate(R3Vector(0, 0, vkPI));
			
			R3Affine finRefXForm = R3identity_affine;
			finRefXForm.Transform(camera);
			finRefXForm.Transform(refXForm);
			
			const ShapeMesh * mesh = m_pDB->GetModel(1)->GetGeomShape(Model3D::m_sMainGeomName)->SafeCastToMesh();
			surfaceMaterial[0] = .5;
			surfaceMaterial[1] = .5;
			surfaceMaterial[2] = 1.;
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial);
			glBegin(GL_TRIANGLES);
			for (int f=0; f<mesh->NFaces(); f++)
			{
				R3Vector n = mesh->FaceNormal(f);
				R3Vector nn(n.X(), n.Z(), n.Y());
				finRefXForm.ApplyInverse(nn);
				nn.Normalize();
				glNormal3d(nn.X(), nn.Y(), nn.Z());
				for (int v=0; v<3; v++)
				{
					R3Point p = mesh->VertexPosition(mesh->VertexOnFace(f, v));
					R3Point np(p.X(), p.Z(), p.Y());
					finRefXForm.Apply(np);
					glVertex3d(np.X(), np.Y(), np.Z());
					
				}
			}
			glEnd();
		}
	}
	
	
	
	if (IsFlagSet(VIS_Axes))
	{
		glDisable(GL_LIGHTING);
		glEnable( GL_LINE_STIPPLE );
		glLineStipple( 3, 0x0101 );
		glLineWidth(3.);
		R3Point p0, p1x, p1y, p1z;
		double s = 3.;
		glBegin(GL_LINES);
		glColor3d(0., 0., 1.);
		p0 = R3Point(0,0,-s);	camera.Apply(p0);	glVertex3d(p0.X(), p0.Y(), p0.Z());
		p1z = R3Point(0,0,s);	camera.Apply(p1z);	glVertex3d(p1z.X(), p1z.Y(), p1z.Z());
		glColor3d(1., 0., 0.);
		p0 = R3Point(-s,0,0);	camera.Apply(p0);	glVertex3d(p0.X(), p0.Y(), p0.Z());
		p1x = R3Point(s,0,0);	camera.Apply(p1x);	glVertex3d(p1x.X(), p1x.Y(), p1x.Z());
		glColor3d(0., 1., 0.);
		p0 = R3Point(0,-s,0);	camera.Apply(p0);	glVertex3d(p0.X(), p0.Y(), p0.Z());
		p1y = R3Point(0,s,0.);	camera.Apply(p1y);	glVertex3d(p1y.X(), p1y.Y(), p1y.Z());
		glEnd();
		glDisable(GL_LINE_STIPPLE);

		window->GetRenderer()->RenderText(p1x, VKString("X"), 1., 0, 0, VisWindowRenderer::TEXT_SIZE_LARGE);
		window->GetRenderer()->RenderText(p1y, VKString("Y"), 0., 1, 0, VisWindowRenderer::TEXT_SIZE_LARGE);
		window->GetRenderer()->RenderText(p1z, VKString("Z"), 0., 0, 1, VisWindowRenderer::TEXT_SIZE_LARGE);
	}
	
	if (IsFlagSet(VIS_Transparent))
		window->GetRenderer()->DrawTransparentFaces(faces, true);
	
	if (!IsFlagSet(VIS_ScreenshotMode))
	{
		VKString text("");
		if (m_iCurrPose>=0)
		{
			text += VKString("Pose ")+VKString::number(m_iCurrPose)+": ";
			if (m_iCurrRegion>=0)
				text += RegionName(m_Poses[m_iCurrPose]->Region(m_iCurrRegion));
		}
		window->GetRenderer()->RenderText(R2Point(-.1, .9), text, 1., 0, 0, VisWindowRenderer::TEXT_SIZE_LARGE);
		if (!done)
			window->GetRenderer()->UpdateScreen();
	}
	return true;
}

void CreateGTApp::DrawPosePnts(DatabaseVisWindow * window, PntCorrsPose * pose, const R3Affine & xform, const R3Affine & camera)
{
	std::map<PntCorrsPose*, int>::iterator iter = m_Pose2ClassifiedPnts.find(pose);
	
	int nregions = (int)pose->NMatches();
	
	std::vector<R3Point> txtPnts;
	VKStringList txtNames;
	
	for (int r=0; r<nregions; r++)
	{
		double size = 1.;
		TargetPrimitive & prim = pose->Prim(r);
		if (prim.IsType(TargetPrimitive::PRIM_None))
			continue;
		
		R3Point pos = prim.GetPosition();
		
		PointType tp = PT_GroundTruth;
		if (iter != m_Pose2ClassifiedPnts.end())
			tp = m_ClassifiedPredicted[iter->second][r];
		
		int rid = pose->Region(r);
		static GLfloat material[4];
		window->MapIntToRadicalColor(rid, material);
		if (tp==PT_Prediction)
			size = .5;
		else if (tp==PT_PredictionOK)
		{	material[0] = 0.; material[1] = 0.; material[2] = 1.;	}
		else if (tp==PT_PredictionTooFar)
		{	material[0] = 1.; material[1] = 0.; material[2] = 0.;	}
		else if (tp==PT_PredictionNotExists)
		{	material[0] = 0.; material[1] = 0.; material[2] = 0.;	}
			
		assert(rid < m_sRegionNames.count());
		
		material[3] = 1.;
		glEnable(GL_LIGHTING);
		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
		

		if (prim.IsType(TargetPrimitive::PRIM_Ground))
			camera.Apply(pos);
		else if (prim.IsType(TargetPrimitive::PRIM_Point))
			xform.Apply(pos);
		else
			assert(false);
		
		R3Sphere(pos, 0.02).Draw();
		
		txtPnts.push_back(pos);
		txtNames.push_back(m_sRegionNames[rid]);
		
	}
	if (IsFlagSet(VIS_PosesPntNames))
	{
		for (int r=0; r<(int)txtPnts.size(); r++)
			window->GetRenderer()->RenderText(txtPnts[r], txtNames[r], 0., .5, 0, VisWindowRenderer::TEXT_SIZE_LARGE);
	}
}

void CreateGTApp::SetSkeleton(Interactor * actor, IKSolver * solver)
{
	m_pActor = actor;
	m_pSolver = solver;
	
	for (int i=0; i<(int)m_Poses.size(); i++)
	{
		if (!m_Poses[i]->IsParamsInit(m_pSolver->NJointParams()))
			m_Poses[i]->InitParams(m_pSolver->NJointParams(), m_pSolver->m_SolverFlags);
	}
}
bool CreateGTApp::DrawSkeleton(DatabaseVisWindow * window, PntCorrsPose * pose, const R3Affine & xform)
{ 
	assert(m_pSolver!=NULL);
	std::vector<R3Point> pnts;
	std::vector<int> edges;
	//m_bRunIK - used to be here...
	bool done = m_pSolver->FillRenderableSkeleton(pose, true, pnts, edges);

	// draw joints
	glEnable(GL_LIGHTING);
	static GLfloat material[4];
	
	int childID = -1;
	if (m_iSelectedSkeletonLink>=0 && m_iSelectedSkeletonLink < (int)edges.size())
		childID = edges[m_iSelectedSkeletonLink*2+1];
	for (int i=0; i<(int)pnts.size(); i++)
	{
		if (childID==i)
		{    material[0] = 0.;    	material[1] = 1.;     	material[2] = 0.;     	material[3] = 1.;}
		else
		{    material[0] = 0.;    	material[1] = 0.;     	material[2] = 1.;     	material[3] = 1.;}
		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
		R3Point pos = pnts[i];
		xform.Apply(pos);
		R3Sphere(pos, 0.01).Draw();
	}
	
	// draw links
	glDisable(GL_LIGHTING);
	glLineWidth(10.);
	glBegin(GL_LINES);
	
	for (int i=0; i<(int)edges.size(); i+=2)
	{
		if (m_iSelectedSkeletonLink==i/2)
			glColor3d(0., 1., 0.);
		else
			glColor3d(1., 0., 0.);
			
		int p1 = edges[i+0];
		int p2 = edges[i+1];
		assert(p1>=0 && p1 < (int)pnts.size() && p2>=0 && p2 < (int)pnts.size());
		R3Point p1pos = pnts[p1];
		R3Point p2pos = pnts[p2];
		xform.Apply(p1pos);
		xform.Apply(p2pos);
		glVertex3d(p1pos.X(), p1pos.Y(), p1pos.Z());
		glVertex3d(p2pos.X(), p2pos.Y(), p2pos.Z());
	}
	glEnd();
	return done;
}

void CreateGTApp::SetScreenshotType(const VKString & sName, int value)
{
	int visFlags = VIS_Model | VIS_PosesPnts | VIS_Transparent;
	visFlags |= VIS_PosesPredictedPnts | VIS_PredictedSkeleton | VIS_GTSkeleton;
	
	if (sName=="plain")
		visFlags = VIS_Model;
	else if (sName=="poses")
		m_iCurrPose = value;
	else if (sName=="predicted_poses")
		m_iCurrPredictedPose = value;
	else
		std::cout<<"[WARNING] CreateGTApp: unknown screenshot type: "<<sName.c_str()<<std::endl;
	
	SetRenderingFlags(visFlags | VIS_ScreenshotMode);
}

int CreateGTApp::NPoses() const
{
	return (int)m_Poses.size();
}

const PntCorrsPose * CreateGTApp::GetPose(int pid) const
{
	return m_Poses[pid];
}

ModelDatabase * CreateGTApp::GetDB()
{
	return m_pDB;
}

double CreateGTApp::GetDistanceScale() const
{
	//std::cout<<"Dist scale: "<<m_fScale<<" / "<<m_fOrigRadius<<" "<<(m_fScale / m_fOrigRadius)<<std::endl;
	return m_fScale;
}

int CreateGTApp::NRegions() const
{
	return m_sRegionNames.count();
}

const VKString & CreateGTApp::RegionName(int rid) const
{
	return m_sRegionNames[rid];
}




