#include "FileManagement.h"
#include "VKString.h"
#include "TimeProfiler.h"

#include "RendererGLUT.h"
#include "RendererOSMesa.h"

#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ShapeDistance.h"

#include "ShapeTriangleSoup.h"
#include "DBApp.h"
#include <vector>
#include "ShapePoint.h"
#include "CreateGTApp.h"
#include "IKSolver.h"
#include "PosePrior.h"
#include "PosePriorGauss.h"
#include "BindingManager.h"

#include "AppScreenshots.h"

#include "InteractorPose.h"
#include "PntCorrsPose.h"


void Evaluate(CreateGTApp * gtApp, const VKString & symmInvFile,
			  const VKString & featPnts, const VKString & predictions, const VKString & outDir);
void Evaluate(CreateGTApp * gtApp, const VKString & evalPose, const VKString & outFile);
double WriteEvalDists(const PntCorrsPose * pose1, const PntCorrsPose * pose2, double distScale,
					  std::ofstream & textStream, std::vector<double> & outDists);

int main(int argc, char** argv )
{
	if (argc<3)
	{
		std::cout<<"Usage: ./"<<argv[0]<<" geomFile posesFile [options]";
		std::cout<<"Params:\n";
		std::cout<<"    -readRegionNames featFile - feature points \n";
		std::cout<<"    -xform xformFile - canonical transformation \n";
		std::cout<<"    -reference human.off - reference human mesh \n";
		std::cout<<"    -outmesh out.off - output mesh \n";
		std::cout<<"    -renderShape outFile - render image\n";
		std::cout<<"    -renderPoses outDir - render images\n";
		std::cout<<"    -readSkeleton skeletonFile - read skeleton (regions names need to be defined) \n";
		std::cout<<"    -interact - open and keep image\n";
		std::cout<<"    -symm_invariant symmFile - evaluate in a symmetry-invariant way\n";
		std::cout<<"    -eval featPnts predictions outdir - predictions at points\n";
		std::cout<<"    -evalpose poseFile outdir - pose predictins at points\n";
		std::cout<<"    -updateGT file - write final gt file\n";
		std::cout<<"    -recomputeIK - recompute inverse kinematics\n";
		std::cout<<"    -normalizedmesh - mesh is pre-normalized\n";
		std::cout<<"    -poseonly\n";
		std::cout<<"    -readPriorAsPose filename - read gaussian prior and take mean as pose\n";
		std::cout<<"    -readPrior filename clampVal variance - read prior (and set clamp values and variance)\n";
		std::cout<<"    -param paramtype p1 p2 ... pN - algorithm params\n";
		std::cout<<"    	wUnassigned value - unassinged points weight\n";
		std::cout<<"    	wEndEffectors value - end effector weight\n";
		std::cout<<"    	wFeatures value - features weight\n";
		std::cout<<"    	wPosePrior value - pose prior weight\n";
		std::cout<<"    	wSymmetryPrior value - symm prior weight\n";
//		std::cout<<"    -evalAgg N dir1 dir2 ... dirN outdir - predictions at points\n";
		exit(0);
	} 
	
	VKString geomFile(argv[1]);
	VKString posesFile(argv[2]);
	VKString readRegionNames = "none";
	VKString xFormFile = "none";
	VKString reference = "none";
	bool interact = false;
	VKString renderShape = "none";
	VKString renderPoses = "none";
	bool poseonly = false;
	
//	VKStringList evalAgg;
	VKString evalFeatPnts = "none";
	VKString evalFeatVals = "none";
	VKString evalOutDir = "none";
	VKString evalPose = "none";
	VKString outMesh = "none";
	
	VKString symmInvariant = "none";
	bool recomputeIK = false;
	VKString updateGTFile = "none";
	
	VKString readSkeleton = "none";
	Interactor * actor = NULL;
	IKSolver * ikSolver = NULL;
	VKString readPriorAsPose = "none";
	VKString readPrior = "none";
	
	double priorClampVal = -1;
	double priorVariance = -1.;
	EnergyParams eparams;
	
	for (int i=3; i<argc; i++)
	{
		if (strcmp(argv[i], "-readRegionNames")==0)
			readRegionNames = argv[++i];
		else if (strcmp(argv[i], "-xform")==0)
			xFormFile = argv[++i];
		else if (strcmp(argv[i], "-outmesh")==0)
			outMesh = argv[++i];
		else if (strcmp(argv[i], "-reference")==0)
			reference = argv[++i];
		else if (strcmp(argv[i], "-interact")==0)
			interact = true;
		else if (strcmp(argv[i], "-updateGT")==0)
			updateGTFile = argv[++i];
		else if (strcmp(argv[i], "-recomputeIK")==0)
			recomputeIK = true;
		else if (strcmp(argv[i], "-normalizedmesh")==0)
			xFormFile = "identity";
		else if (strcmp(argv[i], "-poseonly")==0)
			poseonly = true;
		else if (strcmp(argv[i], "-readPriorAsPose")==0)
			readPriorAsPose = argv[++i];
		else if (strcmp(argv[i+1], "-readRegionStats")==0)
		{ 	eparams.m_sRegionStats = argv[i+2];}
		else if (strcmp(argv[i], "-readPriorClamp")==0)
		{
			readPrior = argv[++i];
			priorClampVal = atof(argv[++i]);
			double maxAngle = atof(argv[++i]);
			if (maxAngle>=0)
				priorVariance = pow(maxAngle, 2.) / (2. * priorClampVal);

		}
		else if (strcmp(argv[i], "-readPrior")==0)
			readPrior = argv[++i];
		else if (strcmp(argv[i], "-readSkeleton")==0)
			readSkeleton = argv[++i];
		else if (strcmp(argv[i], "-renderShape")==0)
			renderShape = argv[++i];
		else if (strcmp(argv[i], "-renderPoses")==0)
			renderPoses = argv[++i];
		else if (strcmp(argv[i], "-eval")==0)
		{
			evalFeatPnts = argv[i+1];
			evalFeatVals = argv[i+2];
			evalOutDir = argv[i+3];
			i+=3;
		}
		else if (strcmp(argv[i], "-evalpose")==0)
		{
			evalPose = argv[i+1];
			evalOutDir = argv[i+2];
			i+=2;
		}
		else if (strcmp(argv[i], "-symm_invariant")==0)
			symmInvariant = argv[++i];
		else if (strcmp(argv[i], "-param")==0)
		{
			if (strcmp(argv[i+1], "wPosePrior")==0)
			{	 eparams.m_wPosePrior = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wEndEffectors")==0)
			{	 eparams.m_wEndEffectorDist = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wSymmetryPrior")==0)
			{	 eparams.m_wSymmetryPrior = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wFeatures")==0)
			{	 eparams.m_wFeatures = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wUnassigned")==0)
			{	 eparams.m_wUnassigned = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wGround")==0)
			{	 eparams.m_wGround = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wContactPairwise")==0)
			{	 eparams.m_wContactPairwise = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wSymmPlane")==0)
			{	 eparams.m_wSymmPlane = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wIsectPenalty")==0)
			{	 eparams.m_wIsectPenalty = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "maxGroundProb")==0)
			{	 eparams.m_fMaxGroundPenalty = -log(atof(argv[i+2]));	i+=2;	}
			else if (strcmp(argv[i+1], "minContactProb")==0)
			{	 eparams.m_fMaxContactPenalty = -log(atof(argv[i+2]));	i+=2;	}
			else
				assert(false);
		}
//		else if (strcmp(argv[i], "-evalAgg")==0)
//		{
//			int ndirs = atoi(argv[i+1]);
//			for (int d=0; d<ndirs; d++)
//				evalAgg.push_back(VKString(argv[i+2+d]));
//			evalOutDir = argv[i+2+ndirs];
//			i += (2+ndirs+1);
//		}
		else
		{
			std::cout<<"[ERROR] Unknown command: "<<argv[i]<<std::endl;
			assert(false);
		}
	}
	
	ModelDatabase * db = new ModelDatabase();
	if (geomFile!="none")
	{
		Model3D * model = Model3D::CreateFromFile(geomFile, ModelDatabase::READ_MESH_AS_SOUP, 0, "mainmesh");
		db->AddModel(model);

	}
	if (reference!="none")
	{
		Model3D * refMesh = Model3D::CreateFromFile(reference, ModelDatabase::READ_MESH_AS_SOUP, 0, "reference");
		db->AddModel(refMesh);
	}

	bool readPose = readPriorAsPose!="none" || FileManagement::FileExists(posesFile.c_str());
	CreateGTApp * gtApp = new CreateGTApp(db, posesFile, xFormFile, outMesh, readPose);
	if (readRegionNames!="none")
		gtApp->SetRegionsFile(readRegionNames);
	
	if (readSkeleton!="none")
	{
		assert(readRegionNames!="none");	// read region names first
		actor = new Interactor(Interactor::ACT_Actor, Interactor::TYPE_PntCorrs);
		PosePrior * prior = NULL;
		if (readPrior!="none")
		{
			prior = PosePrior::ReadPrior(readPrior);
			prior->ReadPointStats(readPrior+".pntstats.txt");
		}
		
		if (priorClampVal>=0 || priorVariance>=0)
		{
			assert(prior!=NULL && prior->PriorType()==PosePrior::PRIOR_GaussAngles);
			PosePriorGauss * gauss = (PosePriorGauss*)prior;
			gauss->SetParams(priorClampVal, priorVariance);
		}
		
		assert(prior!=NULL || (eparams.m_wPosePrior==0 && eparams.m_wSymmetryPrior==0));
		
		ikSolver = new IKSolver(eparams.m_wEndEffectorDist, eparams.m_wPosePrior, eparams.m_wSymmetryPrior, prior);
		ikSolver->ReadSkeleton(readSkeleton, readRegionNames);
		if (prior!=NULL && prior->IsAngles())
			((PriorAngles*)prior)->SetIK(ikSolver);
		gtApp->SetSkeleton(actor, ikSolver);
	}
	
	if (readPriorAsPose!="none")
		gtApp->ReadPriorAsPose(readPriorAsPose);
		
	gtApp->ReadPosesFile();
	int listFlags = DatabaseVisWindow::LISTGP_Models | DatabaseVisWindow::LISTGP_Geoms;
	listFlags |= DatabaseVisWindow::LISTGP_MeshFlags;
	
	int w=-1, h=-1;
	if (renderShape!="none" || renderPoses!="none")
	{  w=1600; h=1200; }

	DatabaseVisWindow * dbVis = new DatabaseVisWindow(db, NULL, listFlags, 1, 1, 0, 0, 1., w, h);

	if (evalFeatVals!="none")
		Evaluate(gtApp, symmInvariant, evalFeatPnts, evalFeatVals, evalOutDir);
	else if (evalPose!="none")
		Evaluate(gtApp, evalPose, evalOutDir);
	
	if (recomputeIK)
	{
		assert(ikSolver!=NULL);
		std::vector<R3Point> pnts;
		std::vector<int> edges;
		int success = 0;
		for (int i=0; i<(int)gtApp->m_Poses.size(); i++)
			success += ikSolver->SolveIKProblem(gtApp->m_Poses[i], true);
		for (int i=0; i<(int)gtApp->m_Predictions.size(); i++)
			success += ikSolver->SolveIKProblem(gtApp->m_Predictions[i], true);
		if (success < (int)(gtApp->m_Poses.size()+gtApp->m_Predictions.size()))
		{
			std::cout<<"[WARNING] ";
			std::cout<<"IK converged for "<<success<<" / "<<(gtApp->m_Poses.size()+gtApp->m_Predictions.size())<<std::endl;
		}
	}
	
	if (updateGTFile!="none")
	{
		gtApp->m_sWriteFile = updateGTFile;
		gtApp->WritePosesFile();
	}
	
#ifdef VISUALIZATION
	if (renderShape!="none" || renderPoses!="none")
	{
		if (interact)
			std::cout<<"[WARNING] Will exit after screenshots..."<<std::endl;

		AppScreenshots * shots = new AppScreenshots(true);
		if (renderShape!="none")
			shots->TakeShotOfApp(gtApp, "plain", 0, renderShape);
		if (renderPoses!="none")
		{
			VKString shotStr = "poses";
			int nposes = gtApp->NPoses();
			
			if (gtApp->NPredictions()>0)
			{
				nposes = gtApp->NPredictions();
				shotStr = "predicted_poses";
			}
			for (int i=0; i<nposes; i++)
				shots->TakeShotOfApp(gtApp, shotStr, i, renderPoses+"/"+VKString::number(i)+".jpg");
		}
		dbVis->SetApp(shots);
#ifdef MESA_RENDERING
		VisWindowRenderer * mesa = new RendererOSMesa(dbVis);
		mesa->UpdateScreen();
#else
		VisWindowRenderer * wndGLUT = new RendererGLUT(argc, argv, dbVis);
#endif
	}
	else if (interact)
	{
		dbVis->SetApp(gtApp);
		VisWindowRenderer * wndGLUT = new RendererGLUT(argc, argv, dbVis);
	}
#else
	std::cout<<"[WARNING] Visualization disabled (compiling flag)"<<std::endl;
#endif
	return 0;
}


void Evaluate(CreateGTApp * gtApp, const VKString & symmInvFile,
			  const VKString & featPnts, const VKString & predictions, const VKString & outDir)
{
	std::vector<int> region2equivclass;
	// check is symmetry-invariant
	std::map<VKString, int> name2equivclass;
	int equivclass = 0;
	if (symmInvFile!="none")
	{
		std::ifstream textStream(symmInvFile.c_str());
		assert(textStream.is_open());
		VKString tempStr;
		while (!textStream.eof())
		{
			tempStr.readLine(textStream);
			VKStringList spLine = tempStr.split(" ");
			for (int i=0; i<spLine.count(); i++)
			{
				if (spLine[i].length()>1)
					name2equivclass[spLine[i]] = equivclass;
			}
			equivclass++;
		}
		textStream.close();
	}
	for (int i=0; i<gtApp->NRegions()+1; i++)
	{
		std::map<VKString, int>::iterator iter = name2equivclass.end();
		if (i < gtApp->NRegions())
			iter = name2equivclass.find(gtApp->RegionName(i));
		
		if (iter != name2equivclass.end())
			region2equivclass.push_back(iter->second);
		else
			region2equivclass.push_back(equivclass++);
	}
	
	// read points
	GeomShape * shape = gtApp->GetDB()->GetModel(0)->GetGeomShape(Model3D::m_sMainGeomName);
	assert(shape!=NULL && shape->SafeCastToMesh()!=NULL);
	ShapePointCloud * pnts = new ShapePointCloud(shape->SafeCastToMesh(), "pnts");
	pnts->ReadGeometryFile(featPnts);
	 
	// read predictions
	ShapeValue * vals = new ShapeValue(pnts);
	vals->ReadShapeValues(predictions);
	
	// normalizations...
	double minval, maxval;
//	double globalMin=-1, globalMax=-1;
	std::vector<bool> exists;
	for (int rid=0; rid<gtApp->NRegions()+1; rid++)
	{
		minval = -1;
		maxval = -1;
		for (int pid=0; pid<pnts->NumPoints(); pid++)
		{
			double v = vals->Value(pnts->GetPoint(pid), rid);
			if (pid==0 || minval > v)
				minval = v;
			if (pid==0 || maxval < v)
				maxval = v;
		}
		if (minval==maxval)
			exists.push_back(false);
		else
			exists.push_back(true);		
	}
	
//	assert(globalMin<globalMax);
	// for each region - create 2D grid
	assert(vals->NDims()==gtApp->NRegions()+1);
	double distScale = gtApp->GetDistanceScale();
	for (int rid=0; rid<gtApp->NRegions()+1; rid++)
	{
		if (!exists[rid])
			continue;
		
		VKString rname = rid < gtApp->NRegions() ? gtApp->RegionName(rid) : "Other";
		VKString outFile = outDir+"/"+rname+".eval_pairs";
		FileManagement::CreateDirsOnPathToFile(outFile.c_str());
		std::ofstream textStream(outFile.c_str());
		assert(textStream.is_open());
		
		//std::cout<<"RName: "<<rname.c_str()<<" "<<pnts->NumPoints()<<std::endl;
		for (int pid=0; pid<pnts->NumPoints(); pid++)
		{
			const ShapePoint & pnt = pnts->GetPoint(pid);
			double v = vals->Value(pnt, rid);
			//std::cout<<pid<<" distscale: "<<distScale<<" value = "<<v<<std::endl;
			
			std::vector<double> dist;
			int poseid = 0;
//			for (int poseid = 0; poseid < gtApp-> NPoses(); poseid++)
			{
				const PntCorrsPose * pose = gtApp->GetPose(poseid);
				assert(poseid==0);	// TODO: need to know which pose dist relates to... (change PR curves as well)
				for (int poseregion = 0; poseregion < (int)pose->NMatches(); poseregion++)
				{
					if (region2equivclass[pose->Region(poseregion)]==region2equivclass[rid])
					{
						R3Point gtpos = pose->Prim(poseregion).GetPosition();
						double currdist = R3Distance(gtpos, pnt.GetPosition()) * distScale;
						dist.push_back(currdist);
					}
				}
			}
			
			textStream<<v<<" ";//<<dist<<"\n";
			for (int d=0; d<(int)dist.size(); d++)
				textStream<<dist[d]<<" ";
			textStream<<"\n";
		}
		textStream.close();
	}
}


double WriteEvalDists(const PntCorrsPose * pose1, const PntCorrsPose * pose2, double distScale,
					  std::ofstream & textStream, std::vector<double> & outDists)
{
	double dist = 0;
	int counter = 0;
	assert(distScale>=0);
	for (int i=0; i<(int)pose1->NMatches(); i++)
	{
		double foundMinDist = -1;
		for (int j=0; j<(int)pose2->NMatches(); j++)
		{
			if (pose1->Region(i) == pose2->Region(j))
			{
				R3Point predPos = pose1->Prim(i).GetPosition();
				R3Point gtPos = pose2->Prim(j).GetPosition();
				double currdist = R3Distance(predPos, gtPos) * distScale;
				textStream<<currdist<<" ";
				foundMinDist = currdist;
				dist += currdist;
				counter++;
			}
		}
		outDists.push_back(foundMinDist);
	}
	return dist / counter;
}

VKString Pad(const VKString & str)
{
	int trgLen = 15;
	return str.CreateStrWithPadding(trgLen, " ");
}

VKString Pad(double val)
{
	return Pad(VKString::number(val));
}

void Evaluate(CreateGTApp * gtApp, const VKString & evalPose, const VKString & outFile)
{
	const ShapeMesh * mesh = gtApp->GetDB()->GetModel(0)->GetGeomShape(Model3D::m_sMainGeomName)->SafeCastToMesh();
	IKSolver * ik = gtApp->m_pSolver;
	
	// read predicted poses
	//PntCorrsPose::CreatePosesFromGTFile(evalPose, mesh, "none", predictedPoses);
	assert(gtApp->m_pSolver!=NULL);
	Interactor * object = new Interactor((ShapeMesh*)mesh, Interactor::ACT_Object);
	BindingManager * bindings = BindingManager::ReadBindings(evalPose, gtApp->m_pActor, object);
	 
	// write file header
	FileManagement::CreateDirsOnPathToFile(outFile.c_str());
	std::ofstream textStream(outFile.c_str());
	assert(textStream.is_open());
//	assert(gtApp->NPoses()==1);
//	textStream<<"NPoses "<<bindings->NBindings()<<"\n";
	textStream<<"NPoses 1\n";	// just one for now
	textStream<<"NJoints "<<ik->m_Joints.size()<<"\n";
	
	// map joints to region (to know whether a point can be a contact point)
	std::vector<int> joint2region;
	for (int i=0; i<(int)ik->m_Joints.size(); i++)
		joint2region.push_back(-1);
	for (int i=0; i<(int)ik->m_Pnts2Skeleton.size(); i++)
	{
		assert(ik->m_Pnts2Skeleton[i] < (int)joint2region.size());
		joint2region[ik->m_Pnts2Skeleton[i]] = i;
	}

	double scale = gtApp->GetDistanceScale();
	assert(scale==1);	// ????

	VKStringList writeBestList;
	double bestError = RN_INFINITY;
	int bestgtid = -1;
	for (int gtpid=0; gtpid<gtApp->NPoses(); gtpid++)
	{
		double error = 0;
		// get ground truth pose
		PntCorrsPose * gtpose = (PntCorrsPose*)gtApp->GetPose(gtpid);
		ik->ForwardKinematics(gtpose);
		
		VKString writeStr = "";
		// for each predicted pose -- evaluate error
		//for (int i=0; i<bindings->NBindings(); i++)
		int i = 0;
		{
			// get predicted pose
			writeStr += VKString("PredictedPose ")+VKString::number(i)+"\n";
			const InteractorPose * predtmp = bindings->GetBinding(i)->GetPose(Interactor::ACT_Actor);
			PntCorrsPose * predicted = (PntCorrsPose*)predtmp;
			ik->ForwardKinematics(predicted);
			
			// go over all possible regions (compare ground truth and predicted)
			for (int j=0; j<(int)ik->m_Joints.size(); j++)
			{
				char gtPntType = 'x';	 // j -- joint, r -- region of interest, s -- surface contact, g -- ground contact
				char predPntType = 'x';  // j -- joint, r -- region of interest, s -- surface contact, g -- ground contact
				double dist = -1.;
				
				// estimate distance
				R3Point gtpos = gtpose->Position(j);
				R3Point predictpos = predicted->Position(j);
				dist = R3Distance(gtpos, predictpos);
				
				int rid = joint2region[j];
				
				// classify ground truth point
				if (rid<0)
					gtPntType ='j';
				else
				{
					error += dist;
					gtPntType = 'r';
					for (int r=0; r<(int)gtpose->NMatches() && gtPntType=='r'; r++)
					{
						if (gtpose->Region(r)==rid)
						{
							if (gtpose->Prim(r).IsType(TargetPrimitive::PRIM_Ground))
								gtPntType = 'g';
							else if (gtpose->Prim(r).IsType(TargetPrimitive::PRIM_Point))
								gtPntType = 's';
							else
								assert(false);
						}
					}
				}
				
				// classify predicted point
				if (rid<0)
					predPntType ='j';
				else
				{
					predPntType = 'r';
					for (int r=0; r<(int)predicted->NMatches() && predPntType=='r'; r++)
					{
						if (predicted->Region(r)==rid)
						{
							if (predicted->Prim(r).IsType(TargetPrimitive::PRIM_Ground))
								predPntType = 'g';
							else if(predicted->Prim(r).IsType(TargetPrimitive::PRIM_Point))
								predPntType = 's';
							else
								assert(false);
						}
					}
				}
								
				writeStr+=VKString::number(rid)+" "+VKString(gtPntType)+" "+VKString(predPntType)+" "+VKString::number(dist)+"\n";
			}
			
			// go over angles -- compare ground truth and real
			assert(gtpose->IsParamsInit(ik->NJointParams()) && predicted->IsParamsInit(ik->NJointParams()));
			writeStr+=VKString("NSkeletonParams ")+VKString::number(ik->NJointParams())+" "+VKString::number(ik->m_iNInitSkeletonParams)+"\n";
			for (int a=0; a<ik->NJointParams(); a++)
			{
				double gta = gtpose->Param(a);
				double pra = predicted->Param(a);
				double angle = PosePointGauss::GetNearestAngle(gta, pra);
				writeStr+=VKString::number(vkAbs(gta-angle))+" ";
			}
			writeStr+="\n";
		}
		writeBestList.push_back(writeStr);
		if (error < bestError || bestgtid==-1)
		{
			bestgtid = gtpid;
			bestError = error;
		}
	}
	assert(bestgtid>=0 && bestgtid<writeBestList.count());
	textStream<<writeBestList[bestgtid].c_str();
	textStream.close();

	
// TODO: visualize predictions vs ground truth poses...
//		gtApp->m_ClassifiedPredicted.push_back(std::vector<CreateGTApp::PointType>());
//		for (int r=0; r<(int)bestDists.size(); r++)
//		{
//			assert(r<(int)bestDists.size());
//			if (bestDists[r]<0)
//				gtApp->m_ClassifiedPredicted[i].push_back(CreateGTApp::PT_PredictionNotExists);
//			else if (bestDists[r] < acceptThreshold)
//				gtApp->m_ClassifiedPredicted[i].push_back(CreateGTApp::PT_PredictionOK);
//			else
//				gtApp->m_ClassifiedPredicted[i].push_back(CreateGTApp::PT_PredictionTooFar);
//		}
//	}
//	gtApp->SetPredictions(predictedPoses);
}




