#include "Classifier.h"
#include "ShapeDistance.h"
#include "VKGraph.h"
#include "DistGeodPnts.h"

TrainingParams::TrainingParams()
{
	m_Classifier = CLA_None;
	m_TrainingData = TRAIN_GaussianAll;
	m_fDistThreshold = -1;
	m_fDist2SigmaScale = -1;
	m_fPositiveValCutOff = -1;
	m_fRatioRandomTrainingData = -1;
}

TrainingDataChoice TrainingParams::Str2TrainDataSetting(const VKString & str)
{
	if (str=="GaussianAll")
		return TRAIN_GaussianAll;
	else if (str=="GaussianRandomAndLabeled")
		return TRAIN_GaussianRandomAndLabeled;
	else if (str=="GaussianRandom")
		return TRAIN_GaussianRandom;
	else if (str=="GaussianLabeledOnly")
		return TRAIN_GaussianLabeledOnly;
	assert(false);
	return TRAIN_GaussianAll;
}

VKString TrainingParams::TrainDataChoice2Str(TrainingDataChoice dataChoice)
{
	switch(dataChoice)
	{
		case TRAIN_GaussianAll:
			return "GaussianAll";
		case TRAIN_GaussianRandomAndLabeled:
			return "GaussianRandomAndLabeled";
		case TRAIN_GaussianRandom:
			return "GaussianRandom";
		case TRAIN_GaussianLabeledOnly:
			return "GaussianLabeledOnly";
	}
	assert(false);
	return "Unknown";
}


VKString Classifier::Type2Str(ClassifierType type)
{
	if (type==CLA_None)
		return "None";
	else if (type==CLA_Dist2Nearest)
		return "Dist2Nearest";
	else if (type==CLA_Dist2NearestGauss)
		return "Dist2NearestGauss";
	else if (type==CLA_JointBoost)
		return "JointBoost";
	else if (type==CLA_Weka)
		return "Weka";
	else if (type==CLA_External)
		return "External";
	else if (type==CLA_GroupWeka)
		return "GroupWeka";
	else
		assert(false);
}

ClassifierType Classifier::Str2Type(const VKString & type)
{
	if (type=="None")
		return CLA_None;
	else if (type=="Dist2Nearest")
		return CLA_Dist2Nearest;
	else if (type=="Dist2NearestGauss")
		return CLA_Dist2NearestGauss;
	else if (type=="JointBoost")
		return CLA_JointBoost;
	else if (type=="Weka")
		return CLA_Weka;
	else if (type=="External")
		return CLA_External;
	else if (type=="GroupWeka")
		return CLA_GroupWeka;
	else
		assert(false);
}


Classifier::Classifier()
{
	m_pDB = new ModelDatabase();
	m_Type = CLA_None;
	m_sMetricName = ShapeDistance::m_sDefaultEuclidean;
	m_Timer.startedProcess("Classifier");
	m_pSolver= NULL;
}

void Classifier::SetIKSolver(IKSolver * solver)
{
	m_pSolver = solver;
}

ModelDatabase * Classifier::GetDB()
{
	return m_pDB;
}

// main algorithms:
void Classifier::Train(const TrainingParams & params)
{
	m_Type = params.m_Classifier;
	//std::cout<<"Train: "<<m_Type<<std::endl;
	if (m_Type!=CLA_GroupWeka)
		FillPerRegionStats();
	
	bool DEBUG_TRUEVALS = false;
	
	if (m_Type==CLA_Dist2Nearest || m_Type==CLA_Dist2NearestGauss)
	{
		assert(false);	// depricated
	}
	else if (m_Type==CLA_Weka)
	{
		std::cout<<"Creating WEKA training data: "<<std::endl;
		// fill ground truth predicted values
		for (int i=0; i<m_sRegionNames.count(); i++)
			FillOnSurfaceMembershipValue(i, params);
		
		// set ground truth values for debugging
		if (DEBUG_TRUEVALS)
		{
			for (int i=0; i<(int)m_Poses.size(); i++)
			{
				Model3D * m = m_pDB->GetModel(i);
				GeomShape * shape = m->GetGeomShape("featpnts"); assert(shape!=NULL);
				ShapeValue * gtPrediction = new ShapeValue(shape, "gtregions", m_sRegionNames.count());
				ShapePointCloud * pnts = shape->SafeCastToPointCloud(); assert(pnts!=NULL);
				shape->AddShapeVal("gtregions", gtPrediction);
				gtPrediction->InitCache(ShapeValue::SVTYPE_PerPointSet);
				assert(i<(int)m_apGTPredictions.size());
				gtPrediction->SetPntsValues(m_apGTPredictions[i], pnts->NumPoints()*m_sRegionNames.count());
			}
		}
		
		// find symmetric groups
		std::vector<std::vector<int> > symmgroups;
		for (int i=0; i<m_sRegionNames.count(); i++)
		{
			m_TrainingData.push_back(NULL);
			symmgroups.push_back(std::vector<int>());
		}
		for (int i=0; i<m_sRegionNames.count(); i++)
		{
			if (NPositiveExamples(i)>0)
			{
				int symmid = GetSymmEqClass(i, NULL);
				if (m_TrainingData[symmid]==NULL)
					m_TrainingData[symmid] = new TrainingData();
				symmgroups[symmid].push_back(i);
			}
		}
		
		// fill data for symmetric groups
		int npositive, nnegative;
		for (int i=0; i<m_sRegionNames.count(); i++)
		{
			bool mainName = false;
			int symmid = GetSymmEqClass(i, &mainName);
			if (NPositiveExamples(i)>0 && mainName)
			{
				assert(m_TrainingData[symmid]!=NULL && symmid<(int)symmgroups.size() && symmgroups[symmid].size()>0);
				AddTrainingExamples(*(m_TrainingData[symmid]), symmid, symmgroups[symmid], params, npositive, nnegative);
				std::cout<<"    "<<m_sRegionNames[i].c_str()<<" ("<<npositive<<" : "<<nnegative<<")"<<std::endl;
			}
		}
	}
	else if (m_Type==CLA_GroupWeka)
	{
		assert(m_Data2File.find("classifier")!=m_Data2File.end());
		VKString outDir = m_Data2File["classifier"];
		
		double totalTime = 0;
		std::map<VKString, std::vector<double*> > region2data;
		int datasize=0;
		
		m_iNPoses = 0;
		m_RegionNGround.clear();
		m_RegionNSurface.clear();
		for (int i=0; i<m_sRegionNames.count(); i++)
		{
			m_RegionNGround.push_back(0);
			m_RegionNSurface.push_back(0);
		}
		
		bool ok;
		bool trainingdataexists = false;
		VKString megaHeader = "";
		TimeProfiler timer;
		std::cout<<"Reading & Aggregating data: "<<(outDir+"/permodel/").c_str()<<std::endl;
		for (int i=0; i<m_pDB->NumModels(); i++)
		{
			if (i%10==0)
				timer.WriteProgress("claprog", i, m_pDB->NumModels());
			VKString objName = m_pDB->GetModel(i)->ModelName();
			VKString inDir = outDir+"/permodel/"+objName+"/";
			
			if (!FileManagement::DirExists(inDir.c_str()))
				inDir = outDir+"/../permodel/"+objName+"/";
			if (!FileManagement::DirExists(inDir.c_str()))
			{
				std::cout<<"[ERROR] Cannot find per model data: "<<inDir.c_str()<<std::endl;
				assert(FileManagement::DirExists(inDir.c_str()));
			}
			std::ifstream inInfoStream((inDir+"/info.txt").c_str());
			std::ifstream inRegionStatsStream((inDir+"/regionstats.txt").c_str());

			if (!(inInfoStream.is_open() && inRegionStatsStream.is_open()))
				std::cout<<"[ERROR] Cannot find files: "<<inDir.c_str()<<"/[info.txt, regionstats.txt]"<<std::endl;
			assert(inInfoStream.is_open() && inRegionStatsStream.is_open());
			
			VKString tempStr;
			tempStr.readLine(inInfoStream);     assert(tempStr=="Type Weka"||tempStr=="Type External");
			tempStr.readToken(inInfoStream);	assert(tempStr=="Timing");
			tempStr.readToken(inInfoStream);
			tempStr = tempStr.replace("s", "");
			totalTime += tempStr.toDouble(&ok);
			assert(ok);
			
			int nposes=0;
			int nregions=0;
			tempStr.readToken(inRegionStatsStream);	assert(tempStr=="NPoses");
			inRegionStatsStream>>nposes;
			m_iNPoses+=nposes;
			tempStr.readToken(inRegionStatsStream);	assert(tempStr=="NRegions");
			inRegionStatsStream>>nregions;
			
			int count;
			tempStr.readToken(inRegionStatsStream); assert(tempStr=="Ground");
			for (int r=0; r<nregions; r++)
			{
				tempStr.readToken(inRegionStatsStream);
				inRegionStatsStream>>count;
				assert(tempStr==m_sRegionNames[r]);
				m_RegionNGround[r]+=count;
			}
			tempStr.readToken(inRegionStatsStream);	assert(tempStr=="Surface");
			for (int r=0; r<nregions; r++)
			{
				tempStr.readToken(inRegionStatsStream);
				inRegionStatsStream>>count;
				assert(tempStr==m_sRegionNames[r]);
				m_RegionNSurface[r]+=count;
			}
			
			tempStr.readToken(inRegionStatsStream);	assert(tempStr=="NPairwiseDists");
			int npairs, from, to, ndists;
			inRegionStatsStream>>npairs;
			for (int p=0; p<npairs; p++)
			{
				inRegionStatsStream>>from>>to>>ndists;
				int pairid = -1;
				for (int ftoid=0; ftoid<(int)m_PairwiseFrom.size() && pairid==-1; ftoid++)
					if (m_PairwiseFrom[ftoid]==from && m_PairwiseTo[ftoid]==to)
						pairid = ftoid;
				
				if (pairid==-1)
				{
					pairid = (int)m_PairwiseFrom.size();
					m_PairwiseFrom.push_back(from);
					m_PairwiseTo.push_back(to);
					m_PairwiseDists.push_back(std::vector<double>());
				}
					
				double dist=0;
				for (int d=0; d<ndists; d++)
				{
					inRegionStatsStream>>dist;
					m_PairwiseDists[pairid].push_back(dist);
				}
			}
			
			inInfoStream.close();
			inRegionStatsStream.close();
			
			
			for (int r=0; r<m_sRegionNames.count(); r++)
			{
				VKString rFile = inDir+"/train_"+m_sRegionNames[r]+".arff";
				std::ifstream rStream(rFile.c_str());
				if (rStream.is_open())
				{
					trainingdataexists = true;
					std::map<VKString, std::vector<double*> >::iterator iter = region2data.find(m_sRegionNames[r]);
					
					if (iter == region2data.end())
						region2data[m_sRegionNames[r]] = std::vector<double*>();
					std::vector<double*> & values = region2data[m_sRegionNames[r]];
					VKString header = "";
					bool isheader = true;
					while (!rStream.eof())
					{
						tempStr = "";
						tempStr.readLine(rStream);
						if (isheader)
						{
							header += tempStr+"\n";
							isheader = tempStr!="@DATA";
						}
						else if (tempStr!="")
						{
							VKStringList strSplit = tempStr.split(" ");
							if (strSplit.count()==1)
								strSplit = tempStr.split(",");
							assert(strSplit.count()>1);
							assert(datasize==0 || strSplit.count()==datasize);
							datasize = strSplit.count();
							int id = (int)values.size();
							values.push_back(new double[datasize]);
							for (int f=0; f<datasize; f++)
							{
								values[id][f] = strSplit[f].toDouble(&ok);
								assert(ok);
								assert((values[id][f]-values[id][f] >=0)==(values[id][f]-values[id][f] >=0));
							}
						}
					}
					assert(!isheader && (megaHeader==header || megaHeader==""));
					megaHeader = header;
					rStream.close();
				}
			}
			
		}
		std::cout<<std::endl;
		std::ofstream infoStream((outDir+"/info.txt").c_str());
		assert(infoStream.is_open());
		infoStream<<"Type Weka\n";
		infoStream<<"Timing "<<totalTime<<"s\n";
		infoStream.close();
		WritePerRegionStats(outDir+"/regionstats.txt");
		
		if (trainingdataexists)
		{
			assert(megaHeader!="");
			assert(datasize>0);
			for (int r=0; r<m_sRegionNames.count(); r++)
			{
				std::map<VKString, std::vector<double*> >::iterator iter = region2data.find(m_sRegionNames[r]);
				if (iter != region2data.end())
				{
					std::ofstream regionStream((outDir+"/train_"+m_sRegionNames[r]+".arff").c_str());
					assert(regionStream.is_open());
					regionStream<<megaHeader.c_str();
					for (int d=0; d<iter->second.size(); d++)
					for (int f=0; f<datasize; f++)
						regionStream<<iter->second[d][f]<<(f==(datasize-1)?"\n":",");
					regionStream.close();
				}
			}
		}
	}
	else
		assert(m_Type==CLA_External);	// todo...
}

void Classifier::Classify(const VKString & objName)
{
	m_aFeatureVariance = NULL;
	int nfeat = m_FeatureIndexMap[(int)m_FeatureIndexMap.size()-1];
	Model3D * m = GetModel(objName);
	GeomShape * shape = m->GetGeomShape("featpnts");
	assert(shape!=NULL);
	ShapePointCloud * pnts = shape->SafeCastToPointCloud();
	assert(pnts!=NULL);
	
	int nlabels = m_sRegionNames.count()+1;
	ShapeValue * prediction = new ShapeValue(shape, "prediction", nlabels);
	prediction->SetDimNames(m_sRegionNames);
	int npnts = pnts->NumPoints();
	double * predictvals = new double[npnts*nlabels];
	double * features = new double[nfeat];
	shape->AddShapeVal("prediction", prediction);
	prediction->InitCache(ShapeValue::SVTYPE_PerPointSet);
	for (int i=0; i<npnts; i++)
	{
		const ShapePoint & p = pnts->GetPoint(i);
		FillFeatVals(m->ModelID(), p, features);
		FillClassifyPoint(features, &(predictvals[i * (nlabels)]));
	}
	prediction->SetPntsValues(predictvals, npnts*nlabels);
	
	delete [] predictvals;
	delete [] features;
}

void Classifier::FillClassifyPoint(const double * features, double * prediction)
{
	assert(false);
	if (m_Type==CLA_Dist2Nearest || m_Type==CLA_Dist2NearestGauss)
	{
		int nfeat = m_FeatureIndexMap[(int)m_FeatureIndexMap.size()-1];
		int nlabels = m_sRegionNames.count()+1;
		for (int i=0; i<nlabels; i++)
			prediction[i] = -RN_INFINITY;
		assert(m_TrainingData.size()==1);
		TrainingData & t = *(m_TrainingData[0]);
		int ntraining = (int)t.m_aFeatures.size();
		FillFeatureVariance();
		for (int i=0; i<ntraining; i++)
		{
			double value=0;
			int label = t.m_Labels[i];
			
			if (m_Type==CLA_Dist2Nearest)
			{
				for (int f=0; f<nfeat; f++)
					value += -pow(features[f]-t.m_aFeatures[i][f], 2.);
			}
			else if (m_Type==CLA_Dist2NearestGauss)
			{
				for (int f=0; f<nfeat; f++)
					value += exp(-pow(features[f]-t.m_aFeatures[i][f], 2.) / (m_aFeatureVariance[f]*2.));
				value/=nfeat;
			}
			prediction[label] = vkMax(value, prediction[label]);
		}		
	}
	else
		assert(false);
}

void Classifier::FillFeatureVariance()
{
	if (m_aFeatureVariance==NULL)
	{
		m_aFeatureVariance = new double[NFeat()];
		double * mean = new double[NFeat()];
		for (int i=0; i<NFeat(); i++)
		{
			m_aFeatureVariance[i] = 0;
			mean[i] = 0;
		}
		
		// find mean
		int count = 0;
		for (int i=0; i<(int)m_TrainingData.size(); i++)
		for (int t=0; t<(int)m_TrainingData[i]->m_aFeatures.size(); t++)
		for (int f=0; f<NFeat(); f++)
		{
			count++;
			mean[f] += m_TrainingData[i]->m_aFeatures[t][f];
		}
		for (int f=0; f<NFeat(); f++)
			mean[f] /= count;
		
		// find variance
		for (int i=0; i<(int)m_TrainingData.size(); i++)
		for (int t=0; t<(int)m_TrainingData[i]->m_aFeatures.size(); t++)
		for (int f=0; f<NFeat(); f++)
			m_aFeatureVariance[f] += pow(m_TrainingData[i]->m_aFeatures[t][f] - mean[f], 2.);

		for (int f=0; f<NFeat(); f++)
			m_aFeatureVariance[f] /= (count-1);

		delete [] mean;
	}
}

// read / write data
void Classifier::WriteAll() const
{
	Classifier * nc = (Classifier*)this;
	// write classifier data
	for (std::map<VKString, VKString>::iterator iter = nc->m_Data2File.begin(); iter != nc->m_Data2File.end(); iter++)
	{
		if (iter->first=="classifier")
			WriteClassifier(iter->second);
		else
			std::cout<<"[WARNING] Don't know how to write: "<<iter->first.c_str()<<" (file="<<iter->second.c_str()<<std::endl;
	}
	// write objects data
	for (std::map<VKString, int>::iterator iter1 = nc->m_ObjName2Maps.begin();
		 iter1 != nc->m_ObjName2Maps.end(); iter1++)
	{
		for (std::map<VKString, VKString>::iterator iter2 = nc->m_ObjData2File[iter1->second].begin();
			 iter2 != nc->m_ObjData2File[iter1->second].end(); iter2++)
		{
			if (iter2->first=="prediction")
				WritePrediction(iter1->first, iter2->second);
			else if (iter2->first=="weka_unclassified")
				WriteWekaUnclassified(iter1->first, iter2->second);
			else
				std::cout<<"[WARNING] Don't know how to write: "<<iter2->first.c_str()<<" (file="<<iter2->second.c_str()<<std::endl;
		}
	}
}

void Classifier::AddWriteMap(const VKString & datatype, const VKString & filename)
{
	m_Data2File[datatype] = filename;
}

void Classifier::AddWriteMap(const VKString & objName, const VKString & datatype, const VKString & filename)
{
	int currid = (int)m_ObjData2File.size();
	std::map<VKString, int>::iterator iter = m_ObjName2Maps.find(objName);
	if (iter==m_ObjName2Maps.end())
		m_ObjName2Maps[objName] = currid;
	else
		currid = iter->second;
	while (currid>=(int)m_ObjData2File.size())
		m_ObjData2File.push_back(std::map<VKString, VKString>());
	m_ObjData2File[currid][datatype] = filename;
}

void Classifier::ReadFeatureMap(const VKString & filename)
{
	VKString tempStr;
	int nfeat;
	std::ifstream textStream(filename.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] Cannot read feature map: "<<filename.c_str()<<std::endl;
	assert(textStream.is_open());
	
	int featIndex = 0;
	while (!textStream.eof())
	{
		tempStr.readToken(textStream);
		textStream>>nfeat;
		if (tempStr.length()>0 && nfeat>0)
		{
			m_FeatureMap.push_back(tempStr);
			m_FeatureIndexMap.push_back(featIndex);
			featIndex += nfeat;
		}
	}
	m_FeatureIndexMap.push_back(featIndex);
	
	textStream.close();
}


void Classifier::ReadClassifier(const VKString & dirname)
{
	VKString infoFile(dirname+"/info.txt");
	VKString objFile(dirname+"/training_models.txt");
	VKString training(dirname+"/training_data.bin");
	VKString regionstatsFile(dirname+"/regionstats.txt");
	
	VKString tempStr;
	std::ifstream textStream(infoFile.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] Cannot read file: "<<infoFile.c_str()<<std::endl;
	assert(textStream.is_open());
	tempStr.readToken(textStream);	assert(tempStr=="Type");
	tempStr.readToken(textStream);
	m_Type = Str2Type(tempStr);
	textStream.close();
	
	ReadObjectNames(objFile);
	if (m_Type!=CLA_Weka && m_Type!=CLA_External)
		ReadTrainingData(training);
	
	ReadPerRegionStats(regionstatsFile);
}

void Classifier::WriteWekaTrainingData(const VKString & dirname) const
{
	
	std::ofstream textStream((dirname+"/weka_info.txt").c_str());
	assert(textStream.is_open());
	for (int i=0; i<m_sRegionNames.count(); i++)
	{
		VKString label = Label(i);
		textStream<<label.c_str()<<" "<<NPositiveExamples(i)<<"\n";
	}
	textStream.close();
	
	//assert(m_sRegionNames.count()==(int)m_TrainingData.size());
	for (int i=0; i<m_sRegionNames.count(); i++)
	{
		bool main = false;
		int symmInvLabelID = GetSymmEqClass(i, &main);
		if (m_TrainingData[symmInvLabelID]!=NULL && main)
			WriteWekaARFF(dirname+"/train_"+m_sRegionNames[i]+".arff", m_TrainingData[symmInvLabelID]);
	}
}

void Classifier::WriteWekaARFF(const VKString & filename, TrainingData * data) const
{
	assert(data!=NULL);
	std::ofstream textStream(filename.c_str());
	assert(textStream.is_open());
	
	textStream<<"@RELATION pnts\n";
	
	int nfeatures = m_FeatureIndexMap[(int)m_FeatureIndexMap.size()-1];
	int currfeat = 0;
	for (int i=0; i<m_FeatureMap.count(); i++)
	{
		for (int nf=currfeat; nf <m_FeatureIndexMap[i+1]; nf++)
			textStream<<"@ATTRIBUTE "<<m_FeatureMap[i].c_str()<<"-"<<(currfeat++)<<" NUMERIC\n";
	}
	while (currfeat < nfeatures)
		textStream<<"@ATTRIBUTE other-"<<(currfeat++)<<" NUMERIC\n";
	textStream<<"@ATTRIBUTE class NUMERIC\n";
	textStream<<"@DATA\n";

	
	assert(data->m_MembershipValue.size()==data->m_Labels.size());
	assert(data->m_MembershipValue.size()==data->m_aFeatures.size());
	for (int i=0; i<(int)data->m_aFeatures.size(); i++)
	{
		for (int f=0; f<nfeatures; f++)
			textStream<<data->m_aFeatures[i][f]<<",";
		//textStream<<data->m_Labels[i]<<"\n";
		if (data->m_Labels[i]!=1)
		{
			std::cout<<"[ERROR] Label: "<<data->m_Labels[i]<<" at "<<i<<" / "<<data->m_Labels.size()<<std::endl;
			std::cout<<"Data: "<<data<<std::endl;
		}
		assert(data->m_Labels[i]==1);
		textStream<<data->m_MembershipValue[i]<<"\n";
	}
	
	textStream.close();
}

void Classifier::WriteClassifier(const VKString & dirname) const
{
	if (m_Type==CLA_GroupWeka)
		return;
	
	((Classifier*)this)->m_Timer.finishedProcess("Classifier");
	FileManagement::CreateDirsOnPath(dirname.c_str());
	
	VKString infoFile(dirname+"/info.txt");
	VKString objFile(dirname+"/training_models.txt");
	VKString training(dirname+"/training_data.bin");
	VKString regionstatsFile(dirname+"/regionstats.txt");
	
	std::ofstream textStream(infoFile.c_str());
	assert(textStream.is_open());
	textStream<<"Type "<<Type2Str(m_Type).c_str()<<"\n";
	textStream<<"Timing "<<m_Timer.ProcessLengthStr("Classifier").c_str()<<"\n";
	textStream.close();
	
	WriteObjectNames(objFile);
	WritePerRegionStats(regionstatsFile);
	
	if (m_Type==CLA_Weka)
		WriteWekaTrainingData(dirname);
	else
		WriteTrainingData(training);
}

void Classifier::WritePerRegionStats(const VKString & filename) const
{
	std::ofstream textStream(filename.c_str());
	assert(textStream.is_open());
	textStream<<"NPoses "<<m_iNPoses<<"\n";
	assert((int)m_RegionNGround.size()==m_sRegionNames.count());
	assert((int)m_RegionNSurface.size()==m_sRegionNames.count());
	textStream<<"NRegions "<<m_sRegionNames.count()<<"\n";
	textStream<<"Ground\n";
	for (int i=0; i<m_sRegionNames.count(); i++)
		textStream<<m_sRegionNames[i].c_str()<<" "<<m_RegionNGround[i]<<"\n";
	textStream<<"Surface\n";
	for (int i=0; i<m_sRegionNames.count(); i++)
		textStream<<m_sRegionNames[i].c_str()<<" "<<m_RegionNSurface[i]<<"\n";

	
	textStream<<"NPairwiseDists "<<m_PairwiseDists.size()<<"\n";
	for (int i=0; i<(int)m_PairwiseDists.size(); i++)
	{
		textStream<<m_PairwiseFrom[i]<<" "<<m_PairwiseTo[i]<<" "<<m_PairwiseDists[i].size()<<"\n";
		for (int j=0; j<(int)m_PairwiseDists[i].size(); j++)
			textStream<<m_PairwiseDists[i][j]<<" ";
		textStream<<"\n";
	}

	
	textStream.close();
}

void Classifier::ReadPerRegionStats(const VKString & filename)
{
	std::ifstream textStream(filename.c_str());
	assert(textStream.is_open());
	VKString tempStr;
	tempStr.readToken(textStream);	assert(tempStr=="NPoses");
	textStream>>m_iNPoses;
	
	int nregions;
	tempStr.readToken(textStream);	assert(tempStr=="NRegions");
	textStream>>nregions;
	assert(nregions==m_sRegionNames.count());
	for (int i=0; i<nregions; i++)
	{
		m_RegionNGround.push_back(0);
		m_RegionNSurface.push_back(0);
	}
	
	tempStr.readToken(textStream);	assert(tempStr=="Ground");
	for (int i=0; i<nregions; i++)
	{
		tempStr.readToken(textStream);		assert(tempStr==m_sRegionNames[i]);
		textStream>>m_RegionNGround[i];
	}
	tempStr.readToken(textStream);	assert(tempStr=="Surface");
	for (int i=0; i<m_sRegionNames.count(); i++)
	{
		tempStr.readToken(textStream);		assert(tempStr==m_sRegionNames[i]);
		textStream>>m_RegionNSurface[i];
	}
	
	tempStr.readToken(textStream);	assert(tempStr=="NPairwiseDists");
	int npairwise;
	textStream>>npairwise;
	for (int i=0; i<npairwise; i++)
	{
		m_PairwiseFrom.push_back(0);
		m_PairwiseTo.push_back(0);
		m_PairwiseDists.push_back(std::vector<double>());
		
		int ndists;
		textStream>>m_PairwiseFrom[i]>>m_PairwiseTo[i]>>ndists;
		for (int j=0; j<ndists; j++)
			m_PairwiseDists[i].push_back(0);
		
		for (int j=0; j<ndists; j++)
			textStream>>m_PairwiseDists[i][j];
	}

	textStream.close();
}

void Classifier::ReadObjectNames(const VKString & filename)
{
	// not necessary...
//	std::ifstream textStream(filename.c_str());
//	assert(textStream.is_open());
//	VKString tempStr;
//	for (int i=0; i<m_pDB->NumModels(); i++)
//	{
//		assert(!textStream.eof());
//		tempStr.readToken(textStream);
//		assert(tempStr==m_pDB->GetModel(i)->ModelName());	// just checking, this function is to avoid errors in indexing...
//	}
//	textStream.close();
}

void Classifier::WriteObjectNames(const VKString & filename) const
{
	std::ofstream textStream(filename.c_str());
	assert(textStream.is_open());
	for (int i=0; i<m_pDB->NumModels(); i++)
		textStream<<m_pDB->GetModel(i)->ModelName().c_str()<<"\n";
	textStream.close();
}

void Classifier::ReadTrainingData(const VKString & filename)
{
	Classifier * nc = (Classifier*)this;
	FILE * file = fopen(filename.c_str(), "rb");
	if (file==NULL)
		std::cout<<"[ERROR] Could not read training data: "<<filename.c_str()<<std::endl;
	assert(file!=NULL);
	int nsets, nfeat, nsamp;
	fread(&nsets, sizeof(int), 1, file);
	fread(&nfeat, sizeof(int), 1, file);
	assert(m_FeatureIndexMap.size()>0);
	assert(nfeat == m_FeatureIndexMap[(int)m_FeatureIndexMap.size()-1]);

	for (int i=0; i<nsets; i++)
	{
		nc->m_TrainingData.push_back(new TrainingData());
		fread(&nsamp, sizeof(int), 1, file);
		for (int s=0; s<nsamp; s++)
		{
			double * featvals = new double[nfeat];
			fread(featvals, sizeof(double), nfeat, file);
			m_TrainingData[i]->m_aFeatures.push_back(featvals);
		}
		for (int s=0; s<nsamp; s++)
		{
			int label;
			fread(&label, sizeof(int), 1, file);
			assert(label==1);
			m_TrainingData[i]->m_Labels.push_back(1);
			//m_TrainingData[i]->m_Labels.push_back(label);
		}
	}
	fclose(file);
}

void Classifier::WriteTrainingData(const VKString & filename) const
{
	FILE * file = fopen(filename.c_str(), "wb");
	assert(file!=NULL);
	int nsets = (int)m_TrainingData.size();
	assert(m_FeatureIndexMap.size()>0);
	int nfeat = m_FeatureIndexMap[(int)m_FeatureIndexMap.size()-1];
	fwrite(&nsets, sizeof(int), 1, file);
	fwrite(&nfeat, sizeof(int), 1, file);
	for (int i=0; i<nsets; i++)
	{
		int nsamp =(int)m_TrainingData[i]->m_aFeatures.size();
		fwrite(&nsamp, sizeof(int), 1, file);
		for (int s=0; s<nsamp; s++)
			fwrite(m_TrainingData[i]->m_aFeatures[s], sizeof(double), nfeat, file);
		for (int s=0; s<nsamp; s++)
			fwrite(&(m_TrainingData[i]->m_Labels[s]), sizeof(int), 1, file);
	}
	fclose(file);
}

void Classifier::ReadGTRegionNames(const VKString & filename)
{
	bool ok;
	std::ifstream textStream(filename.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] File does not exist: "<<filename.c_str()<<std::endl;
	assert(textStream.is_open());
	while (!textStream.eof())
	{
		VKString strLine = "";
		strLine.readLine(textStream);
		if (strLine!="")
		{
			if (strLine.startsWith("#"))
				continue;
			
			VKStringList splitLine = strLine.split(" ");
			assert(splitLine.count()>=3);
			m_sRegionNames.push_back(splitLine[0]);
			m_sRegionTypes.push_back(splitLine[1]);
			m_RegionNPoints.push_back(splitLine[2].toInt(&ok));
			assert(ok);
		}
	}
	textStream.close();
}

Classifier::ObjData & Classifier::GetObjData(const VKString & objName)
{
	std::map<VKString, int>::iterator iter = m_ObjName2ObjReadFiles.find(objName);
	int id = 0;
	if (iter==m_ObjName2ObjReadFiles.end())
	{
		id = (int)m_DatasetFiles.size();
		m_ObjName2ObjReadFiles[objName] = id;
		m_DatasetFiles.push_back(ObjData());
		m_DatasetFiles[id].m_sObjName = objName;
	}
	else
		id = iter->second;
	return m_DatasetFiles[id];
}

void Classifier::ReadEmpty(const VKString & objName)
{
	GetObjData(objName);
	GetModel(objName);
}

void Classifier::ReadMesh(const VKString & objName, const VKString & filename)
{
//	GetObjData(objName).m_sMeshFile = filename;

	
	Model3D * m = GetModel(objName);
	ShapeTriangleSoup * mesh = new ShapeTriangleSoup(filename, Model3D::m_sMainGeomName, m);
	m->AddGeomShape(mesh->GeometryName(), mesh);
	
	int mid = m->ModelID();
	assert(mid == (int)m_AveRad.size() && mid==(int)m_Translate.size());

	R3Point c = mesh->Centroid();
	double r = mesh->AverageRadius(c);
	m_AveRad.push_back(r);
	m_Translate.push_back(R3Vector(c.X(), c.Y(), c.Z()));	
}


void Classifier::ReadFeaturePnts(const VKString & objName, const VKString & filename)
{
//	GetObjData(objName).m_sFeaturePntsFile = filename;
	
	Model3D * m = GetModel(objName);
	ShapePointCloud * points = new ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD, "featpnts", Model3D::m_sMainGeomName, m);
	points->ReadGeometryFile(filename);
	m->AddGeomShape(points->GeometryName(), points);
}

void Classifier::ReadFeatureVals(const VKString & objName, const VKString & valname, const VKString & filename)
{
//	GetObjData(objName).m_sFeatureName2File[valname] = filename;
	
	GeomShape * shape = GetModel(objName)->GetGeomShape("featpnts");
	assert(shape!=NULL);
	ShapeValue * vals = new ShapeValue(shape, valname, -1);
	vals->ReadShapeValues(filename);
	shape->AddShapeVal(valname, vals);
}

void Classifier::ReadMetric(const VKString & objName, const VKString & shapeFile, const VKString & metricFile)
{
//	GetObjData(objName).m_sMetricFile = metricFile;
//	GetObjData(objName).m_sMetricShapeFile = shapeFile;
	
	assert(shapeFile.endsWith(".pts"));	// assume points
	Model3D * m = GetModel(objName);
	ShapePointCloud * points = new ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD, "densepnts", Model3D::m_sMainGeomName, m);
	points->ReadGeometryFile(shapeFile);
	m->AddGeomShape(points->GeometryName(), points);
	ShapeDistance * dist = ShapeDistance::ReadMetric(metricFile, points, "custommetric", 1.);
	points->AddDistanceMetric(dist->MetricName(), dist);
	m_sMetricName = "custommetric";
}

void Classifier::ReadGTRegions(const VKString & objName, const VKString & filename)
{
//	GetObjData(objName).m_sGTRegionFile = filename;
	
	Model3D * m = GetModel(objName);
	GeomShape * shape = m->GetGeomShape(Model3D::m_sMainGeomName);
	assert(shape!=NULL);
	ShapeMesh * mesh = shape->SafeCastToMesh();	assert(mesh!=NULL);

	int mid = m->ModelID();

	while (mid >= m_Poses.size())
		m_Poses.push_back(std::vector<PntCorrsPose*>());

	PntCorrsPose::CreatePosesFromGTFile(filename, mesh, "none", m_Poses[mid]);
}

void Classifier::ReadCanonicalXForm(const VKString & objName, const VKString & filename)
{
//	GetObjData(objName).m_sCanonicalXFormFile = filename;
	
	Model3D * m = GetModel(objName);
	GeomShape * shape = m->GetGeomShape(Model3D::m_sMainGeomName);	assert(shape!=NULL);
	ShapeMesh * mesh = shape->SafeCastToMesh();	assert(mesh!=NULL);
	double x,y,z;
	double rad;
	VKString tempStr;
	std::ifstream xformStream(filename.c_str());
	int mid = m->ModelID();
	while (mid >= (int)m_Scales.size())
		m_Scales.push_back(1);
	while (mid >= (int)m_AnglesUp.size())
		m_AnglesUp.push_back(0);
	
	if (xformStream.is_open())
	{
		tempStr.readToken(xformStream);		assert(tempStr=="Scale");
		xformStream>>m_Scales[mid];
		tempStr.readToken(xformStream);		assert(tempStr=="RotZ");
		xformStream>>m_AnglesUp[mid];
		tempStr.readToken(xformStream);		assert(tempStr=="OrigCentroid");
		
		xformStream>>x>>y>>z;	// not used
		tempStr.readToken(xformStream);		assert(tempStr=="OrigRadius");
		xformStream>>rad;	// not used
		xformStream.close();
	}
}

void Classifier::ReadWekaResults(const VKString & objName, const VKString & filename)
{
	Model3D * m = GetModel(objName);
	GeomShape * shape = m->GetGeomShape("featpnts");
	assert(shape!=NULL);
	ShapePointCloud * pnts = shape->SafeCastToPointCloud();
	assert(pnts!=NULL);

	int npnts = pnts->NumPoints();
	int nlabels = m_sRegionNames.count()+1;
	ShapeValue * prediction = new ShapeValue(shape, "prediction", nlabels);
	shape->AddShapeVal("prediction", prediction);
	double * predictvals = new double[npnts*nlabels];
	double * remappedvals = new double[npnts*nlabels];
	
	for (int i=0; i<npnts*nlabels; i++)
		predictvals[i] = 0;
	for (int i=0; i<npnts*nlabels; i++)
		remappedvals[i] = 0;
	prediction->InitCache(ShapeValue::SVTYPE_PerPointSet);
	
	//std::cout<<"    prediction: ";
	double epsilon = 1E-10;
	for (int i=0; i< m_sRegionNames.count(); i++)
	{
		VKString fName = filename+m_sRegionNames[i]+".txt";
		std::ifstream textStream(fName.c_str());
		if (!textStream.is_open())
			continue;
		assert(textStream.is_open());
		bool mainname = false;
		int symmInvLabelID = GetSymmEqClass(i, &mainname);
		if (!mainname)
		{
			std::cout<<"\n[ERROR] Invalid region (probably due to symmetry invariance): "<<m_sRegionNames[i].c_str()<<std::endl;
			std::cout<<"File that should not exist: "<<fName.c_str()<<std::endl;
		}
		assert(mainname);
		
		VKString tempStr;
		int inst = 0;
		double labels[2] = {RN_INFINITY,-RN_INFINITY};
		while (!textStream.eof())
		{
			tempStr.readLine(textStream);
			bool metaline = ((tempStr=="=== Predictions on test data ===") || (tempStr==""));
			VKStringList entries = tempStr.split(" ");
			if (entries.count()>0)
				metaline = metaline || entries[0].replace(" ", "")=="inst#";
			
			if (!metaline)
			{
				assert(entries.count()==4);
				inst++;
				int id = inst-1;
				bool ok;
				int readinst = entries[0].replace(" ", "").toInt(&ok);
				assert(ok && readinst==inst);
				double value = entries[2].toDouble(&ok);
				assert(ok);
				
				predictvals[id * nlabels + symmInvLabelID] = value;
				if (predictvals[id * nlabels + symmInvLabelID]<epsilon)
					predictvals[id * nlabels + symmInvLabelID] = 0.;
				if (predictvals[id * nlabels + symmInvLabelID]>1)
					predictvals[id * nlabels + symmInvLabelID] = 1.;
				
				labels[0] = vkMin(labels[0], value);
				labels[1] = vkMax(labels[1], value);
			}
		}
		
		//std::cout<<m_sRegionNames[i].c_str()<<"("<<labels[0]<<":"<<labels[1]<<")  ";
		//std::cout<<"      --- Read Weka results: "<<labels[0]<<" "<<labels[1]<<std::endl;
		
		textStream.close();
		if (inst==0)
			std::cout<<"[ERROR] Cannot parse weka file: "<<fName.c_str()<<std::endl;
		assert(inst>0);
	}
	//std::cout<<std::endl;
	
	for (int p=0; p<npnts; p++)
	for (int l=0; l<m_sRegionNames.count(); l++)
	{
		int symmInvLabelID = GetSymmEqClass(l, NULL);
		remappedvals[p * nlabels + l] = predictvals[p * nlabels + symmInvLabelID];
	}
	
	prediction->SetPntsValues(remappedvals, npnts*nlabels);
	delete [] predictvals;
	delete [] remappedvals;
}


void Classifier::WriteWekaUnclassified(const VKString & objName, const VKString & filename) const
{
	// write arff: features, class='?'
	FileManagement::CreateDirsOnPathToFile(filename.c_str());
	
	Classifier * nc = (Classifier*)this;
	GeomShape * shape = nc->GetModel(objName)->GetGeomShape("featpnts");
	assert(shape!=NULL);
	ShapePointCloud * pnts = shape->SafeCastToPointCloud();
	assert(pnts!=NULL);
	
	std::ofstream textStream(filename.c_str());
	assert(textStream.is_open());
	textStream<<"@RELATION pnts-featpnts\n";
	int nfeatures = m_FeatureIndexMap[(int)m_FeatureIndexMap.size()-1];
	int currfeat = 0;
	for (int i=0; i<m_FeatureMap.count(); i++)
	{
		ShapeValue * val = pnts->GetShapeVal(m_FeatureMap[i]);
		assert(val!=NULL);
		int nfeat = val->NDims();
		for (int nf=0; nf < nfeat; nf++)
			textStream<<"@ATTRIBUTE "<<m_FeatureMap[i].c_str()<<"-"<<(currfeat++)<<" NUMERIC\n";
	}
	while (currfeat < nfeatures)
		textStream<<"@ATTRIBUTE other-"<<(currfeat++)<<" NUMERIC\n";
	textStream<<"@ATTRIBUTE class NUMERIC\n";
	textStream<<"@DATA\n";
	
	for (int p=0; p<pnts->NumPoints(); p++)
	{
		for (int i=0; i<m_FeatureMap.count(); i++)
		{
			ShapeValue * val = pnts->GetShapeVal(m_FeatureMap[i]);
			assert(val!=NULL);
			int offset = m_FeatureIndexMap[i];
			int nfeat = val->NDims();
			assert(offset + nfeat <= nfeatures);	// could be that feature list has incorrect feature dimensions
			for (int d=0; d<nfeat; d++)
				textStream<<val->Value(pnts->GetPoint(p), d)<<",";
		}
		textStream<<"?\n";
	}
	textStream.close();
}

void Classifier::WritePrediction(const VKString & objName, const VKString & filename) const
{
	//std::cout<<"Writing prediction: "<<objName.c_str()<<" "<<filename.c_str()<<std::endl;
	FileManagement::CreateDirsOnPathToFile(filename.c_str());
	Classifier * nc = (Classifier*)this;
	GeomShape * shape = nc->GetModel(objName)->GetGeomShape("featpnts");
	assert(shape!=NULL);
	ShapeValue * val = shape->GetShapeVal("prediction");
	assert(val!=NULL);
	val->WriteShapeValues(filename);
}

void Classifier::ReadPrediction(const VKString & objName, const VKString & filename) 
{
	Classifier * nc = (Classifier*)this;
	GeomShape * shape = nc->GetModel(objName)->GetGeomShape("featpnts");
	assert(shape!=NULL);
	
	ShapeValue * val = new ShapeValue(shape, "prediction");
	val->ReadShapeValues(filename.c_str());
	shape->AddShapeVal("prediction", val);
	val->SetDimNames(m_sRegionNames);
}

Model3D * Classifier::GetModel(const VKString & objName)
{
	Model3D * m = m_pDB->GetModel(objName);
	if (m==NULL)
	{
		m = new Model3D(objName, m_pDB->NumModels());
		m_pDB->AddModel(m);
	}
	return m;
}

void Classifier::GetPredictionConfidenceBounds(const VKString & objName, double ** minvals, double ** maxvals) const
{

	Classifier * nc = (Classifier*)this;
	GeomShape * shape = nc->GetModel(objName)->GetGeomShape("featpnts");
	assert(shape!=NULL);
	ShapePointCloud * pnts = shape->SafeCastToPointCloud();
	ShapeValue * predict = pnts->GetShapeVal("prediction");
	assert(predict!=NULL);
	assert(NLabels()==predict->NDims());
	*minvals = new double[NLabels()];
	*maxvals = new double[NLabels()];
	
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		for (int j=0; j<NLabels(); j++)
		{
			double val = predict->Value(pnts->GetPoint(i), j);
			if (i==0)
			{
				(*minvals)[j] = val;
				(*maxvals)[j] = val;
			}
			else
			{
				(*minvals)[j] = vkMin((*minvals)[j], val);
				(*maxvals)[j] = vkMax((*maxvals)[j], val);
			}
		}
	}
//	if (m_Type!=CLA_Dist2Nearest)	// normalized
//	{
//		for (int j=0; j<NLabels(); j++)
//		{
//			if ((*minvals)[j]!=(*maxvals)[j])
//			{
//				(*minvals)[j] = 0;
//				(*maxvals)[j] = 1.;
//			}
//		}
//	}
}

int Classifier::NFeat() const
{
	assert(m_FeatureIndexMap.size()>0);
	return m_FeatureIndexMap[(int)m_FeatureIndexMap.size()-1];
}

int Classifier::NLabels() const
{
	return m_sRegionNames.count()+1;
}

VKString Classifier::Label(int l) const
{
	if (l<m_sRegionNames.count())
		return m_sRegionNames[l];
	else
		return "Other";
}

double Classifier::GetScale(int i) const
{
	if (i>=0 && i < (int)m_Scales.size())
		return m_Scales[i];
	else
		return 1;
}

int Classifier::NPositiveExamples(int label) const
{
	// NOTE: check points -- if not on mesh DO NOT LEARN (e.g. could be legs on ground)
	// go over all gt points
	int nexamples = 0;
	for (int obj=0; obj<(int)m_Poses.size(); obj++)
	for (int pose=0; pose<(int)m_Poses[obj].size(); pose++)
	for (int region=0; region<m_Poses[obj][pose]->NMatches(); region++)
	{
		bool onsurface = m_Poses[obj][pose]->Prim(region).IsType(TargetPrimitive::PRIM_Point);
//		bool ground = m_Poses[obj][pose]->Prim(region).IsType(TargetPrimitive::PRIM_Ground);
		if (m_Poses[obj][pose]->Region(region)==label && onsurface)	// if gt points with 'label' - add
			nexamples++;
	}
	return nexamples;
}

void Classifier::FillPerRegionStats()
{
	assert(m_pSolver!=NULL);
	for (int i=0; i<m_sRegionNames.count(); i++)
	{
		m_RegionNGround.push_back(0);
		m_RegionNSurface.push_back(0);
	}

	m_iNPoses = 0;
	for (int obj=0; obj<(int)m_Poses.size(); obj++)
	for (int pose=0; pose<(int)m_Poses[obj].size(); pose++)
	{
		m_iNPoses++;
		m_pSolver->ForwardKinematics(m_Poses[obj][pose]);
		for (int region=0; region<(int)m_Poses[obj][pose]->NMatches(); region++)
		{
			int rid = m_Poses[obj][pose]->Region(region);
			assert(rid < m_sRegionNames.count());
			if (m_Poses[obj][pose]->Prim(region).IsType(TargetPrimitive::PRIM_Point))
				m_RegionNSurface[rid]++;
			else if (m_Poses[obj][pose]->Prim(region).IsType(TargetPrimitive::PRIM_Ground))
				m_RegionNGround[rid]++;
			else
				assert(false);

			for (int r2=0; r2<(int)m_Poses[obj][pose]->NMatches(); r2++)
			{
				int rfrom = m_Poses[obj][pose]->Region(region);
				int rto = m_Poses[obj][pose]->Region(r2);
				if (rfrom>=rto || m_Poses[obj][pose]->Prim(region).IsType(TargetPrimitive::PRIM_Ground)
					|| m_Poses[obj][pose]->Prim(r2).IsType(TargetPrimitive::PRIM_Ground))
					continue;
				
				int pairid = -1;
				for (int pair=0; pair < (int)m_PairwiseFrom.size() && pairid==-1; pair++)
					if (m_PairwiseFrom[pair]==rfrom && m_PairwiseTo[pair]==rto)
						pairid = pair;
				if (pairid<0)
				{
					pairid = (int)m_PairwiseFrom.size();
					m_PairwiseFrom.push_back(rfrom);
					m_PairwiseTo.push_back(rto);
					m_PairwiseDists.push_back(std::vector<double>());
				}
				
//				R3Point p1 = m_Poses[obj][pose]->GetPosition(m_pSolver->m_Pnts2Skeleton[rfrom]);
//				R3Point p2 = m_Poses[obj][pose]->GetPosition(m_pSolver->m_Pnts2Skeleton[rto]);
				R3Point p1 =m_Poses[obj][pose]->Prim(region).GetPosition();
				R3Point p2 =m_Poses[obj][pose]->Prim(r2).GetPosition();
				double dist = R3Distance(p1, p2);
				m_PairwiseDists[pairid].push_back(dist);
			}
		}
	}
	
}

void Classifier::FillOnSurfaceMembershipValue(int label, const TrainingParams & params)
{
	assert(params.m_fDistThreshold>0);
	// go over all objects
	int nfeatures = m_FeatureIndexMap[(int)m_FeatureIndexMap.size()-1];
	while (m_apGTPredictions.size() < m_Poses.size())
	{
		m_apGTPredictions.push_back(NULL);
		m_aMembershipTypes.push_back(NULL);
		m_aMap2TrainingData.push_back(NULL);
	}
	assert(m_apGTPredictions.size()==m_Poses.size());
	
	for (int obj=0; obj<(int)m_Poses.size(); obj++)
	{
		Model3D * m = m_pDB->GetModel(obj);
		GeomShape * shape = m->GetGeomShape("featpnts");  assert(shape!=NULL);
		ShapePointCloud * pnts = shape->SafeCastToPointCloud();  assert(pnts!=NULL);
		GeomShape * shapemesh = m->GetGeomShape(Model3D::m_sMainGeomName);	assert(shapemesh!=NULL);
		ShapeMesh * mesh = shapemesh->SafeCastToMesh();	assert(mesh!=NULL);
		
		if (m_apGTPredictions[obj]==NULL)
		{
			int nelements = pnts->NumPoints() * m_sRegionNames.count();
			m_NElements.push_back(nelements);
			m_apGTPredictions[obj] = new double[nelements];
			m_aMembershipTypes[obj] = new int[nelements];
			m_aMap2TrainingData[obj] = new int[nelements];
			for (int i=0; i<pnts->NumPoints() * m_sRegionNames.count(); i++)
			{
				m_apGTPredictions[obj][i] = 0;
				m_aMembershipTypes[obj][i] = MEMBER_RegPoint;
				m_aMap2TrainingData[obj][i] = -1;
			}
		}
				
		// old code -- no metric
		
		if (m_sMetricName==ShapeDistance::m_sDefaultEuclidean)
			assert(false);	// might use in the future
		else	// new code -- pseudo-geodesic metric
		{
			//std::cout<<"Getting membership values: "<<label<<std::endl;
			// get dense points for pseudo-geodesics and the metric
			GeomShape * densepntsshape = m->GetGeomShape("densepnts");				assert(densepntsshape!=NULL);
			ShapePointCloud * densepnts = densepntsshape->SafeCastToPointCloud(); 	assert(densepnts!=NULL);
			double * dists = new double[densepnts->NumPoints()];
			ShapeDistance * metric = densepnts->GetDistanceMetric(m_sMetricName);
			assert(metric!=NULL && metric->SafeCastToGraphNN()!=NULL);
			VKGraph * distgraph = (VKGraph*)metric->SafeCastToGraphNN()->GetGraph();
			assert(distgraph!=NULL);
			
			double sigma = params.m_fDistThreshold;
			if (params.m_fDist2SigmaScale>0)
				sigma *= params.m_fDist2SigmaScale;
			else if (params.m_fPositiveValCutOff>0)
				sigma *= 1. / sqrt(2. * -log(params.m_fPositiveValCutOff));
			else
				assert(false);
			assert(sigma>0);
			
			assert(m_Poses[obj].size()>0);
			// go over ground truth points of interest -- list them all
			int pose = 0;
			for (int region=0; region<(int)m_Poses[obj][pose]->NMatches(); region++)
			{
				// check whether point is POI or other gt point
				if (!m_Poses[obj][pose]->Prim(region).IsType(TargetPrimitive::PRIM_Point))
				{
					assert(m_Poses[obj][pose]->Prim(region).IsType(TargetPrimitive::PRIM_Ground));
					continue;
				}
				const ShapePoint & p = m_Poses[obj][pose]->Prim(region).GetPoint();
				MembershipType tp;
				if (m_Poses[obj][pose]->Region(region)==label)
					tp = MEMBER_NearGtPOI;
				else
					tp = MEMBER_NearGtOther;
				
				// get neighborhood for gt point of interest -- fill values (including infinities)
				int nearDense1 = GetNearestDenseAmongGT(obj, pose, region);
				distgraph->ThresholdedShortestPath(nearDense1, params.m_fDistThreshold, dists);
				
				const R3Kdtree<const PointInSet *> * kdTree = pnts->GetKdTree();
				RNArray<const PointInSet *> euclNhd;
				kdTree->FindAll(p.GetPosition(), 0, params.m_fDistThreshold, euclNhd);
				for (int n=0; n<euclNhd.NEntries(); n++)
				{
					int localpointid = euclNhd[n]->id * m_sRegionNames.count() + label;
					assert(localpointid < m_NElements[obj]);
					int nearDense2 = GetNearestDenseAmongPnts(obj, euclNhd[n]->id);
					double d = dists[nearDense2];
					if (d<params.m_fDistThreshold)
					{
						if (tp == MEMBER_NearGtPOI)
						{
							double & mv = m_apGTPredictions[obj][localpointid];
							mv = vkMax(mv, exp(-(d*d) / (2*sigma*sigma)));
						}
						m_aMembershipTypes[obj][localpointid] |= tp;
					}
				}
			}
		}
	}
}

void Classifier::AddTrainingExamples(TrainingData & training, int symmid, const std::vector<int> & symmetryGroup,
									 const TrainingParams & params, int & npositive, int & nnegative)
{
	nnegative = 0;
	npositive = 0;
	int nfeatures = m_FeatureIndexMap[(int)m_FeatureIndexMap.size()-1];

//	std::cout<<"Symm Group: "
	
	// take maxmimum of values in a symmetry group...
	for (int obj=0; obj<(int)m_Poses.size(); obj++)
	{
		std::vector<double *> candidateNegativeFeatures;
		std::vector<int> candidateNegativePntIDs;
		Model3D * m = m_pDB->GetModel(obj);
		GeomShape * shape = m->GetGeomShape("featpnts");  assert(shape!=NULL);
		ShapePointCloud * pnts = shape->SafeCastToPointCloud();  assert(pnts!=NULL);
		int currnpositive = 0;
		int currnnegative = 0;
		double minmember = RN_INFINITY;
		double maxmember = -RN_INFINITY;
		for (int p=0;  p < pnts->NumPoints();  p++)
		{
			// find features...
			double * features = new double[nfeatures];
			for (int i=0; i<m_FeatureMap.count(); i++)
			{	
				ShapeValue * val = pnts->GetShapeVal(m_FeatureMap[i]);
				if (val==NULL)
					std::cout<<"[ERROR] Feature not loaded: Obj: "<<m->ModelName().c_str()<<" "<<m_FeatureMap[i].c_str()<<std::endl;
				assert(val!=NULL);
				int offset = m_FeatureIndexMap[i];
				int nfeat = val->NDims();
				assert(offset + nfeat <= nfeatures);	// could be that feature list has incorrect feature dimensions
				for (int d=0; d<nfeat; d++)
					features[offset+d] = val->Value(pnts->GetPoint(p), d);
			}

			// find membership value
			double maxval = 0;
			int tp = MEMBER_RegPoint;	// flags due to all symmetric things
			for (int i=0; i<(int)symmetryGroup.size(); i++)
			{
				int id = p * m_sRegionNames.count() + symmetryGroup[i];
				assert(id < m_NElements[obj]);
				assert(obj < (int)m_apGTPredictions.size() && symmetryGroup[i] < m_sRegionNames.count());
				maxval = vkMax(maxval, m_apGTPredictions[obj][id]);
				tp = tp | m_aMembershipTypes[obj][id];
			}

			// add training example
			bool addexample = false;
			switch(params.m_TrainingData)
			{
				case TRAIN_GaussianAll:
					addexample = true;
					break;
				case TRAIN_GaussianRandomAndLabeled:
					addexample = ((tp&MEMBER_NearGtPOI)!=0 || (tp&MEMBER_NearGtOther)!=0);
					if ((tp&MEMBER_NearGtOther)!=0)
						maxval = 0;
					if (!addexample)
					{
						candidateNegativeFeatures.push_back(features);
						candidateNegativePntIDs.push_back(p);
					}
					break;
				case TRAIN_GaussianRandom:
					addexample = ((tp&MEMBER_NearGtPOI)!=0);
					if (!addexample)
					{
						candidateNegativeFeatures.push_back(features);
						candidateNegativePntIDs.push_back(p);
					}
					break;
				case TRAIN_GaussianLabeledOnly:
					addexample = ((tp&MEMBER_RegPoint)!=0);
					break;
			}
			
			if (addexample)
			{
				for (int i=0; i<(int)symmetryGroup.size(); i++)
				{
					int id = p * m_sRegionNames.count() + symmetryGroup[i];
					assert(id < m_NElements[obj]);
					m_aMembershipTypes[obj][id] |= MEMBER_Inlcuded;
					m_aMap2TrainingData[obj][id] = ((int)training.m_aFeatures.size());
				}

				training.m_aFeatures.push_back(features);
				//std::cout<<"T1: "<<training.m_Labels.size()<<" -> 1"<<std::endl;
				training.m_Labels.push_back(1);
				
//				std::cout<<"    Membership value: "<<maxval<<std::endl;
				minmember = vkMin(minmember, maxval);
				maxmember = vkMax(maxmember, maxval);
				
				training.m_MembershipValue.push_back(maxval);
				if (maxval==0)
				{
					nnegative++;
					currnnegative++;
				}
				else
				{
					npositive++;
					currnpositive++;
				}
			}
//			else
//				delete [] features;
		}
		
		if (params.m_TrainingData==TRAIN_GaussianRandomAndLabeled || params.m_TrainingData==TRAIN_GaussianRandom)
		{
			for (int ne = 0; ne < currnpositive * params.m_fRatioRandomTrainingData && candidateNegativeFeatures.size()>0; ne++)
			{
				int randid = rand()%((int)candidateNegativeFeatures.size());
				assert(randid>=0 && randid < (int)candidateNegativeFeatures.size());
				assert(candidateNegativePntIDs.size()==candidateNegativeFeatures.size());
				int p = candidateNegativePntIDs[randid];
				training.m_aFeatures.push_back(candidateNegativeFeatures[randid]);
				training.m_Labels.push_back(1);
				training.m_MembershipValue.push_back(0);

				candidateNegativeFeatures.erase(candidateNegativeFeatures.begin()+randid);
				candidateNegativePntIDs.erase(candidateNegativePntIDs.begin()+randid);
				
				for (int i=0; i<(int)symmetryGroup.size(); i++)
				{
					int id = p * m_sRegionNames.count() + symmetryGroup[i];
					assert(id < m_NElements[obj]);
					m_aMembershipTypes[obj][id] |= MEMBER_Inlcuded;
					m_aMap2TrainingData[obj][id] =((int)training.m_aFeatures.size())-1;
				}
				nnegative++;
				currnnegative++;
			}
		}
	}
}

void Classifier::FillFeatVals(int objID, const ShapePoint & pnt, double * features)
{
	int nfeatures = m_FeatureIndexMap[(int)m_FeatureIndexMap.size()-1];
	R3Point p = pnt.GetPosition();
	GeomShape * shape = m_pDB->GetModel(objID)->GetGeomShape("featpnts");
	assert(shape!=NULL);
	ShapePointCloud * pnts = shape->SafeCastToPointCloud();
	assert(pnts!=NULL);
	const PointInSet * closest = pnts->GetKdTree()->FindClosest(p);
	assert(closest!=NULL);
	
	for (int i=0; i<m_FeatureMap.count(); i++)
	{
		ShapeValue * val = pnts->GetShapeVal(m_FeatureMap[i]);
		assert(val!=NULL);
		int offset = m_FeatureIndexMap[i];
		int nfeat = val->NDims();
		assert(offset + nfeat <= nfeatures);	// could be that feature list has incorrect feature dimensions
		for (int d=0; d<nfeat; d++)
			features[offset+d] = val->Value(closest->GetModelPoint(), d);
	}
}


void Classifier::ReadSymmetryEquivalence(const VKString & symmInvFile)
{
	// check is symmetry-invariant
	int equivclass = 0;
	if (symmInvFile!="none")
	{
		std::ifstream textStream(symmInvFile.c_str());
		if (!textStream.is_open())
			std::cout<<"[ERROR] Cannot read: "<<symmInvFile.c_str()<<std::endl;
		assert(textStream.is_open());
		VKString tempStr;
		while (!textStream.eof())
		{
			tempStr.readLine(textStream);
			if (tempStr=="")
				continue;
			VKStringList spLine = tempStr.split(" ");
			for (int i=0; i<spLine.count(); i++)
			{
				if (spLine[i].length()>1)
					m_Name2SymmClass[spLine[i]] = equivclass;
			}
			equivclass++;
			m_SymmEqClassName.push_back(spLine[0]);
		}
		textStream.close();
	}
}


int Classifier::GetSymmEqClass(int labelID, bool * mainName) const
{
	if (m_Name2SymmClass.size()==0)
	{
		if (mainName!=NULL) *mainName = true;
		return labelID;
	}
	assert(labelID>=0 && labelID < m_sRegionNames.count());
	std::map<VKString, int>::iterator iter = ((Classifier*)this)->m_Name2SymmClass.find(m_sRegionNames[labelID]);
	assert(iter != m_Name2SymmClass.end());
	assert(iter->second >= 0 && iter->second < m_SymmEqClassName.count());
	if (mainName!=NULL)
		*mainName = (m_sRegionNames[labelID]==m_SymmEqClassName[iter->second]);
	return iter->second;
}

int Classifier::GetNearestDenseAmongGT(int obj, int pose, int region)
{
	if (m_Gt2DensePnts.size()==0)
	{
		for (int obj=0; obj<(int)m_Poses.size(); obj++)
		{
			Model3D * m = m_pDB->GetModel(obj);
			GeomShape * densepntsshape = m->GetGeomShape("densepnts");				assert(densepntsshape!=NULL);
			ShapePointCloud * densepnts = densepntsshape->SafeCastToPointCloud(); 	assert(densepnts!=NULL);
			const R3Kdtree<const PointInSet *> * kdTree = densepnts->GetKdTree();
			assert(pose>=0 && pose < (int)m_Poses[obj].size());
			m_Gt2DensePnts.push_back(new int[(int)m_Poses[obj][pose]->NMatches()]);
			for (int r=0; r<(int)m_Poses[obj][pose]->NMatches(); r++)
			{
				if (!m_Poses[obj][pose]->Prim(r).IsType(TargetPrimitive::PRIM_Point))
				{
					assert(m_Poses[obj][pose]->Prim(r).IsType(TargetPrimitive::PRIM_Ground));
					continue;	// ignore ground
				}
				R3Point pos = m_Poses[obj][pose]->Prim(r).GetPosition();
				const PointInSet * closest = kdTree->FindClosest(pos);
				assert(closest!=NULL);
				m_Gt2DensePnts[obj][r] = closest->id;
			}
		}
	}
	
	assert(obj < (int)m_Gt2DensePnts.size() && pose==0);
	return m_Gt2DensePnts[obj][region];
}

int Classifier::GetNearestDenseAmongPnts(int obj, int pntid)
{
	if (m_Sparse2DensePnts.size()==0)
	{
		for (int obj=0; obj<(int)m_Poses.size(); obj++)
		{
			Model3D * m = m_pDB->GetModel(obj);
			GeomShape * densepntsshape = m->GetGeomShape("densepnts");				assert(densepntsshape!=NULL);
			ShapePointCloud * densepnts = densepntsshape->SafeCastToPointCloud(); 	assert(densepnts!=NULL);
			GeomShape * pntsShape = m->GetGeomShape("featpnts");					assert(pntsShape!=NULL);
			ShapePointCloud * pnts = pntsShape->SafeCastToPointCloud(); 			assert(pnts!=NULL);
			const R3Kdtree<const PointInSet *> * kdTree = densepnts->GetKdTree();
			assert(m_Poses[obj].size()>=1);
			int pose = 0;
			m_Sparse2DensePnts.push_back(new int[pnts->NumPoints()]);
			for (int p=0; p<pnts->NumPoints(); p++)
			{
				R3Point pos = pnts->PointPosition(p);
				const PointInSet * closest = kdTree->FindClosest(pos);
				assert(closest!=NULL);
				m_Sparse2DensePnts[obj][p] = closest->id;
			}
		}
	}
	
	assert(obj < (int)m_Sparse2DensePnts.size());
	return m_Sparse2DensePnts[obj][pntid];
}





