#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"

#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ShapeTriangleSoup.h"

#include "InteractorPose.h"
#include "PntCorrsPose.h"

#include "TimeProfiler.h"
#include "IKSolver.h"

#ifndef __Classifier_H
#define __Classifier_H

//class GTPose
//{
//	public:
//		std::vector<int> m_RegionID;
//		std::vector<std::vector<ShapePoint*> > m_PoseRegions;
//};

class TrainingData
{
	public:
		std::vector<double *> m_aFeatures;
		std::vector<int> m_Labels;
		std::vector<double> m_MembershipValue;
};

enum ClassifierType
{
	CLA_None,
	CLA_Dist2Nearest,
	CLA_Dist2NearestGauss,
	CLA_JointBoost,
	CLA_Weka,
	CLA_External,
	CLA_GroupWeka
};

enum TrainingDataChoice
{
	TRAIN_GaussianAll,
	TRAIN_GaussianRandomAndLabeled,
	TRAIN_GaussianLabeledOnly,
	TRAIN_GaussianRandom
};

struct TrainingParams
{
	TrainingParams();
	
	static TrainingDataChoice Str2TrainDataSetting(const VKString & str);
	static VKString TrainDataChoice2Str(TrainingDataChoice dataChoice);
	
	ClassifierType m_Classifier;
	TrainingDataChoice m_TrainingData;
	double m_fDistThreshold;
	double m_fDist2SigmaScale;
	double m_fPositiveValCutOff;	// alternative to defining sigma
	double m_fRatioRandomTrainingData;
};



/**
 * Classifier
 */
class Classifier
{
	public:
		friend class VisClassifierApp;
	
	
		static VKString Type2Str(ClassifierType type);
		static ClassifierType Str2Type(const VKString & type);
	
		Classifier();
		ModelDatabase * GetDB();
	
	// main algorithms:
		virtual void Train(const TrainingParams & params );
		virtual void Classify(const VKString & objName);
		virtual void SetIKSolver(IKSolver * solver);
	
	// read / write data
		// writer -> files map
		virtual void WriteAll() const;
		virtual void AddWriteMap(const VKString & datatype, const VKString & filename);
		virtual void AddWriteMap(const VKString & objName, const VKString & datatype, const VKString & filename);
		
		// classifier level
		virtual void ReadClassifier(const VKString & dirname);
		virtual void WriteClassifier(const VKString & dirname) const;
		virtual void ReadObjectNames(const VKString & filename);
		virtual void WriteObjectNames(const VKString & filename) const;
		virtual void ReadTrainingData(const VKString & filename);
		virtual void WriteTrainingData(const VKString & filename) const;
		virtual void ReadSymmetryEquivalence(const VKString & filename) ;
	
		virtual void WriteWekaTrainingData(const VKString & dirname) const;
		virtual void WriteWekaARFF(const VKString & filename, TrainingData * data) const;
	
		virtual void ReadFeatureMap(const VKString & filename);
		virtual void ReadGTRegionNames(const VKString & filename);
	
		// object level
		virtual void ReadMesh(const VKString & objName, const VKString & filename);
		virtual void ReadFeaturePnts(const VKString & objName, const VKString & filename);
		virtual void ReadFeatureVals(const VKString & objName, const VKString & valname, const VKString & filename);
		virtual void ReadGTRegions(const VKString & objName, const VKString & filename);
		virtual void ReadCanonicalXForm(const VKString & objName, const VKString & filename);
		virtual void ReadMetric(const VKString & objName, const VKString & shapeFile, const VKString & metricFile);
		virtual void ReadEmpty(const VKString & objName);
	
		virtual void ReadWekaResults(const VKString & objName, const VKString & filename);
	
		virtual void WriteWekaUnclassified(const VKString & objName, const VKString & filename) const;
		virtual void ReadPrediction(const VKString & objName, const VKString & filename);
		virtual void WritePrediction(const VKString & objName, const VKString & filename) const;
	
		virtual Model3D * GetModel(const VKString & objName);

		virtual void GetPredictionConfidenceBounds(const VKString & objName, double ** minvals, double ** maxvals) const;
		virtual int NFeat() const;
		virtual int NLabels() const;
		virtual VKString Label(int l) const;
		
		virtual double GetScale(int i) const;
	
	protected:
	// training data - read names only
		struct ObjData
		{
			VKString m_sObjName;
			VKString m_sMeshFile;
			VKString m_sFeaturePntsFile;
			std::map<VKString, VKString> m_sFeatureName2File;
			VKString m_sGTRegionFile;
			VKString m_sCanonicalXFormFile;
			VKString m_sMetricFile;
			VKString m_sMetricShapeFile;
		};
		std::map<VKString, int> m_ObjName2ObjReadFiles;
		std::vector<ObjData> m_DatasetFiles;
		virtual ObjData & GetObjData(const VKString & objName);
	
	// classifier essentials
		ModelDatabase * m_pDB;			// data holder
		VKStringList m_FeatureMap;		// which features to use
		std::vector<int> m_FeatureIndexMap;	// offset for each feature value
		ClassifierType m_Type;
		VKString m_sMetricName;
	
	// ground truth data
		std::vector<std::vector<PntCorrsPose * > > m_Poses;
		std::vector<double> m_Scales;
		std::vector<double> m_AnglesUp;
		std::vector<double> m_AveRad;
		std::vector<R3Vector> m_Translate;
	
		VKStringList m_sRegionNames;
		VKStringList m_sRegionTypes;
		std::vector<int> m_RegionNPoints;
	
		// convert ground truth to training data: label - class id, subsample (max number of examples), changeLabelID (new label ID)
		virtual int NPositiveExamples(int label) const;
		virtual void FillOnSurfaceMembershipValue(int label, const TrainingParams & params);
		virtual void AddTrainingExamples(TrainingData & training, int symmid, const std::vector<int> & symmetryGroup,
										 const TrainingParams & params, int & npositive, int & nnegative);
		//virtual void AddMemberValueExample(TrainingData & training, int label, double sigma);
		virtual void FillFeatVals(int objID, const ShapePoint & pnt, double * features);
		virtual void FillClassifyPoint(const double * features, double * prediction);
		virtual void FillFeatureVariance();
	
		virtual int GetNearestDenseAmongGT(int obj, int pose, int region);
		virtual int GetNearestDenseAmongPnts(int obj, int pntid);
	
	// writer
		std::map<VKString, VKString> m_Data2File;
		std::map<VKString, int> m_ObjName2Maps;
		std::vector<std::map<VKString, VKString> > m_ObjData2File;
	
	// Classifiers
		std::vector<TrainingData *> m_TrainingData;		// for all classifiers
		double * m_aFeatureVariance;	// variance in features (for CLA_Dist2NearestGauss classifier)
	
	// Symmetry-invariant classifiers
		virtual int GetSymmEqClass(int labelID, bool * mainName) const;
		std::map<VKString, int> m_Name2SymmClass;
		VKStringList m_SymmEqClassName;
	
	// debug: ground truth predictions
		enum MembershipType
		{
			MEMBER_NoFlags	   = 0,
			MEMBER_RegPoint    = 1<<0,
			MEMBER_NearGtPOI   = 1<<1,
			MEMBER_NearGtOther = 1<<2,
			MEMBER_Inlcuded    = 1<<3,
			MEMBER_NegExample  = 1<<4,
		};
		std::vector<double *> m_apGTPredictions;
		std::vector<int*> m_aMembershipTypes;
		std::vector<int*> m_aMap2TrainingData;
		std::vector<int> m_NElements;
	
	// mapping to dense points
		std::vector<int *> m_Gt2DensePnts;
		std::vector<int *> m_Sparse2DensePnts;
	
		TimeProfiler m_Timer;
	
	// per region stats
		virtual void FillPerRegionStats();
		virtual void WritePerRegionStats(const VKString & filename) const;
		virtual void ReadPerRegionStats(const VKString & filename);
		int m_iNPoses;
		std::vector<int> m_RegionNGround;
		std::vector<int> m_RegionNSurface;
	
		std::vector<std::vector<double> > m_PairwiseDists;
		std::vector<int> m_PairwiseFrom;
		std::vector<int> m_PairwiseTo;
		IKSolver * m_pSolver;
};

#endif



