#include "RendererGLUT.h"
#include "RendererOSMesa.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ShapeDistance.h"
#include "VKString.h"
#include "TimeProfiler.h"
#include "FileManagement.h"

#include "DBApp.h"
#include "AppScreenshots.h"

#include "Classifier.h"
#include "VisClassifierApp.h"
#include "IKSolver.h"

char** ReadArgFile(const VKString & appName, const VKString & argFile, int * outArgc)
{
	ifstream textStream(argFile.c_str());
	assert(textStream.is_open());
	VKString entireFile;
	entireFile.readFromFile(textStream);
	VKStringList splitFile = entireFile.split(" ");
	*outArgc = 1+splitFile.count();
	char ** buffer = new char*[splitFile.count()+1];

	for (int i=0; i<splitFile.count()+1; i++)
	{
		VKString str;
		if (i==0)
			str = appName;
		else
			str = splitFile[i-1];
		buffer[i] = new char[str.length()+1];
		const char * toCopy = str.c_str();
		for (int c=0; c<str.length(); c++)
			buffer[i][c] = toCopy[c];
		buffer[i][str.length()] = 0;
	}
	textStream.close();
	
	return buffer;
}

int main(int argc, char** argv )
{
	if (argc<2)
	{
		std::cout<<"Usage: ./"<<argv[0]<<"[params]"<<std::endl;
		std::cout<<"Params:"<<std::endl;
		std::cout<<"    -w/r type filename - write/read file of 'type' to 'filename'. Types:"<<std::endl;
		std::cout<<"         featnames - read names of feature dimensions"<<std::endl;
		std::cout<<"         gtnames - names of ground truth points"<<std::endl;
		std::cout<<"         classifier - read region classifier"<<std::endl;
		std::cout<<"         symm_invariant - read symmetry-equivalence groups (for training)"<<std::endl;
		std::cout<<"         regionNames - read region names\n";
		std::cout<<"         skeleton - read skeleton\n";
		std::cout<<"    -ro objName type [opt] filename - read file for object. Types:"<<std::endl;
		std::cout<<"         mesh - read object mesh"<<std::endl;
		std::cout<<"         featpnts - read feature points"<<std::endl;
		std::cout<<"         featvals featname - read feature values (also need a name)"<<std::endl;
		std::cout<<"         gtregions - read ground truth regions"<<std::endl;
		std::cout<<"         canonical - read  canonical xform"<<std::endl;
		std::cout<<"         weka_classified - read results printed by weka"<<std::endl;
		std::cout<<"         metric shapefile - read distance metric"<<std::endl;
		std::cout<<"    -wo objName type [opt] filename - for object. Types:"<<std::endl;
		std::cout<<"         prediction - write predicted classes"<<std::endl;
		std::cout<<"         weka_unclassified - write weka ARFF file with no predictions"<<std::endl;
		std::cout<<"         img_predict - all predictions (as smooth values)"<<std::endl;
		std::cout<<"         img_combined - combined images"<<std::endl;
		std::cout<<"         img_segment - segmentation with best predictors (todo..)"<<std::endl;
		std::cout<<"    -vis"<<std::endl;
		std::cout<<"Algorithm params:"<<std::endl;
		std::cout<<"    -train cla  - train classifier cla=[Dist2Nearest, Dist2NearestGauss, JointBoost...]"<<std::endl;
		std::cout<<"    -classify objName  - predict"<<std::endl;
		std::cout<<"    -distThreshold threshold - threshold (metric has to be loaded -ro metric model.metric)"<<std::endl;
		exit(0);
	}
	else if (argc==2)
		argv = ReadArgFile(VKString(argv[0]), VKString(argv[1]), &argc);
	
	TrainingParams trainparams;
	Classifier * cla = new Classifier();
	VKString regionNamesFile = "none";
	VKString readSkeleton = "none";
	
	std::vector<VKString> screenshots;
	std::vector<VKString> screenshotFiles;
	bool vis = false;
	
	// read parameters
	for (int i=1; i<argc; i++)
	{
		if (strcmp(argv[i], "-vis")==0)
			vis = true;
		else if (strcmp(argv[i], "-randseed")==0)
			srand(atoi(argv[++i]));
		else if (strcmp(argv[i], "-r")==0)
		{
			if (strcmp(argv[i+1], "featnames")==0)
				cla->ReadFeatureMap(argv[i+2]);
			else if (strcmp(argv[i+1], "classifier")==0)
				cla->ReadClassifier(argv[i+2]);
			else if (strcmp(argv[i+1], "gtnames")==0)
				cla->ReadGTRegionNames(argv[i+2]);
			else if (strcmp(argv[i+1], "symm_invariant")==0)
				cla->ReadSymmetryEquivalence(argv[i+2]);
			else if (strcmp(argv[i+1], "regionNames")==0)
				regionNamesFile = argv[i+2];
			else if (strcmp(argv[i+1], "skeleton")==0)
				readSkeleton = argv[i+2];
			else
			{
				std::cout<<"[ERROR] Unknown parameter to read "<<argv[i+1]<<std::endl;
				assert(false);
			}
			i+=2;
		}
		else if (strcmp(argv[i], "-w")==0)
		{
			cla->AddWriteMap(argv[i+1], argv[i+2]);
			i+=2;
		}
		else if (strcmp(argv[i], "-ro")==0)
		{
			VKString objName(argv[i+1]);
			VKString dataType(argv[i+2]);
			if (dataType=="mesh")
				std::cout<<"M: "<<objName.c_str()<<", "<<std::flush;
				
			if (dataType=="mesh")
				cla->ReadMesh(objName, argv[i+3]);
			else if (dataType=="featpnts")
				cla->ReadFeaturePnts(objName, argv[i+3]);
			else if (dataType=="featvals")
				{cla->ReadFeatureVals(objName, argv[i+3], argv[i+4]); 	i+=1;}
			else if (dataType=="gtregions")
				cla->ReadGTRegions(objName, argv[i+3]);
			else if (dataType=="canonical")
				cla->ReadCanonicalXForm(objName, argv[i+3]);
			else if (dataType=="weka_classified")
				cla->ReadWekaResults(objName, argv[i+3]);
			else if (dataType=="prediction")
				 cla->ReadPrediction(objName, argv[i+3]);
			else if (dataType=="metric")
				{cla->ReadMetric(objName, argv[i+3], argv[i+4]); 	i+=1;}
			else if (dataType=="empty")
				{cla->ReadEmpty(objName); 	i-=1;}
			else
			{
				std::cout<<"[ERROR] Invalid argument type: "<<dataType.c_str()<<std::endl;
				assert(false);
			}
			i+=3;
		}
		else if (strcmp(argv[i], "-wo")==0)
		{
			VKString objName(argv[i+1]);
			VKString dataType(argv[i+2]);

			if (dataType.startsWith("img_"))	// screenshot
			{
				screenshots.push_back(dataType+"#"+objName);
				screenshotFiles.push_back(argv[i+3]);
			}
			else
				cla->AddWriteMap(objName, dataType, argv[i+3]);
			i+=3;
		}
		else if (strcmp(argv[i], "-train")==0)
		{
			if (regionNamesFile != "none" && readSkeleton!="none")
			{
				IKSolver * solver = new IKSolver(1., 0, 0, NULL);
				solver->ReadSkeleton(readSkeleton, regionNamesFile);
				cla->SetIKSolver(solver);
			}

			cla->Train(trainparams);
		}
		else if (strcmp(argv[i], "-trainCla")==0)
			trainparams.m_Classifier = Classifier::Str2Type(VKString(argv[++i]));
		else if (strcmp(argv[i], "-trainDistThresh")==0)
			trainparams.m_fDistThreshold = atof(argv[++i]);
		else if (strcmp(argv[i], "-trainPositiveValCutoff")==0)
			trainparams.m_fPositiveValCutOff = atof(argv[++i]);
		else if (strcmp(argv[i], "-trainDataChoice")==0)
		{
			trainparams.m_TrainingData = TrainingParams::Str2TrainDataSetting(argv[++i]);
			if (trainparams.m_TrainingData==TRAIN_GaussianRandomAndLabeled)
				trainparams.m_fRatioRandomTrainingData = atof(argv[++i]);
			else if (trainparams.m_TrainingData==TRAIN_GaussianRandom)
				trainparams.m_fRatioRandomTrainingData = atof(argv[++i]);
		}
		else if (strcmp(argv[i], "-classify")==0)
			cla->Classify(VKString(argv[++i]));
		else
		{
			std::cout<<"[FnRegionClassifier.cpp] Invalid Argument: "<<argv[i]<<std::endl;
			assert(false);
		}
	}
	cla->WriteAll();
	
	
#ifdef VISUALIZATION
	if (vis || screenshots.size()>0)
	{
		int listFlags = DatabaseVisWindow::LISTGP_Models | DatabaseVisWindow::LISTGP_Geoms;
		listFlags |= DatabaseVisWindow::LISTGP_MeshFlags | DatabaseVisWindow::LISTGP_GeomVals;
		int w=-1, h=-1;
		if (screenshots.size()>0)
		{  w=1600; h=1200; }
		DatabaseVisWindow * dbVis = new DatabaseVisWindow(cla->GetDB(), NULL, listFlags, 1, 1, 0, 0, 1., w, h);
		VisClassifierApp * claApp = new VisClassifierApp(cla);
		
		if (screenshots.size()==0)
			dbVis->SetApp(claApp);
		else
		{
			AppScreenshots * appScreen = new AppScreenshots(true);
			
			for (int img=0; img < (int)screenshots.size(); img++)
			{
				if (screenshots[img].startsWith("img_predict"))	// need to know number of feature dimensions
				{
					VKString objName = screenshots[img];
					objName.replace("img_predict#", "");
					double * minvals;
					double * maxvals;
					cla->GetPredictionConfidenceBounds(objName, &minvals, &maxvals);
					for (int feat=0; feat<cla->NLabels(); feat++)
					{
						if (minvals[feat]<maxvals[feat])
						{
							VKString fileName = screenshotFiles[img]+VKString::number(feat)+".jpg";
							appScreen->TakeShotOfApp(claApp, screenshots[img], feat, fileName);
						}
					}
					
					FileManagement::CreateDirsOnPathToFile((screenshotFiles[img]+"/info.txt").c_str());
					std::ofstream textStream((screenshotFiles[img]+"/info.txt").c_str());
					if (!textStream.is_open())
						std::cout<<"[ERROR] Cannot create: "<<(screenshotFiles[img]+"/info.txt").c_str()<<std::endl;
					assert(textStream.is_open());
					
					int nlab = cla->NLabels();
					for (int l=0; l<nlab; l++)
						textStream<<cla->Label(l).c_str()<<" "<<minvals[l]<<" "<<maxvals[l]<<"\n";
					textStream.close();
				}
				else if (screenshots[img].startsWith("img_combined"))	// need to know number of feature dimensions
				{
					VisClassifierApp::s_Thresholds.push_back(0.2);
					VisClassifierApp::s_Thresholds.push_back(0.3);
					VisClassifierApp::s_Thresholds.push_back(0.4);
					VisClassifierApp::s_Thresholds.push_back(0.5);

					for (int i=0; i<(int)VisClassifierApp::s_Thresholds.size(); i++)
					{
						double t = VisClassifierApp::s_Thresholds[i];
						appScreen->TakeShotOfApp(claApp, "img_combined", i, screenshotFiles[img]+VKString::number(t)+".jpg");
					}
				}
				else
					appScreen->TakeShotOfApp(claApp, screenshots[img], 0, screenshotFiles[img]);
			}
			dbVis->SetApp(appScreen);
		}
		
#	ifdef MESA_RENDERING
		VisWindowRenderer * wndMesa = new RendererOSMesa(dbVis);
		wndMesa->UpdateScreen();
#	else
		VisWindowRenderer * wndGLUT = new RendererGLUT(argc, argv, dbVis);
#   endif
	}
#else
	std::cout<<"[WARNING] Visualization disabled (compiling flag)"<<std::endl;
#endif
}








