#include "ClusterPriors.h"

// default params
ClusterGridParams::ClusterGridParams()
{
	m_fValCap = 0.005;
}

ClusterGrid::ClusterGrid(int ndims, double * cellSize, double * domainMin, double * domainMax)
{
	m_aNBins = new int[ndims];
	m_afDomainMin = domainMin;
	m_afDomainMax = domainMax;
	m_afCellSize = cellSize;

	m_1DGrid = NULL;
	m_pR3Grid = NULL;

	m_iNDims = ndims;
	if (m_iNDims==1)
	{
		m_aNBins[0] = (domainMax[0] - domainMin[0])/cellSize[0];
		m_1DGrid = new double[m_aNBins[0]];
	}
	else if (m_iNDims==3)
	{
		R3Box bbox(domainMin[0], domainMin[1], domainMin[2], domainMax[0], domainMax[1], domainMax[2]);
		for (int d=0; d<3; d++)
			m_aNBins[d]= (domainMax[d] - domainMin[d]) / cellSize[d];
		m_pR3Grid = new R3Grid(m_aNBins[0], m_aNBins[1], m_aNBins[2], bbox);
		
	}
	else
	{
		std::cout<<"[ERROR] Unsupported dimensionality: "<<ndims<<std::endl;
		assert(false);
	}
}

ClusterGridParams & ClusterGrid::Params()
{
	return m_Params;
}

void ClusterGrid::Cluster(ClusteringProblem & problem)
{
	assert(problem.NFeatures()==m_iNDims);
	// go over all features -- add to grid bin
	LinAlgMatrixReal * input = Features(problem);
	if (m_1DGrid!=NULL)
	{
		for (int i=0; i<m_aNBins[0]; i++)
			m_1DGrid[i] = 0;
	}
	if (m_pR3Grid!=NULL)
		m_pR3Grid->Clear();
	for (int i = 0; i < problem.NNodes(); i++)
	{
		if (m_iNDims==1)
		{
			int bin = GetBin(input->GetVal(i, 0), m_afDomainMin[0], m_afDomainMax[0], m_aNBins[0]);
			m_1DGrid[bin]+=1.;
		}
		else if (m_iNDims==3)
		{
			R3Point gp = m_pR3Grid->GridPosition(R3Point(input->GetVal(i,0), input->GetVal(i,1), input->GetVal(i,2)));
			m_pR3Grid->AddGridValue((int)gp.X(), (int)gp.Y(), (int)gp.Z(), 1.);
		}
		else
			assert(false);
	}
	
	// blur?
	
	// go over all bins -- extract local maximas ('locality' relates to cellSize)
	std::vector<double*> centroids;
	int nclusters = 0;
	int nconsidered = 0;
	double valcap = (double)problem.NNodes() * m_Params.m_fValCap; // vk: const, todo: parameter
	if (m_1DGrid!=NULL)
	{
		for (int i=1; i<m_aNBins[0]-1; i++)
		{
			if (m_1DGrid[i] > valcap && m_1DGrid[i]>m_1DGrid[i+1] && m_1DGrid[i]>m_1DGrid[i-1])
			{
				//double value = (double)i / (double)m_iN1DBins[0]
				nclusters++;
				int cid = (int)centroids.size();
				centroids.push_back(new double[1]);
				double t = (double)i / (double)m_aNBins[0];
				centroids[cid][0] = (1.-t)*m_afDomainMin[0] + t * m_afDomainMax[0];
			}
			nconsidered++;
		}
	}
	else if (m_pR3Grid!=NULL)
	{
		for (int x=1; x<m_pR3Grid->XResolution()-1; x++)
		for (int y=1; y<m_pR3Grid->YResolution()-1; y++)
		for (int z=1; z<m_pR3Grid->ZResolution()-1; z++)
		{
			bool localmax = true;
			double currval = m_pR3Grid->GridValue(x, y, z);
			localmax = currval > valcap;
			for (int xl=-1; xl<=1 && localmax; xl++)
			for (int yl=-1; yl<=1 && localmax; yl++)
			for (int zl=-1; zl<=1 && localmax; zl++)
			{
				if (xl==0 && yl==0 && zl==0)
					continue;
				if (m_pR3Grid->GridValue(x+xl, y+yl, z+zl) >= currval)
					localmax = false;
			}
			if (localmax)
			{
				nclusters++;
				int cid = (int)centroids.size();
				centroids.push_back(new double[m_iNDims]);
				for (int d=0; d<3; d++)
				{
					double t = (double)(d==0 ? x : (d==1 ? y : z)) / (double)m_aNBins[d];
					centroids[cid][d] = (1.-t)*m_afDomainMin[d] + t * m_afDomainMax[d];
				}
			}
			nconsidered++;
		}
	}
	
	//std::cout<<"Simple local max nclusters: "<<nclusters<<" / "<<nconsidered<<" (ndims="<<m_iNDims<<") "<<std::endl;
	// TODO: final pruning:
	//	  -- distance to neighbor (based on gridcell)
	
	// cluster based on distances to centroids
	std::vector<std::vector<ClusteredNode> > & output = Clusters(problem);
	output.clear();
	for (int i=0; i<(int)centroids.size(); i++)
		output.push_back(std::vector<ClusteredNode>());
	
	for (int n=0; n<(int)problem.NNodes(); n++)
	{
		int bestCentroid = -1;
		double minDistance = RN_INFINITY;
		for (int i=0; i<(int)centroids.size(); i++)
		{
			double dist = 0;
			for (int d=0; d<m_iNDims; d++)
				dist += pow(input->GetVal(n, d) - centroids[i][d], 2.);
			if (dist < minDistance || bestCentroid==-1)
			{
				minDistance = dist;
				bestCentroid = i;
			}
		}
		ClusteredNode node;
		node.id = n;
		node.error = (float)minDistance;
		output[bestCentroid].push_back(node);
	}
}

int ClusterGrid::GetBin(double val, double minval, double maxval, int ncells)
{
	int bin =  ncells * (val - minval) / (maxval - minval);
	if (bin>=ncells)
		bin = ncells-1;
	assert(bin>=0);
	return bin;
}

