#include "FileManagement.h"
#include "VKString.h"
#include "Clustering.h"
#include "TimeProfiler.h"

#include "RendererOSMesa.h"
#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ShapeDistance.h"

#include "ShapeTriangleSoup.h"
#include "DBApp.h"
#include <vector>
#include "ShapePoint.h"
#include "IKSolver.h"
#include "PosePrior.h"
#include "PosePriorMixGauss.h"
#include "BindingManager.h"

#include "AppScreenshots.h"

#include "InteractorPose.h"
#include "VisPriorApp.h"
#include "ClusterPriors.h"
#include "PntCorrsPose.h"

struct PosePriorParams
{
	PosePriorParams();
	
	PosePrior::PosePriorType m_Type;
	bool m_bCoupled;
	
	VKString m_sUpdateMagnitude;
	VKString m_sUpdateVariance;
	VKString m_sClusterAlg;
	
	double m_fMeanShiftSupport;
	double m_fConstVariance;
	bool m_bIgnorePalmsAndToes;
};

PosePriorParams::PosePriorParams()
{
	m_Type = PosePrior::PRIOR_GaussAngles;
	m_sUpdateMagnitude = "none";
	m_sUpdateVariance = "none";
	m_sClusterAlg = "meanshift";
	
	m_fMeanShiftSupport = M_PI/20.;
	m_fConstVariance = -1;
	m_bIgnorePalmsAndToes = false;
	m_bCoupled = false;
}

PosePrior * CreatePriorMultiGauss(std::vector<PntCorrsPose*> & trainPoses, IKSolver * solver,
								  std::vector<ClusteringProblem*> & perLinkClusters,
								  const PosePriorParams & params);

PosePrior * CreatePriorMultiGaussCoupled(std::vector<PntCorrsPose*> & trainPoses, IKSolver * solver,
										 std::vector<ClusteringProblem*> & perLinkClusters,
										 const PosePriorParams & params, VKStringList & posenames);

void CreateGrids(std::vector<PntCorrsPose*> & trainPoses, PosePrior * prior, IKSolver ** saveNewSolver,
				 const VKString & skeletonFile, const VKString &regionsFile, EnergyParams & eparams);

int main(int argc, char** argv )
{
	if (argc<2)
	{
		std::cout<<"Usage: ./"<<argv[0]<<" prior.bin [options]"<<std::endl;;
		std::cout<<"Params:"<<std::endl;;
		std::cout<<"   -priortype [gauss1, multigauss, multigausscoupled]: type of shape prior"<<std::endl;
		std::cout<<"   -t pose.gt: training pose"<<std::endl;
		std::cout<<"   -skeleton skeleton.txt regionsFile.txt: read skeleton"<<std::endl;
		std::cout<<"   -writeTiming timingFile "<<std::endl;
		std::cout<<"   -comFile comFile "<<std::endl;
		std::cout<<"   -vis:  visualize"<<std::endl;
		std::cout<<"PriorParams:"<<std::endl;
		std::cout<<"   -ignorePalmsAndToes "<<std::endl;
		std::cout<<"   -updateMagnitude [densityRatio2Best] : set gaussian magnitudes "<<std::endl;
		std::cout<<"   -updateVariance  [const] value : update variances "<<std::endl;
		std::cout<<"   -clusterAlg  [meanshift] : clustering algorithm "<<std::endl;
		std::cout<<"   -meanShiftSupport h : mean shift parameter  "<<std::endl;
		std::cout<<"   -param (energy params), see FnAlign"<<std::endl;
		exit(0);
	}
	
	EnergyParams eparams;
	PosePriorParams params;
	VKString outfile(argv[1]);

	IKSolver * solver = NULL;
	std::vector<PntCorrsPose*> trainPoses;
	bool vis = false;
	TimeProfiler timer;
	VKStringList files2read;
	VKString writeTiming = "none";
	VKString skeletonFile, regionsFile;
	VKStringList comFiles;
	
	for (int i=2; i<argc; i++)
	{
		if (strcmp(argv[i], "-priortype")==0)
		{
			VKString typeStr = argv[++i];
			if (typeStr=="gauss1")
				params.m_Type = PosePrior::PRIOR_GaussAngles;
			else if (typeStr=="multigauss")
				params.m_Type = PosePrior::PRIOR_MultiGauss;
			else if (typeStr=="multigausscoupled")
			{
				params.m_Type = PosePrior::PRIOR_MultiGauss;
				params.m_bCoupled = true;
			}
			else
				assert(false);
		}
		else if (strcmp(argv[i], "-randseed")==0)
			srand(atoi(argv[++i]));
		else if (strcmp(argv[i], "-t")==0)
			files2read.push_back(argv[++i]);
		else if (strcmp(argv[i], "-clusterAlg")==0)
			params.m_sClusterAlg = VKString(argv[++i]);
		else if (strcmp(argv[i], "-comFile")==0)
			comFiles.push_back(VKString(argv[++i]));
		else if (strcmp(argv[i], "-meanShiftSupport")==0)
			params.m_fMeanShiftSupport = atof(argv[++i]);
		else if (strcmp(argv[i], "-updateMagnitude")==0)
			params.m_sUpdateMagnitude = VKString(argv[++i]);
		else if (strcmp(argv[i], "-ignorePalmsAndToes")==0)
			params.m_bIgnorePalmsAndToes = true;
		else if (strcmp(argv[i], "-updateVariance")==0)
		{
			params.m_sUpdateVariance = VKString(argv[++i]);
			if (params.m_sUpdateVariance=="const")
				params.m_sUpdateVariance = atof(argv[++i]);
		}
		else if (strcmp(argv[i], "-skeleton")==0)
		{
			skeletonFile = argv[i+1];
			regionsFile = argv[i+2];
			i+=2;
		}
		else if (strcmp(argv[i], "-writeTiming")==0)
			writeTiming = VKString(argv[++i]);
		else if (strcmp(argv[i], "-vis")==0)
			vis = true;
		else if (strcmp(argv[i+1], "-readRegionStats")==0)
		{ 	eparams.m_sRegionStats = argv[i+2]; i+=2; }
		else if (strcmp(argv[i], "-param")==0)
		{
			if (strcmp(argv[i+1], "wPosePrior")==0)
			{	 eparams.m_wPosePrior = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wEndEffectors")==0)
			{	 eparams.m_wEndEffectorDist = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wSymmetryPrior")==0)
			{	 eparams.m_wSymmetryPrior = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wFeatures")==0)
			{	 eparams.m_wFeatures = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wUnassigned")==0)
			{	 eparams.m_wUnassigned = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wGround")==0)
			{	 eparams.m_wGround = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wContactPairwise")==0)
			{	 eparams.m_wContactPairwise = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wSymmPlane")==0)
			{	 eparams.m_wSymmPlane = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "wIsectPenalty")==0)
			{	 eparams.m_wIsectPenalty = atof(argv[i+2]);	i+=2;	}
			else if (strcmp(argv[i+1], "minGroundProb")==0)
			{	 eparams.m_fMaxGroundPenalty = -log(atof(argv[i+2]));	i+=2;	}
			else if (strcmp(argv[i+1], "minContactProb")==0)
			{	 eparams.m_fMaxContactPenalty = -log(atof(argv[i+2]));	i+=2;	}
			else if (strcmp(argv[i+1], "energy")==0)	// energy flags
			{
				int fid = 2;
				eparams.ClearEnergyFlags();
				while (i + fid < argc)
				{
					EnergyParams::EnergyFlags flag = EnergyParams::Str2EnergyFlag(argv[i+fid]);
					if (flag != EnergyParams::E_Invalid)
					{	eparams.AddEnergyFlag(flag);	fid++;	}
					else {	fid--; 	break; }
				}
				i+=fid;
			}
			else
			{
				std::cout<<"[ERROR] Unknown (param) argument: "<<argv[i+1]<<std::endl;
				assert(false);
			}
		}
		else
		{
			std::cout<<"[ERROR] Unknown argument: "<<argv[i]<<" in "<<argv[0]<<std::endl;
			assert(false);
		}
	}
	
	if (skeletonFile!="none")
	{
		solver = new IKSolver(1.,  0., 0., NULL);
		solver->ReadSkeleton(skeletonFile, regionsFile);
	}
	
	TimeProfiler profiler;
	profiler.startedProcess("trainingprior");
	
	VKStringList posenames;
	std::cout<<"Reading "<<files2read.count()<<" pose files"<<std::endl;
	assert(solver!=NULL);

	std::vector<PntCorrsPose*> unqueTraining;
	for (int i=0; i<files2read.count(); i++)
	{
		timer.WriteProgress("read", i, (int)files2read.count());
		int tp0 = (int)trainPoses.size();
		PntCorrsPose::CreatePosesFromGTFile(files2read[i], NULL, "none", trainPoses);
		int tp1 = (int)trainPoses.size();
		if (tp1>=tp0)
		{
			for (int id=tp0; id<tp1; id++)
				posenames.push_back(VKString::number(id)+"_"+files2read[i].FileName());
			unqueTraining.push_back(trainPoses[tp0]);
		}
	}
	
	std::vector<PntCorrsPose*> inputTraining = trainPoses;
	
	std::cout<<"Training from "<<trainPoses.size()<<" poses"<<std::endl;
	for (int i=0; i<trainPoses.size(); i++)
	{
		timer.WriteProgress("train", i, (int)trainPoses.size());
		solver->FixPoseParameterization(trainPoses[i]);
		solver->ForwardKinematics(trainPoses[i]);
	}
	std::cout<<std::endl;
	
	assert(params.m_Type==PosePrior::PRIOR_MultiGauss);
	std::vector<ClusteringProblem*> perLinkClusters;
	PosePrior * prior = NULL;
	
	if (params.m_Type == PosePrior::PRIOR_MultiGauss && !params.m_bCoupled)
		prior = CreatePriorMultiGauss(trainPoses, solver, perLinkClusters, params);
	else if (params.m_Type == PosePrior::PRIOR_MultiGauss && params.m_bCoupled)
		prior = CreatePriorMultiGaussCoupled(trainPoses, solver, perLinkClusters, params, posenames);
	else
		assert(false);
	prior->RecordPointStats(unqueTraining, solver->m_RegionNames.count(), comFiles);
	
	std::cout<<"Creating grids "<<std::endl;
	CreateGrids(trainPoses, prior, &solver, skeletonFile, regionsFile, eparams);
	
	profiler.finishedProcess("trainingprior");
	if (writeTiming!="none")
	{
		FileManagement::CreateDirsOnPathToFile(writeTiming.c_str());
		std::ofstream textStream(writeTiming.c_str());
		assert(textStream.is_open());
		textStream<<profiler.profile().c_str();
		textStream.close();
	}
	prior->WritePrior(outfile);
	prior->WritePointStats(outfile+".pntstats.txt");

#ifdef VISUALIZATION
	if (vis)
	{
		std::cout<<"Visualizing priors"<<std::endl;
		int w=-1, h=-1;
		int listFlags = DatabaseVisWindow::LISTGP_Models | DatabaseVisWindow::LISTGP_Geoms;
		listFlags |= DatabaseVisWindow::LISTGP_MeshFlags;
		ModelDatabase * db = new ModelDatabase();

		DatabaseVisWindow * dbVis = new DatabaseVisWindow(db, NULL, listFlags, 1, 1, 0, 0, 1., w, h);
		VisPriorApp * priorApp = new VisPriorApp(trainPoses, solver);
		priorApp->SetVisPriors(prior, perLinkClusters);
		dbVis->SetApp(priorApp);
#ifdef MESA_RENDERING
		VisWindowRenderer * mesa = new RendererOSMesa(dbVis);
		mesa->UpdateScreen();
#else
		VisWindowRenderer * wndGLUT = new RendererGLUT(argc, argv, dbVis);
#endif
	}
#else
	std::cout<<"[WARNING] Visualization disabled (compiling flag)"<<std::endl;
#endif
	return 0;
}


PosePrior * CreatePriorMultiGauss(std::vector<PntCorrsPose*> & trainPoses, IKSolver * solver,
								  std::vector<ClusteringProblem*> & perLinkClusters, const PosePriorParams & params)
{
	PosePriorMixGauss * mixPrior = new PosePriorMixGauss();
	mixPrior->SetIK(solver);
	mixPrior->InitDistributions();
	double * features = NULL;
	int currmaxdims = 0;
	std::vector<std::vector<int> > dnaStrings;
	for (int i=0; i<(int)trainPoses.size(); i++)
		dnaStrings.push_back(std::vector<int>());
	
   	// for each link
	for (int l=0; l<(int)solver->m_Links.size(); l++)
	{
		perLinkClusters.push_back(NULL);
		// get the link -- check number of parameters
		SkeletonLink * link = solver->m_Links[l];
		int ndims =  SkeletonLink::NDims(link->m_JointType);
		int offset = solver->m_ParamOffsetByLinkID[l];
		if (ndims==0)
			continue;
		
		// update feature storage
		if (ndims>currmaxdims)
		{
			if (features!=NULL)
				delete [] features;
			features = new double[ndims];
			currmaxdims = ndims;
		}
		
		// define clustering problem
		ClusteringProblem * problem = new ClusteringProblem((int)trainPoses.size(), ndims);
		for (int i=0; i<(int)trainPoses.size(); i++)
		{
			for (int d=0; d<ndims; d++)
				features[d] = trainPoses[i]->Param(offset+d);
			problem->SetFeatures(i, features);
		}
		
		ClusteringAlgorithm * cluster = NULL;
		if (params.m_sClusterAlg=="meanshift")
			cluster = new MeanShiftClustering(params.m_fMeanShiftSupport, 100000, 3);
		else if (params.m_sClusterAlg=="grid")
		{
			// cluster (simple -- merge nearest)
			double * cellSize = new double[ndims];
			double * domainMin = new double[ndims];
			double * domainMax = new double[ndims];
			for (int i=0; i<ndims; i++)
			{
				cellSize[i] = M_PI/20;
				domainMin[i] = -M_PI;
				domainMax[i] = M_PI;
			}
			cluster = new ClusterGrid(ndims, cellSize, domainMin, domainMax);
		}
		else
		{
			std::cout<<"[ERROR] Unknown clustering algorithm: "<<params.m_sClusterAlg.c_str()<<std::endl;
			assert(false);
		}
		
		cluster->Cluster(*problem);
		
		// write some stats for debugging...
		std::cout<<"Link: "<<link->m_pParent->m_sJointName.c_str()<<" -> "<<link->m_pChild->m_sJointName.c_str();
		std::cout<<" ("<<SkeletonLink::NDims(link->m_JointType)<<"):"<<std::endl;
		std::cout<<"     NClusters: "<<problem->NClusters()<<std::endl;
		std::cout<<"     Sizes: ";
		for (int c=0; c<problem->NClusters(); c++)
			std::cout<<problem->ClusterSize(c)<<" ";
		std::cout<<std::endl;
		
		//delete problem;
		//delete cluster;
		perLinkClusters[l]=problem;
		
		// create distribution from clustering
		std::vector<int> densities;
		double max_density = 0;
		std::vector<PosePointGauss*> gaussians;
		int nclusters = problem->NClusters();
		// hands / toes -- not specified in the dataset
		VKString childName = link->m_pChild->m_sJointName;
		bool palmOrToe = (childName == "right_toe") || (childName == "left_toe");
		palmOrToe = palmOrToe || (childName == "right_palm") || (childName=="left_palm");
		if (params.m_bIgnorePalmsAndToes && palmOrToe)
			nclusters = 1;
		for (int c=0; c<nclusters; c++)
		{
			PosePointGauss * gauss = new PosePointGauss(ndims, true);
			const std::vector<ClusteredNode> & cluster = problem->GetCluster(c);
			
			if (params.m_bIgnorePalmsAndToes && palmOrToe)
			{
				for (int d=0; d<ndims; d++)
					features[d] = 0;
				gauss->Update(features);
			}
			else
			{
				for (int i=0; i<(int)cluster.size(); i++)
				{
					for (int d=0; d<ndims; d++)
						features[d] = trainPoses[cluster[i].id]->Param(offset+d);
					gauss->Update(features);
				}
			}
			
			int currg = mixPrior->GetDistr(l)->NGaussians();
			for (int cid=0; cid<(int)cluster.size(); cid++)
			{
				assert(cluster[cid].id >= 0 && cluster[cid].id < (int)dnaStrings.size());
				while (l >= dnaStrings[cluster[cid].id].size() )
					dnaStrings[cluster[cid].id].push_back(-1);
				assert(l>=0 && l < (int)dnaStrings[cluster[cid].id].size());
				dnaStrings[cluster[cid].id][l] = currg;
			}
			
			densities.push_back((int)cluster.size());
			gaussians.push_back(gauss);
			max_density = vkMax(max_density, (int)cluster.size());
			
			mixPrior->AddGaussian(l, gauss);
		}
		if (params.m_sUpdateMagnitude=="densityRatio2Best")
		{
			for (int updG = 0; updG < (int)gaussians.size(); updG++)
				gaussians[updG]->m_fMagnitude = (double)densities[updG] / (double)max_density;
		}
		else
			assert(params.m_sUpdateMagnitude=="none");
		if (params.m_sUpdateVariance=="const")
		{
			for (int updG = 0; updG < (int)gaussians.size(); updG++)
			for (int d=0; d<ndims; d++)
				gaussians[updG]->m_aVariances[d] = params.m_fConstVariance;
		}
		else
			assert(params.m_sUpdateVariance=="none");
	}
	
	// update symmetries
	for (int i=0; i<(int)trainPoses.size(); i++)
		mixPrior->UpdateSymmetry(trainPoses[i], params.m_bIgnorePalmsAndToes);

	// update initial poses
	for (int i=0; i<(int)dnaStrings.size(); i++)
		mixPrior->UpdatePoseDNAs(dnaStrings[i]);

	if (features!=NULL)
		delete [] features;
	return mixPrior;
}


PosePrior * CreatePriorMultiGaussCoupled(std::vector<PntCorrsPose*> & trainPoses, IKSolver * solver,
										 std::vector<ClusteringProblem*> & perLinkClusters,
										 const PosePriorParams & params, VKStringList & posenames)
{
	PosePriorMixGauss * mixPrior = new PosePriorMixGauss();
	mixPrior->SetIK(solver);
	mixPrior->InitDistributions();
	int currmaxdims = 0;
	
	// cluster all poses
	int nfeatures = 0;
	for (int l=0; l<(int)solver->m_Links.size(); l++)
		nfeatures += SkeletonLink::NDims(solver->m_Links[l]->m_JointType);
	
	ClusteringProblem * problem = new ClusteringProblem((int)trainPoses.size(), nfeatures);
	
	double * features = new double[nfeatures];
	
	for (int i=0; i<(int)trainPoses.size(); i++)
	{
		int fid = 0;
		for (int l=0; l<(int)solver->m_Links.size(); l++)
		{
			int ndims =  SkeletonLink::NDims(solver->m_Links[l]->m_JointType);
			int offset = solver->m_ParamOffsetByLinkID[l];
			for (int d=0; d<ndims; d++)
			{
				features[fid] = trainPoses[i]->Param(offset+d);
				fid++;
			}
		}
		problem->SetFeatures(i, features);
	}
	
//	std::cout<<"Poses: "<<std::endl;
//	for (int i=0; i<posenames.count(); i++)
//		std::cout<<"    "<<posenames[i].c_str()<<std::endl;
	
//	std::cout<<"Mean Shift Clustering: "<<std::endl;
	ClusteringAlgorithm * cluster = NULL;
	if (params.m_sClusterAlg=="meanshift")
		cluster = new MeanShiftClustering(params.m_fMeanShiftSupport, 100000, 3);
	else
		assert(false);
	cluster->Cluster(*problem);
	
	// create gaussians
	int nclusters = problem->NClusters();
	for (int l=0; l<(int)solver->m_Links.size(); l++)
		perLinkClusters.push_back(problem);
	for (int c=0; c<nclusters; c++)
	{
		const std::vector<ClusteredNode> & cluster = problem->GetCluster(c);
		for (int l=0; l<(int)solver->m_Links.size(); l++)
		{
			int ndims =  SkeletonLink::NDims(solver->m_Links[l]->m_JointType);
			if (ndims==0)
				continue;
			int offset = solver->m_ParamOffsetByLinkID[l];

			PosePointGauss * gauss = new PosePointGauss(ndims, true);
			VKString childName = solver->m_Links[l]->m_pChild->m_sJointName;
			bool palmOrToe = (childName == "right_toe") || (childName == "left_toe");
			palmOrToe = palmOrToe || (childName == "right_palm") || (childName=="left_palm");

			for (int cid=0; cid<(int)cluster.size(); cid++)
			{
				for (int d=0; d<ndims; d++)
					features[d] = trainPoses[cluster[cid].id]->Param(offset+d);
				gauss->Update(features);
			}

			if (params.m_bIgnorePalmsAndToes && palmOrToe)
			{
				for (int d=0; d<ndims; d++)
					gauss->m_aVariances[d] = -1;
			}
			
			mixPrior->AddGaussian(l, gauss);
		}
	}

	// create dna strings
	std::vector<std::vector<int> > dnaStrings;	// dnaString[poseid][linkid] = gaussian id
	for (int c=0; c<nclusters; c++)
	{
		dnaStrings.push_back(std::vector<int>());
		for (int l=0; l<(int)solver->m_Links.size(); l++)
			dnaStrings[c].push_back(c);
	}

	// write clustering stats
	std::cout<<"NClusters: "<<nclusters<<std::endl;
	int nmaxshow = 100;
	
	for (int c=0; c<nclusters; c++)
	{
		const std::vector<ClusteredNode> & cluster = problem->GetCluster(c);
		std::cout<<"  C"<<c<<" ("<<cluster.size()<<"): ";
		for (int cid=0; cid<(int)cluster.size(); cid++)
		{
			if (cid < nmaxshow)
				std::cout<<posenames[cluster[cid].id].c_str()<<" ";
			else if (cid==nmaxshow)
				std::cout<<"... ";
		}
		std::cout<<std::endl;
	}
	
	for (int i=0; i<(int)dnaStrings.size(); i++)
		mixPrior->UpdatePoseDNAs(dnaStrings[i]);

	// update symmetries
	for (int i=0; i<(int)trainPoses.size(); i++)
		mixPrior->UpdateSymmetry(trainPoses[i], params.m_bIgnorePalmsAndToes);
	
	// update initial poses
	for (int i=0; i<(int)dnaStrings.size(); i++)
		mixPrior->UpdatePoseDNAs(dnaStrings[i]);

	delete [] features;
	assert(params.m_sUpdateVariance=="none");
	assert(params.m_sUpdateMagnitude=="none");
	
	for (int c=0; c<nclusters; c++)
	{
		const std::vector<ClusteredNode> & cluster = problem->GetCluster(c);
		for (int cid=0; cid<(int)cluster.size(); cid++)
			trainPoses[cluster[cid].id]->SetPoseID(c);
	}
	
	return mixPrior;
}


void CreateGrids(std::vector<PntCorrsPose*> & trainPoses, PosePrior * priorIn, IKSolver ** saveNewSolver,
				 const VKString & skeletonFile, const VKString &regionsFile, EnergyParams & eparams)
{
	assert(priorIn->IsAngles());
	PriorAngles * prior =  (PriorAngles*)priorIn;
	//std::cout<<"Weights: "<<eparams.m_wEndEffectorDist<<" "<<eparams.m_wPosePrior<<" "<<eparams.m_wSymmetryPrior<<std::endl;
	IKSolver * solver = new IKSolver(eparams.m_wEndEffectorDist, eparams.m_wPosePrior, eparams.m_wSymmetryPrior, prior);
	solver->ReadSkeleton(skeletonFile, regionsFile);
	prior->SetIK(solver);
	*saveNewSolver = solver;
	
	// create grids
	std::vector<double> energies;
	for (int i=0; i<(int)trainPoses.size(); i++)
		energies.push_back(RN_INFINITY);
	EnergyInfo info;
	for (int t=0; t<(int)trainPoses.size(); t++)
	{
		solver->GetEnergy(trainPoses[t], info);
		energies[t] = info.m_fPosePriorEnergy+info.m_fSymmetryEnergy;
	}
	
	int gridflags = PriorAngles::GRID_CondOnRefPnt | PriorAngles::GRID_CondOnPoseID;
	prior->FillCandidateRegions(trainPoses);
	bool trainFromInput = false;
	if (trainFromInput)	// debug only
		prior->FillGridFromClusteredPoses(gridflags, trainPoses);
	else
	{
		trainPoses.clear();
		prior->SamplePosesAndFillGrid(gridflags, 50000, &trainPoses);
	}
}




