#include "FileManagement.h"
#include "VKString.h"
#include "TimeProfiler.h"

#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ShapeDistance.h"

#include "ShapeTriangleSoup.h"
#include "IKSolver.h"
#include "DBApp.h"
#include <vector>
#include "ShapePoint.h"
#include "PosePriorMixGauss.h"
#include "PntCorrsPose.h"
#include "BindingEnergy.h"
#include "ShapePointCloud.h"

#include "VisPriorApp.h"
VisPriorApp::VisPriorApp(const std::vector<PntCorrsPose*> & poses, IKSolver * solver)
{
	//m_iVisFlags = VIS_Skeleton | VIS_SkeletonNames | VIS_LinkDistrPoints | VIS_LinkDistrAngles | VIS_MeanGauss | VIS_Clustering;
	m_iVisFlags = VIS_Skeleton | VIS_SkeletonNames | VIS_MeanGauss | VIS_Clustering | VIS_EndEffDistr | VIS_EndEffGridDistr;
	m_iVisFlags |= VIS_AveragePose;
	
	m_Poses = poses;
	m_pSolver = solver;
	m_iCurrPose = 0;
	m_iSelectedLink = -1;
	m_fDistrRadius = 0.;
	m_pDistrPoints = NULL;
	m_pAveragePose = NULL;
	m_pPosePrior = NULL;
	m_iNextInitPose = -1;
	m_iSelectedCOI = -1;
	m_iSelectedRefID = -1;
	m_fRetRotation = 0;
	m_fEnergyThreshold = 1.;	// show all
}

void VisPriorApp::SetVisPriors(PosePrior * prior, std::vector<ClusteringProblem*> & angleClusters)
{
	m_pPosePrior = prior;
	m_AngleClusters = angleClusters;
}

bool VisPriorApp::KeyPressed(DatabaseVisWindow * window, const char key, int mouseX, int mouseY)
{
	bool renewPnts = false;
	if (key=='`')
	{
		std::cout<<"HotKeys:"<<std::endl;
		std::cout<<"   `        this help"<<std::endl;
		std::cout<<"   s        toggle skeleton visualization"<<std::endl;
		std::cout<<"   S        toggle skeleton joint names "<<std::endl;
		std::cout<<"   g        toggle (mean) gaussians visualization"<<std::endl;
		std::cout<<"   a        toggle switch to average pose "<<std::endl;
		std::cout<<"   x/X      energy threshold "<<std::endl;
		std::cout<<"   ./,      change pose id"<<std::endl;
		std::cout<<"   h/H      change link id"<<std::endl;
		std::cout<<"   r/R      change radius (smooth visualiaztion of distirbutions)"<<std::endl;
	}
	else if (key=='s')
		Toggle(VIS_Skeleton);
	else if (key=='S')
		Toggle(VIS_SkeletonNames);
	else if (key=='x' || key=='X')
	{
		renewPnts = true;
		m_fEnergyThreshold += (key=='x' ? 1 : -1) * 0.05;
		if (m_fEnergyThreshold<0)
			m_fEnergyThreshold = 0;
		else if (m_fEnergyThreshold>1)
			m_fEnergyThreshold = 1;
		std::cout<<"EThreshold: "<<m_fEnergyThreshold<<std::endl;
	}
	else if (key=='g')
	{
		Toggle(VIS_MeanGauss);
		Toggle(VIS_EndEffGridDistr);
	}
	else if (key=='\'' || key==';')
	{
		m_iSelectedRefID += key==';' ? -1 : 1;
		std::cout<<"RefID: "<<m_iSelectedRefID<<std::endl;
		renewPnts = true;
	}
	else if (key=='[' || key==']')	// advance regions of interest
	{
		m_fRetRotation += (key=='[' ? 1 : -1) * 0.25 * M_PI;
		std::cout<<"RefRot: "<<m_fRetRotation<<std::endl;
	}
	else if (key=='=' || key=='-')
	
	{
		m_iSelectedCOI += key=='=' ? 1 : -1;
		int maxval = 0;
		for (int i=0; i<(int)m_PerPointCluster.size(); i++)
			maxval = vkMax(maxval, m_PerPointCluster[i]);
		if (m_iSelectedCOI<0)
			m_iSelectedCOI = -1;
		else if (m_iSelectedCOI>=maxval)
			m_iSelectedCOI = maxval;
		renewPnts = true;
		std::cout<<"Cluster of interest: "<<m_iSelectedCOI<<" / "<<maxval<<std::endl;
	}
	else if (key=='.' || key==',')	// advance pose
	{
		renewPnts = true;
		m_iCurrPose += key=='.'?1:-1;
		m_iCurrPose = (m_iCurrPose+(int)m_Poses.size()) % (int)m_Poses.size();
		std::cout<<"Pose: "<<m_iCurrPose<<" / "<<m_Poses.size()<<std::endl;
	}
	else if ((key=='h' || key=='H') && m_pSolver!=NULL)	// advance link
	{
		m_iSelectedLink += (key=='h' ? 1 : -1);
		if (m_iSelectedLink == (int)m_pSolver->m_Links.size())
			m_iSelectedLink = 0;
		else if (m_iSelectedLink<0)
			m_iSelectedLink = ((int)m_pSolver->m_Links.size())-1;
		if (m_iSelectedLink==-1)
			std::cout<<"Link: none"<<std::endl;
		else
		{
			SkeletonLink * l = m_pSolver->m_Links[m_iSelectedLink];
			std::cout<<"Link: "<<l->m_pParent->m_sJointName.c_str()<<" -> "<<l->m_pChild->m_sJointName.c_str();
			std::cout<<" ("<<SkeletonLink::NDims(l->m_JointType)<<")"<<std::endl;
		}
		renewPnts = true;
	}
	else if (key=='i')// pose info
	{
		if (m_iCurrPose>=0 && m_iCurrPose<(int)m_Poses.size() && !IsFlagSet(VIS_AveragePose))
			std::cout<<m_Poses[m_iCurrPose]->GetStr(m_pSolver).c_str()<<std::endl;
		else if (IsFlagSet(VIS_AveragePose) && m_pAveragePose!=NULL)
			std::cout<<m_pAveragePose->GetStr(m_pSolver).c_str()<<std::endl;
			
	}
	else if (key=='v')
	{
		if (m_pPosePrior!=NULL && m_pPosePrior->PriorType() == PosePrior::PRIOR_MultiGauss)
		{
			PosePriorMixGauss * prior = (PosePriorMixGauss*)m_pPosePrior;
			std::cout<<prior->GetStr().c_str()<<std::endl;
		}
		else
			std::cout<<"[WARNING] Unknown prior"<<std::endl;
	}
	else if (key=='e' || key=='E')
	{
		PntCorrsPose * pose = NULL;
		if (IsFlagSet(VIS_AveragePose) && m_pAveragePose!=NULL)
			pose = m_pAveragePose;
		else if (m_iCurrPose>=0 && m_iCurrPose < (int)m_Poses.size())
			pose = m_Poses[m_iCurrPose];
		else
			std::cout<<"[WARNING] Pose not defined..."<<std::endl;
		if (pose!=NULL)
		{
			VKString str = BindingEnergy::GetIKOnlyEnergyStr(pose, m_pSolver, m_pPosePrior, key=='E');
			std::cout<<"Energy "<<m_iCurrPose<<": "<<std::endl;
			std::cout<<str.c_str()<<std::endl;
		}
	}
	else if (key=='n')	// next average pose
	{
		renewPnts = true;
		m_iNextInitPose++;
		if (m_iNextInitPose>=m_pPosePrior->NSuggestedInitPoses())
			m_iNextInitPose = -1;
		
		if (m_pPosePrior->NSuggestedInitPoses()==0)
			Toggle(VIS_AveragePose);
		else if (IsFlagSet(VIS_AveragePose) && m_iNextInitPose==-1)
			Toggle(VIS_AveragePose);
		else if (!IsFlagSet(VIS_AveragePose) && m_iNextInitPose>=0)
			Toggle(VIS_AveragePose);
		
		if (m_pAveragePose==NULL)
		{
			m_pAveragePose = new PntCorrsPose();
			m_pSolver->InitPoseToZero(m_pAveragePose);
		}
				
		if (m_pPosePrior!=NULL && m_pPosePrior->NSuggestedInitPoses()>0 && m_iNextInitPose>=0)
		{
			std::cout<<"Init Pose (in IK): "<<m_iNextInitPose<<" / "<<m_pPosePrior->NSuggestedInitPoses()<<std::endl;
			m_pPosePrior->InitPose(m_iNextInitPose, m_pAveragePose);
		}
		else if (IsFlagSet(VIS_AveragePose))
		{
			std::cout<<"No init poses: average pose"<<std::endl;
			m_pSolver->InitPoseToZero(m_pAveragePose);

			// for every pose store angles
			for (int p=0; p<(int)m_Poses.size(); p++)
			{
				for (int l=0; l<(int)m_pSolver->m_Links.size(); l++)
				{
					SkeletonLink * lp = m_pSolver->m_Links[l];
					int ndims = SkeletonLink::NDims(lp->m_JointType);
					int offset = m_pSolver->m_ParamOffsetByLinkID[l];
					for (int d=0; d<ndims; d++)
						m_pAveragePose->Param(offset+d)+=m_Poses[p]->GetParam(offset+d)/(int)m_Poses.size();
				}
			}
		}
		if (m_iCurrPose>=0)
		{
			for (int i=0; i<4; i++)
				m_pAveragePose->Param(i) = 0;
//				m_pAveragePose->Param(i) = m_Poses[m_iCurrPose]->Param(i);
		}
		m_pSolver->ForwardKinematics(m_pAveragePose);
	}
	else if (key=='r' || key=='R')	// render points as functions based on neighborhood size
	{
		m_fDistrRadius += (key=='r'?-0.01:0.01);
		if (m_fDistrRadius<=0)
		{
			m_fDistrRadius = 0;
			SetRenderingFlag(VIS_LinkDistrRadius, false);
		}
		else
			SetRenderingFlag(VIS_LinkDistrRadius, true);
		renewPnts = true;
		std::cout<<"Distribution Radius: "<<m_fDistrRadius<<std::endl;
	}
	else
		return false;

	if (renewPnts && m_pDistrPoints!=NULL)
	{
		delete m_pDistrPoints;
		m_pDistrPoints = NULL;
	}
	return true;
}

bool VisPriorApp::Draw(DatabaseVisWindow * window)
{
	double br = 0, bg = 0, bb = 0;
	window->GetBackgroundColor(&br, &bg, &bb);
	glClearColor(br, bg, bb, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	
	window->LoadStandardSceneProjectionCamera();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glDisable(GL_LIGHTING);
	R3Affine camera = window->GetEmptySubwindowCamera(0);
//	camera.Translate(R3Vector(0, 0, 1));

	if (m_iCurrPose < (int)m_Poses.size() || (IsFlagSet(VIS_AveragePose)&&m_pAveragePose!=NULL))
	{
		PntCorrsPose * pose = m_pAveragePose;
		if (!IsFlagSet(VIS_AveragePose) || m_pAveragePose==NULL)
			pose = m_Poses[m_iCurrPose];

		R3Point centroid(0,0,.8);
		double scale = .7;
//		for (int i=0; i<(int)m_pSolver->m_Joints.size(); i++)
//			centroid += pose->Position(i);
//		centroid/=(int)m_pSolver->m_Joints.size();
//		centroid.SetZ(centroid.Z()+.5);
//		for (int i=0; i<(int)m_pSolver->m_Joints.size(); i++)
//			scale += R3Distance(pose->Position(i), centroid);
//		scale /=(int)m_pSolver->m_Joints.size();
//
		camera.Scale(scale);
		camera.Translate(-R3Vector(centroid.X(), centroid.Y(), centroid.Z()));

		// draw skeleton
		VKStringList pntsNames;
		std::vector<R3Point> namePositions;
		if (IsFlagSet(VIS_Skeleton))
		{
			bool updatePose = false;
			std::vector<R3Point> pnts;
			std::vector<int> edges;
			m_pSolver->FillRenderableSkeleton(pose, updatePose, pnts, edges);
				
			glColor3d(0., 0., .8);
			for (int i=0; i<(int)pnts.size(); i++)
			{
				R3Point p = pnts[i];
				camera.Apply(p);
				R3Sphere(p, 0.007).Draw();
			}
			
			double r, g, b;
			for (int i=0; i<pose->NMatches(); i++)
			{
				int rid = pose->Region(i);
				int jid = m_pSolver->m_Pnts2Skeleton[rid];
				R3Point p = pose->GetPosition(jid);
				camera.Apply(p);
				pntsNames.push_back(m_pSolver->m_Joints[jid]->m_sJointName);
				namePositions.push_back(p);
				window->MapIntToRadicalColor(rid, r, g, b);
				glColor3d(r, g, b);
				R3Sphere(p, 0.02).Draw();
			}
			
			glLineWidth(5.);
			glColor3d(.8, 0, 0);
			glBegin(GL_LINES);
			for (int e=0; e<(int)edges.size(); e+=2)
			{
				R3Point p1 = pnts[edges[e]];
				R3Point p2 = pnts[edges[e+1]];
				camera.Apply(p1);
				camera.Apply(p2);
				glVertex3d(p1.X(), p1.Y(), p1.Z());
				glVertex3d(p2.X(), p2.Y(), p2.Z());
			}
			glEnd();
		}
		
		// draw point distributions
		PreparePntDistr();
		if (m_pDistrPoints!=NULL)
		{
			if (m_pDistrPoints->NumPoints() < 150)
				glPointSize(6.);
			else if (m_pDistrPoints->NumPoints() < 500)
				glPointSize(3.);
			else
				glPointSize(1.);
			
			glDisable(GL_LIGHTING);
			glEnable(GL_COLOR);
			glDisable(GL_DEPTH_TEST);
			glBegin(GL_POINTS);
			for (int p=0; p<m_pDistrPoints->NumPoints(); p++)
			{
				R3Point pos = m_pDistrPoints->PointPosition(p);
				camera.Apply(pos);
				double r=0., g=0., b=0.;
				if((int)m_PerPointCluster.size()==m_pDistrPoints->NumPoints())
				{
					if (m_iSelectedCOI>=0 && m_iSelectedCOI != m_PerPointCluster[p])
						continue;
					window->MapIntToRadicalColor(m_PerPointCluster[p], r, g, b);
				}
				
				assert(m_PerPointCluster[p]>=0);
				glColor3d(r, g, b);
				glVertex3d(pos.X(), pos.Y(), pos.Z());
			}
			glEnd();
			glEnable(GL_DEPTH_TEST);
		}
		
		// draw grid
		if (IsFlagSet(VIS_EndEffGridDistr) && m_iSelectedRefID>=0)
		{
			assert (m_pPosePrior!=NULL && m_pPosePrior->IsAngles());
			PriorAngles * prior = (PriorAngles*)m_pPosePrior;
			PntCorrsPose * cond = new PntCorrsPose();
			cond->SetPoseID(pose->GetPoseID());
			
			// conditioned on contact of interest
			int jid = m_pSolver->m_Pnts2Skeleton[m_iSelectedRefID];
			R3Point refPos = pose->GetPosition(jid);
			cond->AddMatch(m_iSelectedRefID, TargetPrimitive(TargetPrimitive::PRIM_Point));

			if (cond->NMatches()>0)
			{
				//std::cout<<"N Candidate Contacts: "<<prior->NCandidateContacts()<<std::endl;
				std::vector<double> energies;
				for (int i=0; i<prior->NCandidateContacts(pose->GetPoseID()); i++)
				{
					int trgReg = prior->RegionID(pose->GetPoseID(), i);
					if (trgReg==m_iSelectedRefID || (m_iSelectedCOI>=0 && m_iSelectedCOI!=trgReg))
						continue;
					int gridID = prior->GetGridID(trgReg, cond);
					if (gridID==-1)
					{
						std::cout<<"[WARNING] No grid: Trg="<<trgReg<<" Ref: "<<m_iSelectedRefID;
						std::cout<<" PoseID: "<<pose->GetPoseID()<<std::endl;
						//std::cout<<"     Test2: "<<prior->UniqueCondID(trgReg, m_iSelectedRefID, pose->GetPoseID())<<std::endl;
						continue;
					}
					const R3Grid * grid = prior->GetGrid(gridID);
					double maxval = 0;
					double minval = RN_INFINITY;
					for (int g=0; g<grid->NEntries(); g++)
					{
						if (grid->GridValue(g) < RN_INFINITY)
						{
							minval = vkMin(grid->GridValue(g), minval);
							maxval = vkMax(grid->GridValue(g), maxval);
							energies.push_back(grid->GridValue(g));
						}
					}
					std::sort(energies.begin(), energies.end());
					int id = m_fEnergyThreshold * (int)energies.size();
					if (id>=0 && id < (int)energies.size())
						maxval = energies[id];

					double gridSize = 0;
					double origGridSize=0;
					const ShapePointCloud * pnts = prior->GetNonZeroProb(gridID, &origGridSize);
					gridSize = origGridSize;
					
					// find world(grid) --> world(object)
					R3Affine w2wg = R3identity_affine;
					w2wg.ZRotate(m_fRetRotation+pose->Param(3));
					R3Vector sizeRef(1., 1., 1.);
					sizeRef.Normalize();
					sizeRef*=gridSize*0.5;
					camera.Apply(sizeRef);
					gridSize = sizeRef.Length();
					
					// render grid (non-infinity entries)
					GLfloat surfaceMaterial[4];
					glEnable(GL_LIGHTING);
					//std::cout<<"NPnts: "<<pnts->NumPoints()<<std::endl;
					for (int p=0; p<pnts->NumPoints(); p++)
					{
						R3Point pos = pnts->PointPosition(p);
						
						double gridval = PriorAngles::ExactWorldValue(pos, grid);
						
						double prob = 1.-(gridval-minval) / (maxval-minval);
						if (prob<0)
							continue;

						window->MapIntToRadicalColor(trgReg, surfaceMaterial);
						surfaceMaterial[3] = prob;
						surfaceMaterial[0] = prob * surfaceMaterial[0] + (1.-prob);
						surfaceMaterial[1] = prob * surfaceMaterial[1] + (1.-prob);
						surfaceMaterial[2] = prob * surfaceMaterial[2] + (1.-prob);
						
						glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial);
						w2wg.Apply(pos);
						pos += refPos;
						camera.Apply(pos);	// why upside-down?
						//std::cout<<"Draw sphere: ["<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<"] gs="<<gridSize<<std::endl;
						R3Sphere(pos, gridSize).Draw();
					}
				}
			}
		}
		
		// draw names
		if (IsFlagSet(VIS_SkeletonNames))
		{
			VisWindowRenderer::TextSize size = VisWindowRenderer::TEXT_SIZE_SMALL;
			//VisWindowRenderer::TextSize size = VisWindowRenderer::TEXT_SIZE_LARGE;
			for (int i=0; i<pntsNames.count(); i++)
				window->GetRenderer()->RenderText(namePositions[i], pntsNames[i], 1., 0., 0., size);
		}
	}

	DrawGaussian(window, camera);
	DrawAngleHistogram(window);
	return true;
}

void VisPriorApp::DrawAngleHistogram(DatabaseVisWindow * window)
{
	if (IsFlagSet(VIS_LinkDistrAngles) && m_AngleDistr.size()>0)
	{
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0.0, 1.0, 0.0, 1.0, 0., 10.);
		double w = .8;
		double h = .02;
		double my = .99 - h * m_AngleDistr.size();
		double mx = .99 - w;
		
		for (int i=0; i<(int)m_AngleDistr.size(); i++)
			for (int j=0; j<(int)m_AngleDistr[i].size(); j++)
			{
				double v = m_AngleDistr[i][j];
				double nw = w/m_AngleDistr[i].size();
				window->DrawSquare(mx+j*nw, my+i*h, nw, h, 0., 0., 0., 0., 1.-v, 1.-v, 1.-v, 1.);
			}
		window->DrawSquare(mx, my, w, h*m_AngleDistr.size(), 0., 0., 0., 1., 1., 0., 0., 0);
		
		//		VKStringList angles;
		//		angles.push_back("-180");
		//		angles.push_back("-90");
		//		angles.push_back("0");
		//		angles.push_back("90");
		//		angles.push_back("180");
		//		double nh = .99 - 2*h * m_AngleDistr.size() - 3*h;
		//		double nx = .99 - 2*w;
		//		for (int i=0; i<(int)angles.count(); i++)
		//		{
		//			double nw = nx + 2*i*w/(double)angles.count();
		//			window->GetRenderer()->RenderText(R2Point(nw, nh), angles[i],
		//											  0., 0., .5, VisWindowRenderer::TEXT_SIZE_MEDIUM);
		//		}
	}
}

void VisPriorApp::DrawGaussian(DatabaseVisWindow * window, const R3Affine & camera)
{
	if (IsFlagSet(VIS_MeanGauss) && m_pPosePrior!=NULL && m_iSelectedLink>=0 && m_iCurrPose>=0)
	{
		PntCorrsPose * pose = m_pAveragePose;
		if (!IsFlagSet(VIS_AveragePose) || m_pAveragePose==NULL)
			pose = m_Poses[m_iCurrPose];
		
		SkeletonLink * l = m_pSolver->m_Links[m_iSelectedLink];
		int ndims = SkeletonLink::NDims(l->m_JointType);
		int offset = m_pSolver->m_ParamOffsetByLinkID[m_iSelectedLink];
		int jointid = l->m_pChild->m_iJointID;
		
		assert(m_pPosePrior->PriorType()==PosePrior::PRIOR_MultiGauss);
		PosePriorMixGauss * gprior =(PosePriorMixGauss *)m_pPosePrior;
		MultiGauss * mg = gprior->GetDistr(m_iSelectedLink);
		
		std::vector<double> origangles;
		for (int d=0; d<ndims; d++)
			origangles.push_back(pose->Param(offset+d));
		
		for (int gid=0; gid<mg->NGaussians(); gid++)
		{
			PosePointGauss * g = mg->GetGaussian(gid);
			
			assert(ndims==g->m_iNDims);
			for (int d=0; d<g->m_iNDims; d++)
				pose->Param(offset+d) = g->m_aMeans[d];
			
			double variance = 0;
			for (int d=0; d<g->m_iNDims; d++)
				variance += g->m_aVariances[d];
			variance = 0.3 * (sqrt(variance) / (2*M_PI));
			
			m_pSolver->ForwardKinematics(pose);
			R3Point p = pose->Position(jointid);
			camera.Apply(p);
			double cr, cg, cb;
			window->MapIntToRadicalColor(gid, cr, cg, cb);
			glColor3d(cr, cg, cb);
			R3Sphere(p, variance).Draw();
		}
		
		for (int d=0; d<ndims; d++)
			pose->Param(offset+d) = origangles[d];
	}
}

void VisPriorApp::PreparePntDistr()
{
	PreparePntDistrPerRegion();
}

void VisPriorApp::PreparePntDistrPerRegion()
{
	if (m_pDistrPoints!=NULL)
		return;
	m_pDistrPoints = new ShapePointCloud(NULL, "pts");
	m_PerPointCluster.clear();
	m_PerPointDistr.clear();
	m_AngleDistr.clear();
	
	double minEnergy = RN_INFINITY;
	double maxEnergy = 0;
	std::vector<double> energies;
	std::vector<double> energiesSrt;
	int refPoseID = -1;
	if (m_pAveragePose!=NULL && IsFlagSet(VIS_AveragePose))
		refPoseID = m_pAveragePose->GetPoseID();
	else  if (m_iCurrPose>=0)
		refPoseID = m_Poses[m_iCurrPose]->GetPoseID();
	
	
	
	for (int i=0; i<(int)m_Poses.size(); i++)
	{
		EnergyInfo einfo;
		double eval = RN_INFINITY;
		if (m_iSelectedCOI==-1 || m_iSelectedRefID==-1)
		{
			m_pSolver->GetEnergy(m_Poses[i], einfo);
			eval = einfo.m_fPosePriorEnergy + einfo.m_fSymmetryEnergy;
		}
		else
		{
			std::vector<int> trgRegion;
			std::vector<double> trgEnergy;
			trgRegion.push_back(m_iSelectedCOI);
			m_pSolver->GetPerContactEnergy(m_Poses[i], m_iSelectedRefID, trgRegion, trgEnergy);
			eval = trgEnergy[0];
		}
		minEnergy = vkMin(minEnergy, eval);
		maxEnergy = vkMax(maxEnergy, eval);
		energies.push_back(eval);
		energiesSrt.push_back(eval);
	}
	std::sort(energiesSrt.begin(), energiesSrt.end());
	int id = (int)energiesSrt.size() * m_fEnergyThreshold;
	double threshold = RN_INFINITY;
	if (id>=0 && id < (int)energiesSrt.size())
	{
		threshold = energiesSrt[id];
		maxEnergy = threshold;
	}
	
	//std::cout<<"RefID: "<<m_iSelectedRefID<<" avepos: "<<(m_pAveragePose!=NULL)<<std::endl;

	for (int i=0; i<(int)m_Poses.size(); i++)
	{
		if (energies[i]>threshold || (refPoseID>=0 && refPoseID!=m_Poses[i]->GetPoseID()))
			continue;
		
		// normalize pose
		double tx = m_Poses[i]->GetParam(0);
		double ty = m_Poses[i]->GetParam(1);
		double tz = m_Poses[i]->GetParam(2);
		double rot = m_Poses[i]->GetParam(3);
		for (int xf=0; xf<4; xf++)
		{
			if ((m_pAveragePose!=NULL && IsFlagSet(VIS_AveragePose)) || m_iCurrPose<0)
				m_Poses[i]->Param(xf) = 0;
			else
				m_Poses[i]->Param(xf) = m_Poses[m_iCurrPose]->Param(xf);
		}
		m_pSolver->ForwardKinematics(m_Poses[i]);
		
		R3Vector refOffset(0,0,0);
		int skid = -1;
		if (m_iSelectedRefID>=0 && m_pAveragePose!=NULL)
		{
			skid = m_pSolver->m_Pnts2Skeleton[m_iSelectedRefID];
			refOffset = m_pAveragePose->GetPosition(skid) - m_Poses[i]->GetPosition(skid);
//			std::cout<<"Offset["<<skid<<"]: "<<refOffset.X()<<" "<<refOffset.Y()<<" "<<refOffset.Z()<<" "<<std::endl;
		}
		
		
		for (int m=0; m<m_Poses[i]->NMatches(); m++)
		{
			int rid = m_Poses[i]->Region(m);
			R3Point pos = m_Poses[i]->GetPosition(m_pSolver->m_Pnts2Skeleton[rid]);
//			if (m_pSolver->m_Pnts2Skeleton[rid]==skid)
//				std::cout<<"Point["<<skid<<"]: "<<(pos+refOffset).X()<<" "<<(pos+refOffset).Y()<<" "<<(pos+refOffset).Z()<<"] "<<std::endl;
			m_pDistrPoints->AddPoint(ShapePoint(pos+refOffset));
			m_PerPointCluster.push_back(rid);
		}
		m_Poses[i]->Param(0) = tx;
		m_Poses[i]->Param(1) = ty;
		m_Poses[i]->Param(2) = tz;
		m_Poses[i]->Param(3) = rot;
		m_pSolver->ForwardKinematics(m_Poses[i]);
	}
	
	std::cout<<"Energy: "<<minEnergy<<" - "<<maxEnergy<<std::endl;
	
}

void VisPriorApp::PreparePntDistrPerGaussian()
{
	if (m_pDistrPoints!=NULL)
		return;
	m_PerPointDistr.clear();
	m_AngleDistr.clear();
	if (IsFlagSet(VIS_LinkDistrPoints) || IsFlagSet(VIS_LinkDistrRadius))
	{
		bool averagePose = (IsFlagSet(VIS_AveragePose) && m_pAveragePose!=NULL);
		bool selectedPose = m_iCurrPose < (int)m_Poses.size();
		if (m_iSelectedLink >= 0 && m_pSolver!=NULL && (averagePose || selectedPose))
		{
			PntCorrsPose * pose = m_pAveragePose;
			if (!averagePose)
				pose = m_Poses[m_iCurrPose];
			
			m_pDistrPoints = new ShapePointCloud(NULL, "pnts");
			SkeletonLink * l = m_pSolver->m_Links[m_iSelectedLink];
			int ndims = SkeletonLink::NDims(l->m_JointType);
			int offset = m_pSolver->m_ParamOffsetByLinkID[m_iSelectedLink];
			int jointid = l->m_pChild->m_iJointID;
			
			int nbins = 100;
			for (int d=0; d<ndims; d++)
			{
				m_AngleDistr.push_back(std::vector<double>());
				for (int b=0; b<nbins; b++)
					m_AngleDistr[d].push_back(0);
			}
			
			/// DENSITY of angles (decomposed, visualized at the top of the screen)
			// for every pose store angles
			for (int p=0; p<(int)m_Poses.size(); p++)
			for (int d=0; d<ndims; d++)
			{
				int bin = nbins*((m_Poses[p]->Param(offset+d)+M_PI) / (2*M_PI));
				assert(bin<nbins && bin>=0);
				m_AngleDistr[d][bin]+=1.;
			}
			

			std::vector<double> mv;
			for (int i=0; i<(int)m_AngleDistr.size(); i++)
			{
				mv.push_back(0);
				for (int j=0; j<(int)m_AngleDistr[i].size(); j++)
				{
					m_AngleDistr[i][j] = m_AngleDistr[i][j];
					mv[i] = vkMax(mv[i], m_AngleDistr[i][j]);
				}
			}
			
			for (int i=0; i<(int)m_AngleDistr.size(); i++)
			for (int j=0; j<(int)m_AngleDistr[i].size(); j++)
				m_AngleDistr[i][j] = m_AngleDistr[i][j]/mv[i];
			
			// POINTS (draw actual positions of where points might end up)
			std::vector<double> angles;
			for (int p=0; p<(int)m_Poses.size(); p++)
			{
				for (int d=0; d<ndims; d++)
					angles.push_back(m_Poses[p]->Param(offset+d));
			}
			// copy angles
			std::vector<double> myposeangles;
			for (int d=0; d<ndims; d++)
				myposeangles.push_back(pose->Param(offset+d));

			// for all angle parameters visualize where points would end up
			for (int a=0; a<(int)angles.size(); a+=ndims)
			{
				for (int d=0; d<ndims; d++)
					pose->Param(offset+d) = angles[a+d];
				m_pSolver->ForwardKinematics(pose);
				m_pDistrPoints->AddPoint(pose->Position(jointid));
			}
			
			// revert back the angle parameters
			for (int d=0; d<ndims; d++)
				pose->Param(offset+d) = myposeangles[d];
			
//			// density radius
//			if (IsFlagSet(VIS_LinkDistrRadius))
//			{
//				double maxval = 0;
//				for (int p=0; p<m_pDistrPoints->NumPoints(); p++)
//				{
//					RNArray<const PointInSet*> pts;
//					m_pDistrPoints->GetKdTree()->FindAll(m_pDistrPoints->PointPosition(p), 0, m_fDistrRadius, pts);
//					m_PerPointDistr.push_back((double)pts.NEntries());
//					maxval = vkMax(maxval, m_PerPointDistr[p]);
//					
//				}
//				for (int p=0; p<m_pDistrPoints->NumPoints(); p++)
//					m_PerPointDistr[p] /= maxval;
//				std::cout<<"PerPoint Max: "<<maxval<<" / "<<m_pDistrPoints->NumPoints()<<std::endl;
//			}
			
			// clustering
			assert(m_pDistrPoints!=NULL);
			m_PerPointCluster.clear();
			assert(m_iSelectedLink < (int)m_AngleClusters.size());
			if (IsFlagSet(VIS_Clustering) && m_iSelectedLink>=0 && m_AngleClusters[m_iSelectedLink])
			{
				for (int i=0; i<m_pDistrPoints->NumPoints(); i++)
					m_PerPointCluster.push_back(-1);
				
				for (int c=0; c<m_AngleClusters[m_iSelectedLink]->NClusters(); c++)
				{
					const std::vector<ClusteredNode> & clustering = m_AngleClusters[m_iSelectedLink]->GetCluster(c);
					for (int n=0; n<(int)clustering.size(); n++)
						m_PerPointCluster[clustering[n].id] = c;
				}
			}
		}
	}
}


