#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"

#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ShapeTriangleSoup.h"

#include "MetricVisApp.h"
#include "ShapeDistance.h"
#include "VKGraph.h"
#include "DistGeodPnts.h"

MetricVisApp::MetricVisApp(Model3D * model, ShapeMesh * mesh, ShapePointCloud * pnts, ShapeDistance * dist)
{
	m_pModel = model;
	m_pMesh = mesh;
	m_pPnts = pnts;
	m_pDistance = dist;
	m_fNhdRadius = 0.5;
	m_iSelectedPnt = -1;
	m_iVisFlags = VISMETRIC_Mesh | VISMETRIC_Pnts | VISMETRIC_NhdSphere | VISMETRIC_NhdWireframe | VISMETRIC_NhdHighlight;
	//m_iVisFlags = VISMETRIC_Pnts | VISMETRIC_NhdGraph;
}

bool MetricVisApp::KeyPressed(DatabaseVisWindow * window, const char key, int mouseX, int mouseY)
{
	switch(key)
	{
		case '`':
			std::cout<<"Keys:"<<std::endl;
			std::cout<<"    `       - this help menu"<<std::endl;
			std::cout<<"    SPACE   - select point"<<std::endl;
			std::cout<<"    m       - show mesh"<<std::endl;
			std::cout<<"    p       - show points"<<std::endl;
			std::cout<<"    h       - highlight neighborhood"<<std::endl;
			std::cout<<"    s       - render sphere of neighborhood's radius"<<std::endl;
			std::cout<<"    w       - wireframe"<<std::endl;
			std::cout<<"    g       - nearest graph"<<std::endl;
			std::cout<<"    ?       - stats on neighborhood"<<std::endl;
			break;
		case ' ':
		{
			assert(window->GetDrawableModels().size()>0);
			VisModel3D * mvis = window->GetDrawableModels()[0];
			int cameraID = mvis->SubwindowCameraID(0, window);
			mvis->SelectPoint(cameraID, mouseX, mouseY, window);
			const ShapePoint & pnt = mvis->SelectedPoint();
			if (!pnt.Invalid())
			{
				m_iSelectedPnt = pnt.PointID();
				if (m_iSelectedPnt<0)
					m_iSelectedPnt = m_pPnts->GetKdTree()->FindClosest(pnt.GetPosition())->id;
			}
			else
				m_iSelectedPnt = -1;
			break;
		}
		case 'm':
			Toggle(VISMETRIC_Mesh);
			break;
		case 'p':
			Toggle(VISMETRIC_Pnts);
			break;
		case 'h':
			Toggle(VISMETRIC_NhdHighlight);
			break;
		case 's':
			Toggle(VISMETRIC_NhdSphere);
			break;
		case 'w':
			Toggle(VISMETRIC_NhdWireframe);
			break;
		case 'g':
			Toggle(VISMETRIC_NhdGraph);
			break;
		case '?':
			if (m_iSelectedPnt>=0)
			{
				assert(m_pPnts!=NULL && m_pDistance!=NULL);
				std::vector<ShapePoint> neighborhood;
				m_pDistance->NearestNeighborhood(m_pPnts->GetPoint(m_iSelectedPnt), m_fNhdRadius, &neighborhood);
				int nNeighbors=(int)neighborhood.size();
				
				std::cout<<"Neighbors of "<<m_iSelectedPnt<<": ";
				std::cout<<((int)(100*(double)nNeighbors/(double)m_pPnts->NumPoints()))<<"%     ";
				std::cout<<nNeighbors<<" / "<<m_pPnts->NumPoints()<<"     ";
				std::cout<<std::endl;
			}
			break;
		case '=':
		case '-':
			m_fNhdRadius += (key=='=' ? 1:-1)*0.05;
			std::cout<<"Nhd: "<<m_fNhdRadius<<std::endl;
			break;
		case 27:
			exit(0);
	}
	return true;
}


bool MetricVisApp::Draw(DatabaseVisWindow * window)
{
	window->SetBackgroundColor(1);
	double br, bg, bb;
	window->GetBackgroundColor(&br, &bg, &bb);
	glClearColor(br, bg, bb, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	window->SetBackgroundColor(-1);
	
	window->LoadStandardSceneProjectionCamera();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	VisModel3D * mvis = NULL;
//	if (window->GetDrawableModels().size()>0)
	mvis = window->GetDrawableModels()[0];
//	else
//		mvis = window->GetEmptyWindowXForm(0);
	mvis->ClearRenderablgeGeometries();
	if (IsFlagSet(VISMETRIC_Mesh))
		mvis->AddRenderableGeometry("mesh");
	else if (IsFlagSet(VISMETRIC_Pnts))
		mvis->AddRenderableGeometry("pnts");
	
	ShapePoint selectedPos = ShapePoint();
	if (m_iSelectedPnt>=0)
		selectedPos	= m_pPnts->GetPoint(m_iSelectedPnt);
	//ShapePoint selectedPos = ShapePoint(R3Point(0.591219, 0.662949, 0.424532), NULL);
	

	static bool init = false;
	int cid = mvis->SubwindowCameraID(0, window);
	mvis->GetCameraTransform(cid);
	if (!init)
	{
		mvis->AddRenderableGeometry("pnts");
		mvis->SetPointRadius(0.01);
		mvis->Scale(cid, -0.5, window);
	}
	init = true;

	R3Affine camera = mvis->GetCameraTransform(mvis->SubwindowCameraID(0, window));
	
	window->LoadStandardSceneProjectionCamera();
	
	GLfloat surfaceMaterial[4];
	std::vector<RenderableFace> faces;
	
	window->LoadStandardSceneProjectionCamera();
	
	// draw sphere
	glEnable(GL_LIGHTING);
	glShadeModel(GL_SMOOTH);
	if (IsFlagSet(VISMETRIC_NhdWireframe))
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	R3Vector xyz(1,1,1);  xyz.Normalize();
	if (IsFlagSet(VISMETRIC_NhdSphere) && !selectedPos.Invalid() && m_pPnts!=NULL)
	{
		surfaceMaterial[0] = .5;
		surfaceMaterial[1] = .5;
		surfaceMaterial[2] = 1.;
		surfaceMaterial[3] = 1.;
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial);
		R3Point pos = selectedPos.GetPosition();
		R3Point p2 = pos + xyz * m_fNhdRadius;
		camera.Apply(pos); camera.Apply(p2);
		R3Sphere(pos, (pos-p2).Length()).Draw();
	}
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	
	// draw higlighted points
	if (IsFlagSet(VISMETRIC_NhdHighlight) && !selectedPos.Invalid() && m_pPnts!=NULL && m_pDistance!=NULL)
	{
		std::vector<ShapePoint> neighborhood;
		m_pDistance->NearestNeighborhood(selectedPos, m_fNhdRadius, &neighborhood);
		surfaceMaterial[0] = .8;
		surfaceMaterial[1] = 0.;
		surfaceMaterial[2] = 0.;
		surfaceMaterial[3] = 1.;
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial);
		
		for (int i=0; i<(int)neighborhood.size(); i++)
		{
			R3Point pos = neighborhood[i].GetPosition();
			R3Point p2 = pos + xyz * 0.015;
			camera.Apply(pos); camera.Apply(p2);
			R3Sphere(pos, (pos-p2).Length()).Draw();
		}
	}
	
	// draw graph
	if (IsFlagSet(VISMETRIC_NhdGraph))
	{
		DistGeodPnts * distGeod = m_pDistance->SafeCastToGraphNN();
		if (distGeod!=NULL && m_iSelectedPnt>=0)
		{
			VKGraph * graph = (VKGraph *)distGeod->GetGraph();
			assert(graph!=NULL);
			VKGraphNode * node = graph->Node(m_iSelectedPnt);
			glLineWidth(2.);
			glDisable(GL_LIGHTING);
			glColor3d(0., 1., 0.);
			glBegin(GL_LINES);
			R3Point pos1 = selectedPos.GetPosition();
			camera.Apply(pos1);
			for (int e=0; e<node->NEdges(); e++)
			{
				VKGraphNode * nn = node->Neighbor(e);
				R3Point pos2 = m_pPnts->PointPosition(nn->id);
				camera.Apply(pos2);
				glVertex3d(pos1.X(), pos1.Y(), pos1.Z());
				glVertex3d(pos2.X(), pos2.Y(), pos2.Z());
			}
			glEnd();
		}
	}
	
	// draw transparent faces (if any)
	if (faces.size()>0)
		window->GetRenderer()->DrawTransparentFaces(faces, true);
	
	return false;

}


