#include "Clustering.h"
#include "Sortable.h"

/////////////// Clustering problem //////////
ClusteringProblem::ClusteringProblem(int nNodes, int nFeatures)
{
	if (nFeatures>0)
	{
		assert(nNodes > 0);
		m_pFeatures = new LinAlgMatrixReal(nNodes, nFeatures, 0);
	}
	else
		m_pFeatures = NULL;
	m_bVerbose = true;
	m_MetricType = METRIC_FeatDistL2;
}

ClusteringProblem::ClusteringProblem(LinAlgMatrixReal * features)
{
	m_pFeatures = features;
	m_bVerbose = true;
	m_MetricType = METRIC_FeatDistL2;
}

ClusteringProblem::~ClusteringProblem()
{
	if (m_pFeatures!=NULL)
		delete m_pFeatures;
}

void ClusteringProblem::SetMetricType(DistMetricType type)
{
	m_MetricType = type;
	assert(m_MetricType!=METRIC_PairwiseDists || NNodes()==NFeatures());
}

ClusteringProblem::DistMetricType ClusteringProblem::MetricType() const
{
	return m_MetricType;
}


void ClusteringProblem::SetFeatures(int nodeID, const double * features)
{
	assert(m_pFeatures!=NULL);
	for (int c=0; c<m_pFeatures->Cols(); c++)
		(*m_pFeatures)(nodeID,c) = features[c]; 
}

void ClusteringProblem::SetFeatures(int nodeID, const float * features)
{
	assert(m_pFeatures!=NULL);
	for (int c=0; c<m_pFeatures->Cols(); c++)
		(*m_pFeatures)(nodeID,c) = features[c]; 
}

int ClusteringProblem::NNodes()
{
	assert(m_pFeatures!=NULL);
	return m_pFeatures->Rows();
}

int ClusteringProblem::NFeatures()
{
	assert(m_pFeatures!=NULL);
	return m_pFeatures->Cols();
}

int ClusteringProblem::NClusters()
{
	return (int)m_Clusters.size();
}

int ClusteringProblem::ClusterSize(int clusterID)
{
	return (int)GetCluster(clusterID).size();
}

const std::vector<ClusteredNode> & ClusteringProblem::GetCluster(int clusterID)
{
	assert(clusterID < NClusters());
	return m_Clusters[clusterID];
}

void ClusteringProblem::SetNodeErr(int clusterID, int nodeID, double err)
{
	assert(clusterID < (int)m_Clusters.size() && nodeID < m_Clusters[clusterID].size());
	double olderr = m_Clusters[clusterID][nodeID].error;
	m_Clusters[clusterID][nodeID].error = err;
	m_ClusterTightness[clusterID] += (err/m_Clusters[clusterID].size()-olderr);
}

std::vector<int> ClusteringProblem::SortClusters()
{
	std::vector<int> newClusterOrder;
	std::vector<std::vector<ClusteredNode> > clusters;
	std::vector<double> clusterTightness;

	std::vector<Sortable> sortableTightness;
	for (int i=0; i<(int)m_ClusterTightness.size(); i++)
		sortableTightness.push_back(Sortable(m_ClusterTightness[i], NULL, i));
	std::sort(sortableTightness.begin(), sortableTightness.end());

	for (int i=0; i<(int)sortableTightness.size(); i++)
	{
		int cID = sortableTightness[i].id;
		newClusterOrder.push_back(cID);
		clusterTightness.push_back(sortableTightness[i].value);
		std::vector<Sortable> sortableCluster;
		
		for (int j=0; j<(int)m_Clusters[cID].size(); j++)
			sortableCluster.push_back(Sortable(m_Clusters[cID][j].error, NULL, j));
		std::sort(sortableCluster.begin(), sortableCluster.end());

		clusters.push_back(std::vector<ClusteredNode>());
		for (int j=0; j<(int)sortableCluster.size(); j++)
			clusters[i].push_back(m_Clusters[cID][sortableCluster[j].id]);
	}
	m_ClusterTightness = clusterTightness;
	m_Clusters = clusters;
	return newClusterOrder;
}

void ClusteringProblem::ClearSolution()
{
	m_Clusters.clear();
	m_ClusterTightness.clear();
}

void ClusteringProblem::WriteSolution(const char * filename)
{
	FILE * file = fopen(filename, "wb");
	assert(file!=NULL);
	int nClusters = (int)m_Clusters.size();
	assert(m_ClusterTightness.size()==m_Clusters.size());
	fwrite(&nClusters, sizeof(int), 1, file);
	for (int i=0; i<nClusters; i++)
	{
		int clusterSize = (int)m_Clusters[i].size();
		fwrite(&clusterSize, sizeof(int), 1, file);
		float tightness = m_ClusterTightness[i];
		fwrite(&tightness, sizeof(float), 1, file);
		for (int j=0; j<clusterSize; j++)
			fwrite(&(m_Clusters[i][j].id), sizeof(int), 1, file);
		for (int j=0; j<clusterSize; j++)
			fwrite(&(m_Clusters[i][j].error), sizeof(float), 1, file);
	}
	fclose(file);
}

void ClusteringProblem::ReadSolution(const char * filename)
{
	FILE * file = fopen(filename, "rb");
	assert(file!=NULL);
	int nClusters;
	fread(&nClusters, sizeof(int), 1, file);
	for (int i=0; i<nClusters; i++)
	{
		int clusterSize;
		m_Clusters.push_back(std::vector<ClusteredNode>());
		m_ClusterTightness.push_back(0.);
		fread(&clusterSize, sizeof(int), 1, file);
		float tightness = 0;
		fread(&tightness, sizeof(float), 1, file);
		m_ClusterTightness[i] = tightness;
		for (int j=0; j<clusterSize; j++)
			m_Clusters[i].push_back(ClusteredNode());
		for (int j=0; j<clusterSize; j++)
			fread(&(m_Clusters[i][j].id), sizeof(int), 1, file);
		for (int j=0; j<clusterSize; j++)
			fread(&(m_Clusters[i][j].error), sizeof(float), 1, file);
	}
	fclose(file);
}


/////////////// ClusteringAlgorithm //////////
LinAlgMatrixReal * ClusteringAlgorithm::Features(ClusteringProblem & problem)
{
	assert(problem.m_pFeatures!=NULL);	
	return problem.m_pFeatures;
}

std::vector<std::vector<ClusteredNode> > & ClusteringAlgorithm::Clusters(ClusteringProblem & problem)
{
	return problem.m_Clusters;
}

std::vector<double> & ClusteringAlgorithm::ClusterTightness(ClusteringProblem & problem)
{
	return problem.m_ClusterTightness;
}

/////////////// ClusterNearest //////////
ClusterNearest::ClusterNearest()
{
	m_fMaxDistance = -1;
	m_iMaxClusters = -1;
}

ClusterNearest::ClusterNearest(double threshold)
{
	m_fMaxDistance = threshold;
	m_iMaxClusters = -1;
}

ClusterNearest::ClusterNearest(int nClusters)
{
	m_fMaxDistance = -1;
	m_iMaxClusters = nClusters;
}

ClusterNearest::~ClusterNearest()
{
	
}

void ClusterNearest::Cluster(ClusteringProblem & problem)
{
	if (problem.NNodes()>pow(10,6))
	{
		std::cout<<"[ERROR] Too many pairs for clustering. NPairs: "<<(pow(problem.NNodes(),2)/2)<<std::endl;
		assert(false);
	}
	// find closest pairs
	std::vector<int> cluster2MtxID;
	LinAlgMatrixReal cluster2cluster(problem.NNodes(), problem.NNodes());
	LinAlgMatrixReal point2point(problem.NNodes(), problem.NNodes());
	LinAlgMatrixReal & features = *(Features(problem));
	
	for (int i=0; i<problem.NNodes(); i++)
	for (int j=i+1; j<problem.NNodes(); j++)
	{
		if (problem.MetricType()==ClusteringProblem::METRIC_FeatDistL2)
		{
			point2point(i,j) = 0;
			for (int f=0; f<problem.NFeatures(); f++)
				point2point(i,j) += pow(features(i, f)-features(j, f), 2.);
		}
		else if (problem.MetricType()==ClusteringProblem::METRIC_PairwiseDists)
			point2point(i,j) = features(i, j);
		else
			assert(false);
		
		point2point(j,i) = point2point(i,j);
		cluster2cluster(i,j) = point2point(i,j);
		cluster2cluster(j,i) = point2point(i,j);
	}
	
	for (int i=0; i<problem.NNodes(); i++)
		cluster2MtxID.push_back(i);
		
	// iteratively merge closest clusters
	std::vector<int> node2cluster;
	std::vector<std::vector<int> > cluster2node;
	
	for (int i=0; i<problem.NNodes(); i++)
	{
		node2cluster.push_back(i);
		cluster2node.push_back(std::vector<int>());
		cluster2node[i].push_back(i);
	}
	
	bool terminate = (int)cluster2node.size() <= vkMax(m_iMaxClusters,2);
	while (!terminate)
	{
		// FIND CLOSEST
		double minDist = -1;
		int c1=-1, c2=-1;
		for (int i=0; i<cluster2node.size(); i++)
		for (int j=i+1; j<cluster2node.size(); j++)
		{
			int mid1 = cluster2MtxID[i];
			int mid2 = cluster2MtxID[j];
			double cDist = cluster2cluster(mid1, mid2);
			
			if (minDist < 0 || cDist < minDist)
			{
				minDist = cDist;
				c1 = i;
				c2 = j;
			}
		}
		//std::cout<<"Dist: "<<minDist<<" / "<<m_fMaxDistance<<std::endl;
		if (minDist > m_fMaxDistance && m_fMaxDistance>=0)
			terminate = true;
		else
			terminate = ((int)cluster2node.size() <= vkMax(m_iMaxClusters,2));
		
		// MERGE
		assert(c1>=0 && c2>=0 && c1 < c2);
		while (cluster2node[c2].size()>0)
		{
			cluster2node[c1].push_back(cluster2node[c2][0]);
			node2cluster[cluster2node[c2][0]] = c1;
			cluster2node[c2].erase(cluster2node[c2].begin());
		}
		int mid1 = cluster2MtxID[c1];
		int mid2 = cluster2MtxID[c2];
		cluster2node.erase(cluster2node.begin()+c2);
		cluster2MtxID.erase(cluster2MtxID.begin()+c2);
		
		for (int i=0; i<cluster2node.size(); i++)
		{
			int mid3 = cluster2MtxID[i];
			if (mid1==mid3) continue;
			assert(mid1!=mid2 && mid1!=mid3 && mid2!=mid3);
			cluster2cluster(mid1, mid3) = vkMin(cluster2cluster(mid1, mid3), cluster2cluster(mid2, mid3));
			cluster2cluster(mid3, mid1) = cluster2cluster(mid1, mid3);
		}
	}
	
	// write result	
	int nClusters = (int)cluster2node.size();
	int nFeatures = problem.NFeatures();
	LinAlgMatrixReal clusterCenters(nClusters, nFeatures);
	
	for (int i=0; i<(int)cluster2node.size(); i++)
	for (int p=0; p<(int)cluster2node[i].size(); p++)
	for (int f=0; f<nFeatures; f++)
		clusterCenters(i,f) += features(cluster2node[i][p], f) / cluster2node[i].size();

	std::vector<std::vector<Sortable> > unsortedClusters;
	std::vector<double> unsortedErrors;
	std::vector<Sortable> clusterOrder;
	for (int i=0; i<(int)cluster2node.size(); i++)
	{
		unsortedClusters.push_back(std::vector<Sortable>());
		double aveDist = 0;
		for (int j=0; j<(int)cluster2node[i].size(); j++)
		{
			unsortedClusters[i].push_back(Sortable(0, NULL, cluster2node[i][j]));
			unsortedClusters[i][j].value = 0;
			for (int f=0; f<nFeatures; f++)
				unsortedClusters[i][j].value += pow(features(cluster2node[i][j], f)-clusterCenters(i,f), 2.);
			aveDist += unsortedClusters[i][j].value;
		}
		std::sort(unsortedClusters[i].begin(), unsortedClusters[i].end());
		unsortedErrors.push_back(aveDist);
		clusterOrder.push_back(Sortable(-(double)cluster2node[i].size(), NULL, i));
	}
	std::sort(clusterOrder.begin(), clusterOrder.end());
	
	std::vector<std::vector<ClusteredNode> > & clusters = Clusters(problem);
	std::vector<double> & tightness = ClusterTightness(problem);

	for (int i=0; i<(int)clusterOrder.size(); i++)
	{
		clusters.push_back(std::vector<ClusteredNode>());
		int cid = clusterOrder[i].id;
		tightness.push_back(unsortedErrors[cid]);
		for (int j=0; j<(int)unsortedClusters[cid].size(); j++)
		{
			clusters[i].push_back(ClusteredNode());
			clusters[i][j].id = unsortedClusters[cid][j].id;
			clusters[i][j].error = unsortedClusters[cid][j].value;
		}
	}
}


/////////////// K-means clustering //////////
KMeansClustering::KMeansClustering(int nClusters, int maxIterations, 
								   const LinAlgMatrixReal * seedCentroids)
{
	m_iMaxIterations = maxIterations;
	m_iNClusters = nClusters;
	assert(seedCentroids==NULL);
	m_pCentroids = NULL;
	m_CentroidNorms = new int[m_iNClusters];
}

KMeansClustering::~KMeansClustering()
{
	if (m_pCentroids!=NULL)
		delete m_pCentroids;
	delete [] m_CentroidNorms;
}

void KMeansClustering::Cluster(ClusteringProblem & problem)
{
	assert(problem.MetricType()==ClusteringProblem::METRIC_FeatDistL2);
	
	int nNodes = problem.NNodes();
	int nFeatures = problem.NFeatures();
	
	if (problem.m_bVerbose)
		std::cout<<"K-means: N="<<nNodes<<" k="<<m_iNClusters<<" f="<<nFeatures<<std::endl;
	
	// my slow implementation of the standard Lloyd's algorithm
	std::vector<int> clusterIDs;
	for (int i=0; i<nNodes; i++)
		clusterIDs.push_back(-1);

	// initialize memory for centroids
	if (m_pCentroids==NULL || m_pCentroids->Cols() != nFeatures)
	{
		if (m_pCentroids!=NULL)
			delete m_pCentroids;
		m_pCentroids = new LinAlgMatrixReal(m_iNClusters, nFeatures);
	}
	
	LinAlgMatrixReal & features = *(Features(problem));
	
//	// features unnormalized?
//	for (int i=0; i<nNodes; i++)
//	{
//		std::cout<<"Node["<<i<<"] = [";
//		double n = 0;
//		for (int f=0; f<nFeatures; f++)
//		{
//			n += pow(features(i,f), 2.);
//			std::cout<<features(i,f)<<" "<<std::flush;
//		}
//		std::cout<<"] -> "<<n<<std::endl;
//	}

	if (problem.m_bVerbose)
		std::cout<<"   Init centroids"<<std::endl;
	
	// assign random centroids
	std::map<int, bool> takenElements;
	assert(m_iNClusters < nNodes);
	for (int c=0; c<m_iNClusters; c++)
	{
		int rndEl = rand()%nNodes;
		while (takenElements.find(rndEl)!=takenElements.end() && rndEl<0)
			rndEl = rand()%nNodes;
		takenElements[rndEl] = true;
		for (int f=0; f<nFeatures; f++)
			(*m_pCentroids)(c, f) = features(rndEl, f);
	}
	
	// k-means clustering iterations
	int i=0;
	int nUpdate = 100;
	if (problem.m_bVerbose)
		std::cout<<"   Iterating ("<<m_iMaxIterations<<")"<<std::endl;
	for (i=0; i<m_iMaxIterations; i++)
	{
		if (problem.m_bVerbose && i%nUpdate==0 && i!=0)
			std::cout<<"\t["<<i<<" / "<<m_iMaxIterations<<"] "<<std::endl;

		// cluster based on centroids
		bool clusteringChanged = false;
		for (int pID=0; pID<nNodes; pID++)
		{
			double bestDist2 = NUM_DBL_MAX;
			int bestCluster = clusterIDs[pID];
			if (bestCluster>=0)
			{
				bestDist2 = 0;
				for (int f=0; f<nFeatures; f++)
					bestDist2 += pow((*m_pCentroids)(bestCluster, f) - features(pID, f), 2);
			}
			// TODO: this can be done with an ANN library (find nearest centroid)
			for (int c=0; c<m_iNClusters; c++)	
			{
				if (c==bestCluster)
					continue;
				double dist2=0;
				for (int f=0; f<nFeatures /*&& dist2 < bestDist2*/; f++)
					dist2 += pow((*m_pCentroids)(c, f) - features(pID,f), 2);
				
				if (dist2 < bestDist2)
				{
					bestDist2 = dist2;
					bestCluster = c;
				}
			}
			assert(bestCluster>=0);
			if (clusterIDs[pID] != bestCluster)
				clusteringChanged = true;
			clusterIDs[pID] = bestCluster;
		}
		
		if (i==m_iMaxIterations || !clusteringChanged)	// no need to look for centroids
			break;
		
		// recompute centroids
		m_pCentroids->SetValues(0);
		for (int c=0;c<m_iNClusters; c++)
			m_CentroidNorms[c] = 0;
		for (int pID=0; pID<nNodes; pID++)
		{
			m_CentroidNorms[clusterIDs[pID]]++;
			for (int f=0; f<nFeatures; f++)
				(*m_pCentroids)(clusterIDs[pID], f) += features(pID, f);
		}
		for (int c=0; c<m_iNClusters; c++)
		for (int f=0; f<nFeatures; f++)
			(*m_pCentroids)(c, f) /= m_CentroidNorms[c];
	}
	if (problem.m_bVerbose && i>nUpdate)
		std::cout<<std::endl;
	
	// assign clusters
	problem.ClearSolution();
	std::vector<std::vector<ClusteredNode> > & clusters = Clusters(problem);
	std::vector<double> & clusterTightness = ClusterTightness(problem);
	assert(clusters.size()==0 && clusterTightness.size()==0);
	std::vector<std::vector<Sortable> > sortableClusters;
	std::vector<Sortable> sortableTightness;
	for (int i=0; i<m_iNClusters; i++)
	{
		clusters.push_back(std::vector<ClusteredNode>());
		sortableClusters.push_back(std::vector<Sortable>());
		sortableTightness.push_back(Sortable(0, NULL, i));
	}
	
	// compute distances to centroid
	for (int i=0; i<nNodes; i++)
	{
		int cID = clusterIDs[i];
		int nID = (int)sortableClusters[cID].size();
		sortableClusters[cID].push_back(Sortable(0, NULL, i));
		for (int f=0; f<nFeatures; f++)
			sortableClusters[cID][nID].value += pow((*m_pCentroids)(cID, f)-features(i, f), 2.);

		sortableTightness[cID].value += sortableClusters[cID][nID].value;
	}
	
	// sort clusters (by average tightness)
	for (int i=0; i<m_iNClusters; i++)
		sortableTightness[i].value /= (int)sortableClusters[i].size();
	std::sort(sortableTightness.begin(), sortableTightness.end());
	
	// fill clusters (models are sorted by distance to centroid)
	for (int i=0; i<m_iNClusters; i++)
	{		
		int cID = sortableTightness[i].id;
		clusterTightness.push_back(sortableTightness[i].value);
	
		std::sort(sortableClusters[cID].begin(), sortableClusters[cID].end());
		for (int j=0; j<(int)sortableClusters[cID].size(); j++)
		{
			int nID = (int)clusters[i].size();
			clusters[i].push_back(ClusteredNode());
			clusters[i][nID].id = sortableClusters[cID][j].id;
			clusters[i][nID].error = sortableClusters[cID][j].value;
		}
	}
}



/////////////// Mean Shift Clustering //////////
MeanShiftClustering::MeanShiftClustering(double h, int maxIterations, int minclustersize)
{
//	std::cout<<"MSC: "<<h<<" "<<maxIterations<<" "<<minclustersize<<std::endl;
	m_fSupportSigma = h;
	m_iMaxIterations = maxIterations;
	m_iMinClusterSize = minclustersize;
}

MeanShiftClustering::~MeanShiftClustering()
{
	
}

void MeanShiftClustering::Cluster(ClusteringProblem & problem)
{
	assert(problem.MetricType()==ClusteringProblem::METRIC_FeatDistL2);
	
	bool DEBUG_VERBOSE = true;
	if (m_fSupportSigma < 0)	// find h
		assert(false);	// for now...
	double h = m_fSupportSigma;
	double maxDist = 10 * h;
	double terminateDist = h*1E-10;
	double convergedDist = 2*h;
	
	LinAlgMatrixReal & feats = *(Features(problem));
		
	LinAlgMatrixReal x_curr(feats.Rows(), feats.Cols());
	LinAlgMatrixReal x_next(feats.Rows(), feats.Cols());
	
	for (int r=0; r<feats.Rows(); r++)
	for (int c=0; c<feats.Cols(); c++)
		x_curr(r, c) =  feats(r, c);
		
	assert(feats.Rows()==problem.NNodes() && feats.Cols()==problem.NFeatures());
	double maxWindowDist = 0;
	// iterate
	int i=0;
	for (i=0; i<m_iMaxIterations; i++)
	{
		// find updates (also check if termination condition reached)
		maxWindowDist = 0;
		x_next.SetValues(0);
		for (int n1=0; n1<problem.NNodes(); n1++)
		{
			double totalWeight  = 0;
			std::vector<int> neighbors;
			std::vector<double> weights;
			
			// find neighborhood. TODO: r3 search
			for (int n2=0; n2<problem.NNodes(); n2++)
			{
				double dist = 0;
				if (n1!=n2)
				{
					for (int f=0; f<problem.NFeatures(); f++)	// find neighborhood
						dist += pow(x_curr(n1, f)-x_curr(n2, f), 2.);
				}
				neighbors.push_back(n2);
				double w = exp(-dist/h);
				weights.push_back(w);
				totalWeight += w;
				if (dist < convergedDist)
					maxWindowDist = vkMax(maxWindowDist, dist);
			}
			
			// find new position
			for (int nhd=0; nhd<(double)neighbors.size(); nhd++)
			{
				for (int f=0; f<problem.NFeatures(); f++)	// find neighborhood
					x_next(n1, f) += x_curr(neighbors[nhd], f)*weights[nhd]/totalWeight;
			}
		}
		
		// move points
		for (int n=0; n<problem.NNodes(); n++)
		for (int f=0; f<problem.NFeatures(); f++)	// find neighborhood
			x_curr(n,f) = x_next(n,f);
		
		if (maxWindowDist < terminateDist)
			break;
	}
	if (i==m_iMaxIterations)
	{
		std::cout<<"[ERROR] Mean shift clustering failed to converge.";
		std::cout<<"        Threshold: "<<maxWindowDist<<" / "<<terminateDist<<std::endl;
	}
	
	// extract clusters
	std::vector<std::vector<ClusteredNode> > & clusters = Clusters(problem);
	std::vector<int> clusterIDs;
	for (int i=0; i<problem.NNodes(); i++)
		clusterIDs.push_back(-1);
	for (int n1=0; n1<problem.NNodes(); n1++)
	{
		if (clusterIDs[n1]==-1) // add new cluster
		{
			int cid = (int)clusters.size();
			clusters.push_back(std::vector<ClusteredNode>());
			for (int n2=0; n2<problem.NNodes(); n2++)
			{
				double dist = 0;
				if (n1!=n2)
				{
					for (int f=0; f<problem.NFeatures(); f++)	// find neighborhood
						dist += pow(x_curr(n1, f)-x_curr(n2, f), 2.);
				}
				if (sqrt(dist) < convergedDist)
				{
					if (clusterIDs[n2]!=-1)
					{
						std::cout<<"[ERROR] Mean shift aggregation step failed"<<std::endl;
						continue;
					}
					clusterIDs[n2] = cid;
					ClusteredNode node;
					node.id = n2;
					node.error = 0;
					clusters[cid].push_back(node);
				}
			}
		}
	}
	
	// merge small clusters
	if (DEBUG_VERBOSE)
		std::cout<<"Merging small clusters..."<<std::endl;;
	for (int i=0; i<(int)clusters.size(); i++)
	{
		if (DEBUG_VERBOSE)
			std::cout<<"Cluster["<<i<<"]: "<<clusters[i].size()<<" / "<<m_iMinClusterSize<<std::endl;
		if ((int)clusters[i].size()<m_iMinClusterSize)
		{
			assert(clusters[i].size()>0);
			int n1 = clusters[i][0].id;
			
			// find big cluster to merge with
			int bestCluster =-1;
			double bestDistance = 0;
			for (int j=0; j<(int)clusters.size(); j++)
			{
				if (DEBUG_VERBOSE)
					std::cout<<"   Testing: "<<j<<", "<<clusters[j].size()<<": "<<std::flush;
				if (clusters[j].size()>=m_iMinClusterSize)	// acceptable
				{
					assert(clusters[j].size()>0);
					double dist2cluster = 0;
					int n2 = clusters[j][0].id;
					assert(n1!=n2);
					for (int f=0; f<problem.NFeatures(); f++)	// find neighborhood
						dist2cluster += pow(x_curr(n1, f)-x_curr(n2, f), 2.);
					if (DEBUG_VERBOSE)
						std::cout<<"d="<<dist2cluster<<std::endl;

					if (bestCluster==-1 || dist2cluster < bestDistance)
					{
						bestDistance = dist2cluster;
						bestCluster = j;
					}
				}
				else if (DEBUG_VERBOSE)
					std::cout<<std::endl;
			}
			
			if (bestCluster==-1)
				bestCluster = 0;
			assert(bestCluster!=-1);
			if (DEBUG_VERBOSE)
				std::cout<<"   Merging with: "<<bestCluster<<std::endl;
			
			if (i!=bestCluster)
			{
				for (int j=0; j<(int)clusters[i].size(); j++)
					clusters[bestCluster].push_back(clusters[i][j]);
				clusters.erase(clusters.begin()+i);
				i--;
			}
		}
	}
}









