/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim, Princeton University			   *
 *   Author: Vladimir Kim (kim_vladimir@yahoo.com, vk@princeton.edu)       *
 *   Research Supervisor: Tom Funkhouser (funk@cs.princeton.edu)           *
 *                                  ****                                   *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "TimeProfiler.h"
#ifdef CALCULATE_TIMING
	TimeProfiler * TimeProfiler::m_pGlobalSuperProfiler = new TimeProfiler();;
#endif

#include "VKStringList.h"

TimeProfiler::Process::Process()
{
	m_iRecordedUpdateDelta = -1;
	m_iCurrStartTime = -1;
	m_iTotalRunningTime = 0;
}

bool TimeProfiler::m_bNonVerbousProgressBar = false;

TimeProfiler::TimeProfiler()
{
}

void TimeProfiler::startedProcess(const VKString & processID)
{
	if (m_mIDToProcessMap[processID].m_iCurrStartTime!=-1)
		finishedProcess(processID);			//if process is currently running - finish it first
	
	m_mIDToProcessMap[processID].m_iCurrStartTime = clock();
	
	if (m_mIDToProcessMap[processID].m_sProcessName.isEmpty())
		m_mIDToProcessMap[processID].m_sProcessName = processID;
}

void TimeProfiler::finishedProcess(const VKString& processID, bool printTime)
{
	std::map<VKString, Process>::iterator iter = m_mIDToProcessMap.find(processID);

	//assert(iter!=m_mIDToProcessMap.end());
	if (iter==m_mIDToProcessMap.end())
		return;
	
	Process & process = iter->second;
	long long clockStamp = clock();
	if (process.m_iCurrStartTime==-1)
		process.m_iCurrStartTime = clockStamp;
	long long diff = clockStamp - process.m_iCurrStartTime;
	process.m_iTotalRunningTime += (diff);
	process.m_iCurrStartTime = -1;

	if (printTime)
	{
		std::cout<<">>>>>>> Finished: "<<process.m_sProcessName.toStdString();
		VKString totProcSecs = getTotalProcessLengthInSeconds(process.m_sProcessName);
		std::cout<<" - "<<totProcSecs.toStdString()<<std::endl;
	}
}

long long TimeProfiler::getTotalProcessLengthInCycles(const VKString& processID) const
{
	return ((TimeProfiler*)this)->m_mIDToProcessMap[processID].m_iTotalRunningTime;
}

VKString TimeProfiler::getTotalProcessLengthInSeconds(const VKString& processID) const
{
	TimeProfiler * ncThis = ((TimeProfiler*)this);
	long long remainder = ncThis->m_mIDToProcessMap[processID].m_iTotalRunningTime % CLOCKS_PER_SEC;
	remainder = (remainder * 1000) / CLOCKS_PER_SEC;
	VKString retVal= VKString::number(ncThis->m_mIDToProcessMap[processID].m_iTotalRunningTime / CLOCKS_PER_SEC);
	retVal += VKString(".") + VKString::number(remainder) + "s";
	return retVal;

}

int TimeProfiler::getTotalProcessLengthInSecondsInt(const VKString& processID) const
{
	TimeProfiler * ncThis = ((TimeProfiler*)this);
	return ncThis->m_mIDToProcessMap[processID].m_iTotalRunningTime / CLOCKS_PER_SEC;
}

VKString TimeProfiler::ProcessLengthStr(const VKString & processID) const
{
	int sec = getTotalProcessLengthInSecondsInt(processID);
	if (sec==0)
		return "0s";
	
	VKString str;
	if (sec / 3600 > 0)
		str += VKString::number(sec / 3600)+"h ";	
	sec = sec % 3600;

	if (sec / 60 > 0)
		str += VKString::number(sec / 60)+"m ";	
	sec = sec % 60;

	if (sec > 0)
		str += VKString::number(sec)+"s ";
	
	return str;
}

VKString TimeProfiler::profile()
{
	if (m_mIDToProcessMap.size()==0)
		return "Empty Time Profile";
	VKString profile = "Time Profile:\n";
	for (std::map<VKString, Process>::iterator iter = m_mIDToProcessMap.begin(); 
		 iter!=m_mIDToProcessMap.end(); iter++)
	{
		VKString procName = iter->first;
		procName.replace(" ", "_");
		profile += VKString("    ") + procName;
		profile += VKString(" \t ") + getTotalProcessLengthInSeconds(iter->first) + "\n";
	}
	return profile;
}

void TimeProfiler::WriteProgress(const VKString & process, int currProgress, int maxProgress)
{
	WriteProgress(process, (long long)currProgress, (long long)maxProgress);
}

void TimeProfiler::WriteProgress(const VKString & process, long long currProgress, long long maxProgress)
{
	finishedProcess(process);	
	
	if (!m_bNonVerbousProgressBar)
	{
		int TICK_SLOTS = 40;
		int currTicks = TICK_SLOTS * currProgress / maxProgress;
		std::cout<<"\r[";
		for (int i=0; i<TICK_SLOTS; i++)
		{
			if (i < currTicks)
				std::cout<<"+"<<std::flush;
			else
				std::cout<<" "<<std::flush;
		}
		
		long long currTime = getTotalProcessLengthInSecondsInt(process);
		long long remaining = (maxProgress-currProgress);
		long long remainingTime = ((currProgress==0) ? (-1) : (currTime * remaining / currProgress));
		std::cout<<"] N="<<remaining<<" T={"<<currTime<<"s | "<<remainingTime<<"s}    "<<std::flush;
	}
	else
	{
		std::map<VKString, Process>::iterator iter = m_mIDToProcessMap.find(process);
		if (iter==m_mIDToProcessMap.end())
		{
			std::cout<<"Started Iterations: "<<currProgress<<" / "<<maxProgress<<std::endl;
			startedProcess(process);
			m_mIDToProcessMap[process].m_iRecordedUpdateDelta = currProgress;
			finishedProcess(process);
		}
		else
		{
			long long dt = currProgress-iter->second.m_iRecordedUpdateDelta;
			iter->second.m_iRecordedUpdateDelta = dt;
			if (iter->second.m_iRecordedUpdateDelta + dt >=maxProgress)
			{
				VKString currTime = getTotalProcessLengthInSeconds(process);
				std::cout<<"Almost Finished: "<<currProgress<<" / "<<maxProgress<<": ";
				std::cout<<currTime.c_str()<<std::endl;
			}
		}
	}
	startedProcess(process);
}

void TimeProfiler::WriteIntermediateTimes(VKStringList & processes, int N)
{
	std::cout<<"\r[";	
	for (int i=0; i<processes.count(); i++)
	{
		finishedProcess(processes[i]);
		std::cout<<(getTotalProcessLengthInSecondsInt(processes[i]) / N)<<"s\t"<<std::flush;
		startedProcess(processes[i]);	
	}
		
	std::cout<<"] / "<<N<<std::flush;
}







