/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "DBCorrs.h"
#include "Model3D.h"
#include "CorrsPair.h"

CorrsPair::CorrsPair(const VKString & corrName, GeomShape * shapeFrom, 
					 GeomShape * shapeTo, DBCorrs * corrs, CoreCorrsType tp)
: ShapeCorrs(corrName, tp, corrs)
{
	m_pShapeFrom = shapeFrom;
	m_pShapeTo = shapeTo;
	
	m_bCacheFilled = false;
	m_iNCachedFromPnts = 0;
	m_aCachedCorrs = NULL;
	m_aCachedCorrVals = NULL;
}

CorrsPair::~CorrsPair()
{
	
}

VKString CorrsPair::GetCorrsClassName() const
{
	return "CorrsPair";
}

void CorrsPair::Domain(std::vector<const GeomShape *> & domainShapes) const 
{
	domainShapes.push_back(m_pShapeFrom);
}

void CorrsPair::Range(std::vector<const GeomShape *> & rangeShapes) const 
{
	rangeShapes.push_back(m_pShapeTo);
}


GeomShape * CorrsPair::ShapeFrom()
{
	return m_pShapeFrom;
}

GeomShape * CorrsPair::ShapeTo()
{
	return m_pShapeTo;
}

CorrsPair * CorrsPair::Inverse()
{
	return NULL;
}

ShapePointCloud * CorrsPair::SafeCastPntsFrom()
{
	return ShapeFrom()->SafeCastToPointCloud();
}

ShapePointCloud * CorrsPair::SafeCastPntsTo()
{
	return ShapeTo()->SafeCastToPointCloud();
}

void CorrsPair::SetFromShape(GeomShape * shape)
{
	m_pShapeFrom = shape;
}

void CorrsPair::SetToShape(GeomShape * shape)
{
	m_pShapeTo = shape;
}

const Model3D * CorrsPair::ModelFrom() const
{
	return m_pShapeFrom->GetModel();
}

const Model3D * CorrsPair::ModelTo() const
{
	return m_pShapeTo->GetModel();
}

void CorrsPair::WriteShapeCorrs(FILE * file)
{
	WriteShapeCorrsAsCorrsPair(file);
}

void CorrsPair::ReadShapeCorrs(FILE * file)
{
	ReadShapeCorrsAsCorrsPair(file);
}

void CorrsPair::WriteShapeCorrsAsCorrsPair(FILE * file)
{
	//WriteHeader(file, PairMapParams("CorrsPair", GetCorrsName(), ShapeFrom(), ShapeTo()));
	
	VKString cacheFiled;
	VKString valName;
	if (CacheFilled())
		cacheFiled = "CrrPairCached";
	else
		cacheFiled = "CrrPairNOT";

	if (m_aCachedCorrVals==NULL)
		valName = "none";
	else
		valName = "default";

	VKString::WriteString(cacheFiled, file);
	
	if (CacheFilled())
	{
		VKString::WriteString(valName, file);
		fwrite(&m_iNCachedFromPnts, sizeof(int), 1, file);
		// copy data to arrays
		for (int i=0; i<m_iNCachedFromPnts; i++)
			ShapePointCloud::WritePointSet(file, m_aCachedCorrs[i]);

		if (m_aCachedCorrVals!=NULL)
		{
			for (int i=0; i<m_iNCachedFromPnts; i++)
			{
				int nVals = (int)m_aCachedCorrVals[i].size();
				double * vals = new double[nVals]; 
				for (int j=0; j<nVals; j++)
					vals[j] = m_aCachedCorrVals[i][j];
				fwrite(&nVals, sizeof(int), 1, file);
				fwrite(vals, sizeof(double), nVals, file);
				delete [] vals;
			}
		}
	}
	else
		std::cout<<"[WARNING] Writing uncached CorrsPair (not useful)"<<std::endl;
}

void CorrsPair::ReadShapeCorrsAsCorrsPair(FILE * file)
{
	VKString cacheFiled = VKString::ReadString(file);
	if (cacheFiled == "CrrPairCached")
	{
		VKString valName = VKString::ReadString(file);
		fread(&m_iNCachedFromPnts, sizeof(int), 1, file);
		ShapePointCloud * pnts = SafeCastPntsFrom();
		assert(pnts!=NULL);
		assert(m_iNCachedFromPnts==pnts->NumPoints());
		InitCache(valName);
		
		// copy data to arrays
		for (int i=0; i<m_iNCachedFromPnts; i++)
			ShapePointCloud::WritePointSet(file, m_aCachedCorrs[i]);
		
		if (valName=="default")
		{
			assert(m_aCachedCorrVals!=NULL);
			for (int i=0; i<m_iNCachedFromPnts; i++)
			{
				int nVals;
				fread(&nVals, sizeof(int), 1, file);
				double * vals = new double[nVals]; 
				fread(vals, sizeof(double), nVals, file);
				for (int j=0; j<nVals; j++)
					m_aCachedCorrVals[i].push_back(vals[j]);
				delete [] vals;
			}
		}
	}
	else
	{
		assert(cacheFiled=="CrrPairNOT");
		std::cout<<"[WARNING] Reading uncached CorrsPair (not useful) "<<std::endl;
	}
}



int CorrsPair::GetCorrs(const ShapePoint & pnt,
						std::vector<ShapePoint> & corrs, 
						const VKString & corrValName,
						std::vector<double> *corrVals)
{
	if (CacheFilled())
		return GetCachedCorrs(pnt, corrs, corrValName, corrVals);

	assert(false);	// only handles pre-computed stuff
}

void CorrsPair::AutoFillCache(const VKString & corrValName)
{
	ShapePointCloud * pnts = SafeCastPntsFrom();
	// cannot cache if not discretized (at least use all vertices)
	assert(pnts!=NULL);	
	
	InitCache(corrValName);
	
	if (corrValName=="none")
	{
		for (int i=0; i<pnts->NumPoints(); i++)
			GetCorrs(pnts->GetPoint(i), m_aCachedCorrs[i], corrValName, NULL);	
	}
	else
	{
		// todo: cache more than one 'default' corr value
		assert(corrValName=="default");
		for (int i=0; i<pnts->NumPoints(); i++)
			GetCorrs(pnts->GetPoint(i), m_aCachedCorrs[i], 
					 corrValName, &(m_aCachedCorrVals[i]));
	}
	
	SetCacheFilled(true);
}

void CorrsPair::InitCache(const VKString & corrValName)
{
	ShapePointCloud * pnts = SafeCastPntsFrom();
	// cannot cache if not discretized (at least use all vertices)
	assert(pnts!=NULL);	
	m_iNCachedFromPnts = pnts->NumPoints();
	
	m_bCacheFilled = false;
	if (m_aCachedCorrs!=NULL)
		delete m_aCachedCorrs;
	
	if (m_aCachedCorrVals!=NULL)
		delete m_aCachedCorrVals;
	
	m_aCachedCorrs = new std::vector<ShapePoint>[m_iNCachedFromPnts];
	if (corrValName!="none")
		m_aCachedCorrVals = new std::vector<double>[m_iNCachedFromPnts];
}

bool CorrsPair::SetCachedCorrs(int id, const std::vector<ShapePoint> & corrs)
{
	assert(id>=0 && id < m_iNCachedFromPnts);
	m_aCachedCorrs[id] = corrs;
	return CacheFilled();
}

bool CorrsPair::SetCachedVals(int id, const std::vector<double> & vals)
{
	assert(m_aCachedCorrVals!=NULL); 
	assert(id>=0 && id < m_iNCachedFromPnts);
	m_aCachedCorrVals[id] = vals;
	return CacheFilled();
}

bool CorrsPair::GetCachedCorrs(int id,  std::vector<ShapePoint> & corrs)
{
	assert(id>=0 && id < m_iNCachedFromPnts);
	corrs = m_aCachedCorrs[id];
	return CacheFilled();
}

bool CorrsPair::GetCachedVals(int id, std::vector<double> & vals)
{
	assert(m_aCachedCorrVals!=NULL); 
	assert(id>=0 && id < m_iNCachedFromPnts);
	vals = m_aCachedCorrVals[id];
	return CacheFilled();
}

void CorrsPair::SetCacheFilled(bool filled)
{
	m_bCacheFilled = filled;
}

bool CorrsPair::CacheFilled() const
{
	return m_bCacheFilled;
}

int CorrsPair::GetCachedCorrs(const ShapePoint & pnt,
							   std::vector<ShapePoint> & corrs, 
							   const VKString & corrValName,
							   std::vector<double> *corrVals)
{
	assert(CacheFilled());
	ShapePointCloud * pf = SafeCastPntsFrom();
	assert(pf!=NULL);
	assert(pnt.SamePntsSet(pf));
	int id = pnt.PointID();
	
	bool valid = GetCachedCorrs(id,  corrs);
	assert(valid);
	
	if (corrVals!=NULL)
	{
		if (corrValName=="default")
		{
			valid = GetCachedVals(id, *corrVals);
			assert(valid);
		}
		else
			assert(corrValName=="none" && corrVals==NULL);
	}
	return (int)corrs.size();
}


