/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ShapeCorrs_H
#define __ShapeCorrs_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "GeomShape.h"
#include "ShapePoint.h"

class CorrValue;
class DBCorrs;

/** 
 * This is interface for correspondences. 
 * The interface is implemented with 3 core classes:
 *		CorrsCollection		- general 'fuzzy' DB correspondences
 *		CorrsCollectionMap	- general DB correspondences (one-to-one maps)
 *		CorrsPair			- 'fuzzy' correspondences for a pair of models
 *		CorrsPairMap		- Bijective pairwise mapping (one-to-one maps)
 * (each class can cache values and read cached values, thus it
 */
class ShapeCorrs
{
	public:
		enum CoreCorrsType
		{
			CORECORRS_CorrsCollection,
			CORECORRS_CorrsCollectionMap,
			CORECORRS_CorrsPair,
			CORECORRS_CorrsPairMap,
			CORECORRS_CorrsOther
		};
		
		static VKString m_sDefaultCorrValName;
		static VKString CoreCorrsTypeToStr(CoreCorrsType tp);
		static CoreCorrsType StrToCoreCorrsType(const VKString & str);
	
		ShapeCorrs(const VKString & corrsName, 
				   CoreCorrsType corrsType, DBCorrs * corrs);
		virtual ~ShapeCorrs();
		const VKString & GetCorrsName() const;
	
		CoreCorrsType GetCorrsType();
		virtual VKString GetCorrsClassName() const;
	
		virtual void Domain(std::vector<const GeomShape *> & domainShapes ) const = 0 ;
		virtual void Range(std::vector<const GeomShape *> & rangeShapes )  const = 0;
		virtual bool IsShapeSupportedDomain(const GeomShape * shape) const;
		virtual bool IsShapeSupportedRange(const GeomShape * shape) const;
		virtual bool PointInDomain(const ShapePoint & pnt) const;
	
		virtual int GetCorrs(const ShapePoint & pnt,
							 std::vector<ShapePoint> & corrs, 
							 const VKString & corrValName="none",
							 std::vector<double> *corrVals = NULL) = 0;

		virtual int PairwiseCorrs(const ShapePoint & pnt,
								  GeomShape * toShape,
								  std::vector<ShapePoint> & corrs, 
								  const VKString & corrValName = "none",
								  std::vector<double> * corrValue = NULL);	
	
		virtual ShapePoint BestCorr(const ShapePoint & pnt,
									GeomShape * toShape,
									const VKString & corrValName = "default",
									double * corrValue = NULL);	
	
		virtual double DefaultIntegratedValue(const VKString & evalPnts = "none") const;
	
		virtual void WriteShapeCorrs(const VKString & filename);
		virtual void ReadShapeCorrs(const VKString & filename);
	
		virtual void WriteShapeCorrs(FILE * file);
		virtual void ReadShapeCorrs(FILE * file);
	
		virtual void AddCorrValue(const VKString & corrValName, 
								  CorrValue * corrVal);
	
		virtual int NExtraCorrValues() const;
		virtual const VKString & CorrValueName(int corrValID) const; 
		virtual bool CorrValueSupported(const VKString & corrValStr) const;
		virtual bool IsDeafultSupported() const;
	
	protected:
		DBCorrs * m_pDBCorrs;
		virtual double ExtraCorrValue(const ShapePoint & p1, 
									  const ShapePoint & p2,
									  const VKString & corrValName);
		virtual void ExtraCorrValues(const ShapePoint & p1, 
									 std::vector<ShapePoint> & p2,
									 const VKString & corrValName, 
									 std::vector<double> & vals);
		virtual CorrValue * GetExtraCorrValue(const VKString & corrValName);
	
		VKString m_sCorrName;
		CoreCorrsType m_CorrsTp;

		std::vector<CorrValue * > m_CorrVals;
		std::map<VKString, int> m_NameToCorrValID;
};

#endif

