/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/
#include "DistGeodPnts.h"
#include "ShapePointCloud.h"
#include "Sortable.h"
#include "FileManagement.h"

DistGeodPnts::DistGeodPnts(ShapePointCloud * densepnts, const VKString & metricName, double norm)
: ShapeDistance(densepnts, metricName)
{
	m_pGraph = NULL;
	m_pPnts = densepnts;
	m_fNorm = norm;
}

DistGeodPnts::~DistGeodPnts()
{
	//delete m_pGraph;
}

DistGeodPnts * DistGeodPnts::SafeCastToGraphNN()
{
	return this;
}

void DistGeodPnts::InitNNGraph(double distThreshold, int minNN, int maxNN)
{
	assert(m_pGraph==NULL);
	m_pGraph = new VKGraph(m_pPnts->NumPoints());
	const R3Kdtree<const PointInSet *> * kdTree = m_pPnts->GetKdTree();
	bool thresholdGiven = (distThreshold!=-1);
	if (distThreshold==-1)
	{
		R3Point centroid = m_pPnts->Centroid();
		distThreshold = m_pPnts->AverageRadius(centroid) * 0.2;
	}
	for (int i=0; i<m_pPnts->NumPoints(); i++)
	{
		RNArray<const PointInSet *> neighbors;
		R3Point pos = m_pPnts->PointPosition(i);
		// get nearest neighbors
		kdTree->FindAll(pos, 0, distThreshold, neighbors);
		
		// if too few -- re-estimate
		double thresh = distThreshold;
		while (neighbors.NEntries() < minNN)
		{
			neighbors.Empty();
			kdTree->FindAll(pos, 0, thresh, neighbors);
			thresh *= 2.;
		}
		if (!thresholdGiven)
			distThreshold = thresh;
		
		// if too many -- sort and select
		if (neighbors.NEntries() > maxNN)
		{
			std::vector<Sortable> srtByDist;
			for (int n=0; n<neighbors.NEntries(); n++)
			{
				if (neighbors[n]->id!=i)
					srtByDist.push_back(Sortable(R3Distance(pos, neighbors[n]->GetPosition()), NULL, n));
			}
			std::sort(srtByDist.begin(), srtByDist.end());
			std::vector<const PointInSet *> copy;
			for (int n=0; n<maxNN; n++)
				copy.push_back(neighbors[srtByDist[n].id]);
			neighbors.Empty();
			for (int n=0; n<maxNN; n++)
				neighbors.InsertTail(copy[n]);
		}
		
		// add edges to the graph
		for (int n=0; n<neighbors.NEntries(); n++)
		{
			if (neighbors[n]->id!=i && !m_pGraph->Adjacent(i, neighbors[n]->id))
				m_pGraph->AddEdge(i, neighbors[n]->id, R3Distance(pos, neighbors[n]->GetPosition()));
		}
	}
	
	//std::cout<<"Connected components: "<<m_pGraph->FindConnectedComponents()<<std::endl;
	//assert(m_pGraph->FindConnectedComponents()==1);
}

double DistGeodPnts::NormalizedDistance(const ShapePoint & p1, const ShapePoint & p2) const
{
	assert(m_pGraph!=NULL);
	return -1;
}

void DistGeodPnts::NearestNeighborhood(const ShapePoint & p,  double distanceThreshold,
									   std::vector<ShapePoint> * neighborhood, std::vector<double> * normDists) const
{
	assert(m_pGraph!=NULL);
	// SLOW -- reimplement
	const R3Kdtree<const PointInSet *> * kdTree = m_pPnts->GetKdTree();
	const PointInSet * closest = kdTree->FindClosest(p.GetPosition());
	assert(closest!=NULL);
	double * shortestDistances = new double[m_pPnts->NumPoints()];
	//m_pGraph->ShortestPathDistances(closest->id, shortestDistances);
	m_pGraph->ThresholdedShortestPath(closest->id, distanceThreshold, shortestDistances);
	for (int i=0; i<m_pPnts->NumPoints(); i++)
	{
		if (shortestDistances[i] < distanceThreshold)
		{
			neighborhood->push_back(m_pPnts->GetPoint(i));
			if (normDists!=NULL)
				normDists->push_back(shortestDistances[i]);
		}
	}
	delete [] shortestDistances;
}

ShapePoint DistGeodPnts::NearestNeighbor(const ShapePoint & p, double* normDists) const
{
	assert(m_pGraph!=NULL);
	return ShapePoint();
}

const VKGraph * DistGeodPnts::GetGraph() const
{
	return m_pGraph;
}

void DistGeodPnts::WriteMetric(const VKString & filename) const
{
	assert(m_pGraph!=NULL);
	
	FileManagement::CreateDirsOnPathToFile(filename.c_str());
	FILE * file = fopen(filename.c_str(), "wb");
	assert(file!=NULL);
	WriteBasicMetricData(file);
	
	int nnodes = m_pGraph->NumNodes();
	fwrite(&nnodes, sizeof(int), 1, file);
	for (int n=0; n<nnodes; n++)
	{
		VKGraphNode * node = m_pGraph->Node(n);
		int ne = node->NEdges();
		fwrite(&ne, sizeof(int), 1, file);
		for (int e=0; e<node->NEdges(); e++)
			fwrite(&(node->Neighbor(e)->id), sizeof(int), 1, file);
	}
	
	fclose(file);
}

void DistGeodPnts::ReadMetric(FILE * file)
{
	assert(m_pGraph==NULL && m_pPnts!=NULL);
	int nnodes=0;
	fread(&nnodes, sizeof(int), 1, file);
	assert(nnodes==m_pPnts->NumPoints());
	m_pGraph = new VKGraph(nnodes);
	for (int n=0; n<nnodes; n++)
	{
		int ne=0;
		fread(&ne, sizeof(int), 1, file);
		for (int e=0; e<ne; e++)
		{
			int n2 =-1;
			fread(&n2, sizeof(int), 1, file);
			assert(n2>=0);
			double dist = R3Distance(m_pPnts->PointPosition(n), m_pPnts->PointPosition(n2));
			if (!m_pGraph->Adjacent(n, n2))
				m_pGraph->AddEdge(n, n2, dist);
		}
	}
}




