/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/
#include "DistGeodesicDijkstra.h"
#include "ShapeManifoldMesh.h"
#include "Model3D.h"

DistGeodesicDijkstra::DistGeodesicDijkstra(GeomShape * shape, 
										   const VKString & metricName, 
										   int cacheSize, double norm)
: ShapeDistance(shape, metricName)
{
	m_iMaxNumCached = cacheSize;
	m_fNorm = norm;
	
	GeomShape::GeomShapeTypes tp = shape->GetShapeType();
	assert(tp==GeomShape::SHAPE_TYPE_MANIFOLD);
	m_pManifoldMesh = ((ShapeManifoldMesh*) shape)->GetMesh();
	assert(m_pManifoldMesh!=NULL);

	if (m_fNorm<=0)
	{
		m_fNorm = 0;
		for (int f=0; f<m_pManifoldMesh->NFaces(); f++)
			m_fNorm += m_pManifoldMesh->FaceArea(m_pManifoldMesh->Face(f));
		m_fNorm = sqrt(m_fNorm);
		assert(m_fNorm>0);
	}
	
	
	int nVert = m_pManifoldMesh->NVertices();
	m_iMaxNumCached = cacheSize;
	m_PrecomputedRows = new double*[m_iMaxNumCached];
	for (int i=0; i<m_iMaxNumCached; i++)
		m_PrecomputedRows[i] = new double[nVert];
	m_aVertexIDToRow = new int[nVert];
	
	for (int i=0; i<nVert; i++)
		m_aVertexIDToRow[i] = -1;
	
	m_aPrecompData = InitPrecomputation(m_pManifoldMesh, NULL);

}

DistGeodesicDijkstra::~DistGeodesicDijkstra()
{
	
}

DistGeodesicDijkstra * DistGeodesicDijkstra::SafeCastToGeodesic()
{
	return this;
}

double DistGeodesicDijkstra::NormalizedDistance(double distance) const
{
	assert(m_fNorm>0);
	return distance / m_fNorm;
}

double DistGeodesicDijkstra::NormalizedDistance(const ShapePoint & p1, const ShapePoint & p2) const
{
	DistGeodesicDijkstra * ncThis = (DistGeodesicDijkstra*)this;
	
	bool atVtxP1 = false, atVtxP2 = false;
	int v1id = p1.VertexID(&atVtxP1);
	int v2id = p2.VertexID(&atVtxP2);
	
	if (atVtxP1 && atVtxP2)
	{
		if (!IsPrecomputed(v1id))
			ncThis->PrecomputeRow(v1id);
		return NormalizedDistance(ncThis->VertexDistance(v1id, v2id));
	}
	else
	{
		int f1ID;
		const ShapeMesh * mesh1 = NULL;
		double b1[3];
		int v1[3];
		p1.GetBarycentricCoords(&f1ID, &(b1[0]), &(b1[1]), &mesh1);
		b1[2] = 1. - b1[0] - b1[1];
		for (int i=0; i<3; i++)
			v1[i] = mesh1->VertexOnFace(f1ID, i);
		
		int f2ID;
		const ShapeMesh * mesh2 = NULL;
		double b2[3];
		int v2[3];
		p2.GetBarycentricCoords(&f2ID, &(b2[0]), &(b2[1]), &mesh2);
		b2[2] = 1. - b2[0] - b2[1];
		for (int i=0; i<3; i++)
			v2[i] = mesh2->VertexOnFace(f2ID, i);
		
		if (m_pShape!=mesh2 || m_pShape!=mesh1)
		{
			std::cout<<"[ERROR] Cannot find geodesic distance on different meshes.\nMeshes:"<<std::endl;
			std::cout<<"     Dist:   "<<m_pShape->GetModel()->ModelName().c_str();
			std::cout<<"::"<<m_pShape->GeometryName().c_str()<<std::endl;;
			std::cout<<"     From:   "<<mesh1->GetModel()->ModelName().c_str();
			std::cout<<"::"<<mesh1->GeometryName().c_str()<<std::endl;;
			std::cout<<"     To:     "<<mesh2->GetModel()->ModelName().c_str();
			std::cout<<"::"<<mesh2->GeometryName().c_str()<<std::endl;;
		}
		assert(m_pShape==mesh2 && m_pShape==mesh1);
		
		// see if p2 -> p1 is computed
		if (IsPrecomputed(v2[0]) && IsPrecomputed(v2[1]) && IsPrecomputed(v2[2]))
		{
			double d = 0;
			for (int i=0; i<3; i++)
			for (int j=0; j<3; j++)
				if (b2[i]!=0 && b1[j]!=0)
					d += ncThis->VertexDistance(v2[i], v1[j]) * b2[i] * b1[j];
			return NormalizedDistance(d);
		}
		
		// precompute stuff for p1 -> p2
		for (int i=0; i<3; i++)
			if (!IsPrecomputed(v1[i]))
				ncThis->PrecomputeRow(v1[i]);
		
		// fill distances
		double d = 0;
		for (int i=0; i<3; i++)
		for (int j=0; j<3; j++)
			if (b1[i]!=0 && b2[j]!=0)
				d += ncThis->VertexDistance(v1[i], v2[j]) * b1[i] * b2[j];
		return NormalizedDistance(d);
	}
}

//void DistGeodesicDijkstra::NearestNeighborhood(const ShapePoint & p, 
//											   double distanceThreshold,
//											   std::vector<ShapePoint> * neighborhood,
//											   std::vector<double> * normDists) const
//{
//	assert(false);
//}
//
//ShapePoint DistGeodesicDijkstra::NearestNeighbor(const ShapePoint & p, 
//												 double* normDists ) const
//{
//	assert(false);
//}

const double * DistGeodesicDijkstra::GetPrecomputedRow(int vID, int * rowSize) const
{
	assert (IsPrecomputed(vID));
	*rowSize = m_pManifoldMesh->NVertices();
	return m_PrecomputedRows[m_aVertexIDToRow[vID]];
}

bool DistGeodesicDijkstra::IsPrecomputed(int vertexID) const
{
	return m_aVertexIDToRow[vertexID]!=-1;
}

double DistGeodesicDijkstra::VertexDistance(int v1, int v2) const
{
	if (v1==v2)
		return 0;
	assert(v1>=0 && v2>=0);
	if (IsPrecomputed(v1))
		return m_PrecomputedRows[m_aVertexIDToRow[v1]][v2];
	else if (IsPrecomputed(v2))
		return m_PrecomputedRows[m_aVertexIDToRow[v2]][v1];
	else
	{
		assert(false);
		//PrecomputeRow(v1);
		//return m_PrecomputedRows[m_aVertexIDToRow[v1]][v2];
	}
}

void DistGeodesicDijkstra::ComputeVertexPaths(int v1, std::vector<int> & prevVtx) const
{
	DistGeodesicDijkstra * ncThis = (DistGeodesicDijkstra*)this;
	ncThis->m_aPrecompData = InitPrecomputation(ncThis->m_pManifoldMesh, ncThis->m_aPrecompData);
	//PrecomputeFill(m_aPrecompData, m_pManifoldMesh, vertexID, m_PrecomputedRows[rowID]);

	R3MeshVertex *source_vertex = m_pManifoldMesh->Vertex(v1);
	
	// Re-set vertex data
	prevVtx.clear();
	for (int j = 0; j < m_pManifoldMesh->NVertices(); j++)
	{
		ncThis->m_aPrecompData[j].distance = FLT_MAX;
		ncThis->m_aPrecompData[j].heappointer = NULL;
		prevVtx.push_back(-1);
	}
	 
	// Initialize priority queue
	GeoVertexData *data = (GeoVertexData *) m_pManifoldMesh->VertexData(source_vertex);
	assert(data!=NULL);
	data->distance = 0;
	RNHeap<GeoVertexData *> heap(data, &(data->distance), &(data->heappointer));
	heap.Push(data);
	
	// Visit other nodes computing shortest distance
	while (!heap.IsEmpty())
	{
		GeoVertexData *data = heap.Pop();
		R3MeshVertex *vertex = data->vertex;
		int v0id = m_pManifoldMesh->VertexID(vertex);
		
		for (int j = 0; j < m_pManifoldMesh->VertexValence(vertex); j++)
		{
			R3MeshEdge *edge = m_pManifoldMesh->EdgeOnVertex(vertex, j);
			R3MeshVertex *neighbor_vertex = m_pManifoldMesh->VertexAcrossEdge(edge, vertex);
			GeoVertexData *neighbor_data = (GeoVertexData *) m_pManifoldMesh->VertexData(neighbor_vertex);
			RNScalar old_distance = neighbor_data->distance;
			
			RNScalar new_distance = m_pManifoldMesh->EdgeLength(edge) + data->distance;
			assert(new_distance>=0);
			if (new_distance < old_distance)
			{
				int vid = m_pManifoldMesh->VertexID(neighbor_vertex);
				assert(vid < (int)prevVtx.size());
				prevVtx[vid] = v0id;
				neighbor_data->distance = new_distance;
				if (old_distance < FLT_MAX) heap.Update(neighbor_data);
				else heap.Push(neighbor_data);
			}
		}
	}
	
//	if (toVerticesRemember.size()>0)
//	{
//		for (int i=0; i<(int)toVerticesRemember.size(); i++)
//		{
//			R3MeshVertex * vtx= m_pManifoldMesh->Vertex(toVerticesRemember[i]);
//			toDistanceArray[i] = ((GeoVertexData*)mesh->VertexData(vtx))->distance;
//			assert(toDistanceArray[i]>=0);
//		}
//	}
//	else
//	{
//		for (int i=0; i<mesh->NVertices(); i++)
//		{
//			R3MeshVertex * vtx = mesh->Vertex(i);
//			toDistanceArray[i] = ((GeoVertexData*)mesh->VertexData(vtx))->distance;
//			assert(toDistanceArray[i]>=0);
//		}
//	}
}

void DistGeodesicDijkstra::PrecomputeRow(int vertexID)
{
	if (IsPrecomputed(vertexID))
		return;
	
	int rowID=-1;
	if ((int)m_AddedVertices.size() >= m_iMaxNumCached)
	{
		int currVertexID = m_AddedVertices[0];
		rowID = m_aVertexIDToRow[currVertexID];
		m_AddedVertices.erase(m_AddedVertices.begin());
		m_aVertexIDToRow[currVertexID] = -1;
	}
	else	// cache not filled yet 
	{
		rowID = (int)m_AddedVertices.size();
	}
	assert(rowID < m_iMaxNumCached && rowID>=0);
	
	m_AddedVertices.push_back(vertexID);
	assert(m_AddedVertices.size() <= m_iMaxNumCached);
	m_aVertexIDToRow[vertexID] = rowID;
	
	m_aPrecompData = InitPrecomputation(m_pManifoldMesh, m_aPrecompData);
	
	PrecomputeFill(m_aPrecompData, m_pManifoldMesh, vertexID, m_PrecomputedRows[rowID]);
}





DistGeodesicDijkstra::GeoVertexData * 
DistGeodesicDijkstra::InitPrecomputation(R3Mesh * mesh, 
										 DistGeodesicDijkstra::GeoVertexData * vtx_data)
{	
	if (vtx_data==NULL)
		vtx_data = new GeoVertexData [ mesh->NVertices() ];
	assert(vtx_data!=NULL);
	
	// Initialize vertex data
	for (int i = 0; i < mesh->NVertices(); i++) 
	{
		R3MeshVertex *vertex = mesh->Vertex(i);
		mesh->SetVertexData(vertex, &(vtx_data[i]));
		vtx_data[i].vertex = vertex;	
	}
	
	return vtx_data;

}

void DistGeodesicDijkstra::PrecomputeFill(DistGeodesicDijkstra::GeoVertexData * vertex_data, R3Mesh * mesh, int fromVertex,
										  double * toDistanceArray, const std::vector<int> & toVerticesRemember)
{
	R3MeshVertex *source_vertex = mesh->Vertex(fromVertex);
	
	// Re-set vertex data
	for (int j = 0; j < mesh->NVertices(); j++) 
	{
		vertex_data[j].distance = FLT_MAX;
		vertex_data[j].heappointer = NULL;
	}
	
	// Initialize priority queue
	GeoVertexData *data = (GeoVertexData *) mesh->VertexData(source_vertex);
	assert(data!=NULL);
	data->distance = 0;
	RNHeap<GeoVertexData *> heap(data, &(data->distance), &(data->heappointer));
	heap.Push(data);		
	
	// Visit other nodes computing shortest distance
	while (!heap.IsEmpty()) 
	{
		GeoVertexData *data = heap.Pop();
		R3MeshVertex *vertex = data->vertex;
		
		for (int j = 0; j < mesh->VertexValence(vertex); j++) 
		{
			R3MeshEdge *edge = mesh->EdgeOnVertex(vertex, j);
			R3MeshVertex *neighbor_vertex = mesh->VertexAcrossEdge(edge, vertex);
			GeoVertexData *neighbor_data = (GeoVertexData *) mesh->VertexData(neighbor_vertex);
			RNScalar old_distance = neighbor_data->distance;
			
			RNScalar new_distance = mesh->EdgeLength(edge) + data->distance;
			assert(new_distance>=0);
			if (new_distance < old_distance) 
			{
				neighbor_data->distance = new_distance;
				if (old_distance < FLT_MAX) heap.Update(neighbor_data);
				else heap.Push(neighbor_data);
			}
		}
	}
	
	if (toVerticesRemember.size()>0)
	{
		for (int i=0; i<(int)toVerticesRemember.size(); i++)
		{
			R3MeshVertex * vtx= mesh->Vertex(toVerticesRemember[i]);
			toDistanceArray[i] = ((GeoVertexData*)mesh->VertexData(vtx))->distance;
			assert(toDistanceArray[i]>=0);			
		}
	}
	else
	{
		for (int i=0; i<mesh->NVertices(); i++)
		{
			R3MeshVertex * vtx = mesh->Vertex(i);
			toDistanceArray[i] = ((GeoVertexData*)mesh->VertexData(vtx))->distance;	
			assert(toDistanceArray[i]>=0);			
		}
	}
}

void DistGeodesicDijkstra::FillThresholded(DistGeodesicDijkstra::GeoVertexData * vertex_data, 
										   R3Mesh * mesh, int fromVertex,
										   std::map<int, double> & verticesWithValues, 
										   double maxDistance)
{
	assert(false);	// not using for now
	
	R3MeshVertex *source_vertex = mesh->Vertex(fromVertex);
	
	// distance to self
	verticesWithValues[fromVertex] = 0;
	
	// Re-set vertex data
	for (int j = 0; j < mesh->NVertices(); j++) 
	{
		vertex_data[j].distance = FLT_MAX;
		vertex_data[j].heappointer = NULL;
	}
	
	// Initialize priority queue
	GeoVertexData *data = (GeoVertexData *) mesh->VertexData(source_vertex);
	assert(data!=NULL);
	data->distance = 0;
	RNHeap<GeoVertexData *> heap(data, &(data->distance), &(data->heappointer));
	heap.Push(data);		
	
	// Visit other nodes computing shortest distance
	while (!heap.IsEmpty()) 
	{
		GeoVertexData *data = heap.Pop();
		R3MeshVertex *vertex = data->vertex;
		
		for (int j = 0; j < mesh->VertexValence(vertex); j++) 
		{
			R3MeshEdge *edge = mesh->EdgeOnVertex(vertex, j);
			R3MeshVertex *neighbor_vertex = mesh->VertexAcrossEdge(edge, vertex);
			GeoVertexData *neighbor_data=(GeoVertexData*)mesh->VertexData(neighbor_vertex);
			RNScalar old_distance = neighbor_data->distance;
			
			RNScalar new_distance = mesh->EdgeLength(edge) + data->distance;
			assert(new_distance>=0);
			if (new_distance < old_distance && new_distance < maxDistance)
			{
				neighbor_data->distance = new_distance;
				verticesWithValues[mesh->VertexID(neighbor_data->vertex)] = new_distance;
				
				if (old_distance < FLT_MAX) heap.Update(neighbor_data);
				else heap.Push(neighbor_data);
			}
		}
	}	
}

double DistGeodesicDijkstra::GetThresholded(int toVertex, R3Mesh * mesh, 
											std::map<int, double> & verticesWithValues, 
											double maxDistance)
{
	std::map<int, double>::iterator iter = verticesWithValues.find(toVertex);
	if (iter==verticesWithValues.end())
		return maxDistance;
	else
		return iter->second;
}
