/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "ShapeGrid.h"
#include "TimeProfiler.h"
#include "ShapePointCloud.h"



ShapeGrid::ShapeGrid(ShapeMesh * mesh, double rotation, double cellsize)
{
	m_pMesh = mesh;
	m_iNumDilations = 1;;	// 4
	InitGrid(cellsize, pow(1024, 3), rotation);
}

double ShapeGrid::PointDepth(const R3Point & point, bool interpolate) const
{
	R3Point tp = point;
	m_XForm.Apply(tp);
	
	double value = 0;
	if (interpolate)
		value = m_pObjectDepthGrid->WorldValue(tp);	// real penalty
	else
	{
		//R3Point gp = m_pObjectDepthGrid->GridPosition(tp);
		const R3Affine & w2g = m_pObjectDepthGrid->WorldToGridTransformation();
		R3Point gp = tp;
		w2g.Apply(gp);
		
		int x = (int)(gp.X()+0.5);
		int y = (int)(gp.Y()+0.5);
		int z = (int)(gp.Z()+0.5);
		
		if (x>=0 && x < m_pObjectDepthGrid->XResolution()
			&& y>=0 && y < m_pObjectDepthGrid->YResolution()
			&& z>=0 && z < m_pObjectDepthGrid->ZResolution())
			value = m_pObjectDepthGrid->GridValue(x, y, z);
	}
	
	return value*m_fGridCellSize;
}

void ShapeGrid::InitGrid(double cellsize, int maxCells, double rotation)
{
	m_fGridCellSize = cellsize;
	
	TimeProfiler timer;
	
	if (m_pMesh->NFaces()==0)
	{
		std::cout<<"[WARNING] No faces => no intersection penalty"<<std::endl;
		return;
	}
	// TODO: PCA alignment
	
	// Create grid
	R3Box bbox;
	m_XForm = R3identity_affine;
	m_XForm.ZRotate(-rotation);
	for (int i=0; i<m_pMesh->NVertices(); i++)
	{
		R3Point p = m_pMesh->VertexPosition(i);
		m_XForm.Apply(p);
		if (i==0)
			bbox = R3Box(p,p);
		else
			bbox.Union(p);
	}
	
	std::cout<<"BBox: ["<<bbox.XMin()<<" "<<bbox.XMax()<<"], ["<<bbox.YMin()<<" "<<bbox.YMax();
	std::cout<<"], ["<<bbox.ZMin()<<" "<<bbox.ZMax()<<"] "<<std::endl;
	
	bbox = R3Box(R3Point(bbox.XMin() - m_fGridCellSize*6, bbox.YMin() - m_fGridCellSize*6, bbox.ZMin() - m_fGridCellSize*6),
				 R3Point(bbox.XMax() + m_fGridCellSize*6, bbox.YMax() + m_fGridCellSize*6, bbox.ZMax() + m_fGridCellSize*6));
	
	int xres=-1, yres=-1, zres=-1;
	int maxiter = 10000;
	while (maxiter-->=0)
	{
		xres = (bbox.XMax()-bbox.XMin()) / m_fGridCellSize;
		yres = (bbox.YMax()-bbox.YMin()) / m_fGridCellSize;
		zres = (bbox.ZMax()-bbox.ZMin()) / m_fGridCellSize;
		assert(xres>=0 && yres>=0 && zres>=0);
		if (xres*yres*zres > maxCells)
			m_fGridCellSize *= 1.1;
		else
			break;
	}
	assert(maxiter>0);
	
	std::cout<<"Resolution: "<<xres<<" x "<<yres<<" x "<<zres<<std::endl;
	m_pObjectDepthGrid = new R3Grid(xres, yres, zres, bbox);
	R3Grid * mask = new R3Grid(xres, yres, zres, bbox);
	//m_pObjectVoronoiGrid = new R3Grid(xres, yres, zres, bbox);
	//m_pDist2SurfaceGrid = new R3Grid(xres, yres, zres, bbox);
	m_pObjectVoronoiGrid = NULL;
	m_pDist2SurfaceGrid = NULL;
	
	
	//m_pObjectDepthGrid = NULL;
	//R3Grid * mask = NULL;
	//	m_pObjectVoronoiGrid = NULL;
	//	m_pDist2SurfaceGrid = NULL;
	
	
	if (m_pObjectDepthGrid!=NULL)	m_pObjectDepthGrid->Clear(0);
	if (m_pObjectVoronoiGrid!=NULL)	m_pObjectVoronoiGrid->Clear(0);
	if (m_pDist2SurfaceGrid!=NULL)	m_pDist2SurfaceGrid->Clear(0);
	
	// Fill grid values
	if (m_pObjectDepthGrid!=NULL || m_pDist2SurfaceGrid!=NULL)
	for (int f=0; f<m_pMesh->NFaces(); f++)
	{
		R3Point p1 = m_pMesh->VertexPosition(m_pMesh->VertexOnFace(f, 0));
		R3Point p2 = m_pMesh->VertexPosition(m_pMesh->VertexOnFace(f, 1));
		R3Point p3 = m_pMesh->VertexPosition(m_pMesh->VertexOnFace(f, 2));
		m_XForm.Apply(p1);
		m_XForm.Apply(p2);
		m_XForm.Apply(p3);
		
		if (m_pObjectDepthGrid!=NULL)
			m_pObjectDepthGrid->RasterizeWorldTriangle(p1, p2, p3, 1.);
		if (m_pDist2SurfaceGrid!=NULL)
			m_pDist2SurfaceGrid->RasterizeWorldTriangle(p1, p2, p3, 1.);
	}
	
	if (m_pObjectDepthGrid!=NULL)
		m_pObjectDepthGrid->Dilate(m_iNumDilations);
	
	if (m_pObjectVoronoiGrid!=NULL)
	{
		assert(false);
//		ShapePointCloud * densepnts = m_Interactors[Interactor::ACT_Object]->m_pObjectDensePoints;
//		assert(densepnts!=NULL);
//		int fid;
//		double b1, b2;
//		const ShapeMesh * meshcurr=NULL;
//		std::cout<<"Fill Voronoi: "<<std::endl;;
//		for (int pid=0; pid<densepnts->NumPoints(); pid++)
//		{
//			if (pid%1000==0)
//				timer.WriteProgress("isectpoints", pid, densepnts->NumPoints());
//			densepnts->GetPoint(pid).GetBarycentricCoords(&fid, &b1, &b2, &meshcurr);
//			assert(meshcurr==m_pMesh);
//			R3Point gp = densepnts->PointPosition(pid);
//			m_XForm.Apply(gp);
//			R3Point gp2 = m_pObjectVoronoiGrid->GridPosition(gp);
//			gp = gp2;
//			int i = (int)gp.X();
//			int j = (int)gp.Y();
//			int k = (int)gp.Z();
//			if (i>=0 && i<m_pObjectVoronoiGrid->XResolution()
//				&& j>=0 && j<m_pObjectVoronoiGrid->YResolution()
//				&& k>=0 && k<m_pObjectVoronoiGrid->ZResolution())
//			{
//				if (m_pObjectVoronoiGrid->GridValue(i, j, k)==0)
//					m_pObjectVoronoiGrid->SetGridValue(i,j,k, fid+1);
//			}
//		}
//		m_pObjectVoronoiGrid->Voronoi();
	}
	
	if (m_pDist2SurfaceGrid!=NULL)
		assert(false);
//		m_pDist2SurfaceGrid->SquaredDistanceTransform();
	
	// Fill 'inside mask'
	if (mask!=NULL)
		mask->Add(1.);	// set all to 'included' values
	
	if (m_pObjectDepthGrid!=NULL)
	{
		assert(mask!=NULL);
		// simple inside-out segmentation -- take orthogonal wavefront passes
		for (int passID=0; passID<5; passID++)
		{
			std::cout<<"Inside-out pass: "<<passID<<" / 5"<<std::endl;
			double passDirection;
			int passDimension;
			
			if (passID==0)	// top-down
			{
				passDimension = 2;
				passDirection = -1;	// down
			}
			else if (passID==1 || passID==2)	// left-right
			{
				passDimension = 0;
				passDirection = (passID==1) ? 1 : -1;
			}
			else if (passID==3 || passID==4)	// front-back
			{
				passDimension = 1;
				passDirection = (passID==3) ? 1 : -1;
			}
			else
				assert(false);
			
			int dim1 = 0;
			int dim2 = 1;
			
			if (passDimension==0)
				dim1 = 2;
			else if (passDimension==1)
				dim2 = 2;
			
			R3Point gp;
			for (int j=0; j<mask->Resolution(dim1); j++)
				for (int k=0; k<mask->Resolution(dim2); k++)
					for (int i=0; i<mask->Resolution(passDimension); i++)
					{
						if (passDirection<0)
							gp[passDimension] = (double)(mask->Resolution(passDimension) - 1 - i);
						else
							gp[passDimension] = (double)i;
						gp[dim1] = (double)j;
						gp[dim2] = (double)k;
						
						if (m_pObjectDepthGrid->GridValue((int)gp.X(), (int)gp.Y(), (int)gp.Z())>0)
							break;
						
						mask->SetGridValue((int)gp.X(), (int)gp.Y(), (int)gp.Z(), 0.);
					}
		}
		std::vector<int> nhdX;
		std::vector<int> nhdY;
		std::vector<int> nhdZ;
		for (int i=0; i<6; i++)
		{
			nhdX.push_back(0);
			nhdY.push_back(0);
			nhdZ.push_back(0);
		}
		nhdX[0] = -1;
		nhdX[1] = 1;
		nhdY[2] = -1;
		nhdY[3] = 1;
		nhdZ[4] = -1;
		nhdZ[5] = 1;
		
		int nx = mask->Resolution(0);
		int ny = mask->Resolution(1);
		int nz = mask->Resolution(2);
		std::cout<<"Region-growing..."<<std::endl;
		
		//	// region-growing to prune remaining
		for (int i=1; i<nx-1; i++)
			for (int j=1; j<ny-1; j++)
				for (int k=1; k<nz-1; k++)
				{
					// if non-surface and touching "empty" - start region growing
					bool ijk_maskedOut = mask->GridValue(i,j,k)<1;
					bool ijk_onSurface = m_pObjectDepthGrid->GridValue(i,j,k)>0;
					bool visit = !ijk_maskedOut && !ijk_onSurface;
					
					int nminadj = 1;
					int adjacentEmpty = 0;
					for (int n=0; n<6 && adjacentEmpty<nminadj && visit; n++)
					{
						bool nijk_maskedOut = mask->GridValue(i+nhdX[n], j+nhdY[n], k+nhdZ[n])<1;
						bool nijk_onSurface = m_pObjectDepthGrid->GridValue(i+nhdX[n], j+nhdY[n], k+nhdZ[n])>0;
						if (nijk_maskedOut && !nijk_onSurface)
							adjacentEmpty++;
					}
					if (j==1 && k==1)
						timer.WriteProgress("progisectgrid", i, nx);
					std::vector<int> regionGrowing;
					int nvisit = 0;
					if (visit && adjacentEmpty>=nminadj)
					{
						regionGrowing.push_back(i*ny*nz + j*nz + k);
						while (regionGrowing.size()>0)
						{
							int id = regionGrowing[0];
							regionGrowing.erase(regionGrowing.begin());
							int ci = id / (ny*nz);
							int cj = (id - ci * ny*nz)/nz;
							int ck = id - ci*ny*nz - cj*nz;
							bool maskedOut1 = mask->GridValue(ci, cj, ck)<1;
							if (maskedOut1)
								continue;
							nvisit++;
							//mask->SetGridValue(ci,cj,ck, .5);
							mask->SetGridValue(ci,cj,ck, 0);
							
							for (int n=0; n<6; n++)
							{
								int cci = ci+nhdX[n];
								int ccj = cj+nhdY[n];
								int cck = ck+nhdZ[n];
								bool validBounds = cci>=0 && cci<nx && ccj>=0 && ccj<ny && cck>=0 && cck<nz;
								bool onSurface = validBounds && (m_pObjectDepthGrid->GridValue(cci, ccj, cck)>0);
								bool maskedOut = validBounds && (mask->GridValue(cci, ccj, cck)<1);
								if (validBounds && !maskedOut && !onSurface)
									regionGrowing.push_back(cci*ny*nz + ccj*nz + cck);
							}
						}
						//std::cout<<"\nRegion-grow: ["<<i<<" "<<j<<" "<<k<<"] "<<nvisit<<std::endl;
					}
				}
		std::cout<<std::endl;
		
		// multiply with mask
		std::cout<<"Squared Dist Transform"<<std::endl;
		m_pObjectDepthGrid->Clear();
		for (int x=1; x<mask->XResolution()-1; x++)
		for (int y=1; y<mask->YResolution()-1; y++)
		for (int z=1; z<mask->ZResolution()-1; z++)
		{
			if (mask->GridValue(x, y, z)>0)
			{
				for (int n=0; n<(int)nhdX.size(); n++)
				{
					int cci = x+nhdX[n];
					int ccj = y+nhdY[n];
					int cck = z+nhdZ[n];
					bool validBounds = cci>=0 && cci<nx && ccj>=0 && ccj<ny && cck>=0 && cck<nz;
					bool maskedOut = validBounds && (mask->GridValue(cci, ccj, cck)<1);
					if (validBounds && maskedOut)
						m_pObjectDepthGrid->SetGridValue(x, y, z, 1.);
				}
			}
		}
		m_pObjectDepthGrid->SquaredDistanceTransform();
		
		//m_pObjectDepthGrid->DetectEdges();
		//m_pObjectDepthGrid->Threshold(.5, 0, 1.);
		//m_pObjectDepthGrid->SquaredDistanceTransform();
		
		
		std::cout<<"Multiply by mask"<<std::endl;
		m_pObjectDepthGrid->Multiply(*mask);

		std::cout<<"Prepare grid inside points (visualization only)"<<std::endl;
		// inside points (for visualization)
		m_pInsidePoints = new ShapePointCloud(NULL, "gridpnts");
		for (int x=0; x<m_pObjectDepthGrid->XResolution(); x++)
		for (int y=0; y<m_pObjectDepthGrid->YResolution(); y++)
		for (int z=0; z<m_pObjectDepthGrid->ZResolution(); z++)
		{
			if (m_pObjectDepthGrid->GridValue(x, y, z)>0)
			{
				//std::cout<<"Add point: "<<x<<" "<<y<<" "<<z<<std::endl;
				R3Point grid((double) x, (double)y, (double)z);
				m_pObjectDepthGrid->GridToWorldTransformation().Apply(grid);
				m_XForm.ApplyInverse(grid);
				m_pInsidePoints->AddPoint(ShapePoint(grid));
			}
		}

		std::cout<<"Grid Filled Points: "<<m_pInsidePoints->NumPoints()<<std::endl;
		delete mask;
	}
}
