/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "VKStringList.h"
#include "VkFunctions.h"
#include "Sortable.h"

#include "ShapePointCloud.h"
#include "ShapeManifoldMesh.h"
#include "ShapeTriangleSoup.h"
#include "ShapeLabels.h"
#include "ShapeMesh.h"
#include "Model3D.h"


SceneGraphLabel::SceneGraphLabel()
{
	m_pParent = NULL;
	m_bCachedIDs = false;
	
}


ShapeLabels * ShapeLabels::CopyLabels(ShapeLabels * labels, GeomShape * target)
{
	ShapeLabels * copy = new ShapeLabels(target, labels->GetShapeValueName());
	
	// copy tree hierarchy
	std::vector<SceneGraphLabel*> newLabels;
	std::vector<SceneGraphGeometry*> geoms = labels->m_AllGeometries;

	for (int i=0; i<(int)labels->m_AllNodes.size(); i++)
	{
		SceneGraphLabel * newLabel = new SceneGraphLabel();
		newLabel->m_sLabel = labels->m_AllNodes[i]->m_sLabel;
		newLabel->m_iID = i;
		newLabel->m_pParent = NULL;
		newLabels.push_back(newLabel);
	}
	
	for (int i=0; i<(int)labels->m_AllNodes.size(); i++)
	{
		for (int c=0; c<(int)labels->m_AllNodes[i]->m_apChildren.size(); c++)
		{
			int cID = labels->m_AllNodes[i]->m_apChildren[c]->m_iID;
			newLabels[i]->m_apChildren.push_back(newLabels[cID]);
			newLabels[cID]->m_pParent = newLabels[i];
		}
	}
	copy->SetAllLabels(newLabels);
	copy->SetAllGeometries(geoms);
	
	// copy labels
	ShapePointCloud * pnts = target->SafeCastToPointCloud();
	ShapeMesh * mesh = target->SafeCastToMesh();
	if (pnts!=NULL)
	{
		assert(pnts->GetMesh() == labels->m_pMesh);
		std::vector<int> pntSegIDs;
		for (int i=0; i<pnts->NumPoints(); i++)
			pntSegIDs.push_back((int)labels->Value(pnts->GetPoint(i), 0));
		copy->SetPntsSegments(pntSegIDs);
	}
	else if (mesh!=NULL)
	{
		ShapePointCloud * pnts = NULL;
		bool cleanPnts = false;
		if (labels->m_pPnts!=NULL)
			pnts = labels->m_pPnts;
		else if (labels->m_pMesh!=NULL)
		{
			cleanPnts= true;
			pnts = new ShapePointCloud(NULL, "pnts");
			for (int i=0; i<labels->m_pMesh->NFaces(); i++)
				pnts->AddPoint(ShapePoint(labels->m_pMesh, i, 1./3., 1./3., pnts, i));
		}
		else
			assert(false);

		const R3Kdtree<const PointInSet *> * kdTree = pnts->GetKdTree();

		std::vector<int> faceSegIDs;
		
		for (int i=0; i<mesh->NFaces(); i++)
		{
			R3Point pos = mesh->FaceCentroid(i);
			const PointInSet * closest = kdTree->FindClosest(pos);
			assert(closest!=NULL);
			faceSegIDs.push_back(labels->Value(closest->GetModelPoint(), 0));
		}
		copy->SetFaceSegments(faceSegIDs);
		if (cleanPnts)
			delete pnts;
	}
	return copy;
}

ShapeLabels * ShapeLabels::SplitLabelsIntoCC(ShapeLabels * labels)
{
	// assume flat hierarchy and a mesh (for now)
	std::vector<SceneGraphLabel*> newLabels;
	std::vector<int> face2seg;

	assert(labels->GetShape()->SafeCastToMesh()!=NULL);
	ShapeMesh * mesh = labels->GetShape()->SafeCastToMesh();
	for (int f=0; f<mesh->NFaces(); f++)
		face2seg.push_back(0);
		
	int root = labels->RootNode();
	assert(labels->NodeNChildren(root) == labels->NNodes()-1);
	
	SceneGraphLabel * rootPtr = new SceneGraphLabel();
	rootPtr->m_sLabel = labels->NodeLabel(root);
	rootPtr->m_iID = 0;
	rootPtr->m_pParent = NULL;
	newLabels.push_back(rootPtr);

	std::cout<<"Creating CC decomposition of labeling"<<std::endl;
	for (int i=0; i<labels->NodeNChildren(root); i++)
	{
		int cid = labels->NodeChild(root, i);
		
		std::vector<int> allFacesWithLabel = labels->NodeFacesAtNode(cid);
		std::map<int, bool> allFacesWithLabelMap;
		for (int f=0; f<(int)allFacesWithLabel.size(); f++)
			allFacesWithLabelMap[allFacesWithLabel[f]] = true;
		
		// split faces into connected components
		std::vector<std::vector<int> > CCs;
		while (allFacesWithLabelMap.size()>0)
		{
			int fid = allFacesWithLabelMap.begin()->first;
			allFacesWithLabelMap.erase(allFacesWithLabelMap.begin());
			std::vector<int> toVisit;
			toVisit.push_back(fid);
			int cid = (int)CCs.size();
			CCs.push_back(std::vector<int>());

			while (toVisit.size()>0)
			{
				int vfid = toVisit[0];
				toVisit.erase(toVisit.begin());
				CCs[cid].push_back(vfid);
				
				std::map<int, bool> adjacent;
				for (int adjV = 0; adjV < 3; adjV++)
				for (int adjF = 0; adjF < mesh->NAdjacentFaces(mesh->VertexOnFace(vfid, adjV)); adjF++)
					adjacent[mesh->FaceOnVertex(mesh->VertexOnFace(vfid, adjV), adjF)] = true;
				
				for (std::map<int, bool>::iterator nh = adjacent.begin(); nh != adjacent.end(); nh++)
				{
					std::map<int, bool>::iterator origSetIter = allFacesWithLabelMap.find(nh->first);
					if (origSetIter != allFacesWithLabelMap.end())	// actually part of correct label
					{
						toVisit.push_back(nh->first); // => grow CC region
						allFacesWithLabelMap.erase(origSetIter);
					}
				}
			}
		}
		
		// add connected components
		std::cout<<"  "<<labels->NodeLabel(cid).c_str()<<" - "<<CCs.size()<<std::endl;
		for (int c=0; c<(int)CCs.size(); c++)
		{
			SceneGraphLabel * newLabel = new SceneGraphLabel();
			newLabel->m_sLabel = labels->NodeLabel(cid);
			newLabel->m_iID = (int)newLabels.size();
			newLabel->m_pParent = rootPtr;
			rootPtr->m_apChildren.push_back(newLabel);
			newLabels.push_back(newLabel);
			
			for (int finc = 0; finc < (int)CCs[c].size(); finc++)
				face2seg[CCs[c][finc]] = newLabel->m_iID;
		}
	}
	
	ShapeLabels * copy = new ShapeLabels(labels->GetShape(), labels->GetShapeValueName());
	copy->SetAllLabels(newLabels);
	copy->SetFaceSegments(face2seg);
	return copy;
}

ShapeLabels::ShapeLabels(GeomShape * shape, const VKString & valueName)
: ShapeValue(shape, valueName, 1)
{
	
}

ShapeLabels::~ShapeLabels()
{
	
}

ShapeLabels * ShapeLabels::SafeCastToLabels()
{
	return this;
}

const ShapeLabels * ShapeLabels::SafeCastToLabels() const
{
	return this;
}

VKString ShapeLabels::GetClassName() const
{
	return "ShapeLabels";
}

void ShapeLabels::SetFaceSegments(const std::vector<int> & segIDs)
{
	int nFaces = (int)segIDs.size();
	double * faceIDs = new double[nFaces];
	for (int i=0; i<nFaces; i++)
		faceIDs[i] = (double)segIDs[i];
	InitCache(ShapeValue::SVTYPE_PerFace);
	SetFaceValues(faceIDs, nFaces);
	delete [] faceIDs;
	ClearCache();
}

void ShapeLabels::SetPntsSegments(const std::vector<int> & segIDs)
{
	int nVtx = (int)segIDs.size();
	double * perVtxVals = new double[nVtx];
	for (int i=0; i<nVtx; i++)
		perVtxVals[i] = (double)segIDs[i];
	InitCache(ShapeValue::SVTYPE_PerPointSet);
	SetPntsValues(perVtxVals, nVtx);
	delete [] perVtxVals;
	ClearCache();
}

void ShapeLabels::SetAllLabels(std::vector<SceneGraphLabel*> & labels)
{
	m_AllNodes = labels;
}

void ShapeLabels::SetAllGeometries(std::vector<SceneGraphGeometry*> & geoms)
{
	m_AllGeometries = geoms;
}

int ShapeLabels::NNodes() const
{
	return (int)m_AllNodes.size();
}

VKStringList ShapeLabels::GetAllLabels(int segID) const
{
	VKStringList allLabels;
	assert(segID>=0 && segID < (int)m_AllNodes.size());
	SceneGraphLabel * node = m_AllNodes[segID];
	while (node!=NULL)
	{
		allLabels.push_back(node->m_sLabel);
		node = node->m_pParent;
	}
	return allLabels;
}

VKStringList ShapeLabels::GetAllLabels() const
{
	VKStringList allLabels;
	for (int i=0; i<(int)m_AllNodes.size(); i++)
		allLabels.push_back(m_AllNodes[i]->m_sLabel);
	return allLabels;
}

void ShapeLabels::WriteShapeValues(FILE * file)
{
	WriteCachedShapeValues(file);
	int nNodes = (int)m_AllNodes.size();
	fwrite(&nNodes, sizeof(int), 1, file);
	// write all nodes
	for (int i=0; i<(int)m_AllNodes.size(); i++)
	{
		SceneGraphLabel * node = m_AllNodes[i];
		int nChildren = node->m_apChildren.size();
		int * childrenIDs = new int[nChildren];
		for (int cID=0; cID < nChildren; cID++)
			childrenIDs[cID] = node->m_apChildren[cID]->m_iID;
		
		int parentID = -1;
		if (node->m_pParent!=NULL)
			parentID = node->m_pParent->m_iID;
		
		VKString::WriteString(node->m_sLabel, file);
		assert(node->m_iID==i);
		fwrite(&(node->m_iID), sizeof(int), 1, file);
		fwrite(&nChildren, sizeof(int), 1, file);
		fwrite(childrenIDs, sizeof(int), nChildren, file);
		fwrite(&parentID, sizeof(int), 1, file);
		
		int nGeomIDs = node->m_aiGeometryID.size();
		fwrite(&nGeomIDs, sizeof(int), 1, file);
		if (nGeomIDs>0)
		{
			int * geomIDs = node->m_aiGeometryID.data();
			int * firstFaces = node->m_iFirstFaceInGeometry.data();
			assert((int)node->m_iFirstFaceInGeometry.size()==nGeomIDs);
			int * nFaces = node->m_iNFacesInGeometry.data();
			assert((int)node->m_iNFacesInGeometry.size()==nGeomIDs);

			fwrite(geomIDs, sizeof(int), nGeomIDs, file);
			fwrite(firstFaces, sizeof(int), nGeomIDs, file);
			fwrite(nFaces, sizeof(int), nGeomIDs, file);
		}
		delete [] childrenIDs;
	}
	
	// write all geometries
	int nGeometries = (int)m_AllGeometries.size(); 
	fwrite(&nGeometries, sizeof(int), 1, file);
	for (int i=0; i<(int)m_AllGeometries.size(); i++)
		VKString::WriteString(m_AllGeometries[i]->m_sGeometryName, file);
}

void ShapeLabels::ReadShapeValues(FILE * file)
{
	ReadCachedShapeValues(file);
	int nNodes;
	fread(&nNodes, sizeof(int), 1, file);
	
	std::vector<std::vector<int> > perNodeChildren;
	std::vector<int> perNodeParent;
	
	// read ids
	for (int i=0; i<nNodes; i++)
	{
		perNodeChildren.push_back(std::vector<int>());
		
		SceneGraphLabel * node = new SceneGraphLabel();
		int nChildren;
		node->m_sLabel = VKString::ReadString(file);
		fread(&(node->m_iID), sizeof(int), 1, file);
		assert(node->m_iID==i);
		fread(&nChildren, sizeof(int), 1, file);
		int * childrenIDs = new int[nChildren];
		fread(childrenIDs, sizeof(int), nChildren, file);
		int parentID;
		fread(&parentID, sizeof(int), 1, file);
		perNodeParent.push_back(parentID);
		for (int cID=0; cID < nChildren; cID++)
			perNodeChildren[i].push_back(childrenIDs[cID]);
		
		delete [] childrenIDs;
		m_AllNodes.push_back(node);
//		if (node->m_sLabel.startsWith("remove_Person"))
//			std::cout<<"Node: "<<node->m_sLabel.c_str()<<std::endl;
		
		// read node's geometries
		int nGeomIDs = 0;
		fread(&nGeomIDs, sizeof(int), 1, file);
		if (nGeomIDs>0) 
		{
			int * geomIDs = new int[nGeomIDs];
			int * firstFaces = new int[nGeomIDs];
			int * nFaces = new int[nGeomIDs];
			
			fread(geomIDs, sizeof(int), nGeomIDs, file);
			fread(firstFaces, sizeof(int), nGeomIDs, file);
			fread(nFaces, sizeof(int), nGeomIDs, file);
			for (int gID=0; gID<nGeomIDs; gID++)
			{
				node->m_aiGeometryID.push_back(geomIDs[gID]);
				node->m_iFirstFaceInGeometry.push_back(firstFaces[gID]);
				node->m_iNFacesInGeometry.push_back(nFaces[gID]);
//				if (node->m_sLabel.startsWith("remove_Person") && gID==0)
//				{
//					std::cout<<"   G"<<geomIDs[gID]<<": "<<firstFaces[gID]<<" - "<<nFaces[gID]<<std::endl;
//					assert(m_pMesh!=NULL);
//					for (int f=firstFaces[gID]; f<firstFaces[gID]+nFaces[gID]; f++)
//					{
//						R3Point v1 = m_pMesh->VertexPosition(m_pMesh->VertexOnFace(f, 0));
//						R3Point v2 = m_pMesh->VertexPosition(m_pMesh->VertexOnFace(f, 1));
//						R3Point v3 = m_pMesh->VertexPosition(m_pMesh->VertexOnFace(f, 2));
//						std::cout<<"      F"<<f<<": ["<<v1.X()<<" "<<v1.Y()<<" "<<v1.Z()<<"] ";
//						std::cout<<"["<<v2.X()<<" "<<v2.Y()<<" "<<v2.Z()<<"] ";
//						std::cout<<"["<<v3.X()<<" "<<v3.Y()<<" "<<v3.Z()<<"] "<<std::endl;;
//					}
//				}
			}
			delete [] geomIDs;
			delete [] firstFaces;
			delete [] nFaces;		
		}
	}
	
	// assign ptrs
	for (int i=0; i<nNodes; i++)
	{
		SceneGraphLabel * node = m_AllNodes[i];
		for (int cID=0; cID<(int)perNodeChildren[i].size(); cID++)
			node->m_apChildren.push_back(m_AllNodes[perNodeChildren[i][cID]]);
		int parentID = perNodeParent[i];
		if (parentID>=0)
			node->m_pParent = m_AllNodes[parentID];
	}
	
	// read geometries
	int nGeometries = 0;
	fread(&nGeometries, sizeof(int), 1, file);
	for (int i=0; i<nGeometries; i++)
	{
		m_AllGeometries.push_back(new SceneGraphGeometry());
		m_AllGeometries[i]->m_sGeometryName = VKString::ReadString(file);
		m_AllGeometries[i]->m_iID = i;
	}
}

void ShapeLabels::ReadFromPlyFile(const VKString & filename, ShapeLabels * optionalGTLabels)
{
	R3Mesh * plyReader = new R3Mesh();
	plyReader->ReadFile(filename.c_str());
	
	std::vector<SceneGraphLabel*> sceneGraph;
	std::vector<int> faces;
	int nmaxSeg = 0;
	for (int i=0; i<plyReader->NFaces(); i++)
	{
		int segID = plyReader->FaceSegment(plyReader->Face(i));
		faces.push_back(segID+1);
		nmaxSeg = vkMax(nmaxSeg, segID+1);
	}
	SceneGraphLabel * root = new SceneGraphLabel();
	root->m_sLabel = "SD_Root";
	root->m_iID = 0;
	root->m_pParent = NULL;
	sceneGraph.push_back(root);

	for (int i=0; i<nmaxSeg; i++)
	{
		SceneGraphLabel * newNode = new SceneGraphLabel();
		int id = (int)sceneGraph.size() - 1;
		//if (optionalGTLabels==NULL)
		newNode->m_sLabel = VKString("Seg")+VKString::number(id+1);
		newNode->m_iID = (int)sceneGraph.size();
		newNode->m_pParent = root;
		root->m_apChildren.push_back(newNode);
		sceneGraph.push_back(newNode);
	}
	SetFaceSegments(faces);
	SetAllLabels(sceneGraph);

	delete plyReader;
}

void ShapeLabels::ReadFromSegFile(const VKString & filename, const VKString & segNamesFile)
{
	VKStringList segNames;
	if (segNamesFile!="none")
	{
		std::ifstream textStream(segNamesFile.c_str());
		if (!textStream.is_open())
			std::cout<<"[ERROR] Cannot find segment names: "<<segNamesFile.c_str()<<std::endl;
		assert(textStream.is_open());
		
		while (!textStream.eof())
		{
			VKString segName;
			segName.readToken(textStream);
			if (segName!="")
				segNames.push_back(segName);
		}
		textStream.close();
	}
	
	std::ifstream textStream(filename.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] Cannot find segment ids: "<<filename.c_str()<<std::endl;
	assert(textStream.is_open());
	
	assert(m_pMesh!=NULL);
	std::vector<int> faces;

	// create root
	std::vector<SceneGraphLabel*> sceneGraph;
	SceneGraphLabel * root = new SceneGraphLabel();
	root->m_sLabel = "SD_Root";
	root->m_iID = 0;
	root->m_pParent = NULL;
	sceneGraph.push_back(root);
	
	int segID = -1;
	bool ok;	
	int minSegID = 1000000, maxSegID = -1;
	while (!textStream.eof())
	{
		VKString value;
		value.readToken(textStream);
		if (value=="")
			continue;
		segID = value.toInt(&ok);
		if (!ok)
			segID = (int)value.toDouble(&ok);
		assert(ok);
		faces.push_back(segID);
		minSegID = vkMin(minSegID, segID);
		maxSegID = vkMax(maxSegID, segID);
				
		while ((int)sceneGraph.size() <= segID)
		{
			SceneGraphLabel * newNode = new SceneGraphLabel();
			int id = (int)sceneGraph.size() - 1;
			if (id < segNames.count())
				newNode->m_sLabel = segNames[id];
			else
				newNode->m_sLabel = VKString("Seg")+VKString::number(id+1);
			newNode->m_iID = (int)sceneGraph.size();
			newNode->m_pParent = root;
			root->m_apChildren.push_back(newNode);
			sceneGraph.push_back(newNode);
		}
	}
	if (m_pMesh->NFaces()!=(int)faces.size())
		std::cout<<"[ERROR] Faces read not same as faces in mesh: "<<m_pMesh->NFaces()<<" vs "<<faces.size()<<std::endl;
	assert(m_pMesh->NFaces()==(int)faces.size());
	textStream.close();
	
	SetFaceSegments(faces);
	SetAllLabels(sceneGraph);
}

void ShapeLabels::ReadFromLabFile(const VKString & filename)
{
	std::ifstream textStream(filename.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] Cannot find labels: "<<filename.c_str()<<std::endl;
	assert(textStream.is_open());
	
	// skip two lines
	VKString tempStr;
	tempStr.readLine(textStream);
	tempStr.readLine(textStream);
	
	// create root
	std::vector<SceneGraphLabel*> sceneGraph;
	SceneGraphLabel * root = new SceneGraphLabel();
	root->m_sLabel = "SD_Root";
	root->m_iID = 0;
	root->m_pParent = NULL;
	sceneGraph.push_back(root);
	
	// go over all segments -> attach to the root
	// read segment name
	tempStr.readToken(textStream);
	VKString keyword = tempStr;
	tempStr.readLine(textStream);

	assert(m_pMesh!=NULL);
	std::vector<int> faces;
	for (int i=0; i<m_pMesh->NFaces(); i++)
		faces.push_back(0);

	while (!textStream.eof())
	{
		// add new node
		SceneGraphLabel * newNode = new SceneGraphLabel();
		newNode->m_sLabel = keyword;
		newNode->m_iID = (int)sceneGraph.size();
		newNode->m_pParent = root;
		root->m_apChildren.push_back(newNode);
		sceneGraph.push_back(newNode);

		int nFaces = 0;
		while (!textStream.eof())
		{
			// read face ID
			tempStr.readToken(textStream);
			bool validNum = false;
			int fID = tempStr.toInt(&validNum);
			if (validNum)
			{
				assert(fID < (int)faces.size() && fID>=0);
				faces[fID] = newNode->m_iID;
				nFaces++;
			}
			else
			{
				keyword = tempStr;
				tempStr.readLine(textStream);
				break;
			}
		}
	}
	
	SetFaceSegments(faces);
	SetAllLabels(sceneGraph);

	textStream.close();
}

int ShapeLabels::NGeometries() const
{
	return (int)m_AllGeometries.size();
}

const VKString & ShapeLabels::GeometryLabel(int geomID) const
{
	if (geomID >= (int)m_AllGeometries.size() || geomID<0)
	{
		std::cout<<"[ERROR] Cannot obtain label: "<<geomID<<" / "<<m_AllGeometries.size()<<std::endl;
		assert(geomID < (int)m_AllGeometries.size());
	}
	return m_AllGeometries[geomID]->m_sGeometryName;
}


int ShapeLabels::RootNode() const
{
	return 0;
}

void ShapeLabels::InitFlatGraph(int nChildren)
{
	AddNode(-1, "root");
	for (int i=0; i<nChildren; i++)
		AddNode(0, VKString("Seg_")+VKString::number(i));
}

void ShapeLabels::SetNodeLabel(int nodeID, const VKString & nodeLabel)
{
	assert(nodeID>=0 && nodeID < (int)m_AllNodes.size());
	m_AllNodes[nodeID]->m_sLabel = nodeLabel;
}

const VKString & ShapeLabels::NodeLabel(int nodeID) const
{
	assert(nodeID>=0 && nodeID < (int)m_AllNodes.size());
	return m_AllNodes[nodeID]->m_sLabel;
}

int ShapeLabels::NodeNChildren(int nodeID) const
{
	assert(nodeID>=0 && nodeID < (int)m_AllNodes.size());
	return (int)m_AllNodes[nodeID]->m_apChildren.size();
}

int ShapeLabels::NodeChild(int nodeID, int cID) const
{
	assert(nodeID>=0 && nodeID < (int)m_AllNodes.size());
	assert(cID>=0 && cID < (int)m_AllNodes[nodeID]->m_apChildren.size());
	return m_AllNodes[nodeID]->m_apChildren[cID]->m_iID;
}

int ShapeLabels::NodeParent(int nodeID) const
{	
	assert(nodeID>=0 && nodeID < (int)m_AllNodes.size());
	if (m_AllNodes[nodeID]->m_pParent!=NULL)
		return m_AllNodes[nodeID]->m_pParent->m_iID;
	return -1;
}

const std::vector<int> & ShapeLabels::NodeFacesAtNode(int nodeID) const
{
	assert(m_pMesh!=NULL);
	assert(nodeID>=0 && nodeID < (int)m_AllNodes.size());
	if (!m_AllNodes[nodeID]->m_bCachedIDs)	
		((ShapeLabels*)this)->CacheAllLeafIDs();
	assert(m_AllNodes[nodeID]->m_bCachedIDs);
	return m_AllNodes[nodeID]->m_CachedIDs;
}

void ShapeLabels::NodeAllFaces(int nodeID, std::vector<int> & faceIDs) const
{
	assert(m_pMesh!=NULL);
	std::vector<SceneGraphLabel *> nodesToExplore;
	nodesToExplore.push_back(m_AllNodes[nodeID]);
	
	while (nodesToExplore.size() > 0)
	{
		SceneGraphLabel * node = nodesToExplore[0];
		nodesToExplore.erase(nodesToExplore.begin());
		
		// cache if needed
		if (!node->m_bCachedIDs)	
			((ShapeLabels*)this)->CacheAllLeafIDs();
		assert(node->m_bCachedIDs);
		
		// add faces
		for (int i=0; i<(int)node->m_CachedIDs.size(); i++)
			faceIDs.push_back(node->m_CachedIDs[i]);
		
		// recurse
		for (int i=0; i<node->m_apChildren.size(); i++)
		{
			assert(node!=node->m_apChildren[i]);
			nodesToExplore.push_back(node->m_apChildren[i]);
		}
	}
}

void ShapeLabels::NodeAllPoints(int nodeID, std::vector<int> & pointIDs) const
{
	assert(m_pPnts!=NULL);
	std::vector<SceneGraphLabel *> nodesToExplore;
	nodesToExplore.push_back(m_AllNodes[nodeID]);
	
	while (nodesToExplore.size() > 0)
	{
		SceneGraphLabel * node = nodesToExplore[0];
		nodesToExplore.erase(nodesToExplore.begin());
		
		// cache if needed
		if (!node->m_bCachedIDs)	
			((ShapeLabels*)this)->CacheAllLeafIDs();
		assert(node->m_bCachedIDs);
		
		// add ids
		for (int i=0; i<(int)node->m_CachedIDs.size(); i++)
			pointIDs.push_back(node->m_CachedIDs[i]);
		
		// recurse
		for (int i=0; i<node->m_apChildren.size(); i++)
		{
			assert(node!=node->m_apChildren[i]);
			nodesToExplore.push_back(node->m_apChildren[i]);
		}
	}
}

const std::vector<int> & ShapeLabels::NodePointsAtNode(int nodeID) const
{
	assert(m_pPnts!=NULL);
	assert(nodeID>=0 && nodeID < (int)m_AllNodes.size());
	if (!m_AllNodes[nodeID]->m_bCachedIDs)	
		((ShapeLabels*)this)->CacheAllLeafIDs();
	assert(m_AllNodes[nodeID]->m_bCachedIDs);
	return m_AllNodes[nodeID]->m_CachedIDs;
}


void ShapeLabels::CacheAllLeafIDs()
{
	if (m_pMesh!=NULL)
	{
		for (int i=0; i<(int)m_AllNodes.size(); i++)
			m_AllNodes[i]->m_CachedIDs.clear();

		for (int fID=0; fID<m_pMesh->NFaces(); fID++)
		{
			int segID = (int)Value(ShapePoint(m_pMesh, fID, 1./.3, 1./3.), 0);
			if (!(segID>=0 && segID < (int)m_AllNodes.size()))
			{
				std::cout<<"[ERROR] Invalid segID="<<segID<<std::endl;
				assert(segID>=0 && segID < (int)m_AllNodes.size());
			}
			m_AllNodes[segID]->m_CachedIDs.push_back(fID);
		}
		for (int i=0; i<(int)m_AllNodes.size(); i++)
			m_AllNodes[i]->m_bCachedIDs = true;
	}
	else if (m_pPnts!=NULL)
	{
		for (int i=0; i<(int)m_AllNodes.size(); i++)
			m_AllNodes[i]->m_CachedIDs.clear();

		for (int pID=0; pID<m_pPnts->NumPoints(); pID++)
		{
			int segID = (int)Value(m_pPnts->GetPoint(pID), 0);
			if (!(segID>=0 && segID < (int)m_AllNodes.size()))
			{
				std::cout<<"[ERROR] Invalid segID="<<segID<<std::endl;
				assert(segID>=0 && segID < (int)m_AllNodes.size());
			}
			m_AllNodes[segID]->m_CachedIDs.push_back(pID);
		}
		
		for (int i=0; i<(int)m_AllNodes.size(); i++)
			m_AllNodes[i]->m_bCachedIDs = true;
	}
	else
		assert(false);
}

void ShapeLabels::ClearCache()
{
	for (int i=0; i<(int)m_AllNodes.size(); i++)
	{
		m_AllNodes[i]->m_bCachedIDs = false;
		m_AllNodes[i]->m_CachedIDs.clear();
	}
}


int ShapeLabels::MaxDepth(bool writeHierarchy) const
{
	VKString depthInfo = writeHierarchy ? "Computing Max Depth:\n" : "";
	int maxDepth = 0;
	std::vector<SceneGraphLabel *> nodesToExplore;
	std::vector<int> nodesDepth;
	if (m_AllNodes.size()==0)
		return 0;
	nodesToExplore.push_back(m_AllNodes[RootNode()]);
	nodesDepth.push_back(0);
	
	while (nodesToExplore.size() > 0)
	{
		SceneGraphLabel * node = nodesToExplore[0];
		assert(node!=NULL);
		int depth = nodesDepth[0];
		maxDepth = vkMax(maxDepth, depth);
		nodesToExplore.erase(nodesToExplore.begin());
		nodesDepth.erase(nodesDepth.begin());
		
		if (writeHierarchy)
		{
			for (int d=0; d<depth; d++)
				depthInfo+="  ";
			depthInfo += VKString::number(node->m_iID)+": "+node->m_sLabel+"\n";
		}
		
		// recurse
		for (int i=0; i<node->m_apChildren.size(); i++)
		{
			assert(node!=node->m_apChildren[i]);
			if (writeHierarchy)
			{
				nodesToExplore.insert(nodesToExplore.begin(), node->m_apChildren[i]);
				nodesDepth.insert(nodesDepth.begin(), depth+1);
			}
			else
			{
				nodesToExplore.push_back(node->m_apChildren[i]);
				nodesDepth.push_back(depth+1);
			}
		}
	}
	if (writeHierarchy)
		std::cout<<depthInfo.c_str()<<std::endl;
	return maxDepth;
}

int ShapeLabels::AddNode(int parentID, const VKString & nodeLabel)
{
	assert((parentID>=0 || m_AllNodes.size()==0) && parentID<(int)m_AllNodes.size());
	
	SceneGraphLabel * newNode = new SceneGraphLabel();
	if (parentID>=0)
		m_AllNodes[parentID]->m_apChildren.push_back(newNode);

	newNode->m_sLabel = nodeLabel;
	newNode->m_iID = (int)m_AllNodes.size();
	if (parentID>=0)
		newNode->m_pParent = m_AllNodes[parentID];
	else
		newNode->m_pParent = NULL;
	m_AllNodes.push_back(newNode);
	return newNode->m_iID;
}

int ShapeLabels::NodeDepth(int nodeID) const
{
	int depth = 0;
	while (NodeParent(nodeID)>-1)
	{
		nodeID = NodeParent(nodeID);
		depth++;
	}
	return depth;
}

int ShapeLabels::NumNodeGeometries(int nodeID) const
{
	assert(nodeID>=0 && nodeID < (int)m_AllNodes.size());
	return (int)m_AllNodes[nodeID]->m_aiGeometryID.size();
}

int ShapeLabels::GetGeometryID(int nodeID, int geomID, int * f0, int *nf) const
{
	assert(nodeID>=0 && nodeID < (int)m_AllNodes.size());
	assert(geomID>=0 && geomID < (int)m_AllNodes[nodeID]->m_aiGeometryID.size());
	if (f0!=NULL)
		*f0 = m_AllNodes[nodeID]->m_iFirstFaceInGeometry[geomID];
	if (nf!=NULL)
		*nf = m_AllNodes[nodeID]->m_iNFacesInGeometry[geomID];
	
	return m_AllNodes[nodeID]->m_aiGeometryID[geomID];
}

Model3D * ShapeLabels::ExtractNode(int nodeID)
{
	//TODO: copy materials info
	// modify geometry, node labels (all children), materials...
	Model3D * m = new Model3D("none");
	assert(m_pMesh!=NULL);
	ShapeMesh * newMesh = NULL;
	if (m_pMesh->SafeCastToManifold()!=NULL)
		newMesh = new ShapeManifoldMesh(GeomShape::SHAPE_TYPE_MANIFOLD, "none", 
										Model3D::m_sMainGeomName, m);
	else if (m_pMesh->SafeCastToTriangleSoup()!=NULL)
		newMesh = new ShapeTriangleSoup("none", Model3D::m_sMainGeomName, m);
	else
		assert(false);
	ShapeLabels * newLabels = new ShapeLabels(newMesh, GetShapeValueName());
	newMesh->AddShapeVal(GetShapeValueName(), newLabels);
	m->AddGeomShape(Model3D::m_sMainGeomName, newMesh);
	
	std::vector<SceneGraphLabel*> subgraphCopy;
	std::vector<bool> deleteNodes;
	for (int i=0; i<(int)m_AllNodes.size(); i++)
		deleteNodes.push_back(false);
	
	// traverse children: 
	//		record geometry to delete (faces)
	//		reconstruct node structure
	//		create geometry copy (to return)
	// initialize copy data holders
	std::vector<SceneGraphLabel *> copiesOfNodes;
	std::vector<int> copyNodeToOrigNodeID;
	std::vector<int> origNodeIDToNewNodeID;
	std::vector<int> origNodeIDToRemainedNodeID;
	
	std::vector<SceneGraphGeometry *> copiesOfGeometries;
	std::vector<int> origGeomIDToNewGeomID;
	std::vector<int> origVertIDToNewVertID;
	for (int g=0; g<(int)m_AllGeometries.size(); g++)
		origGeomIDToNewGeomID.push_back(-1);
	for (int v=0; v<m_pMesh->NVertices(); v++)
		origVertIDToNewVertID.push_back(-1);
	
	// geometry counter (how many nodes reference a geometry)
	std::vector<int> geometryCounter;
	for (int i=0; i<(int)m_AllGeometries.size(); i++)
		geometryCounter.push_back(0);
	for (int n=0; n<(int)m_AllNodes.size(); n++)
	{
		// clear cached node cache
		m_AllNodes[n]->m_bCachedIDs = false;
		m_AllNodes[n]->m_CachedIDs.clear();
		origNodeIDToNewNodeID.push_back(-1);
		origNodeIDToRemainedNodeID.push_back(-1);

		// for each geometry - count references
		for (int g=0; g<(int)m_AllNodes[n]->m_aiGeometryID.size(); g++)
			geometryCounter[m_AllNodes[n]->m_aiGeometryID[g]]++;
	}	
	
	// initialize traversal
	std::vector<SceneGraphLabel *> nodesToDelete;
	nodesToDelete.push_back(m_AllNodes[nodeID]);
	while(nodesToDelete.size() > 0)
	{
		SceneGraphLabel * node = nodesToDelete[0];
		nodesToDelete.erase(nodesToDelete.begin());
		deleteNodes[node->m_iID] = true;
		
		SceneGraphLabel * newNode = new SceneGraphLabel();
		newNode->m_sLabel = node->m_sLabel;
		newNode->m_iID = (int)copiesOfNodes.size();
		
		// add node, and record how IDs map
		copiesOfNodes.push_back(newNode);
		origNodeIDToNewNodeID[node->m_iID] = newNode->m_iID;
		copyNodeToOrigNodeID.push_back(node->m_iID);

		// update geometry references counter
		for (int g=0; g<(int)node->m_aiGeometryID.size(); g++)
			geometryCounter[node->m_aiGeometryID[g]]--;
				
		// add geometry data...
		for (int g=0; g<(int)node->m_aiGeometryID.size(); g++)
		{
			int origGID = node->m_aiGeometryID[g];
			int &newGID = origGeomIDToNewGeomID[origGID];
			
			// create geometry (descriptor)
			if (newGID==-1)		
			{
				newGID = (int)copiesOfGeometries.size();
				SceneGraphGeometry * newGeom = new SceneGraphGeometry();
				newGeom->m_sGeometryName = m_AllGeometries[origGID]->m_sGeometryName;
				newGeom->m_iID = newGID;
				copiesOfGeometries.push_back(newGeom);
			}

			newNode->m_aiGeometryID.push_back(newGID);
			newNode->m_iFirstFaceInGeometry.push_back(newMesh->NFaces());
			newNode->m_iNFacesInGeometry.push_back(node->m_iNFacesInGeometry[g]);

			// copy geometry (faces / vertices)
			for (int fID=0; fID<node->m_iNFacesInGeometry[g]; fID++)
			{
				int f = node->m_iFirstFaceInGeometry[g] + fID;
				int newVToUse[3];
				// create required vertices
				for (int vID=0; vID<3; vID++)
				{
					int v = m_pMesh->VertexOnFace(f, vID);
					int & newV = origVertIDToNewVertID[v];
					if (newV == -1)	// create vertex
					{
						R3Point vp = m_pMesh->VertexPosition(v);
						R3Vector vn = m_pMesh->VertexNormal(v);
						newV = newMesh->AddVertex(vp.X(), vp.Y(), vp.Z(), vn.X(), vn.Y(), vn.Z());
					}
					newVToUse[vID] = newV;
				}
				// create face
				newMesh->AddFace(newVToUse[0], newVToUse[1], newVToUse[2]);
			}
		}
				
		// recurse
		for (int i=0; i<node->m_apChildren.size(); i++)
			nodesToDelete.push_back(node->m_apChildren[i]);
	}
	
	// update ancestry info for copied nodes: m_apChildren, 
	for (int i=0; i<(int)copiesOfNodes.size(); i++)
	{
		SceneGraphLabel * newNode = copiesOfNodes[i];
		int newID = newNode->m_iID;
		assert(newID == i);

		assert(newID < (int)copyNodeToOrigNodeID.size());
		int origNodeID = copyNodeToOrigNodeID[newID];
		assert(origNodeID>=0 && origNodeID < NNodes());

		if (i==0)
			newNode->m_pParent = NULL;
		else
		{
			int origParentID = NodeParent(origNodeID);
			assert(origParentID>=0 && origParentID < (int)origNodeIDToNewNodeID.size());
			int newParentID = origNodeIDToNewNodeID[origParentID];
			assert(newParentID>=0 && newParentID < (int)copiesOfNodes.size());
			newNode->m_pParent = copiesOfNodes[newParentID];
		}
		
		for (int c=0; c<NodeNChildren(origNodeID); c++)
		{
			int origChildID = NodeChild(origNodeID, c);
			assert(origChildID>=0 && origChildID < (int)origNodeIDToNewNodeID.size());
			int newChildID = origNodeIDToNewNodeID[origChildID];
			assert(newChildID>=0 && newChildID < (int)copiesOfNodes.size());
			newNode->m_apChildren.push_back(copiesOfNodes[newChildID]);
		}
	}
	
	// update new mesh values:
		// copy labels data (TODO: face segment IDs)
	//newLabels->SetFaceSegments(std::vector<int> & segIDs);	// todo...
	newLabels->SetAllLabels(copiesOfNodes);
	newLabels->SetAllGeometries(copiesOfGeometries);
		// copy material values (TODO!)

	// update geometry IDs
	std::vector<int> oldGeomID2NewGeomID;
	for (int i=0; i<(int)m_AllGeometries.size(); i++)
		oldGeomID2NewGeomID.push_back(-1);
	for (int i=0; i<(int)geometryCounter.size(); i++)
	{
		if (geometryCounter[i]==0)
		{
			m_AllGeometries.erase(m_AllGeometries.begin()+i);
			geometryCounter.erase(geometryCounter.begin()+i);
			i--;
		}
		else
		{
			int oldID = m_AllGeometries[i]->m_iID;
			assert(oldID>=0 && oldID < (int)oldGeomID2NewGeomID.size());
			oldGeomID2NewGeomID[oldID] = i;
			m_AllGeometries[i]->m_iID = i;
		}
	}
	
	// delete nodes, update node IDs, record which faces need to be deleted
	
	SceneGraphLabel * parent = m_AllNodes[nodeID]->m_pParent;
	if (parent!=NULL)
	{
		bool detected = false;
		for (int c=0; c<parent->m_apChildren.size(); c++)
			if (parent->m_apChildren[c]==m_AllNodes[nodeID])
			{
				detected = true;
				parent->m_apChildren.erase(parent->m_apChildren.begin()+c);
				break;
			}
		assert(detected);
	}
	
	std::vector<Sortable> facesToDelete;	// pairs: faceFrom...faceTo
	assert(deleteNodes.size()==m_AllNodes.size());
	for (int i=0; i<(int)m_AllNodes.size(); i++)
	{
		if (deleteNodes[i])
		{
			for (int g=0; g<(int)m_AllNodes[i]->m_aiGeometryID.size(); g++)
			{
				int f1 = m_AllNodes[i]->m_iFirstFaceInGeometry[g];
				int nf = m_AllNodes[i]->m_iNFacesInGeometry[g];
				if (nf==0)
					continue;
				int nToDel = (int)facesToDelete.size();
				if (nToDel==0 || facesToDelete[nToDel-1].id!=f1)
					facesToDelete.push_back(Sortable(-(double)f1, NULL, f1+nf-1));
				else
					facesToDelete[nToDel-1].id = f1+nf-1;
			}
			delete m_AllNodes[i];
			m_AllNodes.erase(m_AllNodes.begin()+i);
			deleteNodes.erase(deleteNodes.begin()+i);
			i--;
		}
		else
		{
			origNodeIDToRemainedNodeID[m_AllNodes[i]->m_iID] = i;
			m_AllNodes[i]->m_iID = i;
			std::vector<int> newIDs;
			for (int g=0; g<(int)m_AllNodes[i]->m_aiGeometryID.size(); g++)
				newIDs.push_back(oldGeomID2NewGeomID[m_AllNodes[i]->m_aiGeometryID[g]]);
			m_AllNodes[i]->m_aiGeometryID.clear();
			m_AllNodes[i]->m_aiGeometryID = newIDs;
		}
	}
	
	// delete faces
	std::sort(facesToDelete.begin(), facesToDelete.end());
	for (int i=0; i<facesToDelete.size(); i++)
	{
		int f1 = -(int)facesToDelete[i].value;
		int f2 = facesToDelete[i].id;
		int ndel = (f2-f1+1);
		
		m_pMesh->RemoveFaces(f1, f2);
		for (int n=0; n<(int)m_AllNodes.size(); n++)
		for (int g=0; g<(int)m_AllNodes[n]->m_iFirstFaceInGeometry.size(); g++)
		{
			if (m_AllNodes[n]->m_iFirstFaceInGeometry[g]>f2)
				m_AllNodes[n]->m_iFirstFaceInGeometry[g] -= ndel;
			else
			{
				if (m_AllNodes[n]->m_iNFacesInGeometry[g]==0)
					m_AllNodes[n]->m_iFirstFaceInGeometry[g] = -1;
				assert(m_AllNodes[n]->m_iFirstFaceInGeometry[g]<f1);
			}
		}
	}
	
	// update segment IDs
	assert(m_iNDims==1 && m_iNValues==m_pMesh->NFaces() 
		   && m_CacheType==SVTYPE_PerFace);
	for (int i=0; i<m_iNValues; i++)
	{
		int segID = (int)m_aValues[i];
		segID = origNodeIDToRemainedNodeID[segID];
		assert(segID!=-1);
		m_aValues[i] = (double)segID;
	}

	return m;
}

int ShapeLabels::GetLeafID(int fid) const
{
	return (int)Value(ShapePoint(m_pMesh, fid, .3, .3), 0);
}

