/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "VKString.h"

#include "Model3D.h"
#include "ShapeManifoldMesh.h"

ShapeManifoldMesh::ShapeManifoldMesh(GeomShape::GeomShapeTypes shapeTp,
									 const VKString & geomFile, 
									 const VKString & geomName,
									 Model3D * model)
: ShapeMesh(shapeTp, geomName, model)
{
	m_pMesh = NULL;
	if (geomFile!="none")
		ReadGeometryFile(geomFile);
}

ShapeManifoldMesh::~ShapeManifoldMesh()
{
	if (m_pMesh!=NULL)
		delete m_pMesh;
	
}

ShapeManifoldMesh * ShapeManifoldMesh::SafeCastToManifold()
{
	return this;
}

int ShapeManifoldMesh::AddVertex(double x, double y, double z,
								 double nx, double ny, double nz)
{
	R3MeshVertex * v = GetMesh()->CreateVertex(R3Point(x, y, z));
	if (nx !=0 || ny != 0 || nz != 0)
		GetMesh()->SetVertexNormal(v, R3Vector(nx, ny, nz));
	return GetMesh()->VertexID(v);
}

int ShapeManifoldMesh::VertexValence(int vID) const
{
	return m_pMesh->VertexValence(m_pMesh->Vertex(vID));
}

int ShapeManifoldMesh::NEdges() const
{
	return m_pMesh->NEdges();
}

std::vector<int> ShapeManifoldMesh::AdjacentVertices(int vID) const
{
	std::vector<int> neighbors;
	R3MeshVertex * v = m_pMesh->Vertex(vID);
	for (int i=0; i<m_pMesh->VertexValence(v); i++)
	{
		R3MeshVertex * vadj = m_pMesh->VertexAcrossEdge(m_pMesh->EdgeOnVertex(v, i), v);
		neighbors.push_back(m_pMesh->VertexID(vadj));
	}
	return neighbors;
}

int ShapeManifoldMesh::FaceOnVertex(int vID) const
{
	assert(vID < NVertices());
	const R3MeshFace * face = m_pMesh->FaceOnVertex(m_pMesh->Vertex(vID));
	assert(face!=NULL);
	return m_pMesh->FaceID(face);
}

int ShapeManifoldMesh::NAdjacentFaces(int vID) const
{
	return m_pMesh->VertexValence(m_pMesh->Vertex(vID));
}

int ShapeManifoldMesh::FaceOnVertex(int vID, int nFace) const
{
	const R3MeshEdge * edge = m_pMesh->EdgeOnVertex(m_pMesh->Vertex(vID), nFace);
	assert(edge!=NULL);
	const R3MeshFace * face = m_pMesh->FaceOnVertex(m_pMesh->Vertex(vID), edge);
	if (face==NULL)	// non-manifold
		return -1;
	return m_pMesh->FaceID(face);
}

int ShapeManifoldMesh::AddFace(int v1, int v2, int v3)
{
	// todo: might need to change orientation
	R3MeshFace * face = GetMesh()->CreateFace(m_pMesh->Vertex(v1), m_pMesh->Vertex(v2), m_pMesh->Vertex(v3)); 
	assert(face!=NULL);
	return GetMesh()->FaceID(face);
}

void ShapeManifoldMesh::RemoveFace(int f)
{
	assert(false);
}

void ShapeManifoldMesh::RemoveVertex(int v)
{
	assert(false);
}

void ShapeManifoldMesh::WriteGeometryFile(const VKString & geomFile) const
{
	assert(m_pMesh!=NULL);
	m_pMesh->WriteFile(geomFile.c_str());
}

void ShapeManifoldMesh::ReadGeometryFile(const VKString & geomFile)
{
	//std::cout<<"Reading manifold: "<<geomFile.c_str()<<std::endl;
	assert(m_pMesh==NULL);
	m_pMesh = new R3Mesh();
	
	bool fileWasRead = false;
	
	if (geomFile.endsWith(".off"))	// try reading off binary
	{
		FILE * fp = fopen(geomFile.c_str(), "rb");
		assert(fp!=NULL);
		char header[9];
		header[8] = 0;
		fread(header, sizeof(char), 8, fp);
		
		if (strcmp(header, "OFF1    ")==0 || strcmp(header, "OFF2    ")==0)
		{
			fileWasRead = true;
			bool single = false;
			if (strcmp(header, "OFF1    ")==0)
				single = true;
			else if (strcmp(header, "OFF2    ")==0)
				single = false;
			int nvertices, nfaces, nedges;
			fread(&nvertices, sizeof(int), 1, fp);
			fread(&nfaces, sizeof(int), 1, fp);
			fread(&nedges, sizeof(int), 1, fp);
			
			if (single)
			{
				float * vtxData = new float[3*nvertices];
				fread(vtxData, sizeof(float), 3*nvertices, fp);
				for (int i=0; i<nvertices; i++)
					AddVertex((double)vtxData[3*i+0], (double)vtxData[3*i+1], (double)vtxData[3*i+2]);
				delete [] vtxData;
			}
			else
			{
				double * vtxData = new double[3*nvertices];
				fread(vtxData, sizeof(double), 3*nvertices, fp);
				for (int i=0; i<nvertices; i++)
					AddVertex(vtxData[3*i+0], vtxData[3*i+1], vtxData[3*i+2]);
				delete [] vtxData;
			}
			
			int nIds = 0;
			int vtxIds[3];
			for (int i=0; i<nfaces; i++)
			{
				fread(&nIds, sizeof(int), 1, fp);
				assert(nIds==3);	// only triangles for now
				fread(vtxIds, sizeof(int), 3, fp);
				AddFace(vtxIds[0], vtxIds[1], vtxIds[2]);
			}
		}
		fclose(fp);
	}
	
	if (!fileWasRead)
		m_pMesh->ReadFile(geomFile.c_str());
}

const R3Mesh * ShapeManifoldMesh::GetMesh() const
{
	return m_pMesh;
}

R3Mesh * ShapeManifoldMesh::GetMesh()
{
	return m_pMesh;
}

void ShapeManifoldMesh::SetMesh(R3Mesh * mesh)
{
	assert(m_pMesh==NULL);
	m_pMesh = mesh;
}

// vertices
int ShapeManifoldMesh::NVertices() const
{
	return GetMesh()->NVertices();
}

void ShapeManifoldMesh::SetVertexPosition(int vID, const R3Point & newPos )
{
	delete m_pSearchTree;
	delete m_pSearchTreeMesh;
	m_pSearchTree = NULL;
	m_pSearchTreeMesh = NULL;
	m_SearchTreeToTriangleMap.clear();
	m_SearchTreeToVertexMap.clear();
	
	m_pMesh->SetVertexPosition(m_pMesh->Vertex(vID), newPos);
}

R3Point ShapeManifoldMesh::VertexPosition(int vID) const
{
	const R3Mesh * m = GetMesh();
	return m->VertexPosition(m->Vertex(vID));
}
R3Vector ShapeManifoldMesh::VertexNormal(int vID) const
{
	const R3Mesh * m = GetMesh();
	return m->VertexNormal(m->Vertex(vID));	
}

// faces
int ShapeManifoldMesh::NFaces() const
{
	return GetMesh()->NFaces();
}

int ShapeManifoldMesh::VertexOnFace(int fID, int vID) const
{
	const R3Mesh * m = GetMesh();
	return m->VertexID(m->VertexOnFace(m->Face(fID), vID));
}

R3Point ShapeManifoldMesh::FaceCentroid(int fID) const
{
	const R3Mesh * m = GetMesh();
	return m->FaceCentroid(m->Face(fID));
}

R3Point ShapeManifoldMesh::FacePoint(int fID, double b1, double b2) const
{
	const R3Mesh * m = GetMesh();
	double b[3];
	b[0] = b1;
	b[1] = b2;
	b[2] = 1. - b1 - b2;
	return m->FacePoint(m->Face(fID), b);
}

R3Point ShapeManifoldMesh::FaceBarycentric(int fID, const R3Point & p) const
{
	const R3Mesh * m = GetMesh();
	return m->FaceBarycentric(m->Face(fID), p);
}


R3Vector ShapeManifoldMesh::FaceNormal(int fID) const
{
	const R3Mesh * m = GetMesh();
	return m->FaceNormal(m->Face(fID));
}

double ShapeManifoldMesh::FaceArea(int fID) const
{
	const R3Mesh * m = GetMesh();
	return m->FaceArea(m->Face(fID));
}

void ShapeManifoldMesh::AdjacentFaces(int fID, std::vector<int> & fids) const
{
	const R3Mesh * m = GetMesh();
	
	R3MeshFace * f = m->Face(fID);
	
	for (int i=0; i<3; i++)
	{
		R3MeshFace * fAdj = m->FaceOnFace(f, i);
		if (fAdj!=NULL && fAdj != f)
			fids.push_back(m->FaceID(fAdj));
	}
}

void ShapeManifoldMesh::RearrangeAxes(char newOrder[3], int newSigns[3])
{
	R3Mesh * m = GetMesh();
	for (int i=0; i<m->NVertices(); i++)
	{
		R3Point oldPnt = m->VertexPosition(m->Vertex(i));
		R3Point newPnt;
		for (int d=0; d<3; d++)
		{
			int dim = ((int)newOrder[d]-(int)'x');
			assert(dim>=0 && dim<3);
			newPnt[d] = newSigns[d] * oldPnt[dim];
		}
		m->SetVertexPosition(m->Vertex(i), newPnt);
	}
}


R3MeshSearchTree * ShapeManifoldMesh::CreateSearchTree(R3Mesh ** mesh,
													   std::vector<int> & triangleMap,
													   std::vector<int> & vertexMap) const 
{
	ShapeManifoldMesh * ncThis = (ShapeManifoldMesh*)this;
	*mesh = ncThis->GetMesh();
	return new R3MeshSearchTree(*mesh);
}


ShapePoint ShapeManifoldMesh::NearestPointOnFlatMesh(const R2Point & p) const
{
//	// NOTE: assumes delaunay triangulation (of the plane)		
	R3Point p3d(p.X(), p.Y(), 0);
	return GetShapePointOnMesh(p3d);
}
