/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ShapeMaterials.h"
#include "ShapeMesh.h"
#include "FileManagement.h"

ColorOrTexture::ColorOrTexture()
{
	m_Type = BRDF_Invalid;
	m_iTextureID = -1;
	m_Color.SetRed(0.);
	m_Color.SetGreen(0.);
	m_Color.SetBlue(0.);
}
bool ColorOrTexture::IsTexture()
{
	return m_Type==BRDF_Texture;
}
bool ColorOrTexture::IsColor()
{
	return m_Type==BRDF_Color;
}

void ColorOrTexture::WriteColorOrTexture(FILE * file)
{
	double r = m_Color.R();
	double g = m_Color.G();
	double b = m_Color.B();
	fwrite(&m_Type, sizeof(Type), 1, file);
	fwrite(&m_iTextureID, sizeof(int), 1, file);
	fwrite(&r, sizeof(double), 1, file);
	fwrite(&g, sizeof(double), 1, file);
	fwrite(&b, sizeof(double), 1, file);
}

void ColorOrTexture::ReadColorOrTexture(FILE * file)
{
	double r, g, b;
	fread(&m_Type, sizeof(Type), 1, file);
	fread(&m_iTextureID, sizeof(int), 1, file);
	fread(&r, sizeof(double), 1, file);
	fread(&g, sizeof(double), 1, file);
	fread(&b, sizeof(double), 1, file);
	m_Color.SetRed(r);
	m_Color.SetGreen(g);
	m_Color.SetBlue(b);
}



MaterialBRDF::MaterialBRDF()
{
	m_iMaterialID = -1;
	
	m_fShininess = 0;
	m_Reflectivity = 0;
	m_IndexOfRefraction = 0;
	
	m_iMaterialID = 0;
}

MaterialBRDF::~MaterialBRDF()
{
}


ShapeMaterials::ShapeMaterials(GeomShape * shape, const VKString & valueName)
: ShapeValue(shape, valueName, 3)
{
	m_FacesPerMaterial = NULL;
}

ShapeMaterials::~ShapeMaterials()
{
	for (int i=0; i<(int)m_Materials.size(); i++)
		delete m_Materials[i];
	
	for (int i=0; i<(int)m_Textures.size(); i++)
		delete m_Textures[i];
	
	if (m_FacesPerMaterial!=NULL)
		delete [] m_FacesPerMaterial;
}

VKString ShapeMaterials::GetClassName() const
{
	return "ShapeMaterials";
}

ShapeMaterials * ShapeMaterials::SafeCastToMaterials()
{
	return this;
}

const ShapeMaterials * ShapeMaterials::SafeCastToMaterials() const
{
	return this;
}

int ShapeMaterials::AddMaterial(MaterialBRDF * material)
{
	m_Materials.push_back(material);
	return (int)m_Materials.size()-1;
}

int ShapeMaterials::AddTextureImg(R2Image * img)
{
	m_Textures.push_back(img);
	return (int)m_Textures.size()-1;	
}


void ShapeMaterials::FaceMaterial(int fID, 
								  double * u1, double * v1, 
								  double * u2, double * v2, 
								  double * u3, double * v3, 
								  int * materialID)
{
	double vals1[3];
	double vals2[3];
	double vals3[3];
	GetCachedFaceIndexedValues(fID, 0, vals1);
	GetCachedFaceIndexedValues(fID, 1, vals2);
	GetCachedFaceIndexedValues(fID, 2, vals3);
	
	assert(vals1[2]==vals2[2] && vals1[2]==vals3[2]);
	*u1 = vals1[0];
	*v1 = vals1[1];
	*u2 = vals2[0];
	*v2 = vals2[1];
	*u3 = vals3[0];
	*v3 = vals3[1];
	
	*materialID = (int)vals1[2];
}

void ShapeMaterials::SetFaceMaterials(std::vector<R2Point> & uvCoords,
									  std::vector<int> & perFaceMaterialIDs,
									  std::vector<int> & perFaceUVIds)
{
	// note: i don't expect same uv coords to overlap with multiple different material ids
	
	assert(3*perFaceMaterialIDs.size()==perFaceUVIds.size());
	std::vector<std::vector<int> > perUVMaterialID;
	std::vector<std::vector<int> > newPerUVIndices;
	for (int i=0; i<(int)uvCoords.size(); i++)
	{
		perUVMaterialID.push_back(std::vector<int>());
		newPerUVIndices.push_back(std::vector<int>());
	}
	int noUVID = (int)newPerUVIndices.size();
	perUVMaterialID.push_back(std::vector<int>());
	newPerUVIndices.push_back(std::vector<int>());
	
	int nFaceIndices = (int)perFaceMaterialIDs.size()*3;
	int * indices = new int[nFaceIndices];
	std::vector<double> vals;
	uvCoords.push_back(R2Point(-1, -1));

	
	for (int fID=0; fID<(int)perFaceMaterialIDs.size(); fID++)
	{
		int mID = perFaceMaterialIDs[fID];
		for (int lID=0; lID<3; lID++)
		{
			int uvID = perFaceUVIds[3*fID + lID];
			if (uvID < 0)
				uvID = noUVID;
			assert(uvID>=0 && uvID < (int)perUVMaterialID.size());
			int foundUVID = -1;
			for (int i=0; i<(int)perUVMaterialID[uvID].size() && foundUVID==-1; i++)
				if (perUVMaterialID[uvID][i]==mID)
					foundUVID = newPerUVIndices[uvID][i];
			if (foundUVID==-1)
			{
				perUVMaterialID[uvID].push_back(mID);
				assert(uvID < (int)uvCoords.size());
				foundUVID = (int)vals.size()/3;
				newPerUVIndices[uvID].push_back(foundUVID);
				vals.push_back(uvCoords[uvID].X());
				vals.push_back(uvCoords[uvID].Y());
				vals.push_back((double)mID);
			}
			indices[3*fID + lID] = foundUVID;
		}
	}
	
	int nVals = (int)vals.size();
	double * valsDbl = new double[nVals]; 
	for (int i=0; i<nVals; i++)
		valsDbl[i] = vals[i];
		
	InitCache(ShapeValue::SVTYPE_PerIndexedFaces, nVals/3);
	SetIndexedFaceValues(indices, nFaceIndices, valsDbl, nVals);
	delete [] indices;
	delete [] valsDbl;
}

R2Image * ShapeMaterials::GetTexture(int textureID)
{
	assert(textureID>=0 && textureID < (int)m_Textures.size());
	return m_Textures[textureID];
}

MaterialBRDF * ShapeMaterials::GetMaterial(int materialID)
{
	assert(materialID>=0 && materialID < (int)m_Materials.size());
	return m_Materials[materialID];
}

int ShapeMaterials::NTextures() const
{
	return (int)m_Textures.size();
}

int ShapeMaterials::NMaterials() const
{
	return (int)m_Materials.size();
}

const std::vector<int> & ShapeMaterials::GetAllFaces(int material)
{
	if (m_FacesPerMaterial==NULL)
	{
		double uv[6];
		int mID;
		assert(m_pMesh!=NULL);
		m_FacesPerMaterial = new std::vector<int>[NMaterials()+1];
		for (int i=0; i<m_pMesh->NFaces(); i++)
		{
			FaceMaterial(i, &(uv[0]), &(uv[1]), &(uv[2]), &(uv[3]), 
						 &(uv[4]), &(uv[5]), &mID);
			if (mID==-1)
				mID = NMaterials();
			m_FacesPerMaterial[mID].push_back(i);
		}
	}
	if (material<0 || material > NMaterials())
		material = NMaterials();
	return m_FacesPerMaterial[material];
}

void ShapeMaterials::WriteShapeValues(const VKString & filename)
{	
	// write materials
	FileManagement::CreateDirsOnPathToFile(filename.c_str());
	FILE * file = fopen(filename.c_str(), "wb");
	assert(file!=NULL);
	WriteCachedShapeValues(file);
	int nMaterials = NMaterials();
	int nTextures = NTextures();
	fwrite(&nMaterials, sizeof(int), 1, file);
	fwrite(&nTextures, sizeof(int), 1, file);
	for (int i=0; i<nMaterials; i++)
	{
		m_Materials[i]->m_EmitClr.WriteColorOrTexture(file);
		m_Materials[i]->m_DiffuseClr.WriteColorOrTexture(file);
		m_Materials[i]->m_AmbientClr.WriteColorOrTexture(file);
		m_Materials[i]->m_SpecularClr.WriteColorOrTexture(file);
		fwrite(&(m_Materials[i]->m_fShininess), sizeof(double), 1, file);
		m_Materials[i]->m_ReflectiveClr.WriteColorOrTexture(file);
		fwrite(&(m_Materials[i]->m_Reflectivity), sizeof(double), 1, file);
		m_Materials[i]->m_OpacityClr.WriteColorOrTexture(file);
		fwrite(&(m_Materials[i]->m_IndexOfRefraction), sizeof(double), 1, file);
		fwrite(&(m_Materials[i]->m_iMaterialID), sizeof(int), 1, file);
	}
	fclose(file);
	
	// write images
	VKString fileDir = filename.FilePath();
	for (int i=0; i<nTextures; i++)
	{
		assert(m_Textures[i]->NComponents()==3);
//		int nVals = m_Textures[i]->Width()*m_Textures[i]->Height() * m_Textures[i]->NComponents();
//		unsigned char * pixels = (unsigned char*)m_Textures[i]->Pixels();
//		for (int p=0; p<nVals; p++)
//			if (piexls[p] < 0
		
		m_Textures[i]->Write((fileDir+"/MatTexture"+VKString::number(i)+".jpg").c_str());
	}
}

void ShapeMaterials::ReadShapeValues(const VKString & filename)
{
	// read materials 
	FILE * file = fopen(filename.c_str(), "rb");
	assert(file!=NULL);
	ReadCachedShapeValues(file);
	int nMaterials, nTextures;
	fread(&nMaterials, sizeof(int), 1, file);
	fread(&nTextures, sizeof(int), 1, file);
	for (int i=0; i<nMaterials; i++)
	{
		m_Materials.push_back(new MaterialBRDF());
		m_Materials[i]->m_EmitClr.ReadColorOrTexture(file);
		m_Materials[i]->m_DiffuseClr.ReadColorOrTexture(file);
		m_Materials[i]->m_AmbientClr.ReadColorOrTexture(file);
		m_Materials[i]->m_SpecularClr.ReadColorOrTexture(file);
		fread(&(m_Materials[i]->m_fShininess), sizeof(double), 1, file);
		m_Materials[i]->m_ReflectiveClr.ReadColorOrTexture(file);
		fread(&(m_Materials[i]->m_Reflectivity), sizeof(double), 1, file);
		m_Materials[i]->m_OpacityClr.ReadColorOrTexture(file);
		fread(&(m_Materials[i]->m_IndexOfRefraction), sizeof(double), 1, file);
		fread(&(m_Materials[i]->m_iMaterialID), sizeof(int), 1, file);
	}
	fclose(file);
	
	// read images
	VKString fileDir = filename.FilePath();
	for (int i=0; i<nTextures; i++)
	{
		VKString fName = fileDir+"/MatTexture"+VKString::number(i)+".jpg";
		assert(FileManagement::FileExists(fName.c_str()));
		m_Textures.push_back(new R2Image(fName.c_str()));
		assert(m_Textures[i]->NComponents()==3);
		
	}
}

void ShapeMaterials::RemoveFace(int f)
{
	ShapeValueRemoveFace(f);
	delete [] m_FacesPerMaterial;
	m_FacesPerMaterial = NULL;	// can be recomputed
}

void ShapeMaterials::RemoveFaces(int f1, int f2)
{
	ShapeValueRemoveFaces(f1, f2);
	delete [] m_FacesPerMaterial;
	m_FacesPerMaterial = NULL;	// can be recomputed
}




