/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ShapePoint_H
#define __ShapePoint_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "RNBasics/RNBasics.h"
#include "R2Shapes/R2Shapes.h"
#include "R3Shapes/R3Shapes.h"


class PointDataAtVertex;
class PointDataOnFace;
class PointDataMeshless;

class GeomShape;
class ShapeMesh;
class ShapePointCloud;
class Model3D;

class PointData
{
	public:
		PointData(const R3Point & p );
		virtual ~PointData();
		virtual R3Point GetPosition() const = 0;
		virtual R3Vector GetNormal() const = 0;
		virtual const ShapeMesh * GetMeshGeometry() const = 0;
		virtual PointData * Copy() const = 0;
		
		virtual PointDataAtVertex * AtVtxData();
		virtual PointDataOnFace * OnFaceData();
		virtual PointDataMeshless * MeshlessData();
	
		R3Point m_Point;
};

class PointDataMeshless : public PointData
{
	public:
		PointDataMeshless(const R3Point & p);
		virtual const ShapeMesh * GetMeshGeometry() const;
		virtual R3Point GetPosition() const;
		virtual R3Vector GetNormal() const;
		virtual PointDataMeshless * MeshlessData();
		virtual PointData * Copy() const;
};

class PointDataAtVertex : public PointData
{
	public:
		PointDataAtVertex(int vID, const ShapeMesh * mesh, const R3Point & p = R3Point(0,0,0));
		virtual const ShapeMesh * GetMeshGeometry() const;
		virtual R3Point GetPosition() const;
		virtual R3Vector GetNormal() const;
		virtual PointDataAtVertex * AtVtxData();
		virtual PointData * Copy() const;
	
		int m_iVertID;	
		const ShapeMesh * m_pMesh;
};

class PointDataOnFace : public PointData
{
	public:
		PointDataOnFace(int tID, double b1, double b2, const ShapeMesh * mesh);
		PointDataOnFace(int tID, double b1, double b2, const ShapeMesh * mesh, const R3Point & p);
		virtual const ShapeMesh * GetMeshGeometry() const;
		virtual R3Point GetPosition() const;
		virtual R3Vector GetNormal() const;
		virtual PointDataOnFace * OnFaceData();
		virtual PointData * Copy() const;
	
		double m_Bary[2];
		int m_iTriID;
		const ShapeMesh * m_pMesh;
};


class ShapePoint
{
	public:
		friend class ShapePointCloud;
	
		enum ShapePointType
		{
			SHAPE_POINT_AT_VTX,
			SHAPE_POINT_ON_FACE,
			SHAPE_POINT_MESHLESS
		};
		ShapePoint();
	
		ShapePoint(const ShapePoint & shapePnt);
	
		ShapePoint(const ShapeMesh * mesh, int vID, const R3Point & pos = R3Point(0,0,0),
				   const ShapePointCloud * pnts = NULL, int pID=-1);
		ShapePoint(const ShapeMesh * mesh, int tID, double b1, double b2,
				   const ShapePointCloud * pnts = NULL, int pID=-1);
		ShapePoint(const ShapeMesh * mesh, int tID, double b1, double b2, const R3Point & pos,
				   const ShapePointCloud * pnts = NULL, int pID=-1);
		ShapePoint(const R3Point & p, const ShapePointCloud * pnts = NULL, int pID = -1);
		virtual ~ShapePoint();
		virtual ShapePointType GetPntType() const;
		virtual R3Point GetPosition() const;
		virtual R3Vector GetNormal() const;
		virtual int PointID() const;
		virtual void SetPointID(int pID);
	
	// doesn't work for meshless
		virtual void GetBarycentricCoords(int * fID, double * b1, double * b2,
										  const ShapeMesh ** mesh) const;
		virtual int VertexID(bool * atVtx = NULL) const;
	
		virtual bool operator != (const ShapePoint & p) const;
		virtual bool operator == (const ShapePoint & p) const;

		virtual const ShapeMesh * GetMeshGeometry() const;
		virtual bool SameMeshGeometry(const ShapeMesh * shape ) const;
		virtual bool SameMeshGeometry(const GeomShape * shape ) const;
	
		virtual const ShapePointCloud * GetPntsSet() const;
		virtual void SetPntsSet(const ShapePointCloud * pntsSet);
		virtual bool SamePntsSet(const ShapePointCloud * pntsSet) const;
	
		virtual const Model3D * GetModel() const;

		virtual ShapePoint & operator=(const ShapePoint & shapePnt);
		virtual bool Invalid() const;
	
	protected:
		PointData * m_pPntData;
	
	// might be null if not part of a set
		const ShapePointCloud * m_pPntsSet;
		int m_iIDInSet;
};

#endif

