/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ShapePointCloud_H
#define __ShapePointCloud_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "GeomShape.h"
#include "ShapePoint.h"

class PointCloudAllVertices;
class PointCloudAllFaces;
class ShapeMesh;
class ShapePointCloud;

/**
 * Just for a search algorithm
 */
class PointInSet
{
	public:
		PointInSet();
		R3Point GetPosition() const;
		ShapePoint GetModelPoint() const;

		ShapePointCloud * pnts;
		int id;
};

R3Point global_PointPosition(const PointInSet * pnt, void * data);
R2Point global_PointPosition2D(const PointInSet * pnt, void * data);


/**
 * ShapePointCloud
 */
class ShapePointCloud : public GeomShape
{
	public:
		static VKString m_sDefaultAllVtx;
		static VKString m_sDefaultAllFace;
	
		static VKString ShapeMeshFile(const VKString & pntSetFile);
	
		ShapePointCloud(const ShapeMesh * meshGeom, const VKString & pntSetName);
		ShapePointCloud(GeomShape::GeomShapeTypes tp, const VKString & pntSetName,
						const VKString & meshGeomName, const Model3D * model);
		virtual ~ShapePointCloud();
	
		static PointCloudAllVertices * CreateAllVtxSet(const VKString & pntSetName,
													   const VKString & meshGeomName, Model3D * model);
		static PointCloudAllFaces * CreateAllFacesSet(const VKString & pntSetName,
													  const VKString & meshGeomName, Model3D * model);
		static ShapePointCloud * CreateIterFar(const VKString & pntSetName, const VKString & meshGeomName,
											   Model3D * model, int nPnts, const VKString & distanceMetric,
											   ShapePointCloud * selectFromSet = NULL,
											   int firstPoint = 0, bool verbose = true);
	
		static ShapePointCloud * CreateRandom(const VKString & pntSetName, const VKString & meshGeomName,
											  Model3D * model, int nPnts);
			
		static ShapePointCloud * CreateAGDExtrema(const VKString & pntSetName, const VKString & meshGeomName,
												  Model3D * model, int nPnts = -1, bool verbose = true,
												  int nSampPnts = -1, double minDist = -1, double smoothSigma = -1, 
												  bool minima = false, bool maxima = true);
		static ShapePointCloud * CreateCurvatureExtrema(const VKString & pntSetName,
														const VKString & meshGeomName, Model3D * model, int nPnts);
	
		static ShapePointCloud * CreateFromExternalVtxList(const VKString & pntSetName, const VKString & meshGeomName,
														   Model3D * model, const VKString & externalFile);
	
		virtual void WriteGeometryFile(const VKString & geomFile) const;
		virtual void ReadGeometryFile(const VKString & geomFile);
	
		static void WritePointSet(FILE * file, const std::vector<ShapePoint> & pnts);
	// note: does not add pnts to shapes. partOfSet and mesh can be NULL
		static void ReadPointSet(const ShapeMesh * mesh,
								 const ShapePointCloud * partOfSet,
								 FILE * file, std::vector<ShapePoint> & pnts);
	
		virtual const ShapeMesh * GetMesh() const;
		virtual ShapePoint GetPoint(int pntID) const;
		virtual R3Point PointPosition(int pntID) const;
		virtual R3Vector PointNormal(int pntID) const;
		virtual int AddPoint(const ShapePoint & pnt);
		virtual void ReplacePoint(int id, const ShapePoint & newPnt);
		virtual void RemovePoint(int id);
		virtual void Clear();
		virtual int NumPoints() const;
		virtual ShapePointCloud * SafeCastToPointCloud();
	
		virtual R3Point Centroid() const;
		virtual double AverageRadius(const R3Point & centroid) const;
		virtual R3Triad PCAAxes(R3Vector * upVector = NULL) const;
		virtual R3Triad PCAFromPoints(const std::vector<int> & pntIDs, double * vars = NULL, double * ctr = NULL) const;
		virtual R3Triad ComputePCAAxesNoUp(double * variances=NULL) const;
		virtual R3Triad ComputePCAAxesUpVec(const R3Vector & upVec, double * variances=NULL) const;
		virtual R3Affine PCANormTransform(R3Vector * upVector = NULL, bool trans=true, bool rot=true, bool sca=true) const;

		virtual R3Kdtree<const PointInSet *> * CreateSubsetKdTree(const std::vector<int> & subset) const;
		virtual const R3Kdtree<const PointInSet *> * GetKdTree() const;
		virtual const R2Kdtree<const PointInSet *> * GetKdTree2D() const;
	
		virtual void RearrangeAxes(char newOrder[3], int newSigns[3]);
		
	protected:
		const ShapeMesh * m_pGeometry;
		std::vector<ShapePoint> m_PointSet;
		R3Kdtree<const PointInSet *> * m_pPntsSearch;
		R2Kdtree<const PointInSet *> * m_pPntsSearch2D;
	
		PointInSet * m_PointsInSet;
		
		R3Vector * m_pUpVecForAxes;
		R3Triad * m_pPCAAxes;
};

class PointCloudAllVertices : public ShapePointCloud
{
	public:
		PointCloudAllVertices(const VKString & pntSetName,
							  const VKString & meshGeomName,
							  Model3D * model);
		virtual void WriteGeometryFile(const VKString & geomFile);
		virtual void ReadGeometryFile(const VKString & geomFile);

		virtual ShapePoint GetPoint(int pntID) const;
		virtual R3Point PointPosition(int pntID) const;
		virtual int AddPoint(const ShapePoint & pnt);
		virtual void ReplacePoint(int id, const ShapePoint & newPnt);
		virtual void RemovePoint(int id);
		virtual int NumPoints() const;
		virtual void RearrangeAxes(char newOrder[3], int newSigns[3]);
};

class PointCloudAllFaces : public ShapePointCloud
{
	public:
		PointCloudAllFaces(const VKString & pntSetName,
						   const VKString & meshGeomName,
						   Model3D * model);
		virtual void WriteGeometryFile(const VKString & geomFile);
		virtual void ReadGeometryFile(const VKString & geomFile);

		virtual ShapePoint GetPoint(int pntID) const;
		virtual R3Point PointPosition(int pntID) const;	
		virtual int AddPoint(const ShapePoint & pnt);
		virtual void ReplacePoint(int id, const ShapePoint & newPnt);
		virtual void RemovePoint(int id);
		virtual int NumPoints() const;
		virtual void RearrangeAxes(char newOrder[3], int newSigns[3]);
};

#endif

