/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "FileManagement.h"
#include "VKStringList.h"

#include "ShapeValue.h"
#include "ShapePointCloud.h"
#include "ShapeMesh.h"

VKString ShapeValue::m_sDefaultMaterials = "Materials";
VKString ShapeValue::m_sDefaultSceneGraphLabels = "SceneGraphNodes";

VKString ShapeValue::s_sDefaultConnectedComponents = "MeshConnComp";

ShapeValue::ShapeValue(GeomShape * shape, const VKString & valueName, int nDims)
{
	m_pShape = shape;
	InitPtrs();
	m_sValueName = valueName;
	m_iNDims = nDims;
}

ShapeValue::ShapeValue(GeomShape * shape)
{
	
	m_pShape = shape;
	InitPtrs();
}

void ShapeValue::InitPtrs()
{
	m_sValueName = "none";
	m_iNDims = -1;
	assert(m_pShape!=NULL);
	
	m_aValues = NULL;
	m_iNValues = -1;
	
	m_iNPerFaceIDs = -1;
	m_aPerFaceIDs = NULL;
	
	m_pMesh = m_pShape->SafeCastToMesh();
	m_pPnts = m_pShape->SafeCastToPointCloud();
	
	m_CacheType = SVTYPE_Unknown;
}

ShapeValue::~ShapeValue()
{
	if (m_aValues!=NULL)
		delete [] m_aValues;
	
	if (m_aPerFaceIDs!=NULL)
		delete [] m_aPerFaceIDs;
}

ShapeMaterials * ShapeValue::SafeCastToMaterials()
{
	return NULL;
}

ShapeLabels * ShapeValue::SafeCastToLabels()
{
	return NULL;
}

const ShapeMaterials * ShapeValue::SafeCastToMaterials() const
{
	return NULL;
}

const ShapeLabels * ShapeValue::SafeCastToLabels() const
{
	return NULL;
}


VKString ShapeValue::GetClassName() const
{
	return "ShapeValue";
}

VKString ShapeValue::GetShapeValueName() const
{
	return m_sValueName;
}

GeomShape * ShapeValue::GetShape()
{
	return m_pShape;
}

const GeomShape * ShapeValue::GetShape() const
{
	return m_pShape;
}

int ShapeValue::NDims() const
{
	return m_iNDims;
}

int ShapeValue::NCachedValues() const
{
	return m_iNValues;
}

void ShapeValue::Values(const ShapePoint & pnt, double * storeVals) const
{
	for (int i=0; i<NDims(); i++)
		storeVals[i] = Value(pnt, i);
}

double ShapeValue::Value(const ShapePoint & pnt, int dim) const
{
	return GetCached(pnt, dim);
}

void ShapeValue::InitCache(ShapeValueCacheType cacheType, int cacheSize)
{
	// clear cache
	if (m_aValues!=NULL)
		delete [] m_aValues;
	if (m_aPerFaceIDs!=NULL)
		delete [] m_aPerFaceIDs;
	
	m_CacheType = cacheType;
//	assert(m_aValues==NULL);
	
	if (m_CacheType == SVTYPE_PerFace)
	{
		assert(m_pMesh!=NULL);
		assert(cacheSize==-1 || cacheSize==m_pMesh->NFaces());	// shouldn't be specified
		m_iNValues = m_pMesh->NFaces();
	}
	else if (m_CacheType == SVTYPE_PerIndexedFaces)
	{
		assert(m_pMesh!=NULL);
		assert(cacheSize>0);
		m_iNValues = cacheSize;
		m_iNPerFaceIDs = 3 * m_pMesh->NFaces();
		m_aPerFaceIDs = new int[m_iNPerFaceIDs];
		for (int i=0; i<m_iNPerFaceIDs; i++)
			m_aPerFaceIDs[i] = -1;
	}
	else if (m_CacheType == SVTYPE_PerVertex)
	{
		assert(m_pMesh!=NULL);
		m_iNValues = m_pMesh->NVertices();
	}
	else if (m_CacheType == SVTYPE_PerPointSet)
	{
		assert(m_pPnts!=NULL);
		assert(cacheSize==-1 || cacheSize==m_pPnts->NumPoints());	// shouldn't be specified
		m_iNValues = m_pPnts->NumPoints();
	}
	else
	{
		std::cout<<"[ERROR] Could not init cache: unsupported type "<<cacheType<<std::endl;
		assert(false);
	}

	assert(m_iNValues>0 && m_iNDims>0);
	m_aValues = new double[m_iNValues * m_iNDims];
	for (int i=0; i<m_iNValues * m_iNDims; i++)
		m_aValues[i] = 0;
}

double ShapeValue::GetCached(const ShapePoint & pnt, int dim) const
{
	if (m_CacheType == SVTYPE_PerPointSet)
	{
		assert(m_pPnts!=NULL);
		assert(pnt.SamePntsSet(m_pPnts));
		assert(pnt.PointID()!=-1);
		assert(m_aValues!=NULL);
		int id = pnt.PointID() * m_iNDims + dim;
		assert(id < m_iNDims*m_iNValues);
		return m_aValues[id];
	}
	else if (m_CacheType == SVTYPE_PerFace)
	{
		const ShapeMesh * mesh = NULL; double b1, b2; int fID;
		pnt.GetBarycentricCoords(&fID, &b1, &b2, &mesh);	assert(fID>=0);
		int id = fID * m_iNDims + dim;		assert(id < m_iNDims*m_iNValues);
		return m_aValues[id];
	}
	assert(false);	// todo
	return -1;
}

bool ShapeValue::IsValueSet(const ShapePoint & pnt)
{
	assert(false);
}

void ShapeValue::SetIndexedFaceValues(const int * faceIndices, const int nFaceIndices,
									  const double * values, const int nValues)
{
	assert(m_CacheType==SVTYPE_PerIndexedFaces);
	assert(m_aValues!=NULL);
	assert(m_aPerFaceIDs!=NULL);
	assert(m_pMesh!=NULL);
	assert(nValues==m_iNValues*m_iNDims);
	assert(nFaceIndices==3*m_pMesh->NFaces());
	
	for (int i=0; i<nFaceIndices; i++)
		m_aPerFaceIDs[i] = faceIndices[i];

	for (int i=0; i<nValues; i++)
		m_aValues[i] = values[i];
}

void ShapeValue::SetPntsValues(const double * values, const int nValues)
{
	assert(m_CacheType==SVTYPE_PerPointSet);
	assert(m_pPnts!=NULL);
	assert(nValues==m_pPnts->NumPoints()*m_iNDims);
	for (int i=0; i<nValues; i++)
		m_aValues[i] = values[i];
}

void ShapeValue::SetFaceValues(const double * values, const int nValues)
{
	assert(m_CacheType==SVTYPE_PerFace);
	assert(m_pMesh!=NULL);
	assert(nValues==m_pMesh->NFaces()*m_iNDims);
	for (int i=0; i<nValues; i++)
		m_aValues[i] = values[i];
}

void ShapeValue::GetCachedFaceIndexedValues(int fID, int vID, double * saveVals)
{
	assert(m_aPerFaceIDs!=NULL);
	assert(vID>=0 && vID<3);
	assert(3*fID+vID < m_iNPerFaceIDs);

	int id = m_aPerFaceIDs[3*fID + vID];
	assert(id*m_iNDims+m_iNDims-1 < m_iNValues*m_iNDims);
	for (int i=0; i<m_iNDims; i++)
		saveVals[i] = m_aValues[id * m_iNDims + i];
}

void ShapeValue::SetDimNames(const VKStringList & dimnames)
{
	m_DimNames = dimnames;
}

VKString ShapeValue::GetDimName(int id) const
{
	if (id < m_DimNames.count())
		return m_DimNames[id];
	else
		return m_sValueName + "-"+VKString(id);
}

void ShapeValue::WriteShapeValues(const VKString & filename)
{
	FileManagement::CreateDirsOnPathToFile(filename.c_str());
	if (filename.endsWith(".arff"))			// ascii
	{
		std::ofstream textStream(filename.c_str());
		assert(textStream.is_open());
		
		textStream<<"@RELATION ";
		if (m_pMesh!=NULL)
			textStream<<"mesh-";
		else if (m_pPnts!=NULL)
			textStream<<"pnts-";
		textStream<<m_pShape->GeometryName().c_str()<<"\n";
		
		for (int i=0; i<NDims(); i++)
			textStream<<"@ATTRIBUTE "<<GetDimName(i).c_str()<<" NUMERIC\n";
		
		assert(m_pPnts!=NULL);
		
		textStream<<"@DATA\n";
		for (int i=0; i<m_pPnts->NumPoints(); i++)
		for (int d=0; d<NDims(); d++)
			textStream<<Value(m_pPnts->GetPoint(i), d)<<((d==NDims()-1) ? "\n" : ",");
		textStream.close();
	}
	else if (filename.endsWith(".vals"))	// binary
	{
		FILE * file = fopen(filename.c_str(), "wb");
		if (file==NULL)
			std::cout<<"[ERROR] Cannot write file: "<<filename.c_str()<<std::endl;
		assert(file!=NULL);
		WriteShapeValues(file);
		fclose(file);
	}
	else
	{
		std::cout<<"[ERROR] Unknown file format: "<<filename.c_str()<<std::endl;
		assert(false);
	}
}

void ShapeValue::ReadShapeValues(const VKString & filename)
{
	if (filename.endsWith(".arff"))			// ascii
	{
		std::ifstream textStream(filename.c_str());
		if (!textStream.is_open())
			std::cout<<"[ERROR] Cannot open: "<<filename.c_str()<<std::endl;
		assert(textStream.is_open());
		
		VKString tempStr;
		tempStr.readToken(textStream);
		if (tempStr.startsWith("%"))
		{
			tempStr.readLine(textStream);
			tempStr.readToken(textStream);
		}
		if (!(tempStr=="@RELATION" || tempStr=="@relation"))
			std::cout<<"[ERROR] Failed to parse: "<<filename.c_str()<<", str=|"<<tempStr.c_str()<<"|"<<std::endl;
		assert(tempStr=="@RELATION" || tempStr=="@relation");
		tempStr.readToken(textStream);
		VKStringList splitStr;
//		VKStringList splitStr = tempStr.split("-");
//		assert(splitStr.count()>0);
//		assert(splitStr[0]=="pnts" && m_pPnts!=NULL);
		assert(m_pPnts!=NULL);
		
		int ndims = 0;
		bool datamode = false;
		double * vals = NULL;
		int npnt = 0;
		while (!textStream.eof())
		{
			tempStr.readLine(textStream);
			if (tempStr==""||tempStr=="\r"||tempStr=="\n")
				continue;
			if (tempStr.startsWith("@ATTRIBUTE") || tempStr.startsWith("@attribute"))
			{
				assert(!datamode);
				assert(tempStr.endsWith("NUMERIC") || tempStr.endsWith("real"));
				ndims++;
			}
			else if (tempStr.startsWith("@DATA") || tempStr.startsWith("@data"))
				datamode=true;
			else if (datamode)
			{
				if (vals==NULL)
				{
					assert(m_pPnts!=NULL);
					vals = new double[m_pPnts->NumPoints()*ndims];
				}
				splitStr = tempStr.split(",");
				if (splitStr.count()<ndims)
					splitStr = tempStr.split(" ");
				assert(splitStr.count()==ndims);
				bool ok;
				if (npnt>=m_pPnts->NumPoints())
				{
					std::cout<<"[WARNING] Might be bad parse "<<filename.c_str()<<std::endl;
					std::cout<<"          Too many pnts: "<<npnt<<" / "<<m_pPnts->NumPoints()<<std::endl;
					break;
				}
				assert(npnt < m_pPnts->NumPoints());
				for (int i=0; i<ndims; i++)
				{
					vals[npnt*ndims + i] = splitStr[i].toDouble(&ok);
					if (!ok)
						std::cout<<"[ERROR] Failed to parse: "<<filename.c_str()<<" -- "<<splitStr[i].c_str()<<std::endl;
					assert(ok);
				}
				npnt++;
			}
			else
			{
				std::cout<<"[ERROR] Invalid line: "<<tempStr.c_str()<<" file: "<<filename.c_str()<<std::endl;
				assert(false);
			}
		}
		m_iNDims = ndims;
		assert(vals!=NULL && ndims>0);
		InitCache(SVTYPE_PerPointSet);
		SetPntsValues(vals, m_pPnts->NumPoints()*ndims);
		textStream.close();
	}
	else if (filename.endsWith(".vals"))	// binary
	{
		FILE * file = fopen(filename.c_str(), "rb");
		if (file==NULL)
			std::cout<<"[ERROR] Could not read: "<<filename.c_str()<<std::endl;
		assert(file!=NULL);
		ReadShapeValues(file);
		fclose(file);
	}
}

void ShapeValue::WriteShapeValues(FILE * file)
{	
	WriteCachedShapeValues(file);
}

void ShapeValue::ReadShapeValues(FILE * file)
{
	ReadCachedShapeValues(file);
}

void ShapeValue::WriteCachedShapeValues(FILE * file)
{
	VKString::WriteString(m_sValueName, file);	
	fwrite(&m_CacheType, sizeof(ShapeValueCacheType), 1, file);
	fwrite(&m_iNDims, sizeof(int), 1, file);
	fwrite(&m_iNValues, sizeof(int), 1, file);
	assert(m_aValues!=NULL);
	fwrite(m_aValues, sizeof(double), m_iNDims*m_iNValues, file);
	
	if (m_CacheType==SVTYPE_PerIndexedFaces)
	{
		assert(m_aPerFaceIDs!=NULL);
		fwrite(&m_iNPerFaceIDs, sizeof(int), 1, file);
		fwrite(m_aPerFaceIDs, sizeof(int), m_iNPerFaceIDs, file);
	}
}

void ShapeValue::ReadCachedShapeValues(FILE * file)
{
	if (m_sValueName=="none")
		m_sValueName = VKString::ReadString(file);
	else
		VKString::ReadString(file);
	fread(&m_CacheType, sizeof(ShapeValueCacheType), 1, file);
	fread(&m_iNDims, sizeof(int), 1, file);
	fread(&m_iNValues, sizeof(int), 1, file);
	assert(m_aValues==NULL);
	m_aValues = new double[m_iNDims*m_iNValues];
	fread(m_aValues, sizeof(double), m_iNDims*m_iNValues, file);
	
	if (m_CacheType==SVTYPE_PerIndexedFaces)
	{
		assert(m_aPerFaceIDs==NULL);
		fread(&m_iNPerFaceIDs, sizeof(int), 1, file);
		m_aPerFaceIDs = new int[m_iNPerFaceIDs];
		fread(m_aPerFaceIDs, sizeof(int), m_iNPerFaceIDs, file);
	}
}

void ShapeValue::RemoveFace(int f)
{
	ShapeValueRemoveFace(f);
}

void ShapeValue::RemoveFaces(int f1, int f2)
{
	ShapeValueRemoveFaces(f1, f2);
}

void ShapeValue::RemoveVertex(int v)
{
	ShapeValueRemoveVertex(v);
}

void ShapeValue::ShapeValueRemoveFaces(int f1, int f2)
{
	int ndel = (f2-f1+1);
	if (m_CacheType==SVTYPE_PerFace)
	{
		assert(m_pMesh!=NULL);
		assert(m_iNValues==m_pMesh->NFaces());
		
		m_iNValues=m_iNValues - ndel;
		for (int fi=f1; fi<m_iNValues; fi++)
		for (int d=0; d<m_iNDims; d++)
			m_aValues[fi*m_iNDims + d] = m_aValues[(fi+ndel)*m_iNDims + d];
	}
	else if (m_CacheType==SVTYPE_PerIndexedFaces)
	{
		// todo: remove some indices...
		assert(m_pMesh!=NULL);
		assert(m_iNPerFaceIDs==3*m_pMesh->NFaces());
		m_iNPerFaceIDs=m_iNPerFaceIDs - 3*ndel;
		for (int fi=f1; fi<(m_pMesh->NFaces()-ndel); fi++)
		for (int lv=0; lv<3; lv++)
			m_aPerFaceIDs[3*fi + lv] = m_aPerFaceIDs[3*(fi+ndel) + lv];
	}
	else
	{
		std::cout<<"[ERROR] RemoveFace for unknown cache type: "<<m_CacheType<<std::endl;
		assert(false);
	}
}

void ShapeValue::ShapeValueRemoveFace(int f)
{	
	// NOTE: this is not optimal - it does not clear up
	//			unreferenced values. Maybe do it once, later...
	// NOTE: only care about values that are being written 
	//			(e.g. material properties in SVTYPE_PerIndexedFaces)
	if (m_CacheType==SVTYPE_PerFace)
	{
		assert(m_pMesh!=NULL);
		assert(m_iNValues==m_pMesh->NFaces());
		
		m_iNValues--;
		for (int fi=f; fi<m_iNValues; fi++)
		for (int d=0; d<m_iNDims; d++)
			m_aValues[fi*m_iNDims + d] = m_aValues[(fi+1)*m_iNDims + d];
	}
	else if (m_CacheType==SVTYPE_PerIndexedFaces)
	{
		assert(m_pMesh!=NULL);
		assert(m_iNPerFaceIDs==3*m_pMesh->NFaces());
		m_iNPerFaceIDs-=3;
		for (int fi=f; fi<(m_pMesh->NFaces()-1); fi++)
		for (int lv=0; lv<3; lv++)
			m_aPerFaceIDs[3*fi + lv] = m_aPerFaceIDs[3*(fi+1) + lv];
	}
	else
	{
		std::cout<<"[ERROR] RemoveFace for unknown cache type: "<<m_CacheType<<std::endl;
		assert(false);
	}
}

void ShapeValue::ShapeValueRemoveVertex(int v)
{
	assert(false);
}

