#include "Binding.h"
#include "Interactor.h"
#include "BindingEnergy.h"
#include "IKSolver.h"
#include "ShapePointCloud.h"
#include "ShapeValue.h"
#include "PntCorrsPose.h"

#include "BindingManager.h"


//////////////////////
/////   EnergyInfo
//////////////////////
EnergyInfo::EnergyInfo()
{
	m_fTotalEnergy = -1;
	m_fPosePriorEnergy = -1;
	m_fFeatPointsEnergy = -1;
	
	m_EnergyVis = ENVIS_None;
	m_PruneReasons = PRUNE_None;
	m_Flags2Fill = 0xffffffff;
	
	m_LinkID = -1;
	m_SymmetricLink = -1;

}


//////////////////////
/////   EnergyParams
//////////////////////
EnergyParams::EnergyParams()
{
	m_iEnergyFlags = E_Isect | E_Proximity;

	m_wEndEffectorDist = 0;
	m_wSymmetryPrior = 0;
	m_wPosePrior = 0;
	m_wFeatures = 0;
	m_wUnassigned = 0;
	m_wGround = 0;
	m_wContactPairwise = 0;
	m_wSymmPlane = 0;
	m_wIsectPenalty = 0;
	m_wCoMPenalty = 0;
	m_fMaxContactPenalty = RN_INFINITY;
	m_fMaxGroundPenalty = RN_INFINITY;
	m_sRegionStats = "none";
	
	m_bPointStatsConditionedOnPoseID = false;
}

bool EnergyParams::IsEnergySet(EnergyFlags flag) const
{
	return (m_iEnergyFlags & flag)!=0;
}

void EnergyParams::ClearEnergyFlags()
{
	m_iEnergyFlags = 0;
}

void EnergyParams::AddEnergyFlag(EnergyFlags flag)
{
	m_iEnergyFlags |= flag;
}

EnergyParams::EnergyFlags EnergyParams::Str2EnergyFlag(const VKString & str)
{
	if (str=="E_Isect")
		return E_Isect;
	else if (str=="E_Proximity")
		return E_Proximity;
	else if (str=="E_Stability")
		return E_Stability;
	else if (str=="E_PosePrior")
		return E_PosePrior;
	else if (str=="E_Muscles")
		return E_Muscles;
	else
		return E_Invalid;
}

//////////////////////
/////   BindingEnergy
//////////////////////
BindingEnergy::BindingEnergy(Interactor * actor, Interactor * object)
{
	m_Interactors[Interactor::ACT_Actor] = actor;
	m_Interactors[Interactor::ACT_Object] = object;
	m_pPrior = NULL;
	m_pSolver = NULL;
	m_pIntersectionGrid = NULL;
}

void BindingEnergy::ReadPlanes(const VKString & planesFile, std::vector<ReflectionPlane> & planes)
{
	std::ifstream textStream(planesFile.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] Cannot open planes file: "<<planesFile.c_str()<<std::endl;
	assert(textStream.is_open());
	VKString lineStr;
	double * vals = new double[12];
	while (!textStream.eof())
	{
		lineStr.readLine(textStream);
		if (lineStr=="")
			continue;
		VKStringList lnSplit = lineStr.split(" ");
		assert(lnSplit.count()==12);
		bool ok;
		for (int i=0; i<12; i++)
		{
			vals[i] = lnSplit[i].toDouble(&ok);
			assert(ok);
		}
		ReflectionPlane reflPlane;
		
		reflPlane.m_Plane = R3Plane(R3Point(vals[0], vals[1], vals[2]), R3Vector(vals[3], vals[4], vals[5]));
		reflPlane.m_fRadius = vals[6];
		reflPlane.m_fValue = vals[7];
		reflPlane.m_fRadialValue = vals[8];
		reflPlane.m_fProximityValue = vals[9];
		reflPlane.m_fMirrorValue = vals[10];
		reflPlane.m_fWeight = vals[11];
		
		planes.push_back(reflPlane);
	}
	textStream.close();
}

void BindingEnergy::ReadPlanes(const VKString & planesFile)
{
	if (planesFile!="none" && m_Params.m_wSymmPlane!=0)
		ReadPlanes(planesFile, m_ReflectionPlanes);
}

void BindingEnergy::SetPosePrior(IKSolver * solver, PosePrior * prior)
{
	m_pSolver = solver;
	m_pPrior = prior;
	if (m_pSolver!=NULL && m_pPrior!=NULL)
	{
		m_Params.m_wEndEffectorDist = m_pSolver->m_wEndEffectorDist;
		m_Params.m_wPosePrior = m_pSolver->m_wPosePrior;
		m_Params.m_wSymmetryPrior = m_pSolver->m_wSymmetryPrior;
	}
}


void BindingEnergy::SetParams(const EnergyParams & params, double origrotation)
{
	m_Params = params;
	if (m_pSolver!=NULL)
	{
		assert(m_Params.m_wEndEffectorDist == m_pSolver->m_wEndEffectorDist);
		assert(m_Params.m_wPosePrior == m_pSolver->m_wPosePrior);
		assert(m_Params.m_wSymmetryPrior == m_pSolver->m_wSymmetryPrior);
	}
	if (m_Params.m_sRegionStats!="none")
		ReadRegionStats(m_Params.m_sRegionStats);
	
	if (m_Params.m_wIsectPenalty>0)
		m_pIntersectionGrid = new IntersectionGrid(m_Interactors[Interactor::ACT_Actor],
												   m_Interactors[Interactor::ACT_Object], origrotation);
}


IKSolver * BindingEnergy::GetSolver()
{
	return m_pSolver;
}

void BindingEnergy::ReadRegionStats(const VKString & filename)
{
	std::ifstream textStream(filename.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] Stats file does not exist"<<filename.c_str()<<std::endl;
	assert(textStream.is_open());
	VKString tempStr;
	tempStr.readToken(textStream);	assert(tempStr=="NPoses");
	textStream>>m_iNPoses;
	
	int nregions;
	tempStr.readToken(textStream);	assert(tempStr=="NRegions");
	textStream>>nregions;
	for (int i=0; i<nregions; i++)
	{
		m_RegionNGround.push_back(0);
		m_RegionNSurface.push_back(0);
	}
	
	tempStr.readToken(textStream);	assert(tempStr=="Ground");
	for (int i=0; i<nregions; i++)
	{
		tempStr.readToken(textStream);
		textStream>>m_RegionNGround[i];
	}
	tempStr.readToken(textStream);	assert(tempStr=="Surface");
	for (int i=0; i<nregions; i++)
	{
		tempStr.readToken(textStream);
		textStream>>m_RegionNSurface[i];
	}
	
	tempStr.readToken(textStream);	assert(tempStr=="NPairwiseDists");
	int npairs;
	textStream>>npairs;
	for (int i=0; i<npairs; i++)
	{
		int from, to, ndists;
		textStream>>from>>to>>ndists;
		assert(from<nregions && to<nregions);
		int pairid = from * nregions + to;
		while (pairid>=(int)m_PairwiseDistDistr.size())
			m_PairwiseDistDistr.push_back(NULL);
		
		m_PairwiseDistDistr[pairid] = new PosePointGauss(1, false);
		double dist;
		for (int j=0; j<ndists; j++)
		{
			textStream>>dist;
			m_PairwiseDistDistr[pairid]->Update(&dist);
		}
	}
	
	textStream.close();
}

void BindingEnergy::SetIntersectionGrid(IntersectionGrid * grid)
{
	assert(grid!=NULL);
	m_pIntersectionGrid = grid;
}

IntersectionGrid * BindingEnergy::GetIntersectionGrid()
{
	return m_pIntersectionGrid;
}

void BindingEnergy::FillPairwiseEnergy(const PntCorrsPose * posepts, EnergyInfo & energy) const
{
	IKSolver * solver = ((BindingEnergy*)this)->GetSolver();
	assert(solver!=NULL);
	int nregions = solver->m_RegionNames.count();
	energy.m_fContactPairwiseDistsEnergy = 0;
	for (int r1=0; r1 < posepts->NMatches(); r1++)
	{
		int reg1 = posepts->Region(r1);
		if (!posepts->Prim(r1).IsType(TargetPrimitive::PRIM_Point))
			continue;
		for (int r2=0; r2<posepts->NMatches(); r2++)
		{
			int reg2 = posepts->Region(r2);
			if (reg1>=reg2 || !posepts->Prim(r2).IsType(TargetPrimitive::PRIM_Point))
				continue;
			assert(reg1<nregions && reg2<nregions);
			int pairid = reg1 * nregions + reg2;
			if (pairid < (int)m_PairwiseDistDistr.size() && m_PairwiseDistDistr[pairid]!=NULL)
			{
				assert(m_PairwiseDistDistr[pairid]->m_iNDims==1);
				double mean = m_PairwiseDistDistr[pairid]->m_aMeans[0];
				double var = m_PairwiseDistDistr[pairid]->m_aVariances[0];
				double dist = R3Distance(posepts->Prim(r1).GetPosition(), posepts->Prim(r2).GetPosition());
				double currerr = pow(dist-mean, 2.)/var * m_Params.m_wContactPairwise;
//				if (dist<.2)
//					currerr = RN_INFINITY;/// not sure if necessary
				energy.m_fContactPairwiseDistsEnergy += currerr;
				
				if (energy.m_EnergyVis==ENVIS_ContactPairwise)
				{
					int pntID = solver->m_Pnts2Skeleton[reg1];
					assert (pntID>=0 && pntID<(int)energy.m_PerPointEnergy.size());
					energy.m_PerPointEnergy[pntID] += currerr/2.;
					pntID = solver->m_Pnts2Skeleton[reg2];
					assert (pntID>=0 && pntID<(int)energy.m_PerPointEnergy.size());
					energy.m_PerPointEnergy[pntID] += currerr/2.;
				}
			}
		}
	}
}

double BindingEnergy::ComputeCoMPenalty(const PntCorrsPose * posepts, int planeID, EnergyInfo & energy) const
{
	energy.m_fCoMPenalty = 0;
	if (m_Params.m_wCoMPenalty==0)
		return energy.m_fCoMPenalty;
	assert(m_pPrior!=NULL);
	
	if (planeID<0)
	{
		energy.m_fCoMPenalty = RN_INFINITY;
		return energy.m_fCoMPenalty;
	}
	
//	R3Point point = m_Interactors[Interactor::ACT_Object]->Centroid();  // KEEP IT -- center of mass
	R3Point point = m_ReflectionPlanes[planeID].m_Plane.Point(); 		// center of symmetry
	energy.m_fCoMPenalty = m_pPrior->COMDeviation(posepts, point);
//	std::cout<<"ComPenalty: "<<energy.m_fCoMPenalty<<"*"<<m_Params.m_wCoMPenalty<<std::endl;
	energy.m_fCoMPenalty *= m_Params.m_wCoMPenalty;
	return energy.m_fCoMPenalty;
}

int BindingEnergy::FindBestPlaneOfReflection(std::vector<ReflectionPlane> & planes, const PntCorrsPose * posepts,
											 IKSolver * solver, std::vector<int> & rois, double & energy)
{
	assert(planes.size()>0);
	energy = RN_INFINITY;
	// if prescribed -- use prescribed location
	// if not prescribed -- use joint location
	std::vector<R3Point> rpos;
	std::vector<R3Point> offsurfacepos;
	std::vector<int> rcorr;
	std::vector<bool> onsurface;
	for (int r=0; r<solver->m_RegionNames.count(); r++)	// assign positions due to all joints
	{
		rpos.push_back(R3Point(0,0,0));
		offsurfacepos.push_back(R3Point(0,0,0));
		rcorr.push_back(-1);
		onsurface.push_back(false);
	}

	for (int i=0; i<(int)rois.size(); i++)	// assign positions due to all joints
	{
		int r = rois[i];
		VKString rname = solver->m_RegionNames[r];
		if (rname.startsWith("left_"))
			rname = rname.replace("left_", "right_");
		else if (rname.startsWith("right_"))
			rname = rname.replace("right_", "left_");
		int corr = -1;
		for (int r2=0; r2<solver->m_RegionNames.count(); r2++)
			if (rname==solver->m_RegionNames[r2])
				corr = r2;
		assert(corr>=0);
		rcorr[r]= corr;
		rpos[r] = posepts->GetPosition(solver->m_Pnts2Skeleton[r]);
		offsurfacepos[r] = rpos[r];
	}
	
	for (int m=0; m<posepts->NMatches(); m++)	// re-assign positions for contact points
	{
		int r = posepts->Region(m);
		if (posepts->Prim(m).IsType(TargetPrimitive::PRIM_Point))
		{
			rpos[r] = posepts->Prim(m).GetPosition();
			onsurface[r] = true;
		}
	}
	
	int bestplane = -1;
	bool doNotPenalizePartiallyAssigned = true;
	double ignorePlaneThreshold = RN_INFINITY;	// set to infinity to avoid this check
	for (int p=0; p<(int)planes.size(); p++)	// find the minimal-penalty reflection plane
	{
		double evalue = 0;
		bool ignore = false;
		for (int r=0; r<(int)rcorr.size() && !ignore; r++)
		{
			if (rcorr[r]==-1)	// do not penalize unless roi
				continue;
			
			bool ronsurf = onsurface[r];
			bool corronsurf = onsurface[rcorr[r]];
			
			R3Vector pos = rpos[r] - planes[p].m_Plane.Point();
			R3Vector corrpos = rpos[rcorr[r]] - planes[p].m_Plane.Point();
			
			if (pos.Length()>ignorePlaneThreshold)	// too far...
				ignore = true;
			
			if (doNotPenalizePartiallyAssigned && (ronsurf!=corronsurf))
			{
				pos = offsurfacepos[r] - planes[p].m_Plane.Point();
				corrpos = offsurfacepos[rcorr[r]] - planes[p].m_Plane.Point();
			}
			
			pos.Mirror(planes[p].m_Plane);
			evalue += R3Distance(pos.Point(), corrpos.Point());
		}
		
		if (evalue < energy && !ignore)
		{
			energy = evalue;
			bestplane = p;
		}
	}
	
	return bestplane;
	
	//	// avoid left-right twists
	//	assert(bestplane>=0);
	//	double left_is_positive_distance = 0;
	//	double left_is_negative_distance = 0;
	//	for (int r=0; r<m_pSolver->m_RegionNames.count(); r++)
	//	{
	//		if (!ClassifiedRegion(r))	// irrelevant region
	//			continue;
	//
	//		bool left = m_pSolver->m_RegionNames[r].startsWith("left_");
	//		bool right = m_pSolver->m_RegionNames[r].startsWith("right_");
	//
	//		if (left || right)
	//		{
	//			R3Vector pos = rpos[r] - m_ReflectionPlanes[bestplane].m_Plane.Point();
	//			double dot = pos.Dot(m_ReflectionPlanes[bestplane].m_Plane.Normal());
	//			if (dot<0)	// negative
	//			{
	//				if (left)
	//					left_is_positive_distance += vkAbs(dot);	// left && negative  => left_is_positive==false
	//				else if (right)
	//					left_is_negative_distance += vkAbs(dot);	// right && negative => left_is_negative==false
	//			}
	//			else
	//			{
	//				if (left)
	//					left_is_negative_distance += vkAbs(dot);	// left && positive  => left_is_negative==false
	//				else if (right)
	//					left_is_positive_distance += vkAbs(dot);	// right && positive => left_is_positive==false
	//			}
	//		}
	//	}
	//	energy.m_fSymmPlaneEnergy += vkMin(left_is_negative_distance, left_is_positive_distance)*m_Params.m_wSymmPlane;
}

double BindingEnergy::ComputeReflPlaneEnergy(const PntCorrsPose * posepts, EnergyInfo & energy) const
{
	energy.m_fSymmPlaneEnergy = 0;
	if (m_Params.m_wSymmPlane==0 || m_ReflectionPlanes.size()==0)
		return energy.m_fSymmPlaneEnergy;
	
	std::vector<int> rois;
	for (int r=0; r<m_pSolver->m_RegionNames.count(); r++)	// assign positions due to all joints
		if (ClassifiedRegion(r))
			rois.push_back(r);
	BindingEnergy * nct = (BindingEnergy*)this;
	int bp = FindBestPlaneOfReflection(nct->m_ReflectionPlanes, posepts, nct->m_pSolver, rois, energy.m_fSymmPlaneEnergy);
	double com = ComputeCoMPenalty((PntCorrsPose*) posepts, bp, energy);
	energy.m_fSymmPlaneEnergy *= m_Params.m_wSymmPlane;
	return energy.m_fSymmPlaneEnergy + energy.m_fCoMPenalty;
}

double BindingEnergy::ComputeEnergy(const PntCorrsPose * posepts, EnergyInfo & energy) const
{
	assert(posepts!=NULL);
	energy.m_fTotalEnergy = 0;
	energy.m_fPosePriorEnergy = 0;
	energy.m_fFeatPointsEnergy = 0;
	energy.m_fEndEffectorsEnergy = 0;
	energy.m_fUnassignedPointsEnergy = 0;
	energy.m_fSymmetryEnergy = 0;
	energy.m_fIntersectionEnergy = 0;
	energy.m_fContactPairwiseDistsEnergy = 0;
	energy.m_fGridOverlapMinEnergy = 0;
	energy.m_fSymmPlaneEnergy = 0;
	energy.m_fCoMPenalty = 0;
	
	// fill joint features
	assert(m_pSolver!=NULL);
		
	if (energy.m_EnergyVis!=0)
	{
		energy.m_PerPointEnergy.clear();
		for (int i=0; i<(int)m_pSolver->m_Joints.size(); i++)
			energy.m_PerPointEnergy.push_back(0);
	}
	
	FillPairwiseEnergy(posepts, energy);
	
	// ENERGY TERM: feature points
	const ShapeValue * objvals = NULL;
	const ShapePointCloud * objpnts = NULL;
	if (m_Interactors[Interactor::ACT_Object]!=NULL)
		m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
	if (objpnts!=NULL &&  objvals!=NULL)
	{
		//std::cout<<"NMatches: "<<std::endl;
		for (int i=0; i<(int)posepts->NMatches(); i++)
		{
			int rid = posepts->Region(i);
			const TargetPrimitive & trg = posepts->Prim(i);
			double val = ContactPenalty(posepts->GetPoseID(), rid, trg, objpnts, objvals);
			
			//std::cout<<"RID: "<<rid<<" -> "<<val<<std::endl;

			if (energy.m_EnergyVis==ENVIS_ContactFeatures)
			{
				assert(m_pSolver!=NULL && rid < (int)m_pSolver->m_Pnts2Skeleton.size());
				assert(m_pSolver->m_Pnts2Skeleton[rid] < (int)energy.m_PerPointEnergy.size());
				energy.m_PerPointEnergy[m_pSolver->m_Pnts2Skeleton[rid]] = val;
			}
			energy.m_fFeatPointsEnergy += val;
		}
		
		// for all points that are not contact points
		std::vector<bool> included;
		for (int i=0; i<(int)m_pSolver->m_RegionNames.count(); i++)
			included.push_back(false);
		for (int i=0; i<(int)posepts->NMatches(); i++)
		{
			assert(posepts->Region(i) < (int)included.size());
			included[posepts->Region(i)] = true;
		}
		
		// penalize unassigned
		//std::cout<<"Penalizing unassigned: "<<std::endl;
		for (int i=0; i<(int)included.size(); i++)
		{
			//std::cout<<"   Cla("<<i<<") "<<ClassifiedRegion(i)<<" inc: "<<included[i]<<std::endl;
			if (!included[i] && ClassifiedRegion(i))
			{
				double unassignedPenalty = UnassignedPenalty(posepts->GetPoseID(), i);
				energy.m_fUnassignedPointsEnergy += unassignedPenalty;
				if (energy.m_EnergyVis==ENVIS_Unassigned)
					energy.m_PerPointEnergy[m_pSolver->m_Pnts2Skeleton[i]] = unassignedPenalty;
			}
		}
	}
	
	// ENERGY TERMs: end effectors and pose prior
	assert(m_pSolver->m_wEndEffectorDist == m_Params.m_wEndEffectorDist);
	assert(m_pSolver->m_wPosePrior == m_Params.m_wPosePrior);
	assert(m_pSolver->m_wSymmetryPrior == m_Params.m_wSymmetryPrior);
//	m_pSolver->FillDirectionVector(posepts, NULL, energy, NULL);
	m_pSolver->GetEnergy((PntCorrsPose*) posepts, energy);
	
	if (m_pIntersectionGrid!=NULL && m_Params.m_wIsectPenalty>0)
	{
		m_pIntersectionGrid->ComputeEnergy((PntCorrsPose*) posepts, m_pSolver, energy);
		energy.m_fIntersectionEnergy *= m_Params.m_wIsectPenalty;
	}
	
	ComputeReflPlaneEnergy((PntCorrsPose*) posepts, energy);
	
	PntCorrsPose * temppose = ApproximateMinEnergy((PntCorrsPose*) posepts, energy);
	delete temppose;
	
	// FINAL ENERGY
	energy.m_fTotalEnergy = 0;
	energy.m_fTotalEnergy += energy.m_fPosePriorEnergy + energy.m_fSymmetryEnergy+energy.m_fFeatPointsEnergy;
	energy.m_fTotalEnergy += energy.m_fEndEffectorsEnergy+energy.m_fUnassignedPointsEnergy;
	energy.m_fTotalEnergy += energy.m_fIntersectionEnergy + energy.m_fContactPairwiseDistsEnergy;
	energy.m_fTotalEnergy += energy.m_fSymmPlaneEnergy+energy.m_fCoMPenalty;
	return energy.m_fTotalEnergy;
}

double BindingEnergy::UnassignedPenalty(int poseid, int rid) const
{
	double unassignedPenalty = m_Params.m_fMaxContactPenalty;
	double unassignedProb = UnassignedProbability(poseid, rid);
	if (unassignedProb>0)
		unassignedPenalty = vkMin(m_Params.m_fMaxContactPenalty, -log(unassignedProb));
	unassignedPenalty *= m_Params.m_wUnassigned;
	return unassignedPenalty;
}

void BindingEnergy::OffsetGroundProb(double offset, int regionID)
{
	while (regionID >= (int)m_OffsetGroundProb.size())
		m_OffsetGroundProb.push_back(0);
	m_OffsetGroundProb[regionID] = offset;
}

void BindingEnergy::OffsetUnassignedProb(double offset, int regionID)
{
	while (regionID >= (int)m_OffsetUnassignedProb.size())
		m_OffsetUnassignedProb.push_back(0);
	m_OffsetUnassignedProb[regionID] = offset;
}


double BindingEnergy::ContactPenalty(int poseid, int rid, const TargetPrimitive & trg,
									 const ShapePointCloud * objpnts, const ShapeValue * objvals) const
{
	double val = 0;
//	std::cout<<"Setting penalty..."<<std::endl;
	if (trg.IsType(TargetPrimitive::PRIM_Point))
	{
		int pid = -1;
		if  (trg.GetPoint().SamePntsSet(objpnts))
			pid = trg.GetPoint().PointID();
		else
		{
			const PointInSet * closest = objpnts->GetKdTree()->FindClosest(trg.GetPosition());
			pid = closest->id;
		}
		
		assert(pid>=0);
		double prob = objvals->Value(objpnts->GetPoint(pid), rid);
		
		if (prob>0)	// just to know that it's valid
			val = vkMin(m_Params.m_fMaxContactPenalty, -log(prob)+1E-50);
		else
			val = m_Params.m_fMaxContactPenalty;
		val *= m_Params.m_wFeatures;
	}
	else if (trg.IsType(TargetPrimitive::PRIM_PointMeshless))
		val = m_Params.m_wFeatures * m_Params.m_fMaxContactPenalty;
	else if (trg.IsType(TargetPrimitive::PRIM_Ground))
	{
		double groundProbability = GroundProbability(poseid, rid);
		if (groundProbability>0)
			val = vkMin(m_Params.m_fMaxGroundPenalty, -log(groundProbability))*m_Params.m_wGround;
		else
		{
			if (m_pSolver->m_RegionNames[rid]=="left_toe" || m_pSolver->m_RegionNames[rid]=="right_toe")
				val = m_Params.m_fMaxGroundPenalty*m_Params.m_wGround;
			else
				val = RN_INFINITY;
		}
	}
	else if (trg.IsType(TargetPrimitive::PRIM_PartialUnassigned))
	{
		val = 0;
	}
	else
		assert(false);
	return val;
}


//////// COMMENTED OUT ON JAN 11, 2014 - keep for now
//void BindingEnergy::ApproximateMinEnergy(const PntCorrsPose * posepts, EnergyInfo & energy,
//										 std::vector<int> * regions, std::vector<TargetPrimitive> * matches) const
//{
//	if ((energy.m_Flags2Fill & EnergyParams::E_Approximate)==0)
//		return;
//	std::vector<R3Point> previous;
//	double minDist2Previous = 0.4;
//	
//	const ShapeValue * objvals = NULL;
//	const ShapePointCloud * objpnts = NULL;
//	if (m_Interactors[Interactor::ACT_Object]!=NULL)
//		m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
//	if (objpnts==NULL || objvals==NULL)
//		return;
//	
//	int nsurfaceassigned = 0;
//	int npartialunassigned = 0;
//	double assignedPenalty = 0;
//	R3Point refPos(0,0,0);
//	int refReg = -1;
//	for (int m=0; m<posepts->NMatches(); m++)	// go over matches
//	{
//		if (posepts->Prim(m).IsType(TargetPrimitive::PRIM_Point))
//		{
//			nsurfaceassigned++;
//			refPos = posepts->Prim(m).GetPosition();
//			previous.push_back(refPos);
//			refReg = posepts->Region(m);
//			assignedPenalty += ContactPenalty(posepts->GetPoseID(), refReg, posepts->Prim(m), objpnts, objvals);
//			if (regions!=NULL && matches!=NULL)
//			{
//				regions->push_back(refReg);
//				matches->push_back(posepts->Prim(m));
//			}
//		}
//		else if (posepts->Prim(m).IsType(TargetPrimitive::PRIM_PartialUnassigned))
//			npartialunassigned++;
//	}
//	
//	if (!(npartialunassigned>0 && nsurfaceassigned==1))	// for now a single reference contact...
//		return;
//	
//	// precompute per-contact values (also detect relevant contacts)
//	if (m_Region2Pnts.size()==0)
//	{
//		BindingEnergy * nct = (BindingEnergy*)this;
//		
//		for (int r=0; r<objvals->NDims(); r++)
//		{
//			nct->m_Region2Pnts.push_back(std::vector<int>());
//			nct->m_PntPenalty.push_back(std::vector<double>());
//		}
//
//		for (int p=0; p<objpnts->NumPoints(); p++)
//		{
//			for (int r=0; r<objvals->NDims(); r++)
//			{
//				if (ClassifiedRegion(r))
//				{
//					TargetPrimitive prim(TargetPrimitive::PRIM_Point, objpnts->GetPoint(p));
//					double value = ContactPenalty(posepts->GetPoseID(), r, prim, objpnts, objvals);
//
//					if (value < m_Params.m_fMaxContactPenalty * m_Params.m_wFeatures)
//					{
//						nct->m_Region2Pnts[r].push_back(p);
//						nct->m_PntPenalty[r].push_back(value);
//					}
//				}
//			}
//		}
//		assert(m_PntPenalty.size()==m_Region2Pnts.size());
//	}
//
//	double featErrors = 0;
//	double poseError = 0;
//	
//	assert(m_pPrior!=NULL && m_pPrior->IsAngles());
//	PriorAngles * prior = (PriorAngles*)m_pPrior;
//	
//	R3Affine w2wg = R3identity_affine;
//	w2wg.ZRotate(-posepts->GetParam(3));
//	
//	for (int m=0; m<posepts->NMatches(); m++)	// go over matches
//	{
//		const TargetPrimitive & prim = posepts->Prim(m);
//		// product between grid
//		if (prim.IsType(TargetPrimitive::PRIM_PartialUnassigned))
//		{
//			double bestSum = RN_INFINITY;
//			double bestFeat = RN_INFINITY;
//			double bestPose = RN_INFINITY;
//			
//			int rid = posepts->Region(m);
//			int gid = prior->GetGridID(rid, posepts);
//			
//			
//			//std::cout<<"Region: "<<rid<<std::endl;
//			TargetPrimitive::PrimType type = TargetPrimitive::PRIM_None;
//			
//			int lastAssignment = -1;
//			if (regions!=NULL)
//			{
//				assert(matches!=NULL && matches->size()==regions->size());
//				lastAssignment = (int)regions->size();
//				regions->push_back(rid);
//				matches->push_back(TargetPrimitive(TargetPrimitive::PRIM_PartialUnassigned));
//			}
//			
//			// unassigned penalty as a bound
//			double probUnassigned = UnassignedProbability(posepts->GetPoseID(), rid);
//			if (probUnassigned>0)
//			{
//				bestSum = UnassignedPenalty(posepts->GetPoseID(), rid);
//				bestFeat = UnassignedPenalty(posepts->GetPoseID(), rid);
//				bestPose = 0;
//				type = TargetPrimitive::PRIM_PartialUnassigned;
//				//std::cout<<" Unassigned: "<<bestFeat<<"+0"<<std::endl;
//			}
//			
//			if (gid<0)
//			{
//				//assert(probUnassigned>0);// note: changed jan 1
//				featErrors += bestFeat;
//				poseError += bestPose;
//				continue;
//			}
//			
//			const R3Grid * grid = prior->GetGrid(gid);
//			// ground penalty for grid cells that intersect ground
//			double probGround = GroundProbability(posepts->GetPoseID(), rid);
//			if (probGround>0)
//			{
//				TargetPrimitive groundprim(TargetPrimitive::PRIM_Ground);
//				double gEnergy = ContactPenalty(posepts->GetPoseID(), rid, groundprim, objpnts, objvals);
//				double gBestPose = RN_INFINITY;
//				double gridsize=0;
//				const ShapePointCloud * nzset = prior->GetNonZeroProb(gid, &gridsize);
//				for (int gp=0; gp < nzset->NumPoints(); gp++)
//				{
//					double z = nzset->PointPosition(gp).Z() + refPos.Z();
//					//std::cout<<"Z: "<<z<<" in: "<<(z-gridsize)<<" -> "<<(z+gridsize)<<std::endl;
//					if (z-gridsize<=0 && z+gridsize>=0)
//					{
//						double val = PriorAngles::ExactWorldValue(nzset->PointPosition(gp), grid);
//						if (val < gBestPose)
//							gBestPose = val;
//					}
//				}
//				
//				if (gEnergy+gBestPose < bestSum)
//				{
//					bestSum = gEnergy+gBestPose;
//					bestFeat = gEnergy;
//					bestPose = gBestPose;
//					type = TargetPrimitive::PRIM_Ground;
//					
//					if (lastAssignment>=0 && regions!=NULL && matches!=NULL && lastAssignment<(int)matches->size())
//						(*matches)[lastAssignment] = TargetPrimitive(TargetPrimitive::PRIM_Ground);
//				}
//				//std::cout<<" Ground: "<<bestFeat<<"+"<<bestPose<<std::endl;
//			}
//			
//			
//			assert(rid < (int)m_Region2Pnts.size());
//			bool selectedPoint = false;
//			R3Point selectedPosition(0,0,0);
//			for (int p=0; p<m_Region2Pnts[rid].size(); p++)
//			{
//				R3Point pos = objpnts->PointPosition(m_Region2Pnts[rid][p]);
//				R3Point posorig = pos;
//				bool invalid = false;
//				for (int prev=0; prev<(int)previous.size() && !invalid; prev++)
//					invalid = R3Distance(pos, previous[prev])<minDist2Previous;
//				if (invalid)
//					continue;
//						
//				pos += -refPos;
//				w2wg.Apply(pos);
//				grid->WorldToGridTransformation().Apply(pos);
//				
//				int gi = (int)pos.X();
//				int gj = (int)pos.Y();
//				int gk = (int)pos.Z();
//				
//				if ((gi<0 || gi>=grid->XResolution() || gj<0 || gj>=grid->YResolution() ||
//					gk<0 || gk>=grid->ZResolution() || grid->GridValue(gi, gj, gk) >= RN_INFINITY)==false)
//				{
//					double currsum = grid->GridValue(gi, gj, gk) + m_PntPenalty[rid][p];
//					if (currsum < bestSum)
//					{
//						bestSum = currsum;
//						bestFeat = m_PntPenalty[rid][p];
//						bestPose = grid->GridValue(gi, gj, gk);
//						selectedPoint = true;
//						selectedPosition = posorig;
//						
//						//std::cout<<"Grid["<<gi<<" "<<gj<<" "<<gk<<"] = "<<bestPose<<std::endl;
//						//std::cout<<"GridSize="<<grid->XResolution()<<"x"<<grid->YResolution()<<"x"<<grid->ZResolution()<<std::endl;
//						
//						if (lastAssignment>=0 && regions!=NULL && matches!=NULL && lastAssignment<(int)matches->size())
//							(*matches)[lastAssignment] = TargetPrimitive(TargetPrimitive::PRIM_Point,
//																		 objpnts->GetPoint(m_Region2Pnts[rid][p]));
//						type = TargetPrimitive::PRIM_Point;
//					}
//				}
//			}
//			if (selectedPoint)
//				previous.push_back(selectedPosition);
//			
//			//std::cout<<"    Region["<<rid<<"]: "<<bestSum<<"="<<bestFeat<<" + "<<bestPose<<std::endl;
//			//std::cout<<"    Type="<<type<<std::endl;
//			if (type != TargetPrimitive::PRIM_PartialUnassigned)
//				assert(bestPose>0);
//			featErrors += bestFeat;
//			poseError += bestPose;
//		}
//	}
//	//std::cout<<"FINAL ENERGY: "<<featErrors<<"+"<<poseError<<"+"<<assignedPenalty<<std::endl;
//	//assert(poseError>0);
//	energy.m_fGridOverlapMinEnergy = featErrors + poseError + assignedPenalty;
//}

PntCorrsPose * BindingEnergy::ApproximateMinEnergy(const PntCorrsPose * posepts, EnergyInfo & energy) const
{
	if ((energy.m_Flags2Fill & EnergyParams::E_Approximate)==0)
		return NULL;
	std::vector<R3Point> previous;
	
	const ShapeValue * objvals = NULL;
	const ShapePointCloud * objpnts = NULL;
	if (m_Interactors[Interactor::ACT_Object]!=NULL)
		m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
	if (objpnts==NULL || objvals==NULL)
		return NULL;
	
	int nsurfaceassigned = 0;
	int npartialunassigned = 0;
	double assignedPenalty = 0;
	R3Point refPos(0,0,0);
	int refReg = -1;
	for (int m=0; m<posepts->NMatches(); m++)	// go over matches
	{
		if (posepts->Prim(m).IsType(TargetPrimitive::PRIM_Point))
		{
			nsurfaceassigned++;
			refPos = posepts->Prim(m).GetPosition();
			previous.push_back(refPos);
			refReg = posepts->Region(m);
			assignedPenalty += ContactPenalty(posepts->GetPoseID(), refReg, posepts->Prim(m), objpnts, objvals);
		}
		else if (posepts->Prim(m).IsType(TargetPrimitive::PRIM_PartialUnassigned))
			npartialunassigned++;
	}
	assignedPenalty += ComputeReflPlaneEnergy(posepts, energy);
//	assignedPenalty += ComputeCoMPenalty(posepts, energy);
	
	if (!(npartialunassigned>0 && nsurfaceassigned==1))	// for now a single reference contact...
		return NULL;
	
	// precompute per-contact values (also detect relevant contacts)
	if (m_Region2Pnts.size()==0)
	{
		BindingEnergy * nct = (BindingEnergy*)this;
		
		for (int r=0; r<objvals->NDims(); r++)
		{
			nct->m_Region2Pnts.push_back(std::vector<int>());
			nct->m_PntPenalty.push_back(std::vector<double>());
		}
		
		for (int p=0; p<objpnts->NumPoints(); p++)
		{
			for (int r=0; r<objvals->NDims(); r++)
			{
				if (ClassifiedRegion(r))
				{
					TargetPrimitive prim(TargetPrimitive::PRIM_Point, objpnts->GetPoint(p));
					double value = ContactPenalty(posepts->GetPoseID(), r, prim, objpnts, objvals);
					
					if (value < m_Params.m_fMaxContactPenalty * m_Params.m_wFeatures)
					{
						nct->m_Region2Pnts[r].push_back(p);
						nct->m_PntPenalty[r].push_back(value);
					}
				}
			}
		}
		assert(m_PntPenalty.size()==m_Region2Pnts.size());
	}
	
	double featErrors = 0;
	double poseError = 0;
	
	assert(m_pPrior!=NULL && m_pPrior->IsAngles());
	PriorAngles * prior = (PriorAngles*)m_pPrior;
	
	R3Affine w2wg = R3identity_affine;
	w2wg.ZRotate(-posepts->GetParam(3));
	
	PntCorrsPose * newpose = new PntCorrsPose(m_Interactors[Interactor::ACT_Object]);
	newpose->SetPoseID(posepts->GetPoseID());
	for (int m=0; m<posepts->NMatches(); m++)	// go over unassigned points
		newpose->AddMatch(posepts->Region(m), posepts->Prim(m));
	newpose->InitParams(m_pSolver->NJointParams(), m_pSolver->m_SolverFlags);
	for (int i=0; i<m_pSolver->NJointParams(); i++)
		newpose->Param(i) = posepts->GetParam(i);
	m_pSolver->ForwardKinematics(newpose);

	std::vector<Sortable> remainingoptions;
	remainingoptions.push_back(Sortable(0, NULL, -1));
	while ((int)remainingoptions.size()>0)	// keep assigning partial unassigned
	{
		remainingoptions.clear();
		std::vector<int> matchids;
		std::vector<TargetPrimitive> correspondences;
		std::vector<double> bestfeaterrs;
		std::vector<double> bestposeerrs;
		
		for (int m=0; m<newpose->NMatches(); m++)	// go over unassigned points
		{
			const TargetPrimitive & prim = newpose->Prim(m);
			
			// product between grid
			if (prim.IsType(TargetPrimitive::PRIM_PartialUnassigned))
			{
				double bestFeat, bestPose;
				TargetPrimitive newprim = ApproximateMinEnergyForPartMatch(newpose, posepts, m, bestFeat, bestPose, refPos, w2wg);
				remainingoptions.push_back(Sortable(bestFeat+bestPose, NULL, (int)matchids.size()));
				matchids.push_back(m);
				correspondences.push_back(newprim);
				bestfeaterrs.push_back(bestFeat);
				bestposeerrs.push_back(bestPose);
			}
		}
		
		if (remainingoptions.size()==0)
			break;
		std::sort(remainingoptions.begin(), remainingoptions.end());
		// SLOW VERSION (DO NOT DELETE!!!): especially suitable for intersecting pose distribution grids
		assert(remainingoptions[0].id < (int)matchids.size());
		int m = matchids[remainingoptions[0].id];
		if (correspondences[remainingoptions[0].id].IsType(TargetPrimitive::PRIM_Point)
			|| correspondences[remainingoptions[0].id].IsType(TargetPrimitive::PRIM_Ground))
			newpose->Prim(m) = correspondences[remainingoptions[0].id];
		else
			newpose->Clear(m);	// to avoid infinite loop
		featErrors += bestfeaterrs[remainingoptions[0].id];
		poseError += bestposeerrs[remainingoptions[0].id];
		
//		// FAST VERSION -- for now...
//		for (int a=0; a<(int)remainingoptions.size(); a++)
//		{
//			int m = matchids[remainingoptions[a].id];
//			if ((correspondences[remainingoptions[a].id].IsType(TargetPrimitive::PRIM_Point)
//				|| correspondences[remainingoptions[a].id].IsType(TargetPrimitive::PRIM_Ground))
//				&& IsValidAddition(newpose, newpose->Region(m), correspondences[remainingoptions[a].id]))
//			{
//				newpose->Prim(m) = correspondences[remainingoptions[a].id];
//				featErrors += bestfeaterrs[remainingoptions[a].id];
//				poseError += bestposeerrs[remainingoptions[a].id];
//			}
//			else
//				featErrors += UnassignedPenalty(newpose->GetPoseID(), newpose->Region(m));
//		}
//		break;
	}
	
	// newpose - clear all partial unassigned
	for (int m=0; m<(int)newpose->NMatches(); m++)
	{
		if (newpose->Prim(m).IsType(TargetPrimitive::PRIM_PartialUnassigned))
		{
			newpose->Clear(m);
			m--;
		}
	}

	// newpose - clear symmetric assignments (if one assigned - one unassinged)
	for (int m=0; m<(int)newpose->NMatches(); m++)
	{
		assert(newpose->Prim(m).IsType(TargetPrimitive::PRIM_Point) || newpose->Prim(m).IsType(TargetPrimitive::PRIM_Ground));
		// check if symmetry point is in (check by name)
		int rid = newpose->Region(m);
		assert(rid>=0 && rid<m_pSolver->m_RegionNames.count());
		bool left = m_pSolver->m_RegionNames[rid].startsWith("left_");
		bool right = m_pSolver->m_RegionNames[rid].startsWith("right_");
		if (left || right)
		{
			VKString countername = m_pSolver->m_RegionNames[rid];
			if (left)
				countername = countername.replace("left_", "right_");
			else
				countername = countername.replace("right_", "left_");
			bool ok = false;
			for (int m2=0; m2<(int)newpose->NMatches() && !ok; m2++)
				if (m_pSolver->m_RegionNames[newpose->Region(m2)] == countername)
					ok = true;
			if (!ok)  // unassign asymmetric part
			{
				newpose->Clear(m);
				m--;
			}
		}
	}

	//std::cout<<"FINAL ENERGY: "<<featErrors<<"+"<<poseError<<"+"<<assignedPenalty<<std::endl;
	//assert(poseError>0);
	energy.m_fGridOverlapMinEnergy = featErrors + poseError + assignedPenalty;
	
	
	return newpose;
}

TargetPrimitive BindingEnergy::ApproximateMinEnergyForPartMatch(PntCorrsPose * posepts, const PntCorrsPose * cond,
																int matchID, double & bestFeat, double & bestPose,
																const R3Point & refPos, const R3Affine & w2wg) const
{
	const ShapeValue * objvals = NULL;
	const ShapePointCloud * objpnts = NULL;
	if (m_Interactors[Interactor::ACT_Object]!=NULL)
		m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
	assert (objpnts!=NULL && objvals!=NULL);

	double bestSum = RN_INFINITY;
	bestFeat = RN_INFINITY;
	bestPose = RN_INFINITY;
	assert(m_pPrior!=NULL && m_pPrior->IsAngles());
	PriorAngles * prior = (PriorAngles*)m_pPrior;

	int rid = posepts->Region(matchID);
	int gid = prior->GetGridID(rid, cond);
	
	
	//std::cout<<"Region: "<<rid<<std::endl;
	TargetPrimitive bestPrim(TargetPrimitive::PRIM_None);
	
	// unassigned penalty as a bound
	double probUnassigned = UnassignedProbability(posepts->GetPoseID(), rid);
	//if (probUnassigned>0)	// vkchange: jan15, allways allow to keep some things unassigned
	{
		bestSum = UnassignedPenalty(posepts->GetPoseID(), rid);
		bestFeat = UnassignedPenalty(posepts->GetPoseID(), rid);
		bestPose = 0;
		bestPrim = TargetPrimitive(TargetPrimitive::PRIM_PartialUnassigned);
	}
	
	if (gid<0)
		return bestPrim;
	
	const R3Grid * grid = prior->GetGrid(gid);
	// ground penalty for grid cells that intersect ground
	double probGround = GroundProbability(posepts->GetPoseID(), rid);
	bool isvalid = IsValidAddition(posepts, rid, TargetPrimitive(TargetPrimitive::PRIM_Ground));
	if (probGround>0 && isvalid)
	{
		TargetPrimitive groundprim(TargetPrimitive::PRIM_Ground);
		double gEnergy = ContactPenalty(posepts->GetPoseID(), rid, groundprim, objpnts, objvals);
		double gBestPose = RN_INFINITY;
		double gridsize=0;
		const ShapePointCloud * nzset = prior->GetNonZeroProb(gid, &gridsize);
		for (int gp=0; gp < nzset->NumPoints(); gp++)
		{
			double z = nzset->PointPosition(gp).Z() + refPos.Z();
			//std::cout<<"Z: "<<z<<" in: "<<(z-gridsize)<<" -> "<<(z+gridsize)<<std::endl;
			if (z-gridsize<=0 && z+gridsize>=0)
			{
				double val = PriorAngles::ExactWorldValue(nzset->PointPosition(gp), grid);
				if (val < gBestPose)
					gBestPose = val;
			}
		}
		
		if (gEnergy+gBestPose < bestSum)
		{
			bestSum = gEnergy+gBestPose;
			bestFeat = gEnergy;
			bestPose = gBestPose;
			bestPrim = TargetPrimitive(TargetPrimitive::PRIM_Ground);
		}
	}
	
	assert(rid < (int)m_Region2Pnts.size());
	bool selectedPoint = false;
	R3Point selectedPosition(0,0,0);
	for (int p=0; p<m_Region2Pnts[rid].size(); p++)
	{
		TargetPrimitive candprim(TargetPrimitive::PRIM_Point, objpnts->GetPoint(m_Region2Pnts[rid][p]));
		isvalid = IsValidAddition(posepts, rid, candprim);
		
		R3Point pos = objpnts->PointPosition(m_Region2Pnts[rid][p]);
		R3Point posorig = pos;
		if (!isvalid)
			continue;
		
		pos += -refPos;
		w2wg.Apply(pos);
		grid->WorldToGridTransformation().Apply(pos);
		
		int gi = (int)pos.X();
		int gj = (int)pos.Y();
		int gk = (int)pos.Z();
		
		if ((gi<0 || gi>=grid->XResolution() || gj<0 || gj>=grid->YResolution() ||
			 gk<0 || gk>=grid->ZResolution() || grid->GridValue(gi, gj, gk) >= RN_INFINITY)==false)
		{
			double reflEnergy = ReflPenaltyEnergyIncrease(posepts, rid, candprim);
			//double reflEnergy = 0;
			double currsum = grid->GridValue(gi, gj, gk) + m_PntPenalty[rid][p] + reflEnergy;
			
			if (currsum < bestSum)
			{
				bestSum = currsum;
				bestFeat = m_PntPenalty[rid][p];
				bestPose = grid->GridValue(gi, gj, gk);
				selectedPoint = true;
				selectedPosition = posorig;
				
				bestPrim = TargetPrimitive(TargetPrimitive::PRIM_Point, objpnts->GetPoint(m_Region2Pnts[rid][p]));
			}
		}
	}
	return bestPrim;
}

double BindingEnergy::ReflPenaltyEnergyIncrease(PntCorrsPose * posepts, int regionid, const TargetPrimitive & candidate) const
{
	if (!candidate.IsType(TargetPrimitive::PRIM_Point))
		return 0;
	if (m_pSolver->m_RegionNames[regionid].startsWith("left_") || m_pSolver->m_RegionNames[regionid].startsWith("right_"))
		return 0;	// for now
	EnergyInfo energy;
	double e0 = ComputeReflPlaneEnergy(posepts, energy);
	int m=posepts->NMatches();
	posepts->AddMatch(regionid, candidate);
	double e1 = ComputeReflPlaneEnergy(posepts, energy);
	posepts->Clear(m);
	return e1 - e0;
}

bool BindingEnergy::IsValidAddition(PntCorrsPose * posepts, int rid, const TargetPrimitive & candidate) const
{
	assert(m_pPrior!=NULL && m_pPrior->IsAngles());
	PriorAngles * prior = (PriorAngles*)m_pPrior;

	double minDist2Previous = 0.4;
	bool isvalid = true;
	if (candidate.IsType(TargetPrimitive::PRIM_Point))
	{
		R3Point pos = candidate.GetPosition();
		for (int prev=0; prev<posepts->NMatches() && isvalid; prev++)
		{
			if (posepts->Prim(prev).IsType(TargetPrimitive::PRIM_Point))
				isvalid = R3Distance(pos, posepts->Prim(prev).GetPosition())>minDist2Previous;
		}
	}
	if (!isvalid)
		return false;
	
	int m=posepts->NMatches();
	posepts->AddMatch(rid, candidate);
	isvalid = prior->QuickValidityTest(posepts);
	posepts->Clear(m);
	
	return isvalid;
}

double BindingEnergy::ComputeBindingEnergy(const Binding * b, EnergyInfo & energy) const
{
	// pose probability and trained priors
	const InteractorPose * pose = b->GetPose(Interactor::ACT_Actor);
	assert(pose->GetPoseType()==InteractorPose::TYPE_PntCorrsPose);
	const PntCorrsPose * posepts = (PntCorrsPose*)pose;
	return ComputeEnergy(posepts, energy);
}

double BindingEnergy::SurfaceProbability(int poseid, int regionID) const
{
	if (poseid<0 || !m_Params.m_bPointStatsConditionedOnPoseID)
	{
		assert(m_iNPoses>0 && regionID<(int)m_RegionNGround.size() && regionID<(int)m_RegionNSurface.size());
		return (double)m_RegionNSurface[regionID]/(double)m_iNPoses;
	}
	assert(m_pPrior!=NULL);
	return m_pPrior->SurfaceProbability(poseid, regionID);
}

double BindingEnergy::GroundProbability(int poseid, int regionID) const
{
	double offset = 0;
	if (regionID < (int)m_OffsetGroundProb.size())
		offset = m_OffsetGroundProb[regionID];

	if (poseid<0 || !m_Params.m_bPointStatsConditionedOnPoseID)
	{
		assert(m_iNPoses>0 && regionID<(int)m_RegionNGround.size() && regionID<(int)m_RegionNSurface.size());
		double prob = (double)m_RegionNGround[regionID]/(double)m_iNPoses;
		return vkMin(prob+offset, 1.);
	}
	assert(m_pPrior!=NULL);
		
	return vkMin(m_pPrior->GroundProbability(poseid, regionID)+offset, 1.);
}

double BindingEnergy::UnassignedProbability(int poseid, int regionID) const
{
	double offset = 0;
	if (regionID < (int)m_OffsetUnassignedProb.size())
		offset = m_OffsetUnassignedProb[regionID];

	if (poseid<0 || !m_Params.m_bPointStatsConditionedOnPoseID)
	{
		assert(m_iNPoses>0 && regionID<(int)m_RegionNGround.size() && regionID<(int)m_RegionNSurface.size());
		double prob = ((double)(m_iNPoses - m_RegionNGround[regionID] - m_RegionNSurface[regionID])) / (double)m_iNPoses;
		return vkMin(prob+offset, 1.);
	}
	assert(m_pPrior!=NULL);
	return vkMin(m_pPrior->UnassignedProbability(poseid, regionID)+offset, 1.);
}

bool BindingEnergy::ClassifiedRegion(int regionID) const
{
	if (m_iNPoses>0 && regionID<(int)m_RegionNGround.size() && regionID<(int)m_RegionNSurface.size())
		return ((m_RegionNGround[regionID] + m_RegionNSurface[regionID]) > 0);
	else
	{
		if (m_ClassifiedRegion.size()==0)
		{
			const ShapeValue * objvals = NULL;
			const ShapePointCloud * objpnts = NULL;
			m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
			assert(objpnts!=NULL & objvals!=NULL);

			BindingEnergy * ncThis = (BindingEnergy*)this;
			for (int i=0; i<objvals->NDims(); i++)
				ncThis->m_ClassifiedRegion.push_back(false);
			
			for (int i=0; i<objpnts->NumPoints(); i++)
				for (int d=0; d<objvals->NDims(); d++)
					if (objvals->Value(objpnts->GetPoint(i), d)!=0)
						ncThis->m_ClassifiedRegion[d] = true;
		}
		if (regionID>=(int)m_ClassifiedRegion.size())
			return false;
		return m_ClassifiedRegion[regionID];
	}
	assert(false);
}

VKString BindingEnergy::GetIKOnlyEnergyStr(PntCorrsPose * pose, IKSolver * solver, PosePrior * prior, bool detailed)
{
	Interactor actor(Interactor::ACT_Actor, Interactor::TYPE_PntCorrs);
	BindingEnergy energy(&actor, NULL);
	energy.SetPosePrior(solver, prior);
	
	return energy.GetEnergyStr(pose, detailed);
}

VKString BindingEnergy::GetEnergyStr(BindingManager * bindings, int poseid, bool detailed) const
{
	std::vector<int> bindings2visit;
	if (poseid==-1)
	{
		for (int i=0; i<bindings->NBindings(); i++)
			bindings2visit.push_back(i);
	}
	else
		bindings2visit.push_back(poseid);

	VKString retStr = "";
	for (int i=0; i<(int)bindings2visit.size(); i++)
	{
		if (i>0 || poseid>=0)
			retStr+=VKString("\n==================================\n");
		retStr+=VKString("Pose: ")+VKString::number(bindings2visit[i])+"\n";
		const Binding * b = bindings->GetBinding(bindings2visit[i]);
		
		const InteractorPose * pose = b->GetPose(Interactor::ACT_Actor);
		assert(pose->GetPoseType()==InteractorPose::TYPE_PntCorrsPose);
		const PntCorrsPose * posepts = (PntCorrsPose*)pose;
		retStr += GetEnergyStr(posepts, detailed);
	}
	return retStr;
}

VKString BindingEnergy::GetEnergyStr(const PntCorrsPose * pnts, bool detailed) const
{
	VKString retStr = "";
	EnergyInfo energy;
	ComputeEnergy(pnts, energy);
	
	retStr += VKString("Energy: ")+VKString::number(energy.m_fTotalEnergy)+"\n";
	retStr += VKString("    End Effectors: ")+VKString::number(energy.m_fEndEffectorsEnergy)+"\n";
	retStr += VKString("    Pose Priors:   ")+VKString::number(energy.m_fPosePriorEnergy)+"\n";
	retStr += VKString("    Symm Priors:   ")+VKString::number(energy.m_fSymmetryEnergy)+"\n";
	retStr += VKString("    Features:      ")+VKString::number(energy.m_fFeatPointsEnergy)+"\n";
	retStr += VKString("    Unassigned:    ")+VKString::number(energy.m_fUnassignedPointsEnergy)+"\n";
	retStr += VKString("    Intersection:  ")+VKString::number(energy.m_fIntersectionEnergy)+"\n";
	retStr += VKString("    ContactPairs:  ")+VKString::number(energy.m_fContactPairwiseDistsEnergy)+"\n";
	retStr += VKString("    Symm Plane:    ")+VKString::number(energy.m_fSymmPlaneEnergy)+"\n";
	retStr += VKString("    Centroid:      ")+VKString::number(energy.m_fCoMPenalty)+"\n";
	retStr += VKString("Approximation: ")+"\n";
	retStr += VKString("    GridOverlap:  ")+VKString::number(energy.m_fGridOverlapMinEnergy)+"\n";
	
	if (!detailed)
		return retStr;
	
	std::vector<EnergyVisType> energyFlags;
	energyFlags.push_back(ENVIS_AngleAll);			// priors
	energyFlags.push_back(ENVIS_AngleSymmetry);		// symmetry
	energyFlags.push_back(ENVIS_EndEffectorDist);	// end effectors
	energyFlags.push_back(ENVIS_ContactFeatures);	// contact point features
	energyFlags.push_back(ENVIS_Unassigned);		// unassigned points
	energyFlags.push_back(ENVIS_Intersection);		// intersection
	energyFlags.push_back(ENVIS_ContactPairwise);	// pairwise energy
	
	for (int i=0; i<(int)energyFlags.size(); i++)
	{
		retStr+= VKString("\nDetails: ");
		switch(energyFlags[i])
		{
			case ENVIS_AngleAll:
				retStr+=VKString("Pose Priors");
				break;
			case ENVIS_AngleSymmetry:
				retStr+=VKString("Pose Symmetry");
				break;
			case ENVIS_EndEffectorDist:
				retStr+=VKString("End Effectors");
				break;
			case ENVIS_ContactFeatures:
				retStr+=VKString("Features");
				break;
			case ENVIS_Unassigned:
				retStr+=VKString("Unassigned");
				break;
			case ENVIS_Intersection:
				retStr+=VKString("Intersection");
				break;
			case ENVIS_ContactPairwise:
				retStr+=VKString("ContactPairwise");
				break;
			default:
				retStr+=VKString("Unknown");
				break;
		}
		retStr+="\n";
		
		VKStringList zerovals;
		EnergyInfo energydetails;
		energydetails.m_EnergyVis = energyFlags[i];
		ComputeEnergy(pnts, energydetails);
		if (energydetails.m_PerPointEnergy.size()!=m_pSolver->m_Joints.size())
		{
			retStr+=VKString("    [ERROR] Values not filled ")+VKString::number((int)energydetails.m_PerPointEnergy.size());
			retStr+=VKString(" / ")+VKString::number((int)m_pSolver->m_Joints.size());
		}
		for (int p=0; p<(int)energydetails.m_PerPointEnergy.size(); p++)
		{
			VKString descrStr = "unknown";
			if (energyFlags[i]==ENVIS_AngleAll || energyFlags[i]==ENVIS_AngleSymmetry)	 // link energy
			{
				SkeletonLink * link = m_pSolver->m_Joints[p]->m_pParent;
				if (link==NULL)
					continue;
				descrStr = link->m_pParent->m_sJointName+" -> "+link->m_pChild->m_sJointName;
			}
			else // point energy
				descrStr = m_pSolver->m_Joints[p]->m_sJointName.c_str();
			
			if (energydetails.m_PerPointEnergy[p]==0)
				zerovals.push_back(descrStr);
			else
			{
				descrStr = descrStr.CreateStrWithPadding(35, " ");
				retStr += VKString("    ")+descrStr+": "+VKString::number(energydetails.m_PerPointEnergy[p])+"\n";
			}
		}
	}
	return retStr;
}
