#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"

#include "ModelDatabase.h"
#include "ShapeTriangleSoup.h"

#ifndef __BindingEnergy_H
#define __BindingEnergy_H

#include "Interactor.h"
#include "InteractorPose.h"
#include "Binding.h"
#include "IKSolver.h"
#include "IntersectionGrid.h"
#include "PosePriorGauss.h"

class BindingManager;

/**
 * Energy visualization (?)
 */
enum EnergyVisType
{
	ENVIS_None,
	ENVIS_AngleAll,
	ENVIS_AngleSymmetry,
	ENVIS_Angle1,
	ENVIS_Angle2,
	ENVIS_Angle3,
	ENVIS_EndEffectorDist,
	ENVIS_ContactFeatures,
	ENVIS_Unassigned,
	ENVIS_PotentialTotalError,	// used by genetic algorithm to decide which end effectors to move...
	ENVIS_Intersection,
	ENVIS_ContactPairwise
};

/**
 * Storage for energy
 */
struct EnergyInfo
{
	enum PruneReasons
	{
		PRUNE_None,
		PRUNE_InvalidPairwiseDists,
		PRUNE_IKNotConverged
	};
	EnergyInfo();
	double m_fTotalEnergy;
	double m_fEndEffectorsEnergy;
	double m_fPosePriorEnergy;
	double m_fSymmetryEnergy; 
	double m_fFeatPointsEnergy;
	double m_fUnassignedPointsEnergy;
	double m_fIntersectionEnergy;
	double m_fContactPairwiseDistsEnergy;
	double m_fSymmPlaneEnergy;
	double m_fCoMPenalty;
	
	double m_fGridOverlapMinEnergy; // approximate pose prior + feat points energy
	
	int m_Flags2Fill;
	EnergyVisType m_EnergyVis;
	std::vector<double> m_PerPointEnergy;
	
	PruneReasons m_PruneReasons;
	
	// additional info for pose prior
	int m_LinkID;
	int m_SymmetricLink;
};



struct EnergyParams
{
	enum EnergyFlags
	{
		E_Invalid		= 0,
		E_Isect  		= 1<<1, // NOT IMPLEMENTED
		E_Proximity		= 1<<2,
		E_Stability		= 1<<3, // NOT IMPLEMENTED
		E_PosePrior		= 1<<4,
		E_Muscles		= 1<<5,	// NOT IMPLEMENTED
		E_Approximate   = 1<<6,
	};

	EnergyParams();
	
	bool m_bPointStatsConditionedOnPoseID;
	
	virtual bool IsEnergySet(EnergyFlags flag) const;
	virtual void ClearEnergyFlags();
	virtual void AddEnergyFlag(EnergyFlags flag);
	static EnergyFlags Str2EnergyFlag(const VKString & str);

	double m_wEndEffectorDist;	// same as IK
	double m_wPosePrior;		// same as IK
	double m_wSymmetryPrior;	// same as IK
	double m_wFeatures;
	double m_wUnassigned;
	double m_wGround;
	double m_wContactPairwise;
	double m_wSymmPlane;
	double m_wIsectPenalty;
	double m_wCoMPenalty;
	
	VKString m_sRegionStats;
	
	double m_fMaxContactPenalty;
	double m_fMaxGroundPenalty;

	int m_iEnergyFlags;
};

class ReflectionPlane
{
	public:
	
		R3Plane m_Plane;
		double m_fRadius;
		double m_fValue;
		double m_fRadialValue;
		double m_fProximityValue;
		double m_fMirrorValue;
		double m_fWeight;
};

/**
 * Energy computation
 */
class BindingEnergy
{
	public:
		BindingEnergy(Interactor * actor, Interactor * object);
		virtual void ReadPlanes(const VKString & planesFile);
		static void ReadPlanes(const VKString & planesFile, std::vector<ReflectionPlane> & planes);
//		virtual void ReadCoMDistribution(const VKString & comFile);
		virtual void SetPosePrior(IKSolver * solver, PosePrior * prior);
		virtual void SetIntersectionGrid(IntersectionGrid * grid = NULL);	// can create basic intersection grid w/o params
		virtual double ComputeBindingEnergy(const Binding * binding, EnergyInfo & energy) const;
		virtual double ComputeEnergy(const PntCorrsPose * posepts, EnergyInfo & energy) const;
		static int FindBestPlaneOfReflection(std::vector<ReflectionPlane> & planes, const PntCorrsPose * posepts,
											 IKSolver * solver, std::vector<int> & rois, double & energy);
		virtual double ComputeReflPlaneEnergy(const PntCorrsPose * posepts, EnergyInfo & energy) const;
		virtual double ComputeCoMPenalty(const PntCorrsPose * posepts, int planeID, EnergyInfo & energy) const;
		virtual double ReflPenaltyEnergyIncrease(PntCorrsPose * posepts, int regionid, const TargetPrimitive & match) const;
//		virtual void ApproximateMinEnergy(const PntCorrsPose * posepts, EnergyInfo & energy,
//										  std::vector<int> * regions = NULL, std::vector<TargetPrimitive> * matches = NULL) const;
		virtual PntCorrsPose * ApproximateMinEnergy(const PntCorrsPose * posepts, EnergyInfo & energy) const;
	
		virtual TargetPrimitive ApproximateMinEnergyForPartMatch(PntCorrsPose * posepts, const PntCorrsPose * cond,
																 int match, double & bestFeat, double & bestPose,
																 const R3Point & refPos, const R3Affine & w2wg) const;

		virtual bool IsValidAddition(PntCorrsPose * posepts, int regionid, const TargetPrimitive & match) const;
	
		virtual IKSolver * GetSolver();
		virtual IntersectionGrid * GetIntersectionGrid();
		virtual void SetParams(const EnergyParams & params, double origRotation);
		virtual void ReadRegionStats(const VKString & filename);
	
		virtual void OffsetGroundProb(double offset, int regionID);	// these two are hacky functions for the hybrid method
		virtual void OffsetUnassignedProb(double offset, int regionID); // they might remain unused in the final version

		// energy flags
		virtual bool ClassifiedRegion(int regionID) const;
		virtual double SurfaceProbability(int poseid, int regionID) const;
		virtual double GroundProbability(int poseid, int regionID) const;
		virtual double UnassignedProbability(int poseid, int regionID) const;
	
		virtual void FillPairwiseEnergy(const PntCorrsPose * posepts, EnergyInfo & energy) const;
	
		virtual double ContactPenalty(int poseid, int regionID, const TargetPrimitive & prim,
									  const ShapePointCloud * objpnts, const ShapeValue * objvals) const;
		virtual double UnassignedPenalty(int poseid, int regionID) const;
	
		static VKString GetIKOnlyEnergyStr(PntCorrsPose * pose, IKSolver * solver, PosePrior * prior, bool detailed);
		virtual VKString GetEnergyStr(BindingManager * binding, int poseid = -1, bool detailed = false) const;
		virtual VKString GetEnergyStr(const PntCorrsPose * pose, bool detailed = false) const;

	
//	protected:
		Interactor * m_Interactors[2];
		IKSolver * m_pSolver;		// skeleton
		PosePrior * m_pPrior;	// pose prior
		std::vector<bool> m_ClassifiedRegion;
	
		EnergyParams m_Params;
	
		int m_iNPoses;
		std::vector<int> m_RegionNGround;
		std::vector<int> m_RegionNSurface;
		
		std::vector<double> m_OffsetGroundProb;
		std::vector<double> m_OffsetUnassignedProb;

		std::vector<PosePointGauss * > m_PairwiseDistDistr;

		IntersectionGrid * m_pIntersectionGrid;
	
		std::vector<std::vector<int> > m_Region2Pnts;
		std::vector<std::vector<double> > m_PntPenalty;
	
		std::vector<ReflectionPlane> m_ReflectionPlanes;
};

#endif



