#include "Binding.h"
#include "Interactor.h"
#include "BindingManager.h"

#include "ShapeValue.h"
#include "TimeProfiler.h"
#include "ShapePointCloud.h"
#include "PntCorrsPose.h"

#ifdef INCLUDE_BULLET
	#include "btBulletDynamicsCommon.h"
#endif
//////////////////////////////
//////  PhysicsConstants
//////////////////////////////
PhysicsConstants::PhysicsConstants()
{
	m_SurfaceMargin = 0;
	m_fActorWeight = 1.;
	m_fObjectWeight = 1.;
	m_fGravity = 10.;
	m_fEulerStepSec = 1./100.;

	double fric = .5;
	m_afFriction[FRIC_GroundStatic] = fric;
	m_afFriction[FRIC_GroundRolling] = fric;
	m_afFriction[FRIC_ActorStatic] = fric;
	m_afFriction[FRIC_ActorRolling] = fric;
	m_afFriction[FRIC_ObjectStatic] = fric;
	m_afFriction[FRIC_ObjectRolling] = fric;
}


//////////////////////////////
//////  BindingManager
//////////////////////////////
BindingAlgParams::BindingAlgParams()
{
	// optimization
	m_iOptFlags = OPT_RandomSeeds;
	m_iOptPoseID = -1;
	m_iNSamples = 100;
	m_iNBest = 5;
	
	// physics
	m_iCurrentPhysicsBinding = -1;

	m_fCandidateValueCutoff = 0.2;
	m_fCandidateDistThresh = 0.2;
	m_bDisableIK = false;
	m_iExternalFlags = 0;
	m_bDoNotMap2SamePnt = true;	// check for conflicting assingments...
}

bool BindingAlgParams::IsOptimizationSet(OptFlags flag) const
{
	return (m_iOptFlags & flag)!=0;
}

void BindingAlgParams::ClearOptimizationFlags()
{
	m_iOptFlags = 0;
}

void BindingAlgParams::AddOptimizationFlag(OptFlags flag)
{
	m_iOptFlags |= flag;
}

BindingAlgParams::OptFlags BindingAlgParams::Str2OptFlag(const VKString & str)
{
	if (str=="RandomSeeds")
		return OPT_RandomSeeds;
	else if (str=="Physics")
		return OPT_Physics;
	else if (str=="PriorityDrivenSearch")
		return OPT_PriorityDrivenSearch;
	else if (str=="GreedySearch")
		return OPT_PriorityDrivenSearch;
	else if (str=="DepthFirstSearch")
		return OPT_DepthFirstSearch;
	else if (str=="GeneticSearch")
		return OPT_GeneticSearch;
	else if (str=="BranchAndBound")
		return OPT_BranchAndBound;
	else if (str=="PoseFirstSearch")
		return OPT_PoseFirstSearch;
	else if (str=="BestContacts")
		return OPT_BestContacts;
	else if (str=="External")
		return OPT_External;
	else if (str=="Hybrid")
		return OPT_Hybrid;
	else
		return OPT_Invalid;
}




//////////////////////////////
//////  BindingManager
//////////////////////////////
BindingManager::BindingManager(Interactor * actor, Interactor * object)
{
	m_Interactors[Interactor::ACT_Actor] = actor;
	m_Interactors[Interactor::ACT_Object] = object;
	
#ifdef INCLUDE_BULLET
	m_DynamicsWorld = NULL;
	m_PhysicsConstants.m_SurfaceMargin = 0.06;
#endif

	m_pEnergy = new BindingEnergy(actor, object);
}

void BindingManager::ReadRegionNames(const VKString & regionsFile)
{
	bool ok;
	std::ifstream textStream(regionsFile.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] File does not exist: "<<regionsFile.c_str()<<std::endl;
	assert(textStream.is_open());
	while (!textStream.eof())
	{
		VKString strLine = "";
		strLine.readLine(textStream);
		if (strLine!="")
		{
			if (strLine.startsWith("#"))
				continue;
			
			VKStringList splitLine = strLine.split(" ");
			assert(splitLine.count()>=3);
			m_sRegionNames.push_back(splitLine[0]);
		}
	}
	textStream.close();
}

const VKStringList & BindingManager::RegionNames() const
{
	return m_sRegionNames;
}

int BindingManager::NBindings() const
{
	return (int)m_Bindings.size();
}

const Binding * BindingManager::GetBinding(int bndID) const
{
	assert(bndID>=0 && bndID < (int)m_Bindings.size());
	return m_Bindings[bndID];
}

Binding * BindingManager::GetBinding(int bndID) 
{
	assert(bndID>=0 && bndID < (int)m_Bindings.size());
	return m_Bindings[bndID];
}

const Interactor * BindingManager::GetActor(int id) const
{
	return m_Interactors[id];
}

BindingEnergy * BindingManager::GetEnergy()
{
	return m_pEnergy;
}

void BindingManager::WriteBindings(const VKString & directory) const
{
	FileManagement::CreateDirsOnPath(directory.c_str());
	// write some debug data
	{
		int fid;
		double b1,b2,b3;
		const ShapeMesh * mesh = NULL;
		std::ofstream textStream((directory+"/candidate.rcorrs").c_str());
		assert(textStream.is_open());

		textStream<<"C2FLevels "<<m_CandidateCorrs.size()<<"\n";
		int nregions = 0;
		if (m_CandidateCorrs.size()>0)
			nregions =(int) m_CandidateCorrs[0].size();
		textStream<<"Regions "<<nregions<<"\n";
		for (int c2f=0; c2f<(int)m_CandidateCorrs.size(); c2f++)
		for (int r=0; r<(int)m_CandidateCorrs[c2f].size(); r++)
		{
			assert(r<m_sRegionNames.count());
			textStream<<"Region "<<r<<" "<<m_sRegionNames[r].c_str()<<" NPnts "<<m_CandidateCorrs[c2f][r].size()<<"\n";
			for (int p=0; p<(int)m_CandidateCorrs[c2f][r].size(); p++)
			{
				const ShapePoint & pnt = m_CandidateCorrs[c2f][r][p];
				if (pnt.GetPntType()==ShapePoint::SHAPE_POINT_MESHLESS)
				{
					fid = -1;
					b1 = -1;
					b2 = -1;
					b3 = -1;
				}
				else
				{
					const ShapeMesh * mesh = pnt.GetMeshGeometry();
					if (mesh->NFaces()>0)
					{
						pnt.GetBarycentricCoords(&fid, &b1, &b2, &mesh);
						b3 = 1.-b1-b2;
					}
					else
					{
						fid = pnt.VertexID();
						b1=0; b2=0; b3=0;
					}
				}
				R3Point pos = pnt.GetPosition();
				textStream<<fid<<" "<<b1<<" "<<b2<<" "<<b3<<" "<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<"\n";
			}
		}
		
		textStream.close();
	}
	
	// write final bindings
	FILE * file = fopen((directory+"/bindings.bin").c_str(), "wb");
	assert(file!=NULL);
	int nbindings = (int)m_Bindings.size();
	fwrite(&nbindings, sizeof(int), 1, file);
	for (int i=0; i<nbindings; i++)
		m_Bindings[i]->WriteBinding(file);
	fclose(file);
	
	if (m_sAlgorithmInfo!="")
	{
		std::ofstream textStream((directory+"/info.txt").c_str());
		assert(textStream.is_open());
		textStream<<m_sAlgorithmInfo.c_str();
		textStream.close();
	}
	
//	// write gt corrs
//	if (m_Interactors[Interactor::ACT_Actor]->GetActorType()==Interactor::TYPE_PntCorrs)
//	{
//		std::ofstream textStream((directory+"/bindings.ascii").c_str());
//		assert(textStream.is_open());
//		int nbindings = (int)m_Bindings.size();
//		textStream<<"NPoses "<<nbindings<<"\n";
//		
//		for (int i=0; i<nbindings; i++)
//		{
//			const InteractorPose * actPose = m_Bindings[i]->GetPose(Interactor::ACT_Actor);
//			assert(actPose->GetPoseType() == InteractorPose::TYPE_PntCorrsPose);
//			const PntCorrsPose * actCorrs = (const PntCorrsPose*)actPose;
//			int nregions = (int)actCorrs->NMatches();
//			textStream<<"Pose "<<i<<" Regions "<<nregions<<"\n";
//			int fid;
//			double b1, b2,b3;
//			const ShapeMesh * mesh;
//			for (int j=0; j<nregions; j++)
//			{
//				int rid = actCorrs->m_RegionID[j];
//				textStream<<"Region "<<rid<<" "<<m_sRegionNames[rid].c_str()<<" pnts 1\n";
//				
//				R3Point pos = actCorrs->Prim(j).GetPosition();
//				if (actCorrs->Prim(j).IsType(TargetPrimitive::PRIM_Ground))
//					textStream<<(-1)<<" "<<0<<" "<<0<<" "<<0<<" "<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<"\n";
//				else
//				{
//					assert(actCorrs->Prim(j).IsType(TargetPrimitive::PRIM_Point));
//					actCorrs->Prim(j).GetPoint().GetBarycentricCoords(&fid, &b1, &b2, &mesh);
//					b3 = 1.-b1-b2;
//					textStream<<fid<<" "<<b1<<" "<<b2<<" "<<(1.-b1-b2)<<" "<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<"\n";
//				}
//			}
//		}
//		textStream.close();
//	}
	
	WriteTimer(directory+"/timing.txt");
}

void BindingManager::AddBinding(Binding * binding)
{
	m_Bindings.push_back(binding);
}

void BindingManager::ReadGTPose(const VKString & gtpose, const ShapeMesh* mesh, const VKString & canonicalXForm)
{
	std::vector<PntCorrsPose*> poses;
	if (mesh==NULL && m_Interactors[Interactor::ACT_Object]!=NULL)
		mesh = m_Interactors[Interactor::ACT_Object]->GetMesh(m_Interactors[Interactor::ACT_Object]->GetEnumeratedPose(0));
	PntCorrsPose::CreatePosesFromGTFile(gtpose, mesh, canonicalXForm, poses);
	for (int i=0; i<(int)poses.size(); i++)
	{
		Binding * bind = new Binding(m_Interactors[Interactor::ACT_Actor], m_Interactors[Interactor::ACT_Object]);
		assert(poses[i]!=NULL);
		bind->SetPose(Interactor::ACT_Actor, poses[i]);
		if (m_Interactors[Interactor::ACT_Object]!=NULL)
		{
			const InteractorPose * objpose = m_Interactors[Interactor::ACT_Object]->GetEnumeratedPose(0);
			R3Vector objTrans = R3Vector(0, 0, -m_Interactors[Interactor::ACT_Object]->Ground(objpose));
			bind->SetPose(Interactor::ACT_Object, objpose);
			bind->SetXForm(Interactor::ACT_Object, new RigidAlign(objTrans, R3Vector(0,0,1), 0));
		}
		m_Bindings.push_back(bind);
	}
}

BindingManager * BindingManager::ReadBindings(const VKString & directory, Interactor * actor, Interactor * object)
{
	BindingManager * bindings = new BindingManager(actor, object);
	FILE * file = fopen((directory+"/bindings.bin").c_str(), "rb");
	if (file==NULL)
		std::cout<<"[ERROR] Cannot find: "<<(directory+"/bindings.bin").c_str()<<std::endl;
	assert(file!=NULL);
	int nbindings;
	fread(&nbindings, sizeof(int), 1, file);
	for (int i=0; i<nbindings; i++)
		bindings->m_Bindings.push_back(Binding::ReadBinding(file, actor, object));
	fclose(file);
	
	// read some debug data
	{
		int fid;
		double b1,b2,b3, x, y, z;
		const ShapeMesh * mesh = NULL;
		std::ifstream textStream((directory+"/candidate.rcorrs").c_str());
		VKString tempStr;
		int nregions, c2f_levels, npnts, test;
		if (textStream.is_open())
		{
			tempStr.readToken(textStream);
			if (tempStr=="Regions")
				textStream>>nregions;
			else if (tempStr=="C2FLevels")
			{
				textStream>>c2f_levels;
				tempStr.readToken(textStream);
				textStream>>nregions;
				assert(tempStr=="Regions");
			}
			
			for (int c2f=0; c2f<c2f_levels; c2f++)
			{
				bindings->m_CandidateCorrs.push_back(std::vector<std::vector<ShapePoint> >());
				for (int r=0; r<nregions; r++)
				{
					bindings->m_CandidateCorrs[c2f].push_back(std::vector<ShapePoint>());
					assert(!textStream.eof());
					tempStr.readToken(textStream);		assert(tempStr=="Region");
					textStream>>test;					assert(test==r);
					//assert(r<bindings->m_sRegionNames.count());
					tempStr.readToken(textStream);		//assert(tempStr==bindings->m_sRegionNames[r]);
					tempStr.readToken(textStream);		assert(tempStr=="NPnts");
					textStream>>npnts;
					for (int p=0; p<npnts; p++)
					{
						assert(object->NumEnumeratedPoses()==1);
						const ShapeMesh * mesh = object->GetMesh(object->GetEnumeratedPose(0));
						textStream>>fid>>b1>>b2>>b3>>x>>y>>z;
						if (mesh->NFaces()>0)
							bindings->m_CandidateCorrs[c2f][r].push_back(ShapePoint(mesh, fid, b1, b2, R3Point(x, y, z)));
						else
							bindings->m_CandidateCorrs[c2f][r].push_back(ShapePoint(mesh, fid, R3Point(x, y, z)));
					}
				}
			}
			textStream.close();
		}
	}
	return bindings;
}


#ifdef INCLUDE_BULLET
void BindingManager::InitPhysics()
{
	///collision configuration contains default setup for memory, collision setup
	btCollisionConfiguration * collisions = new btDefaultCollisionConfiguration();
	//m_collisionConfiguration->setConvexConvexMultipointIterations();
	
	///use the default collision dispatcher. For parallel processing you can use a diffent dispatcher (see Extras/BulletMultiThreaded)
	btDispatcher * dispatcher = new	btCollisionDispatcher(collisions);
	
	//m_Broadphase = new btDbvtBroadphase();
	btBroadphaseInterface * broadphase = new btSimpleBroadphase();
	
	///the default constraint solver. For parallel processing you can use a different solver (see Extras/BulletMultiThreaded)
	btConstraintSolver * solver = new btSequentialImpulseConstraintSolver();
	
	m_DynamicsWorld = new btDiscreteDynamicsWorld(dispatcher, broadphase, solver, collisions);
	m_DynamicsWorld->setGravity(btVector3(0, 0, -m_PhysicsConstants.m_fGravity));
	
	///create a few basic rigid bodies
	//btBoxShape* groundShape = new btBoxShape(btVector3(btScalar(20.),btScalar(20.),btScalar(1.)));
	//groundShape->initializePolyhedralFeatures();
	//	btCollisionShape* groundShape = new btStaticPlaneShape(btVector3(0,1,0),50);
	btVector3 normal(0, 0, 1.);
	normal.normalize();
	btCollisionShape* groundShape = new btStaticPlaneShape(normal, 0);// A plane

	// ADD PLANE
	//m_CollisionShapes.push_back(groundShape);
	
	btTransform groundTransform;
	groundTransform.setIdentity();
	//groundTransform.setOrigin(btVector3(-10, -10, -3.));
	
	//We can also use DemoApplication::localCreateRigidBody, but for clarity it is provided here:
	//using motionstate is recommended, it provides interpolation capabilities, and only synchronizes 'active' objects
	btScalar mass(0.f);
	btVector3 localInertia(0,0,0);
	btDefaultMotionState* myMotionState = new btDefaultMotionState(groundTransform);
	btRigidBody::btRigidBodyConstructionInfo rbInfo(mass,myMotionState,groundShape,localInertia);
	btRigidBody* ground = new btRigidBody(rbInfo);
	ground->setFriction(m_PhysicsConstants.m_afFriction[PhysicsConstants::FRIC_GroundStatic]);
	ground->setRollingFriction(m_PhysicsConstants.m_afFriction[PhysicsConstants::FRIC_GroundRolling]);
	
	ground->setCollisionFlags(ground->getCollisionFlags()|btCollisionObject::CF_STATIC_OBJECT);
	
	//add the body to the dynamics world
	m_DynamicsWorld->addRigidBody(ground);	// ground

	// register algorithm
	btGImpactCollisionAlgorithm::registerAlgorithm(static_cast<btCollisionDispatcher *>(dispatcher));
}

void BindingManager::InitPhysicsConfig(int bindingID)
{
	m_iCurrentPhysicsBinding = bindingID;
	if (m_DynamicsWorld==NULL)
		InitPhysics();
	assert(m_DynamicsWorld!=NULL);
	// Cleanup scene (do NOT delete objects). Keep ground plane.
	while (m_DynamicsWorld->getNumCollisionObjects()>1)
	{
		btCollisionObject* obj = m_DynamicsWorld->getCollisionObjectArray()[1];
		m_DynamicsWorld->removeCollisionObject( obj );
		//delete obj;
	}

	while (m_DynamicsWorld->getNumConstraints()>0)
		m_DynamicsWorld->removeConstraint(m_DynamicsWorld->getConstraint(0));
	
	
	// add two actors
	Binding * b = GetBinding(bindingID);
	for (int i=0; i<2; i++)
	{
		if (m_Interactors[i]==NULL)
			continue;
		Interactor::InteractorID aid = (Interactor::InteractorID)i;
		RigidAlign * align = b->GetBindingXForm(aid)->SafeCastToRigidAlign();
		std::vector<btTypedConstraint *> joints;
		std::vector<btRigidBody*> bodies;
		m_Interactors[i]->FillRigidBodies(b->GetPose(aid), m_PhysicsConstants, align, bodies, joints);
		for (int j=0; j<(int)joints.size(); j++)
			m_DynamicsWorld->addConstraint(joints[j], true);
		for (int j=0; j<(int)bodies.size(); j++)
			m_DynamicsWorld->addRigidBody(bodies[j]);
	}
}

void BindingManager::ResetPhysics()
{
	m_iCurrentPhysicsBinding = -2;
}

void BindingManager::GetContactPoints(int bindingID, std::vector<btManifoldPoint> & pnts) const
{
	assert(bindingID>=0 && bindingID < (int)m_Bindings.size());
	if (m_iCurrentPhysicsBinding!=bindingID)
		((BindingManager*)this)->InitPhysicsConfig(bindingID);
	
	int numManifolds = m_DynamicsWorld->getDispatcher()->getNumManifolds();
	for (int i=0;i<numManifolds;i++)
	{
		const btPersistentManifold* contactManifold = m_DynamicsWorld->getDispatcher()->getManifoldByIndexInternal(i);
		const btCollisionObject* obA = static_cast<const btCollisionObject*>(contactManifold->getBody0());
		const btCollisionObject* obB = static_cast<const btCollisionObject*>(contactManifold->getBody1());
		
		int numContacts = contactManifold->getNumContacts();
		for (int j=0;j<numContacts;j++)
			pnts.push_back(contactManifold->getContactPoint(j));
		
		//you can un-comment out this line, and then all points are removed
		//contactManifold->clearManifold();
	}

}

void BindingManager::PrintPhysicsStatus(int bindingID) const
{
	assert(bindingID>=0 && bindingID < (int)m_Bindings.size());
	if (m_iCurrentPhysicsBinding!=bindingID)
		((BindingManager*)this)->InitPhysicsConfig(bindingID);

	int numManifolds = m_DynamicsWorld->getDispatcher()->getNumManifolds();
	std::cout<<"Manifolds: "<<numManifolds<<std::endl;
	for (int i=0;i<numManifolds;i++)
	{
		const btPersistentManifold* contactManifold = m_DynamicsWorld->getDispatcher()->getManifoldByIndexInternal(i);
		const btCollisionObject* obA = static_cast<const btCollisionObject*>(contactManifold->getBody0());
		const btCollisionObject* obB = static_cast<const btCollisionObject*>(contactManifold->getBody1());
		
		int numContacts = contactManifold->getNumContacts();
		for (int j=0;j<numContacts;j++)
		{
			const btManifoldPoint& pt = contactManifold->getContactPoint(j);
			std::cout<<"   M"<<i<<" p"<<j<<": D="<<pt.m_distance1<<" I="<<pt.m_appliedImpulse<<std::endl;
		}
	}
}

void BindingManager::SimulatePhysics(int bindingID, double dt)
{

	if (dt > 20)
	{
		std::cout<<"[WARNING] Simulation is over 20s. Possibly lagging. Exiting..."<<std::endl;
		return;
	}
	// simulate physics
	assert(bindingID>=0 && bindingID < (int)m_Bindings.size());
	if (m_iCurrentPhysicsBinding!=bindingID)
		InitPhysicsConfig(bindingID);
	
	assert(m_DynamicsWorld!=NULL);

	int niter = (int)(dt / m_PhysicsConstants.m_fEulerStepSec);
	double dt_remained = dt - niter * m_PhysicsConstants.m_fEulerStepSec;
	for (int i=0; i<niter; i++)
		m_DynamicsWorld->stepSimulation((float)m_PhysicsConstants.m_fEulerStepSec);
	if (dt_remained >= m_PhysicsConstants.m_fEulerStepSec)
		std::cout<<"[WARNING] Too big step: "<<dt_remained<<" vs "<<m_PhysicsConstants.m_fEulerStepSec<<std::endl;
	if (dt_remained>0)
		m_DynamicsWorld->stepSimulation((float)dt_remained);
 
	// copy state
	//assert(m_DynamicsWorld->getNumCollisionObjects()==3);	// ground plane + object + actor
	
	Binding * bind = GetBinding(bindingID);
	
	// change binding x-forms
	for (int i=0; i<2; i++)
	{
		if (m_Interactors[i]==NULL)
			continue;
		Interactor::InteractorID aid = (Interactor::InteractorID)i;
		RigidAlign * align = bind->GetBindingXForm(aid)->SafeCastToRigidAlign();
		assert(align!=NULL);
		m_Interactors[i]->UpdatePoseFromPhysics((InteractorPose*)bind->GetPose(aid), align);
	}
}

btDynamicsWorld * BindingManager::GetPhysics()
{
	return m_DynamicsWorld;
}

void BindingManager::ExitPhysics()
{
	// TODO...
//	//remove the rigidbodies from the dynamics world and delete them
//	int i;
//	for (i=m_dynamicsWorld->getNumCollisionObjects()-1; i>=0 ;i--)
//	{
//		btCollisionObject* obj = m_dynamicsWorld->getCollisionObjectArray()[i];
//		btRigidBody* body = btRigidBody::upcast(obj);
//		if (body && body->getMotionState())
//		{
//			delete body->getMotionState();
//		}
//		m_dynamicsWorld->removeCollisionObject( obj );
//		delete obj;
//	}
//	
//	//delete collision shapes
//	for (int j=0;j<m_collisionShapes.size();j++)
//	{
//		btCollisionShape* shape = m_collisionShapes[j];
//		delete shape;
//	}
//	m_collisionShapes.clear();
//	
//	delete m_dynamicsWorld;
//	
//	delete m_solver;
//	
//	delete m_broadphase;
//	
//	delete m_dispatcher;
//	
//	delete m_collisionConfiguration;
}
#endif


void BindingManager::WriteAllPosesAsCorrs(const VKString & outGTFile) const
{
	assert(false);
}


void BindingManager::WriteTimer(const VKString & outputFile) const
{
	FileManagement::CreateDirsOnPathToFile(outputFile.c_str());
	std::ofstream textStream(outputFile.c_str());
	assert(textStream.is_open());
	textStream<<((TimeProfiler&)m_Timer).profile().c_str();
	textStream.close();
}








