#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"

#include "ModelDatabase.h"
#include "ShapeTriangleSoup.h"
#include "TimeProfiler.h"
#include "BindingEnergy.h"
#include "PntCorrsPose.h"

#ifndef __BindingManager_H
#define __BindingManager_H

#include "Interactor.h"
#include "Binding.h"

#ifdef INCLUDE_BULLET
	#include "LinearMath/btAlignedObjectArray.h"
	#include "btBulletDynamicsCommon.h"
	#include "BulletCollision/Gimpact/btGImpactShape.h"
	#include "BulletCollision/Gimpact/btGImpactCollisionAlgorithm.h"
#endif


class RandomBindings;

// bullet physics
class btBroadphaseInterface;
class btCollisionShape;
class btOverlappingPairCache;
class btCollisionDispatcher;
class btConstraintSolver;
struct btCollisionAlgorithmCreateFunc;
class btDefaultCollisionConfiguration;
class	btCollisionShape;
class	btDynamicsWorld;
class	btRigidBody;
class	btTypedConstraint;

class PhysicsConstants
{
	public:
		PhysicsConstants();
	
		double m_SurfaceMargin;
	
		double m_fActorWeight;
		double m_fObjectWeight;
		double m_fGravity;
		double m_fEulerStepSec;
		enum FrictionType
		{
			FRIC_GroundStatic	 = 0,
			FRIC_GroundRolling	 = 1,
			FRIC_ActorStatic	 = 2,
			FRIC_ActorRolling	 = 3,
			FRIC_ObjectStatic	 = 4,
			FRIC_ObjectRolling	 = 5,
		};
		double m_afFriction[6];
};


/**
 * Structures used in pose search procedure
 */
struct CandPoint
{
	TargetPrimitive m_Primitive;
	double m_fProbability;
};

class PermutationsSharedData;
struct CorrsPermutation
{
	CorrsPermutation(const CorrsPermutation & copy);
	CorrsPermutation(const PermutationsSharedData * sharedData);
	virtual ~CorrsPermutation();
	bool operator < (const CorrsPermutation & perm) const;	// to use in a map
	
	EnergyInfo m_Energy;
	PntCorrsPose * m_pPose;
	int m_iNLocalRegions;
	int * m_aPermutation;
	PermutationsSharedData * m_pSharedData;
};

struct PermutationsSharedData
{
	std::map<CorrsPermutation, bool> m_VisitedPermutations;
	std::vector<int> m_Local2GlobalRegionID;
	std::vector<std::vector<CandPoint> > m_LocalRegion2Candidates;
};


struct BindingAlgParams
{
	enum OptFlags
	{
		OPT_Invalid					= 0,
		OPT_RandomSeeds 			= 1<<0,	// INVALID
		OPT_Physics 				= 1<<1,	// INVALID
		OPT_PriorityDrivenSearch	= 1<<2,	// INVALID
		OPT_GreedySearch			= 1<<3,	// INVALID
		OPT_DepthFirstSearch		= 1<<4,	// INVALID
		OPT_GeneticSearch			= 1<<5,
		OPT_BranchAndBound			= 1<<6,
		OPT_BestContacts			= 1<<7,
		OPT_External				= 1<<8,
		OPT_Hybrid					= 1<<9,
		OPT_PoseFirstSearch			= 1<<10	// read file in pose format
	};
	
	BindingAlgParams();
	
	// optimization flags
	bool IsOptimizationSet(OptFlags flag) const;
	void ClearOptimizationFlags();
	void AddOptimizationFlag(OptFlags flag);
	static OptFlags Str2OptFlag(const VKString & str);
		
	int m_iOptFlags;
	int m_iOptPoseID;
	int m_iNSamples;				// number of samples (for randomized search)
	int m_iCurrentPhysicsBinding;	// physics-based search

	int m_iInitPopulation;
	int m_iNBest;
	
	double m_fCandidateValueCutoff;
	double m_fCandidateDistThresh;
	
	VKStringList m_sReadPoseFile;
	enum ExtFlags
	{
		EXT_Input			 = 1<<0,
		EXT_SolveIK			 = 1<<1,
		EXT_Project2Corrs	 = 1<<2
	};
	int m_iExternalFlags;
	
	bool m_bDisableIK;
	bool m_bDoNotMap2SamePnt;
	int m_iCoarse2Fine;
	long long m_iNMaxPermutations;
	double m_fPntRadiusStep;
};

/**
 * Manage a set of potential bindings...
 */
class BindingManager
{
	public:
		enum PointType
		{
			PNT_Null    = 0,
			PNT_Isect   = 1,
			PNT_Touch   = 2,
			PNT_TouchAR = 3,	// stronger than above
		};
	
		struct PointInfo
		{
			PointType m_PntType;
			double m_Dist;
			R3Point m_Pos;
			R3Point m_ClosestPoint;
			int m_iClosestFace;
		};
		
		BindingManager(Interactor * actor, Interactor * object);
		virtual void ReadRegionNames(const VKString & regionsFile);
		virtual const VKStringList & RegionNames() const;
	
	// store bindings
		virtual int NBindings() const;
		virtual void ReadGTPose(const VKString & gtpose, const ShapeMesh* mesh = NULL, const VKString & canonicalXForm = "none");
		virtual void AddBinding(Binding * binding);
		virtual const Binding * GetBinding(int bndID) const;
		virtual Binding * GetBinding(int bndID);
		virtual void ComputeBindings(const BindingAlgParams & params);	// main algorithm
		virtual const Interactor * GetActor(int id) const;
		virtual void WriteBindings(const VKString & directory) const;
		static BindingManager * ReadBindings(const VKString & directory, Interactor * actor, Interactor * object);

	// write results
		virtual void WriteAllPosesAsCorrs(const VKString & outGTFile) const;	// see gt format

	// energy
		virtual BindingEnergy * GetEnergy();

	// physics
#ifdef INCLUDE_BULLET
		virtual void ResetPhysics();
		virtual void SimulatePhysics(int bindingID, double dt);
		virtual btDynamicsWorld * GetPhysics();
		virtual void PrintPhysicsStatus(int bindingID) const;
		virtual void GetContactPoints(int bindingID, std::vector<btManifoldPoint> & pnts) const;
#endif

	// DEBUG: candidates
		std::vector<std::vector<std::vector<ShapePoint> > > m_CandidateCorrs;
		std::vector<std::vector<std::vector<double> > > m_CandidateCorrValues;
		virtual void WriteTimer(const VKString & outputFile) const;
	
	protected:
		VKStringList m_sRegionNames;
	
	// object / actor data
		Interactor * m_Interactors[2];
	
	// Store bindings
		std::vector<Binding*> m_Bindings;

	// Energy
		BindingEnergy * m_pEnergy;
			
	// Optimization
		virtual void ComputeBindingsGenetic();
		virtual void ComputeBindingsBranchAndBound();
		virtual void ComputeBindingsBranchAndBound2();	// the actual algorithm (not B&B)
		virtual void ComputeBindingsBestContact();
		virtual void ComputeBindingsExternal();
		virtual void AddGlobalMatchingData();
		virtual void ComputeBindingsPoseFirstSearch();
	
		virtual PntCorrsPose * LocalSearchStep(PntCorrsPose * pose, double nhdSearch = 1.);
	
		virtual void AddClosestMatch(PntCorrsPose * pose, int region, double threshold);
	
		virtual double ComputeEnergy(CorrsPermutation * permutation);
		virtual int SelectFromPDF(const std::vector<double> & values);
		virtual void FillPose(CorrsPermutation * permutation);
		virtual bool RandomMutation(CorrsPermutation * permutation, int locRegion, double searchRadius, bool useProb);
		virtual bool ValidPointAssignment(CorrsPermutation * permutation, int updatedRegion);
		virtual PermutationsSharedData * FindCandidates(double valThreshold, double distThershold, int c2f_level = 0,
														CorrsPermutation * candidate = NULL, double candDistThresh = -1);
		virtual PermutationsSharedData * VotedCandidates(double valThreshold, double distThershold);
	
		virtual void ClusterPoses(std::vector<std::vector<Sortable> > & inputPopulation,
								  std::vector<std::vector<Sortable> > & outputPopulation);
		virtual bool IsSimilar(PntCorrsPose * pose1, PntCorrsPose * pose2, double dist) const;
		
		// branch and bound
		virtual bool NextMutation(CorrsPermutation * permutation);
		virtual bool NextBranch(CorrsPermutation * permutation);
		virtual long long MutationID(const CorrsPermutation * permutation) const ;
		virtual long long NMutations(const PermutationsSharedData * sharedData) const ;
		virtual bool IsComplete(const CorrsPermutation * permutation) const;
	
#ifdef INCLUDE_BULLET
		virtual void InitPhysicsConfig(int bindingID);
		virtual void InitPhysics();
		virtual void ExitPhysics();

		btDynamicsWorld* m_DynamicsWorld;
#endif
		PhysicsConstants m_PhysicsConstants;
		BindingAlgParams m_AlgParams;
		
	// algorithm timing
		TimeProfiler m_Timer;
		VKString m_sAlgorithmInfo;
		
};


#endif

