#include "Binding.h"
#include "Interactor.h"
#include "BindingManager.h"

#include "TimeProfiler.h"

#include "ShapePointCloud.h"
#include "ShapeValue.h"
#include "ShapeValue.h"
#include "VkFunctions.h"
#include "IKSolver.h"
#include "PntCorrsPose.h"

#ifdef INCLUDE_BULLET
	#include "btBulletDynamicsCommon.h"
#endif

CorrsPermutation::CorrsPermutation(const PermutationsSharedData * sharedData)
{
	m_pSharedData = (PermutationsSharedData*)sharedData;
	m_iNLocalRegions = (int)m_pSharedData->m_Local2GlobalRegionID.size();
	m_aPermutation = new int[m_iNLocalRegions];
	for (int i=0; i<m_iNLocalRegions; i++)
		m_aPermutation[i] = 0;	// invalid
	m_Energy.m_fTotalEnergy = -1;
	m_Energy.m_fPosePriorEnergy = -1;
	m_Energy.m_fFeatPointsEnergy = -1;
	m_pPose = NULL;	// unknown
	
}

CorrsPermutation::CorrsPermutation(const CorrsPermutation & copy)
{
	m_iNLocalRegions = copy.m_iNLocalRegions;
	m_aPermutation = new int[m_iNLocalRegions];
	for (int i=0; i<m_iNLocalRegions; i++)
		m_aPermutation[i] = copy.m_aPermutation[i];
	m_Energy = copy.m_Energy;
	m_pPose = copy.m_pPose;
	m_pSharedData = copy.m_pSharedData;
}

CorrsPermutation::~CorrsPermutation()
{
	delete [] m_aPermutation;
}


bool CorrsPermutation::operator < (const CorrsPermutation & perm) const
{
	for (int i=0; i<m_iNLocalRegions; i++)
	{
		if (m_aPermutation[i] < perm.m_aPermutation[i])
			return true;
		else if (m_aPermutation[i] > perm.m_aPermutation[i])
			return false;
	}
	return false;
}

void BindingManager::ComputeBindings(const BindingAlgParams & params)
{
//	std::cout<<"Compute bindings...\n";
//	std::cout<<"   hybrid: "<<m_AlgParams.IsOptimizationSet(BindingAlgParams::OPT_Hybrid)<<std::endl;
//	std::cout<<"   genetic: "<<m_AlgParams.IsOptimizationSet(BindingAlgParams::OPT_GeneticSearch)<<std::endl;
//	std::cout<<"   bb: "<<m_AlgParams.IsOptimizationSet(BindingAlgParams::OPT_BranchAndBound)<<std::endl;
//	std::cout<<"   posefirst: "<<m_AlgParams.IsOptimizationSet(BindingAlgParams::OPT_PoseFirstSearch)<<std::endl;
//	std::cout<<"   bestcontact: "<<m_AlgParams.IsOptimizationSet(BindingAlgParams::OPT_BestContacts)<<std::endl;
//	std::cout<<"   external: "<<m_AlgParams.IsOptimizationSet(BindingAlgParams::OPT_External)<<std::endl;
	
	m_AlgParams = params;
	if (m_AlgParams.IsOptimizationSet(BindingAlgParams::OPT_GeneticSearch))
		ComputeBindingsGenetic();
	else if (m_AlgParams.IsOptimizationSet(BindingAlgParams::OPT_BranchAndBound))
		ComputeBindingsBranchAndBound2();
	else if (m_AlgParams.IsOptimizationSet(BindingAlgParams::OPT_PoseFirstSearch))
		ComputeBindingsPoseFirstSearch();
	else if (m_AlgParams.IsOptimizationSet(BindingAlgParams::OPT_BestContacts))
		ComputeBindingsBestContact();
	else if (m_AlgParams.IsOptimizationSet(BindingAlgParams::OPT_External))
		ComputeBindingsExternal();
	else if (m_AlgParams.IsOptimizationSet(BindingAlgParams::OPT_Hybrid))
	{
		AddGlobalMatchingData();
		ComputeBindingsBranchAndBound2();
	}
	else
		assert(false);
}


int BindingManager::SelectFromPDF(const std::vector<double> & pdf)
{
	double sum = 0;
	for (int i=0; i<(int)pdf.size(); i++)
		sum += pdf[i];
	
	double rnum = ((double)(rand()%10000))/10000.;
	assert(rnum>=0 && rnum<1.);
	double cdf = 0;
	for (int i=0; i<(int)pdf.size(); i++)
	{
		cdf += pdf[i] / sum;
		if (cdf>=rnum)
			return i;
	}
	return ((int)pdf.size())-1;
}

double BindingManager::ComputeEnergy(CorrsPermutation * permutation)
{
	assert(m_pEnergy!=NULL);
	// check if energy exists
	if (permutation->m_Energy.m_fTotalEnergy>=0)
		return permutation->m_Energy.m_fTotalEnergy;
	
	// check if there is a conflicting assignment
	if (m_AlgParams.m_bDoNotMap2SamePnt)
	{
		double threshold = 0.2;	// vk: const
		const ShapePointCloud * objpnts = NULL;
		const ShapeValue * objvals = NULL;
		m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
		assert(objpnts!=NULL & objvals!=NULL);

		PermutationsSharedData * shared = permutation->m_pSharedData;
		std::vector<R3Point> assignedPnts;
		for (int i=0; i<permutation->m_iNLocalRegions; i++)
		{
			if (permutation->m_aPermutation[i]>=0)
			{
				TargetPrimitive & prim = shared->m_LocalRegion2Candidates[i][permutation->m_aPermutation[i]].m_Primitive;
				if (prim.IsType(TargetPrimitive::PRIM_Point))
				{
					R3Point pos = prim.GetPosition();
					for (int p=0; p<(int)assignedPnts.size(); p++)
					{
						if (R3Distance(assignedPnts[p], pos) < threshold)
						{
							permutation->m_Energy.m_fTotalEnergy = RN_INFINITY;
							permutation->m_Energy.m_fFeatPointsEnergy = RN_INFINITY;
							return RN_INFINITY;
						}
					}
					assignedPnts.push_back(prim.GetPosition());
				}
			}
		}
	}
	
	// check if pose exists
//	if (permutation->m_pPose==NULL)
	FillPose(permutation);
	
	// compute energy for a pose
	m_pEnergy->ComputeEnergy(permutation->m_pPose, permutation->m_Energy);
	return permutation->m_Energy.m_fTotalEnergy;
}

PermutationsSharedData * BindingManager::VotedCandidates(double valThreshold, double distThreshold)
{
	assert(false);	// todo
	return NULL;
#if 0
	m_CandidateCorrs.clear();
	m_CandidateCorrValues.clear();
	std::vector<std::vector<std::vector<int> > > pntids;
	
	const ShapePointCloud * objpnts = NULL;
	const ShapeValue * objvals = NULL;
	m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
	assert(objpnts!=NULL & objvals!=NULL);

	const R3Kdtree<const PointInSet *> * kdTree = objpnts->GetKdTree();
	
	m_CandidateCorrs.push_back(std::vector<std::vector<ShapePoint> >());
	m_CandidateCorrValues.push_back(std::vector<std::vector<double> >());
	pntids.push_back(std::vector<std::vector<int> >());

	int c2f_level = 0;
	int nobjpoints = objpnts->NumPoints();		// npoints
	int nactorpoints = objvals->NDims() - 1;	// == nlabels

	IKSolver * solver = GetEnergy()->GetSolver();
	assert(solver!=NULL && solver->m_pPrior!=NULL);
	PosePrior * prior = solver->m_pPrior;

	PntCorrsPose * dummypose = new PntCorrsPose(m_Interactors[Interactor::ACT_Object]);
	double * perRegionVals = new double[objpnts->NumPoints() * nactorpoints];
	for (int i=0; i<objpnts->NumPoints() * nactorpoints; i++)
		perRegionVals[i] = 0;
	
	TimeProfiler timer;
	for (int r=0; r<nactorpoints; r++)	// for each contact point
	{
		// keep candidates for debugging
		m_CandidateCorrs[c2f_level].push_back(std::vector<ShapePoint>());
		m_CandidateCorrValues[c2f_level].push_back(std::vector<double>());
		pntids[c2f_level].push_back(std::vector<int>());
		
		// ignore if there is not training data
		if (!GetEnergy()->ClassifiedRegion(r))
			continue;
		
		double probSurface = GetEnergy()->SurfaceProbability(r);
		double probGround = GetEnergy()->GroundProbability(r);
		double probUnassigned = GetEnergy()->UnassignedProbability(r);
		
		if (probSurface>0)
		{
			std::vector<Sortable> srtPnts;
			for (int p=0; p<nobjpoints; p++)
			{
				timer.WriteProgress("prog", r*nobjpoints + p, nactorpoints*nobjpoints);
				double value = objvals->Value(objpnts->GetPoint(p), r);
				if (value > valThreshold)
				{
					srtPnts.push_back(Sortable(-value, NULL, p));
					
					
//					std::vector<std::vector<int> > candPntIDsForVals;
//					std::vector<std::vector<int> > candRegionIDs;
					for (double rot = -M_PI; rot < M_PI; rot += M_PI / 16.) // for each rotation
					{
						for (int poseid = 0; poseid < prior->NSuggestedInitPoses(); poseid++) // for each possible initial pose
						{
							prior->InitPose(poseid, dummypose);
							dummypose->Param(0) = 0;
							dummypose->Param(1) = 0;
							dummypose->Param(2) = 0;	// translation
							dummypose->Param(3) = rot;	// up-rotation
							solver->ForwardKinematics(dummypose);
							
							// set translation (from pose & rotation)
							assert(r < (int)solver->m_Pnts2Skeleton.size());
							R3Point jpos = dummypose->GetPosition(solver->m_Pnts2Skeleton[r]);
							R3Point trgpos = objpnts->PointPosition(p);
							dummypose->Param(0) = trgpos.X() - jpos.X();
							dummypose->Param(1) = trgpos.Y() - jpos.Y();
							dummypose->Param(2) = trgpos.Z() - jpos.Z();
							solver->ForwardKinematics(dummypose);
							
							// set contacts near limb
							int ngoodactors = 0;
							std::vector<int> pntIDForVals;
							std::vector<int> regionIDs;
							std::vector<double> weights;
							for (int r2=0; r2<nactorpoints; r2++)	// for all other contacts
							{
								double pr2 = GetEnergy()->SurfaceProbability(r2);
								if (r2==r || !GetEnergy()->ClassifiedRegion(r) || pr2==0)
									continue;
								
								assert(r2 < (int)solver->m_Pnts2Skeleton.size());
								R3Point jpos = dummypose->GetPosition(solver->m_Pnts2Skeleton[r2]);
								RNArray<const PointInSet*> nhd;
								kdTree->FindAll(jpos, 0, distThreshold, nhd);
								bool goodactor = false;
								for (int n=0; n<nhd.NEntries(); n++)
								{
									double weight = exp(-pow(R3Distance(jpos, objpnts->PointPosition(nhd[n]->id))/distThreshold, 2.));
									double value = objvals->Value(objpnts->GetPoint(nhd[n]->id), r2);
									if (value > valThreshold)
									{
										//perRegionVals[r2*objpnts->NumPoints() + nhd[n]->id] += weight*value;
										goodactor = true;
										pntIDForVals.push_back(nhd[n]->id);
										regionIDs.push_back(r2);
										weights.push_back(weight*value);
									}
								}
								ngoodactors += (int)goodactor;
							}
							for (int cp=0; cp < (int)pntIDForVals.size(); cp++)
								perRegionVals[regionIDs[cp]*objpnts->NumPoints() + pntIDForVals[cp]] += ngoodactors*weights[cp];
						}
					}
				}
			}
			std::sort(srtPnts.begin(), srtPnts.end());
			
			// add candidate points
			for (int p1=0; p1<(int)srtPnts.size(); p1++)
			{
				m_CandidateCorrs[c2f_level][r].push_back(objpnts->GetPoint(srtPnts[p1].id));
				m_CandidateCorrValues[c2f_level][r].push_back(0);
				pntids[c2f_level][r].push_back(srtPnts[p1].id);
			}
		}
	}
	std::cout<<std::endl;
	for (int r=0; r<m_CandidateCorrValues[c2f_level].size(); r++)
	for (int p=0; p<m_CandidateCorrValues[c2f_level][r].size(); p++)
		m_CandidateCorrValues[c2f_level][r][p] = perRegionVals[r*objpnts->NumPoints()+pntids[c2f_level][r][p]];
		
	delete dummypose;
	delete [] perRegionVals;
	assert(false);	// no return
#endif
	
}

PermutationsSharedData * BindingManager::FindCandidates(double valThreshold, double distThreshold, int c2f_level,
														CorrsPermutation * c2fcounter, double candDistThresh)
{
	assert(false);
	return NULL;
#if 0
	while (c2f_level>=(int)m_CandidateCorrs.size())
	{
		m_CandidateCorrs.push_back(std::vector<std::vector<ShapePoint> >());
		m_CandidateCorrValues.push_back(std::vector<std::vector<double> >());
	}
	m_CandidateCorrs[c2f_level].clear();
	m_CandidateCorrValues[c2f_level].clear();
	
	const ShapePointCloud * objpnts = NULL;
	const ShapeValue * objvals = NULL;
	m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
	assert(objpnts!=NULL & objvals!=NULL);

	int nobjpoints = objpnts->NumPoints();		// npoints
	int nactorpoints = objvals->NDims() - 1;	// == nlabels

	PermutationsSharedData * shared = new PermutationsSharedData();
	
	int localid = -1;
	for (int r=0; r<nactorpoints; r++)	// for each contact point
	{
		// keep candidates for debugging
		m_CandidateCorrs[c2f_level].push_back(std::vector<ShapePoint>());
		m_CandidateCorrValues[c2f_level].push_back(std::vector<double>());
		
		// ignore if there is not training data
		if (!GetEnergy()->ClassifiedRegion(r))
			continue;
		
		// add candidate contacts storage
		localid = (int)shared->m_Local2GlobalRegionID.size();
		shared->m_Local2GlobalRegionID.push_back(r);
		shared->m_LocalRegion2Candidates.push_back(std::vector<CandPoint>());
		
		double probSurface = GetEnergy()->SurfaceProbability(r);
		double probGround = GetEnergy()->GroundProbability(r);
		double probUnassigned = GetEnergy()->UnassignedProbability(r);

		CandPoint candidate;	// keep overwriting this
		assert(localid>=0 && localid <(int)shared->m_LocalRegion2Candidates.size());
		
		// see if there are some potential surface contact points
		if (probSurface>0)
		{
			std::vector<int> excludedcontacts;
			std::vector<Sortable> srtPnts;
			for (int p=0; p<nobjpoints; p++)
			{
				excludedcontacts.push_back(false);
				double value = objvals->Value(objpnts->GetPoint(p), r);
				bool c2fTest = true;
				if (c2fcounter!=NULL && candDistThresh>=0)
				{
					assert(c2fcounter->m_pPose != NULL);
					int jointid = GetEnergy()->GetSolver()->m_Pnts2Skeleton[r];
					R3Point jpos = c2fcounter->m_pPose->GetPosition(jointid);
					c2fTest = (R3Distance(objpnts->PointPosition(p), jpos) < candDistThresh);
				}
				if (value > valThreshold && c2fTest)
					srtPnts.push_back(Sortable(-value, NULL, p));
			}
			std::sort(srtPnts.begin(), srtPnts.end());
			
			// add candidate points
			for (int p1=0; p1<(int)srtPnts.size(); p1++)
			{
				if (excludedcontacts[srtPnts[p1].id]) continue;
				candidate.m_Primitive = TargetPrimitive(TargetPrimitive::PRIM_Point, objpnts->GetPoint(srtPnts[p1].id));
				candidate.m_fProbability = probSurface * objvals->Value(objpnts->GetPoint(srtPnts[p1].id), r);
				shared->m_LocalRegion2Candidates[localid].push_back(candidate);
				m_CandidateCorrs[c2f_level][r].push_back(objpnts->GetPoint(srtPnts[p1].id));
				m_CandidateCorrValues[c2f_level][r].push_back(objvals->Value(objpnts->GetPoint(srtPnts[p1].id), r));
				
				// clear candidates within some distance (if threshold was given)
				if (distThreshold>0)
				{
					RNArray<const PointInSet *> points;
					const R3Kdtree<const PointInSet *> * kdTree = objpnts->GetKdTree();
					kdTree->FindAll(objpnts->PointPosition(srtPnts[p1].id), 0, distThreshold, points);

					for (int p2=0; p2<points.NEntries(); p2++)
						excludedcontacts[points[p2]->id] = true;
				}
			}
		}
		
		// add potential ground correspondences
		if (probGround>0)
		{
			candidate.m_Primitive = TargetPrimitive(TargetPrimitive::PRIM_Ground);
			candidate.m_fProbability = probGround;
			shared->m_LocalRegion2Candidates[localid].push_back(candidate);
		}
		
		// add potential unassigned correspondence
		if (probUnassigned>0)
		{
			candidate.m_Primitive = TargetPrimitive(TargetPrimitive::PRIM_None);
			candidate.m_fProbability = probUnassigned;
			shared->m_LocalRegion2Candidates[localid].push_back(candidate);
		}
	}
	return shared;
#endif
	
}

void BindingManager::FillPose(CorrsPermutation * permutation)
{
//	delete permutation->m_pPose;
	bool reinit = (permutation->m_pPose==NULL);
	if (permutation->m_pPose==NULL)
		permutation->m_pPose = new PntCorrsPose(m_Interactors[Interactor::ACT_Object]);
	permutation->m_Energy.m_fTotalEnergy = -1;
	
	const int * perm = permutation->m_aPermutation;
	const PermutationsSharedData * info = permutation->m_pSharedData;
	assert(permutation->m_pPose->NMatches()==0);
	for (int i=0; i<permutation->m_iNLocalRegions; i++)	// copy current pose
	{
		assert(i<(int)info->m_Local2GlobalRegionID.size() && i < (int)info->m_LocalRegion2Candidates.size());
		int rid = info->m_Local2GlobalRegionID[i];
		if (perm[i]<0)	// unassigned
			permutation->m_pPose->AddMatch(rid, TargetPrimitive(TargetPrimitive::PRIM_PartialUnassigned));
		else
		{
			assert(perm[i]>=0 && perm[i] < (int)info->m_LocalRegion2Candidates[i].size());
			const TargetPrimitive & prim = info->m_LocalRegion2Candidates[i][perm[i]].m_Primitive;
			if (!prim.IsType(TargetPrimitive::PRIM_None))
				permutation->m_pPose->AddMatch(rid, prim);
		}
	}
	IKSolver * solver = GetEnergy()->GetSolver();
	assert(solver!=NULL);
//	std::cout<<"Pose: "<<permutation->m_pPose->NMatches()<<": [ ";
//	for (int i=0; i<permutation->m_pPose->NMatches(); i++)
//		std::cout<<permutation->m_pPose->Prim(i).IsType(TargetPrimitive::PRIM_PartialUnassigned)<<" ";
//	std::cout<<"]"<<std::endl;
	solver->SolveIKProblem(permutation->m_pPose, reinit);
}
			
bool BindingManager::RandomMutation(CorrsPermutation * permutation, int localRegionID, double searchRadius, bool useProb)
{
	
//	const char * RESET = "\033[0m";
//	const char * BLUE  = "\033[34m";
//	std::cout<<"Permutation Before [ ";
//	for (int i=0; i<permutation->m_iNLocalRegions; i++)
//	{
//		if (i==localRegionID)
//			std::cout<<BLUE;
//		std::cout<<permutation->m_aPermutation[i]<<" ";
//		if (i==localRegionID)
//			std::cout<<RESET;
//	}
//	std::cout<<"]"<<std::endl;

	assert(localRegionID>=0 && localRegionID < permutation->m_iNLocalRegions);
	PermutationsSharedData * shared = permutation->m_pSharedData;
	bool valid = false;
	
	int attempts = 10;
	while (!valid && attempts-->0)
	{
		std::vector<int> validcandidates;
		if (searchRadius<0)	// search
		{
			for (int i=0; i<(int)shared->m_LocalRegion2Candidates[localRegionID].size(); i++)
				validcandidates.push_back(i);
		}
		else
		{
			std::vector<CandPoint> & cnds = shared->m_LocalRegion2Candidates[localRegionID];
			const TargetPrimitive & prim = cnds[permutation->m_aPermutation[localRegionID]].m_Primitive;
			if (prim.IsType(TargetPrimitive::PRIM_None) || prim.IsType(TargetPrimitive::PRIM_Ground))
				return false;
			R3Point currpnt = prim.GetPosition();
			for (int i=0; i<(int)cnds.size(); i++)
			{
				const TargetPrimitive & nearprim = cnds[i].m_Primitive;
				if (nearprim.IsType(TargetPrimitive::PRIM_Point) && R3Distance(nearprim.GetPosition(), currpnt) < searchRadius)
					validcandidates.push_back(i);
			}
		}
		if (validcandidates.size()==0)
			break;
		std::sort(validcandidates.begin(), validcandidates.end());
		int randval = rand() % (int)validcandidates.size();
		if (useProb)
		{
			std::vector<double> pdf;
			for (int i=0; i<(int)validcandidates.size(); i++)
				pdf.push_back(shared->m_LocalRegion2Candidates[localRegionID][validcandidates[i]].m_fProbability);
			randval = SelectFromPDF(pdf);
		}
		
		int match = validcandidates[randval];
		assert(match < (int)shared->m_LocalRegion2Candidates[localRegionID].size());
		permutation->m_aPermutation[localRegionID] = match;
		bool assignedBefore = (shared->m_VisitedPermutations.find(*permutation) != shared->m_VisitedPermutations.end());

		valid = !assignedBefore && ValidPointAssignment(permutation, localRegionID);
		if (valid)
			shared->m_VisitedPermutations[*permutation] = true;
	}
	permutation->m_Energy.m_fTotalEnergy = -1;
	permutation->m_pPose = NULL;
	
//	std::cout<<"Permutation After [ ";
//	for (int i=0; i<permutation->m_iNLocalRegions; i++)
//	{
//		if (i==localRegionID)
//			std::cout<<BLUE;
//		std::cout<<permutation->m_aPermutation[i]<<" ";
//		if (i==localRegionID)
//			std::cout<<RESET;
//	}
//	std::cout<<"]"<<std::endl;
	return valid;
}

bool BindingManager::ValidPointAssignment(CorrsPermutation * permutation, int updatedRegion)
{
	return true;
}

void BindingManager::ComputeBindingsGenetic()
{
	assert(false);	// todo
#if 0
	m_Timer.startedProcess("00_GeneticAlgorithm");
	double searchNhd = .5;	// vk: const -- simmulated annealing?
	
	// prepare algorithm
	PermutationsSharedData * sharedData = FindCandidates(m_AlgParams.m_fCandidateValueCutoff, m_AlgParams.m_fCandidateDistThresh);
	
	//	Initial population (completely random)
	std::vector<Sortable> population;
	TimeProfiler timer;
	for (int i=0; i<m_AlgParams.m_iInitPopulation; i++)
	{
		if (i%50==0)
			timer.WriteProgress("search", i, m_AlgParams.m_iInitPopulation + m_AlgParams.m_iNSamples);
		CorrsPermutation * newborn = new CorrsPermutation(sharedData);
		for (int r = 0; r < newborn->m_iNLocalRegions; r++)
			RandomMutation(newborn, r, -1, true);
		population.push_back(Sortable(ComputeEnergy(newborn), newborn));
	}

	// Local refinement
	for (int i=0; i<m_AlgParams.m_iNSamples; i++)
	{
		if (i%50==0)
			timer.WriteProgress("search", m_AlgParams.m_iInitPopulation+i, m_AlgParams.m_iInitPopulation + m_AlgParams.m_iNSamples);
		
		// Only include the best
		std::sort(population.begin(), population.end());
		while ((int)population.size()>m_AlgParams.m_iNBest)
		{
			delete (CorrsPermutation*)population[m_AlgParams.m_iNBest].ptr;
			population.erase(population.begin()+m_AlgParams.m_iNBest);	// memory leak
		}
		
		// pick random mutation
		int npose = rand()%population.size();
		CorrsPermutation * permutation = new CorrsPermutation(*((CorrsPermutation*)population[npose].ptr));

		// pick random region (todo: can pick highest-error region)
		int r = rand()%permutation->m_iNLocalRegions;
		bool valid = RandomMutation(permutation, r, searchNhd, true);
		
		if (valid)
			population.push_back(Sortable(ComputeEnergy(permutation), permutation));
		else
			delete permutation;
		
		// TODO: pick random angle - search based on 'pose prior'?
		
		// TODO: pick random translation
	}
	std::cout<<std::endl;
	
	// only keep the m_iNBest
	std::sort(population.begin(), population.end());
	while ((int)population.size()>m_AlgParams.m_iNBest)
	{
		delete (CorrsPermutation*)population[m_AlgParams.m_iNBest].ptr;
		population.erase(population.begin()+m_AlgParams.m_iNBest);
	}

	// create bindings
	const InteractorPose * objpose = m_Interactors[Interactor::ACT_Object]->GetEnumeratedPose(0);
	R3Vector objTrans = R3Vector(0, 0, -m_Interactors[Interactor::ACT_Object]->Ground(objpose));
	for (int i=0; i<(int)population.size(); i++)
	{
		Binding * binding = new Binding(m_Interactors[Interactor::ACT_Actor], m_Interactors[Interactor::ACT_Object]);
		m_Bindings.push_back(binding);
		binding->SetPose(Interactor::ACT_Object, objpose);
		binding->SetXForm(Interactor::ACT_Object, new RigidAlign(objTrans, R3Vector(0,0,1), 0));
		binding->SetPose(Interactor::ACT_Actor, ((CorrsPermutation*)population[i].ptr)->m_pPose);
		binding->SetEnergy(ComputeEnergy((CorrsPermutation*)population[i].ptr));
	}
	
	m_Timer.finishedProcess("00_GeneticAlgorithm");
#endif
}

///////////// COMPARISONS ////////////
void BindingManager::ComputeBindingsBestContact()
{
	m_Timer.startedProcess("00_BestContact");
	int flags = GetEnergy()->GetSolver()->m_SolverFlags;
	GetEnergy()->GetSolver()->m_SolverFlags = (flags & (~(IKSolver::IK_QuickPruning)));

	const ShapePointCloud * objpnts = NULL;
	const ShapeValue * objvals = NULL;
	m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
	assert(objpnts!=NULL & objvals!=NULL);
	
	int nobjpoints = objpnts->NumPoints();		// npoints
	int nactorpoints = objvals->NDims() - 1;	// == nlabels
	
	PntCorrsPose * finalpose = new PntCorrsPose(m_Interactors[Interactor::ACT_Object]);
	for (int r=0; r<nactorpoints; r++)
	{
		if (!GetEnergy()->ClassifiedRegion(r))
			continue;
		
		int bestpoint = -2;
		double bestvalue = GetEnergy()->UnassignedPenalty(-1, r);
		double value = GetEnergy()->ContactPenalty(-1, r, TargetPrimitive(TargetPrimitive::PRIM_Ground), objpnts, objvals);
		if (value < bestvalue)
		{
			bestvalue = value;
			bestpoint = -1;
		}
		
		for (int p=0; p<nobjpoints; p++)
		{
			TargetPrimitive prim(TargetPrimitive::PRIM_Point, objpnts->GetPoint(p));
			value = GetEnergy()->ContactPenalty(-1, r, prim, objpnts, objvals);
			if (value < bestvalue)
			{
				bestpoint = p;
				bestvalue = value;
			}
		}
		
		if (bestpoint==-1)	// ground
			finalpose->AddMatch(r, TargetPrimitive(TargetPrimitive::PRIM_Ground));
		else if (bestpoint>=0)	// surface
			finalpose->AddMatch(r, TargetPrimitive(TargetPrimitive::PRIM_Point, objpnts->GetPoint(bestpoint)));
	}
	IKSolver * solver = GetEnergy()->GetSolver();
	assert(solver!=NULL);
	solver->SolveIKProblem(finalpose, true);

	
//	PermutationsSharedData * sharedData = FindCandidates(m_AlgParams.m_fCandidateValueCutoff, 0.2);
//	CorrsPermutation * bestpose = new CorrsPermutation(sharedData);
//	for (int r = 0; r < bestpose->m_iNLocalRegions; r++)
//		bestpose->m_aPermutation[r] = -1;
//	
//	for (int i=0; i<(int)sharedData->m_LocalRegion2Candidates.size(); i++)
//	{
//		double bestEnergy = RN_INFINITY;
//		int bestCandidate = -1;
//		for (int j=0; j<(int)sharedData->m_LocalRegion2Candidates[i].size(); j++)
//		{
//			bestpose->m_aPermutation[i] = j;
//			bestpose->m_pPose = NULL;
//			bestpose->m_Energy.m_fTotalEnergy = -1;
//			FillPose(bestpose);
//			m_pEnergy->ComputeEnergy(bestpose->m_pPose, bestpose->m_Energy);
//			double energy = bestpose->m_Energy.m_fFeatPointsEnergy;
//			energy += bestpose->m_Energy.m_fUnassignedPointsEnergy;
//			if (energy < bestEnergy || bestCandidate<0)
//			{
//				bestEnergy = energy;
//				bestCandidate = j;
//			}
//		}
//		assert(bestCandidate>=0);
//		bestpose->m_aPermutation[i] = bestCandidate;
//	}
//	bestpose->m_pPose = NULL;
//	bestpose->m_Energy.m_fTotalEnergy = -1;
//	FillPose(bestpose);
//	PntCorrsPose * finalpose = bestpose->m_pPose;
	
	
	const InteractorPose * objpose = m_Interactors[Interactor::ACT_Object]->GetEnumeratedPose(0);
	R3Vector objTrans = R3Vector(0, 0, -m_Interactors[Interactor::ACT_Object]->Ground(objpose));
	Binding * binding = new Binding(m_Interactors[Interactor::ACT_Actor], m_Interactors[Interactor::ACT_Object]);
	m_Bindings.push_back(binding);
	binding->SetPose(Interactor::ACT_Object, objpose);
	binding->SetXForm(Interactor::ACT_Object, new RigidAlign(objTrans, R3Vector(0,0,1), 0));
	binding->SetPose(Interactor::ACT_Actor, finalpose);
	//binding->SetEnergy(ComputeEnergy(bestpose));
	m_Timer.finishedProcess("00_BestContact");
}

void BindingManager::ComputeBindingsExternal()
{
	m_Timer.startedProcess("00_External");
	if (m_AlgParams.m_sReadPoseFile.count()>0)
	{
		IKSolver * solver = GetEnergy()->GetSolver();
		assert(solver!=NULL);
		for (int i=0; i<m_AlgParams.m_sReadPoseFile.count(); i++)
			ReadGTPose(m_AlgParams.m_sReadPoseFile[i]);
		int firstGTBinding = 0;
		for (int i=0; i<NBindings(); i++)
		{
			Binding * original = GetBinding(firstGTBinding + i);
			PntCorrsPose * oldpose = (PntCorrsPose*)original->GetPose(Interactor::ACT_Actor);
			if (!oldpose->IsParamsInit(solver->NJointParams()))
				oldpose->InitParams(solver->NJointParams(), solver->m_SolverFlags);
			EnergyInfo info;
			solver->ForwardKinematics(oldpose);
			GetEnergy()->ComputeEnergy(oldpose, info);
			original->SetEnergy(info.m_fTotalEnergy);
		}
		int lastGTBinding = NBindings();
		
		Interactor * actor = m_Interactors[Interactor::ACT_Actor];
		Interactor * object = m_Interactors[Interactor::ACT_Object];
		bool solveik = (m_AlgParams.m_iExternalFlags & BindingAlgParams::EXT_SolveIK);
		for (int i=firstGTBinding; i<lastGTBinding && solveik; i++)
		{
			Binding * original = GetBinding(i);
			Binding * projected = new Binding(actor, object);
			AddBinding(projected);
			projected->SetPose(Interactor::ACT_Object, original->GetPose(Interactor::ACT_Object));
			projected->SetXForm(Interactor::ACT_Object, original->GetBindingXForm(Interactor::ACT_Object));
			PntCorrsPose * oldpose = (PntCorrsPose*)original->GetPose(Interactor::ACT_Actor);
			PntCorrsPose * newpose = new PntCorrsPose(*oldpose);
			solver->m_SolverFlags = (solver->m_SolverFlags & (~(IKSolver::IK_QuickPruning)));
			solver->SolveIKProblem(newpose, true);
			projected->SetPose(Interactor::ACT_Actor, newpose);
			EnergyInfo info;
			GetEnergy()->ComputeEnergy(newpose, info);
//			std::cout<<"IKEnergy: "<<info.m_fTotalEnergy<<std::endl;
//			std::cout<<"Params: "<<std::flush;
//			for (int p=0; p<solver->NJointParams(); p++)
//				std::cout<<newpose->Param(p)<<" "<<std::flush;
//			std::cout<<std::endl;
			projected->SetEnergy(info.m_fTotalEnergy);
		}
		
		// update bindings by projecting points to valid corrs
		
		bool projectgt = (m_AlgParams.m_iExternalFlags & BindingAlgParams::EXT_Project2Corrs);
		if (projectgt)
		{
			FindCandidates(m_AlgParams.m_fCandidateValueCutoff, m_AlgParams.m_fCandidateDistThresh);
			double corrDistThreshold = 0.5;
			for (int i=firstGTBinding; i<lastGTBinding && projectgt; i++)
			{
				Binding * original = GetBinding(i);
				PntCorrsPose * oldpose = (PntCorrsPose*)original->GetPose(Interactor::ACT_Actor);

				for (int c2f = 0; c2f < (int)m_CandidateCorrs.size(); c2f++)
				{
					Binding * projected = new Binding(actor, object);
					AddBinding(projected);
					projected->SetPose(Interactor::ACT_Object, original->GetPose(Interactor::ACT_Object));
					projected->SetXForm(Interactor::ACT_Object, original->GetBindingXForm(Interactor::ACT_Object));
					PntCorrsPose * newpose = new PntCorrsPose(object);

					for (int p=0; p<(int)oldpose->NMatches(); p++)
					{
						int rid = oldpose->Region(p);
						if (oldpose->Prim(p).IsType(TargetPrimitive::PRIM_Point))
						{
							R3Point pos = oldpose->Prim(p).GetPosition();
							int nearest = -1;
							double mindist = -1;
							if (rid < (int)m_CandidateCorrs[c2f].size())
							{
								for (int c=0; c<(int)m_CandidateCorrs[c2f][rid].size(); c++)
								{
									double dist = R3Distance(pos, m_CandidateCorrs[c2f][rid][c].GetPosition());
									if (nearest<0 || dist < mindist)
									{
										nearest = c;
										mindist = dist;
									}
								}
							}
							if (nearest>=0 && mindist < corrDistThreshold)
								newpose->AddMatch(rid, TargetPrimitive(TargetPrimitive::PRIM_Point, m_CandidateCorrs[c2f][rid][nearest]));
						}
						else
							newpose->AddMatch(rid, TargetPrimitive(TargetPrimitive::PRIM_Ground));
					}
					solver->SolveIKProblem(newpose, true);
					projected->SetPose(Interactor::ACT_Actor, newpose);
					EnergyInfo info;
					GetEnergy()->ComputeEnergy(newpose, info);
					projected->SetEnergy(info.m_fTotalEnergy);
				}
			}
		}
		bool keepOriginal =(m_AlgParams.m_iExternalFlags & BindingAlgParams::EXT_Input);
		if (!keepOriginal)
			m_Bindings.erase(m_Bindings.begin()+firstGTBinding, m_Bindings.begin()+lastGTBinding);
		std::cout<<"NBindings: "<<m_Bindings.size()<<std::endl;
	}
	m_Timer.finishedProcess("00_External");
}


///////////// BRANCH-AND-BOUND //////////
long long  BindingManager::NMutations(const PermutationsSharedData * shared) const
{
	long long nmutations = 1;
	assert(shared->m_LocalRegion2Candidates.size()>0);
	for (int i=0; i<(int)shared->m_LocalRegion2Candidates.size(); i++)
	{
		assert((int)shared->m_LocalRegion2Candidates[i].size()>0);
		nmutations *= (int)shared->m_LocalRegion2Candidates[i].size();
	}
	return nmutations;
}

long long BindingManager::MutationID(const CorrsPermutation * permutation) const
{
	int mutationID = 0;
	int multiplier = 1;
	PermutationsSharedData * shared = permutation->m_pSharedData;
	assert(shared->m_LocalRegion2Candidates.size()>0);
	for (int i=((int)shared->m_LocalRegion2Candidates.size())-1; i>=0; i--)
	{
		int id = permutation->m_aPermutation[i];
		if (id < 0) id = 0;
		mutationID += id * multiplier;
		
		int nmax = (int)shared->m_LocalRegion2Candidates[i].size();
		assert((int)shared->m_LocalRegion2Candidates[i].size()>0);
		multiplier *= nmax;
	}
	return mutationID;
}

bool BindingManager::NextMutation(CorrsPermutation * permutation)
{
	// find last unassigned
	PermutationsSharedData * shared = permutation->m_pSharedData;
	assert(shared->m_LocalRegion2Candidates.size()>0);
	int last = -1;
	for (last=0; last<(int)shared->m_LocalRegion2Candidates.size(); last++)
	{
		if (last+1 == (int)shared->m_LocalRegion2Candidates.size()
			|| permutation->m_aPermutation[last]==-1)
			break;
	}
	
	// increment right before unassigned
	assert(last>=0 && last < permutation->m_iNLocalRegions);
	permutation->m_aPermutation[last]++;
	
	// if unassigned > max -- recursively increase
	while (last>=0)
	{
		int nmax = (int)shared->m_LocalRegion2Candidates[last].size();
		if (permutation->m_aPermutation[last] >= nmax)
		{
			permutation->m_aPermutation[last] = -1;
			last--;
			if (last >= 0)
				permutation->m_aPermutation[last]++;
		}
		else
			break;
	}
	return last < 0;
}

bool BindingManager::NextBranch(CorrsPermutation * permutation)
{
	// find last unassigned
	PermutationsSharedData * shared = permutation->m_pSharedData;
	assert(shared->m_LocalRegion2Candidates.size()>0);
	int last = -1;
	for (last=0; last<(int)shared->m_LocalRegion2Candidates.size(); last++)
	{
		if (last+1 == (int)shared->m_LocalRegion2Candidates.size()
			|| permutation->m_aPermutation[last+1]==-1)
			break;
	}
	
	// increment right before unassigned
	assert(last>=0 && last < permutation->m_iNLocalRegions);
	permutation->m_aPermutation[last]++;
	
	// if unassigned > max -- recursively increase
	while (last>=0)
	{
		int nmax = (int)shared->m_LocalRegion2Candidates[last].size();
		if (permutation->m_aPermutation[last] >= nmax)
		{
			permutation->m_aPermutation[last] = -1;
			last--;
			if (last >= 0)
				permutation->m_aPermutation[last]++;
		}
		else
			break;
	}
	return last < 0;
}

bool BindingManager::IsComplete(const CorrsPermutation * permutation) const
{
	for (int i=0; i<permutation->m_iNLocalRegions; i++)
		if (permutation->m_aPermutation[i]<0)
			return false;
	return true;
}

void BindingManager::ComputeBindingsBranchAndBound()
{
	assert(false);	// changed point stats
#if 0
	m_Timer.startedProcess("00_BranchAndBound");
	if (m_AlgParams.m_bDisableIK)
	{
		IKSolver * solver = GetEnergy()->GetSolver();
		assert(solver!=NULL);
		solver->m_SolverFlags = solver->m_SolverFlags | IKSolver::IK_DisableIK;
	}
	
	int maxlevels = m_AlgParams.m_iCoarse2Fine;
	double minDistBetweenPnts = 0;
	
	int c2f = 0;
	VKString msg;
	std::vector<Sortable> finalPopulation;
	VKStringList messages;
	CorrsPermutation * counter = NULL;
	CorrsPermutation * bestComplete = NULL;
	double bestEnergy = RN_INFINITY;
	//////// COARSE-TO-FINE-ALGORITHM ////
	for (c2f = 0; c2f < maxlevels; c2f++)
	{
		messages.push_back(VKString("C2F ")+VKString::number(c2f)+"\n");
		std::cout<<messages[messages.count()-1].c_str()<<std::flush;
		
		// enumerate candidate contacts
		PermutationsSharedData * shared = NULL;
		long long nmutations = 0;
		
		nmutations = m_AlgParams.m_iNMaxPermutations;
		double pntNhd = minDistBetweenPnts;
		minDistBetweenPnts = 0;
		double bndNhd = pntNhd * 0.5;
		while (nmutations >= m_AlgParams.m_iNMaxPermutations)
		{
			if (c2f==0)
				shared = FindCandidates(m_AlgParams.m_fCandidateValueCutoff, minDistBetweenPnts, c2f);
			else
				shared = FindCandidates(m_AlgParams.m_fCandidateValueCutoff, minDistBetweenPnts, c2f, counter, pntNhd+bndNhd);
			nmutations = NMutations(shared);
			if (nmutations >= m_AlgParams.m_iNMaxPermutations)
			{
				minDistBetweenPnts += m_AlgParams.m_fPntRadiusStep;
				bndNhd = vkMax(0, bndNhd-m_AlgParams.m_fPntRadiusStep);
			}
			if (minDistBetweenPnts>=m_AlgParams.m_fCandidateDistThresh)
				break;
		}
		
		int nregions = (int)shared->m_LocalRegion2Candidates.size();
		msg  = VKString("    MinP2PDist ")+VKString::number(minDistBetweenPnts)+"\n";
		msg += VKString("    Mutations (")+VKString::number(nmutations)+"): ";
		for (int i=0; i<nregions; i++)
			msg += VKString::number((int)shared->m_LocalRegion2Candidates[i].size())+(i==(nregions-1) ? "\n" : "*");
		messages.push_back(msg);
		std::cout<<messages[messages.count()-1].c_str()<<std::flush;

		// create initial permutation
		std::vector<Sortable> population;
		if (c2f>0)
			population.push_back(Sortable(bestEnergy, counter));
		counter = new CorrsPermutation(shared);	// re-initialize counter
		for (int r = 0; r < counter->m_iNLocalRegions; r++)
			counter->m_aPermutation[r] = -1;

		TimeProfiler timer;
		// explore the solution space
		bool terminate = false;
		int niter = 0;
		long long nquickprunned = 0;
		long long nenergyprunned = 0;
		bool updated = true;
		int lid = messages.count();
		//////// BRANCH AND BOUND ALGORITHM ////
		while (!terminate && niter < m_AlgParams.m_iNSamples)
		{
			long long id = MutationID(counter);
			if (niter%50==0 && updated)
			{
				updated = false;
				if (nmutations>1000000)
					timer.WriteProgress("b&b", id/1000000, nmutations/1000000);
				else
					timer.WriteProgress("b&b", id, nmutations);
				if (niter%250==0)
				{
					std::cout<<"\n     Iter: "<<niter<<" / "<<m_AlgParams.m_iNSamples<<std::endl;
					int r=0;
					r = (int)(10000.*(double)id/(double)nmutations);
					std::cout<<"     Progress:    "<<id<<" / "<<nmutations<<" - "<<(r/100)<<"."<<(r%100)<<"%"<<std::endl;
					r = (int)(10000.*(double)niter/(double)nmutations);
					std::cout<<"     IKCalls:    "<<niter<<" / "<<id<<" - "<<(r/100)<<"."<<(r%100)<<"%"<<std::endl;
					r = (int)(10000.*(double)nquickprunned/(double)nmutations);
					std::cout<<"     QuickPrune:  "<<nquickprunned<<" - "<<(r/100)<<"."<<(r%100)<<"%"<<std::endl;
					r = (int)(10000.*(double)nenergyprunned/(double)nmutations);
					std::cout<<"     EnergyPrune: "<<nenergyprunned<<" - "<<(r/100)<<"."<<(r%100)<<"%"<<std::endl;
					std::cout<<"     State: [ ";
					for (int i=0; i<counter->m_iNLocalRegions; i++)
						std::cout<<counter->m_aPermutation[i]<<" ";
					std::cout<<"]. CurrBest="<<bestEnergy<<std::endl;
				}
			}

			if (c2f>0)	// speed up IK calls for coarse-to-fine at finer resolution levels
			{
				IKSolver * solver = GetEnergy()->GetSolver();
				assert(solver!=NULL && bestComplete!=NULL);
				if (counter->m_pPose==NULL)
				{
					counter->m_pPose = new PntCorrsPose(m_Interactors[Interactor::ACT_Object]);
					counter->m_pPose->InitParams(solver->NJointParams(), solver->m_SolverFlags);
				}

				while (counter->m_pPose->NMatches()>0)
					counter->m_pPose->Clear(0);
				for (int jp=0; jp<solver->NJointParams(); jp++)
					counter->m_pPose->Param(jp) = bestComplete->m_pPose->Param(jp);
			}
			else if (counter->m_pPose!=NULL)
			{
				delete counter->m_pPose;
				counter->m_pPose = NULL;
			}
			double currEnergy = ComputeEnergy(counter);
			if (currEnergy < bestEnergy && IsComplete(counter))
			{
				bestEnergy = currEnergy;
				bestComplete = new CorrsPermutation(*counter);
				population.push_back(Sortable(ComputeEnergy(bestComplete), bestComplete));
			}
			
			if (currEnergy > bestEnergy)
				terminate = NextBranch(counter);
			else
				terminate = NextMutation(counter);
			
			if (currEnergy < RN_INFINITY && !terminate)
			{
				updated = true;
				niter++;
				if (id < MutationID(counter))
					nenergyprunned += (MutationID(counter) - id);
				assert(nenergyprunned>=0);
			}
			else if (!terminate)
			{
				if (id < MutationID(counter))
					nquickprunned += (MutationID(counter) - id);
				assert(nquickprunned>=0);
			}
			
			counter->m_pPose = NULL;
			counter->m_Energy.m_fTotalEnergy = -1;
		}  		//////// END: BRANCH AND BOUND ALGORITHM ////
	
		// pick best element - continue coarse-to-fine
		std::sort(population.begin(), population.end());
		assert(population.size()>0);
		finalPopulation.push_back(Sortable(bestEnergy, (CorrsPermutation*)population[0].ptr));
		std::cout<<std::endl;

		long long id = MutationID(counter);
		if (id==0)
			id = nmutations;
		
		msg = "";
		int rtmp = (int)(10000.*(double)id/(double)nmutations);
		msg += VKString("     Progress:    ")+VKString::number(id)+" - ";
		msg += VKString::number(rtmp/100)+"."+VKString::number(rtmp%100)+"%\n";
		
		rtmp = (int)(10000.*(double)niter/(double)id);
		msg += VKString("     IKCalls:    ")+VKString::number(niter)+" - ";
		msg += VKString::number(rtmp/100)+"."+VKString::number(rtmp%100)+"%\n";
		
		rtmp = (int)(10000.*(double)nquickprunned/(double)nmutations);
		msg += VKString("     QuickPrune:  ")+VKString::number(nquickprunned)+" - ";
		msg += VKString::number(rtmp/100)+"."+VKString::number(rtmp%100)+"%\n";
		
		rtmp = (int)(10000.*(double)nenergyprunned/(double)nmutations);
		msg += VKString("     EnergyPrune: ")+VKString::number(nenergyprunned)+" - ";
		msg += VKString::number(rtmp/100)+"."+VKString::number(rtmp%100)+"%\n";
		msg += VKString("     CurrBest: ")+VKString::number(bestEnergy)+"\n";
		messages.push_back(msg);
		std::cout<<msg.c_str()<<std::flush;
		
		if (niter==m_AlgParams.m_iNSamples)	// max # iterations reached
		{
			msg = "ERROR Not finished!\n";
			msg += "Final State [";
			for (int i=0; i<counter->m_iNLocalRegions; i++)
				msg += VKString::number(counter->m_aPermutation[i])+" ";
			msg += "]";
			messages.push_back(msg);
			std::cout<<msg.c_str()<<std::flush;
			break;
		}
		if (minDistBetweenPnts==0)
		{
			msg = "Converged";
			messages.push_back(msg);
			std::cout<<msg.c_str()<<std::flush;
			break;
		}
		else if (c2f==maxlevels-1)
		{
			msg = VKString("Max C2F Level: ")+VKString::number(c2f);
			messages.push_back(msg);
			std::cout<<msg.c_str()<<std::flush;
			break;
		}
		counter = (CorrsPermutation*)population[0].ptr;
		population.clear();
	}  //////// END: COARSE-TO-FINE-ALGORITHM ////
	
	m_sAlgorithmInfo = "";
	for (int i=0; i<messages.count(); i++)
		m_sAlgorithmInfo += messages[i];
	IKSolver * solver = GetEnergy()->GetSolver();
	assert(solver!=NULL);
	m_sAlgorithmInfo += VKString("\nIKInfo:\n")+solver->m_pIKStats->GetStatsStr();
	const InteractorPose * objpose = m_Interactors[Interactor::ACT_Object]->GetEnumeratedPose(0);
	R3Vector objTrans = R3Vector(0, 0, -m_Interactors[Interactor::ACT_Object]->Ground(objpose));
	for (int i=0; i<(int)finalPopulation.size(); i++)
	{
		int id = ((int)finalPopulation.size()) - 1 - i;
		Binding * binding = new Binding(m_Interactors[Interactor::ACT_Actor], m_Interactors[Interactor::ACT_Object]);
		m_Bindings.push_back(binding);
		binding->SetPose(Interactor::ACT_Object, objpose);
		binding->SetXForm(Interactor::ACT_Object, new RigidAlign(objTrans, R3Vector(0,0,1), 0));
		binding->SetPose(Interactor::ACT_Actor, ((CorrsPermutation*)finalPopulation[id].ptr)->m_pPose);
		binding->SetEnergy(ComputeEnergy((CorrsPermutation*)finalPopulation[id].ptr));
	}
	
	assert(finalPopulation.size()==m_CandidateCorrs.size());
	assert(finalPopulation.size()==m_CandidateCorrValues.size());
	std::vector<std::vector<std::vector<ShapePoint> > > reordered;
	std::vector<std::vector<std::vector<double> > > valsreordered;
	for (int i=0; i<(int)m_CandidateCorrs.size(); i++)
	{
		reordered.push_back(m_CandidateCorrs[(int)m_CandidateCorrs.size()-1-i]);
		valsreordered.push_back(m_CandidateCorrValues[(int)m_CandidateCorrValues.size()-1-i]);
	}
	m_CandidateCorrs = reordered;
	m_CandidateCorrValues = valsreordered;
	
	m_Timer.finishedProcess("00_BranchAndBound");
	
	VotedCandidates(0.1, 1.);
#endif
}


PntCorrsPose* BindingManager::LocalSearchStep(PntCorrsPose * pose, double nhdSearch)
{
	IKSolver * solver = GetEnergy()->GetSolver();
	EnergyInfo energy;
	double bestEnergy = GetEnergy()->ComputeEnergy(pose, energy);
	PntCorrsPose * bestpose = new PntCorrsPose(*pose);
	bool terminate = false;
	int nmaxiter = 100;
	
	// prepare
	const ShapePointCloud * objpnts = NULL;
	const ShapeValue * objvals = NULL;
	m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
	assert(objpnts!=NULL & objvals!=NULL);

	//std::cout<<"Local Search"<<std::endl;
	while(!terminate && nmaxiter-->=0)
	{
		terminate = true;
		for (int i=0; i<pose->NMatches(); i++)
		{
			int r = pose->Region(i);
			TargetPrimitive & p = pose->Prim(i);
			if (p.IsType(TargetPrimitive::PRIM_Point))
			{
				int pid = p.PointID((ShapePointCloud*)objpnts);
				R3Point pos = pose->GetPosition(solver->m_Pnts2Skeleton[r]);
				RNArray<const PointInSet*> nhd;
				objpnts->GetKdTree()->FindAll(pos, 0, nhdSearch, nhd);
				for (int n=0; n<nhd.NEntries(); n++)
				{
					if (pid != nhd[n]->id && objvals->Value(objpnts->GetPoint(nhd[n]->id), r) > m_AlgParams.m_fCandidateValueCutoff)
					{
						PntCorrsPose * currpose = new PntCorrsPose(*pose);
						currpose->Clear(i);
						currpose->AddMatch(r, TargetPrimitive(TargetPrimitive::PRIM_Point, objpnts->GetPoint(nhd[n]->id)));
						solver->SolveIKProblem(currpose, false);
						double currEnergy = GetEnergy()->ComputeEnergy(currpose, energy);
						if (currEnergy < bestEnergy)
						{
							//std::cout<<"   Improved: "<<bestEnergy<<" -> "<<currEnergy<<std::endl;
							bestEnergy = currEnergy;
							delete bestpose;
							bestpose = currpose;
							terminate = false;
						}
						else
							delete currpose;
					}
				}
			}
		}
	}
	
	return bestpose;
}


//////////////// Rigid poses search ////////////////
void BindingManager::ComputeBindingsPoseFirstSearch()
{
	assert(false);
#if 0
	m_Timer.startedProcess("00_PoseFirstSearch");
	double distThreshold = 0.5;
	double valThreshold = m_AlgParams.m_fCandidateValueCutoff;
		
	m_CandidateCorrs.clear();
	m_CandidateCorrValues.clear();
//	std::vector<std::vector<std::vector<int> > > pntids;
	
	const ShapePointCloud * objpnts = NULL;
	const ShapeValue * objvals = NULL;
	m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
	assert(objpnts!=NULL & objvals!=NULL);
	
	const R3Kdtree<const PointInSet *> * kdTree = objpnts->GetKdTree();
	
//	m_CandidateCorrs.push_back(std::vector<std::vector<ShapePoint> >());
//	m_CandidateCorrValues.push_back(std::vector<std::vector<double> >());
//	pntids.push_back(std::vector<std::vector<int> >());
	
	int c2f_level = 0;
	int nobjpoints = objpnts->NumPoints();		// npoints
	int nactorpoints = objvals->NDims() - 1;	// == nlabels
	
	IKSolver * solver = GetEnergy()->GetSolver();
	assert(solver!=NULL && solver->m_pPrior!=NULL);
	PosePrior * prior = solver->m_pPrior;
	
	std::vector<std::vector<Sortable> > population;
	for (int p=0; p<prior->NSuggestedInitPoses(); p++)
		population.push_back(std::vector<Sortable>());

	PntCorrsPose * dummypose = new PntCorrsPose(m_Interactors[Interactor::ACT_Object]);
	
	std::vector<ShapePointCloud *> perregionpnts;
	std::vector<std::vector<int> > origPntIDs;
	for (int r=0; r<nactorpoints; r++)	// for each contact point
	{
		perregionpnts.push_back(NULL);
		origPntIDs.push_back(std::vector<int>());
		if (!GetEnergy()->ClassifiedRegion(r))
			continue;
		if (GetEnergy()->SurfaceProbability(r)>0)
		{
			perregionpnts[r] = new ShapePointCloud(NULL, "pts");
			for (int p=0; p<nobjpoints; p++)
			{
				if (objvals->Value(objpnts->GetPoint(p), r)>valThreshold)
				{
					origPntIDs[r].push_back(p);
					perregionpnts[r]->AddPoint(ShapePoint(objpnts->PointPosition(p)));
				}
			}
		}
	}
		
	
	TimeProfiler timer;
	for (int r=0; r<nactorpoints; r++)	// for each contact point
	{
		// keep candidates for debugging
//		m_CandidateCorrs[c2f_level].push_back(std::vector<ShapePoint>());
//		m_CandidateCorrValues[c2f_level].push_back(std::vector<double>());
//		pntids[c2f_level].push_back(std::vector<int>());
		
		// ignore if there is not training data
		if (!GetEnergy()->ClassifiedRegion(r))
			continue;

		double probSurface = GetEnergy()->SurfaceProbability(r);
		double probGround = GetEnergy()->GroundProbability(r);			
		double probUnassigned = GetEnergy()->UnassignedProbability(r);
		
		if (probSurface>0)
		{
			std::vector<Sortable> srtPnts;
			for (int p=0; p<nobjpoints; p++)
			{
				timer.WriteProgress("prog", r*nobjpoints + p, nactorpoints*nobjpoints);
				double value = objvals->Value(objpnts->GetPoint(p), r);
				if (value > valThreshold)
				{
					srtPnts.push_back(Sortable(-value, NULL, p));
//					int poseid = 1;
					for (int poseid = 0; poseid < prior->NSuggestedInitPoses(); poseid++) // for each possibel initial pose
					{
						PntCorrsPose * bestpose = NULL;
						double bestenergy = RN_INFINITY;
						for (double rot = -M_PI; rot < M_PI; rot += M_PI / 32.)// for each rotation
						{
							prior->InitPose(poseid, dummypose);
							dummypose->Param(0) = 0;
							dummypose->Param(1) = 0;
							dummypose->Param(2) = 0;	// translation
							dummypose->Param(3) = rot;	// up-rotation
							solver->ForwardKinematics(dummypose);
							
							// set translation (from pose & rotation)
							assert(r < (int)solver->m_Pnts2Skeleton.size());
							R3Point jpos = dummypose->GetPosition(solver->m_Pnts2Skeleton[r]);
							R3Point trgpos = objpnts->PointPosition(p);
							dummypose->Param(0) = trgpos.X() - jpos.X();
							dummypose->Param(1) = trgpos.Y() - jpos.Y();
							dummypose->Param(2) = trgpos.Z() - jpos.Z();
							solver->ForwardKinematics(dummypose);
							
							// set contacts near limb
							int ngoodactors = 0;
							PntCorrsPose * currpose = new PntCorrsPose(dummypose);
							currpose->AddMatch(r, TargetPrimitive(TargetPrimitive::PRIM_Point, objpnts->GetPoint(p)));
							for (int r2=0; r2<nactorpoints; r2++)	// for all other contacts
							{
								if (r2==r || !GetEnergy()->ClassifiedRegion(r2))
									continue;
								double surfaceProbR2 = GetEnergy()->SurfaceProbability(r2);
								double groundProbR2 = GetEnergy()->GroundProbability(r2);
								
								// find closes point
								assert(r2 < (int)solver->m_Pnts2Skeleton.size());
								R3Point jpos = dummypose->GetPosition(solver->m_Pnts2Skeleton[r2]);
								double dist2surface = RN_INFINITY;
								int pntid = -1;
								const PointInSet * closest = NULL;
								if (perregionpnts[r2]!=NULL)
									closest = perregionpnts[r2]->GetKdTree()->FindClosest(jpos);
								if (closest!=NULL)
								{
									pntid = origPntIDs[r2][closest->id];
									dist2surface = R3Distance(jpos, objpnts->PointPosition(pntid));
								}
								double dist2ground = jpos.Z();

								if (groundProbR2==0)
									dist2ground = RN_INFINITY;
								if (surfaceProbR2==0)
									dist2surface = RN_INFINITY;
								
								if (dist2ground < RN_INFINITY || dist2surface < RN_INFINITY)
								{
									if (dist2ground < dist2surface)
										currpose->AddMatch(r2, TargetPrimitive(TargetPrimitive::PRIM_Ground));
									else
										currpose->AddMatch(r2, TargetPrimitive(TargetPrimitive::PRIM_Point, objpnts->GetPoint(pntid)));
								}
							}
							EnergyInfo energy;
							GetEnergy()->ComputeEnergy(currpose, energy);
							double eval = energy.m_fFeatPointsEnergy + energy.m_fUnassignedPointsEnergy;
							eval += energy.m_fIntersectionEnergy+energy.m_fContactPairwiseDistsEnergy;
							population[poseid].push_back(Sortable(eval, currpose));
						}
					}
				}
			}
		}
	}
	std::cout<<std::endl;
	
	// Cluster poses
	std::vector<std::vector<Sortable> > clustered;
	std::cout<<"Clustering..."<<std::endl;
	ClusterPoses(population, clustered);
	
	std::cout<<"Full IK Optimization for clusters: "<<clustered.size()<<std::endl;
	int nbest = 1;
	int nmax = 100000;
	std::vector<Sortable> srtClusters;
	EnergyInfo energy;
	for (int i=0; i<vkMin(nmax, (int)clustered.size()); i++)
	{
		std::sort(clustered[i].begin(), clustered[i].end());
		for (int j=0; j<vkMin(nbest, (int)clustered[i].size()); j++)
		{
			timer.WriteProgress("newtm", i * nbest + j, nbest * (int)clustered.size());
			PntCorrsPose * pose = (PntCorrsPose*)clustered[i][j].ptr;
			solver->SolveIKProblem(pose, false);
			srtClusters.push_back(Sortable(GetEnergy()->ComputeEnergy(pose, energy), pose, i));
		}
	}
	std::cout<<std::endl;

	int nbestclusters = 10;
	// Extract candidate poses + candidate contacts for local refinement
	std::sort(srtClusters.begin(), srtClusters.end());
	std::cout<<"Refined Search for best "<<nbestclusters<<std::endl;
	std::vector<Sortable> currpopulation;
	for (int i=0; i<(int)vkMin(nbestclusters, srtClusters.size()); i++)
	{
		PntCorrsPose * improved = LocalSearchStep((PntCorrsPose *)srtClusters[i].ptr);
		double e = GetEnergy()->ComputeEnergy(improved, energy);
		currpopulation.push_back(Sortable(e, improved, srtClusters[i].id));
	}
	
	// final results
	std::sort(currpopulation.begin(), currpopulation.end());
	std::vector<Sortable> finalPopulation;
	int nbestfinal = 5;
	std::cout<<"Store final results "<<nbestfinal<<std::endl;
	for (int i=0; i<(int)vkMin(nbestfinal, currpopulation.size()); i++)
	{
		m_CandidateCorrs.push_back(std::vector<std::vector<ShapePoint> >());
		m_CandidateCorrValues.push_back(std::vector<std::vector<double> >());
		for (int j=0; j<clustered[currpopulation[i].id].size(); j++)
		{
			PntCorrsPose * pose = (PntCorrsPose*)clustered[currpopulation[i].id][j].ptr;
			for (int mid = 0; mid < pose->NMatches(); mid++)
			{
				int r = pose->Region(mid);
				while (r >= (int)m_CandidateCorrs[i].size())
				{
					m_CandidateCorrs[i].push_back(std::vector<ShapePoint>());
					m_CandidateCorrValues[i].push_back(std::vector<double>());
				}
				if (pose->Prim(mid).IsType(TargetPrimitive::PRIM_Point))
					m_CandidateCorrs[i][r].push_back(pose->Prim(mid).GetPoint());
			}
		}
		finalPopulation.push_back(currpopulation[i]);
	}
	
	FindCandidates(valThreshold, 0, (int)finalPopulation.size());

	std::cout<<"Final: "<<finalPopulation.size()<<std::endl;
	
	delete dummypose;
	
	const InteractorPose * objpose = m_Interactors[Interactor::ACT_Object]->GetEnumeratedPose(0);
	R3Vector objTrans = R3Vector(0, 0, -m_Interactors[Interactor::ACT_Object]->Ground(objpose));
	for (int i=0; i<(int)finalPopulation.size(); i++)
	{
//		int id = ((int)finalPopulation.size()) - 1 - i;
		int id = i;
		Binding * binding = new Binding(m_Interactors[Interactor::ACT_Actor], m_Interactors[Interactor::ACT_Object]);
		m_Bindings.push_back(binding);
		binding->SetPose(Interactor::ACT_Object, objpose);
		binding->SetXForm(Interactor::ACT_Object, new RigidAlign(objTrans, R3Vector(0,0,1), 0));
		binding->SetPose(Interactor::ACT_Actor, (PntCorrsPose*)finalPopulation[id].ptr);
		binding->SetEnergy(GetEnergy()->ComputeEnergy((PntCorrsPose*)finalPopulation[id].ptr, energy));
	}
	m_Timer.finishedProcess("00_PoseFirstSearch");
#endif
}

void BindingManager::ClusterPoses(std::vector<std::vector<Sortable> > & inputPopulation,
								  std::vector<std::vector<Sortable> > & outputPopulation)
{
	double posediff = 0.5;
	
	for (int i=0; i<(int)inputPopulation.size(); i++)
	{
		std::sort(inputPopulation[i].begin(), inputPopulation[i].end());
		std::vector<bool> visited;
		for (int j=0; j<(int)inputPopulation[i].size(); j++)
			visited.push_back(false);
		
		for (int j=0; j<(int)inputPopulation[i].size(); j++)
		{
			if (!visited[j])
			{
				int lid = (int)outputPopulation.size();
				outputPopulation.push_back(std::vector<Sortable>());
				for (int k=j; k<(int)inputPopulation[i].size(); k++)
				{
					if (IsSimilar((PntCorrsPose*)inputPopulation[i][j].ptr, (PntCorrsPose*)inputPopulation[i][k].ptr, posediff))
					{
						visited[k] = true;
						outputPopulation[lid].push_back(inputPopulation[i][k]);
					}
				}
				//std::cout<<"Cluster: "<<i<<" - "<<outputPopulation[lid].size()<<std::endl;
			}
		}
	}
}

bool BindingManager::IsSimilar(PntCorrsPose * pose1, PntCorrsPose * pose2, double dist) const
{
	if (pose1==pose2)
		return true;
	
	if (pose1->NMatches()!=pose2->NMatches())
		return false;
	
	for (int i=0; i<pose1->NMatches(); i++)
	{
		TargetPrimitive & prim1 = pose1->Prim(i);
		int r1 = pose1->Region(i);
		bool foundValid = false;
		for (int j=0; j<pose2->NMatches(); j++)
		{
			TargetPrimitive & prim2 = pose2->Prim(j);
			int r2 = pose2->Region(j);
			if (prim1.IsType(TargetPrimitive::PRIM_Point) && prim2.IsType(TargetPrimitive::PRIM_Point) && r1==r2)
			{
				if (R3Distance(prim1.GetPosition(), prim2.GetPosition()) <= dist)
					foundValid = true;
				break;
			}
			else if (prim1.IsType(TargetPrimitive::PRIM_Ground) && prim2.IsType(TargetPrimitive::PRIM_Ground))
			{
				foundValid = true;
				break;
			}
		}
		if (!foundValid)
			return false;
	}
	return true;
}


void BindingManager::AddGlobalMatchingData()
{
	std::cout<<"Add Global Matching Data"<<std::endl;
	assert(m_AlgParams.m_sReadPoseFile.count()==1);
	ReadGTPose(m_AlgParams.m_sReadPoseFile[0]);
	
	bool WEIGHT_BY_POSEPRIOR = false;
	if (WEIGHT_BY_POSEPRIOR)
	{
		assert(false);	// possibly todo...
		//		IKSolver * solver = GetEnergy()->GetSolver();
		//		assert(solver!=NULL);
		//
		//		for (int i=0; i<NBindings(); i++)
		//		{
		//			Binding * original = GetBinding(firstGTBinding + i);
		//
		//			if (!oldpose->IsParamsInit(solver->NJointParams()))
		//				oldpose->InitParams(solver->NJointParams(), solver->m_SolverFlags);
		//			EnergyInfo info;
		//			solver->ForwardKinematics(oldpose);
		//			GetEnergy()->ComputeEnergy(oldpose, info);
		//			original->SetEnergy(info.m_fTotalEnergy);
		//		}
	}
	
	const ShapePointCloud * objpnts = NULL;
	const ShapeValue * objvals = NULL;
	m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
	assert(objpnts!=NULL & objvals!=NULL);
	int nobjpoints = objpnts->NumPoints();		// npoints
	int nactorpoints = objvals->NDims() - 1;	// == nlabels
	
	double * maxvalues = new double[objvals->NDims()];
	for (int i=0; i<(int)objvals->NDims(); i++)
		maxvalues[i] = 0;
	
	double * values2modify = new double[nobjpoints * objvals->NDims()];
	for (int i=0; i<objpnts->NumPoints(); i++)
	{
		objvals->Values(objpnts->GetPoint(i), &(values2modify[i * objvals->NDims()]));
		for (int r=0; r<(int)objvals->NDims(); r++)
			maxvalues[r] = vkMax(maxvalues[r], values2modify[i * objvals->NDims() + r]);
	}
	
//	for (int i=0; i<objpnts->NumPoints(); i++)
//	for (int d=0; d<objvals->NDims(); d++)
//		values2modify[i*objvals->NDims() + d] = 0;

	const R3Kdtree<const PointInSet *> * objpntskdtree = objpnts->GetKdTree();
	
	int nbindings = NBindings();
	double splatDist = 0.5;	// 30cm
	double amplitude = 0.8/nbindings;
	double otherproboffset = .5/nbindings;
	//double falloff = 2.;	// small - sharper
	double falloff = 0.5;	// small - sharper
	for (int i=0; i<NBindings(); i++)	// for each pose - transfer
	{
		std::vector<bool> included;
		for (int r=0; r<nactorpoints; r++)	// for each contact point
			included.push_back(false);
		
		PntCorrsPose * pose = (PntCorrsPose*)GetBinding(i)->GetPose(Interactor::ACT_Actor);
		
		for (int m=0; m<pose->NMatches(); m++)
		{
			int r = pose->Region(m);
			const TargetPrimitive & prim = pose->Prim(m);
			assert(r < nactorpoints);
			included[r] = true;
			if (prim.IsType(TargetPrimitive::PRIM_Point))
			{
				RNArray<const PointInSet *> nhd;
				int closestid = objpntskdtree->FindClosest(prim.GetPosition())->id;
//				values2modify[closestid*objvals->NDims()+r] = amplitude;
				objpntskdtree->FindAll(prim.GetPosition(), 0, splatDist, nhd);
				for (int n=0; n<nhd.NEntries(); n++)
				{
					double dist = R3Distance(objpnts->PointPosition(nhd[n]->id), prim.GetPosition());
					if (nhd[n]->id==closestid)
						values2modify[nhd[n]->id*objvals->NDims()+r] += amplitude*maxvalues[r];
					else
						values2modify[nhd[n]->id*objvals->NDims()+r] += amplitude*maxvalues[r]*exp(-pow(dist/(falloff*splatDist), 2.));
					if (values2modify[nhd[n]->id*objvals->NDims()+r]>1.)
						values2modify[nhd[n]->id*objvals->NDims()+r] = 1.;
				}
			}
			else if (prim.IsType(TargetPrimitive::PRIM_Ground))
				GetEnergy()->OffsetGroundProb(otherproboffset, r);
			else
				assert(false);
		}
		
		for (int r=0; r<nactorpoints; r++)
		{
			if (!GetEnergy()->ClassifiedRegion(r))
				continue;
			if (!included[r])	// unassigned
			{
				std::cout<<"Offset: "<<r<<std::endl;
				GetEnergy()->OffsetUnassignedProb(otherproboffset, r);
			}
		}
	}
	((ShapeValue*)objvals)->SetPntsValues(values2modify, nobjpoints*objvals->NDims());
	m_Bindings.clear();
}


