#include "Binding.h"
#include "Interactor.h"
#include "BindingManager.h"

#include "TimeProfiler.h"

#include "ShapePointCloud.h"
#include "ShapeValue.h"
#include "ShapeValue.h"
#include "VkFunctions.h"
#include "IKSolver.h"
#include "PntCorrsPose.h"

#ifdef INCLUDE_BULLET
	#include "btBulletDynamicsCommon.h"
#endif

// this is for baseline comparison to a rigid pose
void BindingManager::AddClosestMatch(PntCorrsPose * pose, int region, double valThreshold)
{
	IKSolver * solver = GetEnergy()->GetSolver();
	assert(solver!=NULL);
	int jid = solver->m_Pnts2Skeleton[region];
	R3Point pos = pose->GetPosition(jid);

	if (m_CandidateCorrs.size()==0)
	{
		const ShapePointCloud * objpnts = NULL;
		const ShapeValue * objvals = NULL;
		m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
		assert(objpnts!=NULL & objvals!=NULL);
		int nobjpoints = objpnts->NumPoints();		// npoints
		int nactorpoints = objvals->NDims() - 1;	// == nlabels

		int last = (int)m_CandidateCorrs.size();
		m_CandidateCorrs.push_back(std::vector<std::vector<ShapePoint> >());
		m_CandidateCorrValues.push_back(std::vector<std::vector<double> >());
		for (int r=0; r<nactorpoints; r++)
		{
			m_CandidateCorrs[last].push_back(std::vector<ShapePoint>());
			m_CandidateCorrValues[last].push_back(std::vector<double>());
			for (int p=0; p<objpnts->NumPoints(); p++)
			{
				double value = objvals->Value(objpnts->GetPoint(p), r);
				if (value > valThreshold)
				{
					m_CandidateCorrs[last][r].push_back(ShapePoint(objpnts->GetPoint(p)));
					m_CandidateCorrValues[last][r].push_back(value);
				}
			}
		}
	}
	
	int bestpoint = -1;
	double bestDistance = 0.5;	// threshold
	for (int p=0; p<(int)m_CandidateCorrs[0][region].size(); p++)
	{
		double dist = R3Distance(m_CandidateCorrs[0][region][p].GetPosition(), pos);
		if (dist < bestDistance)
		{
			bestDistance = dist;
			bestpoint = p;
		}
	}
	
	double groundProbability = GetEnergy()->GroundProbability(-1, region);
	if (groundProbability>0 && pos.Z()<bestDistance)
		pose->AddMatch(region, TargetPrimitive(TargetPrimitive::PRIM_Ground));
	else if (bestpoint>=0)
		pose->AddMatch(region, TargetPrimitive(TargetPrimitive::PRIM_Point, m_CandidateCorrs[0][region][bestpoint]));
}

// NOTE: this is actually not branch & bound, but our final algorithm... todo: rename
void BindingManager::ComputeBindingsBranchAndBound2()
{
	m_Timer.startedProcess("00_BnB2");
	
	bool DEBUG_FIRST_POSES = false;
	
	double valThreshold = m_AlgParams.m_fCandidateValueCutoff;
	
	// candidates
	const ShapePointCloud * objpnts = NULL;
	const ShapeValue * objvals = NULL;
	m_Interactors[Interactor::ACT_Object]->GetObjectPredictions(&objpnts, &objvals);
	assert(objpnts!=NULL & objvals!=NULL);
	
	int nobjpoints = objpnts->NumPoints();		// npoints
	int nactorpoints = objvals->NDims() - 1;	// == nlabels
	IKSolver * solver = GetEnergy()->GetSolver();
	assert(solver!=NULL && solver->m_pPrior!=NULL);
	PosePrior * prior = solver->m_pPrior;
	
	if (m_AlgParams.m_bDisableIK)
		solver->m_SolverFlags = solver->m_SolverFlags | IKSolver::IK_DisableIK;

	std::vector<Sortable> population;
	
	TimeProfiler timer;
	int ncurrent = 0;
	int nchecked = 0;
	bool terminate = false;
	std::vector<Sortable> backupPopulation;	// just to avoid 0 poses
	std::cout<<"Exploring Search Space (w/ end-effector distributions)"<<std::endl;
	
	// count positive signals
	int countpnts = 0;
	int mmaxpnts = 1000;// never reached...
	while (countpnts==0 || countpnts>mmaxpnts)
	{
		countpnts = 0;
		for (int r=0; r<nactorpoints && !terminate; r++)	// for each contact point
		{
			if (!GetEnergy()->ClassifiedRegion(r) && GetEnergy()->SurfaceProbability(-1, r)>0)
				continue;
			for (int p=0; p<nobjpoints && !terminate; p++)
				if (objvals->Value(objpnts->GetPoint(p), r)>valThreshold)
					countpnts++;
		}
		if (countpnts==0)
		{
			valThreshold/=2.;
			std::cout<<"[WARNING] Found 0 contact points. Reducing threshold: "<<valThreshold<<std::endl;
			if (valThreshold<10e-10)
				break;
		}
		else if (countpnts>mmaxpnts)
		{
			valThreshold*=1.5;
			std::cout<<"[WARNING] Too many contact points. Increasing threshold: "<<valThreshold<<", pnts="<<countpnts<<std::endl;
		}
	}
	std::cout<<"NContacts: "<<countpnts<<std::endl;
	
	for (int r=0; r<nactorpoints && !terminate; r++)	// for each contact point
	{
		// ignore if there is no training data
		if (!GetEnergy()->ClassifiedRegion(r))
			continue;
		
		double probSurface = GetEnergy()->SurfaceProbability(-1, r);	// ever assigned to surface
//		double probGround = GetEnergy()->GroundProbability(-1, r);
//		double probUnassigned = GetEnergy()->UnassignedProbability(-1, r);
		
		if (probSurface>0)
		{
			//std::vector<Sortable> srtPnts;
			for (int p=0; p<nobjpoints && !terminate; p++)
			{
				if ((ncurrent%100) == 0)
					timer.WriteProgress("prog", r*nobjpoints + p, nactorpoints*nobjpoints);
				ncurrent++;
				double value = objvals->Value(objpnts->GetPoint(p), r);
				if (value > valThreshold)
				{
					//srtPnts.push_back(Sortable(-value, NULL, p));
					for (double rot = -M_PI; rot < M_PI && !terminate; rot += M_PI / 32.)// for each rotation
					{
						for (int poseID=0; poseID < prior->NSuggestedInitPoses() && !terminate; poseID++)	// for each pose id
						{
							PntCorrsPose * currpose = new PntCorrsPose(m_Interactors[Interactor::ACT_Object]);
							currpose->SetPoseID(poseID);
							currpose->InitParams(solver->NJointParams(), solver->m_SolverFlags);
							//if (DEBUG_FIRST_POSES)
							prior->InitPose(poseID, currpose);
							currpose->Param(3) = rot;	// up-rotation
							solver->ForwardKinematics(currpose);
							currpose->AddMatch(r, TargetPrimitive(TargetPrimitive::PRIM_Point, objpnts->GetPoint(p)));
							R3Point jpos = currpose->GetPosition(solver->m_Pnts2Skeleton[r]);
							R3Point trgpos = objpnts->PointPosition(p);
							currpose->Param(0) = trgpos.X() - jpos.X();
							currpose->Param(1) = trgpos.Y() - jpos.Y();
							currpose->Param(2) = trgpos.Z() - jpos.Z();
							solver->ForwardKinematics(currpose);
							
							// add unassigned
							for (int rnull=0; rnull<nactorpoints; rnull++)
							{
								if (rnull!=r && GetEnergy()->ClassifiedRegion(rnull))
								{
									if (m_AlgParams.m_bDisableIK)
										AddClosestMatch(currpose, rnull, m_AlgParams.m_fCandidateValueCutoff);
									else
										currpose->AddMatch(rnull, TargetPrimitive(TargetPrimitive::PRIM_PartialUnassigned));
								}
							}
							
							EnergyInfo energyinfo;
							nchecked++;
							double energy = RN_INFINITY;
							if (solver->m_SolverFlags & IKSolver::IK_DisableIK)
								energy = GetEnergy()->ComputeEnergy(currpose, energyinfo);
							else
							{
								energyinfo.m_Flags2Fill |= EnergyParams::E_Approximate;
								currpose = GetEnergy()->ApproximateMinEnergy(currpose, energyinfo);	// note: memory leak... but should be ok...
								energy = energyinfo.m_fGridOverlapMinEnergy;
							}
							if ((energy<RN_INFINITY/* && quickvaliditytest*/) || DEBUG_FIRST_POSES)
								population.push_back(Sortable(energy, currpose));
							else
								backupPopulation.push_back(Sortable(energy, currpose));
						}
					}
				}
			}
		}
	}
	std::cout<<std::endl;
	std::cout<<"Valid Population: "<<population.size()<<" / "<<nchecked<<std::endl;
	if (population.size()==0)
	{
		std::cout<<"No valid poses. Backing up: "<<backupPopulation.size()<<std::endl;
		population = backupPopulation;
	}

	std::cout<<"Exploring Search Space (full IK): "<<std::endl;
	
	std::vector<Sortable> population2;
	EnergyInfo einfo;
	einfo.m_Flags2Fill |= EnergyParams::E_Approximate;
	std::vector<double> minbound;
	std::vector<double> evalues;
	
	//int nfine = 100000;
	int nfine = 10000;
	//int nfine = 500;
	int nlocalsearch = 5;
	bool cheapLocalSearchPrecluster = false;
	bool cheapLocalSearchPostcluster = true;
	bool expensiveLocalSearch = false;
	bool dealWithSymmetric = false;	// not implemented, not used...
	bool resolveIK = true;
	bool clusterRemaining = true;
	
	double bestEnergy = RN_INFINITY;
	bool DEBUG_VIS_BOUNDS = false;
	int nposes = m_AlgParams.m_iNBest;
	if (m_AlgParams.m_bDisableIK)	// keep the best one
	{
		std::sort(population.begin(), population.end());
		population.erase(population.begin()+1, population.end());
	}
	else if (DEBUG_FIRST_POSES)
		nposes = (int)population.size();
	else
	{
		std::sort(population.begin(), population.end());
		for (int i=0; i<vkMin((int)population.size(), nfine); i++)	// top 100
		{
			if (i%100==0)
			{
				timer.WriteProgress("prog2", i, vkMin((int)population.size(), nfine));
				std::cout<<population[i].value<<" / "<<bestEnergy<<std::endl;
			}
			std::vector<int> regions;
			std::vector<TargetPrimitive> matches;
			
			einfo.m_fGridOverlapMinEnergy = population[i].value;
			PntCorrsPose * oldpose = (PntCorrsPose*)population[i].ptr;
			PntCorrsPose * newpose = oldpose;
			if (einfo.m_fGridOverlapMinEnergy > bestEnergy && !DEBUG_VIS_BOUNDS)
				break;	// lower bound worse than best
			
			// check if already exists
			bool alreadyExists = false;
			for (int j=0; j<(int)population2.size() && !alreadyExists; j++)
				alreadyExists = IsSimilar((PntCorrsPose*)population2[j].ptr, newpose, 0);
			
			// run full IK
			if (!alreadyExists)
			{
				solver->SolveIKProblem(newpose, false);
				if (expensiveLocalSearch)
					newpose = LocalSearchStep(newpose, 0.2*1.5);	// TODO: 0.2 == gridsize

				double eval = GetEnergy()->ComputeEnergy(newpose, einfo);
				if (eval < bestEnergy)
					bestEnergy = eval;
				
				double eval2 = RN_INFINITY;
				PntCorrsPose * newpose2 = NULL;
				if (resolveIK)
				{
					newpose2 = new PntCorrsPose(*newpose);
					solver->SolveIKProblem(newpose2, true);
					if (expensiveLocalSearch)
						newpose2 = LocalSearchStep(newpose2, 0.2*1.5);	// TODO: 0.2 == gridsize
					eval2 = GetEnergy()->ComputeEnergy(newpose2, einfo);
					if (eval2 < bestEnergy)
						bestEnergy = eval2;
				}
				
				if (DEBUG_VIS_BOUNDS)
				{
					minbound.push_back(population[i].value);
					evalues.push_back(eval);
				}
				
				population2.push_back(Sortable(eval, newpose));
				if (newpose2!=NULL)
					population2.push_back(Sortable(eval2, newpose2));
				
//				if (dealWithSymmetric)	// if assignment is symmetric -- add its counterpart
//				{
//					for (int symmM=0; symmM<newpose->NMatches(); symmM)
//					{
//						newpose->Region(symmM)
//					}
//				}
			}
			//delete oldpose;	// memory leak - it's' ok
		}
		std::cout<<std::endl;
		if (DEBUG_VIS_BOUNDS)
		{
			std::cout<<"B=[";
			for (int i=0; i<(int)minbound.size(); i++)
				std::cout<<minbound[i]<<" ";
			std::cout<<"];"<<std::endl;
			std::cout<<"A=[";
			for (int i=0; i<(int)evalues.size(); i++)
				std::cout<<evalues[i]<<" ";
			std::cout<<"];"<<std::endl;
			std::cout<<"plot(A, 'r'), hold on, plot(B, 'b');"<<std::endl;
		}
		
		std::cout<<"First Bound: "<<population.size()<<" -> "<<population2.size()<<std::endl;
		population.clear();
		population = population2;
		population2.clear();
		std::sort(population.begin(), population.end());
		
		// local search
		if (cheapLocalSearchPrecluster)
		{
			std::sort(population.begin(), population.end());
			for (int pop=0; pop<vkMin(nlocalsearch, (int)population.size()); pop++)
			{
				PntCorrsPose * updated = LocalSearchStep((PntCorrsPose*)population[pop].ptr, 0.2*1.5); // TODO: 0.2 == gridsize
				double eval = GetEnergy()->ComputeEnergy(updated, einfo);
				population[pop] = Sortable(eval, updated);
			}
			std::sort(population.begin(), population.end());
		}
		
		if (clusterRemaining)
		{
			std::vector<std::vector<Sortable> > input;
			std::vector<std::vector<Sortable> > output;
			input.push_back(std::vector<Sortable>());
			for (int i=0; i<(int)population.size(); i++)
				input[0].push_back(Sortable(population[i].value, population[i].ptr));
			ClusterPoses(input, output);
			population.clear();
			std::cout<<"ClusteredPoses("<<output.size()<<"): ";
			for (int i=0; i<(int)output.size(); i++)
			{
				std::cout<<output[i].size()<<" "<<std::flush;;
				assert(output[i].size()>0);
				population.push_back(output[i][0]);
			}
			std::cout<<std::endl;
		}
		
		// local search
		if (cheapLocalSearchPostcluster)
		{
			std::sort(population.begin(), population.end());
			for (int pop=0; pop<vkMin(nlocalsearch, (int)population.size()); pop++)
			{
				PntCorrsPose * updated = LocalSearchStep((PntCorrsPose*)population[pop].ptr, 0.2*1.5);	// TODO: clustering threshold
				double eval = GetEnergy()->ComputeEnergy(updated, einfo);
				population[pop] = Sortable(eval, updated);
			}
			std::sort(population.begin(), population.end());
		}
	}

	
	
	const InteractorPose * objpose = m_Interactors[Interactor::ACT_Object]->GetEnumeratedPose(0);
	R3Vector objTrans = R3Vector(0, 0, -m_Interactors[Interactor::ACT_Object]->Ground(objpose));

	std::sort(population.begin(), population.end());	// should be useless... investigate
	std::cout<<"Final Poses: "<<population.size()<<std::endl;
	for (int i=0; i<vkMin((int)population.size(), nposes); i++)	// top 100
	{
		Binding * binding = new Binding(m_Interactors[Interactor::ACT_Actor], m_Interactors[Interactor::ACT_Object]);
		m_Bindings.push_back(binding);
		binding->SetPose(Interactor::ACT_Object, objpose);
		binding->SetXForm(Interactor::ACT_Object, new RigidAlign(objTrans, R3Vector(0,0,1), 0));
		binding->SetPose(Interactor::ACT_Actor, (PntCorrsPose*)population[i].ptr);
		//EnergyInfo einfo;
		binding->SetEnergy(GetEnergy()->ComputeEnergy((PntCorrsPose*)population[i].ptr, einfo));
		binding->SetEnergy(population[i].value);
	}
		
	m_CandidateCorrs.clear();
	m_CandidateCorrValues.clear();
	int last = (int)m_CandidateCorrs.size();
	m_CandidateCorrs.push_back(std::vector<std::vector<ShapePoint> >());
	m_CandidateCorrValues.push_back(std::vector<std::vector<double> >());
	for (int r=0; r<nactorpoints; r++)
	{
		m_CandidateCorrs[last].push_back(std::vector<ShapePoint>());
		m_CandidateCorrValues[last].push_back(std::vector<double>());
		std::vector<Sortable> sortedpnts;
		for (int p=0; p<objpnts->NumPoints(); p++)
		{
			double value = objvals->Value(objpnts->GetPoint(p), r);
			if (value > valThreshold)
				sortedpnts.push_back(Sortable(-value, NULL, p));
		}
		
		std::sort(sortedpnts.begin(), sortedpnts.end());
		for (int p=0; p<(int)sortedpnts.size(); p++)
		{
			m_CandidateCorrs[last][r].push_back(ShapePoint(objpnts->GetPoint(sortedpnts[p].id)));
			m_CandidateCorrValues[last][r].push_back(sortedpnts[p].value);
		}
	}
	
	m_Timer.finishedProcess("00_BnB2");
}











