#include "IKSolver.h"

#include "LinAlgVectorReal.h"
#include "LinAlgMatrixReal.h"
#include "BindingManager.h"
#include "PosePriorGauss.h"
#include "PntCorrsPose.h"

IKStats::IKStats()
{
	m_iNSolverCalled = 0;
	m_iNIterations = 0;
	m_iNTermDueToEnegy = 0;
	m_iNTermDueToAngle = 0;
}

VKString IKStats::GetStatsStr() const
{
	VKString str = "";
	str += VKString("NCalls: ")+VKString::number(m_iNSolverCalled)+"\n";
	str += VKString("   Ave NIter: ")+VKString::number(m_iNIterations)+"\n";
	str += VKString("   TermEnergy: ")+VKString::number(m_iNTermDueToEnegy)+"\n";
	str += VKString("   TermAngle: ")+VKString::number(m_iNTermDueToAngle)+"\n";
	return str;
}



IKSolver::IKSolver(double wEndEffector, double wPosePrior, double wPoseSymmetry, PosePrior * prior)
{
	m_pIKStats = new IKStats();
	m_XFormMask =IK_RigidTranslate | IK_RigidRotateAroundUp | IK_RigidRotateFrontBack | IK_RigidRotateLeftRight;
	m_fMaxEnergyErr = 0.01;
	m_fMaxAngleStep = M_PI * 0.2;
	m_fMaxDistStep = 0.02;
//	m_fMaxAngleChange = M_PI * 0.01;
	m_fMaxAngleChange = 0;
	m_iIterations = 1000;
	m_iMaxConvergenceLevels = 6;
	
	m_wEndEffectorDist = wEndEffector;
	m_wPosePrior = wPosePrior;
	m_wSymmetryPrior = wPoseSymmetry;
	m_pPrior = prior;
	m_SolverFlags = 0;
	
	if (m_wEndEffectorDist>0)
		m_SolverFlags = m_SolverFlags | IK_EndEffectors;
	if ((m_wPosePrior>0 || m_wSymmetryPrior>0) && m_pPrior!=NULL)
		m_SolverFlags = m_SolverFlags | IK_PosePrior;
	
//	m_SolverFlags = m_SolverFlags | IK_RigidTranslate | IK_RigidRotateAroundUp | IK_RigidRotateFrontBack | IK_RigidRotateLeftRight;
	m_SolverFlags = m_SolverFlags | IK_RigidTranslate | IK_RigidRotateAroundUp;
	m_SolverFlags = m_SolverFlags | IK_QuickPruning | IK_ConditionedOnPoseID;
	
	
	m_iNInitSkeletonParams = 0;
	if ((m_SolverFlags & m_XFormMask) == (IK_RigidTranslate | IK_RigidRotateAroundUp))
		m_iNInitSkeletonParams = 4;
	else if ((m_SolverFlags & m_XFormMask) == (IK_RigidTranslate | IK_RigidRotateAroundUp | IK_RigidRotateFrontBack | IK_RigidRotateLeftRight))
		m_iNInitSkeletonParams = 16;	// full rigid transform
	else
		assert(false);
}

void IKSolver::FixPoseParameterization(PntCorrsPose * corrs) const
{
	if (!corrs->IsParamsInit(NJointParams()) || corrs->GetPoseParamsFlags()!=(m_SolverFlags&m_XFormMask))
	{
		int fullrigid = IK_RigidTranslate | IK_RigidRotateAroundUp | IK_RigidRotateFrontBack | IK_RigidRotateLeftRight;
		int uponly = IKSolver::IK_RigidTranslate|IKSolver::IK_RigidRotateAroundUp;
		int origninit = -1;
		if (corrs->GetPoseParamsFlags() == uponly)
			origninit = 4;
		else if (corrs->GetPoseParamsFlags()==fullrigid)
			origninit = 16;
		else
			assert(false);
		
		int diff = m_iNInitSkeletonParams-origninit;
		int nparamssrc = NJointParams()-diff;
		assert(corrs->IsParamsInit(nparamssrc));
		std::vector<double> params;
		for (int i=0; i<nparamssrc; i++)
			params.push_back(corrs->GetParam(i));
		corrs->InitParams(NJointParams(), m_SolverFlags);
		for (int i=m_iNInitSkeletonParams; i<NJointParams(); i++)
			corrs->Param(i) = params[i-diff];
	}
	
	for (int i=m_iNInitSkeletonParams; i < NJointParams(); i++)
		corrs->Param(i) = PosePointGauss::GetNearestAngle(0, corrs->Param(i));
}

void IKSolver::ReadSkeleton(const VKString & skeletonFile, const VKString & regionNamesFile)
{
	double scale = 1.;
	// NOTE: proportions in the file obtained from here:
	//  http://mech.utah.edu/ergo/pages/Educational/safety_modules/ctd-anthropometry/index_files/image002.gif
	VKString tempStr;
	// read regions
	std::map<VKString, int> name2regionid;
	std::ifstream textStreamRegions(regionNamesFile.c_str());
	assert(textStreamRegions.is_open());
	while (!textStreamRegions.eof())
	{
		tempStr.readLine(textStreamRegions);
		if (tempStr!="")
		{
			if (tempStr.startsWith("#"))
				continue;
			VKStringList splitLine = tempStr.split(" ");
			assert(splitLine.count()>=3);
			name2regionid[splitLine[0]] = m_RegionNames.count();
			m_RegionNames.push_back(splitLine[0]);
		}
	}
	textStreamRegions.close();
	
	// read skeleton
	std::ifstream textStream(skeletonFile.c_str());
	assert(textStream.is_open());
	
	for (int i=0; i<m_RegionNames.count(); i++)
		m_Pnts2Skeleton.push_back(-1);
	
	std::map<VKString, int> name2jointid;
	m_ParamOffsetByLinkID.clear();
	int param_offset = m_iNInitSkeletonParams;	// offset by translation
	m_ParamOffsetByLinkID.push_back(param_offset);
	tempStr.readLine(textStream);
	VKStringList scaleLine = tempStr.split(" ");
	assert(scaleLine.count()==2 && scaleLine[0]=="scale");
	bool ok;
	scale = scaleLine[1].toDouble(&ok);    assert(ok);
	while (!textStream.eof())
	{
		tempStr.readLine(textStream);
		if (tempStr=="")
			continue;
		
		// read line
		VKStringList strSplit = tempStr.split(" ");
		assert(strSplit.count()>=4);
		VKString parentName = strSplit[0].replace(" ", "");
		VKString childName = strSplit[1].replace(" ", "");
		//std::cout<<"Link: "<<parentName.c_str()<<" "<<childName.c_str()<<std::endl;
		VKString offsetAxis = strSplit[2].replace(" ", "");
		double offset = strSplit[3].replace(" ", "").toDouble(&ok);  assert(ok);
		offset *= scale;
		//bool neg = (offset < 0);
		//offset = vkAbs(offset);	// vk: change
		SkeletonLink::JointType type = SkeletonLink::Str2Joint(strSplit[4].replace(" ", ""));
		VKString rotationAxis = "none";
		std::vector<double> minangles;
		std::vector<double> maxangles;
		if (type==SkeletonLink::JOINT_Hinge)
		{
			assert(strSplit.count()>=8);
			rotationAxis = strSplit[5].replace(" ", "");
			minangles.push_back(strSplit[6].toDouble(&ok)*M_PI);	  assert(ok);
			maxangles.push_back(strSplit[7].toDouble(&ok)*M_PI);	  assert(ok);
		}
		else if (type==SkeletonLink::JOINT_BallAndSocket)
		{
			assert(strSplit.count()>=10);
			minangles.push_back(strSplit[5].toDouble(&ok)*M_PI);	  assert(ok);
			maxangles.push_back(strSplit[6].toDouble(&ok)*M_PI);	  assert(ok);
			minangles.push_back(strSplit[7].toDouble(&ok)*M_PI);	  assert(ok);
			maxangles.push_back(strSplit[8].toDouble(&ok)*M_PI);	  assert(ok);
			minangles.push_back(strSplit[9].toDouble(&ok)*M_PI);	  assert(ok);
			maxangles.push_back(strSplit[10].toDouble(&ok)*M_PI);	  assert(ok);
		}
		
		param_offset += SkeletonLink::NDims(type);
		m_ParamOffsetByLinkID.push_back(param_offset);

		SkeletonJoint * parent = NULL;
		SkeletonJoint * child = NULL;
		
		// add parent node if necessary
		std::map<VKString, int>::iterator iter1 = name2jointid.find(parentName);
		if (iter1 == name2jointid.end())
		{
			parent = new SkeletonJoint();
			parent->m_iJointID = (int)m_Joints.size();
			parent->m_sJointName = parentName;
			parent->m_pParent = NULL;
			name2jointid[parentName] = parent->m_iJointID;
			m_Joints.push_back(parent);
		}
		else
			parent = m_Joints[iter1->second];
		
		// add child node if necessary
		std::map<VKString, int>::iterator iter2 = name2jointid.find(childName);
		if (iter2 == name2jointid.end())
		{
			child = new SkeletonJoint();
			child->m_iJointID = (int)m_Joints.size();
			child->m_sJointName = childName;
			child->m_pParent = NULL;
			name2jointid[childName] = child->m_iJointID;
			m_Joints.push_back(child);
		}
		else
			child = m_Joints[iter2->second];
		
		// set link parameters
		SkeletonLink * link = new SkeletonLink();
		link->m_JointType = type;
		if (offsetAxis=="X")
			link->m_ChildOffset = R3Vector(offset, 0, 0);
		else if (offsetAxis=="Y")
			link->m_ChildOffset = R3Vector(0, offset, 0);
		else if (offsetAxis=="Z")
			link->m_ChildOffset = R3Vector(0, 0, offset);
		
		link->m_RotationAxis = R3Vector(0,0,0);
		if (rotationAxis=="X")
			link->m_RotationAxis = R3Vector(1, 0, 0);
		else if (rotationAxis=="Y")
			link->m_RotationAxis = R3Vector(0, 1, 0);
		else if (rotationAxis=="Z")
			link->m_RotationAxis = R3Vector(0, 0, 1);
		
		link->m_pParent = parent;
		link->m_pChild = child;
		link->m_iLinkID = (int)m_Links.size();
		
		link->m_MinAngleValue = minangles;
		link->m_MaxAngleValue = maxangles;
		
		m_Links.push_back(link);

		// update nodes with the new link
		assert(parent!=NULL && child!=NULL);
		assert(child->m_pParent==NULL);
		child->m_pParent = link;
		parent->m_Children.push_back(link);
	}
	
	for (int i=0; i<(int)m_Pnts2Skeleton.size(); i++)
	{
		if (m_Pnts2Skeleton[i]==-1)
		{
			assert(i<m_RegionNames.count());
			std::map<VKString, int>::iterator iterjoint = name2jointid.find(m_RegionNames[i]);
			if (iterjoint == name2jointid.end())
				std::cout<<"[ERROR] Invalid contact point: "<<m_RegionNames[i].c_str()<<std::endl;
			assert(iterjoint != name2jointid.end());
			m_Pnts2Skeleton[i] = iterjoint->second;
		}
	}
	textStream.close();
}

bool IKSolver::FillRenderableSkeleton(const InteractorPose * poseIn, bool updatePose,
									  std::vector<R3Point> & pnts, std::vector<int> & edges) const
{
	assert(poseIn->GetPoseType()==InteractorPose::TYPE_PntCorrsPose);
	PntCorrsPose * pose = (PntCorrsPose*)poseIn;
	if (updatePose)
		ForwardKinematics(pose);
	for (int i=0; i<(int)m_Joints.size(); i++)
		pnts.push_back(pose->Position(i));
	
	for (int i=0; i<(int)m_Links.size(); i++)
	{
		edges.push_back(m_Links[i]->m_pParent->m_iJointID);
		edges.push_back(m_Links[i]->m_pChild->m_iJointID);
	}
	return true;
}


void IKSolver::InitPoseToSitting(PntCorrsPose * pose) const
{
	InitPoseToZero(pose);
	assert(35 < NJointParams());
	// sitting pose
	// left shoulder
	int o = 13 - 4 + m_iNInitSkeletonParams;
	pose->Param(o) = 0.39 * M_PI;
	pose->Param(o+1) = -0.12 * M_PI;
	pose->Param(o+2) = 0.36 * M_PI;
	
	// left elbow
	pose->Param(o+3) = -0.5 * M_PI;
	
	// right shoulder
	pose->Param(o+7) = -0.39 * M_PI;
	pose->Param(o+8) = -0.12 * M_PI;
	pose->Param(o+9) = -0.36 * M_PI;

	// right elbow
	pose->Param(o+10) = 0.5 * M_PI;

	// left hip
	pose->Param(o+17) = 0;
	pose->Param(o+18) = -0.5 * M_PI;
	pose->Param(o+19) = 0;
	// left knee
	pose->Param(o+20) = 0.5 * M_PI;
	// right hip
	pose->Param(o+22) = 0;
	pose->Param(o+23) = -0.5 * M_PI;
	pose->Param(o+24) = 0;
	// right knee
	pose->Param(o+25) = 0.5 * M_PI;
}

void IKSolver::InitPoseToZero(PntCorrsPose * pose) const
{
	std::cout<<"Init to zero"<<std::endl;
	pose->InitParams(NJointParams(), m_SolverFlags);
	for (int i=0; i<NJointParams(); i++)
		pose->Param(i) = 0;
}

void IKSolver::GetEnergy(const PntCorrsPose * corrs, EnergyInfo & energy) const
{
	if (m_pPrior!=NULL && !m_pPrior->QuickValidityTest(corrs) && (m_SolverFlags&IK_QuickPruning)!=0)
	{
		energy.m_fEndEffectorsEnergy = RN_INFINITY;
		energy.m_fPosePriorEnergy = RN_INFINITY;
		energy.m_fSymmetryEnergy = RN_INFINITY;
	}
	else
		FillDirectionVector(corrs, NULL, energy, NULL);
}

void IKSolver::GetPerContactEnergy(const PntCorrsPose * corrs, int refReg, const std::vector<int> & trgRegions,
								   std::vector<double> & energies) const
{
	int nlinks = (int)m_Links.size();
	
	// create counters for all links
	std::vector<int> linkTraverses;
	std::vector<double> linkPriorEnergy;
	std::vector<double> linkSymmEnergy;
	std::vector<int> linkSymmCoupling;
	for (int i=0; i<nlinks; i++)
	{
		linkTraverses.push_back(0);
		linkPriorEnergy.push_back(0);
		linkSymmEnergy.push_back(0);
		linkSymmCoupling.push_back(-1);
	}
		
	// compute energy along the way (for each link)
	//assert ((m_SolverFlags&IK_PosePrior)!=0 && (m_wPosePrior+m_wSymmetryPrior)>0);
	EnergyInfo info;
	for (int t = 0; t<m_pPrior->NTerms(); t++)
	{
		double eval = m_pPrior->EnergyTerm(corrs, t, info);
		int linkid = info.m_LinkID;
		int symmid = info.m_SymmetricLink;
		
		if (m_pPrior->IsSymmetryTerm(t))
		{
			assert(linkid>=0 && (symmid>=0 || symmid==-2));
			linkSymmEnergy[linkid] = eval * m_wSymmetryPrior;
			linkSymmCoupling[linkid] = symmid;
		}
		else
		{
			assert(linkid>=0);
			linkPriorEnergy[linkid] = eval * m_wPosePrior;
		}
	}
	
	// trace path for each reg ref + update link counters
	std::vector<std::vector<int> > paths;
	for (int i=0; i<(int)trgRegions.size(); i++)
	{
		paths.push_back(std::vector<int>());
		FillPath(refReg, trgRegions[i], paths[i]);
		for (int p=0; p<(int)paths[i].size(); p++)
			linkTraverses[paths[i][p]]++;
	}

	// store energy for each path (region)
	for (int i=0; i<(int)trgRegions.size(); i++)
	{
		double energy = 0;
//		if (refReg>0)
//			std::cout<<"Energy["<<refReg<<", "<<trgRegions[i]<<"]: "<<std::endl;
		for (int p=0; p<(int)paths[i].size(); p++)
		{
			int linkid = paths[i][p];
			int counter = linkTraverses[linkid];
			double ePrior = linkPriorEnergy[linkid];
			double eSymm = linkSymmEnergy[linkid];
			int symmlink = linkSymmCoupling[linkid];
			bool symmCoupleDouble = false;
			bool symmCoupleSingle = false;
			
			// check symmetry
			if (eSymm>0)
			{
				if (symmlink==-2)
					symmCoupleSingle = true;
				else
				{
					assert(symmlink>=0);
					for (int p2=0; p2<(int)paths[i].size(); p2++)
						if (paths[i][p2]==symmlink)
							symmCoupleDouble = true;
				}
			}
				
			// NOTE: sum of energies might not equal to full energy since symm link is rejected sometimes
			//		 should work for all our cases (pelvis -> limbs, limb -> limb)
			energy += ePrior / counter;
			if (symmCoupleSingle)
				energy += eSymm / counter;	// always include penalty
			else if (symmCoupleDouble)
				energy += eSymm;
//			if (refReg>0)
//			{
//				std::cout<<"   "<<LinkStr(linkid).c_str()<<": "<<ePrior<<"/"<<counter;
//				std::cout<<" + "<<(symmCoupleSingle?eSymm:0)<<"/"<<counter<<" + "<<(symmCoupleDouble?eSymm:0)<<std::endl;
//			}
		}
		energies.push_back(energy);
	}
//	if (refReg>0)
//		exit(0);
}

void IKSolver::FillPath(int fromRegion, int toRegion, std::vector<int> & pathLinks) const
{
	if (fromRegion==toRegion)
		return;
	
	// skeleton ids
	int skeletonFrom = m_Pnts2Skeleton[fromRegion];
	int skeletonTo = m_Pnts2Skeleton[toRegion];

	std::vector<int> parentsJoint2Path;
	for (int p=0; p<(int)m_Joints.size(); p++)
		parentsJoint2Path.push_back(-1);
	int joint = skeletonFrom;
	// path from 'from' to root
	while (joint>=0)
	{
		parentsJoint2Path[joint] = (int)pathLinks.size();
		
		if (m_Joints[joint]->m_pParent==NULL)
			joint = -1;
		else
		{
			SkeletonLink * link = m_Joints[joint]->m_pParent;
			pathLinks.push_back(link->m_iLinkID);
			joint = link->m_pParent->m_iJointID;
		}
	}
	
	// path from 'to' to root
	std::vector<int> path2;
	joint = skeletonTo;
	while (joint>=0)
	{
		if (parentsJoint2Path[joint]>=0)	// common ancestor
		{
			pathLinks.erase(pathLinks.begin()+parentsJoint2Path[joint], pathLinks.end());
			for (int p=((int)path2.size())-1; p>=0; p--)
				pathLinks.push_back(path2[p]);
			break;
		}
		SkeletonLink * link = m_Joints[joint]->m_pParent;
		assert(link!=NULL);
		joint = link->m_pParent->m_iJointID;
		path2.push_back(link->m_iLinkID);
	}
}

VKString IKSolver::PathStr(const std::vector<int> & pathLinks) const
{
	VKString str = "";
	for (int i=0; i<(int)pathLinks.size(); i++)
		str +=LinkStr(pathLinks[i])+", ";
	return str;
}

VKString IKSolver::LinkStr(int linkid) const
{
	VKString j1 = m_Links[linkid]->m_pParent->m_sJointName;
	VKString j2 = m_Links[linkid]->m_pChild->m_sJointName;
	return j1+"->"+j2;
}

PntCorrsPose * IKSolver::CreateCopy(const PntCorrsPose * pose) const
{
	int nTheta = NJointParams();
	PntCorrsPose * newpose = new PntCorrsPose(pose);
	bool validparams = pose->IsParamsInit(NJointParams());
	if (validparams)
	{
		newpose->InitParams(NJointParams(), m_SolverFlags);
		for (int i=0; i<NJointParams(); i++)
			newpose->Param(i) = pose->GetParam(i);
	}
	else
		InitPoseToZero(newpose);
	return newpose;
}

void IKSolver::FindBestInitPose(PntCorrsPose * pose) const
{
	if (m_pPrior!=NULL)
	{
		IKConvergenceParams convergence;
		EnergyInfo energy;
		int nE = NTargetParams(pose);
		int nTheta = NJointParams();
		LinAlgVectorReal bestParams(nTheta);
		LinAlgVectorReal direction(nE);
		pose->InitParams(NJointParams(), m_SolverFlags);
		double bestEnergy = RN_INFINITY;
		int niter = m_iIterations;
		((IKSolver*)this)->m_iIterations = 10;
		for (int i=0; i<m_pPrior->NSuggestedInitPoses(); i++)
		{
			m_pPrior->InitPose(i, pose);
			SolveIKProblem(pose, false);
			FillDirectionVector(pose, &direction, energy, &convergence);
			double newenergy = energy.m_fPosePriorEnergy+energy.m_fEndEffectorsEnergy+energy.m_fSymmetryEnergy;
			if (newenergy < bestEnergy)
			{
				bestEnergy = newenergy;
				for (int p=0; p<nTheta; p++)
					bestParams(p) = pose->Param(p);
			}
		}
		((IKSolver*)this)->m_iIterations = niter;
		for (int i=0; i<nTheta; i++)
			pose->Param(i) = bestParams.GetValue(i);
	}
}

bool IKSolver::SolveIKProblem(PntCorrsPose * pose, bool reinit) const
{
	if (!pose->IsParamsInit(NJointParams()))
		pose->InitParams(NJointParams(), m_SolverFlags);
	
	if (!pose->IsPositionsInit((int)m_Joints.size()))
		ForwardKinematics(pose);

	if (m_pPrior!=NULL && (m_SolverFlags&IK_QuickPruning)!=0 && !m_pPrior->QuickValidityTest(pose))
		return false;
	
	int nTheta = NJointParams();
	
	// using: http://groups.csail.mit.edu/drl/journal_club/papers/033005/buss-2004.pdf
	
	int nE = NTargetParams(pose);
	
	if (nE==0)
		return true;	// cannot do anything meaningful
	
	if (reinit)
		m_pIKStats->m_iNSolverCalled++;
	
	int curriter = 0;
	
	double maxdTheta = 0;
	double prev_error = RN_INFINITY;
	LinAlgVectorReal bestParams(nTheta);
	int bestIteration = -1;
	
	double ratio = .98;
	int amortizedDiff = 5;
	EnergyInfo energy;
	
	IKConvergenceParams convergence;
	convergence.m_iCurrentIteration = 0;
	convergence.m_iCurrentConvergenceLevel = 0;
	convergence.m_fCurrDistStep = m_fMaxDistStep;
	convergence.m_fCurrAngleStep = m_fMaxAngleStep;
//	convergence.m_fCurrDistStep = m_fMaxDistStep * pow(2., m_iMaxConvergenceLevels-2);
//	convergence.m_fCurrAngleStep = m_fMaxAngleStep * pow(2., m_iMaxConvergenceLevels-2);
	
	LinAlgMatrixReal jacobian(nE, nTheta);
	LinAlgVectorReal direction(nE);
	LinAlgVectorReal dtheta(nTheta);
	
	if (reinit)
		FindBestInitPose(pose);
	if ((m_SolverFlags & IKSolver::IK_DisableIK)!=0)
	{
		assert(m_pPrior!=NULL);
		FindBestRigidXForm(pose);
		ForwardKinematics(pose);
		return true;
	}
	
	bool converged = false;
	
	double curr_error = RN_INFINITY;
	convergence.m_iCurrentIteration = 0;
	prev_error = RN_INFINITY;
	while (convergence.m_iCurrentIteration++ < m_iIterations)
	{
		m_pIKStats->m_iNIterations++;
		
		// CLEAR JACOBIANS AND DISPLANCEMENTS
		jacobian.SetValues(0);
		for (int i=0; i<direction.Rows(); i++)
			direction.SetValue(i,0);
		for (int i=0; i<dtheta.Rows(); i++)
			dtheta.SetValue(i,0);
		
		// FIND JOINT POSITIONS & VECTORS TO TARGET
		FindBestRigidXForm(pose);	// rigid alignment
		ForwardKinematics(pose);	// forward kinematics
		FillDirectionVector(pose, &direction, energy, &convergence);	// find direction + energy
		curr_error = energy.m_fEndEffectorsEnergy + energy.m_fPosePriorEnergy + energy.m_fSymmetryEnergy;
		
		//std::cout<<"E["<<convergence.m_iCurrentIteration<<"]: "<<prev_error<<" -> "<<curr_error<<" CNT: "<<(curr_error / prev_error >= ratio)<<std::endl;
		// CHECK CONVERGENCE & STEPS
		if (curr_error < prev_error)	// update error
		{
			if (curr_error/prev_error<ratio)
				bestIteration = convergence.m_iCurrentIteration;
			for (int i=0; i<nTheta; i++)
				bestParams(i) = pose->Param(i);
			prev_error = curr_error;
		}
		if (curr_error/prev_error>=ratio && convergence.m_iCurrentIteration - bestIteration > amortizedDiff)
		{
			m_pIKStats->m_iNTermDueToEnegy++;
			break;
		}
		
		// DO GRADIENT DESENT
		FillJacobian(pose, jacobian);
		SolveDeltaTheta(jacobian, direction, dtheta);
		assert(pose->IsParamsInit(NJointParams()));
		assert(dtheta.Rows()==(int)NJointParams());
		maxdTheta = 0;
//		std::cout<<"dTheta: [";
		for (int i=0; i<nTheta; i++)
		{
			maxdTheta = vkMax(maxdTheta, vkAbs(dtheta.GetValue(i)));
			pose->Param(i) += dtheta.GetValue(i);
//			std::cout<<dtheta.GetValue(i)<<" ";
		}
//		std::cout<<std::endl;
		
		if (maxdTheta < m_fMaxAngleChange)
		{
			m_pIKStats->m_iNTermDueToAngle++;
			break;
		}
	}
	assert(bestIteration>=0);
//	std::cout<<"Best: "<<bestIteration<<std::endl;
	for (int i=0; i<nTheta; i++)
		pose->Param(i) = bestParams.GetValue(i);
	return converged;
}

void IKSolver::FindBestRigidXForm2D(PntCorrsPose * pose) const
{
//	std::cout<<"ParamsIn: [";
//	for (int i=0; i<NJointParams(); i++)
//		std::cout<<pose->Param(i)<<" ";
//	std::cout<<"]"<<std::endl;
	
	// forward kinematics without the initial transformation
	assert(pose->IsParamsInit(NJointParams()));
	for (int i=0; i<m_iNInitSkeletonParams; i++)
		pose->Param(i) = 0;
	ForwardKinematics(pose);
	
//	std::cout<<"Params Inter 1: [";
//	for (int i=0; i<NJointParams(); i++)
//		std::cout<<pose->Param(i)<<" ";
//	std::cout<<"]"<<std::endl;

	
	// establish corresponding points
	int npoints, nground;
	NEndEffectors(pose, &npoints, &nground);
	
	R3Point * points1 = new R3Point[npoints];	// skeleton points
	R3Point * points2 = new R3Point[npoints]; 	// object points
	R3Point * groundPoints = new R3Point[nground]; 	// skeleton (to ground)
	
	int idpnts = 0;
	int idground = 0;
	for (int i=0; i<(int)pose->NMatches(); i++)
	{
		int rid = pose->Region(i);
		assert(rid < (int)m_Pnts2Skeleton.size());
		if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Point)
			|| pose->Prim(i).IsType(TargetPrimitive::PRIM_PointMeshless))	// actual point
		{
			assert(idpnts < npoints);
			points1[idpnts] = pose->Position(m_Pnts2Skeleton[rid]);
			points2[idpnts] = pose->Prim(i).GetPosition();
			if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Point))
				pose->m_NormalizingXForm.Apply(points2[idpnts]);	// TODO: depricate
			idpnts++;
		}
		else if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Ground))
		{
			assert(idground < nground);
			groundPoints[idground] = pose->Position(m_Pnts2Skeleton[rid]);
			idground++;
		}
	}
	
	// Align centers
	R3Point center1(0,0,0);	// skeleton centroid
	R3Point center2(0,0,0);	// object centroid
	double height = 0;	// skeleton average height
	double sumPoints = 0;
	for (int i=0; i<npoints; i++)
	{
		double w = 1.;
		center1 += points1[i] * w;
		center2 += points2[i] * w;
		sumPoints += w;
	}
	double sumGround = 0;
	for (int i=0; i<nground; i++)
	{
		double w = 1.;
		height += groundPoints[i].Z() * w;
		sumGround += w;
	}
	if (sumPoints>0)
	{
		center1 /= sumPoints;
		center2 /= sumPoints;
	}
	if (sumGround>0)
	{
		center1[2] = (center1[2]*sumPoints + height)/(sumPoints+sumGround);	// skeleton
		center2[2] = (center2[2]*sumPoints)/(sumPoints+sumGround);	// skeleton
	}
	
	R3Matrix R(0,0,0,0,0,0,0,0,0);
	if (npoints>0)
	{
		// Compute cross-covariance of two point sets
		// NOTE: rotation is not affected by ground plane
		// 		Compute covariance matrix
		RNScalar w;
		RNScalar m[4] = { 0 };
		for (int i = 0; i < npoints; i++)
		{
			R3Vector p1 = (points1[i] - center1);
			R3Vector p2 = (points2[i] - center2);
			w = 1.;
			for(int j = 0; j < 2; j++)
				for(int k = 0; k < 2; k++)
					m[k*2 + j]+=p1[j]*p2[k]*w*w;
		}
		// normalize covariance matrix
		for (int j = 0; j < 4; j++) m[j] /= npoints;
		
		// Calculate SVD of covariance matrix
		RNScalar Um[4];
		RNScalar Wm[2];
		RNScalar Vmt[4];
		RNSvdDecompose(2, 2, m, Um, Wm, Vmt);
		
		// https://sakai.rutgers.edu/access/content/group/7bee3f05-9013-4fc2-8743-3c5078742791/material/svd_ls_rotation.pdf
		R3Matrix Ut(Um[0], Um[2], 0, Um[1], Um[3], 0, 0, 0, 1);
		R3Matrix V(Vmt[0], Vmt[2], 0, Vmt[1], Vmt[3], 0, 0, 0, 1);
		R3Matrix VUt = V * Ut;
		R3Matrix D = R3identity_matrix;
		D[1][1] = R3MatrixDet2(VUt[0][0], VUt[0][1], VUt[1][0], VUt[1][1]);
		R = V * D * Ut;
	}
	
	// resulting transform
	R4Matrix rotation = R4identity_matrix;
	rotation[0][0] = R[0][0];	// cos
	rotation[0][1] = -R[0][1];	// -sin
	rotation[1][0] = -R[1][0];	// sin
	rotation[1][1] = R[1][1];	// cos
	R4Matrix result = R4identity_matrix;
	result.Translate(center2.Vector());
	result.Transform(rotation);
	result.Translate(-center1.Vector());
	
	// store best transformation
	for (int i=0; i<3; i++)
		pose->Param(i) = result[i][3];
	
	// store best rotation
	pose->Param(3) = atan2(result[1][0], result[0][0]);	// sin / cos
	delete [] points1;
	delete [] points2;
	delete [] groundPoints;

//	std::cout<<"ParamsOut: [";
//	for (int i=0; i<NJointParams(); i++)
//		std::cout<<pose->Param(i)<<" ";
//	std::cout<<"]"<<std::endl;

}

void IKSolver::FindBestRigidXFormFull(PntCorrsPose * pose) const
{
	// forward kinematics without the initial transformation
	assert(pose->IsParamsInit(NJointParams()));
	for (int i=0; i<m_iNInitSkeletonParams; i++)
		pose->Param(i) = 0;
	ForwardKinematics(pose);
	
	// establish corresponding points
	int npoints, nground;
	NEndEffectors(pose, &npoints, &nground);
	
	R3Point * points1 = new R3Point[npoints+nground];	// skeleton points
	R3Point * points2 = new R3Point[npoints+nground]; 	// object points
	
	int idpnts = 0;
	
	for (int i=0; i<(int)pose->NMatches(); i++)
	{
		int rid = pose->Region(i);
		assert(rid < (int)m_Pnts2Skeleton.size());
		if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Point)
			|| pose->Prim(i).IsType(TargetPrimitive::PRIM_PointMeshless))	// actual point
		{
			assert(idpnts < npoints);
			points1[idpnts] = pose->GetPosition(m_Pnts2Skeleton[rid]);	// skeleton point
			points2[idpnts] = pose->Prim(i).GetPosition();			// surface point
			if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Point))
				pose->m_NormalizingXForm.Apply(points2[idpnts]);	// TODO: depricate
			idpnts++;
		}
	}

	// ICP (because ground is a surface, not exact correspondence)
	int nmaxicp = 100;
	double prev_error = RN_INFINITY;
	R3Affine xform = R3identity_affine;
	double epsilon = 1e-5;
	for (int iter=0; iter<nmaxicp; iter++)
	{
		int idground = 0;
		double curr_error = 0;
		bool ignore_ground = (npoints>0 && iter==0);
		for (int i=0; i<(int)pose->NMatches(); i++)
		{
			int rid = pose->Region(i);
			assert(rid < (int)m_Pnts2Skeleton.size());
			if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Ground))
			{
				assert(idground < nground);
				points1[npoints+idground] = pose->GetPosition(m_Pnts2Skeleton[rid]);	// skeleton point
				
				points2[npoints+idground] = pose->GetPosition(m_Pnts2Skeleton[rid]);	// ground point
				xform.Apply(points2[npoints+idground]);
				curr_error += vkAbs(points2[npoints+idground].Z());
				points2[npoints+idground].SetZ(0);
				idground++;
			}
		}
		
		std::cout<<"Improvement ["<<iter<<"]: "<<(prev_error-curr_error)<<std::endl;
		
		if (prev_error-curr_error < epsilon && iter > 0)
			break;
		
		if (iter > 0)
			prev_error = curr_error;
		
		R4Matrix m(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);	// recompute matrix
//		if (npoints + nground == 2)
//		{
//			R3Point center1 = R3Point(points2[0].X()+points2[1].X(), points2[0].Y()+points2[1].Y(), points2[0].Z()+points2[1].Z())/2.;
//			R3Point center2 = R3Point(points1[0].X()+points1[1].X(), points1[0].Y()+points1[1].Y(), points1[0].Z()+points1[1].Z())/2.;
//			R3Vector from = points1[1] - center2;
//			R3Vector to = points2[1] - center1;
//
//			m.Translate(center1.Vector());
//			m.Rotate(from, to);
//			m.Translate(-(center2.Vector()));
//		}
//		else if (ignore_ground)
		{
			m = R3AlignPoints(npoints, points2, points1, NULL, true, true, false);
			std::cout<<"Trying to align: "<<npoints<<std::endl;
			for (int i=0; i<npoints; i++)
			{
				std::cout<<"   ["<<points2[i].X()<<" "<<points2[i].Y()<<" "<<points2[i].Z()<<"] ";
				std::cout<<" -> ["<<points1[i].X()<<" "<<points1[i].Y()<<" "<<points1[i].Z()<<"]\n";
			}
		}
//		else
//			m = R3AlignPoints(npoints+nground, points2, points1, NULL, true, true, false);
		xform = R3Affine(m);
		std::cout<<"M:"<<std::endl;
		for (int i=0; i<4; i++)
		for (int j=0; j<4; j++)
			std::cout<<(j==0?"    ":"")<<m[i][j]<<(j==3?"\n":" ");
		
		for (int i=0; i<npoints; i++)
		{
			xform.ApplyInverse(points1[i]);
			std::cout<<"   ["<<points2[i].X()<<" "<<points2[i].Y()<<" "<<points2[i].Z()<<"] ";
			std::cout<<" -> ["<<points1[i].X()<<" "<<points1[i].Y()<<" "<<points1[i].Z()<<"]\n";
		}

//		if (nground==0)
			break;
	}
	
	for (int i=0; i<16; i++)
		pose->Param(i) = xform.Matrix()[i/4][i%4];
	for (int i=0; i<4; i++)
		pose->Param(i*4+i) = pose->Param(i*4+i) - 1.;
	
	delete [] points1;
	delete [] points2;

}

void IKSolver::FindBestRigidXForm(PntCorrsPose * pose) const
{
	if ((m_SolverFlags & m_XFormMask) == (IK_RigidTranslate|IK_RigidRotateAroundUp))
		FindBestRigidXForm2D(pose);
	else if ((m_SolverFlags & m_XFormMask) == (IK_RigidTranslate|IK_RigidRotateAroundUp|IK_RigidRotateFrontBack|IK_RigidRotateLeftRight))
		FindBestRigidXFormFull(pose);
	else
		assert(false);
}



int IKSolver::NJointParams() const
{
	int nparams = m_iNInitSkeletonParams;
	for (int i=0; i<(int)m_Links.size(); i++)
		nparams += SkeletonLink::NDims(m_Links[i]->m_JointType);
	return nparams;
}

int IKSolver::NEndEffectors(const PntCorrsPose * pose, int * npoints, int * nground) const
{
	assert(npoints!=NULL && nground!=NULL);
	*npoints = 0;
	*nground = 0;
	for (int i=0; i<(int)pose->NMatches(); i++)
	{
		if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Point)
			|| pose->Prim(i).IsType(TargetPrimitive::PRIM_PointMeshless))
			(*npoints)++;
		else if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Ground))
			(*nground)++;
	}
	return (*npoints) + (*nground);
}

int IKSolver::NTargetParams(const PntCorrsPose * pose) const
{
	int nparams = 0;
	if (m_SolverFlags&IK_EndEffectors)	// l2 distance for end effectors
	{
		for (int i=0; i<(int)pose->NMatches(); i++)
		{
			if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Point))
				nparams += 3;
			else if (pose->Prim(i).IsType(TargetPrimitive::PRIM_PointMeshless))
				nparams += 3;
			else if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Ground))
				nparams += 1;
			else
				assert(pose->Prim(i).IsType(TargetPrimitive::PRIM_None)
					   || pose->Prim(i).IsType(TargetPrimitive::PRIM_PartialUnassigned));
		}
	}
	if (m_SolverFlags&IK_PosePrior)	// shape prior
	{
		assert(m_pPrior!=NULL);
		nparams += m_pPrior->NTerms();
	}
	return nparams;
}

void IKSolver::SolveDeltaTheta(const LinAlgMatrixReal & J, const LinAlgVectorReal & e, LinAlgVectorReal & dtheta) const
{
	// see: http://groups.csail.mit.edu/drl/journal_club/papers/033005/buss-2004.pdf
	// SOLVE:  dtheta = Jt (J Jt + lambda2 I)^-1 e
	// 		1. Solve for f: (J * Jt + lambda2 * I) * f == e
	// 		2. dtheta = Jt * f

	double lambda2 = pow(0.1, 2.);
	
	LinAlgMatrixReal Jt(J.Cols(), J.Rows());
	J.Transpose(Jt);
	
	LinAlgMatrixReal JJt(J.Rows(), J.Rows());
	J.Multiply(Jt, JJt);
	
	for (int i=0; i<J.Rows(); i++)
		JJt.AddVal(i, i, lambda2);
	LinAlgVectorReal f(J.Rows());
	JJt.Solve(e, f);
	Jt.Multiply(f, dtheta);
}

void IKSolver::FillJacobian(PntCorrsPose * pose, LinAlgMatrixReal & J) const
{
	// Jij = ds_i / dtheta_j
	int nTheta = J.Cols();
	double step = 3.14 / 50.;
	
	ForwardKinematics(pose);
	std::vector<R3Point> s0;
	for (int i=0; i<(int)m_Joints.size(); i++)
		s0.push_back(pose->Position(i));
	
	int currParam = 0;
	assert(pose->IsParamsInit(NJointParams()));
	for (int tid=m_iNInitSkeletonParams; tid < nTheta; tid++)
	{
		// change theta, run forward kinematics
		double origval = pose->Param(tid);
		pose->Param(tid) += step;
		ForwardKinematics(pose);
		
		currParam = 0;
		// fill jacobian
		if (m_SolverFlags&IK_EndEffectors)
		{
			for (int i=0; i<(int)pose->NMatches(); i++)
			{
				if (pose->Prim(i).IsType(TargetPrimitive::PRIM_None)
					|| pose->Prim(i).IsType(TargetPrimitive::PRIM_PartialUnassigned))
					continue;
				assert(currParam < J.Rows());
				assert(pose->Region(i) < (int)m_Pnts2Skeleton.size());
				int skeletonID = m_Pnts2Skeleton[pose->Region(i)];
				R3Vector diff = (pose->Position(skeletonID) - s0[skeletonID]) / step;
				
				if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Point) || pose->Prim(i).IsType(TargetPrimitive::PRIM_PointMeshless))
				{
					J.AddVal(currParam+0, tid, m_wEndEffectorDist*diff.X());
					J.AddVal(currParam+1, tid, m_wEndEffectorDist*diff.Y());
					J.AddVal(currParam+2, tid, m_wEndEffectorDist*diff.Z());
					currParam+=3;
				}
				else if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Ground))
				{
					J.AddVal(currParam+0, tid, m_wEndEffectorDist*diff.Z());
					currParam+=1;
				}
				else
					assert(false);
			}
		}
		pose->Param(tid) = origval;
	}
	
	if (m_SolverFlags&IK_PosePrior)
	{
		std::vector<int> termIDs, paramIDs;
		m_pPrior->NonZeroJacobians(termIDs, paramIDs);
		JacobianInfo info;
		for (int i=0; i<(int)termIDs.size(); i++)
		{
			assert(termIDs[i] <m_pPrior->NTerms());
			double val = m_pPrior->Jacobian(pose, info, termIDs[i], paramIDs[i]);
			
//			if (!m_pPrior->IsSymmetryTerm(termIDs[i]))
//				std::cout<<"   J["<<termIDs[i]<<" "<<paramIDs[i]<<"]: "<<val<<std::endl;
			
			if (m_pPrior->IsSymmetryTerm(termIDs[i]))
				J.AddVal(currParam+termIDs[i], paramIDs[i], m_wSymmetryPrior * val);
			else
				J.AddVal(currParam+termIDs[i], paramIDs[i], m_wPosePrior * val);
		}
		
	}
	ForwardKinematics(pose);
}

bool IKSolver::FillDirectionVector(const PntCorrsPose * pose, LinAlgVectorReal * e, EnergyInfo & info,
								   const IKConvergenceParams * convergence) const
{
	double diffDistThresh = 1.;
	double diffAngleThresh = 1.;
	if (convergence!=NULL)
	{
		diffDistThresh = convergence->m_fCurrDistStep;
		diffAngleThresh = convergence->m_fCurrAngleStep;
	}
	
	double distance = 0;
	int currParam = 0;
	
	info.m_fEndEffectorsEnergy = 0;
	info.m_fPosePriorEnergy = 0;
	info.m_fSymmetryEnergy = 0;
	bool converged = true;
	
	if (info.m_PerPointEnergy.size()==0)
	{
		for (int i=0; i<(int)m_Joints.size(); i++)
			info.m_PerPointEnergy.push_back(0);
	} 
		
	// distance to target points
	if (m_SolverFlags&IK_EndEffectors)
	{
		assert(pose->IsPositionsInit((int)m_Joints.size()));
		for (int i=0; i<(int)pose->NMatches(); i++)
		{
			// target position
			assert(pose->Region(i) < (int)m_Pnts2Skeleton.size());
//			std::cout<<"Get position (ptr="<<pose<<"): "<<i<<std::endl;
//			std::cout<<"     ->  "<<pose->Region(i)<<std::endl;
			int skeletonID = m_Pnts2Skeleton[pose->Region(i)];
//			std::cout<<"     ->  "<<skeletonID<<std::endl;
			R3Point current = pose->GetPosition(skeletonID);
//			std::cout<<"     ->  "<<current.X()<<" "<<current.Y()<<" "<<current.Z()<<" "<<std::endl;

			double diffEnergy = 0;
			double diffLength = 0;
			if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Point)
				||pose->Prim(i).IsType(TargetPrimitive::PRIM_PointMeshless))
			{
				R3Point target = pose->Prim(i).GetPosition();
				pose->m_NormalizingXForm.Apply(target);
				R3Vector diff = target - current;
				diffLength = diff.Length();
				diffEnergy = m_wEndEffectorDist*pow(diff.Length(), 2.);
				
				if (e!=NULL)
				{
					assert(currParam+3 <= e->Rows());
					distance = vkMax(distance, diff.Length());
					if (diff.Length() > diffDistThresh && diffDistThresh>0)
					{
						diff.Normalize();
						diff *= diffDistThresh;
					}	// clamped
					e->SetValue(currParam+0, m_wEndEffectorDist*diff.X());
					e->SetValue(currParam+1, m_wEndEffectorDist*diff.Y());
					e->SetValue(currParam+2, m_wEndEffectorDist*diff.Z());
				}
				currParam += 3;
			}
			else if (pose->Prim(i).IsType(TargetPrimitive::PRIM_Ground))
			{
				double diff = -current.Z();
				diffLength = vkAbs(diff);
				diffEnergy = m_wEndEffectorDist*pow(diff, 2.);
				if (e!=NULL)
				{
					assert(currParam+1 <= e->Rows());
					distance = vkMax(distance, vkAbs(diff));
					if (vkAbs(diff) > diffDistThresh && diffDistThresh>0)
					{
						if (diff>0)
							diff = diffDistThresh;
						else
							diff = -diffDistThresh;
					}
					e->SetValue(currParam, m_wEndEffectorDist*diff);
				}

				currParam += 1;
			}
			else
				assert (pose->Prim(i).IsType(TargetPrimitive::PRIM_None)
						|| pose->Prim(i).IsType(TargetPrimitive::PRIM_PartialUnassigned));
				
			if (diffLength > m_fMaxEnergyErr)
				converged = false;
			
			if (info.m_EnergyVis == ENVIS_EndEffectorDist)
			{
				int rid = pose->Region(i);
				info.m_PerPointEnergy[m_Pnts2Skeleton[rid]] = diffEnergy;
			}
			
			info.m_fEndEffectorsEnergy += diffEnergy;
		}
	}
	
	double angleDiff = 0;
	
	// pose prior
	if (m_SolverFlags&IK_PosePrior)
	{
		for (int t = 0; t<m_pPrior->NTerms(); t++)
		{
			double eval = m_pPrior->EnergyTerm(pose, t, info);
			if (m_pPrior->IsSymmetryTerm(t))
			{
				eval *= m_wSymmetryPrior;
				info.m_fSymmetryEnergy += eval;
			}
			else
			{
				eval *= m_wPosePrior;
				info.m_fPosePriorEnergy += eval;
			}
			if (e!=NULL)
				e->SetValue(currParam + t, -eval);
			
		}
		if (info.m_EnergyVis==ENVIS_AngleAll)
		{
			for (int i=0; i<(int)info.m_PerPointEnergy.size(); i++)
				info.m_PerPointEnergy[i] *= m_wPosePrior;
		}
		else if (info.m_EnergyVis==ENVIS_AngleSymmetry)
		{
			for (int i=0; i<(int)info.m_PerPointEnergy.size(); i++)
				info.m_PerPointEnergy[i] *= m_wSymmetryPrior;
		}
	}
	return converged;
}

void IKSolver::ForwardKinematics(PntCorrsPose * pose, std::vector<R3Affine *> * saveXForms) const
{
	assert(pose->IsParamsInit(NJointParams()));
	// find reference point and traversal order...
	std::vector<int> jointOrder;
	R3Affine xform = R3identity_affine;
	int referenceJoint = FindReferenceNode(pose, xform);
	FillSkeletonNodeOrder(referenceJoint, jointOrder);
	
	// traverse graph
	std::vector<R3Affine *> parentXForms;
	assert(jointOrder.size()>0);
	assert(referenceJoint == jointOrder[0]);
	pose->InitJointPositions((int)jointOrder.size());
	for (int i=0; i<(int)jointOrder.size(); i++)
		parentXForms.push_back(NULL);
	
	xform.Apply(pose->Position(referenceJoint));
	parentXForms[referenceJoint] = new R3Affine(xform);
	for (int i=1; i<(int)jointOrder.size(); i++)
	{
		int nodeid = jointOrder[i];
		assert(nodeid>=0 && nodeid < (int)m_Joints.size());
		SkeletonLink * parentLink = m_Joints[nodeid]->m_pParent;
		assert(parentLink->m_pParent!=NULL);
		int parentid = parentLink->m_pParent->m_iJointID;
		assert(parentid>=0 && parentid < (int)parentXForms.size());
		assert(parentXForms[parentid]!=NULL);	// something wrong with the order
		const R3Affine & parentXForm = *(parentXForms[parentid]);
		R3Affine linkXForm = LinkTransform(parentLink, pose, m_ParamOffsetByLinkID[parentLink->m_iLinkID]);
		R3Affine combined = R3identity_affine;
		
		combined.Transform(parentXForm);
		combined.Transform(linkXForm);
		combined.Apply(pose->Position(nodeid));
		parentXForms[nodeid] = new R3Affine(combined);
	}
	
	if (saveXForms)
		*saveXForms = parentXForms;
	else
	{
		for (int i=0; i<(int)parentXForms.size(); i++)
			delete parentXForms[i];
	}
}

R3Affine IKSolver::LinkTransform(const SkeletonLink * link, PntCorrsPose * pose, int fp) const
{
	R3Affine xform = R3identity_affine;
	if (link->m_JointType == SkeletonLink::JOINT_Hinge)
	{
		pose->Param(fp) = link->GetValidAngle(pose->Param(fp), 0);
		xform.Rotate(link->m_RotationAxis, pose->Param(fp));
	}
	else if (link->m_JointType == SkeletonLink::JOINT_BallAndSocket)
	{
		for (int i=0; i<3; i++)
			pose->Param(fp+i) = link->GetValidAngle(pose->Param(fp+i), i);
		double theta  = pose->Param(fp+0);
		double phi    = pose->Param(fp+1);
		double alpha  = pose->Param(fp+2);
		xform.Rotate(R3Vector(0, 0, 1), R3Vector(sin(phi)*cos(theta), sin(phi)*sin(theta), cos(phi)));
		xform.Rotate(R3Vector(0, 0, 1), alpha);
	}
	xform.Translate(link->m_ChildOffset);
	return xform;
}


int IKSolver::FindReferenceNode(const PntCorrsPose * pose, R3Affine & xform) const
{
	if ((m_SolverFlags&m_XFormMask) == (IK_RigidTranslate|IK_RigidRotateAroundUp))
	{
		xform = R3identity_affine;
		xform.Translate(R3Vector(pose->GetParam(0), pose->GetParam(1), pose->GetParam(2)));
		xform.Rotate(R3Vector(0, 0, 1), pose->GetParam(3));
	}
	else if ((m_SolverFlags&m_XFormMask) == (IK_RigidTranslate|IK_RigidRotateAroundUp|IK_RigidRotateFrontBack|IK_RigidRotateLeftRight))
	{
		xform = R3Affine(R4Matrix(1.+pose->GetParam(0), pose->GetParam(1), pose->GetParam(2), pose->GetParam(3),
								  pose->GetParam(4), 1.+pose->GetParam(5), pose->GetParam(6), pose->GetParam(7),
								  pose->GetParam(8), pose->GetParam(9), 1.+pose->GetParam(10), pose->GetParam(11),
								  pose->GetParam(12), pose->GetParam(13), pose->GetParam(14), 1.+pose->GetParam(15)));
	}
	else
		assert(false);
	
//	double rotUP = 0;
//	double rotFB = 0;
//	
//	if ((m_SolverFlags & IK_RigidRotateAroundUp)!=0)
//	{
//		rotUP = pose->GetParam(offset+0);
//		offset += 1;
//	}
//	
//	if ((m_SolverFlags & IK_RigidRotateFrontBack)!=0)
//	{
//		rotFB = pose->GetParam(offset+0);
//		offset += 1;
//	}
//	
//	if (rotFB!=0)
//	{
//		R3Vector frontback(0, 1, 0);
//		if (rotUP!=0)
//		{
//			R3Affine upRotate = R3identity_affine;
//			upRotate.Rotate(R3Vector(0, 0, 1), rotUP);
//			upRotate.Apply(frontback);
//		}
//		xform.Rotate(frontback, rotFB);
//	}
//	
//	if (rotUP!=0)
//		xform.Rotate(R3Vector(0, 0, 1), rotUP);
		
	return 0;
}

void IKSolver::FillSkeletonNodeOrder(int fixedJoint, std::vector<int> & jointOrder) const
{
	for (int i=0; i<(int)m_Joints.size(); i++)
		jointOrder.push_back(i);
}

/////////////////
////    SkeletonJoint
SkeletonLink::JointType SkeletonLink::Str2Joint(const VKString & typeStr)
{
	if (typeStr=="fixed")
		return JOINT_None;
	else if (typeStr=="hinge")
		return JOINT_Hinge;
	else if (typeStr=="ball_socket")
		return JOINT_BallAndSocket;

	std::cout<<"[ERROR] Invalid Joint Type: "<<typeStr.c_str()<<std::endl;
	assert(false);
}

const char * SkeletonLink::Joint2Str(JointType jointType)
{
	switch (jointType)
	{
		case JOINT_None:
			return "fixed";
		case JOINT_Hinge:
			return "hinge";
		case JOINT_BallAndSocket:
			return "ball_socket";
	}
	assert(false);
	return NULL;
}

int SkeletonLink::NDims(SkeletonLink::JointType type)
{
	switch (type)
	{
		case JOINT_None:
			return 0;
		case JOINT_Hinge:
			return 1;
		case JOINT_BallAndSocket:
			return 3;
	}
	assert(false);
	return -1;
}

double SkeletonLink::GetValidAngle(double angle, int id) const
{
	while (angle < -M_PI)
		angle += 2*M_PI;
	while (angle > M_PI)
		angle -= 2*M_PI;
	
	if (!(angle>=-M_PI && angle <= M_PI))
	{
		std::cout<<"[ERROR] Invalid Angle: "<<angle<<" should be in [-pi, pi]"<<std::endl;
		assert(false);
	}
	assert(id >= 0 && id < (int)m_MinAngleValue.size() && id < (int)m_MaxAngleValue.size());
	
	if (angle < m_MinAngleValue[id] || angle > m_MaxAngleValue[id])
	{
		double dist2min = vkMin(vkAbs(angle-m_MinAngleValue[id]), vkAbs(angle-2*M_PI-m_MinAngleValue[id]));
		double dist2max = vkMin(vkAbs(angle-m_MaxAngleValue[id]), vkAbs(angle+2*M_PI-m_MaxAngleValue[id]));
		if (dist2min < dist2max)
			angle = m_MinAngleValue[id];
		else
			angle = m_MaxAngleValue[id];
	}
	assert(angle >= m_MinAngleValue[id] && angle <= m_MaxAngleValue[id]);
	return angle;
}


