#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"

#include "ModelDatabase.h"
#include "ShapeTriangleSoup.h"


#ifndef __IKSolver_H
#define __IKSolver_H

#include "Interactor.h"
#include "InteractorPose.h"
#include "LinAlgMatrixReal.h"
#include "PosePrior.h"

class Pnts2Joints;
class SkeletonJoint;
class SkeletonLink;
class IKModule;


struct IKConvergenceParams
{
	double m_fCurrDistStep;
	double m_fCurrAngleStep;
	int m_iCurrentConvergenceLevel;
	int m_iCurrentIteration;
};

struct IKStats
{
	IKStats();
	VKString GetStatsStr() const;
	
	int m_iNSolverCalled;
	int m_iNIterations;
	int m_iNTermDueToEnegy;
	int m_iNTermDueToAngle;
};

/**
 * This class implements inverse kinematics
 */
class IKSolver
{
	public:
		enum IKSolverType
		{
			IK_EndEffectors 		 = 1<<0,
			IK_PosePrior    		 = 1<<1,
			IK_RigidTranslate   	 = 1<<2,
			IK_RigidRotateAroundUp	 = 1<<3,
			IK_RigidRotateFrontBack  = 1<<4,	// NOT WORKING
			IK_RigidRotateLeftRight  = 1<<5,	// NOT WORKING
			IK_QuickPruning			 = 1<<6,
			IK_DisableIK			 = 1<<7,
			IK_ConditionedOnPoseID   = 1<<8
		};
		friend class InteractorPosePrior;
	
		IKSolver(double wEndEffector, double wPosePrior, double wPoseSymmetry, PosePrior * prior);
		virtual void ReadSkeleton(const VKString & skeleton, const VKString & regionNames);
		virtual bool SolveIKProblem(PntCorrsPose * input_output, bool reinit) const;
		virtual bool FillRenderableSkeleton(const InteractorPose * pose, bool updatePose,
											std::vector<R3Point> & pnts, std::vector<int> & edges) const;
		
		virtual void GetEnergy(const PntCorrsPose * corrs, EnergyInfo & info) const;
		virtual void GetPerContactEnergy(const PntCorrsPose * corrs, int refReg, const std::vector<int> & trgRegions,
										 std::vector<double> & outEnergies) const;
		virtual void FillPath(int fromRegion, int toRegion, std::vector<int> & pathLinks) const;
		virtual VKString PathStr(const std::vector<int> & pathLinks) const;
		virtual VKString LinkStr(int linkid) const;
	
		virtual void FindBestInitPose(PntCorrsPose * corrs) const;
		virtual void InitPoseToSitting(PntCorrsPose * corrs) const;
		virtual void InitPoseToZero(PntCorrsPose * corrs) const;
		virtual void FixPoseParameterization(PntCorrsPose * corrs) const;
	
//	protected:
		virtual int NJointParams() const;
		virtual int NEndEffectors(const PntCorrsPose * pose, int * npoints, int * nground) const;
		virtual int NTargetParams(const PntCorrsPose * pose) const;
		virtual void SolveDeltaTheta(const LinAlgMatrixReal & J, const LinAlgVectorReal & e, LinAlgVectorReal & dtheta) const;
		virtual void FillJacobian(PntCorrsPose * corrs, LinAlgMatrixReal & J) const;
		virtual bool FillDirectionVector(const PntCorrsPose * target, LinAlgVectorReal * e, EnergyInfo & info,
										 const IKConvergenceParams * convergence) const;
		virtual void ForwardKinematics(PntCorrsPose * pose, std::vector<R3Affine *> * saveXForms = NULL) const;
		virtual void FindBestRigidXForm(PntCorrsPose * pose) const;
		virtual void FindBestRigidXForm2D(PntCorrsPose * pose) const;
		virtual void FindBestRigidXFormFull(PntCorrsPose * pose) const;
		virtual int FindReferenceNode(const PntCorrsPose * pose, R3Affine & xform) const;
		virtual void FillSkeletonNodeOrder(int referenceJoint, std::vector<int> & jointOrder) const;
		virtual R3Affine LinkTransform(const SkeletonLink * link, PntCorrsPose * pose, int fp) const;
		virtual PntCorrsPose * CreateCopy(const PntCorrsPose * pose) const;
		// skeleton and points on a skeleton
		int m_iNInitSkeletonParams;
		VKStringList m_RegionNames;
		std::vector<int> m_ParamOffsetByLinkID;
		std::vector<int> m_Pnts2Skeleton;
		std::vector<SkeletonJoint*> m_Joints;
		std::vector<SkeletonLink*> m_Links;

	// convergence params
		int m_iIterations;
		int m_iMaxConvergenceLevels;
		double m_fMaxDistStep;
		double m_fMaxAngleStep;
		double m_fMaxEnergyErr;
		double m_fMaxAngleChange;
	
	// parameters
		int m_SolverFlags;
		int m_XFormMask;
	
	// weights:
		double m_wEndEffectorDist;
		double m_wPosePrior;
		double m_wSymmetryPrior;
	
	// pose prior
		PosePrior * m_pPrior;
		IKStats * m_pIKStats;
};

/**
 * Skeleton joint
 */
class SkeletonJoint
{
	public:
		int m_iJointID;			
		VKString m_sJointName;
		SkeletonLink * m_pParent;
		std::vector<SkeletonLink*> m_Children;
};

/**
 * Skeleton link
 */
class SkeletonLink
{
	public:
		enum JointType
		{
			JOINT_None 			= 0, 	// no parameters
			JOINT_Hinge			= 1, 	// 1-dimensional angle
			JOINT_BallAndSocket = 2, 	// 2-dimensional angle
		};
		static JointType Str2Joint(const VKString & typeStr);
		static const char * Joint2Str(JointType typeStr);
		static int NDims(JointType type);
		virtual double GetValidAngle(double angle, int id) const;
	
		int m_iLinkID;
		JointType m_JointType;
	
		std::vector<double> m_MinAngleValue;
		std::vector<double> m_MaxAngleValue;
	
		R3Vector m_RotationAxis;
		R3Vector m_ChildOffset;
		SkeletonJoint * m_pParent;
		SkeletonJoint * m_pChild;
};

#endif



