#include "TimeProfiler.h"

#include "Binding.h"
#include "Interactor.h"
#include "BindingManager.h"

#include "ModelDatabase.h"
#include "VkFunctions.h"
#include "ShapePointCloud.h"
#include "Model3D.h"
#include "PntCorrsPose.h"

#ifdef INCLUDE_BULLET
	#include "hacdCircularList.h"
	#include "hacdVector.h"
	#include "hacdICHull.h"
	#include "hacdGraph.h"
	#include "hacdHACD.h"
#endif


Interactor::Interactor(ShapeMesh * mesh, InteractorID type)
{
	m_pCentroid= NULL;
//	m_pPosePrior = NULL;
	m_pObjectDensePoints = NULL;
	m_pObjectPoints= NULL;
	m_pObjectPredictions = NULL;
	m_Type = type;
	m_PoseType = TYPE_Meshes;
	
	m_IndexableMeshPoses.push_back(new IndexedPose(0));
	m_IndexableMeshPoses[0]->m_pModel = new Model3D("", -1);
	m_IndexableMeshPoses[0]->m_pModel->AddGeomShape("Mesh", mesh);
	m_IndexableMeshPoses[0]->m_sPoseName = "singlepose.off";
	ComputeCanonicalXForm();
}

Interactor::Interactor(InteractorID type, ActorType actType)
{
	m_pCentroid= NULL;
//	m_pPosePrior = NULL;
	m_pObjectDensePoints = NULL;
	m_pObjectPoints= NULL;
	m_pObjectPredictions = NULL;

	m_Type = type;
	if (actType==TYPE_RagDoll)
		InitRagDoll();
	else
		m_PoseType = actType;
}

Interactor::Interactor(const VKString & meshPoses, InteractorID type)
{
	m_pCentroid= NULL;
//	m_pPosePrior = NULL;
	m_pObjectDensePoints = NULL;
	m_pObjectPoints= NULL;
	m_pObjectPredictions = NULL;

	m_Type = type;
	m_PoseType = TYPE_Meshes;
	
	ReadPosesFromMeshDir(meshPoses, "none");
	ComputeCanonicalXForm();
}

Interactor::ActorType Interactor::GetActorType() const
{
	return m_PoseType;
}

void Interactor::InitRagDoll()
{
	m_AllPoseXForm = R3identity_affine;
	m_PoseType = TYPE_RagDoll;
}

void Interactor::ReadPosesFromMeshDir(const VKString & meshPoses, const VKString & posesList)
{
	assert(FileManagement::DirExists(meshPoses.c_str()));
	VKStringList allFiles;
	if (posesList=="none")
		allFiles = FileManagement::ListAllFiles(meshPoses.c_str());
	else
	{
		VKString mdl;
		std::ifstream textStream(posesList.c_str());
		assert(textStream.is_open());
		while (!textStream.eof())
		{
			mdl.readLine(textStream);
			mdl.replace("\n", "");
			if (mdl!="")
				allFiles.push_back(mdl);
		}
		textStream.close();
	}
	
	std::cout<<"Reading Poses: \n";
	TimeProfiler timer;
	for (int i=0; i<(int)allFiles.count(); i++)
	{
		timer.WriteProgress("prog", i, allFiles.count());
		if (allFiles[i].endsWith(".txt"))
			continue;
		
		VKString readFile = (meshPoses+"/"+allFiles[i]);
		//std::cout<<"Reading: "<<readFile.c_str()<<std::endl;
		if (!FileManagement::FileExists(readFile.c_str()))
		{
			std::cout<<"[ERROR] Could not read file: "<<readFile.c_str()<<std::endl;
			assert(false);
		}
		int lid =(int)m_IndexableMeshPoses.size();
		m_IndexableMeshPoses.push_back(new IndexedPose(lid));
		m_IndexableMeshPoses[lid]->m_pModel = new Model3D("", -1);
		m_IndexableMeshPoses[lid]->m_pModel->AddGeomShape("Mesh", new ShapeTriangleSoup(readFile));
		m_IndexableMeshPoses[lid]->m_sPoseName = allFiles[i];
	}
	std::cout<<std::endl;
}

double Interactor::Ground(const InteractorPose * pose, const BindingXForm * xform) const
{
	
	double epsilon = 1E-10;
	double minVal;
	for (int i=0; i<NPoints(pose); i++)
	{
		R3Point vp = GetCanonicalPoint(i, pose);
		if (xform!=NULL)
			xform->Apply(vp);
		if (i==0 || vp.Z() < minVal)
			minVal = vp.Z();
	}
	return minVal-epsilon;
}


void Interactor::ReadPosesTransforms(const VKString & canonicalFile)
{
	std::ifstream textStream(canonicalFile.c_str());
	assert(textStream.is_open());
	
	assert(m_IndexableMeshPoses.size()==1);
	
	const ShapeMesh * mesh = GetMesh(m_IndexableMeshPoses[0]);
	R3Point centroid = mesh->Centroid();
	double aveR = mesh->AverageRadius(centroid);

	VKString tempStr;
	VKStringList tempStrSplit;
	bool ok;
	
	tempStr.readLine(textStream);	tempStrSplit = tempStr.split(" ");
	assert(tempStrSplit.count()==2 && tempStrSplit[0]=="Scale");
	double scale = tempStrSplit[1].toDouble(&ok);	assert(ok);
	
	tempStr.readLine(textStream);	tempStrSplit = tempStr.split(" ");
	assert(tempStrSplit.count()==2 && tempStrSplit[0]=="RotZ");
	double rotate = tempStrSplit[1].toDouble(&ok);	assert(ok);
	
	m_IndexableMeshPoses[0]->m_NormalizingXForm = R3identity_affine;
	m_IndexableMeshPoses[0]->m_NormalizingXForm.Scale(scale/aveR);
	m_IndexableMeshPoses[0]->m_NormalizingXForm.Rotate(R3Vector(0,0,1), rotate);	
	m_IndexableMeshPoses[0]->m_NormalizingXForm.Translate(R3Point(0,0,0)-centroid);
	
	textStream.close();

//	ComputeCanonicalXForm();
//	FILE * f = fopen(canonicalFile.c_str(), "rb");
//	if (f==NULL)
//	{
//		m_AllPoseXForm = R3identity_affine;
//		return;
//	}
//	
//	double mtx[16];
//	fread(mtx, sizeof(double), 16, f);
//	m_AllPoseXForm = R3Affine(R4Matrix(mtx));
//	fclose(f);
}

//void Interactor::ReadActiveRegions(const VKString & activeRegionsDir)
//{
//	std::ifstream activeRegionsList((activeRegionsDir+"/rlist.txt").c_str());
//	if (!activeRegionsList.is_open())
//		return;	// no active regions
//	while (!activeRegionsList.eof())
//	{
//		VKString regionName;
//		regionName.readLine(activeRegionsList);
//		regionName.replace("\n", "");
//		if (regionName=="")
//			continue;
//		VKString regionFile = activeRegionsDir+"/"+regionName+".bin";
//		int rid = (int)m_ActiveRegions.size();
//		ActiveRegion * newRegion = ActiveRegion::ReadActiveRegion(regionFile, regionName, rid, this);
//		m_ActiveRegions.push_back(newRegion);
//	}
//	activeRegionsList.close();
//}

R3Point Interactor::Centroid() const
{
	if (m_pCentroid==NULL)
	{
		assert(m_pObjectDensePoints!=NULL);
		((Interactor*)this)->m_pCentroid = new R3Point();
		*(((Interactor*)this)->m_pCentroid) = m_pObjectDensePoints->Centroid();
	}
	return *m_pCentroid;
}

void Interactor::GetObjectPredictions(const ShapePointCloud ** pnts, const ShapeValue ** predictions) const
{
	*pnts = m_pObjectPoints;
	*predictions = m_pObjectPredictions;
}

void Interactor::ReadObjectPoints(const VKString & filename)
{
	assert(NumEnumeratedPoses()==1);
	Model3D *m  = (Model3D*)GetModel(GetEnumeratedPose(0));
	ShapeMesh * mesh = (ShapeMesh * )GetMesh(GetEnumeratedPose(0));
	assert(mesh!=NULL);
	m_pObjectPoints = new ShapePointCloud(mesh, "featpnts");
	m_pObjectPoints->ReadGeometryFile(filename.c_str());
	m->AddGeomShape(m_pObjectPoints->GeometryName(), m_pObjectPoints);
}

void Interactor::ReadObjectDensePoints(const VKString & filename)
{
	assert(NumEnumeratedPoses()==1);
	Model3D *m  = (Model3D*)GetModel(GetEnumeratedPose(0));
	ShapeMesh * mesh = (ShapeMesh * )GetMesh(GetEnumeratedPose(0));
	assert(mesh!=NULL);
	m_pObjectDensePoints = new ShapePointCloud(mesh, "densepnts");
	m_pObjectDensePoints->ReadGeometryFile(filename.c_str());
	m->AddGeomShape(m_pObjectDensePoints->GeometryName(), m_pObjectDensePoints);
}

void Interactor::ReadObjectPredictions(const VKString & filename)
{
	assert(m_pObjectPoints!=NULL);
	m_pObjectPredictions = new ShapeValue(m_pObjectPoints);
	m_pObjectPredictions->ReadShapeValues(filename.c_str());
	m_pObjectPoints->AddShapeVal(m_pObjectPredictions->GetShapeValueName(), m_pObjectPredictions);
}


void Interactor::WritePoseToFile(FILE * file, const InteractorPose * pose) const
{
	pose->WritePose(file);
}

const InteractorPose * Interactor::CreatePoseFromFile(FILE * file, Interactor * object) const
{
	const InteractorPose * pose = InteractorPose::CreatePose(file, object);
	if (pose->GetPoseType()==InteractorPose::TYPE_IndexedPose)
	{
		int pid = ((IndexedPose*)pose)->m_iPoseID;
		assert(pid>=0 && pid < (int)m_IndexableMeshPoses.size());
		delete pose; 
		return m_IndexableMeshPoses[pid];
	}
	return pose;
}


Interactor * Interactor::ReadInteractor(const VKString & analysisDir, InteractorID type)
{
	Interactor * actor = new Interactor(type);

	FILE * iInfo = fopen((analysisDir+"/info.bin").c_str(), "wb");
	assert(iInfo!=NULL);
	fread(&(actor->m_PoseType), sizeof(ActorType), 1, iInfo);
	fclose(iInfo);

	// analysis directory exists: read as much as possible
	assert(FileManagement::DirExists(analysisDir.c_str()));
	if (actor->m_PoseType==TYPE_RagDoll)
		actor->InitRagDoll();
	else if (actor->m_PoseType==TYPE_Meshes)
	{
		// 1. read models + poses
		actor->ReadPosesFromMeshDir(analysisDir+"/meshes/", analysisDir+"/poses.txt");
		
		// 2. read canonical (pose) transformations
		//actor->ReadPosesTransforms(analysisDir+"/canonical_xform.bin");
	}
	// 3. read active regions
	//actor->ReadActiveRegions(analysisDir+"/active_regions/");
	
	return actor;
}

void Interactor::WriteInteractor(const VKString & analysisDir) const
{
	// poses are already there...
	// analysis directory exists: read as much as possible
	FileManagement::CreateDirsOnPath(analysisDir.c_str());
	FILE * iInfo = fopen((analysisDir+"/info.bin").c_str(), "wb");
	assert(iInfo!=NULL);
	fwrite(&m_PoseType, sizeof(ActorType), 1, iInfo);
	fclose(iInfo);
	
	if (m_PoseType==TYPE_Meshes)
	{
		// 1. write models + poses
		WritePosesFromMeshDir(analysisDir+"/meshes/", analysisDir+"/poses.txt");
		// 2. write canonical (pose) transformations
		//Read(analysisDir+"/canonical_xform.bin");
	}
	// 3. write active regions
//	WriteActiveRegions(analysisDir+"/active_regions/");
}

void Interactor::WritePosesFromMeshDir(const VKString & meshPoses, const VKString & posesList) const
{
	FileManagement::CreateDirsOnPath(meshPoses.c_str());
	std::ofstream outPoseList(posesList.c_str());
	assert(outPoseList.is_open());

	// indexable poses
	for (int i=0; i<(int)m_IndexableMeshPoses.size(); i++)
	{
		outPoseList<<m_IndexableMeshPoses[i]->m_sPoseName.c_str()<<"\n";
		VKString outFile = meshPoses+"/"+m_IndexableMeshPoses[i]->m_sPoseName;
		assert(outFile.endsWith(".off"));
		if (!FileManagement::FileExists(outFile.c_str()))	// do not duplicate
			GetMesh(m_IndexableMeshPoses[i])->WriteGeometryFile(outFile);
	}
	outPoseList.close();
}

void Interactor::WritePosesTransforms(const VKString & canonicalFile) const
{
	assert(false);
//	FileManagement::CreateDirsOnPathToFile(canonicalFile.c_str());
//	FILE * f = fopen(canonicalFile.c_str(), "wb");
//	assert(f!=NULL);	
//	double mtx[16];
//	for (int r=0; r<4; r++)
//	for (int c=0; c<4; c++)
//		mtx[r*4 + c] = m_AllPoseXForm.Matrix()[r][c];
//	fwrite(mtx, sizeof(double), 16, f);
//	fclose(f);
}

//void Interactor::WriteActiveRegions(const VKString & activeRegionsDir) const
//{
//	int nregions = NActiveRegions();
//	
//	FileManagement::CreateDirsOnPath(activeRegionsDir.c_str());
//	std::ofstream activeRegionsList((activeRegionsDir+"/rlist.txt").c_str());
//	assert(activeRegionsList.is_open());
//	for (int i=0; i<nregions; i++)
//	{
//		VKString regionName = GetActiveRegion(i)->GetRegionName();
//		activeRegionsList<<regionName.c_str()<<"\n";
//		GetActiveRegion(i)->WriteActiveRegion(activeRegionsDir+"/"+regionName+".bin");
//	}
//	activeRegionsList.close();
//}


void Interactor::AddActiveRegion(ActiveRegion * region)
{
	m_ActiveRegions.push_back(region);
}

int Interactor::NActiveRegions() const
{
	return (int)m_ActiveRegions.size();
}

const ActiveRegion * Interactor::GetActiveRegion(int regionID) const
{
	assert(regionID>=0 && regionID<(int)m_ActiveRegions.size());
	return m_ActiveRegions[regionID];
}

int Interactor::NPoints(const InteractorPose * pose) const
{
	if (m_PoseType==TYPE_Meshes)
		return GetMesh(pose)->NVertices();
#ifdef INCLUDE_BULLET
	else if (m_PoseType==TYPE_RagDoll)
		return RagDollNSamples(pose);
#endif
	else if (m_PoseType==TYPE_PntCorrs)
		return (int)((PntCorrsPose*)pose)->NMatches();
		
	assert(false);	// todo: ragdoll
	return 0;
}

int Interactor::PointID2FaceID(const InteractorPose * pose, int pid) const
{
	if (m_PoseType==TYPE_Meshes)
		return GetMesh(pose)->FaceOnVertex(pid);
	
	//assert(false);
	return -1;	// disabled for now...
}

R3Point Interactor::GetCanonicalPoint(int pid, const InteractorPose * pose) const
{
	R3Point p(0,0,0);
	if (m_PoseType==TYPE_Meshes)
	{
		p = GetMesh(pose)->VertexPosition(pid);
		GetCanonical(pose).Apply(p);
	}
#ifdef INCLUDE_BULLET
	else if (m_PoseType==TYPE_RagDoll)
		p = RagDollSample(pose, pid);
#endif
	
	return p;
}

double Interactor::GetCanonicalPointWeight(int pid, const InteractorPose * pose) const
{
	double area = 0;
	if (m_PoseType==TYPE_Meshes)
	{
		const ShapeMesh * mesh = GetMesh(pose);
		if (pose->m_MeshArea==0)
			((InteractorPose*)pose)->m_MeshArea = mesh->Area();
		for (int f=0; f<mesh->NAdjacentFaces(pid); f++)
			area += mesh->FaceArea(mesh->FaceOnVertex(pid, f));
		area /= 3*pose->m_MeshArea;
	}
#ifdef INCLUDE_BULLET
	else if (m_PoseType==TYPE_RagDoll)
		area = 1. / RagDollNSamples(pose);
#endif
//	else
//		assert(false);
	return area;
}


void Interactor::ComputeCanonicalXForm()
{
	for (int i=0; i<(int)m_IndexableMeshPoses.size(); i++)
	{
		const ShapeMesh * mesh = GetMesh(m_IndexableMeshPoses[i]);
		R3Point centroid = mesh->Centroid();
		double aveR = mesh->AverageRadius(centroid);
		
		m_IndexableMeshPoses[i]->m_NormalizingXForm = R3identity_affine;
		m_IndexableMeshPoses[i]->m_NormalizingXForm.Scale(1./aveR);
		m_IndexableMeshPoses[i]->m_NormalizingXForm.Translate(R3Point(0,0,0)-centroid);
	}
	m_AllPoseXForm = R3identity_affine;
}

const R3MeshSearchTree * Interactor::GetMeshSearchTree(const InteractorPose * pose,
													   const std::vector<int> ** tree2meshMap_vertices,
													   const std::vector<int> ** tree2meshMap_faces) const
{
	if (pose->m_pSearchTree==NULL)
	{
		assert(pose->m_pR3Mesh == NULL);

		const ShapeMesh * fm = GetMesh(pose);
		R3Mesh * mesh = new R3Mesh();		
		std::vector<int> & vertexMap = ((InteractorPose*)pose)->m_Tree2MeshMapVertices;
		std::vector<int> & triangleMap = ((InteractorPose*)pose)->m_Tree2MeshMapFaces;
		
		// create r3mesh 
		for (int tID = 0; tID < fm->NFaces(); tID++)
		{
			int v1 = fm->VertexOnFace(tID, 0);
			int v2 = fm->VertexOnFace(tID, 1);
			int v3 = fm->VertexOnFace(tID, 2);
			
			R3Point v1Pos = fm->VertexPosition(v1);
			R3Point v2Pos = fm->VertexPosition(v2);
			R3Point v3Pos = fm->VertexPosition(v3);
						
			R3MeshVertex * v1Ptr = mesh->CreateVertex(v1Pos);
			R3MeshVertex * v2Ptr = mesh->CreateVertex(v2Pos);
			R3MeshVertex * v3Ptr = mesh->CreateVertex(v3Pos);
			R3MeshFace * facePtr = mesh->CreateFace(v1Ptr, v2Ptr, v3Ptr);
			
			int v1ID = mesh->VertexID(v1Ptr);
			int v2ID = mesh->VertexID(v2Ptr);
			int v3ID = mesh->VertexID(v3Ptr);
			
			int faceID = mesh->FaceID(facePtr);
			
			assert(v1ID==(int)vertexMap.size());
			vertexMap.push_back(v1);
			assert(v2ID==(int)vertexMap.size());
			vertexMap.push_back(v2);
			assert(v3ID==(int)vertexMap.size());
			vertexMap.push_back(v3);
			
			assert(faceID==(int)triangleMap.size());
			triangleMap.push_back(tID);
		}
		
		// convert vertex positions
		for (int i=0; i<mesh->NVertices(); i++)
		{
			R3MeshVertex * v = mesh->Vertex(i);
			R3Point vp = mesh->VertexPosition(v);
			GetCanonical(pose).Apply(vp);
			mesh->SetVertexPosition(v, vp);
		}
		((InteractorPose*)pose)->m_pSearchTree = new R3MeshSearchTree(mesh);
		((InteractorPose*)pose)->m_pR3Mesh = mesh;
	}
	
	assert(tree2meshMap_vertices!=NULL && tree2meshMap_faces!=NULL);
	*tree2meshMap_vertices = &(pose->m_Tree2MeshMapVertices);
	*tree2meshMap_faces = &(pose->m_Tree2MeshMapFaces);

	return pose->m_pSearchTree;
}

double Interactor::Distance(const InteractorPose * pose, const R3Point & worldPoint,
							double maxDist, int * face, R3Point * p) const
{
	if (m_PoseType==TYPE_Meshes)
	{
		R3Point query = worldPoint;
		const std::vector<int> * tree2meshmap_vertices;
		const std::vector<int> * tree2meshmap_faces;
		R3MeshSearchTree * tree = (R3MeshSearchTree *)GetMeshSearchTree(pose, &tree2meshmap_vertices, &tree2meshmap_faces);
		R3MeshIntersection closest;
		tree->FindClosest(query, closest, 0, maxDist);

		if (face!=NULL)
			*face = -1;
		
		if (closest.type==R3_MESH_NULL_TYPE)
			return RN_INFINITY;
		
		if (face!=NULL)
		{
			assert(closest.face!=NULL && pose->m_pR3Mesh!=NULL);
			int fid = pose->m_pR3Mesh->FaceID(closest.face);
			assert(fid>=0 && fid < (int)tree2meshmap_faces->size());
			*face = (*tree2meshmap_faces)[fid];
		}
		
		if (p!=NULL)
			*p = closest.point;
		return R3Distance(closest.point, worldPoint);
	}
	
	assert(false);
	return -1;
}

R3Box Interactor::BoundingBox(const InteractorPose * pose) const
{
	R3Box box(0., 0., 0., 0., 0., 0.);
	const ShapeMesh * mesh = GetMesh(pose);
	const R3Affine & xform = GetCanonical(pose);
	for (int v=0; v<mesh->NVertices(); v++)
	{
		R3Point pos = mesh->VertexPosition(v);
		xform.Apply(pos);
		box.Union(pos);
	}
	return box;
}

const R3Box & Interactor::RelevantVolume(const InteractorPose * pose) const
{
//	if (m_aRelevantVolumes.size()==0)
//	{
//		for (int i=0; i<NPoses(); i++)
//			(((Interactor*)this)->m_aRelevantVolumes).push_back(NULL);
//	}
	if (pose->m_pRelevantVolume==NULL)
	{
		R3Box * box = new R3Box(BoundingBox(pose));
		double offZ = 2.;
		double offXY = 1.;
		box->Union(R3Point(0,0, box->ZMax()+offZ));
		box->Union(R3Point(0,box->YMax()+offXY,0));
		box->Union(R3Point(box->XMax()+offXY,0, 0));
		box->Union(R3Point(0,box->YMin()-offXY,0));
		box->Union(R3Point(box->XMin()-offXY,0, 0));
		((InteractorPose*)pose)->m_pRelevantVolume = box;
	}

	return *(pose->m_pRelevantVolume);
}

void Interactor::Rescale(double factor)
{
	m_AllPoseXForm.Scale(factor);
	for (int c=0; c<(int)m_IndexableMeshPoses.size(); c++)
		m_IndexableMeshPoses[c]->m_CanonicalPoseComputed = false;
}

void Interactor::Reshuffle(char ** axes)
{
	R4Matrix mtx = R4identity_matrix;	mtx[0][0] = 0; 		mtx[1][1] = 0;		mtx[2][2] = 0;
	for (int i=0; i<3; i++)
	{
		double sign = 1.;
		int aoff = 0;
		if (axes[i][0]=='-')
		{
			sign = -1.;
			aoff = 1;
		}
		
		if (axes[i][aoff]=='x')
			mtx[i][0] = sign;
		else if (axes[i][aoff]=='y')
			mtx[i][1] = sign;
		else if (axes[i][aoff]=='z')
			mtx[i][2] = sign;
		else
			assert(false);
	}
	
	R3Affine currXForm = m_AllPoseXForm;
	R3Affine perXForm(mtx);
	perXForm.Transform(currXForm);
	m_AllPoseXForm = perXForm;
	for (int c=0; c<(int)m_IndexableMeshPoses.size(); c++)
		m_IndexableMeshPoses[c]->m_CanonicalPoseComputed = false;
}

#ifdef INCLUDE_BULLET
void Interactor::FillRigidBodies(const InteractorPose * pose, const PhysicsConstants & physics, const RigidAlign * align,
								 std::vector<btRigidBody *> & bodies, std::vector<btTypedConstraint *> & joints) const
{
	btTransform identity;
	identity.setIdentity();
	
	if (pose->m_pPhysicsRigidBody.size()==0)
	{
		if (m_PoseType==TYPE_Meshes)
		{
			btScalar mass(1.);
			if (m_Type==Interactor::ACT_Actor)
				mass = physics.m_fActorWeight;
			else if (m_Type==Interactor::ACT_Object)
				mass = physics.m_fObjectWeight;
			btVector3 localInertia(0,0,0);
			btCollisionShape * shape = GetCollisionShape(pose, physics);
			if (mass>0)
				shape->calculateLocalInertia(mass,localInertia);

			btDefaultMotionState* myMotionState = new btDefaultMotionState(identity);
			btRigidBody::btRigidBodyConstructionInfo rbInfo(mass, myMotionState, shape, localInertia);
			btRigidBody* body = new btRigidBody(rbInfo);
			
			if (m_Type==Interactor::ACT_Actor)
			{
				body->setFriction(physics.m_afFriction[PhysicsConstants::FRIC_ActorStatic]);
				body->setRollingFriction(physics.m_afFriction[PhysicsConstants::FRIC_ActorRolling]);
			}
			else
			{
				assert(m_Type==Interactor::ACT_Object);
				body->setFriction(physics.m_afFriction[PhysicsConstants::FRIC_ObjectStatic]);
				body->setRollingFriction(physics.m_afFriction[PhysicsConstants::FRIC_ObjectRolling]);
	//			body->setCollisionFlags(body->getCollisionFlags()|btCollisionObject::CF_STATIC_OBJECT);
			}
			((InteractorPose*)pose)->m_pPhysicsRigidBody.push_back(body);
			bodies.push_back(body);
		}
		else if (m_PoseType==TYPE_RagDoll)
			CreateRagDoll(((InteractorPose*)pose));
		else
			assert(false);
	}
	
	bodies = pose->m_pPhysicsRigidBody;
	joints = pose->m_Constraints;

	if (align!=NULL)
	{
		// set transformations
		btTransform transform;
		const R3Vector & t = align->GetTranslation();
		const R3Vector & r = align->GetRotationAxis();
		double angle = align->GetRotationAngle();
		transform.setOrigin(btVector3(t.X(), t.Y(), t.Z()));
		transform.setRotation(btQuaternion(btVector3(r.X(), r.Y(), r.Z()), angle));
		
		if (m_PoseType==TYPE_Meshes)
		{
			for (int i=0; i<(int)bodies.size(); i++)
			{
				btTransform currXForm;
				bodies[i]->getMotionState()->getWorldTransform (currXForm);
				bodies[i]->getMotionState()->setWorldTransform(transform);
				bodies[i]->proceedToTransform(transform);
			}
		}
		else if (m_PoseType==TYPE_RagDoll)
		{
			UpdateRagDollPhysicsFromPose((InteractorPose*) pose);
			for (int i=0; i<(int)bodies.size(); i++)
			{
				btTransform currXForm;
				bodies[i]->getMotionState()->getWorldTransform (currXForm);
				bodies[i]->getMotionState()->setWorldTransform(transform*currXForm);
				bodies[i]->proceedToTransform(transform*currXForm);
			}
		}
	}
}

void Interactor::UpdatePoseFromPhysics(InteractorPose * pose, RigidAlign * align) const
{
	// modify pose FK parameters if necessary
	
	assert(pose->m_pPhysicsRigidBody.size()>=1);
	btRigidBody* body = (btRigidBody*)pose->m_pPhysicsRigidBody[0];
	assert(body!=NULL);
	btTransform physicsXForm;
	body->getMotionState()->getWorldTransform (physicsXForm);
	const btVector3 & t = physicsXForm.getOrigin ();
	btQuaternion r = physicsXForm.getRotation ();
	
	double alpha = r.getAngle();
	btVector3 beta = r.getAxis();
	align->SetTranslation(R3Vector(t.x(), t.y(), t.z()));
	align->SetRotation(R3Vector(beta.x(), beta.y(), beta.z()), alpha);
	
	if (m_PoseType==TYPE_RagDoll)
		UpdateRagDollPoseFromPhysics(pose);
//	else
//	{
//		static bool printedWarning = false;
//		if (!printedWarning)
//			std::cout<<"[WARNING] Unknown pose"<<std::endl;
//		printedWarning = true;
//	}
}

btCollisionShape * Interactor::GetCollisionShape(const InteractorPose * pose, const PhysicsConstants & physics) const
{
	if (pose->m_pPhysicsShape==NULL)
	{
		if (m_Type == Interactor::ACT_Object)
			//((InteractorPose*)pose)->m_pPhysicsShape = GetStaticShape(pose, physics);
			((InteractorPose*)pose)->m_pPhysicsShape = GetGridApproxShape(pose, physics);
			//((InteractorPose*)pose)->m_pPhysicsShape = GetUnionConvexShapes(pose, physics);
		else if (m_Type == Interactor::ACT_Actor)
			//((InteractorPose*)pose)->m_pPhysicsShape = GetDynamicShape(pose, physics);
			((InteractorPose*)pose)->m_pPhysicsShape = GetGridApproxShape(pose, physics);
			//((InteractorPose*)pose)->m_pPhysicsShape = GetUnionConvexShapes(pose, physics);
	}
	assert(pose->m_pPhysicsShape!=NULL);
	return pose->m_pPhysicsShape;
}

btBvhTriangleMeshShape * Interactor::GetStaticShape(const InteractorPose * pose, const PhysicsConstants & physics) const
{
	btBvhTriangleMeshShape * shape = new btBvhTriangleMeshShape(CreateBulletMesh(pose, physics), true, true);
	shape->setMargin(0.02);
	return shape;
}

btGImpactMeshShape * Interactor::GetDynamicShape(const InteractorPose * pose, const PhysicsConstants & physics) const
{
	btGImpactMeshShape * shape = new btGImpactMeshShape(CreateBulletMesh(pose, physics));
	shape->postUpdate();
	shape->updateBound();
	shape->setMargin(physics.m_SurfaceMargin);
	return shape;	
}

btCompoundShape * Interactor::GetUnionConvexShapes(const InteractorPose * pose, const PhysicsConstants & physics) const
{
	const ShapeMesh * mesh = GetMesh(pose);
	const R3Affine & aff = GetCanonical(pose);

	// initialize matrix in HACD format
	HACD::Vec3<HACD::Real> * points = new HACD::Vec3<HACD::Real>[mesh->NVertices()];
	HACD::Vec3<long> * triangles = new HACD::Vec3<long>[mesh->NFaces()];
	
	for(int i=0; i<mesh->NVertices(); i++ )
	{
		int index = i*3;
		R3Point p = mesh->VertexPosition(i);
		aff.Apply(p);
		points[i] = HACD::Vec3<HACD::Real>(p.X(), p.Y(), p.Z());
	}
	
	for(int i=0;i<mesh->NFaces();i++)
		triangles[i] = HACD::Vec3<long>(mesh->VertexOnFace(i, 0), mesh->VertexOnFace(i, 1), mesh->VertexOnFace(i, 2));
	
	// initialize convex decomposition
	HACD::HACD myHACD;
	myHACD.SetPoints(&points[0]);
	myHACD.SetNPoints(mesh->NVertices());
	myHACD.SetTriangles(&triangles[0]);
	myHACD.SetNTriangles(mesh->NFaces());
	myHACD.SetCompacityWeight(0.1);
	myHACD.SetVolumeWeight(0.0);

	// HACD parameters
	// Recommended parameters: 2 100 0 0 0 0
	size_t nClusters = 2;
	double concavity = 100;
	bool invert = false;
	bool addExtraDistPoints = false;
	bool addNeighboursDistPoints = false;
	bool addFacesPoints = false;

	myHACD.SetNClusters(nClusters);                     // minimum number of clusters
	myHACD.SetNVerticesPerCH(100);                      // max of 100 vertices per convex-hull
	myHACD.SetConcavity(concavity);                     // maximum concavity
	myHACD.SetAddExtraDistPoints(addExtraDistPoints);
	myHACD.SetAddNeighboursDistPoints(addNeighboursDistPoints);
	myHACD.SetAddFacesPoints(addFacesPoints);

	std::cout<<"Computing convex decomposition (NOT IMPLEMENTED -- WILL CRUSH)..."<<std::endl;	
	myHACD.Compute();
	std::cout<<" -- done!"<<std::endl;
	nClusters = myHACD.GetNClusters();
	
	//myHACD.Save("output.wrl", false);	// can save the decomposition

	
	/////////////////////////////
	btCompoundShape* compound = new btCompoundShape();
	//m_collisionShapes.push_back (compound);
	
//	btTransform trans;
//	trans.setIdentity();
//	
//	for (int c=0;c<nClusters;c++)
//	{
//		//generate convex result
//		size_t nPoints = myHACD.GetNPointsCH(c);
//		size_t nTriangles = myHACD.GetNTrianglesCH(c);
//		
//		btScalar* vertices = new float[nPoints*3];
//		//unsigned int* triangles = new unsigned int[nTriangles*3];
//		
//		HACD::Vec3<HACD::Real> * pointsCH = new HACD::Vec3<HACD::Real>[nPoints];
//		HACD::Vec3<long> * trianglesCH = new HACD::Vec3<long>[nTriangles];
//		myHACD.GetCH(c, pointsCH, trianglesCH);
//		
//		// calculate centroid
//		double c[3] = {0,0,0};
//		for (int v=0; v<nPoints; v++)
//		{
//			c[0] += pointsCH[v].X();
//			c[1] += pointsCH[v].Y();
//			c[2] += pointsCH[v].Z();
//			vertices[v * 3 + 0] = pointsCH[v].X();
//			vertices[v * 3 + 1] = pointsCH[v].Y();
//			vertices[v * 3 + 2] = pointsCH[v].Z();
//		}
//		btTransform transform;
//		transform.setIdentity();
//
//		btConvexHullShape* convexShape = new btConvexHullShape(vertices, nPoints, sizeof(btScalar)*3);
//		compound->addChildShape(transform, convexShape);
//	}
//	/////////////////////////////

	return compound;
}

//void Interactor::SetPosePrior(InteractorPosePrior * prior)
//{
//	m_pPosePrior = prior;
//}
//
//InteractorPosePrior * Interactor::GetPosePrior()
//{
//	return m_pPosePrior;
//}

//IKSolver * Interactor::GetIK()
//{
//	assert(m_pPosePrior->PriorType()==InteractorPosePrior::PRIOR_GaussAngles);
//	return ((PriorGaussAngles*)m_pPosePrior)->GetIK();
//}

btCompoundShape * Interactor::GetGridApproxShape(const InteractorPose * pose, const PhysicsConstants & physics) const
{
	const ShapeMesh * mesh = GetMesh(pose);
	const R3Affine & aff = GetCanonical(pose);

	btCompoundShape * shape = new btCompoundShape();

	// find out grid dimensions
	bool isotropic = true;
	int res = 32;
	int ncells = pow(res, 3);
	R3Box bbox = BoundingBox(pose);
	int resX=-1, resY=-1, resZ=-1;
	double cellSize = pow(bbox.Volume()/ncells, 1./3.);
	if (isotropic)
	{
		double vdefault = pow(bbox.Volume(), 1./3.);
		resX = bbox.XLength() * res / vdefault;
		resY = bbox.YLength() * res / vdefault;
		resZ = bbox.ZLength() * res / vdefault;
	}
	else
	{
		resX = res;
		resY = res;
		resZ = res;
	}
	
	// rasterize shape into a grid
	R3Grid grid(resX, resY, resZ, bbox);
	for (int i=0; i<mesh->NFaces(); i++)
	{
		R3Point p1 = mesh->VertexPosition(mesh->VertexOnFace(i, 0));
		R3Point p2 = mesh->VertexPosition(mesh->VertexOnFace(i, 1));
		R3Point p3 = mesh->VertexPosition(mesh->VertexOnFace(i, 2));
		aff.Apply(p1);
		aff.Apply(p2);
		aff.Apply(p3);
		grid.RasterizeWorldTriangle(p1, p2, p3, 1.);
	}
	grid.SquaredDistanceTransform();
	
	// create spheres from a grid
	int ntotal = 0;
	double distMargin2 = pow(cellSize*2., 2.);
	for (int x=0; x<resX; x++)
	for (int y=0; y<resY; y++)
	for (int z=0; z<resZ; z++)
	{
		double dist2 = grid.GridValue(x, y, z) * cellSize;
		if (dist2 < distMargin2)
		{
			R3Point p = grid.WorldPosition((double)x, (double)y, (double)z);
			btTransform transform;
			transform.setIdentity();
			transform.setOrigin(btVector3(p.X(), p.Y(), p.Z()));
			btSphereShape * sphere = new btSphereShape(cellSize/2);
			shape->addChildShape(transform, sphere);
			ntotal++;
		}
	}
	
//	// TODO: actual grid
//	for (int v=0; v<mesh->NVertices(); v++)
//	{
//		R3Point p = mesh->VertexPosition(v);
//		aff.Apply(p);
//		btTransform transform;
//		transform.setIdentity();
//		transform.setOrigin(btVector3(p.X(), p.Y(), p.Z()));
//		btSphereShape * sphere = new btSphereShape(physics.m_SurfaceMargin);
//		shape->addChildShape(transform, sphere);
//	}
	shape->setMargin(physics.m_SurfaceMargin);
	return shape;
}

btTriangleIndexVertexArray * Interactor::CreateBulletMesh(const InteractorPose * pose, const PhysicsConstants & physics) const
{
	const ShapeMesh * mesh = GetMesh(pose);
	const R3Affine & aff = GetCanonical(pose);
	
	int * indices = new int[3 * mesh->NFaces()];
	btScalar * vertices = new btScalar[3 * mesh->NVertices()];
	
	
	for (int v=0; v<mesh->NVertices(); v++)
	{
		R3Point p = mesh->VertexPosition(v);
		
		aff.Apply(p);
		for (int d=0; d<3; d++)
			vertices[3 * v + d] = p[d];
	}
	
	for (int f=0; f<mesh->NFaces(); f++)
	for (int d=0; d<3; d++)
		indices[3 * f + d] = mesh->VertexOnFace(f, d);
	
	return new btTriangleIndexVertexArray(mesh->NFaces(), indices, 3 * sizeof(int),
										  mesh->NVertices(), vertices, 3 * sizeof(btScalar));
}

btRigidBody * Interactor::CreateRigidBody(btScalar mass, const btTransform& startTransform, btCollisionShape* shape) const
{
	
	bool isDynamic = (mass != 0.f);
	
	btVector3 localInertia(0,0,0);
	if (isDynamic)
		shape->calculateLocalInertia(mass,localInertia);
	
	btDefaultMotionState* myMotionState = new btDefaultMotionState(startTransform);
	
	btRigidBody::btRigidBodyConstructionInfo rbInfo(mass,myMotionState,shape,localInertia);
	btRigidBody* body = new btRigidBody(rbInfo);
	
	return body;
}
#endif

const ShapeMesh * Interactor::GetMesh(const InteractorPose * pose) const
{
	assert(pose->m_pModel!=NULL);
	GeomShape * shape = ((Model3D*)GetModel(pose))->GetGeomShape("Mesh");
	if (shape!=NULL)
		return shape->SafeCastToMesh();
	return NULL;
}

const Model3D * Interactor::GetModel(const InteractorPose * pose) const
{
	assert(pose!=NULL && pose->m_pModel!=NULL);
	return pose->m_pModel;
}

const InteractorPose * Interactor::GenerateRandomPose(const PhysicsConstants & physics) const
{
	InteractorPose * pose = NULL;
	if (m_PoseType==TYPE_Meshes)
	{
		int pid = vkRandInt((int)m_IndexableMeshPoses.size());
		assert(pid>=0 && pid < (int)m_IndexableMeshPoses.size());
		assert(m_IndexableMeshPoses[pid]->m_pModel!=NULL);
		pose = m_IndexableMeshPoses[pid];
	}
#ifdef INCLUDE_BULLET
	else if (m_PoseType==TYPE_RagDoll)
	{
		pose = GenerateRandomFKPose();
	}
	assert(pose!=NULL);
	
	std::vector<btRigidBody *> bodies;
	std::vector<btTypedConstraint *> joints;
	FillRigidBodies(pose, physics, NULL, bodies, joints);
#endif
	return pose;
}

int Interactor::NumEnumeratedPoses() const
{
	return (int)m_IndexableMeshPoses.size();
}

const InteractorPose * Interactor::GetEnumeratedPose(int poseid) const
{
	assert(poseid>=0 && poseid < (int)m_IndexableMeshPoses.size());
	return m_IndexableMeshPoses[poseid];
}

//void Interactor::PoseLogProb(const InteractorPose * pose, EnergyInfo & energy) const
//{
////	if (m_PoseType==TYPE_Meshes)
////		return 0;	// high probability
////	else if (m_pPosePrior==NULL)
////		return 0;	// no prior -> always high probability
////	else
//	if (m_PoseType!=TYPE_Meshes && m_pPosePrior!=NULL)
//		m_pPosePrior->Energy(pose, energy);
//}


//int Interactor::NPoses() const
//{
//	return (int)m_aModels.size();
//}

const char * Interactor::ActorName(InteractorID id)
{
	if (id==ACT_Actor)
		return "Actor";
	else if (id==ACT_Object)
		return "Object";
	else
		return "Invalid";
}

const R3Affine & Interactor::GetCanonical(const InteractorPose * pose) const
{
	if (!pose->m_CanonicalPoseComputed)
	{
		((InteractorPose*)pose)->m_CanonicalPoseComputed = true;
		((InteractorPose*)pose)->m_CanonicalPose = R3identity_affine;
		((InteractorPose*)pose)->m_CanonicalPose.Transform(m_AllPoseXForm);
		((InteractorPose*)pose)->m_CanonicalPose.Transform(pose->m_NormalizingXForm);
	}
	return pose->m_CanonicalPose;
}
