#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"
#include "InteractorPose.h"
//#include "InteractorPosePrior.h"

#include "ModelDatabase.h"
#include "ShapeTriangleSoup.h"

#ifdef INCLUDE_BULLET
	#include "LinearMath/btAlignedObjectArray.h"
	#include "btBulletDynamicsCommon.h"
	#include "BulletCollision/Gimpact/btGImpactShape.h"
	#include "BulletCollision/Gimpact/btGImpactCollisionAlgorithm.h"
#endif

#ifndef __Interactor_H
#define __Interactor_H

class Binding;
class ActiveRegion;
class BindingXForm;
class RigidAlign;
class PhysicsConstants;
struct EnergyInfo;

/**
 * This class describes an actor or an object
 */
class Interactor
{
	public:
		friend class InteractorPose;
	
		enum InteractorID
		{
			ACT_Actor = 0,
			ACT_Object = 1,
			ACT_Invalid = 2
		};
	
		enum ActorType
		{
			TYPE_Meshes 	= 0,
			TYPE_RagDoll 	= 1,
			TYPE_PntCorrs	= 2,
			TYPE_RiggedMesh = 3,
			TYPE_Invalid	= 4
		};

		Interactor(ShapeMesh * mesh, InteractorID type);
		Interactor(const VKString & meshPoses, InteractorID type);
		Interactor(InteractorID type, ActorType actType = TYPE_Invalid);

		virtual ActorType GetActorType() const;
	
		virtual void InitRagDoll();

		virtual R3Point Centroid() const;
		virtual const R3Box & RelevantVolume(const InteractorPose * pose) const;		// this not a trivial thing - data-specific...
		virtual R3Box BoundingBox(const InteractorPose * pose) const;
	
		virtual double Ground(const InteractorPose * pose, const BindingXForm * xform = NULL) const;
		virtual void Reshuffle(char ** axes);
		virtual void Rescale(double factor);
		virtual const R3Affine & GetCanonical(const InteractorPose * pose) const;
		
		virtual int NPoints(const InteractorPose * pose) const;
		virtual int PointID2FaceID(const InteractorPose * pose, int pid) const;
		virtual R3Point GetCanonicalPoint(int pid, const InteractorPose * pose) const;
		virtual double GetCanonicalPointWeight(int pid, const InteractorPose * pose) const;
#ifdef INCLUDE_BULLET
		virtual void FillRigidBodies(const InteractorPose * pose, const PhysicsConstants & physics, const RigidAlign * xform,
									   std::vector<btRigidBody *> & bodies, std::vector<btTypedConstraint *> & joints) const;
		virtual void UpdatePoseFromPhysics(InteractorPose * pose, RigidAlign * xform) const;
#endif
		virtual double Distance(const InteractorPose * pose, const R3Point & worldPoint,
								double maxDist = RN_INFINITY, int * face = NULL, R3Point * p = NULL ) const;
		
		virtual int NumEnumeratedPoses() const;
		virtual const InteractorPose * GetEnumeratedPose(int poseid) const;
	
		virtual void WritePoseToFile(FILE * file, const InteractorPose * pose) const;
		// virtual const InteractorPose * CreatePoseFromCorrs(const PntCorrsPose * corrs) const;
		virtual const InteractorPose * CreatePoseFromFile(FILE * file, Interactor * object) const;
		virtual const InteractorPose * GenerateRandomPose(const PhysicsConstants & physics) const;
//		virtual void PoseLogProb(const InteractorPose * pose, EnergyInfo & energy) const;
	
		virtual const Model3D * GetModel(const InteractorPose * pose) const;
		
		virtual int NActiveRegions() const;
		virtual void AddActiveRegion(ActiveRegion * region);
		virtual const ActiveRegion * GetActiveRegion(int regionID) const;
		
		static Interactor * ReadInteractor(const VKString & analysisDir, InteractorID type);
		virtual void ReadPosesFromMeshDir(const VKString & meshPoses, const VKString & posesList);
		virtual void ReadPosesTransforms(const VKString & canonicalFile);
//		virtual void ReadActiveRegions(const VKString & activeRegionsDir);
	
		virtual void GetObjectPredictions(const ShapePointCloud ** pnts, const ShapeValue ** predictions) const;
		virtual void ReadObjectPoints(const VKString & filename);
		virtual void ReadObjectDensePoints(const VKString & filename);
		virtual void ReadObjectPredictions(const VKString & filename);

		
		virtual void WriteInteractor(const VKString & analysisDir) const;
		virtual void WritePosesFromMeshDir(const VKString & meshPoses, const VKString & posesList) const;
		virtual void WritePosesTransforms(const VKString & canonicalFile) const;
//		virtual void WriteActiveRegions(const VKString & activeRegionsDir) const;

		static const char * ActorName(InteractorID id);
		virtual const ShapeMesh * GetMesh(const InteractorPose * pose) const;

//		virtual void SetPosePrior(InteractorPosePrior * prior);
//		virtual InteractorPosePrior * GetPosePrior();
//		virtual IKSolver * GetIK();
	
	//protected:
		virtual const R3MeshSearchTree * GetMeshSearchTree(const InteractorPose * pose,
														   const std::vector<int> ** tree2meshMap_vertices,
														   const std::vector<int> ** tree2meshMap_faces) const;
		
		virtual void ComputeCanonicalXForm();
		
		R3Point * m_pCentroid;
		//InteractorPosePrior * m_pPosePrior;
	
		R3Affine m_AllPoseXForm;	// default transformation
	
		InteractorID m_Type;	// actor vs object
		ActorType m_PoseType;	// mesh vs ragdoll vs rigged model
	
		ShapePointCloud * m_pObjectPoints;
		ShapeValue * m_pObjectPredictions;
		ShapePointCloud * m_pObjectDensePoints;

		std::vector<InteractorPose*> m_IndexableMeshPoses;	// all meshes
		std::vector<ActiveRegion *> m_ActiveRegions;		// all active regions
	
	// physics (various bullet geometries)
#ifdef INCLUDE_BULLET
		virtual void CreateRagDoll(InteractorPose * pose) const;
		virtual void UpdateRagDollPhysicsFromPose(InteractorPose * pose) const;
		virtual void UpdateRagDollPoseFromPhysics(InteractorPose * pose) const;
		virtual FKPose * GenerateRandomFKPose() const;
		virtual int RagDollNSamples(const InteractorPose * pose) const;
		virtual R3Point RagDollSample(const InteractorPose * pose, int pid) const;
		virtual int FillEvenDisk(ShapePointCloud * pnts, double radius, double height, int axis, int npoints) const;
		virtual int FillCylinderSides(ShapePointCloud * pnts, double radius, double height, int axis, int npoints) const;
		virtual void FillHeightAndRadius(const btCollisionShape * shape, double & r, double & h, int & upAxis) const;
	
		virtual btRigidBody * CreateRigidBody(btScalar mass, const btTransform& startTransform, btCollisionShape* shape) const;
		virtual btCollisionShape * GetCollisionShape(const InteractorPose * pose, const PhysicsConstants & physics) const;
		virtual btTriangleIndexVertexArray * CreateBulletMesh(const InteractorPose * pose, const PhysicsConstants & physics) const;
		virtual btBvhTriangleMeshShape * GetStaticShape(const InteractorPose * pose, const PhysicsConstants & physics) const;
		virtual btGImpactMeshShape * GetDynamicShape(const InteractorPose * pose, const PhysicsConstants & physics) const;
		virtual btCompoundShape * GetGridApproxShape(const InteractorPose * pose, const PhysicsConstants & physics) const;
		virtual btCompoundShape * GetUnionConvexShapes(const InteractorPose * pose, const PhysicsConstants & physics) const;
#endif
};

//////// RAGDOLL STUFF
enum
{
	BODYPART_PELVIS = 0,
	BODYPART_SPINE,
	BODYPART_HEAD,
	
	BODYPART_LEFT_UPPER_LEG,
	BODYPART_LEFT_LOWER_LEG,
	
	BODYPART_RIGHT_UPPER_LEG,
	BODYPART_RIGHT_LOWER_LEG,
	
	BODYPART_LEFT_UPPER_ARM,
	BODYPART_LEFT_LOWER_ARM,
	
	BODYPART_RIGHT_UPPER_ARM,
	BODYPART_RIGHT_LOWER_ARM,
	
	BODYPART_COUNT
};


enum
{
	JOINT_PELVIS_SPINE = 0,
	JOINT_SPINE_HEAD,
	
	JOINT_LEFT_HIP,
	JOINT_LEFT_KNEE,
	
	JOINT_RIGHT_HIP,
	JOINT_RIGHT_KNEE,
	
	JOINT_LEFT_SHOULDER,
	JOINT_LEFT_ELBOW,
	
	JOINT_RIGHT_SHOULDER,
	JOINT_RIGHT_ELBOW,
	
	JOINT_COUNT
};



#endif



