#include "IntersectionGrid.h"
#include "VkFunctions.h"
#include "ShapePointCloud.h"
#include "BindingEnergy.h"
#include "TimeProfiler.h"

IntersectionGrid::IntersectionGrid(Interactor * actor, Interactor* object, double rotation)
{
	// m_iGridFlags = ISECT_DepthPenalty | ISECT_WeightByContacts;
	m_iGridFlags =  ISECT_PenetrationPenalty | ISECT_WeightByNormals | ISECT_WeightByContacts;
	
	m_fGridCellSize = .02;
	m_iNumDilations = 3;;	// 4
	m_fDist2ContactSigma = 0.02;

	m_Interactors[Interactor::ACT_Actor] = actor;
	m_Interactors[Interactor::ACT_Object] = object;
	m_pObjectDepthGrid = NULL;
	m_pObjectVoronoiGrid = NULL;
	m_pDist2SurfaceGrid = NULL;
	m_pInsidePoints = NULL;

	InitGrid(m_fGridCellSize, pow(1024, 3), rotation);
}

void IntersectionGrid::ComputeEnergy(const PntCorrsPose * posepts, IKSolver * solver, EnergyInfo & energy) const
{
	energy.m_fIntersectionEnergy = 0;

	// go over each link
	//double dist2contact_sigma2 = pow(m_fDist2ContactSigma, 2.);
	double distThreshold = 0.1;
	const InteractorPose * meshpose = m_Interactors[Interactor::ACT_Object]->GetEnumeratedPose(0);
	const ShapeMesh * mesh = m_Interactors[Interactor::ACT_Object]->GetMesh(meshpose);
	if (mesh->NFaces()==0)
		return;
	
	for (int l=0; l<(int)solver->m_Links.size(); l++)
	{
		int j1 = solver->m_Links[l]->m_pParent->m_iJointID;
		int j2 = solver->m_Links[l]->m_pChild->m_iJointID;
		R3Point p1 = posepts->GetPosition(j1);
		R3Point p2 = posepts->GetPosition(j2);
		R3Point p1o = p1;
		R3Point p2o = p2;
		m_XForm.Apply(p1);
		m_XForm.Apply(p2);
		
		R3Vector linkorientation = p2o - p1o;
		linkorientation.Normalize();

		double linkpenalty = 0;
		double maxlinkpenalty = 0;
		for (double t=0; t<R3Distance(p1, p2); t+=m_fStep)
		{
			R3Point tp = (1.-t) * p1 + t * p2;
			R3Point tpo = (1.-t) * p1o + t * p2o;
			
			double weight_dist2contact = 1.;
			if (m_iGridFlags&ISECT_WeightByContacts)
			{
				double dist2contact = RN_INFINITY;
				for (int m=0; m<posepts->NMatches(); m++)
				{
					if (posepts->Prim(m).IsType(TargetPrimitive::PRIM_Point))
						dist2contact = vkMin(dist2contact, R3SquaredDistance(posepts->Prim(m).GetPosition(), tpo));
				}
				//weight_dist2contact = 1. - exp(-dist2contact / dist2contact_sigma2);
				if (dist2contact < m_fDist2ContactSigma)
					weight_dist2contact = 0;
				else
					weight_dist2contact = 1;
			}
			
			double depthPenalty = 0;
			if ((m_iGridFlags&ISECT_DepthPenalty))
			{
				assert(m_pObjectDepthGrid!=NULL);
				depthPenalty = m_pObjectDepthGrid->WorldValue(tp);
			}
			
			double weight_alignnormals = 1;
			double penetrationPenalty = 0;
			if ((m_iGridFlags&ISECT_PenetrationPenalty)!=0 || (m_iGridFlags&ISECT_WeightByNormals)!=0)
			{
				assert(mesh!=NULL && m_pDist2SurfaceGrid!=NULL && m_pObjectVoronoiGrid!=NULL);
				if (m_pDist2SurfaceGrid->WorldValue(tp)*m_fGridCellSize < distThreshold)
				{
					
					R3Point gp = m_pObjectVoronoiGrid->GridPosition(tp);
					int i = (int)gp.X();
					int j = (int)gp.Y();
					int k = (int)gp.Z();
					if (i>=0 && i<m_pObjectVoronoiGrid->XResolution()
						&& j>=0 && j<m_pObjectVoronoiGrid->YResolution()
						&& k>=0 && k<m_pObjectVoronoiGrid->ZResolution())
					{
						int fid = m_pObjectVoronoiGrid->GridValue(i, j, k)-1;
						assert(fid>=0 && fid<mesh->NFaces());
						R3Vector surfaceNormal = mesh->FaceNormal(fid);
						weight_alignnormals = vkAbs(surfaceNormal.Dot(linkorientation));
						
						R3TriangleVertex vtx1(mesh->VertexPosition(mesh->VertexOnFace(fid, 0)));
						R3TriangleVertex vtx2(mesh->VertexPosition(mesh->VertexOnFace(fid, 1)));
						R3TriangleVertex vtx3(mesh->VertexPosition(mesh->VertexOnFace(fid, 2)));
						R3Triangle tri(&vtx1, &vtx2, &vtx3);
						
						R3Span span(p1o, p2o);
						if (R3Intersects(tri, span)!=RN_NULL_CLASS_ID)
						{
							penetrationPenalty = 1.;
							//std::cout<<"     Penetration Penalty ("<<solver->LinkStr(l).c_str()<<")";
							//std::cout<<": "<<weight_alignnormals<<"*"<<weight_dist2contact<<std::endl;
						}
					}
					
				}
			}
			
			double currpenalty = depthPenalty*weight_dist2contact;	// depth penalty
			currpenalty += penetrationPenalty*weight_dist2contact*weight_alignnormals;  // intersection penalty
			
			linkpenalty += currpenalty*m_fStep;
			//if (maxlinkpenalty < currpenalty)
			//	std::cout<<"Updating final penalty("<<solver->LinkStr(l).c_str()<<"): "<<currpenalty<<std::endl;
			maxlinkpenalty = vkMax(maxlinkpenalty, currpenalty);
		}
		
		//energy.m_fIntersectionEnergy += vkMin(linkpenalty, .5);
		//energy.m_fIntersectionEnergy += linkpenalty;
		energy.m_fIntersectionEnergy += maxlinkpenalty;
	}
}

double IntersectionGrid::ComputePenalty(const R3Point & point) const
{
	R3Point tp = point;
	m_XForm.Apply(tp);
	return m_pObjectDepthGrid->WorldValue(tp) * m_fStep;	// real penalty
	//return m_pObjectDepthGrid->WorldValue(tp);	// debug
}

const ShapePointCloud * IntersectionGrid::InsidePoints(double * radius) const
{
	assert(m_pInsidePoints!=NULL);
	if (radius!=NULL)
		*radius = m_fStep;
	return m_pInsidePoints;
}

void IntersectionGrid::InitGrid(double step, int maxCells, double rotation)
{
	TimeProfiler timer;
	const InteractorPose * meshpose = m_Interactors[Interactor::ACT_Object]->GetEnumeratedPose(0);
	const ShapeMesh * mesh = m_Interactors[Interactor::ACT_Object]->GetMesh(meshpose);
	
	if (mesh->NFaces()==0)
	{
		std::cout<<"[WARNING] No faces => no intersection penalty"<<std::endl;
		return;
	}
	// TODO: PCA alignment
	
	// Create grid
	R3Box bbox;
	m_XForm = R3identity_affine;
	m_XForm.ZRotate(-rotation);
	for (int i=0; i<mesh->NVertices(); i++)
	{
		R3Point p = mesh->VertexPosition(i);
		m_XForm.Apply(p);
		if (i==0)
			bbox = R3Box(p,p);
		else
			bbox.Union(p);
	}
	
	std::cout<<"BBox: ["<<bbox.XMin()<<" "<<bbox.XMax()<<"], ["<<bbox.YMin()<<" "<<bbox.YMax();
	std::cout<<"], ["<<bbox.ZMin()<<" "<<bbox.ZMax()<<"] "<<std::endl;
	
	bbox = R3Box(R3Point(bbox.XMin() - step*6, bbox.YMin() - step*6, bbox.ZMin() - step*6),
				 R3Point(bbox.XMax() + step*6, bbox.YMax() + step*6, bbox.ZMax() + step*6));
	
	m_fStep = step;
	int xres=-1, yres=-1, zres=-1;
	int maxiter = 10000;
	while (maxiter-->=0)
	{
		xres = (bbox.XMax()-bbox.XMin()) / m_fStep;
		yres = (bbox.YMax()-bbox.YMin()) / m_fStep;
		zres = (bbox.ZMax()-bbox.ZMin()) / m_fStep;
		assert(xres>=0 && yres>=0 && zres>=0);
		if (xres*yres*zres > maxCells)
			m_fStep *= 1.1;
		else
			break;
	}
	assert(maxiter>0);
	
	std::cout<<"Resolution: "<<xres<<" x "<<yres<<" x "<<zres<<std::endl;
//	m_pObjectDepthGrid = new R3Grid(xres, yres, zres, bbox);
//	R3Grid * mask = new R3Grid(xres, yres, zres, bbox);
	m_pObjectVoronoiGrid = new R3Grid(xres, yres, zres, bbox);
	m_pDist2SurfaceGrid = new R3Grid(xres, yres, zres, bbox);

	
	m_pObjectDepthGrid = NULL;
	R3Grid * mask = NULL;
//	m_pObjectVoronoiGrid = NULL;
//	m_pDist2SurfaceGrid = NULL;

	
	if (m_pObjectDepthGrid!=NULL)	m_pObjectDepthGrid->Clear(0);
	if (m_pObjectVoronoiGrid!=NULL)	m_pObjectVoronoiGrid->Clear(0);
	if (m_pDist2SurfaceGrid!=NULL)	m_pDist2SurfaceGrid->Clear(0);
	
	// Fill grid values
	if (m_pObjectDepthGrid!=NULL || m_pDist2SurfaceGrid!=NULL)
	for (int f=0; f<mesh->NFaces(); f++)
	{
		R3Point p1 = mesh->VertexPosition(mesh->VertexOnFace(f, 0));
		R3Point p2 = mesh->VertexPosition(mesh->VertexOnFace(f, 1));
		R3Point p3 = mesh->VertexPosition(mesh->VertexOnFace(f, 2));
		m_XForm.Apply(p1);
		m_XForm.Apply(p2);
		m_XForm.Apply(p3);
		
		if (m_pObjectDepthGrid!=NULL)
			m_pObjectDepthGrid->RasterizeWorldTriangle(p1, p2, p3, 1.);
		if (m_pDist2SurfaceGrid!=NULL)
			m_pDist2SurfaceGrid->RasterizeWorldTriangle(p1, p2, p3, 1.);
	}
	if (m_pObjectDepthGrid!=NULL)
		m_pObjectDepthGrid->Dilate(m_iNumDilations);

	if (m_pObjectVoronoiGrid!=NULL)
	{
		ShapePointCloud * densepnts = m_Interactors[Interactor::ACT_Object]->m_pObjectDensePoints;
		assert(densepnts!=NULL);
		int fid;
		double b1, b2;
		const ShapeMesh * meshcurr=NULL;
		//std::cout<<"Fill Voronoi: "<<std::endl;;
		for (int pid=0; pid<densepnts->NumPoints(); pid++)
		{
//			if (pid%1000==0)
//				timer.WriteProgress("isectpoints", pid, densepnts->NumPoints());
			densepnts->GetPoint(pid).GetBarycentricCoords(&fid, &b1, &b2, &meshcurr);
			assert(meshcurr==mesh);
			R3Point gp = densepnts->PointPosition(pid);
			m_XForm.Apply(gp);
			R3Point gp2 = m_pObjectVoronoiGrid->GridPosition(gp);
			gp = gp2;
			int i = (int)gp.X();
			int j = (int)gp.Y();
			int k = (int)gp.Z();
			if (i>=0 && i<m_pObjectVoronoiGrid->XResolution()
				&& j>=0 && j<m_pObjectVoronoiGrid->YResolution()
				&& k>=0 && k<m_pObjectVoronoiGrid->ZResolution())
			{
				if (m_pObjectVoronoiGrid->GridValue(i, j, k)==0)
					m_pObjectVoronoiGrid->SetGridValue(i,j,k, fid+1);
			}
		}
		m_pObjectVoronoiGrid->Voronoi();
	}
//	std::cout<<std::endl;
//	std::cout<<"Voronoi now: "<<m_pObjectVoronoiGrid->Minimum()<<" - "<<m_pObjectVoronoiGrid->Maximum()<<" ";
//	std::cout<<"NFaces: "<<mesh->NFaces()<<std::endl;
//	assert((m_pObjectVoronoiGrid->Maximum()-1)<mesh->NFaces());
	
	if (m_pDist2SurfaceGrid!=NULL)
		m_pDist2SurfaceGrid->SquaredDistanceTransform();
	
	// Fill 'inside mask'
	if (mask!=NULL)
		mask->Add(1.);	// set all to 'included' values
	
	if (m_pObjectDepthGrid!=NULL)
	{
		assert(false);	// seems like there is memory allocation bug here...
		assert(mask!=NULL);
		// simple inside-out segmentation -- take orthogonal wavefront passes
		for (int passID=0; passID<5; passID++)
		{
			std::cout<<"Inside-out pass: "<<passID<<" / 5"<<std::endl;
			double passDirection;
			int passDimension;
			
			if (passID==0)	// top-down
			{
				passDimension = 2;
				passDirection = -1;	// down
			}
			else if (passID==1 || passID==2)	// left-right
			{
				passDimension = 0;
				passDirection = (passID==1) ? 1 : -1;
			}
			else if (passID==3 || passID==4)	// front-back
			{
				passDimension = 1;
				passDirection = (passID==3) ? 1 : -1;
			}
			else
				assert(false);
			
			int dim1 = 0;
			int dim2 = 1;
			
			if (passDimension==0)
				dim1 = 2;
			else if (passDimension==1)
				dim2 = 2;

			R3Point gp;
			for (int j=0; j<mask->Resolution(dim1); j++)
			for (int k=0; k<mask->Resolution(dim2); k++)
			for (int i=0; i<mask->Resolution(passDimension); i++)
			{
				if (passDirection<0)
					gp[passDimension] = (double)(mask->Resolution(passDimension) - 1 - i);
				else
					gp[passDimension] = (double)i;
				gp[dim1] = (double)j;
				gp[dim2] = (double)k;
							
				if (m_pObjectDepthGrid->GridValue((int)gp.X(), (int)gp.Y(), (int)gp.Z())>0)
					break;
				
				mask->SetGridValue((int)gp.X(), (int)gp.Y(), (int)gp.Z(), 0.);
			}
		}
		std::vector<int> nhdX;
		std::vector<int> nhdY;
		std::vector<int> nhdZ;
		for (int i=0; i<6; i++)
		{
			nhdX.push_back(0);
			nhdY.push_back(0);
			nhdZ.push_back(0);
		}
		nhdX[0] = -1;
		nhdX[1] = 1;
		nhdY[2] = -1;
		nhdY[3] = 1;
		nhdZ[4] = -1;
		nhdZ[5] = 1;
		
		int nx = mask->Resolution(0);
		int ny = mask->Resolution(1);
		int nz = mask->Resolution(2);
		std::cout<<"Region-growing..."<<std::endl;
		
	//	// region-growing to prune remaining
		for (int i=1; i<nx-1; i++)
		for (int j=1; j<ny-1; j++)
		for (int k=1; k<nz-1; k++)
		{
			// if non-surface and touching "empty" - start region growing
			bool ijk_maskedOut = mask->GridValue(i,j,k)<1;
			bool ijk_onSurface = m_pObjectDepthGrid->GridValue(i,j,k)>0;
			bool visit = !ijk_maskedOut && !ijk_onSurface;
			
			int nminadj = 1;
			int adjacentEmpty = 0;
			for (int n=0; n<6 && adjacentEmpty<nminadj && visit; n++)
			{
				bool nijk_maskedOut = mask->GridValue(i+nhdX[n], j+nhdY[n], k+nhdZ[n])<1;
				bool nijk_onSurface = m_pObjectDepthGrid->GridValue(i+nhdX[n], j+nhdY[n], k+nhdZ[n])>0;
				if (nijk_maskedOut && !nijk_onSurface)
					adjacentEmpty++;
			}
			if (j==1 && k==1)
				timer.WriteProgress("progisectgrid", i, nx);
			std::vector<int> regionGrowing;
			int nvisit = 0;
			if (visit && adjacentEmpty>=nminadj)
			{
				regionGrowing.push_back(i*ny*nz + j*nz + k);
				while (regionGrowing.size()>0)
				{
					int id = regionGrowing[0];
					regionGrowing.erase(regionGrowing.begin());
					int ci = id / (ny*nz);
					int cj = (id - ci * ny*nz)/nz;
					int ck = id - ci*ny*nz - cj*nz;
					bool maskedOut1 = mask->GridValue(ci, cj, ck)<1;
					if (maskedOut1)
						continue;
					nvisit++;
					//mask->SetGridValue(ci,cj,ck, .5);
					mask->SetGridValue(ci,cj,ck, 0);
					
					for (int n=0; n<6; n++)
					{
						int cci = ci+nhdX[n];
						int ccj = cj+nhdY[n];
						int cck = ck+nhdZ[n];
						bool validBounds = cci>=0 && cci<nx && ccj>=0 && ccj<ny && cck>=0 && cck<nz;
						bool onSurface = validBounds && (m_pObjectDepthGrid->GridValue(cci, ccj, cck)>0);
						bool maskedOut = validBounds && (mask->GridValue(cci, ccj, cck)<1);
						if (validBounds && !maskedOut && !onSurface)
							regionGrowing.push_back(cci*ny*nz + ccj*nz + cck);
					}
				}
				//std::cout<<"\nRegion-grow: ["<<i<<" "<<j<<" "<<k<<"] "<<nvisit<<std::endl;
			}
		}
		std::cout<<std::endl;
		
		// multiply with mask
		std::cout<<"Squared Dist Transform"<<std::endl;
		m_pObjectDepthGrid->DetectEdges();
		m_pObjectDepthGrid->Threshold(.5, 0, 1.);
		m_pObjectDepthGrid->SquaredDistanceTransform();
		
		std::cout<<"Multiply by mask"<<std::endl;
		m_pObjectDepthGrid->Multiply(*mask);
		//	m_pObjectDepthGrid = mask;
		//	mask = NULL;
		//	double * xxx = new double[1000000];
		//	for (int i=0; i<1000000; i++)
		//		xxx[i] = 0;
		std::cout<<"Prepare grid inside points (visualization only)"<<std::endl;
		// inside points (for visualization)
		m_pInsidePoints = new ShapePointCloud(NULL, "gridpnts");
		for (int x=0; x<m_pObjectDepthGrid->XResolution(); x++)
			for (int y=0; y<m_pObjectDepthGrid->YResolution(); y++)
				for (int z=0; z<m_pObjectDepthGrid->ZResolution(); z++)
				{
					if (m_pObjectDepthGrid->GridValue(x, y, z)>0)
					{
						//std::cout<<"Add point: "<<x<<" "<<y<<" "<<z<<std::endl;
						R3Point grid((double) x, (double)y, (double)z);
						m_pObjectDepthGrid->GridToWorldTransformation().Apply(grid);
						m_XForm.ApplyInverse(grid);
						m_pInsidePoints->AddPoint(ShapePoint(grid));
					}
				}
		
		std::cout<<"Grid Filled Points: "<<m_pInsidePoints->NumPoints()<<std::endl;
		delete mask;
	}
}


