#include "TimeProfiler.h"

#include "Binding.h"
#include "Interactor.h"
#include "InteractorPose.h"
#include "BindingManager.h"
#include "InteractorPose.h"
#include "PntCorrsPose.h"
#include "IKSolver.h"
#include <execinfo.h>

//////////////////////////////
////  TargetPrimitive
//////////////////////////////
TargetPrimitive::TargetPrimitive()
{
	m_Type = PRIM_None;
	m_Point = ShapePoint();
}

TargetPrimitive::TargetPrimitive(PrimType tp, const ShapePoint & pnt)
{
	m_Point = pnt;
	m_Type = tp;
}

bool TargetPrimitive::IsType(PrimType type) const
{
	return m_Type == type;
}

R3Point TargetPrimitive::GetPosition() const
{
	if (!(IsType(PRIM_Point) || IsType(PRIM_Ground) || IsType(PRIM_PointMeshless)))
	{
//		int size = 1000;
//		void * buffer = new void*[size];
//		size = backtrace(&buffer, size);
//		char ** buffStr = backtrace_symbols((void *const *)buffer, size);
//		for (int i=0; i<size; i++)
//			std::cout<<buffStr[i]<<std::endl;
		assert(false);
		//return R3Point(0,0,0);
	}
	if (IsType(PRIM_Ground) && m_Point.Invalid())
		return R3Point(0,0,0);
	
	assert(!m_Point.Invalid());
	return m_Point.GetPosition();
}

const ShapePoint & TargetPrimitive::GetPoint() const
{
	assert(IsType(PRIM_Point));
	assert(!m_Point.Invalid());
	return m_Point;
}


int TargetPrimitive::PointID(ShapePointCloud * pointset) const
{
	assert(IsType(PRIM_Point));
	assert(!m_Point.Invalid());
	if (m_Point.SamePntsSet(pointset) || pointset==NULL)
		return m_Point.PointID();
	return -1;
}



//////////////////////////////
////  PntCorrsPose
//////////////////////////////
PntCorrsPose::PntCorrsPose(const PntCorrsPose * match)
{
	m_iPoseID = match->m_iPoseID;
	m_Type = match->m_Type;
	m_aTargets = match->m_aTargets;
	m_RegionID = match->m_RegionID;
	m_pObject = match->m_pObject;

	
	m_SkeletonParams = match->m_SkeletonParams;
	m_JointPositions = match->m_JointPositions;

	// below also nto useful...
	m_NormalizingXForm = match->m_NormalizingXForm;
	m_CanonicalPoseComputed = match->m_CanonicalPoseComputed;
	m_CanonicalPose = match->m_CanonicalPose;
	m_sPoseName = match->m_sPoseName;
	m_MeshArea = match->m_MeshArea;
	m_iPoseParamsFlags = match->m_iPoseParamsFlags;
}

PntCorrsPose::PntCorrsPose(Interactor * object)
{
	//assert(object!=NULL);
	m_Type = TYPE_PntCorrsPose;
	m_pObject = object;
	m_iPoseParamsFlags = 0;
	m_iPoseID = -1;
}

PntCorrsPose::PntCorrsPose()
{
	m_Type = TYPE_PntCorrsPose;
	m_pObject = NULL;
	m_iPoseParamsFlags = 0;
	m_iPoseID = -1;
}

PntCorrsPose::~PntCorrsPose()
{
}

VKString PntCorrsPose::GetStr(const IKSolver * ik) const
{
	if (!IsParamsInit(ik->NJointParams()))
		return "Pose is not initalized\n";
		
	VKString str = "Pose Params:\n";
	for (int l=0; l<(int)ik->m_Links.size(); l++)
	{
		SkeletonLink * link = ik->m_Links[l];
		int ndims = SkeletonLink::NDims(link->m_JointType);
		if (ndims==0)
			continue;
		VKString descrStr = link->m_pParent->m_sJointName+" -> "+link->m_pChild->m_sJointName;
		str += VKString("")+descrStr.CreateStrWithPadding(30, " ")+": ";
		int offset = ik->m_ParamOffsetByLinkID[l];
		
		str += VKString(" [  ");
		for (int d=0; d<ndims; d++)
			str += VKString::number(GetParam(offset+d)).CreateStrWithPadding(13, " ");
		str += "]\n";
	}
	return str;
}

void PntCorrsPose::WritePose(FILE * file) const
{
	fwrite(&m_Type, sizeof(PoseType), 1, file);
	
	int npnts = (int)m_aTargets.size();
	fwrite(&npnts, sizeof(int), 1, file);
	
	int fid,pid;
	double bary[3];
	double pos[3];
	const ShapeMesh * mesh;
	// write points
	for (int i=0; i<npnts; i++)
	{
		int rid = m_RegionID[i];
		fwrite(&rid, sizeof(int), 1, file);
		//std::cout<<rid<<" -> "<<m_aTargets[i].m_Type<<std::endl;
		
		R3Point pospnt(0,0,0);
		if (m_aTargets[i].IsType(TargetPrimitive::PRIM_Point) || m_aTargets[i].IsType(TargetPrimitive::PRIM_Ground))
			pospnt = m_aTargets[i].GetPosition();
		if (m_aTargets[i].IsType(TargetPrimitive::PRIM_PartialUnassigned))
		{
			fid = -2;
			pid = -2;
			bary[0] = 0; bary[1] = 0; bary[2] = 0;
		}
		else if (m_aTargets[i].IsType(TargetPrimitive::PRIM_Point))
		{
			mesh = m_aTargets[i].GetPoint().GetMeshGeometry();
			if (mesh->NFaces()>0)
			{
				m_aTargets[i].GetPoint().GetBarycentricCoords(&fid, &(bary[0]), &(bary[1]), &mesh);
				bary[2] = 1. - bary[0] - bary[1];
			}
			else
			{
				fid =m_aTargets[i].GetPoint().VertexID();
				bary[0] = 0; bary[1] = 0; bary[2] = 0;
			}
			pid = m_aTargets[i].PointID(NULL);
		}
		else
		{
			assert(m_aTargets[i].IsType(TargetPrimitive::PRIM_Ground));
			pid = -1;
			fid  = -1;
			bary[0] = 0; bary[1] = 0; bary[2] = 0;
		}
		pos[0] = pospnt.X();
		pos[1] = pospnt.Y();
		pos[2] = pospnt.Z();
		fwrite(&pid, sizeof(int), 1, file);
		fwrite(&fid, sizeof(int), 1, file);
		fwrite(bary, sizeof(double), 3, file);
		fwrite(pos, sizeof(double), 3, file);
	}
	
	// write skeleton parameters (angles and joint positions)
	int nskeletonparams = (int)m_SkeletonParams.size();
	int njointpositions = (int)m_JointPositions.size();
	fwrite(&nskeletonparams, sizeof(int), 1, file);
	for (int i=0; i<nskeletonparams; i++)
		fwrite(&(m_SkeletonParams[i]), sizeof(double), 1, file);
	
	fwrite(&njointpositions, sizeof(int), 1, file);
	for (int i=0; i<njointpositions; i++)
	{
		pos[0] = m_JointPositions[i].X();
		pos[1] = m_JointPositions[i].Y();
		pos[2] = m_JointPositions[i].Z();
		fwrite(pos, sizeof(double), 3, file);
	}
	fwrite(&m_iPoseParamsFlags, sizeof(int), 1, file);
}

void PntCorrsPose::ReadPose(FILE * file)
{
	assert(m_pObject!=NULL && m_pObject->NumEnumeratedPoses()==1);
	const ShapeMesh * mesh = m_pObject->GetMesh(m_pObject->GetEnumeratedPose(0));
	int npnts;
	int fid, pid, rid;
	double bary[3];
	double pos[3];
	
	fread(&npnts, sizeof(int), 1, file);
	for (int i=0; i<npnts; i++)
	{
		fread(&rid, sizeof(int), 1, file);
		fread(&pid, sizeof(int), 1, file);
		fread(&fid, sizeof(int), 1, file);
		fread(bary, sizeof(double), 3, file);
		fread(pos, sizeof(double), 3, file);
		
		m_RegionID.push_back(rid);
		int lid = (int)m_aTargets.size();
		if (fid >= 0 && mesh->NFaces()>0)
			m_aTargets.push_back(TargetPrimitive(TargetPrimitive::PRIM_Point, ShapePoint(mesh, fid, bary[0], bary[1], NULL, pid)));
		else if (fid >= 0 && mesh->NFaces()==0)
			m_aTargets.push_back(TargetPrimitive(TargetPrimitive::PRIM_Point, ShapePoint(mesh, fid, pos, NULL, pid)));
		else if (pid==-2 && fid==-2)
			m_aTargets.push_back(TargetPrimitive(TargetPrimitive::PRIM_PartialUnassigned, ShapePoint(mesh, fid, pos, NULL, pid)));
		else
			m_aTargets.push_back(TargetPrimitive(TargetPrimitive::PRIM_Ground, ShapePoint(R3Point(pos[0], pos[1], pos[2]), NULL)));
	}
	
	
	// read skeleton parameters (angles and joint positions)
	int nskeletonparams = 0;
	int njointpositions = 0;
	fread(&nskeletonparams, sizeof(int), 1, file);
	assert(!feof(file));
	for (int i=0; i<nskeletonparams; i++)
		m_SkeletonParams.push_back(0);
	for (int i=0; i<nskeletonparams; i++)
		fread(&(m_SkeletonParams[i]), sizeof(double), 1, file);
	
	
	fread(&njointpositions, sizeof(int), 1, file);
	assert(!feof(file));
	for (int i=0; i<njointpositions; i++)
	{
		fread(pos, sizeof(double), 3, file);
		m_JointPositions.push_back(R3Point(pos[0], pos[1], pos[2]));
	}
	
	if (!feof(file))
		fread(&m_iPoseParamsFlags, sizeof(int), 1, file);
}

void PntCorrsPose::CreatePosesFromGTFile(const VKString & gtFile, const VKString & meshFile, const VKString & xform,
										 std::vector<PntCorrsPose*> & poses)
{
	//ShapeTriangleSoup * mesh = new ShapeTriangleSoup();
	//mesh->ReadGeometryFile(meshFile);
	CreatePosesFromGTFile(gtFile, NULL, xform, poses);
}

void PntCorrsPose::CreatePosesFromGTFile(const VKString & gtFile, const ShapeMesh * mesh,
										 const VKString & xform, std::vector<PntCorrsPose*> & poses)
{
	double scale = 1.;
	VKString tempStr;
	if (xform!="none")
	{
		std::ifstream xformStream(xform.c_str());
		if (!xformStream.is_open())
			std::cout<<"[ERROR] Cannot read "<<xform.c_str()<<std::endl;
		assert(xformStream.is_open());
		tempStr.readToken(xformStream);		assert(tempStr=="Scale");
		xformStream>>scale;
		xformStream.close();
	}
	
	int nposes, poseid, nregions, regionid, npnts, fid;
	double b1, b2, b3, x, y, z;
	
	
	std::ifstream textStream(gtFile.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] Cannot read "<<gtFile.c_str()<<std::endl;
	assert(textStream.is_open());
	tempStr.readToken(textStream);	assert(tempStr=="NPoses");
	textStream>>nposes;
	
	for (int i=0; i<nposes; i++)
	{
		PntCorrsPose * pose = new PntCorrsPose((Interactor *)NULL);
		tempStr.readToken(textStream);	assert(tempStr=="Pose");
		textStream>>poseid;		assert(poseid==i);
		tempStr.readToken(textStream);	assert(tempStr=="Regions");
		textStream>>nregions;
		for (int r=0; r<nregions; r++)
		{
			tempStr.readToken(textStream);  assert(tempStr=="Region");
			textStream>>regionid;
			tempStr.readToken(textStream);
			tempStr.readToken(textStream);	assert(tempStr=="pnts");
			textStream>>npnts;
			
			// take average point...
			R3Point avePnt(0,0,0);
			for (int p=0; p<npnts; p++)
			{
				textStream>>fid>>b1>>b2>>b3>>x>>y>>z;
				avePnt += R3Vector(x,y,z);
			}
			R3Point newpnt = (avePnt / npnts) * scale;
			pose->m_RegionID.push_back(regionid);
			int id = (int)pose->m_aTargets.size();
			//pose->m_aFeaturePoints.push_back(std::vector<ShapePoint>());
			if (fid==-1)
				pose->m_aTargets.push_back(TargetPrimitive(TargetPrimitive::PRIM_Ground, ShapePoint(newpnt)));
			else if (mesh==NULL)
				pose->m_aTargets.push_back(TargetPrimitive(TargetPrimitive::PRIM_Point, ShapePoint(newpnt)));
			else if (mesh->NFaces()>0)	// face
				pose->m_aTargets.push_back(TargetPrimitive(TargetPrimitive::PRIM_Point, ShapePoint(mesh, fid, b1, b2)));
			else	// vertex
				pose->m_aTargets.push_back(TargetPrimitive(TargetPrimitive::PRIM_Point, ShapePoint(mesh, fid)));
		}
		
		// read pose angles
		tempStr.readToken(textStream);  assert(tempStr=="NSkeleton:");
		int nskeletonparams;
		textStream>>nskeletonparams;
		pose->m_SkeletonParams.clear();
		for (int i=0; i<nskeletonparams; i++)
			pose->m_SkeletonParams.push_back(0);
		for (int i=0; i<nskeletonparams; i++)
			textStream>>pose->m_SkeletonParams[i];
		
		tempStr.readToken(textStream);
		if (tempStr=="ParamFlags:")
		{
			int nflags = 0;
			textStream>>nflags;
			pose->m_iPoseParamsFlags = 0;
			for (int p=0; p<nflags; p++)
			{
				tempStr.readToken(textStream);
				if (tempStr=="Translate")
					pose->m_iPoseParamsFlags = pose->m_iPoseParamsFlags | IKSolver::IK_RigidTranslate;
				else if (tempStr=="RotateAroundUp")
					pose->m_iPoseParamsFlags = pose->m_iPoseParamsFlags | IKSolver::IK_RigidRotateAroundUp;
				else if (tempStr=="RotateLeftRight")
					pose->m_iPoseParamsFlags = pose->m_iPoseParamsFlags | IKSolver::IK_RigidRotateLeftRight;
				else if (tempStr=="RotateFrontBack")
					pose->m_iPoseParamsFlags = pose->m_iPoseParamsFlags | IKSolver::IK_RigidRotateFrontBack;
			}
		}
		else
			pose->m_iPoseParamsFlags = IKSolver::IK_RigidTranslate | IKSolver::IK_RigidRotateAroundUp;
		
		poses.push_back(pose);
	}
	textStream.close();
}

bool PntCorrsPose::WritePosesGTFile(const VKString & gtFile, const std::vector<PntCorrsPose*> & poses,
									const VKStringList & regionNames, const VKStringList & regionTypes,
									const std::vector<int> & regionNPnts)
{
	bool allPoints = true;
	FileManagement::CreateDirsOnPathToFile(gtFile.c_str());
	std::ofstream textStream(gtFile.c_str());
	assert(textStream.is_open());
	int nposes = (int)poses.size();
	textStream<<"NPoses "<<nposes<<"\n";
	int fid;
	double b1, b2, b3;
	const ShapeMesh * tempmesh;
	R3Point pos;
	for (int i=0; i<nposes; i++)
	{
		PntCorrsPose * pose = poses[i];
		// compute nregions
		int nregions = 0;
		for (int r=0; r<(int)pose->NMatches(); r++)
		{
			if (pose->Prim(r).IsType(TargetPrimitive::PRIM_None)==false)
				nregions++;
		}
		// write regions
		textStream<<"Pose "<<i<<" Regions "<<nregions<<"\n";
		for (int r=0; r<(int)pose->NMatches(); r++)
		{
			int rid = pose->m_RegionID[r];
			
			if (pose->Prim(r).IsType(TargetPrimitive::PRIM_None))
			{
				if (r < pose->NMatches()-1)
					allPoints = false;
				continue;
			}
			nregions--;
			if (regionNames.count()>0)
				textStream<<"Region "<<rid<<" "<<regionNames[rid].c_str()<<" "<<regionTypes[rid].c_str()<<" "<<1<<"\n";
			else
				textStream<<"Region "<<rid<<" ignore_not_err pnts 1\n";
			
			if (pose->m_aTargets[r].IsType(TargetPrimitive::PRIM_Point))
			{
				const ShapeMesh * mesh = pose->m_aTargets[r].GetPoint().GetMeshGeometry();
				if (mesh->NFaces()>0)
				{
					pose->m_aTargets[r].GetPoint().GetBarycentricCoords(&fid, &b1, &b2, &tempmesh);
					b3 = 1. - b1 - b2;
				}
				else
				{
					fid = pose->m_aTargets[r].GetPoint().VertexID();
					b1 = 0; b2 = 0; b3 = 0;
				}
			}
			else if (pose->m_aTargets[r].IsType(TargetPrimitive::PRIM_PointMeshless))
			{
				fid = -2;
				b1 = 0.;
				b2 = 0;
				b3 = 0;
			}
			else
			{
				assert(pose->m_aTargets[r].IsType(TargetPrimitive::PRIM_Ground));
				fid = -1;
				b1 = 0.;
				b2 = 0;
				b3 = 0;
			}
			pos = pose->m_aTargets[r].GetPosition();
			
			textStream<<fid<<" "<<b1<<" "<<b2<<" "<<b3<<" "<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<"\n";
		}
		assert(nregions==0);
		
		// write pose angles
		int nskeletonparams = (int)pose->m_SkeletonParams.size();
		textStream<<"NSkeleton: "<<nskeletonparams<<"\n";
		for (int s=0; s<nskeletonparams; s++)
			textStream<<pose->m_SkeletonParams[s]<<((s==nskeletonparams-1)?"\n":" ");
		
		// write explanation for angles
		VKStringList paramFlagsStr;
		if ((pose->m_iPoseParamsFlags & IKSolver::IK_RigidTranslate))
			paramFlagsStr.push_back("Translate");
		if ((pose->m_iPoseParamsFlags & IKSolver::IK_RigidRotateLeftRight))
			paramFlagsStr.push_back("RotateLeftRight");
		if ((pose->m_iPoseParamsFlags & IKSolver::IK_RigidRotateAroundUp))
			paramFlagsStr.push_back("RotateAroundUp");
		if ((pose->m_iPoseParamsFlags & IKSolver::IK_RigidRotateFrontBack))
			paramFlagsStr.push_back("RotateFrontBack");
		textStream<<"ParamFlags: "<<paramFlagsStr.count()<<"\n";
		for (int p=0; p<paramFlagsStr.count(); p++)
			textStream<<paramFlagsStr[p].c_str()<<((p==paramFlagsStr.count()-1)?"\n":" ");
	}
	textStream.close();
	return allPoints;
}

int PntCorrsPose::GetPoseParamsFlags() const
{
	return m_iPoseParamsFlags;
}

double PntCorrsPose::PoseDistanceScale() const
{
	if (m_pObject==NULL)
		return 1.;
	R3Vector scale(0,0,1);
	m_pObject->GetEnumeratedPose(0)->m_NormalizingXForm.Apply(scale);
	return scale.Length();
}

int PntCorrsPose::NMatches() const
{
	return (int)m_aTargets.size();
}

TargetPrimitive & PntCorrsPose::Prim(int matchID)
{
	assert(matchID>=0 && matchID < (int)m_aTargets.size());
	return m_aTargets[matchID];
}
const TargetPrimitive & PntCorrsPose::Prim(int matchID) const
{
	assert(matchID>=0 && matchID < (int)m_aTargets.size());
	return m_aTargets[matchID];
}
int PntCorrsPose::Region(int matchID) const
{
	assert(matchID>=0 && matchID < (int)m_RegionID.size());
	return m_RegionID[matchID];
}

int & PntCorrsPose::Region(int matchID)
{
	assert(matchID>=0 && matchID < (int)m_RegionID.size());
	return m_RegionID[matchID];	
}

int PntCorrsPose::AddMatch(int regionID, const TargetPrimitive & trg)
{
	int id = (int)m_aTargets.size();
	m_aTargets.push_back(trg);
	m_RegionID.push_back(regionID);
	return id;
}

void PntCorrsPose::InitParams(int nparams, int flags)
{
	m_SkeletonParams.clear();
	for (int i=0; i<nparams; i++)
		m_SkeletonParams.push_back(0);
	m_iPoseParamsFlags = flags;
}

void PntCorrsPose::InitJointPositions(int njoints)
{
	m_JointPositions.clear();
	for (int i=0; i<njoints; i++)
		m_JointPositions.push_back(R3Point(0,0,0));
}

bool PntCorrsPose::IsParamsInit(int nparams) const
{
	return nparams == (int)m_SkeletonParams.size();
}

bool PntCorrsPose::IsPositionsInit(int npositions) const
{
	return npositions==(int)m_JointPositions.size();
}

double & PntCorrsPose::Param(int skeletonParam)
{
	assert(skeletonParam>=0 && skeletonParam < (int)m_SkeletonParams.size());
	return m_SkeletonParams[skeletonParam];
}

R3Point & PntCorrsPose::Position(int jointID)
{
	assert(jointID>=0 && jointID < (int)m_JointPositions.size());
	return m_JointPositions[jointID];
}

R3Point PntCorrsPose::GetPosition(int jointID) const
{
	assert(jointID>=0 && jointID < (int)m_JointPositions.size());
	return m_JointPositions[jointID];
}

double PntCorrsPose::GetParam(int skeletonParam) const
{
	assert(skeletonParam>=0 && skeletonParam < (int)m_SkeletonParams.size());
	return m_SkeletonParams[skeletonParam];
}

void PntCorrsPose::Clear(int matchID)
{
	assert(matchID>=0 && matchID < (int)m_aTargets.size());
	m_aTargets.erase(m_aTargets.begin()+matchID);
	m_RegionID.erase(m_RegionID.begin()+matchID);
}

int PntCorrsPose::GetPoseID() const
{
	return m_iPoseID;
}

void PntCorrsPose::SetPoseID(int poseID)
{
	m_iPoseID = poseID;
}

