	#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"

#include "ModelDatabase.h"
#include "ShapeTriangleSoup.h"


#ifndef __InteractorPosePrior_H
#define __InteractorPosePrior_H

// forward declarations
class InteractorPose;
class PntCorrsPose;
class IKSolver;
struct EnergyInfo;

struct JacobianInfo
{
	public:
		JacobianInfo();
};

/**
 * Gaussian distribution
 */
class PosePointGauss
{
public:
	PosePointGauss(int ndims, bool angles = true);
	
	virtual double Sample(int dimid, double uniform1, double uniform2) const;
	
	virtual void Update(const double * value);
	virtual void UpdateVarianceFromZero(const double * value);
	
	virtual void WriteDistribution(FILE * file) const;
	virtual void ReadDistribution(FILE * file);
	
	static PosePointGauss * CreateDistribution(FILE * file);
	static double GetNearestAngle(double mean, double angle) ;
	
	double * m_aMeans;
	double * m_aVariances;
	int m_iNSamples;
	int m_iNDims;
	bool m_bAngles;
	
	double m_fMagnitude;
};

/**
 * Evaluates the pose probability
 */
class PosePrior
{
	public:
		enum PosePriorType
		{
			PRIOR_GaussAngles 		= 1,
			PRIOR_MultiGauss		= 2
		};
		PosePrior();
		virtual void Train(const std::vector<const InteractorPose *> & pose);
		virtual void WritePrior(const VKString & filename) const;
		virtual void ReadPrior(FILE * file);
		virtual PosePriorType PriorType() const;
	
		virtual bool IsAngles() const;
	
		virtual bool QuickValidityTest(const InteractorPose * pose) const;
		virtual double COMDeviation(const PntCorrsPose * pose, const R3Point & objCOM) const;
	
	// for ik solver
		virtual int NTerms() const;
		virtual bool IsSymmetryTerm(int termID) const;
		virtual double EnergyTerm(const InteractorPose * pose, int termID, EnergyInfo & info) const;
	
		virtual double Jacobian(const InteractorPose * pose, JacobianInfo & info, int termID, int  paramID) const;
		virtual void NonZeroJacobians(std::vector<int> & termIDs, std::vector<int> & paramIDs) const;
		virtual int NSuggestedInitPoses() const;
		virtual void InitPose(int poseID, InteractorPose * pose) const;
		virtual void RandomSample(InteractorPose * pose) const;
	
		static PosePrior * ReadPrior(const VKString & filename);

		virtual void RecordPointStats(const std::vector<PntCorrsPose*> & poses, int nregions, const VKStringList & comFiles);
		virtual double SurfaceProbability(int poseid, int regionid) const;
		virtual double GroundProbability(int poseid, int regionid) const;
		virtual double UnassignedProbability(int poseid, int regionid) const;
	
		virtual void WritePointStats(const VKString & filename) const;
		virtual void ReadPointStats(const VKString & filename);
	
//	protected:
		std::vector<int> m_iNPoses;
		std::vector<std::vector<int> > m_RegionNGround;
		std::vector<std::vector<int> > m_RegionNSurface;
		
		std::vector<PosePointGauss *> m_comDistrPerStyle;
		PosePointGauss * m_comDistrGlobal;

};

/**
 * Kinematic model w/ prior on angles
 */
class PriorAngles : public PosePrior
{
	public:
		enum ProbGridStyle
		{
			GRID_None = 0,
			GRID_CondOnRefPnt = 1<<0,
			GRID_CondOnPoseID = 1<<1,
			GRID_CondOnSymm   = 1<<2
		};
	
		PriorAngles();
		virtual void SetIK(IKSolver * solver);
		virtual IKSolver * GetIK();
		virtual bool IsAngles() const;
		virtual double Jacobian(const InteractorPose * pose, JacobianInfo & info, int termID, int paramID) const;
	
		static const PntCorrsPose * CastPose(const InteractorPose * pose);
		virtual bool QuickValidityTest(const InteractorPose * pose) const;
		virtual bool ValidPairwiseDistances(const PntCorrsPose * pose) const;
			
		virtual void ReadGridData(FILE * file);
		virtual void WriteGridData(FILE * file) const;
	
		virtual void SamplePosesAndFillGrid(int gridFlags, int nsamplesPerGaussian, std::vector<PntCorrsPose*> * saveposes);
		virtual void FillGridFromClusteredPoses(int gridFlags, const std::vector<PntCorrsPose *> & clustered);
		virtual void FillCandidateRegions(const std::vector<PntCorrsPose *> & poses);
		static double ExactWorldValue(const R3Point & wp, const R3Grid * grid);
		
		virtual int GetGridID(int regionID, const PntCorrsPose * cond) const;
		virtual const R3Grid * GetGrid(int gridID) const;
		virtual const ShapePointCloud * GetNonZeroProb(int gridID, double * gridsize = NULL) const;
		virtual int NCandidateContacts(int poseid) const;
		virtual int RegionID(int poseid, int candidateID) const;
		virtual bool IsCompatible(int poseID, int regionID) const;
	
	protected:
		int m_iGridStyleFlags;
		
		// conditionals
		int m_iNCondContactPnts;
		int m_iNCondPoses;
		int m_iNCondSymm;	// multiplier depends on contact point ID
		double m_fGridCellSize;
		
		std::vector<std::vector<int> > m_CandidateRegions;	// conditioned on pose id
		std::map<int, int> m_CondID2GridID;
		std::vector<R3Grid*> m_ContactProbGrids;
		std::vector<ShapePointCloud*> m_NonZeroGridPnts;
	
		IKSolver * m_pIK;
		std::vector<std::map<int, double > > m_CachedPairwiseDists;  // prune implausible pauses (points too far)

		virtual int UniqueCondID(int targetRegion, int referenceRegion, int poseID, int symmetryRegion=0) const;
};

#endif



