#include "PosePrior.h"
#include "InteractorPose.h"
#include "VkFunctions.h"
#include "IKSolver.h"
#include "BindingManager.h"
#include "PosePriorGauss.h"
#include "PntCorrsPose.h"

////////
// PriorGaussAngles
PosePriorGauss::PosePriorGauss()
{
	m_fMaxPenalty = RN_INFINITY;
	m_iNTotal = 0;
}

PosePrior::PosePriorType PosePriorGauss::PriorType() const
{
	return PRIOR_GaussAngles;
}

int PosePriorGauss::NTerms() const
{
	assert(m_pIK!=NULL);
	int nparams =0;
	for (int i=0; i<(int)m_pIK->m_Links.size(); i++)
		nparams += SkeletonLink::NDims(m_pIK->m_Links[i]->m_JointType);
	return nparams;
}

double PosePriorGauss::EnergyTerm(const InteractorPose * poseIn, int termID, EnergyInfo & info) const
{
	int dimid, linkid;
	GetAngleDistrID(termID, linkid, dimid);
	const PosePointGauss * gauss = m_AngleDistributions[linkid];
	assert(gauss!=NULL);
	assert(m_pIK!=NULL);
	int offset_skeleton = m_pIK->m_ParamOffsetByLinkID[linkid];
	
	const PntCorrsPose * pose = CastPose(poseIn);
	
	double mean = gauss->m_aMeans[dimid];
	double variance = gauss->m_aVariances[dimid];
	double angle = pose->GetParam(offset_skeleton + dimid);
	angle = PosePointGauss::GetNearestAngle(mean, angle);
	
	double retval = 0;
//	std::cout<<"Variance: "<<variance<<std::endl;
	if (variance==0)
	{
//		if (angle != mean)
//			return m_fMaxPenalty;
//		else
		retval = 0;
	}
	else
		retval = pow(angle - mean, 2.) / (2.*variance);
	
	if (info.m_EnergyVis==ENVIS_AngleAll)
	{
		int jointid = m_pIK->m_Links[linkid]->m_pParent->m_iJointID;
		assert(jointid < (int)info.m_PerPointEnergy.size());
		info.m_PerPointEnergy[jointid] += retval;
	}

	return retval;
}

void PosePriorGauss::GetAngleDistrID(int termID, int & getLinkID, int & getDimID) const
{
	if (m_TermID2DistrID.size()==0)	// fill mapping
	{
		PosePriorGauss * ncThis = (PosePriorGauss*)this;

		assert(m_pIK!=NULL);
		int nparams =0;
		for (int i=0; i<(int)m_pIK->m_Links.size(); i++)
		{
			int ndims = SkeletonLink::NDims(m_pIK->m_Links[i]->m_JointType);
//			std::cout<<"Gauss["<<i<<"]: "<<m_AngleDistributions[i]<<" - "<<ndims<<std::endl;
			for (int d=0; d<ndims; d++)
			{
//				std::cout<<"    term: "<<((int)ncThis->m_TermID2DistrID.size())<<" -> "<<i<<std::endl;;
				ncThis->m_TermID2DistrID.push_back(i);
				ncThis->m_TermID2DimID.push_back(d);
			}
		}
	}
	assert(termID>=0 && termID < (int)m_TermID2DistrID.size());
	getLinkID = m_TermID2DistrID[termID];
	assert(getLinkID < (int)m_AngleDistributions.size());
	getDimID = m_TermID2DimID[termID];
//	std::cout<<"getDimID="<<getDimID<<" / "<<m_AngleDistributions[getLinkID]->m_iNDims<<std::endl;
	assert(getDimID < m_AngleDistributions[getLinkID]->m_iNDims);
//	std::cout<<"Retrieval: "<<termID<<" -> "<<getLinkID<<": "<<m_AngleDistributions[getLinkID]<<std::endl;
}

void PosePriorGauss::Train(const std::vector<const InteractorPose *> & poses)
{
	// gaussian distribution on angles
	assert(m_pIK!=NULL);
	double loccoord[3];
	
	assert(m_pIK->m_Links.size()<=m_pIK->m_ParamOffsetByLinkID.size());
	m_AngleDistributions.clear();
	for (int l=0; l<(int)m_pIK->m_Links.size(); l++)
	{
		int ndims = SkeletonLink::NDims(m_pIK->m_Links[l]->m_JointType);
		assert(ndims>=0 && ndims <= 3);
		if (ndims>0)
			m_AngleDistributions.push_back(new PosePointGauss(ndims));
		else
			m_AngleDistributions.push_back(NULL);
	}
	
	m_iNTotal = 0;
	for (int i=0; i<(int)poses.size(); i++)
	{
		assert(poses[i]->GetPoseType()==InteractorPose::TYPE_PntCorrsPose);
		const PntCorrsPose * pose = (const PntCorrsPose *)poses[i];
		
		// update gaussians
		for (int l=0; l<(int)m_pIK->m_Links.size(); l++)
		{
			if (m_AngleDistributions[l]!=NULL)
			{
				int offset = m_pIK->m_ParamOffsetByLinkID[l];
				int ndims = SkeletonLink::NDims(m_pIK->m_Links[l]->m_JointType);
				for (int d=0; d<ndims; d++)
					loccoord[d] = pose->GetParam(offset+d);
				m_AngleDistributions[l]->Update(loccoord);
			}
		}
		
		// update assigned / unassigned count
		for (int r=0; r<(int)pose->NMatches(); r++)
		{
			while (pose->Region(r)>=(int)m_Assigned.size())
				m_Assigned.push_back(0);
			m_Assigned[pose->Region(r)]++;
		}
		m_iNTotal++;
	}
	
	//	// TEST PROBABILITIES:
	//	std::cout<<"Testing probabilities of training data:"<<std::endl;
	//	//for (int i=0; i<(int)poses.size(); i++)
	//	int i=8;
	//	std::cout<<"Pose "<<i<<" / "<<poses.size()<<": "<<LogProb(poses[i])<<std::endl;
}

void PosePriorGauss::NonZeroJacobians(std::vector<int> & termIDs, std::vector<int> & paramIDs) const
{
	int nangles = m_pIK->NJointParams() - m_pIK->m_iNInitSkeletonParams;
	for (int i=0; i<nangles; i++)
	{
		termIDs.push_back(i);
		paramIDs.push_back(m_pIK->m_iNInitSkeletonParams + i);
	}
	assert(nangles==(int)termIDs.size());
}


void PosePriorGauss::WritePrior(const VKString & filename) const
{
	FileManagement::CreateDirsOnPathToFile(filename.c_str());
	FILE * file = fopen(filename.c_str(), "wb");
	assert(file!=NULL);
	
	int ndistr = (int)m_AngleDistributions.size();
	PosePriorType tp = PriorType();
	fwrite(&tp, sizeof(PosePriorType), 1, file);
	
	fwrite(&ndistr, sizeof(int), 1, file);
	for (int i=0; i<ndistr; i++)
	{
		bool included = (m_AngleDistributions[i]!=NULL);
		fwrite(&included, sizeof(bool), 1, file);
		if (included)
			m_AngleDistributions[i]->WriteDistribution(file);
	}
	
	int nassigned = (int)m_Assigned.size();
	fwrite(&nassigned, sizeof(int), 1, file);
	for (int i=0; i<nassigned; i++)
		fwrite(&(m_Assigned[i]), sizeof(int), 1, file);
	fwrite(&m_iNTotal, sizeof(int), 1, file);
	
	fclose(file);
}

void PosePriorGauss::ReadPrior(FILE * file)
{
	assert(file!=NULL);
	
	int ndistr;
	fread(&ndistr, sizeof(int), 1, file);
	bool included;
	for (int i=0; i<ndistr; i++)
	{
		fread(&included, sizeof(bool), 1, file);
		if (included)
			m_AngleDistributions.push_back(PosePointGauss::CreateDistribution(file));
		else
			m_AngleDistributions.push_back(NULL);
	}
	
	int nassigned = 0;
	fread(&nassigned, sizeof(int), 1, file);
	for (int i=0; i<nassigned; i++)
		m_Assigned.push_back(0);
	for (int i=0; i<nassigned; i++)
		fread(&(m_Assigned[i]), sizeof(int), 1, file);
	fread(&m_iNTotal, sizeof(int), 1, file);
}


void PosePriorGauss::SetParams(double maxPenalty, double variance)
{
	m_fMaxPenalty = maxPenalty;
	if (variance>=0)
	{
		for (int i=0; i<(int)m_AngleDistributions.size(); i++)
		{
			if (m_AngleDistributions[i]!=NULL)
			{
				for (int d=0; d<m_AngleDistributions[i]->m_iNDims; d++)
					m_AngleDistributions[i]->m_aVariances[d] = variance;
			}
		}
	}
}


