#include "PosePrior.h"
#include "InteractorPose.h"
#include "VkFunctions.h"
#include "IKSolver.h"
#include "BindingManager.h"
#include "PosePriorGauss.h"
#include "PosePriorMixGauss.h"
#include "PntCorrsPose.h"

////////
// PosePriorMixGauss
//
PosePriorMixGauss::PosePriorMixGauss()
{
	m_fMaxPenalty = RN_INFINITY;
	m_iTermTypes = PRIORTERM_Gaussians | PRIORTERM_Symmetry;
}

void PosePriorMixGauss::SetTerms(int terms)
{
	m_iTermTypes = terms;
}

PosePrior::PosePriorType PosePriorMixGauss::PriorType() const
{
	return PRIOR_MultiGauss;
}

void PosePriorMixGauss::InitDistributions()
{
	assert(m_pIK!=NULL);
	for (int i=0; i<(int)m_pIK->m_Links.size(); i++)
		m_LinkDistributions.push_back(new MultiGauss());
}

MultiGauss * PosePriorMixGauss::GetDistr(int linkid)
{
	assert(linkid>=0 && linkid < (int)m_LinkDistributions.size());
	return m_LinkDistributions[linkid];
}

void PosePriorMixGauss::AddGaussian(int linkid, PosePointGauss * gauss)
{
	assert(linkid>=0 && linkid < (int)m_LinkDistributions.size());
	m_LinkDistributions[linkid]->AddGaussian(gauss);
}

int PosePriorMixGauss::NTerms() const
{
	if (m_NonZeroLinkDistr.size()==0)
	{
		for (int i=0; i<(int)m_LinkDistributions.size(); i++)
			if (m_LinkDistributions[i]->NGaussians()>0)
				((PosePriorMixGauss*)this)->m_NonZeroLinkDistr.push_back(i);
	}
	return NSymmetryCouplings() + (int)m_NonZeroLinkDistr.size();
}

bool PosePriorMixGauss::IsSymmetryTerm(int termID) const
{
	int maxterm = NTerms() - NSymmetryCouplings();
	return termID >= maxterm;
}


double PosePriorMixGauss::EnergyTerm(const InteractorPose * poseIn, int termID, EnergyInfo & info) const
{
	int linkID = -1;
	// go over all parameters (links) -- find right energy term
	NTerms();	// initialize non-zero if necessary
	assert(m_NonZeroLinkDistr.size()>0);
	double retval = 0;
	
	const PntCorrsPose * pose = CastPose(poseIn);
	//double epsilon = 10e-10;	// avoid numerical instability for small variance
	double epsilon = 0;
	bool takemax = false;
	int poseid = pose->GetPoseID();
	// link energy
	if (!IsSymmetryTerm(termID))
	{
		double totalprob = 0;
		linkID = m_NonZeroLinkDistr[termID];
		// compute energy due to currLinkID
		
		assert(linkID>=0 && linkID<(int)m_pIK->m_ParamOffsetByLinkID.size());
		info.m_LinkID = linkID;
		info.m_SymmetricLink = -1;

		int offset = m_pIK->m_ParamOffsetByLinkID[linkID];
		int nfirst = 0;
		int nlast = m_LinkDistributions[linkID]->NGaussians();
		if (((m_pIK->m_SolverFlags & IKSolver::IK_ConditionedOnPoseID)!=0) && poseid>=0)
		{
			assert(poseid>=nfirst && poseid<nlast);
			nfirst = poseid;
			nlast = nfirst + 1;
		}
		
		for (int g=nfirst; g<nlast; g++)
		{
			PosePointGauss * gauss = m_LinkDistributions[linkID]->GetGaussian(g);
			double logprob = 0;
			for (int d=0; d<gauss->m_iNDims; d++)
			{
				double angle = pose->GetParam(offset + d);
				angle = PosePointGauss::GetNearestAngle(gauss->m_aMeans[d], angle);
				
				if (gauss->m_aVariances[d]<=epsilon)
				{
					if(angle != gauss->m_aMeans[d] && gauss->m_aVariances[d]>=0)
						logprob += m_fMaxPenalty;
					else
						logprob += 0;
				}
				else
					logprob += pow(gauss->m_aMeans[d] - angle, 2.)/gauss->m_aVariances[d];
			}
			if (takemax)
				totalprob = vkMax(totalprob, gauss->m_fMagnitude * exp(-logprob));
			else
				totalprob += gauss->m_fMagnitude * exp(-logprob);
		}
		if (totalprob>=1.)
			retval = 10E-35;
		else if (totalprob==0)
			retval = m_fMaxPenalty;
		else
			retval = -log(totalprob);
		//std::cout<<"Term"<<termID<<" = "<<retval<<std::endl;
	}
	else if (IsSymmetryTerm(termID))	// symmetry energy
	{
		int symmTermID  = (termID - (int)m_NonZeroLinkDistr.size());
		assert(symmTermID>=0 && symmTermID < (int)NSymmetryCouplings());
		assert(symmTermID < (int)m_LinkSymmetry.size());
		int nparams = NParamsForSymmJoint(symmTermID);
		linkID = m_SymmLinkLeft[symmTermID];
		
		info.m_LinkID = m_SymmLinkLeft[symmTermID];
		info.m_SymmetricLink = m_SymmLinkRight[symmTermID];

		
		PosePointGauss * gauss = m_LinkSymmetry[symmTermID];
		for (int d=0; d<nparams; d++)
		{
			double symmdiff = SignedDistToSymmetry(symmTermID, d, pose);
			if (gauss->m_aVariances[d]<=epsilon)	// avoid numerical instability for small variance
			{
				if(symmdiff != gauss->m_aMeans[d] && gauss->m_aVariances[d]>=0)
					retval += m_fMaxPenalty;
				else
					retval += 10E-35;
			}
			else
				retval += pow(gauss->m_aMeans[d] - symmdiff, 2.)/gauss->m_aVariances[d];
		}
		//std::cout<<"SymTerm"<<termID<<" = "<<retval<<std::endl;
	}
	else
		assert(false);

	if ((info.m_EnergyVis==ENVIS_AngleAll && !IsSymmetryTerm(termID))
		|| (info.m_EnergyVis==ENVIS_AngleSymmetry && IsSymmetryTerm(termID)))
	{
		assert(linkID>=0);
		int jointid = m_pIK->m_Links[linkID]->m_pChild->m_iJointID;
		assert(jointid < (int)info.m_PerPointEnergy.size());
		info.m_PerPointEnergy[jointid] += retval;
	}

	return retval;
}

void PosePriorMixGauss::NonZeroJacobians(std::vector<int> & termIDs, std::vector<int> & paramIDs) const
{
	// go over all parameters (links) -- find right energy term
	int termID = 0;
	for (termID=0; termID<(int)m_NonZeroLinkDistr.size(); termID++)
	{
		int linkID = m_NonZeroLinkDistr[termID];
		PosePointGauss * gauss = m_LinkDistributions[linkID]->GetGaussian(0);
		int offset = m_pIK->m_ParamOffsetByLinkID[linkID];
		
		for (int d=0; d<gauss->m_iNDims; d++)
		{
			paramIDs.push_back(offset+d);
			termIDs.push_back(termID);
		}
	}
	
	// symmetry
	for (int i=0; i<(int)m_SymmLinkLeft.size(); i++)
	{
		int left = m_SymmLinkLeft[i];
		int right = m_SymmLinkRight[i];

		int ndims = NParamsForSymmJoint(i);
		int offset = m_pIK->m_ParamOffsetByLinkID[left];
		for (int d=0; d<ndims; d++)
		{
			int pid = SymmParamID(i, d);
			paramIDs.push_back(offset+pid);
			termIDs.push_back(termID);
		}

		if (right>=0)
		{
			offset = m_pIK->m_ParamOffsetByLinkID[right];
			for (int d=0; d<ndims; d++)
			{
				int pid = SymmParamID(i, d);
				paramIDs.push_back(offset+pid);
				termIDs.push_back(termID);
			}
		}
		termID++;
	}
}

void PosePriorMixGauss::WritePrior(const VKString & filename) const
{
	FILE * f = fopen(filename.c_str(), "wb");
	assert(f!=NULL);
	PosePriorType tp = PriorType();
	fwrite(&tp, sizeof(PosePriorType), 1, f);

	// write distribution parameters
	int nlinks = (int)m_LinkDistributions.size();
	fwrite(&nlinks, sizeof(int), 1, f);	// write links
	for (int l=0; l<nlinks; l++)	// for each link
	{
		int ngaussians = m_LinkDistributions[l]->NGaussians();
		fwrite(&ngaussians, sizeof(int), 1, f);	// write gaussians
		for (int g=0; g < ngaussians; g++)	// for each gaussian distribution (in curr link)
			m_LinkDistributions[l]->GetGaussian(g)->WriteDistribution(f);
	}
	
	// write symmetry parameters
	int nsymmlinks = (int)m_LinkSymmetry.size();
	fwrite(&nsymmlinks, sizeof(int), 1, f);
	for (int l=0; l<nsymmlinks; l++)
		m_LinkSymmetry[l]->WriteDistribution(f);
	
	// write initial pose parameters
	int ninitposes = (int)m_GaussianPoseDNAs.size();
	fwrite(&ninitposes, sizeof(int), 1, f);	// write links
	for (int i=0; i<ninitposes; i++)
	{
		int ngaussians = (int)m_GaussianPoseDNAs[i].size();
		fwrite(&ngaussians, sizeof(int), 1, f);	// write gaussians
		for (int g=0; g<ngaussians; g++)
			fwrite(&(m_GaussianPoseDNAs[i][g]), sizeof(int), 1, f);	// write mean id
		fwrite(&(m_PerDNAPoseCounter[i]), sizeof(int), 1, f);	// write # examples
	}

	// write grid
	WriteGridData(f);
	
	fclose(f);
}

void PosePriorMixGauss::ReadPrior(FILE * f)
{
	assert(f!=NULL);
	int nlinks=0;
	fread(&nlinks, sizeof(int), 1, f);	// read links
	assert(m_LinkDistributions.size()==0);
	for (int l=0; l<nlinks; l++)	// for each link
	{
		m_LinkDistributions.push_back(new MultiGauss());
		int ngaussians=0;
		fread(&ngaussians, sizeof(int), 1, f);	// read gaussians
		for (int g=0; g < ngaussians; g++)	// for each gaussian read distribution (in curr link)
			m_LinkDistributions[l]->AddGaussian(PosePointGauss::CreateDistribution(f));
	}
	
	// read symmetry parameters
	int nsymmlinks;
	fread(&nsymmlinks, sizeof(int), 1, f);
	for (int l=0; l<nsymmlinks; l++)
		m_LinkSymmetry.push_back(PosePointGauss::CreateDistribution(f));
		
	// read initial pose parameters
	int ninitposes = (int)m_GaussianPoseDNAs.size();
	fread(&ninitposes, sizeof(int), 1, f);	// read links
	for (int i=0; i<ninitposes; i++)
	{
		m_GaussianPoseDNAs.push_back(std::vector<int>());
		m_PerDNAPoseCounter.push_back(0);
		int ngaussians = (int)m_GaussianPoseDNAs[i].size();
		fread(&ngaussians, sizeof(int), 1, f);	// read gaussians
		for (int g=0; g<ngaussians; g++)
		{
			m_GaussianPoseDNAs[i].push_back(0);
			fread(&(m_GaussianPoseDNAs[i][g]), sizeof(int), 1, f);	// read mean id
		}
		fread(&(m_PerDNAPoseCounter[i]), sizeof(int), 1, f);	// read # examples
	}
	
	ReadGridData(f);
}

int PosePriorMixGauss::NSuggestedInitPoses() const
{
	return (int)m_GaussianPoseDNAs.size();
}

void PosePriorMixGauss::InitPose(int poseID, InteractorPose * poseIn) const
{
	PntCorrsPose * pose = (PntCorrsPose*)CastPose(poseIn);
	assert(m_pIK!=NULL);
	pose->InitParams(m_pIK->NJointParams(), m_pIK->m_SolverFlags);
	pose->SetPoseID(poseID);
	
	assert(poseID < (int)m_GaussianPoseDNAs.size());
	assert(m_pIK->m_Links.size() == m_GaussianPoseDNAs[poseID].size());
	assert(m_pIK->m_Links.size() == m_LinkDistributions.size());
	
//	std::cout<<"Init pose: "<<poseID<<". DNA = [";
	for (int linkID=0; linkID<(int)m_GaussianPoseDNAs[poseID].size(); linkID++)
	{
		int offset = m_pIK->m_ParamOffsetByLinkID[linkID];
		int ndims = SkeletonLink::NDims(m_pIK->m_Links[linkID]->m_JointType);
		if (ndims>0)
		{
			int meanid = m_GaussianPoseDNAs[poseID][linkID];
//			std::cout<<meanid<<" "<<std::flush;
			const PosePointGauss * gaussian = m_LinkDistributions[linkID]->GetGaussian(meanid);
			for (int d=0; d<ndims; d++)
				pose->Param(offset + d) = gaussian->m_aMeans[d];
		}
	}
//	std::cout<<"]"<<std::endl;
}

void PosePriorMixGauss::RandomSample(InteractorPose * poseIn) const
{
	PntCorrsPose * pose = (PntCorrsPose*)CastPose(poseIn);
	assert(m_pIK!=NULL);
	pose->InitParams(m_pIK->NJointParams(), m_pIK->m_SolverFlags);
	
	int poseid = pose->GetPoseID();
	assert(poseid>=0);
	
	// set according to pose probability
	for (int linkID=0; linkID < (int)m_pIK->m_Links.size(); linkID++)
	{
		int offset = m_pIK->m_ParamOffsetByLinkID[linkID];
		int ndims = SkeletonLink::NDims(m_pIK->m_Links[linkID]->m_JointType);
		if (ndims>0)
		{
			const PosePointGauss * gaussian = m_LinkDistributions[linkID]->GetGaussian(poseid);
			for (int d=0; d<ndims; d++)
			{
				double u1 = (double)rand() / (double)RAND_MAX;
				double u2 = (double)rand() / (double)RAND_MAX;
				pose->Param(offset + d) = gaussian->Sample(d, u1, u2);
			}
		}
	}
	
	// randomly select and fill according to symmetry probability
	
}


void PosePriorMixGauss::UpdatePoseDNAs(std::vector<int> & dnaStr)
{
	// go over links: find best matching gaussians
	VKString jointStr = "";
	for (int i=0; i<(int)dnaStr.size(); i++)
		jointStr += VKString::number(dnaStr[i])+"_";
	
	std::map<VKString, int>::iterator iter = m_Str2DNA.find(jointStr);
	if (iter==m_Str2DNA.end())
	{
		int dnaID = (int)m_GaussianPoseDNAs.size();
		m_GaussianPoseDNAs.push_back(dnaStr);
		m_PerDNAPoseCounter.push_back(1);
		m_Str2DNA[jointStr] = dnaID;
	}
	else
		m_PerDNAPoseCounter[iter->second]++;
}

void PosePriorMixGauss::SortDNAs()
{
	std::vector<Sortable> srtDNAs;

	std::vector<std::vector<int> > newDNAs;
	std::vector<int> newCounters;

	for (int i=0; i<(int)m_PerDNAPoseCounter.size(); i++)
		srtDNAs.push_back(Sortable(m_PerDNAPoseCounter[i], NULL, i));
	std::sort(srtDNAs.begin(), srtDNAs.end());

	m_Str2DNA.clear();
	for (int i=0; i<srtDNAs.size(); i++)
	{
		newDNAs.push_back(m_GaussianPoseDNAs[srtDNAs[i].id]);
		newCounters.push_back(m_PerDNAPoseCounter[srtDNAs[i].id]);
		
		VKString str = "";
		for (int j=0; j<(int)m_GaussianPoseDNAs[srtDNAs[i].id].size(); j++)
			str += VKString::number(m_GaussianPoseDNAs[srtDNAs[i].id][j])+"_";
		m_Str2DNA[str] = i;
	}
	
	m_GaussianPoseDNAs = newDNAs;
	m_PerDNAPoseCounter = newCounters;
}

void PosePriorMixGauss::UpdateSymmetry(const InteractorPose * trainingPoseIn, bool noToesPalms)
{
	if (m_SymmLinkLeft.size()==0)
		InitSymmetryCoupling();

	const PntCorrsPose * pose = CastPose(trainingPoseIn);
	
	if (m_LinkSymmetry.size()==0)
	{
		for (int i=0; i<NSymmetryCouplings(); i++)
			m_LinkSymmetry.push_back(new PosePointGauss(NParamsForSymmJoint(i), true));
	}

	int maxndims = 3;
	double * diffs = new double[maxndims];
	for (int i=0; i<NSymmetryCouplings(); i++)
	{
		assert(NParamsForSymmJoint(i) <= maxndims);
		
		VKString childName = m_pIK->m_Links[m_SymmLinkLeft[i]]->m_pChild->m_sJointName;
		bool palmOrToe = (childName == "right_toe") || (childName == "left_toe");
		palmOrToe = palmOrToe || (childName == "right_palm") || (childName=="left_palm");

		if (noToesPalms && palmOrToe)
		{
			for (int d=0; d<NParamsForSymmJoint(i); d++)
				m_LinkSymmetry[i]->m_aVariances[d] = -1;
		}
		else
		{
			for (int d=0; d<NParamsForSymmJoint(i); d++)
				diffs[d] = SignedDistToSymmetry(i, d, pose);
			m_LinkSymmetry[i]->UpdateVarianceFromZero(diffs);
		}
	}
	delete [] diffs;
	
}

int PosePriorMixGauss::NSymmetryCouplings() const
{
	if (m_SymmLinkLeft.size()==0)
		((PosePriorMixGauss*)this)->InitSymmetryCoupling();
	return (int)m_PerParamRelationships.size();
}

int PosePriorMixGauss::NParamsForSymmJoint(int symmID) const
{
	if (m_SymmLinkLeft.size()==0)
		((PosePriorMixGauss*)this)->InitSymmetryCoupling();

	assert(symmID < (int)m_SymmLinkLeft.size() && m_SymmLinkLeft.size()==m_SymmLinkRight.size());
	int linkID = m_SymmLinkLeft[symmID];
	
	if (m_SymmLinkRight[symmID]==-2)
		return 2;
	assert(linkID>=0 && linkID < (int)m_pIK->m_Links.size());
	return SkeletonLink::NDims(m_pIK->m_Links[linkID]->m_JointType);
}

int PosePriorMixGauss::SymmParamID(int symmID, int paramID) const
{
	if (m_SymmLinkRight[symmID]==-2)
	{
		if (paramID==0)
			return 0;
		else if (paramID==1)
			return 2;
		else
			assert(false);
	}
	return paramID;
}

double PosePriorMixGauss::SignedDistToSymmetry(int symmID, int paramID, const InteractorPose * poseIn) const
{
	paramID = SymmParamID(symmID, paramID);
	const PntCorrsPose * pose = CastPose(poseIn);
	if (m_SymmLinkLeft.size()==0)
		((PosePriorMixGauss*)this)->InitSymmetryCoupling();
	
	int left = m_SymmLinkLeft[symmID];
	int right = m_SymmLinkRight[symmID];
	double diff = 0;
	assert(left>=0);
	int oleft = m_pIK->m_ParamOffsetByLinkID[left];
	
	switch(m_PerParamRelationships[symmID][paramID])
	{
		case SYM_None:
			diff = 0;
			break;
		case SYM_Equal:
		case SYM_DiffSign:
		{
			assert(right>=0);
			int oright = m_pIK->m_ParamOffsetByLinkID[right];
			double aright = pose->GetParam(oright+paramID);
			double aleft = pose->GetParam(oleft+paramID);
			if (m_PerParamRelationships[symmID][paramID] == SYM_DiffSign)
				aright = PosePointGauss::GetNearestAngle(aleft, -aright);
			else
				aright = PosePointGauss::GetNearestAngle(aleft, aright);
			diff = aright - pose->GetParam(oleft+paramID);
			break;
		}
		case SYM_Zero:
			assert(right==-2);
			assert(paramID==0 || paramID==2);
			diff = -PosePointGauss::GetNearestAngle(0, pose->GetParam(oleft));
			break;
		default:
			assert(false);
	}
	return diff;
}


void PosePriorMixGauss::InitSymmetryCoupling()
{
	m_SymmLinkLeft.clear();	// left (or center parameters)
	m_SymmLinkRight.clear();	// right
	m_PerParamRelationships.clear();

	// left-right symmetry
	std::map<VKString, int> name2coupleid;
	for (int j=0; j<(int)m_pIK->m_Links.size(); j++)
	{
		int ndims = SkeletonLink::NDims(m_pIK->m_Links[j]->m_JointType);
		if (ndims==0)
			continue;
		
		VKString childName = m_pIK->m_Links[j]->m_pChild->m_sJointName;
		int symmgroup = -1;
		if (childName.startsWith("left"))
			symmgroup = 0;
		else if (childName.startsWith("right"))
			symmgroup = 1;
		else
			symmgroup = 2;
		
		VKString nameonly = childName.replace("right_", "").replace("left_", "");
		std::map<VKString, int>::iterator iter = name2coupleid.find(nameonly);
		
		int symmID = -1;
		if (iter==name2coupleid.end())	// add new symmetry group
		{
			symmID = (int)m_SymmLinkLeft.size();
			m_SymmLinkLeft.push_back(-1);
			m_SymmLinkRight.push_back(-1);
			m_PerParamRelationships.push_back(std::vector<SymmType>());
			for (int d=0; d<ndims; d++)
				m_PerParamRelationships[symmID].push_back(SYM_None);
			name2coupleid[nameonly] = symmID;
			
			if (symmgroup==0 || symmgroup==1)	// left angle ==right angle
			{
				for (int d=0; d<ndims; d++)
				{
					SymmType tp = SYM_Equal;
					if (d!=1 && nameonly!="leg2" && nameonly != "toe")	// vk: hardcoded - read from file
						tp = SYM_DiffSign;
					m_PerParamRelationships[symmID][d] = tp;
				}
			}
			else	// center angle is zero
			{
				assert(symmgroup==2);
				m_SymmLinkRight[symmID] = -2;
				m_PerParamRelationships[symmID][0] = SYM_Zero;	// NOT SURE
				m_PerParamRelationships[symmID][2] = SYM_Zero;	// NOT SURE
			}
		}
		else	// add node to symmetry group
			symmID = iter->second;
		
		if (symmgroup==0)		// left
			m_SymmLinkLeft[symmID] = j;
		else if (symmgroup==1)	// right
			m_SymmLinkRight[symmID] = j;
		else	// center
			m_SymmLinkLeft[symmID] = j;
	}
}

void PosePriorMixGauss::SymmCoupling(int symmID, int * left, int * right) const
{
	assert(symmID>=0 && symmID < (int)m_SymmLinkLeft.size());
	*left = m_SymmLinkLeft[symmID];
	*right = m_SymmLinkRight[symmID];
}

VKString PosePriorMixGauss::GetStr() const
{
	VKString str="";
	str += "Mixture of Gaussians:\n";
	str += "==> PosePrior: \n";
	for (int i=0; i<(int)m_pIK->m_Links.size(); i++)
	{
		MultiGauss * multigauss = ((PosePriorMixGauss*)this)->GetDistr(i);
		if (multigauss->NGaussians()>0)
		{
			SkeletonLink * link = m_pIK->m_Links[i];
			VKString descrStr = link->m_pParent->m_sJointName+" -> "+link->m_pChild->m_sJointName;
			descrStr = descrStr.CreateStrWithPadding(35, " ");
			str += VKString("    ")+descrStr+"\n";
			for (int g=0; g<multigauss->NGaussians(); g++)
			{
				const PosePointGauss * gauss = multigauss->GetGaussian(g);
				str += VKString("         Mean")+g+" = [";
				for (int d=0; d<gauss->m_iNDims; d++)
					str += VKString::number(gauss->m_aMeans[d]).CreateStrWithPadding(14, " ");
				str += VKString("]\n");
				str += VKString("         Vars")+g+" = [";
				for (int d=0; d<gauss->m_iNDims; d++)
					str += VKString::number(gauss->m_aVariances[d]).CreateStrWithPadding(14, " ");
				str += VKString("]\n");
			}
		}
	}
	
	str+="==> SymmetryPrior: \n";
	for (int i=0; i<NSymmetryCouplings(); i++)
	{
		SkeletonLink * link = m_pIK->m_Links[m_SymmLinkLeft[i]];
		VKString descrStr = link->m_pParent->m_sJointName+" -> "+link->m_pChild->m_sJointName;
		descrStr = descrStr.CreateStrWithPadding(35, " ");
		str += VKString("    ")+descrStr+"\n";

		const PosePointGauss * gauss = m_LinkSymmetry[i];
		str += VKString("         Mean = [");
		for (int d=0; d<NParamsForSymmJoint(i); d++)
			str += VKString::number(gauss->m_aMeans[d]).CreateStrWithPadding(14, " ");
		str += VKString("]\n");
		str += VKString("         Vars = [");
		for (int d=0; d<NParamsForSymmJoint(i); d++)
			str += VKString::number(gauss->m_aVariances[d]).CreateStrWithPadding(14, " ");
		str += VKString("]\n");
	}
	str += "\n";
	return str;
}

////////
// MultiGauss
//
MultiGauss::MultiGauss()
{
	
}

void MultiGauss::AddGaussian(PosePointGauss * gauss)
{
	m_Gaussians.push_back(gauss);
}

int MultiGauss::NGaussians() const
{
	return (int)m_Gaussians.size();
}

PosePointGauss * MultiGauss::GetGaussian(int g)
{
	assert(g>=0 && g < (int)m_Gaussians.size());
	return m_Gaussians[g];
}

const PosePointGauss * MultiGauss::GetGaussian(int g) const
{
	assert(g>=0 && g < (int)m_Gaussians.size());
	return m_Gaussians[g];
}



