/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "VkFunctions.h"
#include "Sortable.h"
#include "TimeProfiler.h"

#include "ShapeLocalPCA.h"
#include "ShapeDistance.h"


ShapeLocalPCA::ShapeLocalPCA(GeomShape * shape, const VKString & valueName, double radius, int kNN, bool normRad,
							 const std::vector<int> & subset)
: ShapeValue(shape, valueName, 7)
	// 4 values for axes (normalizes): axis1 (2 values), axis2 (2 values), axis3 = axis1 x axis2
	// 3 values for variance
{
	m_fRadius = radius;
	m_ikNN = kNN;
	
	m_bNormRad = normRad;
	m_fRadScale = -1;
	
	m_Subset = subset;
	assert(m_pPnts!=NULL);
	
	if (m_Subset.size()==0)
		m_pSubsetKdTree = m_pPnts->GetKdTree();
	else
		m_pSubsetKdTree = m_pPnts->CreateSubsetKdTree(m_Subset);
	m_pDensePoints = m_pPnts;
}

ShapeLocalPCA::ShapeLocalPCA(ShapePointCloud * pnts, const VKString & valueName, double radius,
							 ShapePointCloud * densepnts, ShapeDistance * metric)
: ShapeValue(pnts, valueName, 7)
{
	m_fRadius = radius;
	m_ikNN = -1;
	m_bNormRad = false;
	m_fRadScale = 1;
	assert(m_pPnts!=NULL);
	m_pMetric = metric;
	m_pDensePoints = densepnts;
}

ShapeLocalPCA::~ShapeLocalPCA()
{
	
}

VKString ShapeLocalPCA::GetClassName() const
{
	return "ShapeLocalPCA";
}

void ShapeLocalPCA::GetNhd(const R3Point & pos, const R3Kdtree<const PointInSet *> * kdTree,
						   double rad, int kNN, std::vector<int> & ids)
{
	int niter = 1000;
	RNArray<const PointInSet *> nhd;
	kdTree->FindAll(pos, 0, rad, nhd);
	if (kNN==-1)
	{
		for (int j=0; j<nhd.NEntries(); j++)
			ids.push_back(nhd[j]->id);
	}
	else
	{
		assert(rad>0);
		while (kNN > nhd.NEntries() && niter-->=0)
		{
			nhd.Empty();
			rad*=2.;
			kdTree->FindAll(pos, 0, rad, nhd);
		}
		std::vector<Sortable> srtNhd;
		for (int j=0; j<nhd.NEntries(); j++)
			srtNhd.push_back(Sortable(R3Distance(pos, nhd[j]->GetPosition()), NULL, nhd[j]->id));
		std::sort(srtNhd.begin(), srtNhd.end());
		for (int j=0; j<vkMin(kNN, (int)srtNhd.size()); j++)
			ids.push_back(srtNhd[j].id);
	}		
}

void ShapeLocalPCA::ComputePntsPCA(bool verbouse)
{
	assert(m_pPnts!=NULL);

	double variance[3];
	R3Triad axes;

	double * allVals = new double[7*m_pPnts->NumPoints()];
	TimeProfiler timer;
	int ntoofewpnts = 0;
	for (int i=0; i<m_pPnts->NumPoints(); i++)
	{
		if (verbouse)
			timer.WriteProgress("prog", i, m_pPnts->NumPoints());
		std::vector<int> ids;
		if (m_pMetric==NULL)
		{
			double rad = m_fRadius;
			if (m_bNormRad)
			{
				const R3Point & ctr = m_pPnts->Centroid();
				rad = m_pPnts->AverageRadius(ctr);
				assert(rad>0);
				
				if (m_fRadScale==-1)
				{
					if (m_bNormRad)
						m_fRadScale = rad;
					else
						m_fRadScale = 1.;
				}
				rad *= m_fRadius;
			}
			assert(m_pSubsetKdTree!=NULL);
			GetNhd(m_pPnts->PointPosition(i), m_pSubsetKdTree, rad, m_ikNN, ids);
		}
		else
		{
			std::vector<ShapePoint> neighborhood;
			//std::vector<double> normDists	-- distances
			m_pMetric->NearestNeighborhood(m_pPnts->GetPoint(i), m_fRadius, &neighborhood);
			for (int nn=0; nn<(int)neighborhood.size(); nn++)
				ids.push_back(neighborhood[nn].PointID());
		}
		
		
		bool enoughpnts = (int)ids.size()>=vkMax(6,m_ikNN);
		if (!enoughpnts)	// too few points for meaningful PCA analysis
		{
			ntoofewpnts++;
			for (int d=0; d<7; d++)
				allVals[i*7 + d] = 1;
			//assert(enoughpnts);
		}
		else	// sufficient samples
		{
			axes = m_pDensePoints->PCAFromPoints(ids, variance);
			axes.Normalize();
		
			allVals[i*7 + 0] = axes[0][0];
			allVals[i*7 + 1] = axes[0][1];
			allVals[i*7 + 2] = axes[1][0];
			allVals[i*7 + 3] = axes[1][1];
			allVals[i*7 + 4] = sqrt(variance[0]);
			allVals[i*7 + 5] = sqrt(variance[1]);
			allVals[i*7 + 6] = sqrt(variance[2]);		
		
			double epsilon = 10E-10;
			bool validTriad = vkAbs(axes[0].Length()-1.) < epsilon && vkAbs(axes[1].Length()-1.) < epsilon ;
			validTriad = validTriad && (vkAbs(axes[2].Length()-1.) < epsilon);
			if (!validTriad)
			{
				std::cout<<"[ERROR] Invalid triad: "<<std::endl;
				std::cout<<"        a1=["<<axes[0].X()<<" "<<axes[0].Y()<<" "<<axes[0].Z()<<"] "<<std::endl;
				std::cout<<"        a2=["<<axes[1].X()<<" "<<axes[1].Y()<<" "<<axes[1].Z()<<"] "<<std::endl;
				std::cout<<"        a3=["<<axes[2].X()<<" "<<axes[2].Y()<<" "<<axes[2].Z()<<"] "<<std::endl;
				assert(validTriad);
			}
		}
	}
	if (verbouse)
		std::cout<<std::endl;
	InitCache(ShapeValue::SVTYPE_PerPointSet);
	SetPntsValues(allVals, 7*m_pPnts->NumPoints());
	
	if (ntoofewpnts>0 && verbouse)
		std::cout<<"[WARNING] Insufficient neighborhoods: "<<ntoofewpnts<<" / "<<m_pPnts->NumPoints()<<std::endl;
}

R3Triad ShapeLocalPCA::GetPCA(const ShapePoint & pnt, double * stdDev) const
{
	static double pcaVals[7];
	Values(pnt, pcaVals);
	
	if (pcaVals[4] == -1)
	{
		stdDev[0] = -1;
		return R3Triad();
	}
	assert(m_fRadScale!=-1);
	
	R3Vector a1(pcaVals[0], pcaVals[1], sqrt(vkMax(0, 1. - pow(pcaVals[0], 2.) - pow(pcaVals[1], 2.))));
	R3Vector a2(pcaVals[2], pcaVals[3], sqrt(vkMax(0, 1. - pow(pcaVals[2], 2.) - pow(pcaVals[3], 2.))));
	R3Vector a3 = a1 % a2;
	a3.Normalize();
	
	stdDev[0] = pcaVals[4] / m_fRadScale;
	stdDev[1] = pcaVals[5] / m_fRadScale;
	stdDev[2] = pcaVals[6] / m_fRadScale;
	return R3Triad(a1, a2, a3);
}

double ShapeLocalPCA::GetPntsRadius() const
{
	return m_fRadScale;
}

int ShapeLocalPCA::GetKNN() const
{
	return m_ikNN;
}

bool ShapeLocalPCA::IsNormalized() const
{
	return m_bNormRad;
}

double ShapeLocalPCA::GetRadiusScale() const
{
	return m_fRadius;
}


