/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "VisWindow.h"
#include "VisButtonsMenu.h"

/////////// VisButton ///////////
VisButton::VisButton(const VKString & buttonName, 
					 int initState, int nStates)
{
	m_sButtonName = buttonName;
	m_iButtonState = initState;
	m_iMaximalStateValue = nStates;
}

int VisButton::GetState() const
{
	return m_iButtonState;
}

void VisButton::ChangeState(int change)
{
	m_iButtonState += change;
	if (m_iButtonState >= m_iMaximalStateValue)
		m_iButtonState = 0;
	else if (m_iButtonState < 0)
		m_iButtonState = m_iMaximalStateValue-1;
}

/////////// VisButtonsGroup ///////////
VisButtonsGroup::VisButtonsGroup(const VKString & groupname,
								 ButtonsGroupType groupType)
{
	m_sGroupName = groupname;
	m_GroupType = groupType;
	m_iFirstButtonPage = 0;
}

VisButtonsGroup::~VisButtonsGroup()
{
	FreeButtons();
}
void VisButtonsGroup::FreeButtons()
{
	for (int i=0; i<(int)m_Buttons.size(); i++)
		delete m_Buttons[i];
	m_Buttons.clear();
}

void VisButtonsGroup::AddOnOffButtons(const VKStringList & buttonNames)
{
	for (int i=0; i<buttonNames.count(); i++)
		m_Buttons.push_back(new VisButton(((VKStringList&)buttonNames)[i], 0, 2));
}

void VisButtonsGroup::AddButton(const VKString & buttonName, int nStates)
{
	m_Buttons.push_back(new VisButton(buttonName, 0, nStates));
}

////////// VisButtonsMenu ///////////
VisButtonsMenu::VisButtonsMenu()
{
	m_iRenderTextForGroup = -1;
	m_iMaxButtons = 20;
	
	m_fTextSpacing = 0.07;
	m_fTextXOffset = 1.2;
	m_fTextYOffset = 0.1;
}

VisButtonsMenu::~VisButtonsMenu()
{
	
}

void VisButtonsMenu::AddButtonGroup(VisButtonsGroup * group)
{
	m_ButtonGroups.push_back(group);
}

void VisButtonsMenu::SetRenderingParams(double textSpacing, 
										double textXOffset, 
										double textYOffset,
										int nMaxButtons)
{
	if (textSpacing!=-1)
		m_fTextSpacing = textSpacing;
	if (textXOffset!=-1)
		m_fTextXOffset = textXOffset;
	if (textYOffset!=-1)
		m_fTextYOffset = textYOffset;
	if (nMaxButtons!=-1)
		m_iMaxButtons = nMaxButtons;
}

void VisButtonsMenu::FillRenderablePos()
{
	int nRadioButtons = (int)m_ButtonGroups.size();
	
	double wMinX = 0.;
	double wMaxX = 1.;
	double wMinY = 0.;
	double wMaxY = 1.;
	double height = wMaxY - wMinY;
	double width = wMaxX - wMinX;
	
	double yMin = wMinY + m_fTextSpacing;
	double yMax = wMaxY - m_fTextSpacing;
	
	double bHeight = yMax-yMin;
	double bWidth = width / (double)nRadioButtons;
	
	double bSizeW = bWidth * .7;
	double bSizeH = bHeight / (double)m_iMaxButtons;
	
	for (int i=0; i<nRadioButtons; i++)
	{
		// draw radio button names 
		double xGroupName = wMinX + bWidth * i + bWidth/4.;
		double yGroupName = yMin;
		
		VisButtonsGroup * gp = GetGroup(i);
		gp->m_fHeaderMinX = xGroupName; 
		gp->m_fHeaderWidth = bWidth;
		gp->m_fHeaderMinY = yGroupName - .5 * bSizeH;
		gp->m_fHeaderHeight = m_fTextSpacing;
		
		gp->m_fFooterMinX = xGroupName - bWidth *.2;
		gp->m_fFooterWidth = bWidth;
		gp->m_fFooterMinY = yMax + m_fTextSpacing; 
		gp->m_fFooterHeight = m_fTextSpacing;
		
		std::vector<VisButton*> & btns = gp->m_Buttons;
		int start = gp->m_iFirstButtonPage;
		for (int b=start; b < vkMin((int)btns.size(), start + m_iMaxButtons); b++)
		{
			double bx = xGroupName - bWidth/8.;
			double by = ((double)(b - start)) / (double)m_iMaxButtons;
			by = yMin + m_fTextSpacing*.5 + by * bHeight;
			btns[b]->m_fMinX = bx;
			btns[b]->m_fMinY = by;
			btns[b]->m_fWidth = bSizeW;
			btns[b]->m_fHeight = bSizeH;			
		}
	}
}

void VisButtonsMenu::Draw(VisWindow * window)
{	
	glLineWidth(1.);
	FillRenderablePos();
	VisWindowRenderer::TextSize fontS = VisWindowRenderer::TEXT_SIZE_SMALL;
	for (int i=0; i<(int)m_ButtonGroups.size(); i++)
	{
		// draw group names 
		VisButtonsGroup * gp = GetGroup(i);
		R3Point pHeader = R3Point(gp->m_fHeaderMinX, 1. - gp->m_fHeaderMinY, 0);
		window->GetRenderer()->RenderText(pHeader, gp->m_sGroupName, 1., 0., 0., fontS);
		
		// highlight group name is selected
		if (m_iRenderTextForGroup==i)
		{
			glColor3d(1., 0., 0.);
			glBegin(GL_LINES);
			double ypos = pHeader.Y() - .2*gp->m_fHeaderHeight;
			glVertex2d(pHeader.X(), ypos);
			glVertex2d(pHeader.X() + gp->m_fHeaderWidth / 2., ypos);
			glEnd();
		}
		
		// draw buttons 
		std::vector<VisButton*> & btns = gp->m_Buttons;
		int start = gp->m_iFirstButtonPage;
		for (int b=start; b < vkMin((int)btns.size(), start + m_iMaxButtons); b++)
		{			
			double bx = btns[b]->m_fMinX;
			double by = 1. - btns[b]->m_fMinY;
			double bw = btns[b]->m_fWidth;
			double bh = btns[b]->m_fHeight;

			bool sel = (btns[b]->m_iButtonState>0);
			if (btns[b]->m_iMaximalStateValue == 2)	// on/off
				window->DrawSquare(bx, by, bw, bh, .2, .2, .5, 1., .2, .2, .5, sel ? 1. : 0);
			else
			{
				
				if (sel)
					window->DrawArrows(bx, by, bw, bh, .2, .2, .5, 1., .4, .4, .5, 1., true);
				else
					window->DrawArrows(bx, by, bw, bh, .2, .2, 1., 1., .4, .4, 1., 1., true);
			}
			
			// draw button names (if selected)
			if (m_iRenderTextForGroup==i)
			{
				VKString bName = btns[b]->m_sButtonName;
				if (btns[b]->m_iMaximalStateValue > 2)
				{
					int cS = btns[b]->m_iButtonState;
					int mS = btns[b]->m_iMaximalStateValue;
					bName += VKString(" : ")+VKString::number(cS) + " / " + VKString::number(mS);
				}
				
				double nbx = m_fTextXOffset;
				//double nby = m_fTextYOffset+(1.-(1.-by)*1.3);
				double nby =  1. - (m_fTextYOffset+(b-start)*m_fTextSpacing*1.3);
				double textClr[3] = {.2, .2, .5};
				if (!sel)
				{	textClr[0] = .5;	textClr[1] = .5;	textClr[2] = .8;	}
				window->GetRenderer()->RenderText(R3Point(nbx, nby, 0), 
												  bName, textClr[0], textClr[1], textClr[2], fontS);
				glColor3d(1., .2, .5);
				glBegin(GL_LINES);
				glVertex2d(bx+.5*bw, by+.5*bh);
				glVertex2d(nbx - .2*bw, nby + .5*bh);
				glEnd();
			}
		}
		
		
		// draw page scrolling
		if ((int)btns.size() >= m_iMaxButtons)
		{
			double bh = btns[0]->m_fHeight;
			double centerY = (gp->m_fFooterMinY-1.5*bh);
			double topY = centerY + bh*.4;
			window->DrawArrows(gp->m_fFooterMinX, 1.-topY, 
							   btns[0]->m_fWidth, bh*.4 * 2,
							   .5, .2, .2, 1., .2, .5, .2, 1., true);
			
			R3Point pFooter = R3Point(gp->m_fFooterMinX, 1. - gp->m_fFooterMinY, 0);
			VKString text = VKString::number(start);
			window->GetRenderer()->RenderText(pFooter, text, 1., 0., 0., fontS);
		}
	}	
}

bool VisButtonsMenu::ClickInWindow(double x, double y, 
								   int & groupID, int & buttonID, int & dir)
{
	FillRenderablePos();
	
	bool buttonClicked = false;
	
	for (int i=0; i<(int)m_ButtonGroups.size() && !buttonClicked; i++)
	{
		VisButtonsGroup * gp = GetGroup(i);

		// see if header clicked
		double hx = gp->m_fHeaderMinX;
		double hw = gp->m_fHeaderWidth;
		double hy = 1. - gp->m_fHeaderMinY;
		double hh = gp->m_fHeaderHeight;
		if (x >= hx && x <= (hx + hw) && y>= hy && y <= (hy + hh))
		{
			if (m_iRenderTextForGroup!=i)
				m_iRenderTextForGroup = i;
			else
				m_iRenderTextForGroup = -1;
		}
		
		std::vector<VisButton*> & btns = gp->m_Buttons;
		int & start = gp->m_iFirstButtonPage;
		
		// see if button clicked
		for (int b=start; b < vkMin((int)btns.size(), start + m_iMaxButtons); b++)
		{			
			double bx = btns[b]->m_fMinX;
			double by = 1. - btns[b]->m_fMinY;
			double bw = btns[b]->m_fWidth;
			double bh = btns[b]->m_fHeight;
			
			if (x >= bx && x <= (bx + bw) && y>= by && y <= (by + bh))
			{
				groupID = i;
				buttonID = b;
				buttonClicked = true;
				
				dir = (x < (bx + .5*bw)) ? -1 : 1;
				break;
			}
		}
		
		// see if footer clicked
		double fx = gp->m_fFooterMinX;
		double fw = gp->m_fFooterWidth;
		double fy = 1. - gp->m_fFooterMinY;
		double fh = gp->m_fFooterHeight;
		if (x >= fx && x <= (fx + fw) && y>= fy && y <= (fy + fh))
		{
			if ((x - fx)/fw < .5)
				ChangePage(i, -1);
			else
				ChangePage(i, 1);
		}
	}
	return buttonClicked;
}

void VisButtonsMenu::ChangePage(int groupID, int advance)
{
	VisButtonsGroup * gp = GetGroup(groupID);
	int oldVal = gp->m_iFirstButtonPage;
	gp->m_iFirstButtonPage += advance * m_iMaxButtons;
	
	if (gp->m_iFirstButtonPage < 0)
	{
		gp->m_iFirstButtonPage = (int)gp->m_Buttons.size() - m_iMaxButtons;
		if (gp->m_iFirstButtonPage < 0 || oldVal > 0)
			gp->m_iFirstButtonPage = 0;
	}
	else if (gp->m_iFirstButtonPage>=(int)gp->m_Buttons.size())
		gp->m_iFirstButtonPage = 0;
}

void VisButtonsMenu::ClearGroup(int groupID)
{
	VisButtonsGroup * gp = GetGroup(groupID);
	gp->m_iFirstButtonPage = 0;
	gp->FreeButtons();
}

void VisButtonsMenu::ChangeState(int groupID, int buttonID, int change)
{
	if (GetGroup(groupID)->m_GroupType==VisButtonsGroup::BUTGROUP_RadioButtons)
	{
		std::vector<VisButton*> & btns = GetGroup(groupID)->m_Buttons;
		for (int i=0; i<(int)btns.size(); i++)
			if (i!=buttonID)
				btns[i]->m_iButtonState = 0;
	}

	GetButton(groupID, buttonID)->ChangeState(change);
}

const VKString & VisButtonsMenu::ButtonName(int groupID, int buttonID)
{
	const VKString & val = GetButton(groupID, buttonID)->m_sButtonName;
	return val;
}


int VisButtonsMenu::State(int groupID, int buttonID)
{
	int val = GetButton(groupID, buttonID)->m_iButtonState;
	return val;
}

VisButtonsGroup * VisButtonsMenu::GetGroup(int groupID)
{
	assert(groupID >=0 && groupID < (int)m_ButtonGroups.size());
	return m_ButtonGroups[groupID];
}

VisButton * VisButtonsMenu::GetButton(int groupID, int buttonID)
{
	
	VisButtonsGroup * group = GetGroup(groupID);
	int nMaxButtons = (int)group->m_Buttons.size();
	if (buttonID < 0 || buttonID>=nMaxButtons)
	{
		std::cout<<"[ERROR] Invalid button id: "<<buttonID<<" / "<<nMaxButtons;
		std::cout<<": gp="<<group->m_sGroupName.c_str()<<std::endl;
	}
	assert(buttonID>=0 && buttonID < nMaxButtons);
	return group->m_Buttons[buttonID];
}

