/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "VisModel3D.h"
#include "VisWindow.h"
#include "DatabaseVisWindow.h"
#include "ShapeValue.h"
#include "ShapeMaterials.h"
#include "ShapeImg.h"
#include "ShapeGrid.h"

// NOTE: not fully implemented
//#define USE_SHADERS

#ifdef USE_SHADERS

// static draw functions
void VisModel3D::DrawMeshWithSampledValues(const ShapeMesh * mesh, 
										   VisWindow * window, const R3Affine & camera, 
										   ShapePointCloud * pnts, double * values, 
										   double minValue, double maxValue)
{	
	static bool initialized = false;
	static GLuint shaderProgram;
	if (!initialized)
	{
		initialized = true;
		GLuint fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
		GLuint vertexShader = glCreateShader(GL_VERTEX_SHADER);
		
		std::ifstream fileFragment("shaders/values_fragment.shader");
		std::ifstream fileVertex("shaders/default_vertex.shader");
		assert(fileFragment.is_open() && fileVertex.is_open());
		
		VKString fragmentShaderCode(fileFragment); 
		VKString vertexShaderCode(fileVertex);		
		const char * fragmentShaderData = fragmentShaderCode.c_str();
		const char * vertexShaderData = vertexShaderCode.c_str();
		glShaderSource(fragmentShader, 1, &fragmentShaderData, NULL);
		glShaderSource(vertexShader, 1, &vertexShaderData, NULL);
		glCompileShader(fragmentShader);
		glCompileShader(vertexShader);
		shaderProgram = glCreateProgram();
		glAttachShader(shaderProgram, fragmentShader);
		glAttachShader(shaderProgram, vertexShader);
		glLinkProgram(shaderProgram);
		glUseProgram(shaderProgram);
		
		//glDetachShader(shaderProgram, fragmentShader);
		//glDetachShader(shaderProgram, fragmentShader);
		//glDeleteShader(fragmentShader);
		//glDeleteShader(fragmentShader);
		//glDeleteProgram(shaderProgram);
	}
	
	GLint shaderValues = glGetUniformLocation(shaderProgram, "shaderValues");
	
	GLfloat surfaceMaterial[4];
	window->StandardSurfaceColor(surfaceMaterial);
	surfaceMaterial[0] = .8;
	surfaceMaterial[1] = .8;
	surfaceMaterial[2] = 1.;
	
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial); 

	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);

	glEnable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);
	glEnable(GL_NORMALIZE);
	glEnable(GL_RESCALE_NORMAL);
	window->LoadStandardSceneProjectionCamera();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	camera.Load();

	// TODO: can get uniform location with array as a name variable[15]
	// has to be fixed-size... :(
	// void glGetUniformLocation()	
	//	void glUniform1f(GLint location, GLfloat v0);
	// http://www.lighthouse3d.com/tutorials/glsl-tutorial/
	// Actually, seems like going through 128 points per fragment is way too expensive...
	glBegin(GL_TRIANGLES);
	for (int f=0; f < mesh->NFaces(); f++)
	{
		for (int j=0; j<3; j++)
		{
			R3Vector n = mesh->FaceNormal(f);
			R3Point p = mesh->VertexPosition(mesh->VertexOnFace(f, j));
			glNormal3d(n.X(), n.Y(), n.Z());
			glVertex3d(p.X(), p.Y(), p.Z());
		}
	}
	glEnd();	
}
#endif

void VisModel3D::DrawPrepareCamera(VisWindow * window, const R3Affine & camera)
{
	window->LoadStandardSceneProjectionCamera();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	camera.Load();
}

void VisModel3D::DrawPrepareBasicFaces(VisWindow * window, const R3Affine & camera)
{
	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	
	// TODO: do it properly w/ gfx card (i.e. upload geometry)
	glEnable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);
	glEnable(GL_NORMALIZE);
	glEnable(GL_RESCALE_NORMAL);
	glDisable(GL_TEXTURE_2D);
	
	DrawPrepareCamera(window, camera);
}

void VisModel3D::DrawGeomWithLabels(const GeomShape * geom, const ShapeLabels * labels, 
									VisWindow * window, const R3Affine & camera, int maxDepth)
{
	glPointSize(4.);
	const ShapeMesh * mesh = geom->SafeCastToMesh();
	const ShapePointCloud * pnts = geom->SafeCastToPointCloud();
	assert(mesh!=NULL || pnts!=NULL);
	
	DrawPrepareBasicFaces(window, camera);
	std::vector<int> nodesToExplore;
	std::vector<int> nodesDepth;
	
	if (maxDepth==-1)
		maxDepth = labels->MaxDepth();

	bool highlightOnly = false;
	if (maxDepth >= LABELS_OPTVAL_IS_NODEDEPTH && maxDepth < LABELS_OPTVAL_IS_NODEID)
	{
		nodesToExplore.push_back(labels->RootNode());
		nodesDepth.push_back(0);
	}
	else if (maxDepth >= LABELS_OPTVAL_IS_NODEID)
	{
		highlightOnly = true;
		int nodeID = maxDepth - LABELS_OPTVAL_IS_NODEID;
//		std::cout<<"Node: "<<nodeID<<std::endl;
		nodesToExplore.push_back(nodeID);
		maxDepth = labels->NodeDepth(nodeID);
		nodesDepth.push_back(maxDepth);
	}
	
	if (mesh!=NULL)
	{
		glEnable(GL_LIGHTING);
		glDisable(GL_COLOR);
	}
	else if (pnts!=NULL)
	{
		glDisable(GL_LIGHTING);
		glEnable(GL_COLOR);
	}
	
//	glLineWidth(10);
//	glDisable(GL_LIGHTING);
//	glBegin(GL_LINES);
//	glColor3d(1.,0, 0);
//	glVertex3d(0,0,0);
//	glVertex3d(0,0,100);
//	glColor3d(0.,0, 1.);
//	glVertex3d(0,0,0);
//	glVertex3d(0,100,0);
//	glEnd();
	
	VKStringList segLabels;
	std::vector<R3Point> segCentroids;
	while(nodesToExplore.size()>0)
	{
		assert(nodesToExplore.size()==nodesDepth.size());
		int depth = nodesDepth[0];
		int nodeID = nodesToExplore[0];
		VKString segLabel = labels->NodeLabel(nodeID);
		nodesToExplore.erase(nodesToExplore.begin());
		nodesDepth.erase(nodesDepth.begin());
		bool pruneGeom = false;
//		bool pruneGeom = segLabel.startsWith("remove_") || segLabel.startsWith("tiny_");
//		if (pruneGeom)
//			continue;
		//bool pruneGeom = !segLabel.startsWith("object_");

		
		int nChildren = labels->NodeNChildren(nodeID);
		const std::vector<int> * atNodeIDs = NULL;
		if (mesh!=NULL) 
			atNodeIDs = &(labels->NodeFacesAtNode(nodeID));
		else if (pnts!=NULL)
			atNodeIDs = &(labels->NodePointsAtNode(nodeID));
		assert(atNodeIDs!=NULL);
		
		// at depth or has associated geometry: render, terminate
		if (depth==maxDepth || atNodeIDs->size()>0)	
		{
			std::vector<int> idsStorage;
			const std::vector<int> * idsToRender = NULL;
			
			if (depth<maxDepth)	// only because found some faces earlier than deepest recursion
				idsToRender = atNodeIDs;
			else
			{
				if (mesh!=NULL)
					labels->NodeAllFaces(nodeID, idsStorage);
				else if (pnts!=NULL)
					labels->NodeAllPoints(nodeID, idsStorage);
				else
					assert(false);
				idsToRender = &idsStorage;
			}
			assert(idsToRender!=NULL);
			
			static GLfloat surfaceMaterial[4];
			surfaceMaterial[3] = 1.;
			if (highlightOnly)
				window->StandardSurfaceColor(surfaceMaterial);
			else if (nodeID>0)	// not root
				window->MapIntToRadicalColor(nodeID, surfaceMaterial);
			else
				window->StandardSurfaceColor(surfaceMaterial);
			R3Point centroid(0,0,0);
				
			if (mesh!=NULL && !pruneGeom)
			{
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial); 
				glBegin(GL_TRIANGLES);
				for (int i=0; i<(int)(*idsToRender).size(); i++)
				{
					int fID = (*idsToRender)[i];
					for (int j=0; j<3; j++)
					{
						int vID = mesh->VertexOnFace(fID, j);					
						R3Vector n = mesh->FaceNormal(fID);
						R3Point p = mesh->VertexPosition(vID);
						glNormal3d(n.X(), n.Y(), n.Z());
						glVertex3d(p.X(), p.Y(), p.Z());
					}
					centroid += mesh->FacePoint(fID, 1./3., 1./3.);
				}
				glEnd();
			}
			else if (pnts!=NULL && !pruneGeom)
			{
				glColor3d(surfaceMaterial[0], surfaceMaterial[1], surfaceMaterial[2]);
				glBegin(GL_POINTS);
				for (int i=0; i<(int)(*idsToRender).size(); i++)
				{
					R3Point pos = pnts->PointPosition((*idsToRender)[i]);
					glVertex3d(pos.X(), pos.Y(), pos.Z());
					centroid += pos;
				}
				glEnd();
			}
			else
				assert(pruneGeom);
			
			centroid/= (int)(*idsToRender).size();
			
			
			// add geometry names (no need for now...)
//			for (int i=0; i<labels->NumNodeGeometries(nodeID); i++)
//			{
//				segLabel += VKString(" g")+VKString::number(i)+":";
//				segLabel += labels->GeometryLabel(labels->GetGeometryID(nodeID, i));
//			}
			if (!pruneGeom)
			{
				segLabels.push_back(segLabel);
				segCentroids.push_back(centroid);
			}
		}
		
		// go level depper
		if (depth+1 <= maxDepth)
		{
			for (int cID=0; cID<nChildren; cID++)
			{
				nodesDepth.push_back(depth+1);
				nodesToExplore.push_back(labels->NodeChild(nodeID, cID));
			}
		}
	}
	
//	for (int i=0; i < segLabels.count(); i++)
//		 window->GetRenderer()->RenderText(segCentroids[i], segLabels[i], .5, 0, 0,
//										   VisWindowRenderer::TEXT_SIZE_SMALL);
}

void VisModel3D::DrawMeshWithLabeledGeometry(const ShapeMesh * mesh, ShapeLabels * labels, 
											 VisWindow * window, const R3Affine & camera)
{
	DrawPrepareBasicFaces(window, camera);
	std::vector<int> nodesToExplore;
	std::vector<int> nodesDepth;
		
	nodesToExplore.push_back(labels->RootNode());
	nodesDepth.push_back(0);
	
	glDisable(GL_COLOR);
	VKStringList segLabels;
	std::vector<R3Point> segCentroids;
	int globalGeomID = 0;
	//std::cout<<"[deleteme] DrawLabeled..."<<std::endl;
	
	while(nodesToExplore.size()>0)
	{
		assert(nodesToExplore.size()==nodesDepth.size());
		int depth = nodesDepth[0];
		int nodeID = nodesToExplore[0];
		nodesToExplore.erase(nodesToExplore.begin());
		nodesDepth.erase(nodesDepth.begin());
		
		VKString nodeLabel = labels->NodeLabel(nodeID);
		bool pruneGeom = false;
//		bool pruneGeom = nodeLabel.startsWith("remove_") || nodeLabel.startsWith("tiny_");
//		if (pruneGeom)
//			continue;

		
		int nChildren = labels->NodeNChildren(nodeID);		
		if (labels->NumNodeGeometries(nodeID)>0)
		{
			static GLfloat surfaceMaterial[4];
			surfaceMaterial[3] = 1.;
			
			glBegin(GL_TRIANGLES);
			for (int gID=0; gID<labels->NumNodeGeometries(nodeID); gID++)
			{
				R3Point centroid(0,0,0);
				window->MapIntToRadicalColor(globalGeomID, surfaceMaterial);
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial); 
				int f0, nf;
				int g = labels->GetGeometryID(nodeID, gID, &f0, &nf);
				for (int fID=f0; fID<(f0+nf); fID++)
				{
					for (int j=0; j<3; j++)
					{
						int vID = mesh->VertexOnFace(fID, j);					
						R3Vector n = mesh->FaceNormal(fID);
						R3Point p = mesh->VertexPosition(vID);
						glNormal3d(n.X(), n.Y(), n.Z());
						glVertex3d(p.X(), p.Y(), p.Z());
					}
					centroid += mesh->FacePoint(fID, 1./3., 1./3.);
				}
				centroid/= nf;
				globalGeomID++;
				VKString segLabel = labels->GeometryLabel(labels->GetGeometryID(nodeID, gID));
				segLabels.push_back(segLabel);
				segCentroids.push_back(centroid);
			}
			glEnd();
		}
	
		// go level depper
		for (int cID=0; cID<nChildren; cID++)
		{
			nodesDepth.push_back(depth+1);
			nodesToExplore.push_back(labels->NodeChild(nodeID, cID));
		}
	}
	
	for (int i=0; i < segLabels.count(); i++)
		window->GetRenderer()->RenderText(segCentroids[i], segLabels[i], .5, 0, 0,
										  VisWindowRenderer::TEXT_SIZE_SMALL);
}

void VisModel3D::DrawMeshFaces(const ShapeMesh * shape, bool drawMaterials,
							   VisWindow * window, const R3Affine & camera,
							   GLuint * textures, int nTextures, bool colorMesh, 
							   double r, double g, double b, double a)
{
	//std::cout<<"Draw mesh faces"<<std::endl;
	ShapeValue * value = ((ShapeMesh*)shape)->GetShapeVal(ShapeValue::m_sDefaultMaterials);
	ShapeMaterials * materials = NULL;
	if (value!=NULL && drawMaterials)
		materials = value->SafeCastToMaterials();
	
	DrawPrepareBasicFaces(window, camera);
	
	int nMaterials = 0;
	if (materials!=NULL)
		nMaterials = materials->NMaterials();
	for (int matID=0; matID<=nMaterials; matID++)
	{
		// initialize material
		int textureID = -1;
		InitMaterial(materials, matID, window, &textureID, textures, colorMesh, r, g, b, a);
		
		// find faces to render
		const std::vector<int> & faces = ((materials==NULL) ? std::vector<int>()
										  : materials->GetAllFaces(matID));
				
		int nFaces = (int)faces.size();
		if (materials==NULL)
			nFaces = shape->NFaces();
		
		// render faces
		double uv[6];
		int mid;
		if (nFaces>0)
		{
			glBegin(GL_TRIANGLES);
			for (int f=0; f < nFaces; f++)
			{
				int fID = f;
				if (faces.size()>0)
					fID = faces[f];
				if (materials!=NULL && textureID!=-1)
					materials->FaceMaterial(fID, &(uv[0]), &(uv[1]), &(uv[2]), &(uv[3]),
											&(uv[4]), &(uv[5]), &mid);
				for (int j=0; j<3; j++)
				{
					int vID = shape->VertexOnFace(fID, j);
					
					//R3Vector n = shape->VertexNormal(vID);	// fix it: might not be right w/ triangle soups
					R3Vector n = shape->FaceNormal(fID);
					R3Point p = shape->VertexPosition(vID);
					glNormal3d(n.X(), n.Y(), n.Z());
					if (materials!=NULL && textureID!=-1)
						glTexCoord2f(uv[2*j+0], uv[2*j+1]);
					glVertex3d(p.X(), p.Y(), p.Z());
				}
			}
			glEnd();
		}
		else
		{
			glDisable(GL_LIGHTING);
			glColor3d(.2, .2, .2);
			glBegin(GL_POINTS);
			for (int v=0; v<shape->NVertices(); v++)
			{
				R3Point p = shape->VertexPosition(v);
				glVertex3d(p.X(), p.Y(), p.Z());
			}
			glEnd();
		}
	}
}

void VisModel3D::DrawMeshNormals(const ShapeMesh * mesh, VisWindow * window, const R3Affine & camera,
								 double r, double g, double b, double length)
{
	DrawPrepareBasicFaces(window, camera);
	glDisable(GL_LIGHTING);
	glColor3d(r, g, b);
	
	bool withInsideOut = true;
	
	ShapeGrid * test = NULL;
	if (withInsideOut)
	{
		glColor3d(.3, .3, .3);
		static std::map<const ShapeMesh*, ShapeGrid *> testmap;
		if (testmap.find(mesh)==testmap.end())
			testmap[mesh] = new ShapeGrid((ShapeMesh*)mesh, 0, 0.02);
		test = testmap[mesh];
		//glBegin(GL_POINTS);
		for (int i=0; i<test->m_pInsidePoints->NumPoints(); i++)
		{
			R3Point p = test->m_pInsidePoints->PointPosition(i);
			//glVertex3d(p.X(), p.Y(), p.Z());
			R3Sphere(p, 0.01 * test->PointDepth(p)).Draw();
		}
		//glEnd();
	}
	
	for (int f=0; f < mesh->NFaces(); f++)
	{
		R3Point p = mesh->FacePoint(f, 0.3, 0.3);
		R3Vector n = mesh->FaceNormal(f);
		
//		if (test!=NULL)
//		{
//			n.Normalize(); 	n *= 0.5 * test->m_fGridCellSize;
//			double depthAtPoint = test->PointDepth(p-n);
//			double depthAtOffset = test->PointDepth(p+n);
//			if (depthAtPoint < depthAtOffset)
//			{
//				glColor3d(0, 0, 1.);
//				n = -n;
//			}
//			else if (depthAtPoint==depthAtOffset)
//				glColor3d(0, 0, 0);
//			else
//				glColor3d(0, 1., 0);
//		}

//		n.Normalize();
//		R3Sphere(p-n*length*0.1, length*0.05).Draw();
//		glBegin(GL_LINES);
//		p = p - n*length*0.1;
//		glVertex3d(p.X(), p.Y(), p.Z());
//		p = p + n*length;
//		glVertex3d(p.X(), p.Y(), p.Z());
//		glEnd();
	}
	
}

void VisModel3D::DrawMeshEdges(const ShapeMesh * shape, VisWindow * window, const R3Affine & camera,
							   bool colorMesh, double r, double g, double b, double a)
{
	DrawPrepareCamera(window, camera);
	
	glDisable(GL_LIGHTING);
	if (!colorMesh)
		glColor3d(1, 0, 0);
	else	// colored mesh - draw edges in brighter (or darker) version of the color
	{
		if (r==0 && g==0 && b==0) { r=.05; g=.05; b=.05;}
		double lum = 0.2126 * r + 0.7152 * g + 0.0722 * b;
		double s = 1.;
		if (lum < .5)	s = 1.2;
		else s = .8;
		glColor3d(r*s, g*s, b*s);
	}
	
	glBegin(GL_LINES);
	for (int i=0; i<shape->NFaces(); i++)
	{
		int v1ID = shape->VertexOnFace(i, 0);
		int v2ID = shape->VertexOnFace(i, 1);
		int v3ID = shape->VertexOnFace(i, 2);
		//R3Vector n = shape->FaceNormal(i);
		R3Point p1 = shape->VertexPosition(v1ID);
		R3Point p2 = shape->VertexPosition(v2ID);
		R3Point p3 = shape->VertexPosition(v3ID);
		//glNormal3d(n.X(), n.Y(), n.Z());
		glVertex3d(p1.X(), p1.Y(), p1.Z());
		glVertex3d(p2.X(), p2.Y(), p2.Z());
		glVertex3d(p2.X(), p2.Y(), p2.Z());
		glVertex3d(p3.X(), p3.Y(), p3.Z());
		glVertex3d(p3.X(), p3.Y(), p3.Z());
		glVertex3d(p1.X(), p1.Y(), p1.Z());
	}
	glEnd();
}


void VisModel3D::DrawThumbnail(GLuint textureID, double cx, double cy, double dx, double dy)
{
	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glBindTexture( GL_TEXTURE_2D, textureID);
	glColor3ub(255,255,255);
	glDisable(GL_BLEND); 
	glDisable(GL_COLOR_MATERIAL);
	glEnable( GL_TEXTURE_2D );
	
	glBegin(GL_QUADS);
	glNormal3d(0, 0, 1.);
	glTexCoord2f(0, 0);
	glVertex3d(cx-dx/2., cy-dy/2., 0);
	glTexCoord2f(0, 1);
	glVertex3d(cx-dx/2., cy+dy/2., 0);
	glTexCoord2f(1, 1);
	glVertex3d(cx+dx/2., cy+dy/2., 0);
	glTexCoord2f(1, 0);
	glVertex3d(cx+dx/2., cy-dy/2., 0);
	glEnd();
	
	glDisable( GL_TEXTURE_2D );
}


void VisModel3D::InitMaterial(ShapeMaterials * materials, int matID,
							  VisWindow * window, int * textureID, GLuint * textures,
							  bool colorMesh, double r, double g, double b , double a)
{
	*textureID = -1;
	static GLfloat surfaceMaterial[4];
	if (colorMesh)
	{
		surfaceMaterial[0] = r;
		surfaceMaterial[1] = g;
		surfaceMaterial[2] = b;
		surfaceMaterial[3] = a;

		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial); 
	}
	else if (materials!=NULL)
	{
		if (matID==-1 || matID>=materials->NMaterials())
		{
			window->StandardSurfaceColor(surfaceMaterial);
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial); 
		}
		else
		{
			surfaceMaterial[3] = 1.;
			MaterialBRDF * brdf = materials->GetMaterial(matID);
			assert(brdf!=NULL);

			if (brdf->m_DiffuseClr.IsColor())
			{
				for (int c=0; c<3; c++)
					surfaceMaterial[c] = brdf->m_DiffuseClr.m_Color[c];
				glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, surfaceMaterial); 
			}
			else if (brdf->m_DiffuseClr.IsTexture() && (*textureID==-1) && textures!=NULL)
				*textureID = brdf->m_DiffuseClr.m_iTextureID;
			else
			{
				window->StandardSurfaceColor(surfaceMaterial);
				glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, surfaceMaterial); 
			}
			
			if (brdf->m_AmbientClr.IsColor())
			{
				for (int c=0; c<3; c++)
					surfaceMaterial[c] = brdf->m_AmbientClr.m_Color[c];
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, surfaceMaterial); 
			}
			else if (brdf->m_AmbientClr.IsTexture() && (*textureID==-1) && textures!=NULL)
				*textureID = brdf->m_AmbientClr.m_iTextureID;
			else
			{
				window->StandardSurfaceColor(surfaceMaterial);
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, surfaceMaterial); 
			}

			if (brdf->m_SpecularClr.IsColor())
			{
				for (int c=0; c<3; c++)
					surfaceMaterial[c] = brdf->m_SpecularClr.m_Color[c];
				glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, surfaceMaterial); 
			}
			else if (brdf->m_SpecularClr.IsTexture() && (*textureID==-1) && textures!=NULL)
				*textureID = brdf->m_SpecularClr.m_iTextureID;
			else
			{
				surfaceMaterial[0] = 0;	surfaceMaterial[1] = 0;	surfaceMaterial[2] = 0;
				glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, surfaceMaterial); 
			}

			GLfloat shine = brdf->m_fShininess>0 ? brdf->m_fShininess : 0;
			glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, &shine);
			
			if (brdf->m_EmitClr.IsColor())
			{
				for (int c=0; c<3; c++)
					surfaceMaterial[c] = brdf->m_EmitClr.m_Color[c];
				glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, surfaceMaterial); 
			}
			else if (brdf->m_EmitClr.IsTexture() && (*textureID==-1) && textures!=NULL)
				*textureID = brdf->m_EmitClr.m_iTextureID;
			else
			{
				surfaceMaterial[0] = 0;	surfaceMaterial[1] = 0;	surfaceMaterial[2] = 0;
				glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, surfaceMaterial); 
			}
		}
	}
	else
	{
		window->StandardSurfaceColor(surfaceMaterial);
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial); 
	}
	
	if (*textureID>=0)
	{
		window->StandardSurfaceColor(surfaceMaterial);
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial); 

		assert(materials!=NULL);
		int nTextures = materials->NTextures();
		assert(*textureID < nTextures);
		glBindTexture( GL_TEXTURE_2D, textures[*textureID]);
		glColor3ub(255,255,255);
		glDisable(GL_BLEND); 
		glEnable(GL_LIGHTING); 
		glDisable(GL_COLOR_MATERIAL);
		glEnable( GL_TEXTURE_2D );
	}
	else
	{
		glDisable(GL_COLOR_MATERIAL);
		glEnable(GL_BLEND); 
		glEnable(GL_LIGHTING); 
		glDisable( GL_TEXTURE_2D );
	}
}

void VisModel3D::DrawPntsWithScalar(const ShapePointCloud * pnts, bool drawIDs, bool drawAsDots, bool drawAsSphers,
									bool drawNormals, VisWindow * window, const R3Affine & camera, double meshrad, double rScale,
									double * perPointVals, double minValue, double maxValue, double thresh)
{
	meshrad/=4.;
	//std::cout<<"Params: "<<minValue<<" - "<<maxValue<<" "<<thresh<<std::endl;
	// todo: bool drawIDs, bool drawNormals,
	DrawPrepareCamera(window, camera);
	glEnable(GL_DEPTH_TEST);
	
	// TODO: do it properly w/ gfx card
	glEnable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);
	static GLfloat material[4];
	if ((pnts->NumPoints() > 10000 || drawAsDots) && !drawAsSphers)
	{
		glDisable(GL_LIGHTING);
		glColor3d(.2, .2, .2);
		//glPointSize(pnts->NumPoints()>500 ? 1 : r);
		//glPointSize(r);
		glPointSize(1);
	
		//std::cout<<"Drawing.. "<<pnts->NumPoints()<<std::endl;
		glBegin(GL_POINTS);
		for (int i=0; i<pnts->NumPoints(); i++)
		{
			R3Vector n = pnts->PointNormal(i);
			n.Normalize();
			R3Point p = pnts->PointPosition(i);// + (0.005*r) * n;
			if (perPointVals!=NULL)
			{
				window->MapScalarToColor(perPointVals[i], minValue, maxValue, material);
				glColor3d(material[0], material[1], material[2]);
			}
			glVertex3d(p.X(), p.Y(), p.Z());
		}
		glEnd();
	}
	else 
	{
		double radius = meshrad * vkAbs(rScale);
		material[3] = 1.;
		glEnable(GL_LIGHTING);
		for (int i=0; i<pnts->NumPoints(); i++)
		{
			double r, g, b;
			double fRad = radius;
			if (perPointVals==NULL)
			{
				if (pnts->NumPoints()<=18)
					window->MapIntToRadicalColor(i, material);
				else
					window->MapIntToColor(i, material);
			}
			else
			{
				window->MapScalarToColor(perPointVals[i], minValue, maxValue, material);
				if (rScale>=0)
					fRad = radius * (perPointVals[i]-minValue) / (maxValue-minValue);
				else
					fRad = radius;
			}
			
			glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
			if (fRad>0 && (perPointVals==NULL || perPointVals[i] >= thresh))
				R3Sphere(pnts->PointPosition(i), fRad).Draw();
		}
	}
}

void VisModel3D::DrawGeomsToIntersectionBuffer(const std::vector<GeomShape*> & geomSets, VisWindow * window,
											   double r, double rScale, const R3Affine & camera,  const BufferID & bufferID)
{
	DrawPrepareCamera(window, camera);

#ifdef GLUT_RENDERING
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, bufferID.m_iFrameBufferObject);
    glViewport( 0, 0, bufferID.m_iWidth, bufferID.m_iHeight);
    // render points and meshes
    // disable all interpolation
    // instead of colors: 
    //          r - pntID or triID (lowest 2 bits)
    //          g - pntID or triID (next 2 bits)
    //          b - pntID or triID (highest 2 bits)
    //          a - geom ID (0 - no set, so subtract 1 to get id, can be mesh or point)
    //          pntID:		0xXXBBGGRR
	//			pntSetID:	0xAAXXXXXX
	
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);
    glDisable(GL_POLYGON_SMOOTH);
    glShadeModel(GL_FLAT);
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_CULL_FACE);
    glDisable(GL_DITHER);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    	
	glClearColor(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT 
			| GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	
	double radius = r * rScale;
	for (int i=0; i<(int)geomSets.size(); i++)
	{
		const GeomShape * geom = geomSets[i];
		assert(i < 254);
		unsigned char a = (unsigned char)(i+1);
		const ShapePointCloud * pnts = geom->SafeCastToPointCloud();
		const ShapeMesh * mesh = geom->SafeCastToMesh();
		
		int nPrimitves = 0;
		if (mesh!=NULL)
			nPrimitves = mesh->NFaces();
		else if (pnts!=NULL)
			nPrimitves = pnts->NumPoints();
		else
			assert(false);
		
		for (int pID=0; pID<nPrimitves; pID++)
		{
			// check: do we have enough buffer bits?
			//assert((pID>>24)==0); 
			if ((pID>>24)!=0)
			{
				static bool err = true;
				VKString mName = geom->GetModel()->ModelName();
				if (err)
					std::cout<<"[WARNING] Not enough isec buffer bits "<<mName.c_str()<<std::endl;
				err = false;
				continue;
			}

			// choose the color
			unsigned char r = (char)pID;
			unsigned char g = (char)(pID>>8);
			unsigned char b = (char)(pID>>16);
			glColor4ub(r, g, b, a);
			
			// draw primitive
			if (mesh!=NULL)
			{
				R3Point p1 = mesh->FacePoint(pID, 1., 0.);
				R3Point p2 = mesh->FacePoint(pID, 0., 1.);
				R3Point p3 = mesh->FacePoint(pID, 0., 0.);
				glBegin(GL_TRIANGLES);
				glVertex3d(p1.X(), p1.Y(), p1.Z());
				glVertex3d(p2.X(), p2.Y(), p2.Z());
				glVertex3d(p3.X(), p3.Y(), p3.Z());
				glEnd();
			}
			else if (pnts!=NULL)
			{
				R3Point p = pnts->PointPosition(pID);			
				R3Sphere(p, radius).Draw();
			}
			else
				assert(false);
		}
	}    
	window->LoadCurrentBuffer();
#endif
}

////// VisModel3D class ///////
VisModel3D::VisModel3D(Model3D * model, int drawFlags)
{
	//m_fNewPointRadius = 0;
	m_fRenderValueThreshold = 0.;
	m_fRenderValueMinimum = -1;
	m_fRenderValueMaximum = -1;
	//m_bRenderIntersectionBuffers = true;
	m_iEnabledIntersectionBuffers = (IBUF_Pnts | IBUF_Meshes);
	
	m_SelectedPoint = ShapePoint();
	
	//m_fPointRelativeRadius = .03;
	m_fPointRelativeRadius = .03;
	m_iDrawFlags = drawFlags;

	m_pModel = model;
	
	if (m_pModel!=NULL && m_pModel->GeometryExists(Model3D::m_sMainGeomName))
	{
		GeomShape * mainGeom = m_pModel->GetGeomShape(Model3D::m_sMainGeomName);
		assert(mainGeom!=NULL);
		ShapeMesh * mesh = mainGeom->SafeCastToMesh();
		if (mesh!=NULL)
		{
//			m_Origin = mesh->Centroid();
//			m_fRadius = mesh->AverageRadius(m_Origin);
//			UpdateCamera(R3Vector(0, 0, 1), R3Vector(0, -1, 0), 0.5);
			R3Box bbox;
			for (int i=0; i<mesh->NVertices(); i++)
			{
				R3Point vp = mesh->VertexPosition(i);
				if (i==0)
					bbox = R3Box(vp, vp);
				else
					bbox.Union(vp);
			}
			m_Origin = bbox.Centroid();
			m_fRadius = bbox.DiagonalLength();
			UpdateCamera(R3Vector(0, 0, 1), R3Vector(0, -1., 0), 0.5);
			//UpdateCamera(R3Vector(0, 0, 1), R3Vector(-.5, -.5, 0), 4.);
		}
		else	// main could have been a point cloud
			assert(false);
	}
	else
	{
		m_Origin = R3Point(0,0,0);
		m_fRadius = 1;
		UpdateCamera(R3Vector(0, 0, 1), R3Vector(0, -1, 0), 1);
	}
}

VisModel3D::~VisModel3D()
{
	FreeMemory();
}

void VisModel3D::FreeMemory()
{
	std::map<VKString, GeometryTextures*>::iterator iter;
	for (iter = m_GeometryTextures.begin(); iter != m_GeometryTextures.end(); iter++)
	{
		if (iter->second->m_iNTextures > 0)
		{
			glDeleteTextures(iter->second->m_iNTextures, iter->second->m_mMaterialImageToTextureID );
			delete [] iter->second->m_mMaterialImageToTextureID;
		}
	}
	m_GeometryTextures.clear();
}

GeometryTextures * VisModel3D::CreateTextures(GeomShape * geom)
{
	GeometryTextures * newTextures = new GeometryTextures();
	newTextures->m_iNTextures = 0;
	newTextures->m_mMaterialImageToTextureID = NULL;
	newTextures->w = NULL;
	newTextures->h = NULL;
	bool wrap = true;
	
	ShapeMesh * mesh = geom->SafeCastToMesh();
	ShapeImg * thumb = geom->SafeCastToImg();
	std::vector<R2Image*> images;
	if (mesh!=NULL)	// mesh textures
	{
		ShapeValue * value = mesh->GetShapeVal(ShapeValue::m_sDefaultMaterials);
		ShapeMaterials * materials = NULL;
		if (value!=NULL)
			materials = value->SafeCastToMaterials();
		if (materials!=NULL)
		{
			// init texture names and map
			newTextures->m_iNTextures = materials->NTextures();
			if (newTextures->m_iNTextures==0)
				return newTextures;
			newTextures->m_mMaterialImageToTextureID = new GLuint[newTextures->m_iNTextures];
			glGenTextures(newTextures->m_iNTextures, newTextures->m_mMaterialImageToTextureID );
			
			for (int tID=0; tID < materials->NTextures(); tID++)
				images.push_back(materials->GetTexture(tID));
		}
	}
	else if (thumb!=NULL)	// thumbnail
	{
		newTextures->m_iNTextures = 1;
		images.push_back(thumb->GetImage());
	}
	
	if (newTextures->m_iNTextures > 0)
	{
		newTextures->m_mMaterialImageToTextureID = new GLuint[newTextures->m_iNTextures];
		glGenTextures(newTextures->m_iNTextures, newTextures->m_mMaterialImageToTextureID );
		newTextures->w = new int [newTextures->m_iNTextures];
		newTextures->h = new int [newTextures->m_iNTextures];
		
		for (int tID=0; tID < (int)images.size(); tID++)
		{
			R2Image * img = images[tID];
			assert(img!=NULL);
			newTextures->w[tID] = img->Width();	//?? vkchange: maybe change to width
			newTextures->h[tID] = img->Height();
			
			// put image to GPU
			glBindTexture( GL_TEXTURE_2D, newTextures->m_mMaterialImageToTextureID[tID] );
			glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
			glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
			glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
			
			glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap ? GL_REPEAT : GL_CLAMP );
			glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap ? GL_REPEAT : GL_CLAMP );
			
			assert(img->NComponents()==3);
			gluBuild2DMipmaps(GL_TEXTURE_2D, 3, img->Width(), img->Height(),
							  GL_RGB, GL_UNSIGNED_BYTE, img->Pixels() );
		}
	}
	return newTextures;
}

GeometryTextures * VisModel3D::GetTextures(const VKString & geomName)
{
	assert(m_pModel!=NULL);
	
	std::map<VKString, GeometryTextures*>::iterator iter = m_GeometryTextures.find(geomName);
	
	if (iter != m_GeometryTextures.end())	// texture initialized
		return iter->second;

	GeomShape * geom = m_pModel->GetGeomShape(geomName);
	assert(geom!=NULL);
	
	// initialize texture
	m_GeometryTextures[geomName] = CreateTextures(geom);
	
	return m_GeometryTextures[geomName];
}

void VisModel3D::Draw(DatabaseVisWindow * window, const R3Affine & camera, int bufferID, std::vector<RenderableFace> * trFaces)
{
	assert(m_pModel!=NULL);
	std::vector<GeomShape*> isectBuffer;
	std::vector<int> geom2BufferID;
	for (int i=0; i<m_RenderableGeometries.count(); i++)
	{
		// geometries
		GeomShape * geom = m_pModel->GetGeomShape(m_RenderableGeometries[i]);
		if (geom==NULL)
			return;
		if (geom==NULL)
		{
			std::cout<<"[ERROR] No geom: "<<m_RenderableGeometries[i].c_str();
			std::cout<<" in "<<m_pModel->ModelName().c_str()<<std::endl;
		}
		assert(geom!=NULL);
		ShapeMesh * meshGeom = geom->SafeCastToMesh();
		ShapePointCloud * pntsGeom = geom->SafeCastToPointCloud();
		ShapeImg * imgGeom = geom->SafeCastToImg();
		
		// draw geom values (if appropriate)
		//std::cout<<"Draw: "<<geom->GeometryName().c_str()<<std::endl;
		bool drew = false;
		for (int j=0; j<m_RenderableGeomValues.count(); j++)
			drew = DrawGeomVals(geom, m_RenderableGeomValues[j], m_RenderableGeomValsOptional[j], window, camera);
		
		// check if need to draw intersection buffer
		if (((meshGeom!=NULL && IsFlagOn(DRAW_MESHES))			// mesh
			 || (pntsGeom!=NULL && IsFlagOn(DRAW_POINTS)))		// or points
			&& (m_iEnabledIntersectionBuffers & IBUF_Meshes))	// i-sect buffer enabled
		{
			isectBuffer.push_back(geom);
			geom2BufferID.push_back(i);
		}
		
		
		// draw geometries: meshes, points, images, etc...
		if (meshGeom!=NULL && IsFlagOn(DRAW_MESHES))		// draw mesh
		{
			GeometryTextures * textures = GetTextures(m_RenderableGeometries[i]);
			if (IsFlagOn(DRAW_FACES) && !drew)
				DrawMeshFaces(meshGeom, IsFlagOn(DRAW_MATERIALS), window, camera,
							  textures->m_mMaterialImageToTextureID, textures->m_iNTextures);
			if (IsFlagOn(DRAW_NORMALS))
				DrawMeshNormals(meshGeom, window, camera, 1., 0., 0., 0.1);
			if (IsFlagOn(DRAW_EDGES))
				DrawMeshEdges(meshGeom, window, camera,
							  textures->m_mMaterialImageToTextureID, textures->m_iNTextures);
			// unimplemented: DRAW_VERTICES, DRAW_IDs, DRAW_NORMALS
		}
		else if (pntsGeom!=NULL && IsFlagOn(DRAW_POINTS) && !drew)	// point set
		{
			DrawPntsWithScalar(pntsGeom, IsFlagOn(DRAW_IDs), IsFlagOn(DRAW_PDOTS), IsFlagOn(DRAW_PSPHERES),
							   IsFlagOn(DRAW_NORMALS), window, camera, m_fRadius, m_fPointRelativeRadius, NULL, 0, 1);
		}
		else if (imgGeom!=NULL)		// thumbnail
			DrawGeomImg(imgGeom, GetTextures(m_RenderableGeometries[i]), window, bufferID);
	}
	
	
	
	// draw intersection buffers
	DrawIsectBuffers(window, bufferID, isectBuffer, geom2BufferID, &camera);
	
	// draw selected point
	if (!m_SelectedPoint.Invalid())
	{
		R3Point p = m_SelectedPoint.GetPosition();
		
		static GLfloat material[4];
		double radius = m_fRadius * m_fPointRelativeRadius*1.5;
		material[0] = 0.;
		material[1] = 0.;
		material[2] = .7;
		material[3] = 1.;
		glEnable(GL_LIGHTING);
		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
		R3Sphere(p, radius).Draw();
	}
}

void VisModel3D::Draw(DatabaseVisWindow * window, int cameraID)
{	
	Draw(window, GetCameraTransform(cameraID), cameraID);
}

void VisModel3D::DrawGeomImg(ShapeImg * imgGeom, GeometryTextures * textures, 
							 DatabaseVisWindow * window, int cameraID)
{
	double cx, cy, dx, dy, s = -1;
	for (std::map<int,int>::iterator iter = m_mSubwindowCamera.begin(); 
		 iter != m_mSubwindowCamera.end(); iter++)
		if (cameraID == iter->second)
			window->GetWindowParams(iter->first, &cx, &cy, &s, &dx, &dy);
	if (s>=0)
	{
//		GeometryTextures * textures = GetTextures(m_RenderableGeometries[i]);
		assert(textures->m_iNTextures==1);
		
		dx /= window->Width();
		dy /= window->Height();
		
		double imgAspect = ((double)textures->h[0]) / ((double)textures->w[0]);
		double wndAspect = dy / dx;
		
		if (imgAspect < wndAspect)
			dy = dx * imgAspect;
		else
			dx = dy / imgAspect;
		
		cx /= window->Width();
		cy /= window->Height();
		dx *= .9;
		dy *= .9;
		
		DrawThumbnail(textures->m_mMaterialImageToTextureID[0], cx, cy, dx, dy);
	}
}

void VisModel3D::DrawIsectBuffers(DatabaseVisWindow * window, int bufferID, const std::vector<GeomShape*> & geoms,
								  const std::vector<int> & geomIDs, const R3Affine * cameraPtr)
{
	assert(cameraPtr!=NULL);
	const R3Affine & camera = *cameraPtr;
	std::vector<int> newGeomIds = geomIDs;
	while (newGeomIds.size() < geoms.size())
	{
		int currID = (int)newGeomIds.size();
		GeomShape * shape = geoms[currID];
		int selID = -1;
		for (int i=0; i<NumRenderableGeometries() && selID==-1; i++)
			if (GetRenderableGeometry(i)==shape)
				selID = i;
		if (selID==-1)
		{
			std::cout<<"[WARNING] Isect buffer draw failed (geom= "<<shape->GeometryName().c_str()<<")"<<std::flush;
			return;
		}
		newGeomIds.push_back(selID);
	}
		
	if (geoms.size()>0)
	{
		//R3Affine camera = cameraPtr!=NULL ? *cameraPtr : GetCameraTransform(cameraID);
		std::map<int, int>::iterator iter = m_CameraIDToIntBufferID.find(bufferID);
		BufferID * intBuff = NULL;
		if (iter == m_CameraIDToIntBufferID.end())
		{
			int intBuffID = (int)m_IntersectionBuffers.size();
			m_IntersectionBuffers.push_back(window->CreateBuffer());
			m_CameraIDToIntBufferID[bufferID] = intBuffID;
			intBuff = &(m_IntersectionBuffers[intBuffID]);
			m_IntersectionGeomIDs.push_back(newGeomIds);
		}
		else
		{
			intBuff = &(m_IntersectionBuffers[iter->second]);
			m_IntersectionGeomIDs[iter->second] = newGeomIds;
		}
		double r1 = m_fRadius, r2 = m_fPointRelativeRadius;
		DrawGeomsToIntersectionBuffer(geoms, window, r1, r2, camera, *intBuff);
	}
}

bool VisModel3D::DrawGeomVals(GeomShape * geom, const VKString & geomValsName,
							  int optVal, DatabaseVisWindow * window, const R3Affine & camera)
{
	bool doNotRedraw = false;
	ShapeMesh * geomMesh = geom->SafeCastToMesh();
	ShapePointCloud * geomPnts = geom->SafeCastToPointCloud();
	ShapeValue * value = geom->GetShapeVal(geomValsName);
	if (value!=NULL)
	{
		ShapeMaterials * material = value->SafeCastToMaterials();
		ShapeLabels * labels = value->SafeCastToLabels();
		
		if (material!=NULL)		// materials
		{
			// do nothing? (default materials can be rendered with meshflags...)
		}
		else if (labels!=NULL)	// segmentation
		{
			//if (optVal<=labels->MaxDepth() && optVal )
				DrawGeomWithLabels(geom, labels, window, camera, optVal);
			//else if (geomMesh!=NULL)
			//	DrawMeshWithLabeledGeometry(geomMesh, labels, window, camera);
			doNotRedraw = true;
		}
		else
		{
			if (geomPnts!=NULL)
			{
				static double * ppv = NULL;
				static int size = -1;
				
				double minval = -1;
				double maxval = -1;
				
				if (size<geomPnts->NumPoints())
				{
					delete [] ppv;
					ppv = new double[geomPnts->NumPoints()];
					size = geomPnts->NumPoints();
				}
				
				int featid = optVal;
				for (int i=0; i<geomPnts->NumPoints(); i++)
				{
					double val = value->Value(geomPnts->GetPoint(i), featid);
					if (i==0)
					{
						minval = val;
						maxval = val;
					}
					else
					{
						minval = vkMin(minval, val);
						maxval = vkMax(maxval, val); 
					}
					ppv[i] = val;
				}
				if (m_fRenderValueMinimum>=0)
					minval = m_fRenderValueMinimum;
				if (m_fRenderValueMaximum>=0)
					maxval = m_fRenderValueMaximum;
				DrawPntsWithScalar(geomPnts, IsFlagOn(DRAW_IDs), IsFlagOn(DRAW_PDOTS), IsFlagOn(DRAW_PSPHERES),
								   IsFlagOn(DRAW_NORMALS), window, camera, m_fRadius,
								   //m_fNewPointRadius <=0 ? m_fPointRelativeRadius : -m_fNewPointRadius,
								   m_fPointRelativeRadius,
								   ppv, minval, maxval, m_fRenderValueThreshold);
				doNotRedraw = true;
			}
			else if (geomMesh!=NULL)
			{
				// nothing for now..
			}
		}
	}
	return doNotRedraw;
}

// WHAT TO RENDER
void VisModel3D::SetIntersectionBuffer(IntersectBufferType bufferTp, bool enable)
{
	if (enable)
		m_iEnabledIntersectionBuffers |= bufferTp;
	else
		m_iEnabledIntersectionBuffers &= (~bufferTp);
}

bool VisModel3D::IsFlagOn(DrawFlags drawFlag)
{
	return (m_iDrawFlags & drawFlag) != 0;
}

void VisModel3D::ToggleFlag(DrawFlags drawFlag)
{
	if (IsFlagOn(drawFlag))
		m_iDrawFlags = m_iDrawFlags & (~drawFlag);
	else
		m_iDrawFlags = m_iDrawFlags | drawFlag;
}

void VisModel3D::SetFlag(DrawFlags drawFlag, bool enable)
{
	if (IsFlagOn(drawFlag)!=enable)
		ToggleFlag(drawFlag);

}

void VisModel3D::AddRenderableGeomValue(const VKString & geomValueName, int optVal)
{
	int posID = m_RenderableGeomValues.containsGetID(geomValueName);
	if (posID==-1)
	{
		m_RenderableGeomValues.push_back(geomValueName);
		m_RenderableGeomValsOptional.push_back(optVal);
	}
	else
		m_RenderableGeomValsOptional[posID] = optVal;
}

void VisModel3D::RemoveRenderableGeomValue(const VKString & geomValueName)
{
	if (geomValueName=="all")
	{
		m_RenderableGeomValues.clear();
		m_RenderableGeomValsOptional.clear();
	}

	int posID = m_RenderableGeomValues.containsGetID(geomValueName);
	if (posID>=0)
	{
		m_RenderableGeomValues.remove(posID);
		m_RenderableGeomValsOptional.erase(m_RenderableGeomValsOptional.begin()+posID);
	}
}

int VisModel3D::NumRenderableValues() const
{
	return (int)m_RenderableGeomValues.count();
}

const VKString & VisModel3D::GetRenderableGeomValueName( int gvid, int * optval) const
{
	if (optval!=NULL && gvid < (int)m_RenderableGeomValsOptional.size())
		*optval = m_RenderableGeomValsOptional[gvid];
	return m_RenderableGeomValues[gvid];
}


void VisModel3D::AddRenderableGeometry(const VKString & geometryName)
{
	if (!m_RenderableGeometries.contains(geometryName))
		m_RenderableGeometries.push_back(geometryName);
}

void VisModel3D::ClearRenderablgeGeometries()
{
	m_RenderableGeometries.clear();
}

int VisModel3D::NumRenderableGeometries()
{
	return (int)m_RenderableGeometries.count();
}

GeomShape * VisModel3D::GetRenderableGeometry(int geomID)
{
	assert(m_pModel!=NULL);
	assert(geomID>=0 && geomID < m_RenderableGeometries.count());
	return m_pModel->GetGeomShape(m_RenderableGeometries[geomID]);
}

void VisModel3D::RemoveRenderableGeometry(const VKString & geometryName)
{
	if (m_RenderableGeometries.contains(geometryName))
		m_RenderableGeometries.remove(geometryName);
}

Model3D * VisModel3D::GetModel()
{
	assert(m_pModel!=NULL);
	return m_pModel;
}

// CAMERAS
int VisModel3D::SubwindowCameraID(int subwindowID, DatabaseVisWindow * window)
{
	std::map<int, int>::iterator iter = m_mSubwindowCamera.find(subwindowID);
	if (iter==m_mSubwindowCamera.end())
	{
		int camID = CreateNewCamera();
		m_mSubwindowCamera[subwindowID] = camID;
		double cx, cy, s, dx, dy;
		
		window->GetWindowParams(subwindowID, &cx, &cy, &s, &dx, &dy);
		
		dx /= window->Width();
		dy /= window->Height();
		cx /= window->Width();
		cy /= window->Height();
		m_CameraPostMap[camID].Translate(R3Vector(cx*1.2, cy, 0.));
		m_CameraPostMap[camID].Scale(s*.75);
		return camID;
	}
	else
		return iter->second;
}

int VisModel3D::CreateNewCamera()
{
	//R3Triad triad(m_Towards, m_Up); 
	R3Triad triad(R3Vector(1,0,0), R3Vector(0,1,0), R3Vector(0,0,1));
	triad.Rotate(triad.Axis(1), -m_Up);
	triad.Rotate(triad.Axis(2), 3.14 / 4.);
	triad.Rotate(R3Vector(1.,0,0), 3.14 / 6.);
	
	//R3Vector newUp
	
	m_CamerasCoord.push_back(R3CoordSystem(R3Point(0,0,0), triad));
	m_CamerasScale.push_back(m_fRelativeScale);

	int camID = (int)m_CameraPostMap.size();
	m_CameraPostMap.push_back(R3Affine(R4identity_matrix));
	m_CameraPostMap[camID].Translate(R3Vector(0, 0, -2.));
	//m_CameraPostMap[camID].Rotate(m_Up, 3.14/4);
	
	return camID;
}

double VisModel3D::GetMeshRadius() const
{
	return m_fRadius;
}

double VisModel3D::GetRelativePntRadius() const
{
	return m_fPointRelativeRadius;
}

R3Affine VisModel3D::GetCameraTransform(int cameraID, double * scale) const
{
	assert(cameraID>=0 && cameraID<(int)m_CamerasCoord.size());
	
	R3Affine m0(R4identity_matrix);
		
	// subwindow positioning
	m0.Transform(m_CameraPostMap[cameraID]);

	// camera stuff:
	R3Point p0 = m_CamerasCoord[cameraID].Origin();
	R3Vector o(p0.X(), p0.Y(), p0.Z());
	R3Affine axesRot(m_CamerasCoord[cameraID].Axes().Matrix());
	m0.Translate(o);
	m0.Transform(axesRot);
	m0.Scale(m_CamerasScale[cameraID]);

	if (scale!=NULL)
		*scale = m_CamerasScale[cameraID];
	
	// object-space transformations (manual)
	if (cameraID < (int)m_ObjSpaceTransform.size())
		m0.Transform(m_ObjSpaceTransform[cameraID]);

	// normalize mesh
	if (cameraID < (int)m_NormalizeMeshForCamera.size() && m_NormalizeMeshForCamera[cameraID])
	{
		m0.Scale(1. / (2 * m_fRadius));
		m0.Translate(-R3Vector(m_Origin.X(), m_Origin.Y(), m_Origin.Z()));
	}

	return m0;	
}

//R3Affine VisModel3D::GetCameraTransform(const R3Affine & insertObjectSpaceXForm,
//										int cameraID, double * scale) const
//{
//	assert(cameraID>=0 && cameraID<(int)m_CamerasCoord.size());
//	
//	R3Affine m0(R4identity_matrix);
//	
//	// subwindow positioning
//	m0.Transform(m_CameraPostMap[cameraID]);
//	
//	// camera stuff:
//	R3Point p0 = m_CamerasCoord[cameraID].Origin();
//	R3Vector o(p0.X(), p0.Y(), p0.Z());
//	R3Affine axesRot(m_CamerasCoord[cameraID].Axes().Matrix());
//	m0.Translate(o);
//	m0.Transform(axesRot);
//	m0.Scale(m_CamerasScale[cameraID]);
//	
//	if (scale!=NULL)
//		*scale = m_CamerasScale[cameraID];
//	
//	// initial scaling (use provided map)
//	m0.Scale(1. / (2. * m_fRadius));
//	m0.Translate(-R3Vector(m_Origin.X(), m_Origin.Y(), m_Origin.Z()));
//	
//	m0.Transform(insertObjectSpaceXForm);
//	
//	return m0;	
//}

void VisModel3D::WindowResize(int newW, int newH)
{
	std::cout<<"[WARNING] Resize not supported"<<std::endl;
}

void VisModel3D::Rotate(int cameraID, R3Vector axis,  double radians, VisWindow * window)
{
	assert(cameraID>=0 && cameraID<(int)m_CamerasCoord.size());
	m_CamerasCoord[cameraID].Rotate(axis, radians);
}

void VisModel3D::Rotate(int cameraID, int x, int y, int dx, int dy, VisWindow * window)
{
	assert(cameraID>=0 && cameraID<(int)m_CamerasCoord.size());		
	
	if ((dx == 0) && (dy == 0)) return;

	R3CoordSystem & cs = m_CamerasCoord[cameraID];
//	R3Vector right = cs.Axes().Axis(RN_X);
//	R3Vector up = cs.Axes().Axis(RN_Y);
	
	R3Vector right(1., 0, 0);
	R3Vector up(0, 1., 0);
	
    RNLength vx = (double) dx / (double) window->Width();
    RNLength vy = (double) dy / (double) window->Height();
    RNAngle theta = 4.0 * (fabs(vx) + fabs(vy));
    R3Vector vector = (right * vx) + (up * vy);
    R3Vector axis = (right % up) % vector;
    axis.Normalize();
	
	R3Triad axes = cs.Axes();
	axes.Rotate(axis, theta);
	cs.SetAxes(axes);
}

void VisModel3D::Translate(int cameraID, int x, int y, int dx, int dy, VisWindow * window)
{
	assert(cameraID>=0 && cameraID<(int)m_CamerasCoord.size());		
	if ((dx == 0) && (dy == 0)) return;
	R3CoordSystem & cs = m_CamerasCoord[cameraID];
	double s = m_CamerasScale[cameraID];
//	R3Vector right = -cs.Axes().Axis(RN_X);
//	R3Vector up = -cs.Axes().Axis(RN_Y);
	R3Vector right = R3Vector(1., 0, 0);
	R3Vector up = R3Vector(0, 1., 0);
	double vx = 3. * (double) dx / (double) window->Width() / s;
	double vy = 3. * (double) dy / (double) window->Height() / s;
	cs.Translate(right * vx  + up * vy);
}

void VisModel3D::Scale(int cameraID, int x, int y, int dx, int dy, VisWindow * window)
{
	assert(cameraID>=0 && cameraID<(int)m_CamerasScale.size());
	if (dy==0) return;
	double ds=(double)dy/(double)window->Height();	
	Scale(cameraID, ds, window);
}

void VisModel3D::Scale(int cameraID, double ds, VisWindow * window)
{
	assert(cameraID>=0 && cameraID<(int)m_CamerasScale.size());
	m_CamerasScale[cameraID] *= (1. + ds);
}

void VisModel3D::SetValueLimits(double minval, double maxval, double thresh)
{
	m_fRenderValueMinimum = minval;
	m_fRenderValueMaximum = maxval;
	m_fRenderValueThreshold = thresh;
}

void VisModel3D::SetPointRadius(double newpointradius)
{
	//m_fNewPointRadius = newpointradius;
	m_fPointRelativeRadius = newpointradius;
}

void VisModel3D::UpdateCamera(const R3Vector & upVec, const R3Vector & towardsVec, double relativeScale)
{
	//assert(cameraID>=0 && cameraID<(int)m_Cameras.size());
	m_Towards = towardsVec;
	m_Towards.Normalize();
	m_Up = upVec;
	m_Up.Normalize();
	m_fRelativeScale = relativeScale;

	m_mSubwindowCamera.clear();
	m_CamerasCoord.clear();
	m_CamerasScale.clear();
	m_CameraPostMap.clear();
}

void VisModel3D::SetObjSpaceTransform(int cameraID, const R3Affine & preTransform, bool normalizeMesh)
{
	while (cameraID>=(int)m_ObjSpaceTransform.size())
		m_ObjSpaceTransform.push_back(R3identity_affine);
	m_ObjSpaceTransform[cameraID] = preTransform;

	while (cameraID>=(int)m_NormalizeMeshForCamera.size())
		m_NormalizeMeshForCamera.push_back(true);
	m_NormalizeMeshForCamera[cameraID] = normalizeMesh;
}

void VisModel3D::QuickIntersectTriangleID(int cameraID, int x, int y, int w, int h,
										  std::vector<int> & fids, VisWindow * window)
{
	std::map<int, int>::iterator iter = m_CameraIDToIntBufferID.find(cameraID);
	if (iter==m_CameraIDToIntBufferID.end())
	{
		std::cout<<"[WARNING] Intersection buffer does not exist: "<<cameraID<<std::endl;
		return;
	}
#ifdef GLUT_RENDERING
	BufferID & bufferID = m_IntersectionBuffers[iter->second];
	
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, bufferID.m_iFrameBufferObject);
    glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    assert(bufferID.m_iWidth >= viewport[2]);
    assert(bufferID.m_iHeight >= viewport[3]);
	
	x = vkMax(x, 0);
	y = vkMax(y, 0);
	w = vkMin(w, window->Width()-1 - x);
	h = vkMin(w, window->Height()-1 - y);

    glReadPixels(x, y, w, h, GL_RGBA, GL_UNSIGNED_BYTE, bufferID.m_aColorBuffer);
	window->LoadCurrentBuffer();
	// something to intersect
	for (int i=0; i<w*h; i++)
	{
		int r = (int)bufferID.m_aColorBuffer[i*4+0];
		int g = (int)bufferID.m_aColorBuffer[i*4+1];
		int b = (int)bufferID.m_aColorBuffer[i*4+2];
		int a = (int)bufferID.m_aColorBuffer[i*4+3];
		if (a>0)
			fids.push_back(r | (g<<8) | (b<<16));
	}
#endif
	
}

ShapePoint VisModel3D::QuickIntersection(int cameraID, const R3Affine & xform, int x, int y, VisWindow * window)
{
	std::map<int, int>::iterator iter = m_CameraIDToIntBufferID.find(cameraID);
	if (iter==m_CameraIDToIntBufferID.end())
	{
		std::cout<<"[WARNING] Intersection buffer does not exist: "<<cameraID<<std::endl;
		return ShapePoint();
	}
#ifndef GLUT_RENDERING
	return ShapePoint();
#else
	BufferID & bufferID = m_IntersectionBuffers[iter->second];
	std::vector<int> setsIDs = m_IntersectionGeomIDs[iter->second];
	
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, bufferID.m_iFrameBufferObject);
    glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    assert(bufferID.m_iWidth >= viewport[2]);
    assert(bufferID.m_iHeight >= viewport[3]);
	
    glReadPixels(0, 0, bufferID.m_iWidth, bufferID.m_iHeight,
				 GL_RGBA, GL_UNSIGNED_BYTE, bufferID.m_aColorBuffer);
	window->LoadCurrentBuffer();
    
    // find triangle from color: r, g, b (if a = 0) at x, y
    int id = y * bufferID.m_iWidth + x;
    int r = bufferID.m_aColorBuffer[4 * id + 0];
    int g = bufferID.m_aColorBuffer[4 * id + 1];
    int b = bufferID.m_aColorBuffer[4 * id + 2];
    int a = bufferID.m_aColorBuffer[4 * id + 3];
	
	// something to intersect
    if (a > 0)
    {
		int pSetID = a - 1;
		assert(pSetID < (int)setsIDs.size());
		pSetID = setsIDs[pSetID];
		if (pSetID<0 || pSetID >= NumRenderableGeometries())
		{
			std::cout<<"[WARNING] Invalid geometry id: ";
			std::cout<<pSetID<<" / "<<NumRenderableGeometries()<<std::endl;
			return ShapePoint();
		}
		GeomShape * shape = GetRenderableGeometry(pSetID);
		assert(shape!=NULL);
		ShapePointCloud * pnts = shape->SafeCastToPointCloud();
		ShapeMesh * mesh = shape->SafeCastToMesh();
		int pID = r | (g<<8) | (b<<16);
		
		if (mesh!=NULL)	// find point on a triangle
		{
			assert(pID>=0 && pID < mesh->NFaces());
			window->LoadStandardSceneProjectionCamera();
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			xform.Load();
			
			GLdouble modelview[16];
			GLdouble projection[16];
			glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
			glGetDoublev(GL_PROJECTION_MATRIX, projection);
			GLfloat fwinZ;
			glReadPixels(x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &fwinZ);
			GLdouble winX = x, winY = y, winZ = fwinZ;
			GLdouble posX, posY, posZ;
			gluUnProject( winX, winY, winZ, modelview, projection, viewport, &posX, &posY, &posZ);
			
			R3Point b = mesh->FaceBarycentric(pID, R3Point(posX, posY, posZ));
			// NOTE: pID might be a bit unstable - maybe look for neighboring triangles
			//			otherwise b - might not be valid barycentric coordinates
			return ShapePoint(mesh, pID, b.X(), b.Y());
		}
		else if (pnts!=NULL) // just point id
		{
			assert(pID>=0 && pID < pnts->NumPoints());
			return pnts->GetPoint(pID);
		}
		else
		{
			VKString mName = shape->GetModel()->ModelName();
			std::cout<<"[WARNING] Unrecognized w/ ID: "<<pSetID<<" : "<<mName.c_str()<<std::endl;
		}
    }
	return ShapePoint();
#endif
}

ShapePoint VisModel3D::QuickIntersection(int cameraID, int x, int y, VisWindow * window)
{
	return QuickIntersection(cameraID, GetCameraTransform(cameraID), x, y, window);
}

void VisModel3D::SelectPoint(const ShapePoint & pnt)
{
	m_SelectedPoint = pnt;	
	if (m_SelectedPoint.Invalid())
		std::cout<<"Unselected"<<std::endl;
	else
	{
		int fID;
		double b1, b2;
		const ShapeMesh * msh;
		m_SelectedPoint.GetBarycentricCoords(&fID, &b1, &b2, &msh);
		std::cout<<"Selected. Face: "<<fID;
		std::cout<<" id="<<m_SelectedPoint.PointID();
		R3Point pos = m_SelectedPoint.GetPosition();
		std::cout<<" pos=["<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<"]"<<std::endl;
		
	}
}

void VisModel3D::SelectPoint(int cameraID, int x, int y, VisWindow * window)
{
	SelectPoint(QuickIntersection(cameraID, x, y, window));
}

void VisModel3D::UnselectPoint()
{
	m_SelectedPoint = ShapePoint();
}

const ShapePoint & VisModel3D::SelectedPoint() const
{
	return m_SelectedPoint;
}

void VisModel3D::WriteCamera(FILE * file) const
{
	int nCameras = (int)m_CamerasCoord.size();
	fwrite(&nCameras, sizeof(int), 1, file);
	double xyz[3];	
	for (int i=0; i<nCameras; i++)
	{
		double axes[9];
		for (int d=0; d<3; d++)
		{
			xyz[d] = m_CamerasCoord[i].Origin()[d];
			for (int vd=0; vd<3; vd++)
				axes[d*3 + vd] = m_CamerasCoord[i].Axes()[d][vd];
		}
		fwrite(xyz, sizeof(double), 3, file);
		fwrite(axes, sizeof(double), 9, file);
		fwrite(&(m_CamerasScale[i]), sizeof(double), 1, file);
		
		double mtx[16];
		for (int r=0; r<4; r++)
		for (int c=0; c<4; c++)	
			mtx[r*4+c] = m_CameraPostMap[i].Matrix()[r][c];
		fwrite(mtx, sizeof(double), 16, file);
	}
	
	for (int d=0; d<3; d++)
		xyz[d] = m_Origin[d];
	fwrite(xyz, sizeof(double), 3, file);
	fwrite(&m_fRadius, sizeof(double), 1, file);

	for (int d=0; d<3; d++)
		xyz[d] = m_Towards[d];
	fwrite(xyz, sizeof(double), 3, file);
	
	for (int d=0; d<3; d++)
		xyz[d] = m_Up[d];
	fwrite(xyz, sizeof(double), 3, file);
	
	fwrite(&m_fRelativeScale, sizeof(double), 1, file);
	
	int nPairs = (int)m_mSubwindowCamera.size();
	fwrite(&nPairs, sizeof(int), 1, file);
	for (std::map<int, int>::iterator iter = ((VisModel3D*)this)->m_mSubwindowCamera.begin(); 
		 iter != m_mSubwindowCamera.end(); iter++)
	{
		fwrite(&(iter->first), sizeof(int), 1, file);
		fwrite(&(iter->second), sizeof(int), 1, file);
	}
}

void VisModel3D::ReadCamera(FILE * file)
{
	m_CamerasCoord.clear();
	m_mSubwindowCamera.clear();
	m_CamerasScale.clear();
	m_CameraPostMap.clear();
	
	int nCameras;
	fread(&nCameras, sizeof(int), 1, file);
	double xyz[3];	
	double axes[9];
	double mtx[16];
	for (int i=0; i<nCameras; i++)
	{
		m_CamerasScale.push_back(1);
		fread(xyz, sizeof(double), 3, file);
		fread(axes, sizeof(double), 9, file);
		fread(&(m_CamerasScale[i]), sizeof(double), 1, file);
		
		R3Point orig(xyz[0], xyz[1], xyz[2]);
		R3Vector v1(axes[0], axes[1], axes[2]);
		R3Vector v2(axes[3], axes[4], axes[5]);
		R3Vector v3(axes[6], axes[7], axes[8]);
		
		m_CamerasCoord.push_back(R3CoordSystem(orig, R3Triad(v1, v2, v3)));
		
		fread(mtx, sizeof(double), 16, file);
		m_CameraPostMap.push_back(R3Affine(R4Matrix(mtx)));
	}
	
	fread(xyz, sizeof(double), 3, file);
	m_Origin = R3Point(xyz[0], xyz[1], xyz[2]);
	
	fread(&m_fRadius, sizeof(double), 1, file);

	fread(xyz, sizeof(double), 3, file);
	m_Towards = R3Vector(xyz[0], xyz[1], xyz[2]);
	
	fread(xyz, sizeof(double), 3, file);
	m_Up = R3Vector(xyz[0], xyz[1], xyz[2]);
	
	fread(&m_fRelativeScale, sizeof(double), 1, file);
	
	int nPairs;
	fread(&nPairs, sizeof(int), 1, file);
	int n1, n2;
	for (int i=0; i<nPairs; i++)
	{
		fread(&n1, sizeof(int), 1, file);
		fread(&n2, sizeof(int), 1, file);
		m_mSubwindowCamera[n1] = n2;
	}
}
