/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

/*******************************************************/
/*******************************************************/
/*******************************************************/
/****** Renderer Implementation (abstract class) *******/
/*******************************************************/
/*******************************************************/
/*******************************************************/
#include "VisWindow.h"
#include "Sortable.h"

RenderableFace::RenderableFace()
{
	vtx[0] = R3Point(0,0,0);
	vtx[1] = R3Point(0,0,0);
	vtx[2] = R3Point(0,0,0);
	norm = R3Vector(0,0,0);
	r = 0.;
	g = 0.;
	b = 0.;
	a = 0.;
	m_FaceType = RENDFACE_Triangle;
	m_fPointRadius = 0;
}

RenderableFace::RenderableFace(const R3Point & pos, FaceType type, double pointRadius,
							   double r, double g, double b, double a)
{
	vtx[0] = pos;
	m_fPointRadius = pointRadius;
	m_FaceType = type;
	this->r = r;
	this->g = g;
	this->b = b;
	this->a = a;
}

RenderableFace::RenderableFace(const R3Point & p1, const R3Point & p2, const R3Point & p3,
							   double r, double g, double b, double a, const R3Vector & norm)
{
	vtx[0] = p1;
	vtx[1] = p2;
	vtx[2] = p3;
	m_FaceType = RENDFACE_Triangle;
	this->r = r;
	this->g = g;
	this->b = b;
	this->a = a;
	this->norm = norm;
	if (norm.Length()==0)
	{
		R3Vector v1 = p2 - p1;
		R3Vector v2 = p3 - p1;
		this->norm = v1 % v2;
		this->norm.Normalize();
	}
}


VisWindowRenderer::VisWindowRenderer(VisWindow * window)
{
	window->SetRenderer(this);
	m_pWindow = window;	
}

void VisWindowRenderer::TerminateWindow()
{
}

void VisWindowRenderer::UpdateScreen()
{
}

void VisWindowRenderer::RenderText(R2Point pnt, const VKString & text, 
								   double r, double g, double b, TextSize size) const
{
	static bool warnOnce = true;
	if (warnOnce)
	{
		std::cout<<"[WARNING] Text Rendering not implemented."<<std::endl;
		std::cout<<"\tText='"<<text.c_str()<<"'"<<std::endl;
		warnOnce = false;
	}
}

void VisWindowRenderer::RenderText(R3Point pnt, const VKString & text, 
								   double r, double g, double b, TextSize size) const
{
	static bool warnOnce = true;
	if (warnOnce)
	{
		std::cout<<"[WARNING] Text Rendering not implemented."<<std::endl;
		std::cout<<"\tText='"<<text.c_str()<<"'"<<std::endl;
		warnOnce = false;
	}
}

void VisWindowRenderer::CreateBuffer(BufferID & bufferID)
{
#ifdef GLUT_RENDERING
	int width = m_pWindow->Width();
	int height = m_pWindow->Height();
	glGenRenderbuffers(1, &bufferID.m_iFrameBufferObject);
	
	// attach depth / color to the buffer object
	glBindFramebuffer(GL_FRAMEBUFFER, bufferID.m_iFrameBufferObject);
	
	glGenRenderbuffers(1, &bufferID.m_iDepthBuffer);
	glBindRenderbuffer(GL_RENDERBUFFER, bufferID.m_iDepthBuffer);
	glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, width, height);
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, 
							  GL_RENDERBUFFER, bufferID.m_iDepthBuffer);
	
	glGenRenderbuffers(1, &bufferID.m_iColorBuffer);
	glBindRenderbuffer(GL_RENDERBUFFER, bufferID.m_iColorBuffer);
	if (VisWindow::BUFFER_COLOR_COMPONENTS==4)
		glRenderbufferStorage(GL_RENDERBUFFER, GL_RGBA, width, height);
//	else if (BUFFER_COLOR_COMPONENTS==3)
//		glRenderbufferStorage(GL_RENDERBUFFER, GL_RGB, width, height);
	else
		assert(false);
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, 
							  GL_RENDERBUFFER, bufferID.m_iColorBuffer);
	
	glGenTextures( 1, &bufferID.m_iTextureBuffer);
	glBindTexture( GL_TEXTURE_2D, bufferID.m_iTextureBuffer );
	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL );
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); 
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, 
						   GL_TEXTURE_2D, bufferID.m_iTextureBuffer, 0 );
	
	bufferID.m_iWidth = width;
	bufferID.m_iHeight = height;
	bufferID.m_aColorBuffer = new unsigned char[width*height*VisWindow::BUFFER_COLOR_COMPONENTS];
	for (int i=0; i<width*height*VisWindow::BUFFER_COLOR_COMPONENTS; i++)
		bufferID.m_aColorBuffer[i] = 0;
	
	assert(glCheckFramebufferStatus(GL_FRAMEBUFFER)==GL_FRAMEBUFFER_COMPLETE);
	// ERROR codes: GL_FRAMEBUFFER_UNSUPPORTED, GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT, 
	//      GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER, GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER
	//      GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE, FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT
#else
	std::cout<<"[ERROR] GLUT renderer is disabled"<<std::endl;
	assert(false);
#endif
}

void VisWindowRenderer::LoadBuffer(BufferID & bufferID)
{
#ifdef GLUT_RENDERING
	glBindFramebuffer(GL_FRAMEBUFFER, bufferID.m_iFrameBufferObject);
    glViewport( 0, 0, bufferID.m_iWidth, bufferID.m_iHeight );
#else
	std::cout<<"[ERROR] GLUT renderer is disabled"<<std::endl;
	assert(false);
#endif
}

void VisWindowRenderer::FillDrawBuffer(BufferID & bufferID)
{
#ifdef GLUT_RENDERING
	glReadBuffer(GL_COLOR_ATTACHMENT0);
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    assert(bufferID.m_iWidth >= viewport[2]);
    assert(bufferID.m_iHeight >= viewport[3]);
	
	if (VisWindow::BUFFER_COLOR_COMPONENTS==4)
		glReadPixels(0, 0, bufferID.m_iWidth, bufferID.m_iHeight, 
					 GL_RGBA, GL_UNSIGNED_BYTE, bufferID.m_aColorBuffer);
//	else if (VisWindow::BUFFER_COLOR_COMPONENTS==3)
//		glReadPixels(0, 0, bufferID.m_iWidth, bufferID.m_iHeight, 
//					 GL_RGB, GL_UNSIGNED_BYTE, bufferID.m_aColorBuffer);
	else
		assert(false);	
#else
	std::cout<<"[ERROR] GLUT renderer is disabled"<<std::endl;
	assert(false);
#endif
}

void VisWindowRenderer::DrawTransparentFaces(const std::vector<RenderableFace> & faces, bool lighting) const
{
	std::vector<Sortable> sortedFaces;
	if (faces.size()==0)
		return;
	
	for (int f=0; f<(int)faces.size(); f++)
	{
		double val = 0;
		if (faces[f].m_FaceType==RenderableFace::RENDFACE_Triangle)
			val = vkMax(vkMax(faces[f].vtx[0].Z(),faces[f].vtx[1].Z()), faces[f].vtx[2].Z());
		else if (faces[f].m_FaceType==RenderableFace::RENDFACE_Sphere)
			val = faces[f].vtx[0].Z();
		else if (faces[f].m_FaceType==RenderableFace::RENDFACE_Point)
			val = faces[f].vtx[0].Z();
		else
			assert(false);
		sortedFaces.push_back(Sortable(val, NULL, f));
	}
	std::sort(sortedFaces.begin(), sortedFaces.end());
	
	if (lighting)
	{
		glEnable(GL_LIGHTING);
		glDisable(GL_COLOR);
	}
	else
	{
		glDisable(GL_LIGHTING);
		glEnable(GL_COLOR);
	}
	glEnable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_DEPTH_BUFFER_BIT);
	
	glDisable(GL_CULL_FACE);
	
	for (int sf=0; sf<(int)sortedFaces.size(); sf++)
	{
		int f = sortedFaces[sf].id;
		
		if (faces[f].m_FaceType==RenderableFace::RENDFACE_Triangle)
		{
			glBegin(GL_TRIANGLES);
			if (lighting)
			{
				static GLfloat surfaceMaterial[4];
				surfaceMaterial[0] = faces[f].r;
				surfaceMaterial[1] = faces[f].g;
				surfaceMaterial[2] = faces[f].b;
				surfaceMaterial[3] = faces[f].a;
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial);
				glNormal3d(faces[f].norm.X(), faces[f].norm.Y(), faces[f].norm.Z());
			}
			else
				glColor4d(faces[f].r, faces[f].g, faces[f].b, faces[f].a);

			for (int i=0; i<3; i++)
				glVertex3d(faces[f].vtx[i].X(), faces[f].vtx[i].Y(), faces[f].vtx[i].Z());
			glEnd();
		}
		else if (faces[f].m_FaceType==RenderableFace::RENDFACE_Sphere)
		{
			if (lighting)
			{
				static GLfloat surfaceMaterial[4];
				surfaceMaterial[0] = faces[f].r;
				surfaceMaterial[1] = faces[f].g;
				surfaceMaterial[2] = faces[f].b;
				surfaceMaterial[3] = faces[f].a;
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial);
			}

			R3Sphere(faces[f].vtx[0], faces[f].m_fPointRadius).Draw();
		}
		else if (faces[f].m_FaceType==RenderableFace::RENDFACE_Point)
		{
			glPointSize(faces[f].m_fPointRadius);
			glBegin(GL_POINTS);
			glColor4d(faces[f].r, faces[f].g, faces[f].b, faces[f].a);
			glVertex3d(faces[f].vtx[0].X(),faces[f].vtx[0].Y(),faces[f].vtx[0].Z());
			glEnd();
		}
	}
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_DEPTH_BUFFER_BIT);

}

/********************************************************/
/********************************************************/
/********************************************************/
/****** General Visualization Window (abstract class) ***/
/********************************************************/
/********************************************************/
/********************************************************/

const int VisWindow::BUFFER_COLOR_COMPONENTS = 4;

VisWindow::VisWindow(int width, int height)
{
	m_fStandardAlpha = 1.;
	
	if (height<0)
		m_iHeight = 600;
	else
		m_iHeight = height;
	
	if (width<0)
		m_iWidth = 800;
	else
		m_iWidth = width;

	m_iCurrentFramebufferID = 0;
	m_Buffers.push_back(BufferID());
	m_Buffers[0].m_iFrameBufferObject = 0;
	m_Buffers[0].m_iDepthBuffer = 0;
	m_Buffers[0].m_iColorBuffer = 0;
	m_Buffers[0].m_iTextureBuffer = 0;
	m_Buffers[0].m_aColorBuffer = new unsigned char[4 * m_iWidth * m_iHeight];
	m_Buffers[0].m_iWidth = m_iWidth;
	m_Buffers[0].m_iHeight = m_iHeight;
	

	m_iSpecialKeys = 0;
	m_MouseStates[MOUSE_LEFT]	= MOUSE_BT_UP;
	m_MouseStates[MOUSE_RIGHT]	= MOUSE_BT_UP;
	m_MouseStates[MOUSE_MIDDLE] = MOUSE_BT_UP;
	
	m_MulticolorSequence.push_back(RNRgb(0, 0, 1));
	m_MulticolorSequence.push_back(RNRgb(.6, .6, 1));
	m_MulticolorSequence.push_back(RNRgb(0, 1, 0));
	m_MulticolorSequence.push_back(RNRgb(1, 1, 0));
	m_MulticolorSequence.push_back(RNRgb(1, 0, 0));
		
	m_bGlobalColors = true;;
	m_bDefaultDrawIsocontours = false;
	m_fMinPercentile = 0;
	m_fMaxPercentile = 1;	
	
}

void VisWindow::LoadStandardSceneProjectionCamera(const R3Affine & mview) const
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//gluPerspective( 45,  (double)Width()/(double)Height(), -1., 20.0 ); 
	gluPerspective(50, (double)Width()/(double)Height(), .1, 10 );
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	mview.Load();
}

void VisWindow::SetRenderer(VisWindowRenderer * renderer)
{
	m_pRenderer = renderer;
}

VisWindowRenderer * VisWindow::GetRenderer()
{
	return m_pRenderer;
}

const VisWindowRenderer * VisWindow::GetRenderer() const
{
	return m_pRenderer;
}


void VisWindow::KeyPressed(const char key, int mouseX, int mouseY)
{
	std::cout<<"[WARNING] VisWindow: KeyPressed should be implemented in a child"<<std::endl;
}

void VisWindow::KeyReleased(const char key, int mouseX, int mouseY)
{
	std::cout<<"[WARNING] VisWindow: KeyReleased should be implemented in a child"<<std::endl;
}

void VisWindow::MouseButtonEvent(int x, int y, MouseButton mouseButtonID, 
								 MouseButtonState newState)
{
	UpdateMouseButtons(mouseButtonID, newState);
}

void VisWindow::MouseMotionEvent(int x, int y, int dx, int dy)
{
	// it's ok not to do anything
}

void VisWindow::InitializeGL()
{
	// it's ok to do nothing
}

void VisWindow::Resize(int w, int h)
{
	std::cout<<"[WARNING] VisWindow: Resize should be implemented in a child"<<std::endl;
}

void VisWindow::UpdateMouseButtons(MouseButton mouseButtonID, MouseButtonState newState)
{	
	m_MouseStates[mouseButtonID] = newState;
}

bool VisWindow::IsMouseButtonDown(MouseButton bt)
{
	return m_MouseStates[bt];
}

void VisWindow::SetSpecialKey(int specialKeyFlags)
{
	m_iSpecialKeys = specialKeyFlags;
}

bool VisWindow::Shift()
{
	return (m_iSpecialKeys & SPECIAL_SHIFT) != 0;
}

bool VisWindow::Ctrl()
{
	return (m_iSpecialKeys & SPECIAL_CTRL) != 0;
}

bool VisWindow::Alt()
{
	return (m_iSpecialKeys & SPECIAL_ALT) != 0;
}
	
	
int VisWindow::Width() const
{
	return m_iWidth;
}

int VisWindow::Height() const
{
	return m_iHeight;
}


/**
 * Maps a value to some visible color: blue(small) -> light-blue -> green -> yellow -> red(high)
 * @param value
 * @param minValue
 * @param maxValue		range
 * @param r
 * @param g
 * @param b				output color
 */
void VisWindow::MapScalarToColor(double value, double minValue, double maxValue, 
								 double & r, double & g, double & b) const
{
	double scaledVal = 0;	// scale to [0, 1]
	if (minValue != maxValue)
		scaledVal = (value-minValue) / (maxValue-minValue);
	// linear greyscale
	//	r = scaledVal;
	//	g = scaledVal;
	//	b = scaledVal;	
	
//	// linear blue -> red
//	int nPrev = (int)floor(scaledVal * ((int)m_MulticolorSequence.size()-1));
//	int nNext = (int)ceil(scaledVal * ((int)m_MulticolorSequence.size()-1));
//	double t = 1;
//	if (nNext!=nPrev)
//		t = ((double)nNext - (scaledVal * (double)m_MulticolorSequence.size())) / (nNext-nPrev);
//	
//	assert(nNext < (int)m_MulticolorSequence.size() && nNext>=0);
//	assert(nPrev < (int)m_MulticolorSequence.size() && nPrev>=0);
//	
//	r = m_MulticolorSequence[nPrev].R() * t + m_MulticolorSequence[nNext].R() * (1-t);
//	g = m_MulticolorSequence[nPrev].G() * t + m_MulticolorSequence[nNext].G() * (1-t);
//	b = m_MulticolorSequence[nPrev].B() * t + m_MulticolorSequence[nNext].B() * (1-t);	
	
	double rMax = 0.1, rMin = 0.99;
	double gMax = 0.19, gMin = 0.99;
	double bMax = 0.55,  bMin = 0.99;
	// use Wil's color scheme: white to blue
	r = scaledVal * rMax + (1-scaledVal) * rMin;
	g = scaledVal * gMax + (1-scaledVal) * gMin;
	b = scaledVal * bMax + (1-scaledVal) * bMin;
}

/**
 * maps integer to a color (colors very different from each other, but fewer than MapIntToColor)
 */
void VisWindow::MapIntToRadicalColor(int i, double & r, double & g, double & b) const
{
//	if (i==1)
//		i = 2;
//	else if (i==2)
//		i = 1;
//	else if (i==3)
//		i = 4;
//	else if (i==4)
//		i = 6;
//	else if (i==5)
//		i = 3;
//	else if (i==6)
//		i = 5;	// chair m816 consistent w/ chair m808
	
	
	if (i<0)
	{
		r=0;g=0;b=0;
		return;
	}
	int NMAX = 20;
	i = i%NMAX;
	
	/*switch (i)
	{
		case 0:
			r=64; g=61; b=58;
			break;
		case 1:
			r=81; g=149; b=72;
			break;
		case 2:
			r=136; g=196; b=37;
			break;
		case 3:
			r=6; g=53; b=64;
			break;
		case 4:
			r=242; g=137; b=113;
			break;
		case 5:
			r=31; g=129; b=129;
			break;
		case 6:
			r=242; g=188; b=121;
			break;
		case 7:
			r=191; g=27; b=57;
			break;
		case 8:
			r=115; g=2; b=64;
			break;
		case 9:
			r=23; g=98; b=115;
			break;
	}*/
	
	switch(i)
	{
		case 0:
			r=0; g=1; b=0;
			break;
		case 1:
			r=0; g=0; b=1;
			break;
		case 2:
			r=.25; g=.5; b=0;
			break;
		case 3:
			r=0; g=.8; b=.8;
			break;
		case 4:
			r=.7; g=0; b=.8;
			break;
		case 5:
			r=.6; g=.6; b=0;
			//r=0; g = 0; b = 0;
			break;
		case 6:
			r=.2; g=.3; b=.5;
			break;
		case 7:
			r=0; g=0; b=.5;
			break;
		case 8:
			r=.8; g=.4; b=.4;
			break;
		case 9:
			r=.8; g=0; b=0;
			break;
// below is kind of random			
		case 10:
			r=.24; g=1; b=.5;
			break;
		case 11:
			r=.55; g=.11; b=.26;
			break;
		case 12:
			r=.12; g=.65; b=.85;
			break;
		case 13:
			r=.85; g=.5; b=.25;
			break;			
		case 14:
			r=.45; g=.21; b=.1;
			break;
		case 15:
			r=.0; g=.49; b=.2;
			break;
		case 16:
			r=.30; g=.79; b=.3;
			break;
		case 17:
			r=.50; g=.18; b=.4;
			break;
		case 18:
			r=.73; g=.65; b=.5;
			break;
		case 19:
			r=.70; g=.3; b=0;
			break;
		default:
			r=0.2; g=0.2; b=0.2;
	}
	//std::cout<<"Clr: "<<i<<" -> ["<<r<<" "<<g<<" "<<b<<"]"<<std::endl;
	//r = 1.; g= 1.; b = 1.;
//	r/=255.;
//	g/=255.;
//	b/=255.;
}

void VisWindow::MapIntToRadicalColor(int i, GLfloat material[4]) const
{
	double r, g, b;
	MapIntToRadicalColor(i, r, g, b);
	material[0] = r;
	material[1] = g;
	material[2] = b;
	material[3] = GetAlpha();
}

/**
 * Maps integer to a color
 */
void VisWindow::MapIntToColor(int i, double & r, double & g, double & b) const
{
	i+=2;
	int ri = (31 + 48 * ((3*i) % 5));
	int gi = (31 + 32 * ((5*i) % 7));
	int bi = (31 + 19 * ((7*i) % 11));
	r = ((double)ri) / 256.0;
	g = ((double)gi) / 256.0;
	b = ((double)bi) / 256.0;	
	//std::cout<<i<<" -> ["<<ri<<" "<<gi<<" "<<bi<<"]"<<std::endl;
}

double VisWindow::GetAlpha() const
{
	return m_fStandardAlpha;
}

void VisWindow::SetAlpha(double alpha)
{
	m_fStandardAlpha = alpha;
}

/**
 * Standard surface color - grayish, with some alpha
 */
void VisWindow::StandardSurfaceColor(double & r, double & g, double & b, double & a) const
{
	r = .6;
	g = .6;
	b = .6;
	a = GetAlpha();
}

/**
 * maps integer to a color
 */
void VisWindow::MapIntToColor(int i, GLfloat material[4]) const
{
	double r, g, b;
	MapIntToColor(i, r, g, b);
	material[0] = r;
	material[1] = g;	
	material[2] = b;
	material[3] = 1;	
}

/**
 * Standard surface color - grayish with some alpha
 */
void VisWindow::StandardSurfaceColor(GLfloat material[4]) const
{
	double r, g, b, a;
	StandardSurfaceColor(r, g, b, a);
	material[0] = r;
	material[1] = g;
	material[2] = b;
	material[3] = a;	//GetAlpha();	
}

/**
 * Map scalar to color
 */
void VisWindow::MapScalarToColor(double value, double minValue, double maxValue, GLfloat material[4]) const
{
	double r, g, b, a;
	StandardSurfaceColor(r, g, b, a);
	MapScalarToColor(value, minValue, maxValue, r, g, b);
	material[0] = r;
	material[1] = g;
	material[2] = b;
	material[3] = a;
}

void VisWindow::StartRenderSubwindow(int top, int left, double width, double height)
{
//	VKString windowName;
//	if (top==1)
//		windowName="Top";
//	else if (top==0)
//		windowName="Bottom";
//	else if (top==2)
//		windowName = "";
//	
//	if (left==1)
//		windowName+="Left";
//	else if (left==0)
//		windowName+="Right";
//
//	windowName+="SubWindow";
		
	double transX, transY;
	if (top==1)
		transY = 1 - height;
	else if (top==0)
		transY = 0;
	else if (top==2)
		transY = .5 - height/2.;
	else
		assert(false);
	
	if (left==1)
		transX = 0;
	else if (left==0)
		transX = 1 - width;
	else if (left==2)
		transX = .5 - width/2.;
	else
		assert(false);	

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-0.0, 1.0, -0.0, 1.0, 0., 10.);

	glTranslated(transX, transY, 0);
	glScaled(width, height, 1);

	DrawSquare(0., 0., 1., 1., 0., .2, .5, 1., 1., 1., 1., .9);
	
}

void VisWindow::DrawArrows(double minX, double minY, 
							 double width, double height,
							 double lr, double lg, double lb, double la,
							 double rr, double rg, double rb, double ra,
							 bool fill)
{
	double centerX = minX + .5 * width;
	double centerY = minY + .5 * height;
	
	if (fill)
	{
		glBegin(GL_TRIANGLES);
		glColor4d(lr, lg, lb, la);
		glVertex2d(minX, centerY);
		glVertex2d(centerX, minY);
		glVertex2d(centerX, minY + height);
		
		glColor4d(rr, rg, rb, ra);
		glVertex2d(minX + width, centerY);
		glVertex2d(centerX, minY);
		glVertex2d(centerX, minY + height);
		glEnd();	
	}
	else
	{
		glBegin(GL_LINES);
		glColor4d(lr, lg, lb, la);
		glVertex2d(minX, centerY);
		glVertex2d(centerX, minY);
		glVertex2d(centerX, minY);
		glVertex2d(centerX, minY + height);
		
		glColor4d(rr, rg, rb, ra);
		glVertex2d(minX + width, centerY);
		glVertex2d(centerX, minY);
		glVertex2d(centerX, minY);
		glVertex2d(centerX, minY + height);		

		glColor4d(0, 0, 0, 1.);		
		glVertex2d(centerX, minY);		
		glVertex2d(centerX, minY + height);
		glEnd();
	}
}

void VisWindow::DrawSquare(double minX, double minY, 
						   double width, double height,
						   double r, double g, double b, double a, 
						   double fillR, double fillG, double fillB, double fillA)
{
//	std::cout<<"Draw: ["<<minX<<" "<<minY<<"] : ["<<width<<" "<<height<<"]"<<std::endl;
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glDisable(GL_DEPTH_TEST);
	
	glDisable(GL_LIGHTING);
	glEnable (GL_BLEND); 
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	if (fillA>0)
	{
		glBegin(GL_QUADS);
		glColor4d(fillR, fillG, fillB, fillA);
		glVertex2d(minX, minY);
		glVertex2d(minX+width, minY);
		glVertex2d(minX+width, minY+height);
		glVertex2d(minX, minY+height);
		glEnd();
	}
	
	if (a>0)
	{
		glBegin(GL_LINE_STRIP);
		glColor4d(r, g, b, a);
		glVertex2d(minX, minY);
		glVertex2d(minX+width, minY);
		glVertex2d(minX+width, minY+height);
		glVertex2d(minX, minY+height);
		glVertex2d(minX, minY);	
		glEnd();
	}
	
	glEnable(GL_DEPTH_TEST);
}


bool VisWindow::InSubWindow(int x, int y, double * xWnd, double * yWnd,
							double width, double height, int wndFlags)
{	
	double xRel = (double)x / (double)Width();
	double yRel = (double)y / (double)Height();	
			
	double minX=0, maxX=0, minY=0, maxY=0;
	
	if (wndFlags & WND_LEFT)
	{
		minX = 0.;
		maxX = width;
	}
	else if (wndFlags & WND_VCENTER)
	{
		minX = .5 - width/2.;
		maxX = .5 + width/2.;
	}
	else if (wndFlags & WND_RIGHT)
	{
		minX = 1.-width;
		maxX = 1.;
	}
	
	if (wndFlags & WND_BOTTOM)
	{
		minY = 0.;
		maxY = height;
	}
	else if (wndFlags & WND_HCENTER)
	{
		minY = .5 - height/2.;
		maxY = .5 + height/2.;
	}
	else if (wndFlags & WND_TOP)
	{
		minY = 1.-height;
		maxY = 1.;
	}

	if (xRel >= minX && xRel<=maxX && yRel>=minY && yRel<=maxY)
	{
		*xWnd = (xRel - minX) / (maxX - minX);
		*yWnd = (yRel - minY) / (maxY - minY);
		return true;
	}
	return false;
}

//void VisWindow::ClearAllGlobalValues()
//{
//	m_GlobalValuePool.clear();
//}

//void VisWindow::AddToGlobalValuePool(const VKString & valueName, ModelPerVertexValues * values)
//{
//	m_GlobalValuePool[valueName].push_back(values);
//}

//void VisWindow::NormalizePoolAndCalculateColors(const VKString & valueName,
//												const VKString & interpolate)
//{
//	if (m_GlobalValuePool.find(valueName) == m_GlobalValuePool.end())
//		return;
//	assert(m_GlobalValuePool.find(valueName) != m_GlobalValuePool.end());
//	std::vector<ModelPerVertexValues*> & perVtxValues = m_GlobalValuePool[valueName];
//	
//	std::vector<double> allValues;
//	for (int i=0; i<(int)perVtxValues.size(); i++)
//		perVtxValues[i]->GetAllValues(allValues);
//
//	double minValue = 0;
//	double maxValue = 1;
//	
//	if (allValues.size()>0)
//	{
//		std::sort(allValues.begin(), allValues.end());
//		int minID = (int)floor(allValues.size() * m_fMinPercentile);
//		int maxID = (int)ceil(allValues.size() * m_fMaxPercentile);
//		assert(minID >= 0 && maxID >= 0);
//		minValue = allValues[vkMin(minID, (int)allValues.size()-1)];
//		maxValue = allValues[vkMin(maxID, (int)allValues.size()-1)];
//	}
//	
//	if (interpolate!="none")
//	{
//		for (int i=0; i<(int)perVtxValues.size(); i++)
//			perVtxValues[i]->Interpolate(interpolate);
//	}
//
//	for (int i=0; i<(int)perVtxValues.size(); i++)
//		perVtxValues[i]->InitColorsValueRange(minValue, maxValue, m_bDefaultDrawIsocontours, this);
//}


BufferID VisWindow::CreateBuffer()
{
	BufferID bufferID;
#ifdef VISUALIZATION
	m_pRenderer->CreateBuffer(bufferID);	
#endif
	return bufferID;
}

bool VisWindow::InMainWndBuffer()
{
	return m_iCurrentFramebufferID==0;
}

void VisWindow::LoadCurrentBuffer()
{
#ifdef VISUALIZATION
	assert(m_iCurrentFramebufferID>=0);
	BufferID & bufferID = m_Buffers[m_iCurrentFramebufferID];
	m_pRenderer->LoadBuffer(bufferID);
#endif
}

void VisWindow::StartDrawToBuffer()
{
#ifdef VISUALIZATION
	m_iCurrentFramebufferID++;
	while (m_iCurrentFramebufferID >= (int)m_Buffers.size())
		m_Buffers.push_back(CreateBuffer());
	LoadCurrentBuffer();
#endif
}

BufferID & VisWindow::EndDrawToBuffer()
{
#ifndef VISUALIZATION
	static BufferID bufferID;
#else
	assert(m_iCurrentFramebufferID>=0);
	BufferID & bufferID = m_Buffers[m_iCurrentFramebufferID];
	m_pRenderer->FillDrawBuffer(bufferID);
	m_iCurrentFramebufferID--;
	LoadCurrentBuffer();	
#endif
	return bufferID;
}





